{
        Program Name : AEL
        Package Name : Subroutine Library - Layer 1
        Purpose      : Standard Application Event Logger
        Institution  : Conroy & Conroy Co.
        Date Written : 31-January-2003
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2005 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          These classes provide a default application event logger.
}

unit AEL ;

interface

uses { Layer 0... }
     _DebugIn, { PSDebug_Interface }

     { Layer 1... }
     _AEL ; { TApplication_Event_Logger }

{$I EDEFINES.INC}

{$IFDEF DYNAMIC_OBJECTS}
type TStandard_Application_Event_Logger = class( TApplication_Event_Logger )
                                              public { API... }
                                                  procedure Log_Event( const S : Ansistring ) ;
                                                      override ;
                                                  procedure Trace_Enter( const S : Ansistring ) ;
                                                      override ;
                                                  procedure Trace_Exit( const S : Ansistring ) ;
                                                      override ;
                                          end ;

type TCOM_Standard_Application_Event_Logger = class( TCOM_Application_Event_Logger )
                                                  public { API... }
                                                      procedure Log_Event( S : PAnsiChar ) ;
                                                          override ;
                                                      procedure Trace_Enter( S : PAnsiChar ) ;
                                                          override ;
                                                      procedure Trace_Exit( S : PAnsiChar ) ;
                                                          override ;
                                              end ;
{$ENDIF}

type TSStandard_Application_Event_Logger = object( TSApplication_Event_Logger )
                                               public { Constructors and destructors... }
                                                   constructor Init ;

                                               public { API... }
                                                   function Debugger : PSDebug_Interface ;
                                                       virtual ;
                                                   procedure Log_Event( const S : Ansistring ) ;
                                                       virtual ;
                                                   procedure Trace_Enter( const S : Ansistring ) ;
                                                       virtual ;
                                                   procedure Trace_Exit( const S : Ansistring ) ;
                                                       virtual ;
                                           end ;
     PSStandard_Application_Event_Logger = ^TSStandard_Application_Event_Logger ;

{$IFDEF DYNAMIC_OBJECTS}
function Static_COM_Event_Logger_Wrapper( EL : TCOM_Application_Event_Logger ) : PSApplication_Event_Logger ;
{$ENDIF}

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}

implementation

{$IFDEF DYNAMIC_OBJECTS}
uses Sysutils ; // lowercase
{$ENDIF}

{ TApplication_Event_Logger methods... }

{ Constructors and destructors... }

constructor TSStandard_Application_Event_Logger.Init ;

begin
    { Default does nothing }

    inherited Init ;
end ;


{ API... }

function TSStandard_Application_Event_Logger.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;


procedure TSStandard_Application_Event_Logger.Log_Event( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;


procedure TSStandard_Application_Event_Logger.Trace_Enter( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;


procedure TSStandard_Application_Event_Logger.Trace_Exit( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;


{$IFDEF DYNAMIC_OBJECTS}
{ TStandard_Application_Event_Logger methods... }

procedure TStandard_Application_Event_Logger.Log_Event( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;


procedure TStandard_Application_Event_Logger.Trace_Enter( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;


procedure TStandard_Application_Event_Logger.Trace_Exit( const S : Ansistring ) ;

begin
    { Default does nothing }
end ;



{ TCOM_Standard_Application_Event_Logger methods... }

procedure TCOM_Standard_Application_Event_Logger.Log_Event( S : PAnsiChar ) ; stdcall ;

begin
    { Default does nothing }
end ;


procedure TCOM_Standard_Application_Event_Logger.Trace_Enter( S : PAnsiChar ) ; stdcall ;

begin
    { Default does nothing }
end ;


procedure TCOM_Standard_Application_Event_Logger.Trace_Exit( S : PAnsiChar ) ; stdcall ;

begin
    { Default does nothing }
end ;



type TStatic_COM_Event_Logger_Wrapper = object( TSApplication_Event_Logger )
                                            private
                                                COM : TCOM_Application_Event_Logger ;

                                            public
                                                function Debugger : PSDebug_Interface ;
                                                    virtual ;
                                                procedure Log_Event( const S : Ansistring ) ;
                                                    virtual ;
                                                procedure Trace_Enter( const S : Ansistring ) ;
                                                    virtual ;
                                                procedure Trace_Exit( const S : Ansistring ) ;
                                                    virtual ;
                                        end ;
     PStatic_COM_Event_Logger_Wrapper = ^TStatic_COM_Event_Logger_Wrapper ;

function TStatic_COM_Event_Logger_Wrapper.Debugger : PSDebug_Interface ;

begin
    Result := nil ;
end ;


procedure TStatic_COM_Event_Logger_Wrapper.Log_Event( const S : Ansistring ) ;

begin
    COM.Log_Event( PAnsiChar( S ) ) ;
end ;


procedure TStatic_COM_Event_Logger_Wrapper.Trace_Enter( const S : Ansistring ) ;

begin
    COM.Trace_Enter( PAnsiChar( S ) ) ;
end ;


procedure TStatic_COM_Event_Logger_Wrapper.Trace_Exit( const S : Ansistring ) ;

begin
    COM.Trace_Exit( PAnsiChar( S ) ) ;
end ;


function Static_COM_Event_Logger_Wrapper( EL : TCOM_Application_Event_Logger ) : PSApplication_Event_Logger ;

begin
    Result := new( PStatic_COM_Event_Logger_Wrapper, Init ) ;
    PStatic_COM_Event_Logger_Wrapper( Result )^.COM := EL ;
end ;
{$ENDIF}


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


type Test_COM_Application_Event_Logger = class( TCOM_Application_Event_Logger )
                                             public
                                                 Last_Routine : Ansistring ;

                                                 function Is_Class( Name : PAnsiChar ) : boolean ;
                                                     override ;
                                                 procedure Log_Event( S : PAnsiChar ) ;
                                                     override ;
                                                 procedure Trace_Enter( S : PAnsiChar ) ;
                                                     override ;
                                                 procedure Trace_Exit( S : PAnsiChar ) ;
                                                     override ;
                                         end ;

function Test_COM_Application_Event_Logger.Is_Class( Name : PAnsiChar ) : boolean ;

var S : Ansistring ;

begin
    S := Name ;
    Result := lowercase( S ) = 'test_com_application_even_logger' ;
end ;


procedure Test_COM_Application_Event_Logger.Log_Event( S : PAnsiChar ) ;

begin
    Last_Routine := 'Log_Event' ;
end ;


procedure Test_COM_Application_Event_Logger.Trace_Enter( S : PAnsiChar ) ;

begin
    Last_Routine := 'Trace_Enter' ;
end ;


procedure Test_COM_Application_Event_Logger.Trace_Exit( S : PAnsiChar ) ;

begin
    Last_Routine := 'Trace_Exit' ;
end ;



procedure Test_Wrapper ;

var Test_COM : Test_COM_Application_Event_Logger ;
    Test : PSApplication_Event_Logger ;

begin
    Test_COM := Test_COM_Application_Event_Logger.Create ;
    Test := Static_COM_Event_Logger_Wrapper( Test_COM ) ;

    Test.Log_Event( '' ) ;
    if( Test_COM.Last_Routine <> 'Log_Event' ) then
    begin
        Error ;
    end ;
    Test.Trace_Enter( '' ) ;
    if( Test_COM.Last_Routine <> 'Trace_Enter' ) then
    begin
        Error ;
    end ;
    Test.Trace_Exit( '' ) ;
    if( Test_COM.Last_Routine <> 'Trace_Exit' ) then
    begin
        Error ;
    end ;
    if( Test.Debugger <> nil ) then
    begin
        Error ;
    end ;

    dispose( Test, Done ) ;
    Test_COM.Free ;
end ;


procedure Test_Unit ;

begin
    Test_Wrapper ;
end ;
{$ENDIF}



end.
