{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
        Program Name : CCStandard
        Package Name : Subroutine Library - Layer 3
        Purpose      : Standard Pascal routines
        Institution  : Conroy & Conroy Co.
        Date Written : 28-Nov-88
        Written By   : Alan Conroy
        Version      : 2.1I

        Copyright (c) 1988-2007 by Alan Conroy.  Released to the Public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

         21-Mar-2002   EAC            Made Edit work more like BASIC-Plus.

          1-Nov-2002   EAC            Moved Boolean_To_String to ustrings unit.
                                      Added Format_Time,
                                      Time_To_String, and Parse_Time.

          6-Nov-2002   EAC            Moved time functions to dates unit.

          1-Apr-2005   EAC            Refactoring.

          9-Aug-2006   EAC            Fix insert in TInteger64_List.

         17-Jan-2007   EAC            Added Tag to TInteger64_List.

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit contains standard Pascal support routines for Conroy &
        Conroy Co.
}

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$IFDEF FPC_OR_DELPHI}
{$DEFINE WINDOWS}
{$ENDIF}

unit CCStandard ;

interface

uses
{$IFDEF FPC_OR_DELPHI}
     { VCL... }
     IntList64, { TInteger64_List }
{$ENDIF}
     { C&C Subroutine Library Layer 0... }
     _Types ; { Ansistring }

{ Standard function declarations... }
procedure CC_Log( const X : Ansistring ) ;
function Command_Line : Ansistring ;
function Compare_Memory( var X, Y ; Size : longint ) : boolean ; { Compare memory }
procedure Error ; { Error handler }

function Append_Path( const S1, S2 : string ) : string ;

{ Return position of extension in passed filename.  Returns 0 if no extension. }
function Extension_Pos( const X : Ansistring ) : integer ;

{ Return position of extension in passed filename.  Returns length + 1 if no
  extension. }
function Implicit_Extension_Pos( const X : Ansistring ) : integer ;

function OS_ID : Ansistring ; { Returns name and version of current operating system }
function PID : longint ; { Current process ID }
{$IFNDEF Windows}
function MSDOS_Error : Integer ; { Return DOS extended error code }
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
function MaxAvail : longint ;
{$ENDIF}
function Bits_For_Value( Value : longint ) : integer ;
{$IFDEF FPC_OR_DELPHI}
function XBits_For_Value( Value : int64 ) : integer ;

type TString_Object = class { Simply a means of casting a string to an object }
                          public
                              Text : Ansistring ;

                              constructor Create( const S : Ansistring ) ;
                      end ;

{$ENDIF}

var Exitsave : Pointer ;

{$IFDEF FPC_OR_DELPHI}
type TInteger64_List = IntList64.TInteger64_List ;
{$ENDIF}

implementation

uses { Borland... }
{$IFDEF MSWINDOWS}
     Windows, { GlobalMemoryStatus }
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
     SysUtils,
{$ELSE}
     Strings, { strpas }
{$ENDIF}
     { Subroutine Library - Layer 0... }
     L10n, { Text_* }
     TypeDefs, { tByte_Array }

     { Subroutine Library - Layer 1... }
     _AEL, { TApplication_Event_Logger }
     _Applic, { Application_Manager }
     _CCStandard, { Bit_Values }
     Compatib, { FileExists }
     CommonUt, { Clear_IOResult }
     UE, { ERT }

     { Subroutine Library - Layer 3... }
     Dates, { Time }
     O_S ; { OS }

{$IFDEF FPC_OR_DELPHI}
function MaxAvail : longint ;

{$IFDEF MSWINDOWS}
var Status : TMemoryStatus ;
    Temp : cardinal ;
{$ENDIF}

begin
{$IFDEF MSWINDOWS}
    GlobalMemoryStatus( Status ) ;
    Temp := Status.dwAvailVirtual ;
    if( Temp > $7FFFFFFF ) then
    begin
        Temp := $7FFFFFFF ;
    end ;
    MaxAvail := Temp ;
{$ELSE}
    MaxAvail := 0 ;
{$ENDIF}
end ;


{ TString_Object methods... }

constructor TString_Object.Create( const S : Ansistring ) ;

begin
    inherited Create ;

    Text := S ;
end ;
{$ENDIF}


function Bits_For_Value( Value : longint ) : integer ;

var Loop : integer ;

begin
    if( Value < 0 ) then
    begin
        Value := -Value ;
    end ;

    for Loop := 0 to 31 do
    begin
        if( Value <= Bit_Values[ Loop ] ) then
        begin
            Bits_For_Value := Loop + 1 ;
            exit ;
        end ;
    end ;
    Bits_For_Value := 32 ;
end ;


{$IFDEF FPC_OR_DELPHI}
function XBits_For_Value( Value : int64 ) : integer ;

var Loop : integer ;

begin
    if( Value < 0 ) then
    begin
        Value := -Value ;
    end ;

    for Loop := 0 to 63 do
    begin
        if( Value <= XBit_Values[ Loop ] ) then
        begin
            XBits_For_Value := Loop + 1 ;
            exit ;
        end ;
    end ;
    XBits_For_Value := 32 ;
end ;
{$ENDIF}


{$IFDEF CC_Debug}
{var Allocation_Count : integer ;}


procedure Start_Memory_Test ;

begin
    {TODO}
end ;


procedure End_Memory_Test ;

begin
    {TODO}
end ;
{$ENDIF}


{ Unit API... }

function Command_Line : Ansistring ;

begin
    Command_Line := OS^.Command_Line ;
end ;


{$IFNDEF Windows}
function MSDOS_Error : Integer ;

begin
    MSDOS_Error := OS^.Last_Error.Get_Error ;
end ;
{$ENDIF}


procedure Error ; { Standard error handler procedure }

begin
    Exitproc := Exitsave ; { Restore old vector }
    if( Erroraddr = nil ) then
    begin
        Exit ; { Normal exit }
    end ;
    Erroraddr := nil ; { We are reporting the error }
    Writeln( Ert( Exitcode ) )
end ;


var Have_CCLog : boolean ;
    CCLog : Ansistring ;

procedure CC_Log( const X : Ansistring ) ; { Log a message }

var F : text ;
    S : Ansistring ;

begin
{$IFDEF EXCEPTIONS}
  try
{$ENDIF}
    if( not Have_CCLog ) then
    begin
        Have_CCLog := True ;
        CCLog := OS^.Get_Symbol( 'CCLOG' ) ;
    end ;
    if( length( CCLog ) > 0 ) then { Symbol is defined }
    begin
        assign( F, CCLog ) ;
        if( FileExists( CCLog ) ) then
        begin
            {$I-}
            append( F ) ;
            {$I+}
        end else
        begin
            {$I-}
            rewrite( F ) ;
            {$I+}
        end ;
        if( IOResult <> 0 ) then
        begin                                                                                        
            exit ; { Error opening file }
        end ;

        S := Application_Manager^.Get_Name ;
        if( length( S ) > 0 ) then
        begin
            S := S + ' ' ;
        end ;
        if( length( Application_Manager^.Get_Version ) > 0 ) then
        begin
            S := S + 'V' + Application_Manager^.Get_Version ;
        end ;
        {$I-}
        writeln( F, S, ' on ', Date, ' at ', Time, ': ', X ) ;
        {$I+}
        Clear_IOResult ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
    end ; { if( length( S ) > 0 ) }
{$IFDEF EXCEPTIONS}
  except { Don't let exception here end program }
  end ;
{$ENDIF}
end ; { CC_Log }


function Compare_Memory( var X, Y ; Size : longint ) : boolean ; { Compare memory }

begin
    Compare_Memory := True ; { Assume success }
    while( Size > 0 ) do
    begin
        dec( Size ) ;
        if( tByte_Array( X )[ Size ] <> tByte_Array( Y )[ Size ] ) then
        begin
            Compare_Memory := False ;
            exit ;
        end ;
    end ;
end ;


function Append_Path( const S1, S2 : string ) : string ;

begin
    Result := S1 ;
    if( copy( S1, length( S1 ), 1 ) <> '\' ) then
    begin
        Result := Result + '\' ;
    end ;
    if( copy( S2, 1, 1 ) = '\' ) then
    begin
        Result := Result + copy( S2, 2, length( S2 ) ) ;
    end else
    begin
        Result := Result + S2 ;
    end ;
end ;


function Extension_Pos( const X : Ansistring ) : integer ;
{ Return position of dot in filename extension, 0 if none. }

var Dummy : integer ;

begin
    Dummy := length( X ) ; { Start at end of filename }
    while( ( Dummy > 0 ) and ( X[ Dummy ] <> '.' ) and ( X[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ; { Work our way backwards until found or done }
    end ;
    if( ( Dummy > 0 ) and ( X[ Dummy ] <> '.' ) ) then { Not found }
    begin
        Dummy := 0 ;
    end ;
    Extension_Pos := Dummy ; { Return result }
end ;


function Implicit_Extension_Pos( const X : Ansistring ) : integer ;

var Dummy : integer ;

begin
    Dummy := Extension_Pos( X ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := length( X ) + 1 ;
    end ;
    Implicit_Extension_Pos := Dummy ;
end ;


function OS_ID : Ansistring ; { Returns name and version of current operating system }

begin
    OS_ID := OS^.Name + ' V' + OS^.Version ;
end ;


function PID : longint ; { Current process ID }

begin
    PID := OS^.PID ;
end ;


{ Unit functions }


type TStandard_Application_Event_Logger = object( TSApplication_Event_Logger )
                                              public { Constructors and destructors... }
                                                  constructor Init ;
                                                  destructor Done ; virtual ;

                                              public { API... }
                                                  Old_Event_Logger : PSApplication_Event_Logger ;

                                                  procedure Log_Event( const S : Ansistring ) ;
                                                      virtual ;
                                                  procedure Trace_Enter( const S : Ansistring ) ;
                                                      virtual ;
                                                  procedure Trace_Exit( const S : Ansistring ) ;
                                                      virtual ;
                                          end ;
     PStandard_Application_Event_Logger = ^TStandard_Application_Event_Logger ;

{ TStandard_Application_Event_Logger methods... }

{ Constructors and destructors... }

constructor TStandard_Application_Event_Logger.Init ;

begin
    inherited Init ;

    Old_Event_Logger := nil ;
end ;


destructor TStandard_Application_Event_Logger.Done ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Detach ;
        Old_Event_Logger := nil ;
    end ;

    inherited Done ;
end ;


{ API... }

procedure TStandard_Application_Event_Logger.Log_Event( const S : Ansistring ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Log_Event( S ) ;
    end ;
    CC_Log( S ) ;
end ;


procedure TStandard_Application_Event_Logger.Trace_Enter( const S : Ansistring ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Trace_Enter( S ) ;
    end ;
    CC_Log( 'Entering ' + S ) ;
end ;


procedure TStandard_Application_Event_Logger.Trace_Exit( const S : Ansistring ) ;

begin
    if( Old_Event_Logger <> nil ) then
    begin
        Old_Event_Logger^.Trace_Exit( S ) ;
    end ;
    CC_Log( 'Exiting ' + S ) ;
end ;


var AEL : PStandard_Application_Event_Logger ;

procedure Unit_Initialize ;

begin
    { Log our entry... }
    AEL := nil ;
    if( Application_Manager <> nil ) then
    begin
        AEL := new( PStandard_Application_Event_Logger, Init ) ;
        AEL^.Old_Event_Logger := Application_Manager^.Get_Event_Logger ;
        AEL^.Old_Event_Logger^.Attach ;
        Application_Manager^.Set_Event_Logger( AEL ) ;
    end ;

    Have_CCLog := False ;
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Enter( 'Standard.Unit_Initialize' ) ;
    end ;

    { Initialize data... }
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Exit( 'Standard.Unit_Initialize' ) ;
    end ;
end ; { Unit_Initialize }


procedure Unit_Finalization ;

begin
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Enter( 'Standard.Unit_Finalization' ) ;
    end ;
    if( AEL <> nil ) then
    begin
        AEL^.Trace_Exit( 'Standard.Unit_Finalization' ) ;
    end ;
end ;


{$IFDEF UNIT_INIT_AND_FINAL}
initialization
{$ELSE}
begin
{$ENDIF}
    Unit_Initialize ;

{$IFDEF UNIT_INIT_AND_FINAL}
finalization
    Unit_Finalization ;
{$ENDIF}
end.
