/*
        Program Name : TCommon_COM_Interface
        Package Name : Subroutine Library - Layer 0
        Purpose      : Abstract virtual common COM interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 21-March-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2015 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TCommon_COM_Interface is a base class definition for COM objects used
        in the Application Foundation.
*/

#include "COMInter.hpp"

using namespace std;

// TCommon_COM_Interface methods...

// Constructors and destructors...
    #ifdef CC_Track_Resources
TCommon_COM_Interface::TCommon_COM_Interface()
{
    TBase_COM_Interface() ;

    Add_Resource( RT_Object, (int)this) ;
} ;
    #endif


TCommon_COM_Interface::~TCommon_COM_Interface(void)
{
    Set_Last_Error( NULL ) ;
    #ifdef CC_Track_Resources
    Remove_Resource( RT_Object, integer( self ) ) ;
    #endif
} ;


// API...

tUnified_Exception* TCommon_COM_Interface::Initialize()
{
    Reference_Count = 0 ;
    return NULL;
} ;


tUnified_Exception* TCommon_COM_Interface::Terminate()
{
    Set_Last_Error( NULL ) ;
    free(this) ;
    return NULL;
} ;


void TCommon_COM_Interface::Attach()
{
    Reference_Count++ ;
} ;


bool TCommon_COM_Interface::Detach()
{
    Reference_Count-- ;
    if( Reference_Count < 1 )
    {
        Terminate() ;
        return true;
    } ;
    return false;
} ;


int32 TCommon_COM_Interface::Interface_Version()
{
    return 0 ;
} ;


int32 TCommon_COM_Interface::Facility_ID()
{
    return -1 ;
} ;


int32 TCommon_COM_Interface::Facility_Version()
{
    return 0 ;
} ;


TDebug_Interface* TCommon_COM_Interface::Debugger()
{
    return NULL ;
} ;


tUnified_Exception* TCommon_COM_Interface::Last_Error()
{
    return _Last_Error ;
} ;


void* TCommon_COM_Interface::Extension()
{
    return NULL ;
} ;


// Internal utility routines...

tUnified_Exception* TCommon_COM_Interface::Set_Last_Error( tUnified_Exception* E )
{
    if( E != NULL )
    {
        E->Attach() ;
    } ;
    if( _Last_Error != NULL )
    {
        _Last_Error->Detach() ;
    } ;
    _Last_Error = E ;
    return E ;
} ;



// TCommon_COM_Interface2 methods...

// API...

tUnified_Exception* TCommon_COM_Interface2::Initialize()
{
    Reference_Count = 0 ;
    return NULL;
} ;


tUnified_Exception* TCommon_COM_Interface2::Terminate()
{
    free(this) ;
    return NULL;
} ;


void TCommon_COM_Interface2::Attach()
{
    Reference_Count++ ;
} ;


void TCommon_COM_Interface2::Detach()
{
    Reference_Count-- ;
    if( Reference_Count < 1 )
    {
        Terminate() ;
    } ;
} ;


int32 TCommon_COM_Interface2::Interface_Version()
{
    return 0 ;
} ;


int32 TCommon_COM_Interface2::Facility_ID()
{
    return -1 ;
} ;


int32 TCommon_COM_Interface2::Facility_Version()
{
    return 0 ;
} ;


TDebug_Interface* TCommon_COM_Interface2::Debugger()
{
    return NULL ;
} ;


void* TCommon_COM_Interface2::Extension()
{
    return NULL ;
} ;
