/*
        Program Name : TCommon_COM_Interface
        Package Name : Subroutine Library - Layer 0
        Purpose      : Abstract virtual common COM interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 21-March-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2015 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TCommon_COM_Interface is a base class definition for COM objects used
        in the Application Foundation.
*/

#ifndef COMINTER_HPP
	#define COMINTER_HPP

#include "eDefines.hpp"
#include "_COMInter.hpp"
#include "_DebugIn.hpp"
#include "_UE.hpp"

class TCommon_COM_Interface: public TBase_COM_Interface
{
    public: // API...
        #ifdef CC_Track_Resources
		TCommon_COM_Interface() ;
		#endif
        ~TCommon_COM_Interface() ;

    protected: // Instance data...
        int32 Reference_Count ;
        tUnified_Exception* _Last_Error ;

    public: // API...
        // Initialize the object.  Returns result of initialization attempt.
        tUnified_Exception* __stdcall Initialize() ;

        /* Terminate the use of the object.  Returns
                                       result of termination.  This generally
                                       should not be used - instead, use the
                                       Detach method. */
        tUnified_Exception* __stdcall Terminate() ;

        // Increments the reference count for the object.
        void __stdcall Attach() ;

        /* Decrements the reference count for the
           object.  When count reaches 0, the object is destroyed. */
        bool __stdcall Detach() ;

        // Version of the COM interface for this object.
        int32 __stdcall Interface_Version() ;

        /* Facility ID for the facility represented by this object.  Returns -1 if no
           facility assigned. */
        int32 __stdcall Facility_ID() ;

        /* Facility version.  Result only has meaning if Facility_ID doesn't return
           -1. */
        int32 __stdcall Facility_Version() ;

        // Returns a debugging interface for the object.  Returns NULL if not supported.
        TDebug_Interface* __stdcall Debugger() ;

        //{ Returns last error.
        tUnified_Exception* __stdcall Last_Error() ;

        /* Returns a pointer to an object with extended common COM methods for object.
           Always returns NULL for now. */
        void* __stdcall Extension() ;

    // Internal utility routines...
        /* Note: Don't make any of these virtual, as that could mess up VMT layout.  This class
           is simply to provide implementation of the abstract base class. */
        tUnified_Exception* Set_Last_Error( tUnified_Exception* E ) ;
} ; // TCommon_COM_Interface

class PCommon_COM_Interface: public TCommon_COM_Interface{} ;

class TBase_COM_Interface2
{
    public: // API...
        // Initialize the object.  Returns result of initialization attempt.
        virtual tUnified_Exception* __stdcall Initialize() = 0 ;

        /* Terminate the use of the object.  Returns result of termination.  This generally
           should not be used - instead, use the Detach method. */
        virtual tUnified_Exception* __stdcall Terminate() = 0 ;

        // Increments the reference count for the object.
        virtual void __stdcall Attach() = 0 ;

        /* Decrements the reference count for the object.  When count reaches 0, the object
           is destroyed. */
        virtual void __stdcall Detach() = 0 ;

        // Version of the COM interface for this object.
        virtual int32 Interface_Version() = 0 ;

        /* Facility ID for the facility represented by this object.  Returns -1 if no
           facility assigned. */
        virtual int32 Facility_ID() = 0 ;

        /* Facility version.  Result only has
           meaning if Facility_ID doesn't return -1. */
        virtual int32 __stdcall Facility_Version() = 0 ;

        /* Returns a debugging interface for the 
		   object.  Returns NULL if not supported. */
        virtual TDebug_Interface* __stdcall Debugger() = 0 ;
                                       
        /* Returns a pointer to an object with extended common COM methods for object.
           Always returns NULL for now. */
        virtual void* __stdcall Extension() = 0 ;
} ; // TBase_COM_Interface2

class TCommon_COM_Interface2: public TBase_COM_Interface2
{
    private: // Instance data...
        int Reference_Count ;

    public: // API...
        // Initialize the object.  Returns result of initialization attempt.
        tUnified_Exception* __stdcall Initialize() ;

        /* Terminate the use of the object.  Returns
           result of termination.  This generally
           should not be used - instead, use the Detach method. */
        tUnified_Exception* __stdcall Terminate() ;

        // Increments the reference count for the object.
        void Attach() ;

        /* Decrements the reference count for the
           object.  When count reaches 0, the object is destroyed. */
        void Detach() ;

        // Version of the COM interface for this object.
        int32 Interface_Version() ;

        /* Facility ID for the facility represented
           by this object.  Returns -1 if no facility assigned. */
        int32 __stdcall Facility_ID() ;

        // Facility version.  Result only has meaning if Facility_ID doesn't return -1.
        int32 __stdcall Facility_Version() ;

        // Returns a debugging interface for the object.  Returns NULL if not supported.
        TDebug_Interface* __stdcall Debugger() ;

        /* Returns a pointer to an object with extended common COM methods for object.
           Always returns NULL for now. */
        void* __stdcall Extension() ;
} ; // TCommon_COM_Interface

class PCommon_COM_Interface2: public TCommon_COM_Interface2{} ;

#endif
