{
        Program Name : CommBase
        Package Name : Subroutine Library - Layer 0
        Purpose      : Common base objects (TBase_Object and TBase_Delphi_Object)
        Institution  : Conroy & Conroy Co.
        Date Written : 19-Mar-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit CommBase ;

interface

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$IFNDEF FPC_OR_DELPHI}
  {$UNDEF CC_Balance}
{$ENDIF}

uses { C&C... }
{$IFDEF CC_Balance}
     Classes, { TList }
{$ENDIF}
     _Common, { TObject }
{$IFNDEF MOBILE}
     _DebugIn, { TDebug_Interface }
{$ENDIF}
     _Propert, { TProperties }
     _Streams, { TSStream }
     _UE ; { tUnified_Exception }

{ Serialization/deserialization options... }
const SDO_Visual = 1 ; { Visual aspects }
const SDO_Deep = 2 ; { Include serialization of embedded/reference objects }
{ 4-32768 = reserved for future use }
const SDO_User = 65536 ; { This flag (and following) reserved for the user }

{$IFDEF DYNAMIC_OBJECTS}
type TBase_Delphi_Object = class( TDelphi_Object )
                               private { Instance data... }
                                    _Reference_Count : integer ;
                                    _Properties : TDynamic_Properties ;
{$IFDEF CC_Debug}
                                    Sanity : string[ 19 ] ;
{$ENDIF}
{$IFDEF CC_Balance}
                                    Local_Serial_Number : integer ;
{$ENDIF}
                                    _Last_Error : tUnified_Exception ;

                               public { Constructors and destructors... }
                                    constructor Create ;
                                    destructor Destroy ; override ;

                               public { API... }
                                    procedure Attach ; override ;
                                    procedure Detach ; override ;
                                    function Is_Class( const Name : string ) : boolean ;
                                        override ;

                               protected { Internal utility }
                                    function Set_Last_Error( X : TUnified_Exception ) : TUnified_Exception ;
                                        virtual ;

                               public { API... }
                                    function Reference_Count : integer ;

                                    function Clone : TDelphi_Object ; virtual ;

                                    function Properties : TDynamic_Properties ;
                                        virtual ;
{$IFNDEF MOBILE}
                                    function Debugger : TDebug_Interface ;
                                        virtual ;
{$ENDIF}
                                    procedure Test ; virtual ;

                                    function Deserialize( Stream : TStream ;
                                        Options : longint ) : boolean ;
                                        virtual ;
                                    function Serialize( Stream : TStream ;
                                        Options : longint ) : boolean ;
                                        virtual ;
                                    function Load( Stream : TStream ) : boolean ;
                                        virtual ;
                                    function Persist( Stream : TStream ) : boolean ;
                                        virtual ;
                                    function Last_Error : TUnified_Exception ;
                                        virtual ;
                           end ; { TBase_Delphi_Object }
{$ENDIF}

{$IFDEF CC_Balance}
var Objects : TList = nil ;
    Delphi_Objects : TList = nil ;
    Object_Serial_Number : integer = 0 ;
{$ENDIF}

type PBase_Object = ^TBase_Object ;
     TBase_Object = object( TObject )
                        private { Instance data... }
                             _Memory_Type : TMemory_Type ;
                             _Reference_Count : longint ;
                             _Properties : PSDynamic_Properties ;
{$IFDEF CC_Debug}
                             Base_Object_Sanity : string[ 12 ] ;
{$ENDIF}
{$IFDEF CC_Balance}
                             Local_Serial_Number : integer ;
{$ENDIF}
                        protected
                             _Last_Error : PSUnified_Exception ;

                        public { Constructors and destructors... }
                             constructor Init ;
                             destructor Done ; virtual ;
                             procedure Free ;

                        protected { Internal... }
                            function Set_Last_Error( X : PSUnified_Exception ) : PSUnified_Exception ;
                                virtual ;

                        public { API... }
                             procedure Attach ; virtual ;
                             procedure Detach ; virtual ;
                             function Classname : string ; virtual ;
                             function Is_Class( const Name : string ) : boolean ;
                                 virtual ;
                             procedure Set_Memory_Type( MT : TMemory_Type ) ;
                             function Get_Memory_Type : TMemory_Type ; virtual ;
                             function Reference_Count : integer ;

                             function Clone : PBase_Object ; virtual ;
                             function Properties : PSDynamic_Properties ;
                                 virtual ;
{$IFNDEF MOBILE}
                             function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}
                             procedure Test ; virtual ;

                             function Deserialize( Stream : PSStream ; Options : longint ) : boolean ;
                                 virtual ;
                             function Serialize( Stream : PSStream ; Options : longint ) : boolean ;
                                 virtual ;
                             function Load( Stream : PSStream ) : boolean ;
                                 virtual ;
                             function Persist( Stream : PSStream ) : boolean ;
                                 virtual ;
                             function Last_Error : PSUnified_Exception ;
                                 virtual ;
                    end ; { TBase_Object }

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Subroutine Library... }
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}
     _ObjectF, { Object_Factory }
     L10n ; { Text_* }

{$IFDEF DYNAMIC_OBJECTS}
{ TBase_Delphi_Object methods... }

{ Constructors and destructors... }

constructor TBase_Delphi_Object.Create ;

begin
    inherited Create ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}

  {$IFDEF CC_Debug}
    Sanity := 'TBase_Delphi_Object' ;
  {$ENDIF}
{$IFDEF CC_Balance}
    Local_Serial_Number := Object_Serial_Number ;
    inc( Object_Serial_Number ) ;
    Delphi_Objects.Add( Self ) ;
{$ENDIF}
end ;


destructor TBase_Delphi_Object.Destroy ;

begin
  {$IFDEF CC_Debug}
    Sanity := '                   ' ;
  {$ENDIF}
    if( _Properties <> nil ) then
    begin
        _Properties.Detach ;
    end ;
{$IFDEF CC_Balance}
    Delphi_Objects.Remove( Self ) ;
{$ENDIF}
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}

    inherited Destroy ;
end ;


{ API... }

procedure TBase_Delphi_Object.Attach ; stdcall ;

begin
    inc( _Reference_Count ) ;
end ;


procedure TBase_Delphi_Object.Detach ; stdcall ;

begin
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
        Free ;
    end ;
end ;


function TBase_Delphi_Object.Reference_Count : integer ;

begin
    Reference_Count := _Reference_Count ;
end ;


function TBase_Delphi_Object.Is_Class( const Name : string ) : boolean ; stdcall ;

begin
    try
        Is_Class := ( Name = Classname ) ;
    except
        Is_Class := False ;
    end ;
end ;


function TBase_Delphi_Object.Clone : TDelphi_Object ;

begin
    Clone := nil ;
end ;


function TBase_Delphi_Object.Properties : TDynamic_Properties ;

begin
    if( _Properties = nil ) then
    begin
        _Properties := Object_Factory^.Dynamic_Properties ;
    end ;
    Properties := _Properties ;
end ;


procedure TBase_Delphi_Object.Test ;

var I : integer ;

begin
    I := Reference_Count ;
    Attach ;
    if( I + 1 <> Reference_Count ) then
    begin
        writeln( Text_Error_Attach_Failed ) ;
    end ;
    if( I = 0 ) then { No previous attachments }
    begin
        Attach ;
        inc( I ) ;
    end ;
    Detach ;
    if( I <> Reference_Count ) then
    begin
        writeln( Text_Error_Detach_Failed ) ;
    end ;
end ;


{$IFNDEF MOBILE}
function TBase_Delphi_Object.Debugger : TDebug_Interface ;

begin
    Debugger := nil ;
end ;
{$ENDIF}


function TBase_Delphi_Object.Deserialize( Stream : TStream ; Options : longint ) : boolean ;

begin
    Deserialize := False ;
end ;


function TBase_Delphi_Object.Persist( Stream : TStream ) : boolean ;

begin
    Persist := False ;
end ;


function TBase_Delphi_Object.Serialize( Stream : TStream ; Options : longint ) : boolean ;

begin
    Serialize := False ;
end ;


function TBase_Delphi_Object.Load( Stream : TStream ) : boolean ;

begin
    Load := False ;
end ;


function TBase_Delphi_Object.Set_Last_Error( X : TUnified_Exception ) : TUnified_Exception ;

begin
    Result := X ;
    if( X = _Last_Error ) then
    begin
        exit ;
    end ;
    if( _Last_Error <> nil ) then
    begin
        _Last_Error.Detach ;
    end ;
    _Last_Error := X ;
    if( _Last_Error <> nil ) then
    begin
        _Last_Error.Attach ;
    end ;
end ;


function TBase_Delphi_Object.Last_Error : TUnified_Exception ;

begin
    Result := _Last_Error ;
end ;
{$ENDIF}



{ TObject methods... }

constructor TBase_Object.Init ;

begin
    inherited Init ;
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Object, integer( @self ) ) ;
{$ENDIF}

    _Reference_Count := 0 ;
    _Memory_Type := MT_Unknown ;
    _Properties := nil ;
{$IFDEF CC_Debug}
    Base_Object_Sanity := 'TBase_Object' ;
{$ENDIF}
{$IFDEF CC_Balance}
    Local_Serial_Number := Object_Serial_Number ;
    inc( Object_Serial_Number ) ;
    Objects.Add( @Self ) ;
{$ENDIF}
end ;


destructor TBase_Object.Done ;

begin
{$IFDEF CC_Debug}
    Base_Object_Sanity := '            ' ;
{$ENDIF}
    if( _Properties <> nil ) then
    begin
        _Properties^.Detach ;
    end ;
{$IFDEF CC_Balance}
    Objects.Remove( @Self ) ;
{$ENDIF}
{$IFDEF CC_Track_Resources}
    Add_Resource( RT_Object, integer( @self ) ) ;
{$ENDIF}

    inherited Done ;
end ;


procedure TBase_Object.Free ;

begin
    if( @self <> nil ) then
    begin
{$IFDEF FPC}
        if( _Memory_Type = MT_Heap ) then
        begin
            dispose( PBase_Object( @self ), Done ) ;
        end else
        begin
            Done ;
        end ;
{$ELSE}
        Done ;
  {$IFDEF DYNAMIC_OBJECTS}
        if( _Memory_Type = MT_Heap ) then
        begin
            dispose( PBase_Object( @self ) ) ;
        end ;
  {$ENDIF}
{$ENDIF}
    end ;
end ;


procedure TBase_Object.Attach ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    inc( _Reference_Count ) ;
end ;


procedure TBase_Object.Detach ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
{$IFDEF CC_Debug}
        if( _Reference_Count < 0 ) then
        begin
            writeln( 'Overly detached object' ) ;
        end ;
{$ENDIF}
        Free ;
    end ;
end ;


function TBase_Object.Reference_Count : integer ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Reference_Count := _Reference_Count ;
end ;


function TBase_Object.Clone : PBase_Object ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Clone := nil ;
end ;


function TBase_Object.Properties : PSDynamic_Properties ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    if( _Properties = nil ) then
    begin
        _Properties := Object_Factory^.Static_Dynamic_Properties ;
    end ;
    Properties := _Properties ;
end ;


function TBase_Object.Classname : string ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Classname := 'TBase_Object' ;
end ;


function TBase_Object.Is_Class( const Name : string ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    if( Name = Classname ) then
    begin
        Is_Class := True ;
    end else
    begin
        Is_Class := inherited Is_Class( Name ) ;
    end ;
end ;


procedure TBase_Object.Set_Memory_Type( MT : TMemory_Type ) ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    _Memory_Type := MT ;
end ;


function TBase_Object.Get_Memory_Type : TMemory_Type ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Get_Memory_Type := _Memory_Type ;
end ;


function TBase_Object.Deserialize( Stream : PSStream ; Options : longint ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Deserialize := False ;
end ;


function TBase_Object.Persist( Stream : PSStream ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Persist := False ;
end ;


function TBase_Object.Serialize( Stream : PSStream ; Options : longint ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Serialize := False ;
end ;


function TBase_Object.Load( Stream : PSStream ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Load := False ;
end ;


procedure TBase_Object.Test ;

var I : integer ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    I := Reference_Count ;
    Attach ;
    if( I + 1 <> Reference_Count ) then
    begin
        writeln( Text_Error_Attach_Failed ) ;
    end ;
    if( I = 0 ) then { No previous attachments }
    begin
        Attach ;
        inc( I ) ;
    end ;
    Detach ;
    if( I <> Reference_Count ) then
    begin
        writeln( Text_Error_Detach_Failed ) ;
    end ;
end ;


{$IFNDEF MOBILE}
function TBase_Object.Debugger : PSDebug_Interface ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Debugger := nil ;
end ;
{$ENDIF}


function TBase_Object.Set_Last_Error( X : PSUnified_Exception ) : PSUnified_Exception ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Set_Last_Error := X ;
end ;


function TBase_Object.Last_Error : PSUnified_Exception ;

begin
{$IFDEF CC_Debug}
    if( Base_Object_Sanity <> 'TBase_Object' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Last_Error := _Last_Error ;
end ;


{$IFDEF CC_Regression_Test}
procedure Test_Unit ;

var TBDO : TBase_Delphi_Object ;
    TBO : TBase_Object ;

begin
  {$IFDEF DYNAMIC_OBJECTS}
    TBDO := TBase_Delphi_Object.Create ;
    TBDO.Test ;
    TBDO.Free ;
  {$ENDIF}

    TBO.Init ;
    TBO.Test ;
    TBO.Done ;
end ;
{$ENDIF}



{$IFDEF CC_Balance}
procedure Dump_Balances ;

var F : textfile ;
    Loop : integer ;

begin
    assignfile( F, 'c:\balance.txt' ) ;
    {$I-}
    append( F ) ;
    {$I+}
    if( IOResult <> 0 ) then
    begin
        rewrite( F ) ;
    end ;
    for Loop := 0 to Objects.Count - 1 do
    begin
        writeln( F, PBase_Object( Objects[ Loop ] )^.Classname, ' ', PBase_Object( Objects[ Loop ] )^.Local_Serial_Number ) ;
    end ;
    for Loop := 0 to Delphi_Objects.Count - 1 do
    begin
        writeln( F, 'Delphi_Object->', PBase_Object( Delphi_Objects[ Loop ] )^.Local_Serial_Number ) ;
    end ;
    closefile( F ) ;
end ;


initialization
    Objects := TList.Create ;
    Delphi_Objects := TList.Create ;

finalization
    Dump_Balances ;
    Objects.Free ;
    Objects := nil ;
    Delphi_Objects.Free ;
    Delphi_Objects := nil ;
{$ENDIF}
end.

