unit DX_Player ;

interface

uses // Windows...
     DirectShow,
     Messages,
     MPlayer ; // TMPModes

type TFilterArray = array[ 0..65535 ] of IBaseFilter ;
type TDX_Player = class
                      public // Constructors and destructors...
                          constructor Create( Handle : THandle ) ; virtual ;

                      private // Instance data...
                          GraphBuilder : IGraphBuilder ;
                          MediaEventEx : IMediaEventEx ;
                          MediaSeeking : IMediaSeeking ;
                          MediaControl : IMediaControl ;
                          Source : IBaseFilter ;
                          _Mode : TMPModes ;

                      protected // Internal methods...
                          function Load_File( Filename : Widestring ) : IBaseFilter ;

                      protected // Property handlers...
                          procedure Set_Filename( File_Name : string ) ; virtual ;
                          function Get_Position : Cardinal ;
                          procedure Set_Position( Value : Cardinal ) ;
                          procedure Set_TimeFormat( Value : TGUID ) ;
                          function Get_TimeFormat : TGUID ;
                          function Get_Length : Cardinal ;

                      public // API...
                          procedure Pause ;
                          procedure Stop ;
                          procedure Close ;
                          procedure Play ;
                          function Mode : TMPModes ;
                          function Error : integer ;

                          property Filename : string
                              write Set_Filename ;
                          property Position : Cardinal
                              read Get_Position
                              write Set_Position ;
                          property TimeFormat : TGUID
                              read Get_TimeFormat
                              write Set_TimeFormat ;
                          property Length : Cardinal
                              read Get_Length ;
                  end ;

implementation

uses // Windows...
     ActiveX, // CoCreateInstance
     Windows, // LPCWSTR

     // Borland...
     SysUtils ; // Allocmem

constructor TDX_Player.Create( Handle : THandle ) ;

begin
    inherited Create ;

    // Create an IGraphBuilder object...
    CoCreateInstance( CLSID_FilterGraph, nil, CLSCTX_INPROC, IID_IGraphBuilder,
        GraphBuilder ) ;
    // Get the IMediaControl Interface
    GraphBuilder.QueryInterface( IID_IMediaControl, MediaControl ) ;
    // Get the IMediaSeeking Interface
    GraphBuilder.QueryInterface( IID_IMediaSeeking, MediaSeeking ) ;
    // Get the IMediaEventEx Interface
    GraphBuilder.QueryInterface( IID_IMediaEventEx, MediaEventEx ) ;

    // Tell DirectShow to send any events to the main window
    MediaEventEx.SetNotifyWindow( Handle, WM_APP, 0 ) ;
end ;


function TDX_Player.Load_File( Filename : WideString ) : IBaseFilter ;

var h : HRESULT ;
    F : LPCWSTR ;

begin
    F := pwidechar( Filename ) ;
    h := GraphBuilder.AddSourceFilter( F, 'Default', Result ) ;
    if( h < 0 ) then
    begin
        Result := nil ;
    end ;
end ;



procedure TDX_Player.Set_Filename( File_name : string ) ;

var Pin : IPin ;
    FilterEnum : IEnumFilters ;
    Filter_Count, P : integer ;
    Filters : ^TFilterArray ;
    Filename : Widestring ;

begin
    Filename := widestring( File_Name ) ;
    if( Source = nil ) then
    begin
        Source := Load_File( Filename ) ;
    end ;
    Source.FindPin( Pwidechar( widestring( 'Output' ) ), Pin ) ;

    // Stop the graph
    MediaControl.Stop ;

    // Break any filter connections. We do this by removing and then adding each of the filters in the graph.
    GraphBuilder.EnumFilters( FilterEnum ) ;

    // Need to know how many filters. If we add/remove filters during the
    // enumeration we'll invalidate the enumerator
    Filter_Count := 0 ;
    while( FilterEnum.Skip( 1 ) = S_OK ) do
    begin
        inc( Filter_Count ) ;
    end ;

    // Allocate space, then pull out all of the
    Filters := allocmem( sizeof( IBaseFilter ) * Filter_Count ) ;

    FilterEnum.Reset ;

    P := 0 ;
    while( ( P < Filter_Count ) and ( FilterEnum.Next( 1, Filters[ P ], nil ) = S_OK ) ) do
    begin
        inc( P ) ;
    end ;

    for P := 0 to Filter_Count - 1 do
    begin
        GraphBuilder.RemoveFilter( Filters[ P ] ) ;

        // Put the filter back
        GraphBuilder.AddFilter( Filters[ P ], nil ) ;
    end ;
    freemem( Filters ) ;

    // Make the new filter connection...
    GraphBuilder.Render( Pin ) ;
    Pin._Release ;

    // Seek back to the start...
    Position := 0 ;
end ;


procedure TDX_Player.Pause ;

begin
    MediaControl.Pause ;
    _Mode := mpPaused ;
end ;


procedure TDX_Player.Close ;

begin
    Stop ;
end ;


procedure TDX_Player.Stop ;

begin
    MediaControl.Pause ;
    _Mode := mpStopped ;
end ;


procedure TDX_Player.Play ;

begin
    MediaControl.Run ;
    _Mode := mpPlaying ;
end ;


function TDX_Player.Get_Position : Cardinal ;

var i : int64 ;

begin
    MediaSeeking.GetCurrentPosition( i ) ;
    Result := i div 10000000 ;
end ;


procedure TDX_Player.Set_Position( Value : Cardinal ) ;

var i : int64 ;

begin
    i := Value ;
    i := i * 10000000 ;
    MediaSeeking.SetPositions( i, AM_SEEKING_AbsolutePositioning,
        i, AM_SEEKING_NoPositioning ) ;
end ;


function TDX_Player.Get_Length : Cardinal ;

var i : int64 ;

begin
    MediaSeeking.GetDuration( i ) ;
    Result := i div 10000000 ;
end ;


procedure TDX_Player.Set_TimeFormat( Value : TGUID ) ;

begin
    MediaSeeking.SetTimeFormat( Value ) ;
end ;


function TDX_Player.Get_TimeFormat : TGUID ;

begin
    MediaSeeking.GetTimeFormat( Result ) ;
end ;


function TDX_Player.Mode : TMPModes ;

begin
    if( Position >= Length ) then
    begin
        Stop ;
    end ;
	Result := _Mode ;
end ;


function TDX_Player.Error : integer ;

begin
    Result := 0 ; // No errors at this time
end ;


end.
