{
        Program Name : DataDis
        Package Name : Subroutine Library - Layer 4
        Purpose      : Data display window
        Institution  : Conroy & Conroy Co.
        Date Written : 27-October-2009
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The TData_Display is a control used to display data from an array
        interface in various formats.
}

unit DataDis ;

interface

uses // Borland...
     Classes, // TComponent
     ExtCtrls, // TPanel
     Menus, // TPopupMenu
     Messages, // WM_Paint
     StdCtrls, // TScrollbar

     // C&C Subroutine Library - Layer 1...
     ArrayInt, // TArray_Interface

     // C&C Subroutine Library - Layer 4...
     ArrayDis ; // TArray_Display

type TData_Display = class( TPanel )
                         public // Constructors and destructors...
                             constructor Create( AOwner : TComponent ) ;
                                 override ;
                             destructor Destroy ; override ;

                         private // Private data...
                             AD : TArray_Display ;
                             __Array : TCOM_Array_Interface ;
                             EBCDIC1 : TMenuItem ;
                             Signed1 : TMenuItem ;
                             Binary1 : TMenuItem ;
                             Octal1 : TMenuItem ;
                             Decimal1 : TMenuItem ;
                             Hexadecimal1 : TMenuItem ;
                             Radix50 : TMenuItem ;
                             Word1 : TMenuItem ;
                             Long1 : TMenuItem ;
                             Double1 : TMenuItem ;
                             Byte1 : TMenuItem ;
                             Other1 : TMenuItem ;
                             Scrollbar1 : TScrollbar ;

                         protected // Callbacks...
                             procedure Binary1Click(Sender: TObject);
                             procedure Octal1Click(Sender: TObject);
                             procedure Decimal1Click(Sender: TObject);
                             procedure Hexadecimal1Click(Sender: TObject);
                             procedure Other1Click(Sender: TObject);
                             procedure Byte1Click(Sender: TObject);
                             procedure Word1Click(Sender: TObject);
                             procedure Long1Click(Sender: TObject);
                             procedure Double1Click(Sender: TObject);
                             procedure Gotoaddress1Click(Sender: TObject);
                             procedure Find1Click(Sender: TObject);
                             procedure Radix50Click(Sender: TObject);
                             procedure EBCDIC1Click( Sender : TObject ) ;
                             procedure Signed1Click( Sender : TObject ) ;
                             procedure Modify1Click( Sender : TObject ) ;

                         protected // Property handlers...
                            procedure ScrollBarChange( Sender : TObject ) ;

                             procedure Set_Array( Value : TCOM_Array_Interface ) ;

                         public // API...
                             Display_Panel : TPanel ;

                             procedure wmPaint( var Message ) ;
                                 message WM_Paint ;

                             procedure Set_Number_Size( Size : integer ) ;

                             property _Array : TCOM_Array_Interface
                                 read __Array
                                 write Set_Array ;
                     end ;

implementation

uses // Borland...
     Windows, // TRect
     Controls, // mrOK
     Dialogs, // ShowMessage
     Forms, // sbVertical
     SysUtils, // uppercase

     // C&C Subroutine Library - Layer 0...
     L10n, // Text_*

     // C&C Subroutine Library - Layer 1...
     CVT, // CVTB
     EBCDICs, // From_EBCDIC

     // C&C Subroutine Library - Layer 4...
     Find_Data_Dialog, // Find_Form
     Goto_Address, // TGoto_Address_Form
     ModMem, // Modify_Memory_Dialog
     Radix_Dialog ; // Radix_Form

type TPaint_Panel = class( TPanel )
                        public // API...
                            AD : TArray_Display ;

                            procedure wmPaint( var Message ) ;
                                message WM_Paint ;
                    end ;

procedure TPaint_Panel.wmPaint( var Message ) ;

begin
    if( AD <> nil ) then
    begin
        AD.Display ;
    end ;
    ValidateRect( Handle, nil ) ;
end ;



// TData_Display methods...

// Constructors and destructors...

constructor TData_Display.Create( AOwner : TComponent ) ;

    function Create_Menu( C : string ; O : TNotifyEvent ) : TMenuItem ;

    begin
        Result := TMenuItem.Create( AOwner ) ;
        Result.Caption := C ;
        Result.OnClick := O ;
    end ;

var M : TMenuItem ;

begin
    inherited Create( AOwner ) ;

    Scrollbar1 := TScrollbar.Create( AOwner ) ;
    Scrollbar1.Parent := self ;
    Scrollbar1.Align := alRight ;
    Scrollbar1.Kind := sbVertical ;
    Scrollbar1.TabStop := False ;
    Scrollbar1.OnChange := ScrollBarChange ;
    Display_Panel := TPaint_Panel.Create( AOwner ) ;
    Display_Panel.Parent := self ;
    Display_Panel.Align := alClient ;
    TPaint_Panel( Display_Panel ).AD := AD ;

    Display_Panel.PopupMenu := TPopupMenu.Create( AOwner ) ;
    M := Create_Menu( Text_Radix_50_Amp, nil ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;
        Binary1 := Create_Menu( Text_Binary_Amp, Binary1Click ) ;
        Binary1.GroupIndex := 2 ;
        Binary1.RadioItem := True ;
        M.Add( Binary1 ) ;

        Octal1 := Create_Menu( Text_Octal_Amp, Octal1Click ) ;
        Octal1.GroupIndex := 2 ;
        Octal1.RadioItem := True ;
        M.Add( Octal1 ) ;

        Decimal1 := Create_Menu( Text_Decimal_Amp, Decimal1Click ) ;
        Decimal1.GroupIndex := 2 ;
        Decimal1.RadioItem := True ;
        Decimal1.Checked := True ;
        M.Add( Decimal1 ) ;

        Hexadecimal1 := Create_Menu( Text_Hexadecimal, Hexadecimal1Click ) ;
        Hexadecimal1.GroupIndex := 2 ;
        Hexadecimal1.RadioItem := True ;
        M.Add( Hexadecimal1 ) ;

        Radix50 := Create_Menu( Text_Radix_50, Radix50Click ) ;
        Radix50.GroupIndex := 2 ;
        Radix50.RadioItem := True ;
        M.Add( Radix50 ) ;

        Other1 := Create_Menu( Text_Other_Ellipse, Other1Click ) ;
        Other1.GroupIndex := 2 ;
        Other1.RadioItem := True ;
        M.Add( Other1 ) ;

    Signed1 := Create_Menu( Text_Signed_Amp, Signed1Click ) ;
    Display_Panel.PopupMenu.Items.Add( Signed1 ) ;

    M := Create_Menu( Text_Size_Amp, nil ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;
        Byte1 := Create_Menu( Text_Byte_Amp, Byte1Click ) ;
        Byte1.RadioItem := True ;
        Byte1.Checked := True ;
        Byte1.GroupIndex := 1 ;
        M.Add( Byte1 ) ;

        Word1 := Create_Menu( Text_Word_Amp, Word1Click ) ;
        Word1.GroupIndex := 1 ;
        Word1.RadioItem := True ;
        M.Add( Word1 ) ;

        Long1 := Create_Menu( Text_Long_Amp, Long1Click ) ;
        Long1.GroupIndex := 1 ;
        Long1.RadioItem := True ;
        M.Add( Long1 ) ;

        Double1 := Create_Menu( Text_Quad_Amp, Double1Click ) ;
        Double1.GroupIndex := 1 ;
        Double1.RadioItem := True ;
        Double1.Visible := False ;
        M.Add( Double1 ) ;

    EBCDIC1 := Create_Menu( Text_EBCDIC_Amp, EBCDIC1Click ) ;
    EBCDIC1.Checked := True ;
    Display_Panel.PopupMenu.Items.Add( EBCDIC1 ) ;

    M := Create_Menu( '-', nil ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;

    M := Create_Menu( Text_Goto_Address_Ellipse, Gotoaddress1Click ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;

    M := Create_Menu( Text_Find_Ellipse, Find1Click ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;

    M := Create_Menu( Text_Modify_Ellipse, Modify1Click ) ;
    Display_Panel.PopupMenu.Items.Add( M ) ;
end ; // TData_Display.Create


destructor TData_Display.Destroy ;

begin
    AD.Free ;

    inherited Destroy ;
end ;


// Callbacks...

procedure TData_Display.Modify1Click( Sender : TObject ) ;

var Address : int64 ;
    Base : integer ;
    Data, Work : string ;
    Dummy, Loop : integer ;
    V : int64 ;

begin
    Modify_Memory_Dialog.Data.Text :=
        CVTB( 10, AD.Base, inttostr( _Array.Get_Byte( Modify_Memory_Dialog.Address_Value ) ) ) ;
    Modify_Memory_Dialog.Base := AD.Base ;
    Modify_Memory_Dialog.OK_Button.Enabled := True ;
    Modify_Memory_Dialog.Size := AD.Number_Size ;
    if( Modify_Memory_Dialog.ShowModal = mrOK ) then
    begin
        Address := Modify_Memory_Dialog.Address_Value ;
        Data := Modify_Memory_Dialog.Data.Text ;
        if( Modify_Memory_Dialog.Literal_Text.Checked ) then
        begin
            for Dummy := 1 to length( Data ) do
            begin
                _Array.Set_Byte( Address + Dummy - 1, ord( Data[ Dummy ] ) ) ;
            end ;
        end else
        begin
            while( length( Data ) > 0 ) do
            begin
                Dummy := pos( ' ', Data + ' ' ) ;
                Work := copy( Data, 1, Dummy - 1 ) ;
                Data := copy( Data, Dummy + 1, length( Work ) ) ;
                Base := AD.Base ;
                Translate_Value( Work, 0, Base, V ) ;
                for Loop := 1 to Modify_Memory_Dialog.Size do
                begin
                    _Array.Set_Byte( Address, V and $FF ) ;
                    V := V shr 8 ;
                    inc( Address ) ;
                end ;
            end ;
        end ;
        Display_Panel.Repaint ;
    end ;
end ; // TData_Display.Modify1Click


procedure TData_Display.EBCDIC1Click( Sender : TObject ) ;

begin
    if( EBCDIC1.Caption = Text_ASCII_Amp ) then
    begin
        EBCDIC1.Caption := Text_EBCDIC_Amp ;
        AD.Ascii_Ebcdic := 0 ;
    end else
    begin
        EBCDIC1.Caption := Text_ASCII_Amp ;
        AD.Ascii_Ebcdic := 1 ;
    end ;
    Display_Panel.Repaint ;
end ;


procedure TData_Display.Signed1Click( Sender : TObject ) ;

begin
    Signed1.Checked := not Signed1.Checked ;
    AD.Unsigned := not Signed1.Checked ;
    Display_Panel.Repaint ;
end ;


procedure TData_Display.Binary1Click( Sender : TObject ) ;

begin
    AD.Base := 2 ;
    Display_Panel.Repaint ;
    Binary1.Checked := True ;
end ;


procedure TData_Display.Octal1Click( Sender : TObject ) ;

begin
    AD.Base := 8 ;
    Display_Panel.Repaint ;
    Octal1.Checked := True ;
end ;


procedure TData_Display.Decimal1Click( Sender : TObject ) ;

begin
    AD.Base := 10 ;
    Display_Panel.Repaint ;
    Decimal1.Checked := True ;
end ;


procedure TData_Display.Hexadecimal1Click( Sender : TObject ) ;

begin
    AD.Base := 16 ;
    Display_Panel.Repaint ;
    Hexadecimal1.Checked := True ;
end ;


procedure TData_Display.Other1Click( Sender : TObject ) ;

begin
    Radix_Form.Spin_Edit.Value := AD.Base ;
    if( Radix_Form.ShowModal = mrOK ) then
    begin
        AD.Base := Radix_Form.Spin_Edit.Value ;
        Display_Panel.Repaint ;
        case AD.Base of
            2 : Binary1.Checked := True ;
            8 : Octal1.Checked := True ;
            10 : Decimal1.Checked := True ;
            16 : Hexadecimal1.Checked := True ;
            else Other1.Checked := True ;
        end ;
    end ;
end ;


procedure TData_Display.Byte1Click( Sender : TObject ) ;

begin
    Set_Number_Size( 1 ) ;
end ;


procedure TData_Display.Word1Click( Sender : TObject ) ;

begin
    Set_Number_Size( 2 ) ;
end ;


procedure TData_Display.Long1Click( Sender : TObject ) ;

begin
    Set_Number_Size( 4 ) ;
end ;


procedure TData_Display.Double1Click( Sender : TObject ) ;

begin
    Set_Number_Size( 8 ) ;
end ;


procedure TData_Display.Gotoaddress1Click( Sender : TObject ) ;

begin
    Goto_Address_Form.Base := AD.Base ;
    if( Goto_Address_Form.ShowModal = mrOK ) then
    begin
        AD.Home_Index := strtoint( Cvtb( AD.Base, 10, Goto_Address_Form.ComboBox1.Text ) ) ;
        Display_Panel.Repaint ;
    end ;
end ;


procedure TData_Display.Find1Click( Sender : TObject ) ;

var B : char ;
    I : int64 ;
    Matched : integer ; // Number of bytes matched
    S : string ;

begin
    Find_Form.Base := AD.Base ;
    if( Find_Form.ShowModal = mrOK ) then
    begin
        S := Find_Form.Find_Image ;
        if( Find_Form.Literal_Text.Checked ) then
        begin
            if( not Find_Form.Case_Sensitive.Checked ) then
            begin
                S := uppercase( S ) ;
            end ;
            if( EBCDIC1.Caption = Text_EBCDIC_Amp ) then
            begin
                S := From_EBCDIC( S ) ;
            end ;
        end ;

        // Now search...
        I := AD.Home_Index ;
        Matched := 0 ;
        while( ( I >= 0 ) and ( I <= _Array.High_Bound( 0 ) ) ) do
        begin
            B := char( _Array.Get_Byte( I ) ) ;
            if( not Find_Form.Case_Sensitive.Checked ) then
            begin
                B := upcase( B ) ;
            end ;
            if( B <> S[ Matched + 1 ] ) then
            begin
                Matched := 0 ;
            end else
            begin
                inc( Matched ) ;
            end ;
            if( Matched = length( S ) ) then // Found entire data
            begin
                AD.Home_Index := I ;
                exit ;
            end ;
            I := I + 1 ;
        end ;
        ShowMessage( Text_Nothing_Found ) ;
    end ;
end ; // TData_Display.Find1Click


procedure TData_Display.Radix50Click( Sender : TObject ) ;

begin
    AD.Base := 50 ;
    Display_Panel.Repaint ;
    Radix50.Checked := True ;
end ;


// Property handlers...

procedure TData_Display.Set_Array( Value : TCOM_Array_Interface ) ;

begin
    AD.Free ;
    AD := TArray_Display.Create( Display_Panel, Value, Font.Handle, Color, Font.Color ) ;
    TPaint_Panel( Display_Panel ).AD := AD ;
    __Array := Value ;
    Display_Panel.Repaint ;
end ;


// API...

procedure TData_Display.wmPaint( var Message ) ;

var Range : int64 ;

begin
    if( AD <> nil ) then
    begin
        Display_Panel.Repaint ;
        ValidateRect( Handle, nil ) ;

        Scrollbar1.LargeChange := AD.LPP ;
        if( _Array <> nil ) then
        begin
            Range := _Array.High_Bound( 0 ) ;
            Range := Range - _Array.Low_Bound( 0 ) + 1 ;
            if( Range < 0 ) then // Overflow
            begin
                Range := $7FFFFFFF ;
            end ;
            Range := ( Range div AD.Max_Data ) * AD.LPP - 1 ;
            if( Range > $7FFFFFFF ) then
            begin
                Range := $7FFFFFFF ;
            end ;
            if( Range < 0 ) then
            begin
                Range := 0 ;
            end ;
            Scrollbar1.Max := Range ;
        end ;
    end ;
end ; // TData_Display.wmPaint


procedure TData_Display.ScrollBarChange( Sender : TObject ) ;

var Start : int64 ;

begin
    Start := Scrollbar1.Position ;
    Start := Start * AD.Max_Data ;
    Start := Start div AD.LPP ;
    AD.Home_Index := Start ;
    Display_Panel.Repaint ;
end ;


procedure TData_Display.Set_Number_Size( Size : integer ) ;

begin
    AD.Number_Size := Size ;
    case Size of
        2 : Word1.Checked := True ;
        4 : Long1.Checked := True ;
        8 : Double1.Checked := True ;
        else Byte1.Checked := True ;
    end ;
    Display_Panel.Repaint ;
end ;


end.
