<?php
 /*
        Program Name   : Express
        Package Name   : Subroutine Library - Layer 4
        Purpose        : Expression evaluator
        Institution    : Conroy & Conroy Co.
        Date written   : 25-Aug-2002
        Written by     : Alan Conroy
        Version        : 1.0I

        Copyright(C) 2002-2016 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *******************************************************
        *                                                     *
        *      M O D I F I C A T I O N   H I S T O R Y        *
        *                                                     *
        *******************************************************

              DATE    BY      REASON

          12-Oct-2002 EAC     Fix protection violation caused
                              by trailing operator.
                              Fix handling of E-notation.
           2-Nov-2002 EAC     Handle division by 0.
                              Fix compilation problems in BP.
                              Change Get_Symbol parameters to longint.
                              Fix error in calculation of HypArcCsc.
                              Fix problems with handling E-notation.
                              Code cleanup.
                              Fix returning of error code.
          30-Nov-2002 EAC     Properly handle an expression
                              which is simply a number.
                              Fixed unary minus on functions.
                              Initialize strings for Delphi.
          4-Apr-2006  EAC     Added SHR && SHL operators.
          6-Dec-2006  EAC     Fix bug that swallowed tokens
                              when processing a function name
                              that ended with "E".
          13-Apr-2007 EAC     Fix handling of alphanumeric
                              operators when following numeric literal is not
                              separated from operator with white space.
                              Increased "D" operator precedence.
          11-Jun-2008 EAC     Added NOT right-unary operator.
          9-Apr-2012  Klaus   Fix some minor bugs, add support for string
                              functions && operators
          14-Apr-2012 EAC     Fixed bug that sometimes mangled the expression
                              stack when using functions.
          22-Sep-2016 EAC     Added SF function.

        *******************************************************
        *                                                     *
        *      P R O G R A M   P U R P O S E                  *
        *                                                     *
        *******************************************************

          This unit contains classes && routines to evaluate expressions.

          This code was culled from the Sirius compiler expression code.
*/

require_once('EDEFINES.INC');
require_once('Parse.php'); // TSParser

/*
uses // Subroutine Library - Layer 1...
     Maths, // Power
     Num1s, // Num1s
*/

// Symbol flags...
DEFINE('SF_Right_Unary_Operator',1);
DEFINE('SF_Left_Unary_Operator',2);
DEFINE('SF_Binary_Operator',3);
DEFINE('SF_Function',4);

DEFINE('Node_Literal',0);
DEFINE('Node_Function',1);
DEFINE('Node_Operator',2);

class tExpression_Node
{
    public $Node_Type = 0 ; // Class of node
    public $Operator = 0 ; // Operator or function value
    public $Precedence = 0 ; // Precedence of node
    public $Back ;
    public $Left ;
    public $Right ;// pExpression_Node ;
    public $Literal = '' ;
} // tExpression_Node

// Operations...
DEFINE('Op_Negate_Stack',1);
DEFINE('Op_Add',2);
DEFINE('Op_Subtract',3);
DEFINE('Op_Multiply',4);
DEFINE('Op_Divide',5);
DEFINE('Op_Percent',6);
DEFINE('Op_Exponent',7);
DEFINE('Op_Factorial',8);

DEFINE('Op_Rad',9);
DEFINE('Op_Deg',10);
DEFINE('Op_Abs',11);
DEFINE('Op_Cos',12);
DEFINE('Op_Cot',13);
DEFINE('Op_Csc',14);
DEFINE('Op_Exp',15);
DEFINE('Op_Int',16);
DEFINE('Op_Log',17);
DEFINE('Op_Sec',18);
DEFINE('Op_Sin',19);
DEFINE('Op_Tan',20);
DEFINE('Op_Fix',21);
DEFINE('Op_Sqr',22);
DEFINE('Op_Sgn',23);
DEFINE('Op_Fib',24);

DEFINE('Op_Min',25);
DEFINE('Op_Max',2) ;

DEFINE('Op_Arccos',27);
DEFINE('Op_Arccot',28);
DEFINE('Op_Arccsc',29);
DEFINE('Op_Arcsec',30);
DEFINE('Op_Arcsin',31);
DEFINE('Op_Arctan',32);
DEFINE('Op_Hypcos',33);
DEFINE('Op_Hypcot',34);
DEFINE('Op_Hypcsc',35);
DEFINE('Op_Hypsec',36);
DEFINE('Op_Hypsin',37);
DEFINE('Op_Hyptan',38);
DEFINE('Op_Hyparccos',39);
DEFINE('Op_Hyparccot',40);
DEFINE('Op_HypArcCsc',41);
DEFINE('Op_Hyparcsec',42);
DEFINE('Op_Hyparcsin',43);
DEFINE('Op_Hyparctan',44);
DEFINE('Op_Log10',45);
DEFINE('Op_Rnd',46);

DEFINE('Op_Mod',47);
DEFINE('Op_Rem',48);
DEFINE('Op_D',49);
DEFINE('Op_Div',50);

DEFINE('Op_Root',51);
DEFINE('Op_PI',52);

DEFINE('Op_Shift_Left',53);
DEFINE('Op_Shift_Right',54);
DEFINE('Op_And',55);
DEFINE('Op_Or',56);
DEFINE('Op_Xor',57);
DEFINE('Op_Complement',58);

DEFINE('Op_Round',59);

DEFINE('Op_SF',60);

// Errors...
DEFINE('Error_Expected_Comma',1);
DEFINE('Error_Syntax_Error',2);
DEFINE('Error_Illegal_Symbol',3);
DEFINE('Error_Expected_Left_Parenthesis',4);
DEFINE('Error_Division_By_Zero',5);
DEFINE('Error_Illegal_Number_for_factorial',6);
DEFINE('Error_Floating_point_error',7);
DEFINE('Error_Illegal_Number',8);
DEFINE('Error_Illegal_Argument',9);
DEFINE('Error_Imaginary_Number',10);
DEFINE('Error_Illegal_Argument_in_LOG',11);
DEFINE('Error_Argument_Too_large_in_EXP',12);
DEFINE('Error_Missing_Right_Parenthesis',13);
DEFINE('Error_Integer_To_Large',14);
DEFINE('Error_Expression_Too_Complex',15);
DEFINE('Error_Internal_Error',16);

function IntPower($b,$e)
{
	return pow($b,$e);
}


function trunc($x)
{
	return floor($x);
}


function get_numeric_text_presentation()
{
    return array( 'Decimal' => '.' );
}


function RoundToDig($MyValue, $NDigit)
{
  $Decim = IntPower(10, floor($NDigit));
  if( $MyValue > 0 )
  {
  	  return floor(($MyValue / $Decim) + 0.5) * $Decim;
  }
  elseif( $MyValue < 0 )
  {
    return floor(($MyValue / $Decim) - 0.5) * $Decim;
  }
  else
  {
    return 0;
  }
} // RoundToDig


function EERT( $X )
{
    switch( $X )
    {
		case Error_Expected_Comma: return 'Expected Comma' ;
        case Error_Syntax_Error: return 'Syntax Error' ;
        case Error_Illegal_Symbol: return 'Illegal Symbol' ;
        case Error_Expected_Left_Parenthesis: return 'Expected Left Parenthesis' ;
        case Error_Division_By_Zero: return 'Division By Zero' ;
        case Error_Illegal_Number_for_factorial: return 'Illegal Number for factorial' ;
        case Error_Floating_point_error: return 'Floating point error' ;
        case Error_Illegal_Number: return 'Illegal Number' ;
        case Error_Illegal_Argument: return 'Illegal Argument' ;
        case Error_Imaginary_Number: return 'Imaginary Number' ;
        case Error_Illegal_Argument_in_LOG: return 'Illegal Argument in LOG' ;
        case Error_Argument_Too_large_in_EXP: return 'Argument Too large in EXP' ;
        case Error_Missing_Right_Parenthesis: return 'Missing Right Parenthesis' ;
        case Error_Integer_To_Large: return 'Integer too large' ;
        case Error_Expression_Too_Complex: return 'Expression too complex' ;
        case Error_Internal_Error: return 'Internal error' ;
    }
} // EERT



// TSExpression_Evaluator methods...
class TExpression_Evaluator
{
        // Instance data...
        private $_Error = 0 ;
        private $Literal_Text = '' ;
        private $Parser = new TParser() ;
        private $_Position = 0 ; // Position of error
        private $Stack = array() ;
        private $SStack = array() ;
        private $SSP = 0 ;    // String Stack Pointer
        private $Parameter_Delimiter = ',' ;
        private $Decimal_Char = ' ' ;
        private $Whole_Length = 0 ; // Length of full expression to evaluate


        // API...

        public function ERT( $X )
        {
            return EERT( $X ) ;
        }


        public function Get_Symbol( $X, &$Flags, &$Address, &$Value, &$_Result )
        {
            $_Result = 0 ; // Assume success
            if( $X == '+' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Add ;
                $Value = 16000 ; // Precedence
                return ;
            }
            if( $X == '-' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Subtract ;
                $Value = 16000 ; // Precedence
                return ;
            }
            if( $X == '*' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Multiply ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == '/' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Divide ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == '^' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Exponent ;
                $Value = 26000 ; // Precedence
                return ;
            }
            if( $X == 'SHL' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Shift_Left ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == 'SHR' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Shift_Right ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == 'AND' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_And ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == 'OR' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Or ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == 'XOR' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Xor ;
                $Value = 24000 ; // Precedence
                return ;
            }
            if( $X == 'MIN' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Min ;
                $Value = 10000 ; // Precedence
                return ;
            }
            if( $X == 'MAX' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Max ;
                $Value = 10000 ; // Precedence
                return ;
            }
            if( $X == '%' )
            {
                $Flags = SF_Left_Unary_Operator ;
                $Address = Op_Percent ;
                $Value = 28000 ;
                return ;
            }
            if( $X == '!' )
            {
                $Flags = SF_Left_Unary_Operator ;
                $Address = Op_Factorial ;
                $Value = 28000 ;
                return ;
            }
            if( $X == 'NOT' )
            {
                $Flags = SF_Right_Unary_Operator ;
                $Address = Op_Complement ;
                $Value = 28000 ;
                return ;
            }
            if( $X == 'MOD' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Mod ; // Modulus
                $Value = 24000 ;
                return ;
            }
            if( $X == 'REM' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Rem ; // Remainder
                $Value = 24000 ;
                return ;
            }
            if( $X == 'D' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_D ;
                $Value = 26000 ;
                return ;
            }
            if( $X == 'DIV' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Div ;
                $Value = 24000 ;
                return ;
            }

            if( $X == 'ROOT' )
            {
                $Flags = SF_Function ;
                $Address = Op_Root ;
                $Value = 2 ;
                return ;
            }
            if( $X == 'RAD' )
            {
                $Flags = SF_Function ;
                $Address = Op_Rad ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'RND' )
            {
                $Flags = SF_Function ;
                $Address = Op_Rnd ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'DEG' )
            {
                $Flags = SF_Function ;
                $Address = Op_Deg ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ABS' )
            {
                $Flags = SF_Function ;
                $Address = Op_Abs ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'COS' )
            {
                $Flags = SF_Function ;
                $Address = Op_Cos ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'COT' )
            {
                $Flags = SF_Function ;
                $Address = Op_Cot ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'CSC' )
            {
                $Flags = SF_Function ;
                $Address = Op_Csc ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'EXP' )
            {
                $Flags = SF_Function ;
                $Address = Op_Exp ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'INT' )
            {
                $Flags = SF_Function ;
                $Address = Op_Int ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'LOG' )
            {
                $Flags = SF_Function ;
                $Address = Op_Log ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'SEC' )
            {
                $Flags = SF_Function ;
                $Address = Op_Sec ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'SIN' )
            {
                $Flags = SF_Function ;
                $Address = Op_Sin ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'TAN' )
            {
                $Flags = SF_Function ;
                $Address = Op_Tan ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'FIX' )
            {
                $Flags = SF_Function ;
                $Address = Op_Fix ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'SQR' )
            {
                $Flags = SF_Function ;
                $Address = Op_Sqr ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'SGN' )
            {
                $Flags = SF_Function ;
                $Address = Op_Sgn ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'FIB' )
            {
                $Flags = SF_Function ;
                $Address = Op_FIB ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCCOS' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcCos ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCCOT' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcCot ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCCSC' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcCsc ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCSEC' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcSec ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCSIN' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcSin ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'ARCTAN' )
            {
                $Flags = SF_Function ;
                $Address = Op_ArcTan ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPCOS' )
            {
                $Flags = SF_Function ;
                $Address = Op_HypCos ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPCOT' )
            {
                $Flags = SF_Function ;
                $Address = Op_HypCot ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPCSC' )
            {
                $Flags = SF_Function ;
                $Address = Op_HypCsc ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPSEC' )
            {
                $Flags = SF_Function ;
                $Address = Op_HypSec ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPSIN' )
            {
                $Flags = SF_Function ;
                $Address = Op_Hypsin ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'HYPTAN' )
            {
                $Flags = SF_Function ;
                $Address = Op_HypTan ;
                $Value = 1 ;
                return ;
            }
            if( ( $X == 'HYPARCCOS' ) || ( $X == 'ARCHYPCOS' ) )
            {
                $Flags = SF_Function ;
                $Address = Op_Hyparccos ;
                $Value = 1 ;
                return ;
            }
            if( ( $X == 'HYPARCCOT' ) || ( $X == 'ARCHYPCOT' ) )
            {
                $Flags = SF_Function ;
                $Address = Op_Hyparccot ;
                $Value = 1 ;
                return ;
            }
            if( ( $X == 'HYPARCSEC' ) || ( $X == 'ARCHYPSEC' ) )
            {
                $Flags = SF_Function ;
                $Address = Op_Hyparcsec ;
                $Value = 1 ;
                return ;
            }
            if( ( $X == 'HYPARCSIN' ) || ( $X == 'ARCHYPSIN' ) )
            {
                $Flags = SF_Function ;
                $Address = Op_Hyparcsin ;
                $Value = 1 ;
                return ;
            }
            if( ( $X == 'HYPARCTAN' ) || ( $X == 'ARCHYPTAN' ) )
            {
                $Flags = SF_Function ;
                $Address = Op_Hyparctan ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'LOG10' )
            {
                $Flags = SF_Function ;
                $Address = Op_Log10 ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'SF' )
            {
                $Flags = SF_Function ;
                $Address = Op_SF ;
                $Value = 1 ;
                return ;
            }
            if( $X == 'PI' )
            {
                $Flags = SF_Function ;
                $Address = Op_Pi ;
                $Value = 0 ;
                return ;
            }
            if( $X == '#' )
            {
                $Flags = SF_Binary_Operator ;
                $Address = Op_Round ;
                $Value = 26000 ; // Precedence
                return ;
            }

            $_Result = 1 ; // Error
        } // Get_Symbol


        public function SPush( $ST )
        {
            $this->$SSP = $this->$SSP + 1 ;
            $this->$SStack[ $this->$SSP ] = $ST ;
        }


        public function Push( $F )
        {
            $this->$Stack[ 0 ] = $this->$Stack[ 0 ] + 1 ;
            $this->$Stack[ floor( $this->$Stack[ 0 ] ) ] = $F ;
        }


        public function Function_Reference( $Value )
        // Process function parameters.  Value == parameter count.
        {
            if( $Value != 0 ) // if parameters
            {
                $Y = Parser.Token() ;
                $G = 0 ; // Number of parameters that code was generated for
                if( Pos( $Y, '(//[' ) == 0 ) // No parameters specified
                {
                    Parser.Put_Token( Y ) ; // Restore token
                    if( $Value > 0 )
                    // One || more required parameters
                    {
                        Set_Error( Error_Expected_Left_Parenthesis ) ;
                        return False;
                    }
                }
                else // Parenthesis found
                {
                    $F = 1 ; // Number of parameters
                    while( $F <= abs( $Value ) ) // Until out of parameters
                    {
                        // Process next parameter...
                        $this->_Error = 0 ;
                        if( not Expression )
                        {
                            return False;
                        }
                        Inc( $G ) ;
                        $Y = Parser.Token() ; // Get next token
                        if(
                            ( $Y != Parameter_Delimiter )
                            &&
                            ( $F <= abs( $Value ) )
                          ) // No more parameters
                        {
                            if( Pos( $Y, '}])' ) == 0 )
                            {
                                Parser.Put_Token( Y ) ;
                            }
                            if( $Value < 0 ) // Variable number of parameters
                            {
                                $F = abs( $Value ) + 1 ; // Get actual number
                            }
                            elseif( $F < $Value ) // There should be more parameters
                            {
                                $this->Set_Error( Error_Expected_Comma ) ;
                                return False;
                            }
                        }
                        Inc( $F ) ; // Increment parameter count
                    } //while( F <= abs( $Value ) )}
                } //if}
                if( $Value < 0 ) // Variable number of parameters - push count of them
                {
                    $this->Push( G ) ;
                }
                if( ( Pos( $Y, ']})' ) == 0 ) && ( $G > 0 ) ) // No parenthesis when needed
                {
                    $this->Set_Error( Error_Missing_right_parenthesis ) ;
                    return False;
                }
            } // if Last_Symbol^.Size != 0
            return True ;
        } // Function_Reference


        function Zero_Expression_Tree( X : pExpression_Node )
        // Zero storage stored at tree with root X
        {
        var Current, Previous : pExpression_Node ;

            // Start at root
            Current = X ;
            while( Current != nil )
            {
                // Go all the way left
                while( Current^.Left != nil )
                {
                    Current = Current^.Left ;
                }
                if( Current^.Right != nil )
                {
                    Current = Current^.Right ;
                }
                else
                {
                    Previous = Current^.Back ;
                    if( Previous != nil )
                    {
                        if( Previous^.Left == Current )
                        {
                            Previous^.Left = nil ;
                        }
                        else
                        {
                            Previous^.Right = nil ;
                        }
                    }
                    switch( Current^.Node_Type )
					{
						case Node_Literal
						{
							strdispose( Current^.Literal ) ;
						}
                    }
                    dispose( Current ) ;
                    Current = Previous ;
                }
            } // while( Current != nil )
        } // Zero_Expression_Tree


        public function Numval( S : string ; var Dummy : integer )
        {
        var E : extended ;
            I : integer ;

            // Normalize to decimal...
            if( Decimal_Char != '.' )
            {
                I = pos( Decimal_Char, S ) ;
                if( I > 0 )
                {
                    S[ I ] = '.' ;
                }
            }

            // Normalize to locale...
            Dummy = pos( '.', S ) ;
            if( Dummy > 0 )
            {
            	$Numeric_Text_Presentation = get_numeric_text_presentation();
                S = copy( S, 1, Dummy - 1 ) +
                    $Numeric_Text_Presentation['Decimal'] +
                    copy( S, Dummy + 1, strlen( S ) ) ;
            }

            // Convert...
            val( S, E, Dummy ) ;
            Numval = E ;
        }


        public function Set_Error( $E )
        {
            if( $this->$_Error == 0 ) // Not already set
            {
                $this->$_Error = $E ;
            }
        }


        function Optimize_Tree( X : pExpression_Node )
        // Optimize an expression tree.
        {
        var Dummy : integer ;
            L, R : string ;
            LX, RX, Temp : extended ;
            LI, RI : longint ;
            Res : integer ;

            // Optimize sub-trees
            if( X^.Left != nil )
            {
                Optimize_Tree( X^.Left ) ;
            }
            if( X^.Right != nil )
            {
                Optimize_Tree( X^.Right ) ;
            }

            // See if we can optimize...
            if( X^.Node_Type != Node_Operator )
            {
                return ; // Can't optimize this any further
            }
            LX = 0 ; // Avoid compiler warning
            RX = 0 ; // Ditto
            if( X^.Left != nil )
            {
                if( X^.Left^.Node_Type != Node_Literal )
                {
                    return ;
                }
                L = strpas( X^.Left^.Literal ) ;
                LX = Numval( L, Dummy ) ;

            }
            if( X^.Right != nil )
            {
                if( X^.Right^.Node_Type != Node_Literal )
                {
                    return ;
                }
                R = strpas( X^.Right^.Literal ) ;
                RX = Numval( R, Dummy ) ;
            }

            switch( X^.Operator )
				case Op_Negate_Stack
				{
					LX = -LX ;
					break;
				}
                case Op_Add
                {
                	LX = LX + RX ;
					break;
				}
                case Op_Subtract
                {
                	LX = LX - RX ;
					break;
				}
                case Op_Multiply
                {
                	LX = LX * RX ;
					break;
				}
                case Op_Divide
                    {
                        if( RX == 0.0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                        }
                        else
                        {
                            LX = LX / RX ;
                        }
						break;
                    }
                case Op_Shift_Right
                    {
                        if( RX >= 1 )
                        {
                            Temp = 2 ;
                            while( RX >= 2 )
                            {
                                RX = RX - 1 ;
                                Temp = Temp + Temp ;
                            }
                            LX = LX / Temp ;
                        }
						break;
                    }
                case Op_Shift_Left
                    {
                        if( RX >= 1 )
                        {
                            Temp = 2 ;
                            while( RX >= 2 )
                            {
                                RX = RX - 1 ;
                                Temp = Temp + Temp ;
                            }
                            LX = LX * Temp ;
                        }
						break;
                    }
                case Op_And
                    {
                        if(
                            ( int( LX ) > high( LI ) )
                            or
                            ( int( LX ) < low( LI ) )
                            or
                            ( int( RX ) > high( RI ) )
                            or
                            ( int( RX ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                        }
                        else
                        {
                            LI = floor( int( LX ) ) ;
                            RI = floor( int( RX ) ) ;
                            LI = LI && RI ;
                            LX = LI ;
                        }
						break;
                    }
                case Op_Or
                    {
                        if(
                            ( int( LX ) > high( LI ) )
                            or
                            ( int( LX ) < low( LI ) )
                            or
                            ( int( RX ) > high( RI ) )
                            or
                            ( int( RX ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                        }
                        else
                        {
                            LI = floor( int( LX ) ) ;
                            RI = floor( int( RX ) ) ;
                            LI = LI || RI ;
                            LX = LI ;
                        }
						break;
                    }
                case Op_Xor
                    {
                        if(
                            ( int( LX ) > high( LI ) )
                            or
                            ( int( LX ) < low( LI ) )
                            or
                            ( int( RX ) > high( RI ) )
                            or
                            ( int( RX ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                        }
                        else
                        {
                            LI = floor( int( LX ) ) ;
                            RI = floor( int( RX ) ) ;
                            LI = LI xor RI ;
                            LX = LI ;
                        }
						break;
                    }
                case Op_Percent
                {
                	LX = LX / 100.0 ;
					break;
				}
                case Op_Exponent
                    {
                        LX = Power( LX, RX, Res ) ;
                        Set_Error( Res ) ;
						break;
                    }
                case Op_Factorial
                {
                	LX = Factorial( LX ) ;
					break;
				}
                case Op_Complement
                    {
                        if(
                            ( int( RX ) > high( LI ) )
                            or
                            ( int( RX ) < low( LI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                        }
                        else
                        {
                            LI = floor( int( RX ) ) ;
                            LI = not LI ;
                            LX = LI ;
                        }
						break;
                    }
                case Op_Max
                    {
                        if( LX < RX ) LX = RX ;
						break;
                    }
                case Op_Min
                    {
                        if( LX > RX ) LX = RX ;
						break;
                    }
                else return ;
            } // case X.Operator of

            L = num2( LX ) ;
            L = L + #0 ;
            X^.Literal = strnew( @L[ 1 ] ) ;
            X^.Node_Type = Node_Literal ;
            if( X^.Left != nil )
            {
                strdispose( X^.Left^.Literal ) ;
                dispose( X^.Left ) ;
                X^.Left = nil ;
            }
            if( X^.Right != nil )
            {
                strdispose( X^.Right^.Literal ) ;
                dispose( X^.Right ) ;
                X^.Right = nil ;
            }
        } // Optimize_Tree


        public function Pop()
        {
            Pop = Stack[ floor( Stack[ 0 ] ) ] ;
            if( Stack[ 0 ] == 0 )
            {
                Set_Error( Error_Internal_Error ) ;
            }
            else
            {
                Stack[ 0 ] = Stack[ 0 ] - 1 ;
            }
        }


        public function SPop()
        {
            SPop = SStack[ SSP ] ;
            if( SSP == 0 )
            {
                Set_Error( Error_Internal_Error ) ;
            }
            else
            {
                SSP = SSP - 1 ;
            }
        }


        public function Calculate( Operator : integer )
        {
        var Dummy : integer ;
            A1, A2, A3, E1, E2, Temp : extended ;
            LI, RI : longint ;
            Res : integer ;

            Calculate = False ; // Assume no errors
            switch( Operator )
                Op_Negate_Stack
                    {
                        Push( -Pop ) ;
						break;
                    }
                Op_Add
                    {
                        Push( Pop + Pop ) ;
						break;
                    }
                Op_Subtract
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        Push( E1 - E2 ) ;
						break;
                    }
                Op_Multiply
                    {
                        Push( Pop * Pop ) ;
                    	break;
					}
                Op_Divide
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 == 0.0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( E1 / E2 ) ;
						break;
                    }
                Op_Shift_Left
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 >= 1 )
                        {
                            Temp = 2 ;
                            while( E2 >= 2 )
                            {
                                E2 = E2 - 1 ;
                                Temp = Temp + Temp ;
                            }
                            Push( E1 * Temp ) ;
                        }
						break;
                    }
                Op_Shift_Right
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 >= 1 )
                        {
                            Temp = 2 ;
                            while( E2 >= 2 )
                            {
                                E2 = E2 - 1 ;
                                Temp = Temp + Temp ;
                            }
                            Push( E1 / Temp ) ;
                        }
						break;
                    }
                Op_And
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if(
                            ( int( E1 ) > high( LI ) )
                            or
                            ( int( E1 ) < low( LI ) )
                            or
                            ( int( E2 ) > high( RI ) )
                            or
                            ( int( E2 ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                            Calculate = True ;
                            return ;
                        }
                        LI = floor( int( E1 ) ) ;
                        RI = floor( int( E2 ) ) ;
                        LI = LI && RI ;
                        E1 = LI ;
                        Push( E1 ) ;
						break;
                    }
                Op_Or
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if(
                            ( int( E1 ) > high( LI ) )
                            or
                            ( int( E1 ) < low( LI ) )
                            or
                            ( int( E2 ) > high( RI ) )
                            or
                            ( int( E2 ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                            Calculate = True ;
                            return ;
                        }
                        LI = floor( int( E1 ) ) ;
                        RI = floor( int( E2 ) ) ;
                        LI = LI || RI ;
                        E1 = LI ;
                        Push( E1 ) ;
						break;
                    }
                Op_Xor
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if(
                            ( int( E1 ) > high( LI ) )
                            or
                            ( int( E1 ) < low( LI ) )
                            or
                            ( int( E2 ) > high( RI ) )
                            or
                            ( int( E2 ) < low( RI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                            Calculate = True ;
                            return ;
                        }
                        LI = floor( int( E1 ) ) ;
                        RI = floor( int( E2 ) ) ;
                        LI = LI xor RI ;
                        E1 = LI ;
                        Push( E1 ) ;
						break;
                    }
                Op_Percent
                    {
                        Push( Pop / 100.0 ) ;
						break;
                    }
                Op_Exponent
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        E1 = Power( E1, E2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( E1 ) ;
						break;
                    }
                Op_Round
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        E1 = RoundToDig( E1, E2 ) ;
                        Push( E1 ) ;
						break;
                    }
                Op_Factorial
                    {
                        Push( Factorial( Pop ) ) ;
						break;
                    }
                Op_Complement
                    {
                        E1 = Pop ;
                        if(
                            ( int( E1 ) > high( LI ) )
                            or
                            ( int( E1 ) < low( LI ) )
                          )
                        {
                            Set_Error( Error_Integer_To_Large ) ;
                            Calculate = True ;
                            return ;
                        }
                        LI = floor( int( E1 ) ) ;
                        LI = not LI ;
                        E1 = LI ;
                        Push( E1 ) ;
						break;
                    }
                Op_Max
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E1 > E2 )
                        {
                            E1 = E2 ;
                        }
                        Push( E1 ) ;
						break;
                    }
                Op_Min
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E1 < E2 )
                        {
                            E1 = E2 ;
                        }
                        Push( E1 ) ;
						break;
                    }

                Op_Rad
                    {
                        Push( Pop / ( 180. / Pi ) ) ;
						break;
                    }
                Op_Deg
                    {
                        Push( Pop * ( 180. / Pi ) ) ;
						break;
                    }
                Op_Abs
                    {
                        Push( Abs( Pop ) ) ;
						break;
                    }
                Op_Cos
                    {
                        Push( Cos( Pop ) ) ;
						break;
                    }
                Op_Cot
                    {
                        Push( 1 / Tan( Pop ) ) ;
						break;
                    }
                Op_Csc
                    {
                        Push( 1. / Sin( Pop ) ) ;
						break;
                    }
                Op_Exp
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Exp( E1 ) ) ;
						break;
                    }
                Op_Int
                    {
                        Push( Int( Pop ) ) ;
						break;
                    }
                Op_Log
                    {
                        E1 = Pop ;
                        if( E1 <= 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Ln( E1 ) ) ;
						break;
                    }
                Op_Sec
                    {
                        Push( 1. / Cos( Pop ) ) ;
						break;
                    }
                Op_Sin
                    {
                        Push( Sin( Pop ) ) ;
						break;
                    }
                Op_Tan
                    {
                        Push( Tan( Pop ) ) ;
						break;
                    }
                Op_Fix
                    {
                        Push( floor( Pop ) ) ;
						break;
                    }
                Op_Sqr
                    {
                        E1 = Pop ;
                        if( E1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Sqrt( E1 ) ) ;
						break;
                    }
                Op_Sgn
                    {
                        Push( Sgn( Pop ) ) ;
						break;
                    }
                Op_Fib
                    {
                        E1 = Pop ;
                        if( E1 > 182 ) // No fibonacci > 182
                        {
                            Set_Error( Error_Floating_point_error ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = Int( Abs( E1 ) ) ;
                        if( E1 < 2 )
                        {
                            Push( 1 ) ;
                        }
                        else
                        {
                            A1 = 1 ;
                            A2 = 1 ;
                            for Dummy = 2 to floor( E1 )
                            {
                                A3 = A2 ;
                                A2 = A1 ;
                                A1 = A2 + A3
                            }
                            Push( A1 + A2 ) ;
                        }
						break;
                    }
                Op_Arccos
                    {
                        E1 = Pop ;
                        A1 = Power( E1, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = 1 - A1 ;
                        if( A1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Arctan( Sqrt( A1 ) / E1 ) ) ;
						break;
                    }
                Op_ArcCot
                    {
                        E1 = Pop ;
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Arctan( 1. / E1 ) ) ;
						break;
                    }
                Op_ArcCsc
                    {
                        E1 = Power( Pop, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = E1 - 1 ;
                        if( E1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Arctan( 1. / Sqrt( E1 ) ) ) ;
						break;
                    }
                Op_ArcSec
                    {
                        E1 = Power( Pop, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = E1 - 1 ;
                        if( E1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Arctan( Sqrt( E1 ) ) ) ;
						break;
                    }
                Op_Arcsin
                    {
                        E1 = Pop ;
                        A1 = Power( E1, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = 1 - A1 ;
                        if( A1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( A1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            return ;
                        }
                        Push( Arcsin( E1 ) ) ;
						break;
                    }
                Op_ArcTan
                    {
                        Push( Arctan( Pop ) ) ;
						break;
                    }
                Op_Hypcos
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Hypcos( E1 ) ) ;
						break;
                    }
                Op_HypCot
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = Exp( E1 ) - Exp( -E1 ) ;
                        if( A1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypCot( E1 ) ) ;
						break;
                    }
                Op_Hypcsc
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = Exp( E1 ) - Exp( -E1 ) ;
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( 2. / E1 ) ;
						break;
                    }
                Op_HypSec
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = Exp( E1 ) + Exp( -E1 ) ;
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( 2. / E1 ) ;
						break;
                    }
                Op_HypSin
                    {
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypSin( E1 ) ) ;
						break;
                    }
                Op_HypTan
                    {
						break;
                        E1 = Pop ;
                        if( ( E1 > 88 ) || ( E1 < -88 ) )
                        {
                            Set_Error( Error_Argument_Too_large_in_EXP ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = Exp( E1 ) + Exp( -E1 ) ;
                        if( A1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypTan( E1 ) ) ;
						break;
                    }
                Op_HypArcCos
                    {
                        E1 = Pop ;
                        A1 = Power( E1, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = A1 - 1 ;
                        if( A1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = E1 + Sqrt( A1 ) ;
                        if( E1 <= 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Ln( E1 ) ) ;
						break;
                    }
                Op_SF // Superfactorial
                    {
                        E1 = Pop ;
                        Push( SuperFactorial( E1 ) ) ;
						break;
                    }
                Op_HypArcCot
                    {
                        E1 = Pop ;
                        if( E1 - 1 <= 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypArcCot( E1 ) ) ;
						break;
                    }
                Op_HypArcCsc
                    {
                        E1 = Pop ;
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( E1 < 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = Power( E1, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypArcCsc( E1 ) ) ;
						break;
                    }
                Op_HypArcSec
                    {
                        E1 = Pop ;
                        if( E1 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( E1 < 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = Power( E1, 2, Res ) ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        A1 = 1 / A1 - 1 ;
                        if( A1 < 0 )
                        {
                            Set_Error( 10 ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypArcSec( E1 ) ) ;
						break;
                    }
                Op_HypArcSin
                    {
                        E1 = Pop ;
                        A1 = Power( E1, 2, Res ) + 1 ;
                        if( Res != 0 )
                        {
                            Set_Error( Res ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( A1 < 0 )
                        {
                            Set_Error( Error_Imaginary_Number ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = E1 + Sqrt( A1 ) ;
                        if( E1 <= 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Ln( E1 ) ) ;
						break;
                    }
                Op_HypArcTan
                    {
                        E1 = Pop ;
                        if( abs( E1 ) >= 1 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( HypArcTan( E1 ) ) ;
						break;
                    }
                Op_Log10
                    {
                        E1 = Pop ;
                        if( E1 <= 0 )
                        {
                            Set_Error( Error_Illegal_Argument_in_LOG ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Log10( E1 ) ) ;
						break;
                    }
                Op_PI
                    {
                        Push( Pi ) ;
						break;
                    }
                Op_Rnd
                    {
                        Ensure_Randomness ;
                        Push( random( floor( Pop ) ) ) ;
						break;
                    }
                Op_Root
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 != int( E2 ) )
                        {
                            Set_Error( Error_Illegal_argument ) ;
                            Calculate = True ;
                            return ;
                        }
                        if( E1 < 0 ) // Negative value
                        {
                            if( ( floor( E2 ) && 1 ) == 0 ) // Even number
                            {
                                Set_Error( Error_Imaginary_number ) ;
                                Calculate = True ;
                                return ;
                            }
                        }
                        Push( Root( E1, floor( E2 ) ) ) ;
						break;
                    }

                Op_Mod
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( E1 - Int( E1 / E2 ) * E2 ) ;
						break;
                    }
                Op_Rem
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        E1 = E1 / E2 ;
                        Push( E1 - floor( E1 ) ) ;
						break;
                    }
                Op_Div
                    {
                        E2 = Pop ;
                        E1 = Pop ;
                        if( E2 == 0 )
                        {
                            Set_Error( Error_Division_By_Zero ) ;
                            Calculate = True ;
                            return ;
                        }
                        Push( Int( Int( E1 ) / Int( E2 ) ) ) ;
						break;
                    }
                Op_D
                    {
                        Ensure_Randomness ;
                        E2 = Pop ;
                        E1 = Pop ;
                        A2 = 0 ;
                        while( E1 > 0 )
                        {
                            A2 = A2 + random( floor( E2 ) ) + 1 ;
                            E1 = E1 - 1 ;
                        }
                        Push( A2 ) ;
                    }
            }
        } // Calculate


        function Generate_Tree( X : pExpression_Node ) // Evaluate tree with root X.
        {
        var Current, Previous : pExpression_Node ;
            Dummy : integer ;
            F : extended ;
            S : string ;

            // Start at root
            Current = X ;
            while( Current != nil )
            {
                // Go all the way left
                while( Current^.Left != nil )
                {
                    Current = Current^.Left ;
                }

                // Go right if no more left
                if( Current^.Right != nil )
                {
                    Current = Current^.Right ;
                }
                else
                {
                    // Process terminal node
                    Previous = Current^.Back ;
                    if( Previous != nil )
                    {
                        if( Previous^.Left == Current ) // On left of parent
                        {
                            Previous^.Left = nil ;
                        }
                        else
                        {
                            Previous^.Right = nil ;
                        }
                    }
                    switch( Current^.Node_Type )
						case Node_Operator
						case Node_Function
                            {
                                if( Calculate( Current^.Operator ) )
                                {
                                    Stack[ 0 ] = 0 ;
                                    return ;
                                }
								break;
                            }
						case Node_Literal
						{
                            S = strpas( Current^.Literal ) ;
                            F = Numval( S, Dummy ) ;
                            Push( F ) ;
                            strdispose( Current^.Literal ) ;
                        }
                    }
                    dispose( Current ) ;
                    Current = Previous ;
                }
            } // while( Current != nil )
        } // Generate_Tree


        function Number()
        {
        var B : Integer ;
            Negate : boolean ;
            S, Saved : string ;

            Saved = Parser.Non_Terminals.Text ;
            //Parser.Non_Terminals = '0123456789.E' ;}
            Literal_Text = Parser.Token ;
            Negate = False ;
            while( Literal_Text == '+' ) // Eat unary plus
            {
                Literal_Text = Parser.Token ;
            }
            while( Literal_Text == '-' ) // Accumulate unary minus
            {
                Negate = not Negate ;
                Literal_Text = Parser.Token ;
            }
            if( Literal_Text == Decimal_Char )
            {
                Literal_Text = '0' + Decimal_Char ;
            }
            if( copy( Literal_Text, strlen( Literal_Text ), 1 ) == 'E' )
            {
                S = Parser.Token ;
                if( ( S == '-' ) || ( S == '+' ) )
                {
                    Literal_Text = Literal_Text + S + Parser.Token ;
                }
                else
                {
                    Parser.Put_Token( S ) ;
                }
            }
        Loop:
              Numval( Literal_Text, B ) ; // See if a valid numer
            if( B == 0 ) // So far, so good
            {
                if( not Parser.Token_EOL ) // End of line is end of token
                {
                    B = strlen( Literal_Text ) ;
                    Literal_Text = Literal_Text + Parser.Token() ; // See if some more of a numeric specification
                    if( copy( Literal_Text, strlen( Literal_Text ), 1 ) == 'E' )
                    {
                        S = Parser.Token() ;
                        if( ( S == '-' ) || ( S == '+' ) )
                        {
                            Literal_Text = Literal_Text + S + Parser.Token() ;
                        }
                    }
                    if( B != strlen( Literal_Text ) ) // Source file didn't end
                    {
                        goto Loop ;
                    }
                }
            }
            if( copy( Literal_Text, 1, 1 ) == 'E' )
            {
                B = 2 ;
            }
            if( ( B > 1 ) && ( Literal_Text[ B - 1 ] == 'E' ) )
            {
                dec( B ) ;
            }
            if( B > 0 )
            {
                Parser.Put_Token( Copy( Literal_Text, B, strlen( Literal_Text ) - B + 1 ) ) ;
            }
            if( B == 1 )
            {
                Number = False ;
            }
            else
            {
                if( B > 0 )
                {
                    Literal_Text = Copy( Literal_Text, 1, B - 1 ) ;
                }
                if( Negate )
                {
                    Literal_Text = '-' + Literal_Text ;
                }
                Number = True ;
            }
            Parser.Non_Terminals.Assign( Saved ) ;
        } // Number


        function Get_Subexpression()
        // Get a subexpression (term) && return pointer to subexpression tree
        {
        var Subexpression, Temp : pExpression_Node ;
            Operator : pExpression_Node ;
            X : string ;
            AA : array[ 0..255 ] of char ;
            Address : int64 ;
            Flags : longint ;
            _Result, Value : longint ;

            Subexpression = nil ;
        Loop:
            X = Parser.Token() ; // Get next token
            if( strlen( X ) == 0 )
            {
                Set_Error( Error_Syntax_Error ) ;
                Get_Subexpression = Subexpression ;
                return ;
            }
            if( pos( X[ 1 ], '(//[' ) > 0 ) // Start of parenthesis}
            {
                Temp = Parse_Expression ;
                if( Subexpression == nil ) // First part of subexpression
                {
                    Subexpression = Temp ;
                
				}
				else
                {
                    Operator = Subexpression ;
                    while( Operator^.Left != nil )
                    {
                        Operator = Operator^.Left ;
                    }
                    Operator^.Left = Subexpression ;
                }
                X = Parser.Token() ;
                if( ( X == '' ) || ( pos( X[ 1 ], ')}]' ) == 0 ) ) // No closing parenthesis}
                {
                    Parser.Put_Token( X ) ;
                    Set_Error( Error_Missing_Right_Parenthesis ) ;
                    Get_Subexpression = nil ;
                    Zero_Expression_Tree( Subexpression ) ;
                    return ;
                }
                goto Operator_Loop ;
            }

            if( X == '+' ) // Unary plus
            {
                goto Loop ; // Ignore these
            }

            if( X == '-' ) // Unary minus
            {
                Temp = Get_Subexpression ; // Get subexpression
                if( Temp == nil )
                {
                    Set_Error( Error_Syntax_Error ) ;
                    Get_Subexpression = Subexpression ;
                    return ;
                }
                new( Operator ) ;
                Operator^.Back = nil ;
                Operator^.Left = Temp ;
                Operator^.Right = nil ;
                Operator^.Node_Type = Node_Operator ;
                Operator^.Precedence = 28000 ;
                Operator^.Operator = Op_Negate_Stack ;
                Temp^.Back = Operator ;
                if( Subexpression == nil ) // Nothing in tree yet
                {
                    Subexpression = Operator ;
                }
				else
                {
                    Temp = Subexpression ;
                    while( Temp^.Left != nil )
                    {
                        Temp = Temp^.Left ;
                    }
                    Temp^.Left = Operator ;
                    Operator^.Back = Temp ;
                }
                goto Operator_Loop ;
            }

            Parser.Put_Token( X ) ;
            if( Number ) // Numeric literal
            {
                new( Temp ) ;
                Temp^.Node_Type = Node_Literal ;
                X = Literal_Text ;
                Temp^.Left = nil ;
                Temp^.Right = nil ;
                Temp^.Back = nil ;
                Temp^.Literal = strnew( strpCopy( AA, Literal_Text ) ) ;
                Subexpression = Temp ;
                goto Operator_Loop ;
            }

            // Must be a variable, function, || an error
            X = Parser.Token() ;
            Get_Symbol( X, Flags, Address, Value, _Result ) ;
            if( _Result != 0 )
            {
                Set_Error( Error_Syntax_Error ) ;
                Get_Subexpression = Subexpression ;
                return ;
            }
            if( ( Flags && SF_Binary_Operator ) == SF_Right_Unary_Operator )
            {
                new( Operator ) ;
                Operator^.Back = nil ;
                Operator^.Left = nil ;
                Operator^.Node_Type = Node_Operator ;
                Operator^.Precedence = Value ;
                Operator^.Operator = Address ;
                Temp = Get_Subexpression ; // Get subexpression
                if( Temp == nil )
                {
                    Set_Error( Error_Syntax_Error ) ;
                    Get_Subexpression = Subexpression ;
                    return ;
                }
                Operator^.Right = Temp ;
                Temp^.Back = Operator ;
                if( Subexpression == nil ) // Nothing in tree yet
                {
                    Subexpression = Operator ;
                
				}
				else
                {
                    Temp = Subexpression ;
                    while( Temp^.Left != nil )
                    {
                        Temp = Temp^.Left ;
                    }
                    Temp^.Left = Operator ;
                    Operator^.Back = Temp ;
                }
                goto Operator_Loop ;
            }
            if( ( Flags && SF_Function ) != 0 )
            // Function reference.  Function references have two pieces of code:
            //  1) invocation of the function (including the call, && creation of
            //     temporary string results), && 2) parameter passing.
            {
                // Evaluate function...
                Function_Reference( Value ) ; // Process function parameters
                Calculate( Address ) ; // Now do function evaluation

                // Create new expression node for function call...
                new( Temp ) ;
                Temp^.Node_Type = Node_Literal ;
                X = Num2( Pop ) ; // Pop function value from stack
                Temp^.Left = nil ;
                Temp^.Right = nil ;
                Temp^.Back = nil ;
                Temp^.Literal = strnew( strpCopy( AA, X ) ) ;
                Subexpression = Temp ;
                goto Operator_Loop ;
            }
            Parser.Put_Token( X ) ;
            X = Parser.Token() ; // Pull back from token stream

            new( Temp ) ;
            Temp^.Node_Type = Node_Function ;
            Temp^.Back = nil ;
            Temp^.Left = nil ;
            Temp^.Right = nil ;
            Temp^.Operator = Address ;
            if( Subexpression == nil ) // Nothing in tree yet
            {
                Subexpression = Temp ;
            }
            else
            {
                Operator = Subexpression ;
                while( Operator^.Left != nil )
                {
                    Operator = Operator^.Left ;
                }
                Operator^.Left = Temp ;
                Temp^.Back = Operator ;
            }

            // Process all left-unary operators
        Operator_Loop:
            X = Parser.Token() ;
            if( strlen( X ) > 0 )
            {
                Get_Symbol( X, Flags, Address, Value, _Result ) ;
                if(
                    ( _Result == 0 )
                    &&
                    ( ( Flags && SF_Binary_Operator ) == SF_Left_Unary_Operator )
                  ) // Left-unary operator
                {
                    // Append unary operator to subexpression tree
                    new( Operator ) ;
                    Operator^.Precedence = Value ;
                    Operator^.Left = Subexpression ;
                    Operator^.Right = nil ;
                    Operator^.Back = nil ;
                    Subexpression^.Back = Operator ;
                    Operator^.Node_Type = Node_Operator ;
                    Operator^.Operator = Address ;
                    Subexpression = Operator ;
                    goto Operator_Loop ; // Go back for next unary operator
                }
                Parser.Put_Token( X ) ; // Push token
            }

            Get_Subexpression = Subexpression ;
        } // Get_Subexpression


        function Parse_Expression()
        // Parse an expression into an expression tree, && return pointer to root
        {
        var Operator_Precedence : integer ;
            Operator : string ;
            Left_Subexpression, Current, Previous, Root : pExpression_Node ;
            Flags, Value : longint ;
            Address : int64 ;
            _Result : longint ;
            S : string ;

            // Perform expression setup
            Current = nil ;
            Root = nil ;

            // Get first subexpression
            Left_Subexpression = Get_Subexpression ;

            // Process until end of expression
            while( Left_Subexpression != nil )
            {
        Operator_Loop:
                // Get operator
                Operator = Parser.Token() ;
                if( strlen( Operator ) == 0 )
                {
                    _Result = 1 ;
                    Flags = 0 ;
                
				}
				else
                {
                    Get_Symbol( Operator, Flags, Address, Value, _Result ) ;
                }

                // This code handles the case where an operator is alphanumeric and
                  not separated from following numeric literal by a space.
                S = Operator ;
                while(
                       ( strlen( S ) > 0 )
                       &&
                       ( pos( S[ strlen( S ) ], '0123456789' ) > 0 )
                       &&
                       ( _Result != 0 )
                     )
                {
                    S = copy( S, 1, strlen( S ) - 1 ) ; // Trim last digit
                    Get_Symbol( S, Flags, Address, Value, _Result ) ;
                    if( _Result == 0 )
                    {
                        if( ( Flags && SF_Binary_Operator ) == SF_Binary_Operator )
                        {
                            Parser.Put_Token( copy( Operator, strlen( S ) + 1, strlen( Operator ) ) ) ;
                            // Return digits
                            Operator = copy( Operator, 1, strlen( S ) ) ; // Strip digits
                        }
                    }
                }

                if(
                    ( _Result != 0 )
                    ||
                    ( ( Flags && SF_Binary_Operator ) == 0 )
                  ) // Invalid operator
                {
                    Parser.Put_Token( Operator ) ; // Return token
                    if( Current != nil )
                    {
                        Current^.Right = Left_Subexpression ;
                        Left_Subexpression^.Back = Current ;
                    
					}
					else
                    {
                        Root = Left_Subexpression ;
                    }
                    Set_Error( Error_Syntax_Error ) ;
                    Parse_Expression = Root ;
                    return ;
                }
                if( ( Flags && SF_Binary_Operator ) != SF_Binary_Operator )
                // Unary operator or missing binary operator
                {
                    if( ( Flags && SF_Binary_Operator ) != SF_Left_Unary_Operator )
                    // Left-unary is OK
                    {
                        Parser.Put_Token( Operator ) ; // Return token
                        if( Current == nil )
                        {
                            Parse_Expression = Left_Subexpression ;
                        ]
                        else
                        {
                            Current^.Right = Left_Subexpression ;
                            Left_Subexpression^.Back = Current ;
                            Parse_Expression = Root ;
                        }
                        Set_Error( Error_Syntax_Error ) ;
                        return ;
                    }
                }

                // Create node for operator
                if( Root == nil ) // First thing in tree
                {
                    new( Root ) ;
                    Current = Root ;
                    Previous = nil ;
                }
                else
                {
                    Previous = Current ; // Save previous current node
                    new( Current ) ;
                }
                Current^.Precedence = Value ;
                Current^.Left = Left_Subexpression ;
                Current^.Right = nil ;
                Current^.Back = Previous ;
                Current^.Node_Type = Node_Operator ;
                Current^.Operator = Address ;
                Left_Subexpression^.Back = Current ;

                // Insert operator node into existing tree
                if( Previous != nil ) // Tree exists
                {
                    if( Previous^.Left == nil )
                    {
                        Previous^.Left = Current ;
                    }
                    else
                    {
                        Previous^.Right = Current ;
                    }
                } // if( Previous != nil )

                // Handle precedence
                if( _Result == 0 )
                {
                    Operator_Precedence = Value ;
                    if( ( Current^.Back != nil ) && // There was a previous operator
                        ( Current^.Back^.Precedence >= Operator_Precedence )
                        // Current operator has less precedence than the last
                          (Operators with same precedence evaluate from left to right,
                           thus, the following has an implied lower precedence)
                       )
                    {
                        // Back up to the proper node
                        Previous = Current ;
                        while(
                               ( Previous^.Back != nil )
                               &&
                               ( Previous^.Back^.Precedence >= Operator_Precedence )
                             )
                        {
                            Previous = Previous^.Back ;
                        }

                        // Move current node to proper location in tree...
                        // Unlink from current location
                        if( Current^.Back^.Left == Current )
                        {
                            Current^.Back^.Left = Current^.Left ;
                        }
						else
                        {
                            Current^.Back^.Right = Current^.Left ;
                        }
                        if( Current^.Left != nil )
                        {
                            Current^.Left^.Back = Current^.Back ;
                        }
                        Current^.Back = Previous^.Back ;

                        // Relink to new location
                        if( Previous == Root ) // Current becomes root node
                        {
                            Current^.Left = Root ;
                            Root^.Back = Current ;
                            Root = Current ;
                            Current^.Back = nil ;
                        }
						else
                        {
                            if( Previous^.Back^.Left == Previous )
                            {
                                Previous^.Back^.Left = Current ;
                            }
                            else
                            {
                                Previous^.Back^.Right = Current ;
                            }
                            Current^.Left = Previous ;
                            Previous^.Back = Current ;
                        }
                    } // Current operator has less precedence than the last
                } // if( _Result == 0 )

                // Get next subexpression && loop
                Left_Subexpression = Get_Subexpression ;
            } // while( Left_Subexpression != nil )
            Parse_Expression = Root ;
        } // Parse_Expression


        function Expression() // Generate code for an expression.
        {
        var _Result : pExpression_Node ;

            _Result = Parse_Expression ;
            Expression = ( _Result != nil  ) ;
            if( _Result != nil )
            {
                // Optimize the expression tree
                $this->$_Error = 0 ;
                Optimize_Tree( _Result ) ;
                if( $this->$_Error != 0 ) // Some error during optimization
                {
                    return ;
                }

                // Generate code
                Generate_Tree( _Result ) ;
            }
        } // Expression


        public function Evaluate( $X )
        {
        var Dummy : integer ;

            // Setup...
            $this->$Stack[ 0 ] = 0 ; // Clear stack
            $this->$SSP = 0 ;
            $this->$_Error = 0 ;

            $E = Numval( $X, Dummy ) ;
            if( Dummy == 0 ) // A numeric constant
            {
                Push( $E ) ;
                return True;
            }

            Parser.Init ;
            Parser.Put_Token( $X ) ;
            $this->$Whole_Length = strlen( $X ) ;
            Evaluate = Expression ;
            if(
                ( $this->$_Error == 0 ) // Haven't already flagged an error
                &&
                (
                  ( Parser.Token() != '' )
                  ||
                  ( ( $this->$Stack[ 0 ] != 1 ) && ( $this->$SSP == 0 ) )
                )
              ) // Left-over unparsable text
            {
                $this->Set_Error( Error_Syntax_Error ) ;
            }
            if( $this->$_Error != 0 )
            {
                return False ;
            }
            $this->$_Position = $this->$Whole_Length - strlen( Parser.Grab_Line() );
            Parser.Done ;
        } // Evaluate


        public function IsResultString()
        {
            return ($this->$SSP == 1);
        }


        public function Get_Result()
        {
            return $this->$Stack[ 1 ] ;
        }


        public function Get_SResult()
        {
            return $this->$SStack[ 1 ] ;
        }


        public function Get_Dump()
        {
            s = inttostr(SSP)+': '+SStack[ 1 ]+','+SStack[ 2 ]+','+SStack[ 3 ]+','+SStack[ 4 ]+chr(13)+chr(10) ;
            s = s + inttostr(floor(stack[0]))+': '+Floattostr(stack[1])+' '+Floattostr(stack[2])+' ';
            s = s + Floattostr(stack[3])+' '+Floattostr(stack[4])+' ';
            return s + Floattostr(stack[5])+' '+Floattostr(stack[6]);
        }


        public function _Last_Error()
        {
            return $this->$_Error ;
        }


        public function Get_Position()
        {
            return $this->$_Position ;
        }


        public function Put_Token( $X )
        {
            Parser.Put_Token( $X ) ;
        }


        public function Token()
        {
            return Parser.Token() ;
        }


        public function Set_Parameter_Delimiter( $S )
        {
            $this->$Parameter_Delimiter = $S ;
        }


        public function Set_Decimal_Char( $C )
        {
            $this->$Decimal_Char = $C ;
        }
} // TExpression_Evaluator


function Eval( $Expression, &$_Result, &$Imaginary, &$E, &$P )
{
    $Evaluator = new TExpression_Evaluator() ;

    $Imaginary = 0 ;
    if( $Evaluator->Evaluate( $Expression ) )
    {
        $_Result = $Evaluator->Get_Result() ;
        $E = 0 ;
        $P = 0 ;
    }
    else
    {
        $_Result = 0 ;
        $P = $Evaluator->Get_Position() ;
        $E = $Evaluator->_Last_Error() ;
    }
} // Eval
?>
