{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
              Program Name : FiP
              Package Name : Subroutine Library - Layer 1
              Purpose      : File Processor wrapper
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-08
              Written By   : Alan Conroy
              Version      : 2.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2009 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines wrappers for the static File Processor object.
}

unit FiP ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

{$IFDEF FPC_OR_DELPHI}
  {$DEFINE WINDOWS}
{$ENDIF}

uses { Subroutine Library - Layer 0... }
     _UE, { PSUnified_Exception }

     { Subroutine Library - Layer 1... }
     _FiP, { TCOM_File_Processor }
     _Files, { PSFile }
     _FileSys, { TFind_Record }

     { Subroutine Library - Layer 3... }
     FileSys ; { PSFile_System_Server }

{$IFDEF FPC_OR_DELPHI}
function COM_File_Processor : TCOM_File_Processor ;
{$ENDIF}

implementation

uses { Subroutine Library - Layer 0... }
     _Types, { Ansistring }

     { Subroutine Library - Layer 1... }
     _Stores { PSStore_Server }
{$IFDEF FPC_OR_DELPHI}
     ,SysUtils { lowercase }
{$ENDIF}
     ;

{$IFDEF FPC_OR_DELPHI}
type TCOM_File_Processor_Wrapper = class( TCOM_File_Processor )
                                      public // Constructors and destructors...
                                          constructor Create( FP : PSFile_Processor ) ;

                                      private // Instance data...
                                        Temp : Ansistring ;
                                        _FP : PSFile_Processor ;

                                      public { API... }
                                        function Is_Class( Name : PAnsiChar ) : boolean ;
                                            override ;
                                        function Lookup( Filename : PAnsiChar ;
                                            var Find_Record : TFind_Record ) : PAnsiChar ;
                                            override ;
                                        function Open( const Name : PAnsiChar ; Mode : integer ;
                                            var UEC : PSUnified_Exception ) : PSFile ;
                                            override ;
                                        procedure _Initialize( const Nam, ID : PAnsiChar ;
                                            Flags : longint ; var UEC : PSUnified_Exception ) ;
                                            override ;
                                        function DeviceSize( Nam : PAnsiChar ) : int64 ;
                                            override ;
                                        function DeviceFree( Nam : PAnsiChar ) : int64 ;
                                            override ;
                                        function MKDir( Nam : PAnsiChar ) : longint ;
                                            override ;
                                        procedure Close_File_System( Name : PAnsiChar ) ;
                                            override ;
                                        function Open_File_System( Name : PAnsiChar ;
                                            Flags : integer ;
                                            var Res : PSUnified_Exception ) : PSFile_System ;
                                            override ;
                                        function File_System_Open( Nam : PAnsiChar ) : boolean ;
                                            override ;
                                        { Scan a file specification and fill the Info record.  Separator is the path
                                          separator to use in parsing. }
                                        procedure Filespec_String_Scan( Name, Separator : PAnsiChar ;
                                            var Info : TFSSS ) ; override ;
                                        function Query_Filename( Default, Prompt : PAnsiChar ;
                                            Mode : TFilename_Query_Mode ;
                                            Flags : integer ) : PAnsiChar ;
                                            override ;
                                        procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;
                                            override ;
                                        procedure Register_Store_Server( SS : PSStore_Server ) ;
                                            override ;
                                        function File_Size( Nam : PAnsiChar ) : int64 ;
                                            override ;
                                   end ; { TCOM_File_Processor_Wrapper }

// Constructors and destructors...

constructor TCOM_File_Processor_Wrapper.Create( FP : PSFile_Processor ) ;

begin
    inherited Create ;

    _FP := FP ;
end ;


// API...

function TCOM_File_Processor_Wrapper.Is_Class( Name : PAnsiChar ) : boolean ;

var S : string ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tcom_file_processor_wrapper' ;
end ;


function TCOM_File_Processor_Wrapper.Lookup( Filename : PAnsiChar ;
    var Find_Record : TFind_Record ) : PAnsiChar ;

begin
    Temp := _FP^.Lookup( Ansistring( Filename ), Find_Record ) ;
    Result := PAnsiChar( Temp ) ;
end ;


function TCOM_File_Processor_Wrapper.Open( const Name : PAnsiChar ; Mode : integer ;
    var UEC : PSUnified_Exception ) : PSFile ;

begin
    Result := _FP^.Open( Ansistring( Name ), Mode, UEC ) ;
end ;


procedure TCOM_File_Processor_Wrapper._Initialize( const Nam, ID : PAnsiChar ;
    Flags : longint ; var UEC : PSUnified_Exception ) ;

begin
    _FP^.Initialize( Ansistring( Nam ), Ansistring( ID ), Flags, UEC ) ;
end ;


function TCOM_File_Processor_Wrapper.DeviceSize( Nam : PAnsiChar ) : int64 ;

begin
    Result := _FP^.DeviceSize( Ansistring( Nam ) ) ;
end ;


function TCOM_File_Processor_Wrapper.DeviceFree( Nam : PAnsiChar ) : int64 ;

begin
    Result := _FP^.DeviceFree( Ansistring( Nam ) ) ;
end ;


function TCOM_File_Processor_Wrapper.MKDir( Nam : PAnsiChar ) : longint ;

begin
    Result := _FP^.MkDir( Ansistring( Nam ) ) ;
end ;


procedure TCOM_File_Processor_Wrapper.Close_File_System( Name : PAnsiChar ) ;

begin
    _FP^.Close_File_System( Ansistring( Name ) ) ;
end ;


function TCOM_File_Processor_Wrapper.Open_File_System( Name : PAnsiChar ;
    Flags : integer ; var Res : PSUnified_Exception ) : PSFile_System ;

begin
    Result := _FP^.Open_File_System( Ansistring( Name ), Flags, Res ) ;
end ;


function TCOM_File_Processor_Wrapper.File_System_Open( Nam : PAnsiChar ) : boolean ;

begin
    Result := _FP^.File_System_Open( Ansistring( Nam ) ) ;
end ;


procedure TCOM_File_Processor_Wrapper.Filespec_String_Scan( Name, Separator : PAnsiChar ;
    var Info : TFSSS ) ;

begin
    _FP^.Filespec_String_Scan( Ansistring( Name ), Ansistring( Separator ), Info ) ;
end ;


function TCOM_File_Processor_Wrapper.Query_Filename( Default, Prompt : PAnsiChar ;
    Mode : TFilename_Query_Mode ;
    Flags : integer ) : PAnsiChar ;

begin
    Temp := _FP^.Query_Filename( Ansistring( Default ), Ansistring( Prompt ), Mode, Flags ) ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TCOM_File_Processor_Wrapper.Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;

begin
    _FP^.Register_File_System_Server( FSS ) ;
end ;


procedure TCOM_File_Processor_Wrapper.Register_Store_Server( SS : PSStore_Server ) ;

begin
    _FP^.Register_Store_Server( SS ) ;
end ;


function TCOM_File_Processor_Wrapper.File_Size( Nam : PAnsiChar ) : int64 ;

begin
    Result := _FP^.File_Size( Ansistring( Nam ) ) ;
end ;


var FPW : TCOM_File_Processor = nil ;

function COM_File_Processor : TCOM_File_Processor ;

begin
    if( FPW = nil ) then
    begin
        FPW := TCOM_File_Processor_Wrapper.Create( _FiP.File_Processor ) ;
    end ;
    Result := FPW ;
end ;

initialization

finalization
    FPW.Free ;
    FPW := nil ;
{$ENDIF}
end.
