{
              Program Name : GetChar
              Package Name : Subroutine Library Layer 3
              Purpose      : Character I/O routines
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-1988
              Written By   : Alan Conroy
              Version      : 3.0

	      Copyright (C) 1988 by Alan Conroy.  Released to the Public Domain.

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These functions handle special keyboard input requests.
}

unit Getchars ;

interface

{$I EDEFINES.INC}

uses { Subroutine Library - Layer 0... }
     CommBase, { TBase_Object }
     _Types ; { Small_String }

type TIO_Source = object( TBase_Object )
                      public { API... }
                          procedure Clear( Handle : integer ) ; virtual ;
                          { Clear any pending I/O }
                  end ;

type TInput_Source = object( TIO_Source )
                         private { Instance data... }
                             Input_Buffer : Ansistring ;

                         public { API... }
                             constructor Init ;

                             procedure Clear( Handle : integer ) ; virtual ; { Clear any pending I/O }

                             { Forces input to the source. }
                             function Force_Input( S : Ansistring ) : boolean ;
                                 virtual ;

                             { Forces a special non-ASCII value. }
                             function Force_Special_Input( S : Ansistring ) : boolean ;
                                 virtual ;

                             { Returns a single "raw" character. }
                             function Get_Char( Option, Handle : integer ) : Ansichar ;
                                 virtual ;

                             { Returns a single "cooked" character. }
                             function Inkey( Handle : integer ) : Ansistring ;
                                 virtual ;

                             { Returns True if any input is waiting. }
                             function Pending_Input( Handle : integer ) : boolean ;
                                 virtual ;
                     end ;

type TDefault_Input = object( TInput_Source )
                         public { API... }
                             procedure Clear( Handle : integer ) ; virtual ; { Clear any pending I/O }

                             { Forces input to the source. }
                             function Force_Input( S : Ansistring ) : boolean ;
                                 virtual ;

                             { Forces a special non-ASCII value. }
                             function Force_Special_Input( S : Ansistring ) : boolean ;
                                 virtual ;

                             { Returns a single "raw" character. }
                             function Get_Char( Option, Handle : integer ) : Ansichar ;
                                 virtual ;

                             { Returns a single "cooked" character. }
                             function Inkey( Handle : integer ) : Ansistring ;
                                 virtual ;

                             { Returns True if any input is waiting. }
                             function Pending_Input( Handle : integer ) : boolean ;
                                 virtual ;
                      end ;

function Input_Line( I : TInput_Source ; X : integer ) : Small_String ;

var Default_Input : TDefault_Input ;

{$I EDEFINES.INC}

implementation

uses { Borland... }
     Windows, { PeekMessage }
{$IFDEF FPC_OR_DELPHI}
     Messages, { WM_Char }
     Forms, { Application }
{$ENDIF}

     { Subroutine Library - Layer 0... }
     _ASCII, { CR }
     TypeDefs, { Max_String_Length }

     { Subroutine Library - Layer 1... }
     CommonUt, { Clear_IOResult }

     { Subroutine Library - Layer 3... }
     O_S ; { OS }

{ TIO_Source methods... }

{ API... }

procedure TIO_Source.Clear( Handle : integer ) ;

begin
    { This routine left intentionally blank... }
end ;


{ TInput_Source methods... }

constructor TInput_Source.Init ;

begin
    inherited Init ;

{$IFDEF DYNAMIC_STRINGS}
    Initialize( Input_Buffer ) ;
{$ELSE}
    Input_Buffer := '' ;
{$ENDIF}
end ;


{ API... }

procedure TInput_Source.Clear( Handle : integer ) ;

begin
    inherited Clear( Handle ) ;
    Input_Buffer := '' ;
end ;


function TInput_Source.Force_Input( S : Ansistring ) : boolean ;

begin
    if( length( S ) + length( Input_Buffer ) > Max_String_Length ) then
    begin
        Force_Input := False ;
        exit ;
    end ;
    Force_Input := True ;
    Input_Buffer := Input_Buffer + S ;
end ;


function TInput_Source.Force_Special_Input( S : Ansistring ) : boolean ;

begin
    Force_Special_Input := False ; { We only understand ASCII }
end ;


function TInput_Source.Get_Char( Option, Handle : integer ) : Ansichar ;

begin
    if( length( Input_Buffer ) > 0 ) then
    begin
        Get_Char := Input_Buffer[ 1 ] ;
        Input_Buffer := copy( Input_Buffer, 2, length( Input_Buffer ) ) ;
    end else
    begin
        Get_Char := #0 ;
    end ;
end ;


function TInput_Source.Inkey( Handle : integer ) : Ansistring ;

begin
    InKey := Get_Char( 0, Handle ) ;
end ;


function TInput_Source.Pending_Input( Handle : integer ) : boolean ;

begin
    Pending_Input := ( length( Input_Buffer ) > 0 ) ;
end ;



{ TDefault_Input methods... }

{ API... }


{ Places input into the buffer, as if the user typed it.  It returns true if
  success or false if the buffer is full.  }
function TDefault_Input.Force_Input( S : Ansistring ) : boolean ;

begin
{$IFDEF FPC_OR_DELPHI}
    while( length( S ) > 0 ) do
    begin
	    PostMessage( 0, WM_CHAR, ord( S[ 1 ] ), 0 ) ;
        S := copy( S, 2, length( S ) ) ;
    end ;
    Force_Input := True ;
{$ELSE}
    Force_Input := inherited Force_Input( S ) ;
{$ENDIF}
end ; { Force_Input }


function TDefault_Input.Force_Special_Input( S : Ansistring ) : boolean ;
{ This function allows forcing a single special character.  The parameter is
  the key name of the character to force to the input buffer. }

var SS : Ansistring ;

    procedure Convert ;

    begin
        SS := '' ;
        if( S = 'ALTBACK' ) then
        begin
            SS := NUL + Ansichar( 14 ) ;
            exit ;
        end ;
        if( S = 'A[' ) then
        begin
            SS := NUL + Ansichar( 26 ) ;
            exit ;
        end ;
        if( S = 'A]' ) then
        begin
            SS := NUL + Ansichar( 27 ) ;
            exit ;
        end ;
        if( S = 'A;' ) then
        begin
            SS := NUL + Ansichar( 39 ) ;
            exit ;
        end ;
        if( S = 'A"' ) then
        begin
            SS := NUL + Ansichar( 40 ) ;
            exit ;
        end ;
        if( S = 'A`' ) then
        begin
            SS := NUL + Ansichar( 41 ) ;
            exit ;
        end ;
        if( S = 'A\' ) then
        begin
            SS := NUL + Ansichar( 43 ) ;
            exit ;
        end ;
        if( S = 'A,' ) then
        begin
            SS := NUL + Ansichar( 51 ) ;
            exit ;
        end ;
        if( S = 'A.' ) then
        begin
            SS := NUL + Ansichar( 52 ) ;
            exit ;
        end ;
        if( S = 'A/' ) then
        begin
            SS := NUL + Ansichar( 53 ) ;
            exit ;
        end ;
        if( S = 'AKP-' ) then
        begin
            SS := NUL + Ansichar( 74 ) ;
            exit ;
        end ;
        if( S = 'AKP+' ) then
        begin
            SS := NUL + Ansichar( 78 ) ;
            exit ;
        end ;
        if( S = 'CKP-' ) then
        begin
            SS := NUL + Ansichar( 142 ) ;
            exit ;
        end ;
        if( S = 'CKP+' ) then
        begin
            SS := NUL + Ansichar( 144 ) ;
            exit ;
        end ;
        if( S = 'AKP/' ) then
        begin
            SS := NUL + Ansichar( 149 ) ;
            exit ;
        end ;
        if( S = 'ALTHOME' ) then
        begin
            SS := NUL + Ansichar( 151 ) ;
            exit ;
        end ;
        if( S = 'ALTUP' ) then
        begin
            SS := NUL + Ansichar( 152 ) ;
            exit ;
        end ;
        if( S = 'ALTPAGEUP' ) then
        begin
            SS := NUL + Ansichar( 153 ) ;
            exit ;
        end ;
        if( S = 'ALTLEFT' ) then
        begin
            SS := NUL + Ansichar( 155 ) ;
            exit ;
        end ;
        if( S = 'ALTRIGHT' ) then
        begin
            SS := NUL + Ansichar( 157 ) ;
            exit ;
        end ;
        if( S = 'ALTEND' ) then
        begin
            SS := NUL + Ansichar( 159 ) ;
            exit ;
        end ;
        if( S = 'ALTDOWN' ) then
        begin
            SS := NUL + Ansichar( 160 ) ;
            exit ;
        end ;
        if( S = 'ALTPAGEDOWN' ) then
        begin
            SS := NUL + Ansichar( 161 ) ;
            exit ;
        end ;
        if( S = 'ALTINSERT' ) then
        begin
            SS := NUL + Ansichar( 162 ) ;
            exit ;
        end ;
        if( S = 'ALTDELETE' ) then
        begin
            SS := NUL + Ansichar( 163 ) ;
            exit ;
        end ;
        if( S = 'AKP/' ) then
        begin
            SS := NUL + Ansichar( 164 ) ;
            exit ;
        end ;
        if( S = 'ALTENTER' ) then
        begin
            SS := NUL + Ansichar( 166 ) ;
            exit ;
        end ;
        if( S = 'CKP4' ) then
        begin
            SS := NUL + Ansichar( 115 ) ;
            exit ;
        end ;
        if( S = 'CKP6' ) then
        begin
            SS := NUL + Ansichar( 116 ) ;
            exit ;
        end ;
        if( S = 'CKP8' ) then
        begin
            SS := NUL + Ansichar( 141 ) ;
            exit ;
        end ;
        if( S = 'CKP2' ) then
        begin
            SS := NUL + Ansichar( 145 ) ;
            exit ;
        end ;
        if( S = 'CKP9' ) then
        begin
            SS := NUL + Ansichar( 132 ) ;
            exit ;
        end ;
        if( S = 'CF11' ) then
        begin
            SS := NUL + Ansichar( 137 ) ;
            exit ;
        end ;
        if( S = 'CF12' ) then
        begin
            SS := NUL + Ansichar( 138 ) ;
            exit ;
        end ;
        if( S = 'AF11' ) then
        begin
            SS := NUL + Ansichar( 139 ) ;
            exit ;
        end ;
        if( S = 'AF12' ) then
        begin
            SS := NUL + Ansichar( 140 ) ;
            exit ;
        end ;
        if( S = 'A8' ) then
        begin
            SS := NUL + Ansichar( 127 ) ;
            exit ;
        end ;
        if( S = 'A9' ) then
        begin
            SS := NUL + Ansichar( 128 ) ;
            exit ;
        end ;
        if( S = 'A0' ) then
        begin
            SS := NUL + Ansichar( 129 ) ;
            exit ;
        end ;
        if( S = 'A-' ) then
        begin
            SS := NUL + Ansichar( 130 ) ;
            exit ;
        end ;
        if( S = 'A=' ) then
        begin
            SS := NUL + Ansichar( 131 ) ;
            exit ;
        end ;
        if( S = 'AQ' ) then
        begin
            SS := NUL + NUL ;
            exit ;
        end ;
        if( S = 'AS' ) then
        begin
            SS := NUL + US ;
            exit ;
        end ;
        if( S = 'AA' ) then
        begin
            SS := NUL + RS ;
            exit ;
        end ;
        if( S = 'UP' ) then
        begin
            SS := NUL + 'H' ;
            exit ;
        end ;
        if( S = 'DOWN' ) then
        begin
            SS := NUL + 'P' ;
            exit ;
        end ;
        if( S = 'RIGHT' ) then
        begin
            SS := NUL + 'M' ;
            exit ;
        end ;
        if( S = 'LEFT' ) then
        begin
            SS := NUL + 'K' ;
            exit ;
        end ;
        if( S = 'DELETE' ) then
        begin
            SS := NUL + 'S' ;
            exit ;
        end ;
        if( S = 'E1' ) then { Shift F1, etc }
        begin
            SS := NUL + 'T' ;
            exit ;
        end ;
        if( S = 'E2' ) then
        begin
            SS := NUL + 'U' ;
            exit ;
        end ;
        if( S = 'E3' ) then
        begin
            SS := NUL + 'V' ;
            exit ;
        end ;
        if( S = 'E4' ) then
        begin
            SS := NUL + 'W' ;
            exit ;
        end ;
        if( S = 'E5' ) then
        begin
            SS := NUL + 'X' ;
            exit ;
        end ;
        if( S = 'E6' ) then
        begin
            SS := NUL + 'Y' ;
            exit ;
        end ;
        if( S = 'E7' ) then
        begin
            SS := NUL + 'Z' ;
            exit ;
        end ;
        if( S = 'E8' ) then
        begin
            SS := NUL + '[' ;
            exit ;
        end ;
        if( S = 'E9' ) then
        begin
            SS := NUL + '\' ;
            exit ;
        end ;
        if( S = 'E10' ) then
        begin
            SS := NUL + ']' ;
            exit ;
        end ;
        if( S = 'E11' ) then
        begin
            SS := NUL + #135 ;
            exit ;
        end ;
        if( S = 'E12' ) then
        begin
            SS := NUL + #136 ;
            exit ;
        end ;
        if( S = 'F1' ) then
        begin
            SS := NUL + ';' ;
            exit ;
        end ;
        if( S = 'F2' ) then
        begin
            SS := NUL + '<' ;
            exit ;
        end ;
        if( S = 'F3' ) then
        begin
            SS := NUL + '=' ;
            exit ;
        end ;
        if( S = 'F4' ) then
        begin
            SS := NUL + '>' ;
            exit ;
        end ;
        if( S = 'F5' ) then
        begin
            SS := NUL + '?' ;
            exit ;
        end ;
        if( S = 'F6' ) then
        begin
            SS := NUL + '@' ;
            exit ;
        end ;
        if( S = 'F7' ) then
        begin
            SS := NUL + 'A' ;
            exit ;
        end ;
        if( S = 'F8' ) then
        begin
            SS := NUL + 'B' ;
            exit ;
        end ;
        if( S = 'F9' ) then
        begin
            SS := NUL + 'C' ;
            exit ;
        end ;
        if( S = 'F10' ) then
        begin
            SS := NUL + 'D' ;
            exit ;
        end ;
        if( S = 'F11' ) then
        begin
            SS := NUL + AnsiChar( 133 ) ;
            exit ;
        end ;
        if( S = 'F12' ) then
        begin
            SS := NUL + AnsiChar( 134 ) ;
            exit ;
        end ;
        if( S = 'F13' ) then
        begin
            SS := NUL + 'V' ;
            exit ;
        end ;
        if( S = 'F14' ) then
        begin
            SS := NUL + 'W' ;
            exit ;
        end ;
        if( S = 'F15' ) then
        begin
            SS := NUL + 'X' ;
            exit ;
        end ;
        if( S = 'F16' ) then
        begin
            SS := NUL + 'Y' ;
            exit ;
        end ;
        if( S = 'F17' ) then
        begin
            SS := NUL + 'Z' ;
            exit ;
        end ;
        if( S = 'F18' ) then
        begin
            SS := NUL + '[' ;
            exit ;
        end ;
        if( S = 'F19' ) then
        begin
            SS := NUL + '\' ;
            exit ;
        end ;
        if( S = 'F20' ) then
        begin
            SS := NUL + ']' ;
            exit ;
        end ;
        if( S = 'AF1' ) then
        begin
            SS := NUL + 'h' ;
            exit ;
        end ;
        if( S = 'AF2' ) then
        begin
            SS := NUL + 'i' ;
            exit ;
        end ;
        if( S = 'AF3' ) then
        begin
            SS := NUL + 'j' ;
            exit ;
        end ;
        if( S = 'AF4' ) then
        begin
            SS := NUL + 'k' ;
            exit ;
        end ;
        if( S = 'AF5' ) then
        begin
            SS := NUL + 'l' ;
            exit ;
        end ;
        if( S = 'AF6' ) then
        begin
            SS := NUL + 'm' ;
            exit ;
        end ;
        if( S = 'AF7' ) then
        begin
            SS := NUL + 'n' ;
            exit ;
        end ;
        if( S = 'AF8' ) then
        begin
            SS := NUL + 'o' ;
            exit ;
        end ;
        if( S = 'AF9' ) then
        begin
            SS := NUL + 'p' ;
            exit ;
        end ;
        if( S = 'AF10' ) then
        begin
            SS := NUL + 'q' ;
            exit ;
        end ;
        if( S = 'AF11' ) then
        begin
            SS := NUL + #139 ;
            exit ;
        end ;
        if( S = 'AF12' ) then
        begin
            SS := NUL + #140 ;
            exit ;
        end ;
        if( S = 'CF1' ) then
        begin
            SS := NUL + '^' ;
            exit ;
        end ;
        if( S = 'CF2' ) then
        begin
            SS := NUL + '_' ;
            exit ;
        end ;
        if( S = 'CF3' ) then
        begin
            SS := NUL + '`' ;
            exit ;
        end ;
        if( S = 'CF4' ) then
        begin
            SS := NUL + 'a' ;
            exit ;
        end ;
        if( S = 'CF5' ) then
        begin
            SS := NUL + 'b' ;
            exit ;
        end ;
        if( S = 'CF6' ) then
        begin
            SS := NUL + 'c' ;
            exit ;
        end ;
        if( S = 'CF7' ) then
        begin
            SS := NUL + 'd' ;
            exit ;
        end ;
        if( S = 'CF8' ) then
        begin
            SS := NUL + 'e' ;
            exit ;
        end ;
        if( S = 'CF9' ) then
        begin
            SS := NUL + 'f' ;
            exit ;
        end ;
        if( S = 'CF10' ) then
        begin
            SS := NUL + 'g' ;
            exit ;
        end ;
        if( S = 'BACK' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( ( S = 'KP0' ) or ( S = 'INSERT' ) ) then
        begin
            SS := NUL + 'R' ;
            exit ;
        end ;
        if( S = 'KP9' ) then
        begin
            SS := NUL + 'I' ;
            exit ;
        end ;
        if( (S = 'HOME' ) or ( S = 'KP7' ) ) then
        begin
            SS := NUL + 'G' ;
            exit ;
        end ;
        if( S = 'CKP4' ) then
        begin
            SS := NUL + 's' ;
            exit ;
        end ;
        if( S = 'CKP6' ) then
        begin
            SS := NUL + 't' ;
            exit ;
        end ;
        if( ( S = 'KP1' ) or ( S = 'END' ) ) then
        begin
            SS := NUL + 'O' ;
            exit ;
        end ;
        if( S = 'KP3' ) then
        begin
            SS := NUL + 'Q' ;
            exit ;
        end ;
        if( S = 'CKP7' ) then
        begin
            SS := NUL + 'w' ;
            exit ;
        end ;
        if( S = 'CKP1' ) then
        begin
            SS := NUL + 'u' ;
            exit ;
        end ;
        if( S = 'CKP3' ) then
        begin
            SS := NUL + 'v' ;
            exit ;
        end ;
        if( S = 'A1' ) then
        begin
            SS := NUL + 'x' ;
            exit ;
        end ;
        if( S = 'A2' ) then
        begin
            SS := NUL + 'y' ;
            exit ;
        end ;
        if( S = 'A3' ) then
        begin
            SS := NUL + 'z' ;
            exit ;
        end ;
        if( S = 'A4' ) then
        begin
            SS := NUL + '{' ;
            exit ;
        end ;
        if( S = 'A5' ) then
        begin
            SS := NUL + '|' ;
            exit ;
        end ;
        if( S = 'A6' ) then
        begin
            SS := NUL + '}' ;
            exit ;
        end ;
        if( S = 'A7' ) then
        begin
            SS := NUL + '~' ;
            exit ;
        end ;
        if( S = 'AB' ) then
        begin
            SS := NUL + '0' ;
            exit ;
        end ;
        if( S = 'AC' ) then
        begin
            SS := NUL + '.' ;
            exit ;
        end ;
        if( S = 'AD' ) then
        begin
            SS := NUL + ' ' ;
            exit ;
        end ;
        if( S = 'AE' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AF' ) then
        begin
            SS := NUL + '!' ;
            exit ;
        end ;
        if( S = 'AG' ) then
        begin
            SS := NUL + '"' ;
            exit ;
        end ;
        if( S = 'AH' ) then
        begin
            SS := NUL + '#' ;
            exit ;
        end ;
        if( S = 'AI' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AJ' ) then
        begin
            SS := NUL + '$' ;
            exit ;
        end ;
        if( S = 'AK' ) then
        begin
            SS := NUL + '%' ;
            exit ;
        end ;
        if( S = 'AL' ) then
        begin
            SS := NUL + '&' ;
            exit ;
        end ;
        if( S = 'AM' ) then
        begin
            SS := NUL + '2' ;
            exit ;
        end ;
        if( S = 'AN' ) then
        begin
            SS := NUL + '1' ;
            exit ;
        end ;
        if( S = 'AO' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AP' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AQ' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AR' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AT' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AU' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AV' ) then
        begin
            SS := NUL + '/' ;
            exit ;
        end ;
        if( S = 'AW' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AX' ) then
        begin
            SS := NUL + '-' ;
            exit ;
        end ;
        if( S = 'AY' ) then
        begin
            SS := NUL + '' ;
            exit ;
        end ;
        if( S = 'AZ' ) then
        begin
            SS := NUL + ',' ;
            exit ;
        end ;

        if( S = 'UP' ) then
        begin
            SS := ESC + '[A' ;
            exit ;
        end ;
        if( S = 'DOWN' ) then
        begin
            SS := ESC + '[B' ;
            exit ;
        end ;
        if( S = 'RIGHT' ) then
        begin
            SS := ESC + '[C' ;
            exit ;
        end ;
        if( S = 'LEFT' ) then
        begin
            SS := ESC + '[D' ;
            exit ;
        end ;
        if( S = 'COMPOSE' ) then
        begin
            SS := ESC + '[10' ;
            exit ;
        end ;
        if( S = 'F9' ) then
        begin
            SS := ESC + '[00' ;
            exit ;
        end ;
        if( S = 'AF1' ) then
        begin
            SS := ESC + '[1~a' ;
            exit ;
        end ;
        if( S = 'AF2' ) then
        begin
            SS := ESC + '[2~a' ;
            exit ;
        end ;
        if( S = 'AF3' ) then
        begin
            SS := ESC + '[3~a' ;
            exit ;
        end ;
        if( S = 'F18' ) then
        begin
            SS := ESC + '[12a' ;
            exit ;
        end ;

        if( S = 'F19' ) then
        begin
            SS := ESC + '[13a' ;
            exit ;
        end ;
        if( S = 'F14' ) then
        begin
            SS := '[?6a' ;
            exit ;
        end ;
        if( S = 'F6' ) then
        begin
            SS := '[17a' ;
            exit ;
        end ;
        if( S = 'F17' ) then
        begin
            SS := '[31a' ;
            exit ;
        end ;
        if( S = 'F10' ) then
        begin
            SS := '[21a' ;
            exit ;
        end ;
        if( S = 'F20' ) then
        begin
            SS := '[34a' ;
            exit ;
        end ;
        if( S = 'F4' ) then
        begin
            SS := '[24a' ;
            exit ;
        end ;
        if( S = 'F7' ) then
        begin
            SS := '[18a' ;
            exit ;
        end ;
        if( S = 'F15' ) then
        begin
            SS := '[28a' ;
            exit ;
        end ;
        if( S = 'F8' ) then
        begin
            SS := '[19a' ;
            exit ;
        end ;
        if( S = 'F16' ) then
        begin
            SS := '[29a' ;
            exit ;
        end ;

        if( S = 'AF4' ) then
        begin
            SS := ESC + '[4' ;
            exit ;
        end ;
        if( S = 'AF5' ) then
        begin
            SS := ESC + '[5' ;
            exit ;
        end ;
        if( S = 'AF6' ) then
        begin
            SS := ESC + '[6' ;
            exit ;
        end ;

        if( S = 'PF1' ) then
        begin
            SS := ESC + 'OP' ;
            exit ;
        end ;
        if( S = 'PF2' ) then
        begin
            SS := ESC + 'OQ' ;
            exit ;
        end ;
        if( S = 'PF3' ) then
        begin
            SS := ESC + 'OR' ;
            exit ;
        end ;
        if( S = 'PF4' ) then
        begin
            SS := ESC + 'OS' ;
            exit ;
        end ;
        if( S = 'ENTER' ) then
        begin
            SS := ESC + 'OM' ;
            exit ;
        end ;
        if( S = 'KP0' ) then
        begin
            SS := ESC + 'Op' ;
            exit ;
        end ;
        if( S = 'KP1' ) then
        begin
            SS := ESC + 'Oq' ;
            exit ;
        end ;
        if( S = 'KP2' ) then
        begin
            SS := ESC + 'Or' ;
            exit ;
        end ;
        if( S = 'KP3' ) then
        begin
            SS := ESC + 'Os' ;
            exit ;
        end ;
        if( S = 'KP4' ) then
        begin
            SS := ESC + 'Ot' ;
            exit ;
        end ;
        if( S = 'KP5' ) then
        begin
            SS := ESC + 'Ou' ;
            exit ;
        end ;
        if( S = 'KP6' ) then
        begin
            SS := ESC + 'Ov' ;
            exit ;
        end ;
        if( S = 'KP7' ) then
        begin
            SS := ESC + 'Ow' ;
            exit ;
        end ;
        if( S = 'KP8' ) then
        begin
            SS := ESC + 'Ox' ;
            exit ;
        end ;
        if( S = 'KP9' ) then
        begin
            SS := ESC + 'Oy' ;
            exit ;
        end ;
        if( S = 'KP-' ) then
        begin
            SS := ESC + 'Om' ;
            exit ;
        end ;
        if( S = 'KP,' ) then
        begin
            SS := ESC + 'Ol' ;
            exit ;
        end ;
        if( ( S = 'KP.' ) or ( S = 'DELETE' ) ) then
        begin
            SS := ESC + 'On' ;
            exit ;
        end ;
    end ;

begin
    Convert ;
    if( length( Input_Buffer ) + length( SS ) > Max_String_Length ) then
    begin
        Force_Special_Input := False ;
        exit ;
    end ;
    Force_Special_Input := True ;
    Input_Buffer := Input_Buffer + SS ;
end ; { Force_Special_Input }


function TDefault_Input.Get_Char( Option, Handle : integer ) : Ansichar ;
{ This function returns a single character typed at the keyboard.  If Option is
  non-zero, control returns immediately whether or not a character was waiting
  (in which case the returned character is zero, if no input was waiting). }

var B : boolean ;
    C : Ansichar ;

begin
    C := inherited Get_Char( Option, Handle ) ; { Get buffered input }
    if( C <> #0 ) then { Got something }
    begin
        Get_Char := C ;
        exit ;
    end ;
    Get_Char := OS^.Get_Char( ( Option = 0 ), B, Handle ) ;
end ; { TDefault_Input.Get_Char }


procedure TDefault_Input.Clear( Handle : integer ) ; { Clear input buffer }

begin
    inherited Clear( Handle ) ;

    OS^.Cancel_Typeahead( Handle ) ;
end ;


function TDefault_Input.Inkey( Handle : integer ) : Ansistring ;

var Command : string[ 11 ] ; { Input or command line }
{$IFDEF CONSOLE}
    Dummy : DWORD ;
    Count : integer ;
    _Handle : integer ;
    Input_Data : TInputRecord ;
    W : integer ;
{$ENDIF}

    procedure Do_Window ;

    begin
        Command[ 1 ] := Get_Char( 0, Handle ) ;
        if( ( Command[ 1 ] <> NUL ) and ( Command[ 1 ] <> ESC ) ) then
        { Normal character }
        begin
            Command[ 0 ] := #1 ;
            Inkey := Command ;
            exit ;
        end ;

        if( Command[ 1 ] = NUL ) then { IBM character }
        begin
            Command[ 2 ] := Get_Char( 1, Handle ) ;
            if( Command[ 2 ] = NUL ) then
            begin
                Command := 'AQ' ;
            end else
            if( Command[ 2 ] = US ) then
            begin
                Command := 'AS' ;
            end else
            if( Command[ 2 ] = RS ) then
            begin
                Command := 'AA' ;
            end else
            case Command[ 2 ] of
                #14 : Command := 'ALTBACK' ;
                #26 : Command := 'A[' ;
                #27 : Command := 'A]' ;
                #39 : Command := 'A;' ;
                #40 : Command := 'A"' ;
                #41 : Command := 'A`' ;
                #43 : Command := 'A\' ;
                #51 : Command := 'A,' ;
                #52 : Command := 'A.' ;
                #53 : Command := 'A/' ;
                #74 : Command := 'AKP-' ;
                #78 : Command := 'AKP+' ;
                #142 : Command := 'CKP-' ;
                #144 : Command := 'CKP+' ;
                #149 : Command := 'AKP/' ;
                #151 : Command := 'ALTHOME' ;
                #152 : Command := 'ALTUP' ;
                #153 : Command := 'ALTPAGEUP' ;
                #155 : Command := 'ALTLEFT' ;
                #157 : Command := 'ALTRIGHT' ;
                #159 : Command := 'ALTEND' ;
                #160 : Command := 'ALTDOWN' ;
                #161 : Command := 'ALTPAGEDOWN' ;
                #162 : Command := 'ALTINSERT' ;
                #163 : Command := 'ALTDELETE' ;
                #164 : Command := 'AKP/' ;
                #166 : Command := 'ALTENTER' ;

                #127 : Command := 'A8' ;
                #128 : Command := 'A9' ;
                #129 : Command := 'A0' ;
                #130 : Command := 'A-' ;
                #131 : Command := 'A=' ;
                #132 : Command := 'CKP9' ;
                #133 : Command := 'F11' ;
                #134 : Command := 'F12' ;
                #135 : Command := 'E11' ;
                #136 : Command := 'E12' ;
                #137 : Command := 'CF11' ;
                #138 : Command := 'CF12' ;
                #139 : Command := 'AF11' ;
                #140 : Command := 'AF12' ;
                #141 : Command := 'CKP8' ;
                #145 : Command := 'CKP2' ;
                'H' : Command := 'UP' ;
                'P' : Command := 'DOWN' ;
                'M' : Command := 'RIGHT' ;
                'K' : Command := 'LEFT' ;
                'S' : Command := 'DELETE' ;
                ';' : Command := 'F1' ;
                '<' : Command := 'F2' ;
                '=' : Command := 'F3' ;
                '>' : Command := 'F4' ;
                '?' : Command := 'F5' ;
                '@' : Command := 'F6' ;
                'A' : Command := 'F7' ;
                'B' : Command := 'F8' ;
                'C' : Command := 'F9' ;
                'D' : Command := 'F10' ;
                'T' : Command := 'E1' ; { Shift F1, etc }
                'U' : Command := 'E2' ;
                'V' : Command := 'E3' ;
                'W' : Command := 'E4' ;
                'X' : Command := 'E5' ;
                'Y' : Command := 'E6' ;
                'Z' : Command := 'E7' ;
                '[' : Command := 'E8' ;
                '\' : Command := 'E9' ;
                ']' : Command := 'E10' ;
                'h' : Command := 'AF1' ; { Alt F1, etc }
                'i' : Command := 'AF2' ;
                'j' : Command := 'AF3' ;
                'k' : Command := 'AF4' ;
                'l' : Command := 'AF5' ;
                'm' : Command := 'AF6' ;
                'n' : Command := 'AF7' ;
                'o' : Command := 'AF8' ;
                'p' : Command := 'AF9' ;
                'q' : Command := 'AF10' ;
                '^' : Command := 'CF1' ; { Control F1 is CF1, etc }
                '_' : Command := 'CF2' ;
                '`' : Command := 'CF3' ;
                'a' : Command := 'CF4' ;
                'b' : Command := 'CF5' ;
                'c' : Command := 'CF6' ;
                'd' : Command := 'CF7' ;
                'e' : Command := 'CF8' ;
                'f' : Command := 'CF9' ;
                'g' : Command := 'CF10' ;
                '' : Command := 'BACK' ; { Back tab }
                'R' : Command := 'INSERT' ;
                'I' : Command := 'KP9' ; { PAGEUP }
                'G' : Command := 'HOME' ;
                's' : Command := 'CKP4' ;
                't' : Command := 'CKP6' ;
                'O' : Command := 'END' ;
                'Q' : Command := 'KP3' ; { PAGEDOWN }
                'w' : Command := 'CKP7' ;
                'u' : Command := 'CKP1' ;
                'v' : Command := 'CKP3' ;
                'x' : Command := 'A1' ;
                'y' : Command := 'A2' ;
                'z' : Command := 'A3' ;
                '{' : Command := 'A4' ;
                '|' : Command := 'A5' ;
                '}' : Command := 'A6' ;
                '~' : Command := 'A7' ;
                '0' : Command := 'AB' ;
                '.' : Command := 'AC' ;
                ' ' : Command := 'AD' ;
                '' : Command := 'AE' ;
                '!' : Command := 'AF' ;
                '"' : Command := 'AG' ;
                '#' : Command := 'AH' ;
                '' : Command := 'AI' ;
                '$' : Command := 'AJ' ;
                '%' : Command := 'AK' ;
                '&' : Command := 'AL' ;
                '2' : Command := 'AM' ;
                '1' : Command := 'AN' ;
                '' : Command := 'AO' ;
                '' : Command := 'AP' ;
                '' : Command := 'AQ' ;
                '' : Command := 'AR' ;
                '' : Command := 'AT' ;
                '' : Command := 'AU' ;
                '/' : Command := 'AV' ;
                '' : Command := 'AW' ;
                '-' : Command := 'AX' ;
                '' : Command := 'AY' ;
                ',' : Command := 'AZ' ;
                else exit ;
            end ;
            Inkey := Command ;
            exit ;
        end ;

        { Escape character... }
        Command[ 1 ] := Get_Char( 1, Handle ) ;
        if( Command[ 1 ] = NUL ) then { Just an escape }
        begin
            Command := ESC ;
            Inkey := Command ;
            exit ;
        end ;

        { Special character... }
        Command[ 2 ] := Get_Char( 0, Handle ) ; { Get third character }
        if( Command[ 1 ] = AnsiChar( 91 ) ) {"["} Then
        begin
            case Command[ 2 ] Of
                'A' : Command := 'UP' ;
                'B' : Command := 'DOWN' ;
                'C' : Command := 'RIGHT' ;
                'D' : Command := 'LEFT' ;
                '1'..'3' : begin
                               Command[ 3 ] := Get_Char( 1, Handle ) ;
                               Command[ 4 ] := Get_Char( 1, Handle ) ;
                               case Command[ 3 ] of
                                   '~' : Command := 'E' + Command[ 2 ] ;
                                   '0' : if( Command[ 2 ] = '1' ) then
                                         begin
                                             Command := 'COMPOSE' ;
                                         end else
                                         begin
                                             Command := 'F9' ;
                                         end ;
                                   '2' : Command := 'F18' ;
                                   '3' : Command := 'F19' ;
                                   '6' : Command := 'F14' ;
                                   '7' : Command := 'F6' ;
                                   '1' : if( Command[ 2 ] = '3' ) then Command := 'F17'
                                             Else Command := 'F10' ;
                                   '4' : if( Command[ 2 ] = '3' ) then Command := 'F20'
                                             Else Command := 'F4' ;
                                   '8' : if( Command[ 2 ] = '1' ) then Command := 'F7'
                                             Else Command := 'F15' ;
                                   '9' : if( Command[ 2 ] = '1' ) then Command := 'F8'
                                             Else Command := 'F16' ;
                                   else exit ;
                               end ;
                               Inkey := Command ;
                               exit ;
                           end ;
                '4' : Command := 'E4' ;
                '5' : Command := 'E5' ;
                '6' : Command := 'E6' ;
                else exit ;
            end ;
            Inkey := Command ;
            exit ;
        end ;
        if( Command[ 1 ] = 'W' ) then { Windows virtual key code }
        begin
            case ord( Command[ 2 ] ) of
                1 : Command := 'LBUTTON' ;
                2 : Command := 'RBUTTON' ;
                3 : Command := 'CANCEL' ;
                4 : Command := 'MBUTTON' ;
                12 : Command := 'CLEAR' ;
                $10 : Command := 'SHIFT' ;
                17 : Command := 'CONTROL' ;
                18 : Command := 'MENU' ;
                19 : Command := 'PAUSE' ;
                20 : Command := 'CAPITAL' ;
                21 : Command := 'KANA' ;
                23 : Command := 'JUNJA' ;
                24 : Command := 'FINAL' ;
                25 : Command := 'HANJA' ;
                28 : Command := 'CONVERT' ;
                29 : Command := 'NONCONVERT' ;
                30 : Command := 'ACCEPT' ;
                31 : Command := 'MODECHANGE' ;
                33 : Command := 'PRIOR' ;
                34 : Command := 'NEXT' ;
                35 : Command := 'END' ;
                36 : Command := 'HOME' ;
                37 : Command := 'LEFT' ;
                38 : Command := 'UP' ;
                39 : Command := 'RIGHT' ;
                40 : Command := 'DOWN' ;
                41 : Command := 'SELECT' ;
                42 : Command := 'PRINT' ;
                43 : Command := 'EXECUTE' ;
                44 : Command := 'SNAPSHOT' ;
                45 : Command := 'INSERT' ;
                46 : Command := 'DELETE' ;
                47 : Command := 'HELP' ;
                91 : Command := 'LWIN' ;
                92 : Command := 'RWIN' ;
                93 : Command := 'APPS' ;
                112 : Command := 'F1' ;
                113 : Command := 'F2' ;
                114 : Command := 'F3' ;
                115 : Command := 'F4' ;
                116 : Command := 'F5' ;
                117 : Command := 'F6' ;
                118 : Command := 'F7' ;
                119 : Command := 'F8' ;
                120 : Command := 'F9' ;
                121 : Command := 'F10' ;
                122 : Command := 'F11' ;
                123 : Command := 'F12' ;
                124 : Command := 'F13' ;
                125 : Command := 'F14' ;
                126 : Command := 'F15' ;
                127 : Command := 'F16' ;
                128 : Command := 'F17' ;
                129 : Command := 'F18' ;
                130 : Command := 'F19' ;
                131 : Command := 'F20' ;
                132 : Command := 'F21' ;
                133 : Command := 'F22' ;
                134 : Command := 'F23' ;
                135 : Command := 'F24' ;
                144 : Command := 'NUMLOCK' ;
                145 : Command := 'SCROLL' ;
                229 : Command := 'PROCESSKEY' ;
                246 : Command := 'ATTN' ;
                247 : Command := 'CRSEL' ;
                248 : Command := 'EXSEL' ;
                249 : Command := 'EREOF' ;
                250 : Command := 'PLAY' ;
                251 : Command := 'ZOOM' ;
                252 : Command := 'NONAME' ;
                253 : Command := 'PA1' ;
                254 : Command := 'OEMCLEAR' ;
                else exit ;
            end ;
            Inkey := Command ;
            exit ;
        end ;
        if( Command[ 1 ] = 'O' ) then
        begin
            case Command[ 2 ] of
                'M' : Command := 'ENTER' ;
                'P' : Command := 'PF1' ;
                'Q' : Command := 'PF2' ;
                'R' : Command := 'PF3' ;
                'S' : Command := 'PF4' ;
                'l' : Command := 'KP,' ;
                'm' : Command := 'KP-' ;
                'n' : Command := 'KP.' ;
                'p' : Command := 'KP0' ;
                'q' : Command := 'KP1' ;
                'r' : Command := 'KP2' ;
                's' : Command := 'KP3' ;
                't' : Command := 'KP4' ;
                'u' : Command := 'KP5' ;
                'v' : Command := 'KP6' ;
                'w' : Command := 'KP7' ;
                'x' : Command := 'KP8' ;
                'y' : Command := 'KP9' ;
                else exit ;
            end ;
            Inkey := Command ;
        end ;
    end ; // .Do_Window


{$IFDEF CONSOLE}
    procedure Do_Console ;

    label Loop ;

    var Msg : TMsg ;

    begin
    Loop:
        Command := inherited Get_Char( 0, Handle ) ;
        if( Command <> #0 ) then { Pending input }
        begin
            W := 0 ;
        end else
        begin
            Command := '' ;
            GetNumberOfConsoleInputEvents( _Handle, Dummy ) ;
            while( Dummy = 0 ) do
            begin
                Application.ProcessMessages ; { So other thigns can happen while we wait }
                GetNumberOfConsoleInputEvents( _Handle, Dummy ) ;
                if( ( Dummy = 0 ) and ( Handle = 0 ) ) then
                begin
                    { If there are any windows open, the input may go there
                      instead of through the console, meaning that this loop
                      will never end, even though there is pending input, since
                      GetNumberOfConsoleInputEvents always returns 0. }
                    if( PeekMessage( Msg, 0, WM_KEYFIRST, WM_KEYLAST, PM_NOREMOVE ) ) then
                    begin
                        Do_Window ;
                        exit ;
                    end ;
                end ;
            end ;

            { Get the data... }
            ReadConsoleInput( _Handle, Input_Data, 1, Dummy ) ;
            if(
                ( Input_Data.EventType <> KEY_EVENT )
                or
                ( not Input_Data.Event.KeyEvent.bKeyDown )
              ) then
            begin
                goto Loop ; { Skip anything but key-down }
            end ;

            Command := Input_Data.Event.KeyEvent.AsciiChar ;
            W := Input_Data.Event.KeyEvent.wVirtualKeyCode ;

            { Handle repeats... }
            Count := Input_Data.Event.KeyEvent.wRepeatCount - 1 ;
            Dummy := length( Input_Buffer ) + Count ;
            setlength( Input_Buffer, Dummy ) ;
            while( Count > 0 ) do
            begin
                Input_Buffer[ Dummy ] := Command[ 1 ] ;
                dec( Dummy ) ;
                dec( Count ) ;
            end ;
        end ; { if( Command <> #0 ) }
        case W of
            VK_LBUTTON : Command := 'LBUTTON' ;
            VK_RBUTTON : Command := 'RBUTTON' ;
            VK_CANCEL : Command := 'CANCEL' ;
            VK_MBUTTON : Command := 'MBUTTON' ;
            VK_BACK : Command := BS ;
            VK_TAB : Command := HT ;
            VK_CLEAR : Command := 'CLEAR' ;
            VK_RETURN : Command := CR ;
            VK_SHIFT : Command := 'SHIFT' ;
            VK_CONTROL : Command := 'CONTROL' ;
            VK_MENU : Command := 'ALT' ;
            VK_PAUSE : Command := 'PAUSE' ;
            VK_CAPITAL : Command := 'CAPS' ;
            VK_ESCAPE : Command := ESC ;
            VK_SPACE : Command := ' ' ;
            VK_PRIOR : Command := 'PAGEUP' ;
            VK_NEXT : Command := 'PAGEDOWN' ;
            VK_END : Command := 'END' ;
            VK_HOME : Command := 'HOME' ;
            VK_LEFT : Command := 'LEFT' ;
            VK_UP : Command := 'UP' ;
            VK_RIGHT : Command := 'RIGHT' ;
            VK_DOWN : Command := 'DOWN' ;
            VK_SELECT : Command := 'SELECT' ;
            VK_PRINT : Command := 'PRINT' ;
            VK_EXECUTE : Command := 'EXECUTE' ;
            VK_SNAPSHOT : Command := 'PRINT' ;
            VK_INSERT : Command := 'INSERT' ;
            VK_DELETE : Command := 'DELETE' ;
            VK_HELP : Command := 'HELP' ;
            VK_LWIN : Command := 'LWIN' ;
            VK_RWIN : Command := 'RWIN' ;
            VK_APPS : Command := 'APPS' ;
            VK_NUMPAD0 : Command := 'KP0' ;
            VK_NUMPAD1 : Command := 'KP1' ;
            VK_NUMPAD2 : Command := 'KP2' ;
            VK_NUMPAD3 : Command := 'KP3' ;
            VK_NUMPAD4 : Command := 'KP4' ;
            VK_NUMPAD5 : Command := 'KP5' ;
            VK_NUMPAD6 : Command := 'KP6' ;
            VK_NUMPAD7 : Command := 'KP7' ;
            VK_NUMPAD8 : Command := 'KP8' ;
            VK_NUMPAD9 : Command := 'KP9' ;
            VK_MULTIPLY : Command := 'KP*' ;
            VK_ADD : Command := 'KP+' ;
            VK_SEPARATOR : Command := 'ENTER' ;
            VK_SUBTRACT : Command := 'KP-' ;
            VK_DECIMAL : Command := 'KP.' ;
            VK_DIVIDE : Command := 'KP/' ;
            VK_F1 : Command := 'F1' ;
            VK_F2 : Command := 'F2' ;
            VK_F3 : Command := 'F3' ;
            VK_F4 : Command := 'F4' ;
            VK_F5 : Command := 'F5' ;
            VK_F6 : Command := 'F6' ;
            VK_F7 : Command := 'F7' ;
            VK_F8 : Command := 'F8' ;
            VK_F9 : Command := 'F9' ;
            VK_F10 : Command := 'F10' ;
            VK_F11 : Command := 'F11' ;
            VK_F12 : Command := 'F12' ;
            VK_F13 : Command := 'F13' ;
            VK_F14 : Command := 'F14' ;
            VK_F15 : Command := 'F15' ;
            VK_F16 : Command := 'F16' ;
            VK_F17 : Command := 'F17' ;
            VK_F18 : Command := 'F18' ;
            VK_F19 : Command := 'F19' ;
            VK_F20 : Command := 'F20' ;
            VK_F21 : Command := 'F21' ;
            VK_F22 : Command := 'F22' ;
            VK_F23 : Command := 'F23' ;
            VK_F24 : Command := 'F24' ;
            VK_NUMLOCK : Command := 'NUMLOCK' ;
            VK_SCROLL : Command := 'SCROLL' ;
            VK_PROCESSKEY : Command := 'PROCESS' ;
            VK_ATTN : Command := 'ATTN' ;
            VK_CRSEL : Command := 'CRSEL' ;
            VK_EXSEL : Command := 'EXSEL' ;
            VK_EREOF : Command := 'EREOF' ;
            VK_PLAY : Command := 'PLAY' ;
            VK_ZOOM : Command := 'ZOOM' ;
            VK_PA1 : Command := 'PA1' ;
            VK_OEM_CLEAR : Command := 'CLEAR' ;
        end ;
        if( length( Command ) = 0 ) then
        begin
            goto Loop ;
        end ;
        case Input_Data.Event.KeyEvent.dwControlKeyState of
            LEFT_ALT_PRESSED, RIGHT_ALT_PRESSED : if( ( length( Command ) > 1 ) and ( copy( Command, 1, 2 ) <> 'KP' ) ) then
                                                  begin
                                                      Command := 'ALT' + Command ;
                                                  end else
                                                  begin
                                                      Command := 'A' + Command ;
                                                  end ;
            LEFT_CTRL_PRESSED, RIGHT_CTRL_PRESSED : if( length( Command ) > 1 ) then
                                                    begin
                                                        Command := 'C' + Command ;
                                                    end ;
        end ;
    end ;
{$ENDIF}

begin
    if( Handle = 0 ) then
    begin
{$IFDEF CONSOLE}
        _Handle := GetStdHandle( STD_INPUT_HANDLE ) ;
        Do_Console ;
        InKey := Command ;
{$ELSE}
        Do_Window ;
{$ENDIF}
    end else
    begin
{$IFDEF FPC_OR_DELPHI}
        if( IsWindow( Handle ) ) then
        begin
            Do_Window ;
        end else
        begin
  {$IFDEF CONSOLE}
            _Handle := Handle ;
            Do_Console ;
            InKey := Command ;
  {$ENDIF}
        end ;
{$ENDIF}
    end ;
end ; { TDefault_Input.InKey }


function TDefault_Input.Pending_Input( Handle : integer ) : boolean ;

var B : boolean ;
    C : Ansichar ;

begin
    B := inherited Pending_Input( Handle ) ; { Handle any buffered input }
    if( B ) then
    begin
        Pending_Input := True ;
        exit ;
    end ;
    C := OS^.Get_Char( False, B, Handle ) ;
    if( B ) then { There was some input }
    begin
        Input_Buffer := Input_Buffer + C ;
        Pending_Input := True ;
        exit ;
    end ;
    Pending_Input := False ;
end ; { TDefault_Input.Pending_Input }



function Input_Line( I : TInput_Source ; X : Integer ) : Small_String ;
{ This function returns a string of input after taking it with or without echo.
  Ending delimeters are returned as well. }

    procedure _Write( const S : Ansistring ) ;

    begin
        {$I-}
        write( S ) ;
        {$I+}
        Clear_IOResult ;
    end ;


    procedure _Writeln( const S : Ansistring ) ;

    begin
        {$I-}
        writeln( S ) ;
        {$I+}
        Clear_IOResult ;
    end ;

var Accumulate : Small_String ;
    Count : Integer ;
    Reg : Ansichar ;
    Reg_Ax : Integer ;

label Loop ;

begin
    Count := 1 ;
Loop:
    Reg := I.Get_Char( 0, 0 ) ;
    Reg_Ax := Ord( Reg ) And 127 ;
    Accumulate[ Count ] := AnsiChar( Reg_Ax ) ;
    Count := Count + 1 ;
    if( ( Reg_Ax = 10 ) or ( Reg_Ax = 13 ) or ( Reg_Ax = 12 ) or ( Reg_Ax = 27 ) ) then
    { LF, CR, FF, ESC are delimiters }
    begin
        if( X <> 0 ) then
        begin
            if( Reg_Ax = 27 ) then
            begin
                _Write( '$' ) ;
            end else
            begin
                _Write( Reg ) ;
            end ;
        end ;
        if( Reg_Ax = 13 ) then
        begin
            Accumulate[ Count ] := LF ;
            Count := Count + 1 ;
            if( X <> 0 ) Then
            begin
                _Write( AnsiChar( 10 ) ) ;
            end ;
        end ;
        if( Reg_Ax = 10 ) then
        begin
            Accumulate[ Count ] := CR ;
            Count := Count + 1 ;
            if( X <> 0 ) then
            begin
                _Write( CR ) ;
            end ;
        end ;
        setlength( Accumulate, Count - 1 ) ;
        Input_Line := Accumulate ;
        exit ;
    end ;
    if( Reg_Ax = 21 ) then { Control-U }
    begin
        Count := 1 ;
        if( X <> 0 ) then
        begin
            _Writeln( '^U' ) ;
        end ;
        goto Loop ;
    end ;
    if( Reg_Ax = 127 ) then { Delete }
    begin
        Count := Count - 2 ;
        if( Count < 1 ) then
        begin
            Count := 1 ;
        end ;
        if( X <> 0 ) then
        begin
            _Write( ' ' ) ;
        end ;
        goto Loop ;
    end ;
    if( Count = 254 ) then
    begin
        _Write( '' ) ; { End of input buffer }
        Count := Count - 1 ;
        goto Loop ;
    end ;
    if( X <> 0 ) then
    begin
        _Write( Reg ) ;
    end ;
    goto Loop ;
end ; { Input_Line }


begin
    Default_Input.Init ;
end.


