{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
        Program Name    : Hexs
        Package Name    : Subroutine Library - Layer 0
        Purpose         : Hexadecimal conversion routines
        Institution     : Conroy & Conroy Co.
        Date written    : 10-Dec-85
        Written by      : Alan Conroy
        Version         : 1.0E

        Copyright (C) 1985-2019 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

      ***************************************************************
      *                                                             *
      *      M O D I F I C A T I O N   H I S T O R Y                *
      *                                                             *
      ***************************************************************

          DATE        BY            REASON

        25-Jul-2019   EAC           Extracted from CVT

      *******************************************************
      *                                                     *
      *          P R O G R A M   P U R P O S E              *
      *                                                     *
      *******************************************************

      This module contains routines for converting hexadecimal data.

      CODE CERTIFIED - LEVEL 1 by EAC on 24-Sep-2005.  Untested on Mobile.
}

unit Hexs ;

interface

uses { Layer 0... }
     _Types ; { int64 }

{$I EDEFINES.INC}

{ A fast hex converter.  Returns an integer whose lower 2 bytes contain the
  ASCII values which are the hex digits for the passed byte value. }
function FHex( Value : byte ) : integer ;

function From_Hex( Value : Ansistring ) : int64 ;

function Hex( Value : byte ) : Ansistring ;

function Hex_String( S : ansistring ) : ansistring ;

{ Indicates whether passed string contains valid digits for Hexadecimal. }
function Valid_Hex( Value : Ansistring ) : Boolean ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses { Layer 0... }
     CommonUt ; { Low_String_Index }

{ Unit internal routines... }

const Hex_Values : Ansistring = '0123456789ABCDEF' ;

{ API.. }

function From_Hex( Value : Ansistring ) : int64 ;

var H, I : integer ;

begin
    // Validate and normalize value...
    Result := 0 ;
    Value := Edit( Value, 16 ) ; // Remove all spaces
    if( not Valid_Hex( Value ) ) then
    begin
        exit ;
    end ;
    H := High_String_Index( Value ) ;
    if( ( H > 0 ) and ( ord( Value[ H ] ) = 0 ) ) then
    begin
        dec( H ) ; // Don't include trailing null
    end ;
    if( odd( H + 1 - Low_String_Index ) ) then
    begin
        Value := '0' + Value ;
        inc( H ) ;
    end ;

    // Convert value...
    while( H > 0 ) do
    begin
        I := pos( Value[ Low_String_Index ], Hex_Values ) - 1 + Low_String_Index ;
        Result := ( Result shl 4 ) or ( I - Low_String_Index ) ;
        I := pos( Value[ Low_String_Index + 1 ], Hex_Values ) - 1 + Low_String_Index ;
        Result := ( Result shl 4 ) or ( I - Low_String_Index ) ;
        Value := copy( Value, 3, length( Value ) ) ;
        H := H - 2 ;
    end ;
end ; { From_Hex }


function Hex( Value : byte ) : Ansistring ;

begin
    Hex := Hex_Values[ ( ( Value shr 4 ) and $F ) + Low_String_Index ] + Hex_Values[ ( Value and $F ) + Low_String_Index ] ;
end ;


function Hex_String( S : ansistring ) : ansistring ;

var I : integer ;
    Res : Ansistring ;

begin
    Res := '' ;
    for I := 1 to length( S ) do
    begin
        Res := Hex( ord( S[ I ] ) ) + Res ;
    end ;
    Hex_String := Res ;
end ;


{ A fast hex converter.  Returns an integer whose lower 2 bytes contain the
  ASCII values which are the hex digits for the passed byte value. }
function FHex( Value : byte ) : integer ;

begin
    FHex := ord( Hex_Values[ ( Value and $F ) + Low_String_Index ] ) +
        ( ord( Hex_Values[ ( ( Value shr 4 ) and $F ) + Low_String_Index ] ) shl 8 ) ;
end ;


const Base_List = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&?()*+~|' ;

function Valid_Hex( Value : Ansistring ) : Boolean ;

var H, I, I1 : Integer ;

begin
    H := High_String_Index( Value ) ;
    if( ( H > 0 ) and ( ord( Value[ H ] ) = 0 ) ) then
    begin
        dec( H ) ; // Don't include trailing null
    end ;
    for I := Low_String_Index to H do
    begin
        I1 := Pos( upcase( Value[ I ] ), Base_List ) ;
        if( ( I1 = 0 ) or ( I1 > 16 ) ) then { Verify that the value is valid }
        begin
            Valid_Hex := false ;
            exit ;
        end ;
    end ;
    Valid_Hex := true ;
end ;


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_FHex ;

var I : integer ;
    S : string[ 2 ] ;

begin
    S := '  ' ;
    I := FHex( 255 ) ;
    move( I, S[ 1 ], 2 ) ;
    if( S <> 'FF' ) then
    begin
        Error ;
    end ;
    I := FHex( 0 ) ;
    move( I, S[ 1 ], 2 ) ;
    if( S <> '00' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
    Test_FHex ;
end ;
{$ENDIF}


end.
