{*******************************************************}
{                                                       }
{         Delphi Volume Control Library v.1             }
{                                                       }
{         Copyright (c) 2003 archie@emapa.pl            }
{                                                       }
{*******************************************************}

unit Mixer ;

interface

uses // Borland...
     Classes, // TList
     MMSystem, // TMixerControl
     SysUtils ;

type TMixer = class
                  private // Instance data...
                      Name : String ;
                      Description : String ;
                      _Min_Volume : Integer ;
                      _Max_Volume : Integer ;
                      Handle : integer ;

                  protected // Property handlers...
                      function Get_Volume : integer ;
                      procedure Set_Volume( Value : integer ) ;

                  public // API...
                      property Volume : integer
                          read Get_Volume
                          write Set_Volume ;
              end ; // TMixer
     TMaster_Mixer = class
                         public // Constructors and destructors...
                             constructor Create ;
                             destructor Destroy ; override ;

                         private // Instance data...
                             List : TList ;

                         protected // Property handlers...
                             function Get_Volume : integer ;
                             procedure Set_Volume( Value : integer ) ;

                         public // API...
                             property Volume : integer
                                 read Get_Volume
                                 write Set_Volume ;
                     end ; // TMaster_Mixer

function Master_Mixer : TMaster_Mixer ;

implementation

var Default_Mixer : TMaster_Mixer = nil ;

function Master_Mixer : TMaster_Mixer ;

begin
    if( Default_Mixer = nil ) then
    begin
        Default_Mixer := TMaster_Mixer.Create ;
    end ;
    Result := Default_Mixer ;
end ;



// TMixer methods...

// Property handlers...

function TMixer.Get_Volume : integer ;

var E : Integer ;
    MC : TMixerControl ;
    MCD : TMixerControlDetails ;
    MCD_LT : TMixerControlDetailsListText ;
    ML : TMixerLine ;
    MLC : TMixerLineControls ;

begin
    // Setup...
    Result := 0 ;
    fillChar( ML, SizeOf( ML ), 0 ) ;
    ML.cbStruct := sizeof( ML ) ;
    ML.dwComponentType := MIXERLINE_COMPONENTTYPE_DST_SPEAKERS ; // Master volume

    FillChar( MLC, SizeOf( MLC ), 0 ) ;
    MLC.cbStruct := sizeof( MLC ) ;
    MLC.dwLineID := ML.dwLineID ;
    MLC.dwControlType := MIXERCONTROL_CONTROLTYPE_VOLUME ;
    MLC.cControls := 1 ;

    FillChar( MC, SizeOf( MC ), 0 ) ;
    MC.cbStruct := sizeof( MC ) ;

    MLC.cbmxctrl := sizeof( MC ) ;
    MLC.pamxctrl := @MC ;

    E := mixerGetLineControls( Handle, @MLC, MIXER_GETLINECONTROLSF_ONEBYTYPE ) ;
    if( E = MMSYSERR_NOERROR ) then
    begin
        _Min_Volume := MC.Bounds.lMinimum ;
        _Max_Volume := MC.Bounds.lMaximum ;

        FillChar( MCD, SizeOf( MCD ), 0 ) ;
        MCD.cbStruct := SizeOf( MCD ) ;
        MCD.dwControlID := MC.dwControlID ;

        FillChar( MCD_LT, SizeOf( MCD_LT ), 0 ) ;

        MCD.cbDetails := SizeOf( MCD_LT ) ;
        MCD.paDetails := @MCD_LT ;
        MCD.cChannels := 1 ;

        mixerGetControlDetails( Handle, @MCD, MIXER_GETCONTROLDETAILSF_VALUE ) ;

        Result := abs( MCD_LT.dwParam1 ) ;
    end ; // if( E = MMSYSERR_NOERROR )
end ; // TMixer.Get_Volume


procedure TMixer.Set_Volume( Value : integer ) ;

var E : Integer ;
    MC : TMixerControl ;
    MCD : TMixerControlDetails ;
    MCD_LT : TMixerControlDetailsListText ;
    ML : TMixerLine ;
    MLC : TMixerLineControls ;

begin
    // Setup...
    fillChar( ML, SizeOf( ML ), 0 ) ;
    ML.cbStruct := sizeof( ML ) ;
    ML.dwComponentType := MIXERLINE_COMPONENTTYPE_DST_SPEAKERS ; // Master volume

    FillChar( MLC, SizeOf( MLC ), 0 ) ;
    MLC.cbStruct := sizeof( MLC ) ;
    MLC.dwLineID := ML.dwLineID ;
    MLC.dwControlType := MIXERCONTROL_CONTROLTYPE_VOLUME ;
    MLC.cControls := 1 ;

    FillChar( MC, SizeOf( MC ), 0 ) ;
    MC.cbStruct := sizeof( MC ) ;

    MLC.cbmxctrl := sizeof( MC ) ;
    MLC.pamxctrl := @MC ;

    E := mixerGetLineControls( Handle, @MLC, MIXER_GETLINECONTROLSF_ONEBYTYPE ) ;
    if( E = MMSYSERR_NOERROR ) then
    begin
        _Min_Volume := MC.Bounds.lMinimum ;
        _Max_Volume := MC.Bounds.lMaximum ;

        FillChar( MCD, SizeOf( MCD ), 0 ) ;
        MCD.cbStruct := SizeOf( MCD ) ;
        MCD.dwControlID := MC.dwControlID ;

        FillChar( MCD_LT, SizeOf( MCD_LT ), 0 ) ;

        MCD.cbDetails := SizeOf( MCD_LT ) ;
        MCD.paDetails := @MCD_LT ;
        MCD.cChannels := 1 ;

        mixerGetControlDetails( Handle, @MCD, MIXER_GETCONTROLDETAILSF_VALUE ) ;

        MCD_LT.dwParam1 := Value ;
        mixerSetControlDetails( Handle, @MCD, MIXER_GETCONTROLDETAILSF_VALUE ) ;
    end ; // if( E = MMSYSERR_NOERROR )
end ; // TMixer.Set_Volume


// TMaster_Mixer methods...

// Constructors and destructors...

constructor TMaster_Mixer.Create ;

var E : Integer ;
    I, M : Integer ;
    Mix : TMixer ;
    ML : TMixerLine ;

begin
    inherited Create ;

    List := TList.Create ;
    for I := 0 to mixerGetNumDevs - 1 do
    begin
        //here goes callbacks (events) instead 0,0 if u want to monitor the peak
        E := mixerOpen(@M, I, 0, 0, 0);
        if( E = MMSYSERR_NOERROR ) then
        begin
            fillChar( ML, SizeOf( ML ), 0 ) ;
            ML.cbStruct := sizeof( ML ) ;
            ML.dwComponentType := MIXERLINE_COMPONENTTYPE_DST_SPEAKERS ; // Master volume

            E := mixerGetLineInfo( M, @ML, MIXER_GETLINEINFOF_COMPONENTTYPE ) ;
            if( E = MMSYSERR_NOERROR ) then
            begin
                Mix := TMixer.Create ;
                List.Add( Mix ) ;
                Mix.Handle := M ;
                Mix.Name := ML.szShortName ;
                Mix.Description := ML.szName ;
            end else
            begin
                mixerClose( M ) ;
            end ;
        end ; // if( E = MMSYSERR_NOERROR )
    end ; // for I := 0 to mixerGetNumDevs - 1
end ; // TMaster_Mixer.Create


destructor TMaster_Mixer.Destroy ;

var Loop : integer ;

begin
    for Loop := 0 to List.Count - 1 do
    begin
        TMixer( List[ Loop ] ).Free ;
        List[ Loop ] := nil ;
    end ;
    List.Free ;
    List := nil ;

    inherited Destroy ;
end ;


// Property handlers...

procedure TMaster_Mixer.Set_Volume( Value : Integer ) ;

var Loop : integer ;

var V : record
            case boolean of
                false : ( L : longint ) ;
                true : ( Left, Right : word ) ;
        end ;

begin
    if( List.Count = 0 ) then // No mixers...
    begin
        V.Left := Value ;
        V.Right := Value ;
        auxSetVolume( 5, V.L ) ;
        exit ;
    end ;

    for Loop := 0 to List.Count - 1 do
    begin
        TMixer( List[ Loop ] ).Volume := Value ;
    end ;
end ;


function TMaster_Mixer.Get_Volume : integer ;

var Count, Loop : integer ;

    V : record
            case boolean of
                false : ( L : longint ) ;
                true : ( Left, Right : word ) ;
        end ;

begin
    Result := 0 ;
    if( List.Count = 0 ) then // No mixers...
    begin
        AuxGetVolume( 5, @V ) ;
        Result := ( V.Left + V.Right ) shr 9 ;
        exit ;
    end ;

    Count := 0 ;
    for Loop := 0 to List.Count - 1 do
    begin
        inc( Count ) ;
        Result := Result + TMixer( List[ Loop ] ).Volume ;
    end ;
    if( Count > 0 ) then
    begin
        Result := Result div Count ;
    end ;
end ;



begin
    Default_Mixer.Free ;
    Default_Mixer := nil ;
end.
