{
        Program Name : OSWinDeb
        Package Name : Subroutine Library - Layer 3
        Purpose      : MS Windows Operating system debugger
        Institution  : Conroy & Conroy Co.
        Date Written : 14-Oct-2004
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2004 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

         24-Feb-2006    RS          Fix abend on iterating fonts.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit OSWinDeb ;

interface

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$I EDEFINES.INC}

uses // Subroutine Library - Layer 0...
     _DebugIn, // TText_Debugger

     // Subroutine Library - Layer 1...
     DebugInt ; // TText_Debugger

type TMS_Windows_Debugger = class( TCOM_Debug_Iterator )
                                public // API...
                                    procedure Build_Children ; override ;
                                    function Has_Children : boolean ; override ;
                            end ;

implementation

uses // VCL...
     Graphics, // TColor
     Windows, // HKEY_LOCAL_MACHINE
     ShellAPI, // SHGetFileInfo
     ShlObj, // SHGetSpecialFolder
     Classes, // TStringList
     CommCtrl, // ImageList_GetIconSize
     Registry, // TRegistry
     SysUtils, // inttostr

     // Subroutine Library - Layer 1...
     CommonUt, // TInteger_List

     // Subroutine Library - Layer 3...
     O_S ; // Swap_File_List

{$IFDEF FPC}
const JOHAB_CHARSET = 130 ;
const MAC_CHARSET = 77 ;
const VIETNAMESE_CHARSET = 163 ;
const LOCALE_IDEFAULTMACCODEPAGE = $1011 ;
const LOCALE_SNATIVEDIGITS       = $13;
const LOCALE_SINTLSYMBOL         = $15;
const LOCALE_IINTLCURRDIGITS     = $1A;
const LOCALE_ITIMEMARKPOSN       = $1005;
const LOCALE_ICENTURY            = $24;
const LOCALE_FONTSIGNATURE       = $58;
const NUMMARKERS                 = 20 ;
{$ENDIF}

{
TODO:
Three Windows functions enable an application to obtain lists of resource types,
names, and languages in a specified module. The EnumResourceTypes function
provides a list of resource types found in the module, the EnumResourceNames
function provides the name of each resource within a given type, and the
EnumResourceLanguages function provides the language of each resource of a given
name and type. These functions and their associated callback functions enable
applications to create a list of all resources in a module. This process is
described in Creating a Resource List.
}
function Registry_Description( Registry : TRegistry ;
    const Title : string ) : string ;

var Data_Type : TRegDataType ;
    Dummy : integer ;
    S, T, Temp : string ;

begin
    Data_Type := Registry.GetDataType( Title ) ;
    case Data_Type of
        rdString : T := 'string' ;
        rdExpandString : T := 'expandstring' ;
        rdInteger : T := 'integer' ;
        rdBinary : T := 'binary' ;
        else T := 'unknown' ;
    end ;
    T := T + ' ' + inttostr( Registry.GetDataSize( Title ) ) + ' bytes' ;
    case Data_Type of
        rdString,
          rdExpandString : S := '"' + Registry.ReadString( Title ) + '"' ;
        rdInteger :
            begin
                Dummy := Registry.ReadInteger( Title ) ;
                case Registry.GetDataSize( Title ) of
                    1 : Dummy := Dummy and $FF ;
                    2 : Dummy := Dummy and $FFFF ;
                    3 : Dummy := Dummy and $FFFFFF ;
                end ;
                S := inttostr( Dummy ) ;
            end ;
        else // Treat everything else as binary
            begin
                Dummy := Registry.GetDataSize( Title ) ;
                if( Data_Type = rdExpandString ) then
                begin
                    setlength( Temp, Dummy ) ;
                    Registry.ReadBinaryData( Title, PChar( Temp )^, Dummy ) ;
                    S := '"' + Temp + '" ' ;
                end else
                begin
                    S := '' ;
                end ;
                if( Dummy > 127 ) then
                begin
                    Dummy := 127 ; // No need to generate hex data for more than this due to later truncation
                end ;
                setlength( Temp, Dummy ) ;
                Registry.ReadBinaryData( Title, PChar( Temp )^, Dummy ) ;
                S := '' ;
                for Dummy := 1 to length( Temp ) do
                begin
                    S := S + inttohex( ord( Temp[ Dummy ] ), 2 ) + ' ' ;
                end ;
            end ;
    end ;
    S := Title + ' : ' + T + ' = ' + S ;
    if( length( S ) > 255 ) then
    begin
        S := copy( S, 1, 252 ) + '...' ;
    end ;
    Result := S ;
end ; // Registry_Description


type TRegistry_Key_Debugger = class( TText_Debugger )
                                  public // Constructors and destructors...
                                      constructor Create ;
                                      destructor Destroy ; override ;

                                  private
                                      Registry : TRegistry ;

                                  public // API...
                                      Key : Cardinal ;
                                      Paths : TStringList ;

                                      function Child( Index : longint ) : PDebug_Interface ;
                                          override ;
                                      function Count : longint ; override ;
                              end ;

// Constructors and destructors...

constructor TRegistry_Key_Debugger.Create ;

begin
    inherited Create ;

    Paths := TStringList.Create ;
end ;


destructor TRegistry_Key_Debugger.Destroy ;

begin
    Registry.Free ;
    Paths.Free ;

    inherited Destroy ;
end ;


function TRegistry_Key_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Loop : integer ;
    S : string ;
    Strings : TStringList ;

begin
    if( Registry = nil ) then
    begin
        Registry := TRegistry.Create ;
        Registry.RootKey := Key ;
        Registry.OpenKeyReadOnly( '' ) ;
        for Loop := 0 to Paths.Count - 1 do
        begin
            Registry.OpenKeyReadOnly( Paths[ Loop ] ) ;
        end ;
    end ;

    Result := TText_Debugger.Create ;
    Strings := TStringList.Create ;
    Registry.GetValueNames( Strings ) ;
    if( Index < Strings.Count ) then
    begin
        S := Registry_Description( Registry, Strings[ Index ] ) ;
    end else
    begin
        Index := Index - Strings.Count ;
        Strings.Free ;
        Strings := TStringList.Create ;
        Registry.GetKeyNames( Strings ) ;
        S := Strings[ Index ] ;
        Strings.Free ;
        Result.Free ;
        Result := TRegistry_Key_Debugger.Create ;
        TRegistry_Key_Debugger( Result ).Key := Key ;
        for Loop := 0 to Paths.Count - 1 do
        begin
            TRegistry_Key_Debugger( Result ).Paths.Add( Paths[ Loop ] ) ;
        end ;
        TRegistry_Key_Debugger( Result ).Paths.Add( S ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TRegistry_Key_Debugger.Count : longint ;

var C : integer ;
    Loop : integer ;
    Strings : TStringList ;

begin
    if( Registry = nil ) then
    begin
        Registry := TRegistry.Create ;
        Registry.RootKey := Key ;
        Registry.OpenKeyReadOnly( '' ) ;
        for Loop := 0 to Paths.Count - 1 do
        begin
            Registry.OpenKeyReadOnly( Paths[ Loop ] ) ;
        end ;
    end ;

    Strings := TStringList.Create ;
    Registry.GetValueNames( Strings ) ;
    C := Strings.Count ;
    Strings.Free ;
    Strings := TStringList.Create ;
    Registry.GetKeyNames( Strings ) ;
    C := C + Strings.Count ;
    Strings.Free ;
    Count := C ;
end ;



type TRegistry_Root_Key_Debugger = class( TText_Debugger )
                              public // Constructors and destructors...
                                  destructor Destroy ; override ;

                              private
                                  Registry : TRegistry ;

                              public // API...
                                  Key : Cardinal ;

                                  function Child( Index : longint ) : PDebug_Interface ;
                                      override ;
                                  function Count : longint ; override ;
                          end ;

// Constructors and destructors...

destructor TRegistry_Root_Key_Debugger.Destroy ;

begin
    Registry.Free ;

    inherited Destroy ;
end ;


function TRegistry_Root_Key_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;
    Strings : TStringList ;

begin
    if( Registry = nil ) then
    begin
        Registry := TRegistry.Create ;
        Registry.RootKey := Key ;
        Registry.OpenKeyReadOnly( '' ) ;
    end ;

    Result := TText_Debugger.Create ;
    Strings := TStringList.Create ;
    Registry.GetValueNames( Strings ) ;
    if( Index < Strings.Count ) then
    begin
        S := Registry_Description( Registry, Strings[ Index ] ) ;
    end else
    begin
        Index := Index - Strings.Count ;
        Strings.Free ;
        Strings := TStringList.Create ;
        Registry.GetKeyNames( Strings ) ;
        S := Strings[ Index ] ;
        Strings.Free ;
        Result.Free ;
        Result := TRegistry_Key_Debugger.Create ;
        TRegistry_Key_Debugger( Result ).Key := Key ;
        TRegistry_Key_Debugger( Result ).Paths.Add( S ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TRegistry_Root_Key_Debugger.Count : longint ;

var C : integer ;
    Strings : TStringList ;

begin
    if( Registry = nil ) then
    begin
        Registry := TRegistry.Create ;
        Registry.RootKey := Key ;
        Registry.OpenKeyReadOnly( '' ) ;
    end ;

    Strings := TStringList.Create ;
    Registry.GetValueNames( Strings ) ;
    C := Strings.Count ;
    Strings.Free ;
    Strings := TStringList.Create ;
    Registry.GetKeyNames( Strings ) ;
    C := C + Strings.Count ;
    Strings.Free ;
    Count := C ;
end ;



type TSystem_Parameters_Debugger = class( TText_Debugger )
                                    public // API...
                                        procedure Build_Children ; override ;
                                        function Has_Children : boolean ; override ;
                                end ;

procedure TSystem_Parameters_Debugger.Build_Children ;

    procedure Add_Line( S, S1 : string ) ;

    var Res : TText_Debugger ;

    begin
        Res := TText_Debugger.Create ;
        Res.Title := PChar( S + ' = ' + S1 ) ;
        _Children.AddObject( S, Res ) ;
    end ;

var I : integer ;

begin
    // Accessibility parameters...
//SPI_GETAUDIODESCRIPTION
//SPI_SETACCESSTIMEOUT
    I := 0 ;
{
    if( SystemParametersInfo( SPI_GETCLIENTAREAANIMATION, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Client Area Animation', Boolean_To_String( boolean( I ) ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETDISABLEOVERLAPPEDCONTENT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Disable overlapped content', Boolean_To_String( boolean( I ) ) ) ;
    end ;
}
    if( SystemParametersInfo( SPI_GETFOCUSBORDERHEIGHT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Focus border height', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETFOCUSBORDERWIDTH, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Focus border width', inttostr( I ) ) ;
    end ;
{
    if( SystemParametersInfo( SPI_GETLOGICALDPIOVERRIDE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Logical DPI override', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMESSAGEDURATION, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Message duration', inttostr( I ) ) ;
    end ;
}
    I := 0 ;
    if( SystemParametersInfo( SPI_GETMOUSECLICKLOCK, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse Click Lock', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMOUSECLICKLOCKTIME, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse Click Lock Time', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMOUSESONAR, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse Sonar', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETMOUSEVANISH, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse Sonar', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSHOWSOUNDS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Show Sounds', Boolean_To_String( I ) ) ;
    end ;

    // Desktop parameters...
//SPI_GETDESKWALLPAPER
    I := 0 ;
{
    if( SystemParametersInfo( SPI_GETCLEARTYPE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'ClearType', Boolean_To_String( I ) ) ;
    end ;
}
    I := 0 ;
    if( SystemParametersInfo( SPI_GETDROPSHADOW, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Drop Shadow', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETFLATMENU, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Flat menus', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETFONTSMOOTHINGCONTRAST, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Font Smoothing Contrast', inttostr( I ) ) ;
    end ;
{
    if( SystemParametersInfo( SPI_GETFONTSMOOTHINGORIENTATION, 0, @I, 0 ) ) then
    begin
        case I of
            FE_FONTSMOOTHINGORIENTATIONBGR : Add_Line( 'Font Smoothing Orientation', 'BGR' ) ;
            FE_FONTSMOOTHINGORIENTATIONRGB : Add_Line( 'Font Smoothing Orientation', 'RGB' ) ;
        end ;
    end ;
}
    if( SystemParametersInfo( SPI_GETFONTSMOOTHINGTYPE, 0, @I, 0 ) ) then
    begin
        case I of
            FE_FONTSMOOTHINGSTANDARD : Add_Line( 'Font Smoothing Type', 'Standard' ) ;
            FE_FONTSMOOTHINGCLEARTYPE : Add_Line( 'Font Smoothing Type', 'ClearType' ) ;
        end ;
    end ;

    // Icon parameters...
    if( SystemParametersInfo( SPI_ICONHORIZONTALSPACING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Icon Horizontal Spacing', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_ICONVERTICALSPACING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Icon Vertical Spacing', inttostr( I ) ) ;
    end ;

    // Input parameters...
{
    I := 0 ;
    if( SystemParametersInfo( SPI_GETBLOCKSENDINPUTRESETS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Block Send Input Resets', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETCONTACTVISUALIZATION, 0, @I, 0 ) ) then
    begin
        case I of
            0 : Add_Line( 'Contact Visualization', 'Off' ) ;
            1 : Add_Line( 'Contact Visualization', 'On' ) ;
            2 : Add_Line( 'Contact Visualization', 'Presentation mode' ) ;
        end ;
    end ;
    if( SystemParametersInfo( SPI_GETGESTUREVISUALIZATION, 0, @I, 0 ) ) then
    begin
        case I of
            0 : Add_Line( 'Gesture Visualization', 'Off' ) ;
            1 : Add_Line( 'Gesture Visualization', 'Tap' ) ;
            2 : Add_Line( 'Gesture Visualization', 'Double tap' ) ;
            4 : Add_Line( 'Gesture Visualization', 'Press and tap' ) ;
            8 : Add_Line( 'Gesture Visualization', 'Press and hold' ) ;
            $10 : Add_Line( 'Gesture Visualization', 'Right tap' ) ;
            $1F : Add_Line( 'Gesture Visualization', 'On' ) ;
        end ;
    end ;
}
    I := 0 ;
    if( SystemParametersInfo( SPI_GETKEYBOARDCUES, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Keyboard Cues', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMOUSESPEED, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse speed', inttostr( I ) ) ;
    end ;
{
    if( SystemParametersInfo( SPI_GETMOUSEWHEELROUTING, 0, @I, 0 ) ) then
    begin
        case I of
            0 : Add_Line( 'Mouse wheel routine', 'Focus' ) ;
            1 : Add_Line( 'Mouse wheel routine', 'Hybrid' ) ;
        end ;
    end ;
    if( SystemParametersInfo( SPI_GETPENVISUALIZATION, 0, @I, 0 ) ) then
    begin
        case I of
            0 : Add_Line( 'Pen visualization', 'Off' ) ;
            1 : Add_Line( 'Pen visualization', 'Tap' ) ;
            2 : Add_Line( 'Pen visualization', 'Double tap' ) ;
            $20 : Add_Line( 'Pen visualization', 'Cursor' ) ;
            $23 : Add_Line( 'Pen visualization', 'On' ) ;
        end ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSYSTEMLANGUAGEBAR, 0, @I, 0 ) ) then
    begin
        Add_Line( 'System Language Var', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETTHREADLOCALINPUTSETTINGS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Thread Local Input Settings', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETWHEELSCROLLCHARS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Wheel Scroll Chars', inttostr( I ) ) ;
    end ;
}

    // Menu parameters...
    I := 0 ;
    if( SystemParametersInfo( SPI_GETMENUFADE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Menu Fade', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMENUSHOWDELAY, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Menu Show Delay', inttostr( I ) ) ;
    end ;

    // Power parameters...
    I := 0 ;
    if( SystemParametersInfo( SPI_GETLOWPOWERACTIVE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Low power active', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETLOWPOWERTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Low power timeout', inttostr( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETPOWEROFFACTIVE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Power off active', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETPOWEROFFTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Power off timeout', inttostr( I ) ) ;
    end ;

    // Screen-saver parameters...
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSCREENSAVERRUNNING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Screen saver running', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
{
    if( SystemParametersInfo( SPI_GETSCREENSAVESECURE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Screen saver requires password', Boolean_To_String( I ) ) ;
    end ;
}

    // Timeouts...
{
    if( SystemParametersInfo( SPI_GETHUNGAPPTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Hung app timeout', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETWAITTOKILLTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Wait to kill timeout, inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETWAITTOKILLSERVICETIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Wait to kill service timeout, inttostr( I ) ) ;
    end ;
}

    // UI effects....
    I := 0 ;
    if( SystemParametersInfo( SPI_GETCOMBOBOXANIMATION, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Combobox Animation', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETCURSORSHADOW, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Cursor Shadow', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETGRADIENTCAPTIONS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Gradient Captions', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETHOTTRACKING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Hot Tracking', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETLISTBOXSMOOTHSCROLLING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Listbox smooth scrolling', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETMENUANIMATION, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Menu animation', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSELECTIONFADE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Selection fade', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETTOOLTIPANIMATION, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Tool tip animation', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETTOOLTIPFADE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Tool tip fade', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETUIEFFECTS, 0, @I, 0 ) ) then
    begin
        Add_Line( 'UI Effects', Boolean_To_String( I ) ) ;
    end ;

    // Windows parameters...
    I := 0 ;
    if( SystemParametersInfo( SPI_GETACTIVEWINDOWTRACKING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Active window tracking', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETACTIVEWNDTRKZORDER, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Active window tracking Z-order', Boolean_To_String( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETACTIVEWNDTRKTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Active window tracking delay', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETCARETWIDTH, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Caret width', inttostr( I ) ) ;
    end ;
    I := 0 ;
{
    if( SystemParametersInfo( SPI_GETDOCKMOVING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Dock moving', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETDRAGFROMMAXIMIZE, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Drag from maximize', Boolean_To_String( I ) ) ;
    end ;
}
    if( SystemParametersInfo( SPI_GETFOREGROUNDFLASHCOUNT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Foreground flash count', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETFOREGROUNDLOCKTIMEOUT, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Foreground lock timeout', inttostr( I ) ) ;
    end ;
{
    if( SystemParametersInfo( SPI_GETMOUSEDOCKTHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse dock threshold', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMOUSEDRAGOUTTHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse drag out threshold', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETMOUSESIDEMOVETHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Mouse side move threshold', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETPENDOCKTHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Pen dock threshold', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETPENDRAGOUTTHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Pen drag out threshold', inttostr( I ) ) ;
    end ;
    if( SystemParametersInfo( SPI_GETPENSIDEMOVETHRESHOLD, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Pen side move threshold', inttostr( I ) ) ;
    end ;
}
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSHOWIMEUI, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Show IME UI', Boolean_To_String( I ) ) ;
    end ;
{
    I := 0 ;
    if( SystemParametersInfo( SPI_GETSNAPSIZING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Snap sizing', Boolean_To_String( I ) ) ;
    end ;
    I := 0 ;
    if( SystemParametersInfo( SPI_GETWINARRANGING, 0, @I, 0 ) ) then
    begin
        Add_Line( 'Window arrangement', Boolean_To_String( I ) ) ;
    end ;
}
end ;


function TSystem_Parameters_Debugger.Has_Children : boolean ;

begin
    Result := True ;
end ;



type TSystem_Metrics_Debugger = class( TText_Debugger )
                                    public // API...
                                        function Child( Index : longint ) : PDebug_Interface ;
                                            override ;
                                        function Count : longint ; override ;
                                end ;

function TSystem_Metrics_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Dummy : integer ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := 'ARRANGE = ' + inttostr( GetSystemMetrics( SM_ARRANGE ) ) ;
            end ;
        1 : begin
                Dummy := GetSystemMetrics( SM_CLEANBOOT ) ;
                S := 'CLEANBOOT = ' + inttostr( Dummy ) ;
                case Dummy of
                    0 : S := S + ' (Normal boot)' ;
                    1 : S := S + ' (Fail-safe boot)' ;
                    2 : S := S + ' (Fail-safe with network)' ;
                end ;
            end ;
        2 : begin
                Dummy := GetSystemMetrics( SM_CMOUSEBUTTONS ) ;
                S := 'CMOUSEBUTTONS = ' + inttostr( Dummy ) ;
                if( Dummy = 0 ) then
                begin
                    S := S + ' (no mouse installed)' ;
                end ;
            end ;
        3 : begin
                S := 'CXBORDER = ' + inttostr( GetSystemMetrics( SM_CXBORDER ) ) ;
            end ;
        4 : begin
                S := 'CYBORDER = ' + inttostr( GetSystemMetrics( SM_CYBORDER ) ) ;
            end ;
        5 : begin
                S := 'CXCURSOR = ' + inttostr( GetSystemMetrics( SM_CXCURSOR ) ) ;
            end ;
        6 : begin
                S := 'CYCURSOR = ' + inttostr( GetSystemMetrics( SM_CYCURSOR ) ) ;
            end ;
        7 : begin
                S := 'CXDLGFRAME = ' + inttostr( GetSystemMetrics( SM_CXDLGFRAME ) ) ;
            end ;
        8 : begin
                S := 'CYDLGFRAME = ' + inttostr( GetSystemMetrics( SM_CYDLGFRAME ) ) ;
            end ;
        9 : begin
                S := 'CXDOUBLECLK = ' + inttostr( GetSystemMetrics( SM_CXDOUBLECLK ) ) ;
            end ;
        10 : begin
                 S := 'CYDOUBLECLK = ' + inttostr( GetSystemMetrics( SM_CYDOUBLECLK ) ) ;
             end ;
        11 : begin
                 S := 'CXDRAG = ' + inttostr( GetSystemMetrics( SM_CXDRAG ) ) ;
             end ;
        12 : begin
                 S := 'CYDRAG = ' + inttostr( GetSystemMetrics( SM_CYDRAG ) ) ;
             end ;
        13 : begin
                 S := 'CXEDGE = ' + inttostr( GetSystemMetrics( SM_CXEDGE ) ) ;
             end ;
        14 : begin
                 S := 'CYEDGE = ' + inttostr( GetSystemMetrics( SM_CYEDGE ) ) ;
             end ;
        15 : begin
                 S := 'CXFIXEDFRAME = ' + inttostr( GetSystemMetrics( SM_CXFIXEDFRAME ) ) ;
             end ;
        16 : begin
                 S := 'CYFIXEDFRAME = ' + inttostr( GetSystemMetrics( SM_CYFIXEDFRAME ) ) ;
             end ;
        17 : begin
                 S := 'CXFRAME = ' + inttostr( GetSystemMetrics( SM_CXFRAME ) ) ;
             end ;
        18 : begin
                 S := 'CYFRAME = ' + inttostr( GetSystemMetrics( SM_CYFRAME ) ) ;
             end ;
        19 : begin
                 S := 'CXFULLSCREEN = ' + inttostr( GetSystemMetrics( SM_CXFULLSCREEN ) ) ;
             end ;
        20 : begin
                 S := 'CYFULLSCREEN = ' + inttostr( GetSystemMetrics( SM_CYFULLSCREEN ) ) ;
             end ;
        21 : begin
                 S := 'CXHSCROLL = ' + inttostr( GetSystemMetrics( SM_CXHSCROLL ) ) ;
             end ;
        22 : begin
                 S := 'CYHSCROLL = ' + inttostr( GetSystemMetrics( SM_CYHSCROLL ) ) ;
             end ;
        23 : begin
                 S := 'CXHTHUMB = ' + inttostr( GetSystemMetrics( SM_CXHTHUMB ) ) ;
             end ;
        24 : begin
                 S := 'CXICON = ' + inttostr( GetSystemMetrics( SM_CXICON ) ) ;
             end ;
        25 : begin
                 S := 'CYICON = ' + inttostr( GetSystemMetrics( SM_CYICON ) ) ;
             end ;
        26 : begin
                 S := 'CXICONSPACING = ' + inttostr( GetSystemMetrics( SM_CXICONSPACING ) ) ;
             end ;
        27 : begin
                 S := 'CYICONSPACING = ' + inttostr( GetSystemMetrics( SM_CYICONSPACING ) ) ;
             end ;
        28 : begin
                 S := 'CXMAXIMIZED = ' + inttostr( GetSystemMetrics( SM_CXMAXIMIZED ) ) ;
             end ;
        29 : begin
                 S := 'CYMAXIMIZED = ' + inttostr( GetSystemMetrics( SM_CYMAXIMIZED ) ) ;
             end ;
        30 : begin
                 S := 'CXMAXTRACK = ' + inttostr( GetSystemMetrics( SM_CXMAXTRACK ) ) ;
             end ;
        31 : begin
                 S := 'CYMAXTRACK = ' + inttostr( GetSystemMetrics( SM_CYMAXTRACK ) ) ;
             end ;
        32 : begin
                 S := 'CXMENUCHECK = ' + inttostr( GetSystemMetrics( SM_CXMENUCHECK ) ) ;
             end ;
        33 : begin
                 S := 'CYMENUCHECK = ' + inttostr( GetSystemMetrics( SM_CYMENUCHECK ) ) ;
             end ;
        34 : begin
                 S := 'CXMENUSIZE = ' + inttostr( GetSystemMetrics( SM_CXMENUSIZE ) ) ;
             end ;
        35 : begin
                 S := 'CYMENUSIZE = ' + inttostr( GetSystemMetrics( SM_CYMENUSIZE ) ) ;
             end ;
        36 : begin
                 S := 'CXMIN = ' + inttostr( GetSystemMetrics( SM_CXMIN ) ) ;
             end ;
        37 : begin
                 S := 'CYMIN = ' + inttostr( GetSystemMetrics( SM_CYMIN ) ) ;
             end ;
        38 : begin
                 S := 'CXMINIMIZED = ' + inttostr( GetSystemMetrics( SM_CXMINIMIZED ) ) ;
             end ;
        39 : begin
                 S := 'CYMINIMIZED = ' + inttostr( GetSystemMetrics( SM_CYMINIMIZED ) ) ;
             end ;
        40 : begin
                 S := 'CXMINSPACING = ' + inttostr( GetSystemMetrics( SM_CXMINSPACING ) ) ;
             end ;
        41 : begin
                 S := 'CYMINSPACING = ' + inttostr( GetSystemMetrics( SM_CYMINSPACING ) ) ;
             end ;
        42 : begin
                 S := 'CXMINTRACK = ' + inttostr( GetSystemMetrics( SM_CXMINTRACK ) ) ;
             end ;
        43 : begin
                 S := 'CYMINTRACK = ' + inttostr( GetSystemMetrics( SM_CYMINTRACK ) ) ;
             end ;
        44 : begin
                 S := 'CXSCREEN = ' + inttostr( GetSystemMetrics( SM_CXSCREEN ) ) ;
             end ;
        45 : begin
                 S := 'CYSCREEN = ' + inttostr( GetSystemMetrics( SM_CYSCREEN ) ) ;
             end ;
        46 : begin
                 S := 'CXSIZE = ' + inttostr( GetSystemMetrics( SM_CXSIZE ) ) ;
             end ;
        47 : begin
                 S := 'CYSIZE = ' + inttostr( GetSystemMetrics( SM_CYSIZE ) ) ;
             end ;
        48 : begin
                 S := 'CXSIZEFRAME = ' + inttostr( GetSystemMetrics( SM_CXSIZEFRAME ) ) ;
             end ;
        49 : begin
                 S := 'CYSIZEFRAME = ' + inttostr( GetSystemMetrics( SM_CYSIZEFRAME ) ) ;
             end ;
        50 : begin
                 S := 'CXSMICON = ' + inttostr( GetSystemMetrics( SM_CXSMICON ) ) ;
             end ;
        51 : begin
                 S := 'CYSMICON = ' + inttostr( GetSystemMetrics( SM_CYSMICON ) ) ;
             end ;
        52 : begin
                 S := 'CXSMSIZE = ' + inttostr( GetSystemMetrics( SM_CXSMSIZE ) ) ;
             end ;
        53 : begin
                 S := 'CYSMSIZE = ' + inttostr( GetSystemMetrics( SM_CYSMSIZE ) ) ;
             end ;
        54 : begin
                 S := 'CXVSCROLL = ' + inttostr( GetSystemMetrics( SM_CXVSCROLL ) ) ;
             end ;
        55 : begin
                 S := 'CYVSCROLL = ' + inttostr( GetSystemMetrics( SM_CYVSCROLL ) ) ;
             end ;
        56 : begin
                 S := 'CYCAPTION = ' + inttostr( GetSystemMetrics( SM_CYCAPTION ) ) ;
             end ;
        57 : begin
                 S := 'CYKANJIWINDOW = ' + inttostr( GetSystemMetrics( SM_CYKANJIWINDOW ) ) ;
             end ;
        58 : begin
                 S := 'CYMENU = ' + inttostr( GetSystemMetrics( SM_CYMENU ) ) ;
             end ;
        59 : begin
                 S := 'CYSMCAPTION = ' + inttostr( GetSystemMetrics( SM_CYSMCAPTION ) ) ;
             end ;
        60 : begin
                 S := 'CYVTHUMB = ' + inttostr( GetSystemMetrics( SM_CYVTHUMB ) ) ;
             end ;
        61 : begin
                 S := 'DBCSENABLED = ' + inttostr( GetSystemMetrics( SM_DBCSENABLED ) ) ;
             end ;
        62 : begin
                 S := 'DEBUG = ' + inttostr( GetSystemMetrics( SM_DEBUG ) ) ;
             end ;
        63 : begin
                 S := 'MENUDROPALIGNMENT = ' + inttostr( GetSystemMetrics( SM_MENUDROPALIGNMENT ) ) ;
             end ;
        64 : begin
                 S := 'MIDEASTENABLED = ' + inttostr( GetSystemMetrics( SM_MIDEASTENABLED ) ) ;
             end ;
        65 : begin
                 S := 'MOUSEPRESENT = ' + inttostr( GetSystemMetrics( SM_MOUSEPRESENT ) ) ;
             end ;
        66 : begin
                 S := 'MOUSEWHEELPRESENT = ' + inttostr( GetSystemMetrics( SM_MOUSEWHEELPRESENT ) ) ;
             end ;
        67 : begin
                 S := 'NETWORK = ' + inttostr( GetSystemMetrics( SM_NETWORK ) ) ;
             end ;
        68 : begin
                 S := 'PENWINDOWS = ' + inttostr( GetSystemMetrics( SM_PENWINDOWS ) ) ;
             end ;
        69 : begin
                 S := 'SECURE = ' + inttostr( GetSystemMetrics( SM_SECURE ) ) ;
             end ;
        70 : begin
                 S := 'SHOWSOUNDS = ' + inttostr( GetSystemMetrics( SM_SHOWSOUNDS ) ) ;
             end ;
        71 : begin
                 S := 'SLOWMACHINE = ' + inttostr( GetSystemMetrics( SM_SLOWMACHINE ) ) ;
             end ;
        72 : begin
                 S := 'SWAPBUTTON = ' + inttostr( GetSystemMetrics( SM_SWAPBUTTON ) ) ;
             end ;
     end ;
     Result.Title := PChar( S ) ;
end ;


function TSystem_Metrics_Debugger.Count : longint ;

begin
   Count := 73 ;
end ;


type TRegistry_Debugger = class( TText_Debugger )
                              public // API...
                                  function Child( Index : longint ) : PDebug_Interface ;
                                      override ;
                                  function Count : longint ; override ;
                          end ;

function TRegistry_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TRegistry_Root_Key_Debugger.Create ;
    case Index of
        0 : begin
                TRegistry_Root_Key_Debugger( Result ).Key := HKEY_LOCAL_MACHINE ;
                S := 'HKEY_LOCAL_MACHINE' ;
            end ;
        1 : begin
                TRegistry_Root_Key_Debugger( Result ).Key := HKEY_USERS ;
                S := 'HKEY_USERS' ;
            end ;
     end ;
     Result.Title := PChar( S ) ;
end ;


function TRegistry_Debugger.Count : longint ;

begin
    Count := 2 ;
end ;


function Add_Flag( Value, Mask : int64 ; const S, Name : string ) : string ;

begin
    Result := S ;
    if( ( Value and Mask ) = Mask ) then
    begin
        if( length( S ) > 0 ) then
        begin
            Result := Result + ', ' ;
        end ;
        Result := Result + Name ;
    end ;
end ;


function RasterCaps_To_String( Caps : integer ) : string ;

begin
    Result := Add_Flag( Caps, RC_BANDING, '', 'requires banding' ) ;
    Result := Add_Flag( Caps, RC_BITBLT, Result, 'bitblt' ) ;
    Result := Add_Flag( Caps, RC_BITMAP64, Result, 'bitmaps > 64K' ) ;
    Result := Add_Flag( Caps, RC_DI_BITMAP, Result, 'SetDIBits/GetDIBits' ) ;
    Result := Add_Flag( Caps, RC_DIBTODEV, Result, 'SetDIBitsToDevice' ) ;
    Result := Add_Flag( Caps, RC_FLOODFILL, Result, 'flood fills' ) ;
    Result := Add_Flag( Caps, RC_GDI20_OUTPUT, Result, 'Windows 2.0 GDI' ) ;
    Result := Add_Flag( Caps, RC_PALETTE, Result, 'palette-based' ) ;
    Result := Add_Flag( Caps, RC_SCALING, Result, 'scaling' ) ;
    Result := Add_Flag( Caps, RC_STRETCHBLT, Result, 'StretchBlt' ) ;
    Result := Add_Flag( Caps, RC_STRETCHDIB, Result, 'StretchDIBits' ) ;
end ;


function CurveCaps_To_String( caps : integer ) : string ;

begin
    if( Caps = CC_NONE ) then
    begin
        Result := 'curves not supported' ;
        exit ;
    end ;
    Result := Add_Flag( Caps, CC_CIRCLES, '', 'circles' ) ;
    Result := Add_Flag( Caps, CC_PIE, Result, 'pie wedges' ) ;
    Result := Add_Flag( Caps, CC_CHORD, Result, 'chord arcs' ) ;
    Result := Add_Flag( Caps, CC_ELLIPSES, Result, 'ellipses' ) ;
    Result := Add_Flag( Caps, CC_WIDE, Result, 'wide borders' ) ;
    Result := Add_Flag( Caps, CC_STYLED, Result, 'styled borders' ) ;
    Result := Add_Flag( Caps, CC_WIDESTYLED, Result, 'wide and styled borders' ) ;
    Result := Add_Flag( Caps, CC_INTERIORS, Result, 'interiors' ) ;
    Result := Add_Flag( Caps, CC_ROUNDRECT, Result, 'rounded rectangles' ) ;
end ;


function LineCaps_To_String( Caps : integer ) : string ;

begin
    if( Caps = LC_NONE ) then
    begin
        Result := 'lines not supported' ;
        exit ;
    end ;
    Result := Add_Flag( Caps, LC_POLYLINE, '', 'polyline' ) ;
    Result := Add_Flag( Caps, LC_MARKER, Result, 'marker' ) ;
    Result := Add_Flag( Caps, LC_POLYMARKER, Result, 'multiple markers' ) ;
    Result := Add_Flag( Caps, LC_WIDE, Result, 'wide lines' ) ;
    Result := Add_Flag( Caps, LC_STYLED, Result, 'styled lines' ) ;
    Result := Add_Flag( Caps, LC_WIDESTYLED, Result, 'wide and styled lines' ) ;
    Result := Add_Flag( Caps, LC_INTERIORS, Result, 'interiors' ) ;
end ;


function PolygonCaps_To_String( Caps : integer ) : string ;

begin
    if( Caps = PC_NONE ) then
    begin
        Result := 'polygons not supported' ;
        exit ;
    end ;
    Result := Add_Flag( Caps, PC_POLYGON, '', 'alternate-fill polygons' ) ;
    Result := Add_Flag( Caps, PC_RECTANGLE, Result, 'rectangles' ) ;
    Result := Add_Flag( Caps, PC_WINDPOLYGON, Result, 'winding-fill polygons' ) ;
    Result := Add_Flag( Caps, PC_SCANLINE, Result, 'single scanline' ) ;
    Result := Add_Flag( Caps, PC_WIDE, Result, 'wide borders' ) ;
    Result := Add_Flag( Caps, PC_STYLED, Result, 'styled borders' ) ;
    Result := Add_Flag( Caps, PC_WIDESTYLED, Result, 'wide and styled borders' ) ;
    Result := Add_Flag( Caps, PC_INTERIORS, Result, 'interiors' ) ;
end ;


function TextCaps_To_String( Caps : integer ) : string ;

begin
    Result := Add_Flag( Caps, TC_OP_CHARACTER, '', 'character output precision' ) ;
    Result := Add_Flag( Caps, TC_OP_STROKE, Result, 'stroke output precision' ) ;
    Result := Add_Flag( Caps, TC_CP_STROKE, Result, 'stroke clip precision' ) ;
    Result := Add_Flag( Caps, TC_CR_90, Result, '90-degree character rotation' ) ;
    Result := Add_Flag( Caps, TC_CR_ANY, Result, 'any character rotation' ) ;
    Result := Add_Flag( Caps, TC_SF_X_YINDEP, Result, 'independant scaling' ) ;
    Result := Add_Flag( Caps, TC_SA_DOUBLE, Result, 'doubled character for scaling' ) ;
    Result := Add_Flag( Caps, TC_SA_INTEGER, Result, 'integer multiples only for character scaling' ) ;
    Result := Add_Flag( Caps, TC_SA_CONTIN, Result, 'any multiples for exact character scaling' ) ;
    Result := Add_Flag( Caps, TC_EA_DOUBLE, Result, 'double-weight characters' ) ;
    Result := Add_Flag( Caps, TC_IA_ABLE, Result, 'italicize' ) ;
    Result := Add_Flag( Caps, TC_UA_ABLE, Result, 'underline' ) ;
    Result := Add_Flag( Caps, TC_SO_ABLE, Result, 'strikeouts' ) ;
    Result := Add_Flag( Caps, TC_RA_ABLE, Result, 'raster fonts' ) ;
    Result := Add_Flag( Caps, TC_VA_ABLE, Result, 'vector fonts' ) ;
    Result := Add_Flag( Caps, TC_SCROLLBLT, Result, 'bit-block scrolling' ) ;
end ;


type TDisplay_Debugger = class( TText_Debugger )
                              public // API...
                                  function Child( Index : longint ) : PDebug_Interface ;
                                      override ;
                                  function Count : longint ; override ;
                          end ;

function TDisplay_Debugger.Child( Index : longint ) : PDebug_Interface ;

var DC : HDC ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    DC := CreateDC( 'DISPLAY', 'DISPLAY', nil, nil ) ;
    try
        case Index of
            0 : S := 'DRIVERVERSION = ' + inttostr( GetDeviceCaps( DC, DRIVERVERSION ) ) ;
            1 : case GetDeviceCaps( DC, TECHNOLOGY ) of
                    DT_PLOTTER : S := 'TECHNOLOGY = Vector plotter' ;
                    DT_RASDISPLAY : S := 'TECHNOLOGY = Raster display' ;
                    DT_RASPRINTER : S := 'TECHNOLOGY = Raster printer' ;
                    DT_RASCAMERA : S := 'TECHNOLOGY = Raster camera' ;
                    DT_CHARSTREAM : S := 'TECHNOLOGY = Character stream' ;
                    DT_METAFILE : S := 'TECHNOLOGY = Metafile' ;
                    DT_DISPFILE : S := 'TECHNOLOGY = Display file' ;
                    else S := 'TECHNOLOGY = ' + inttostr( GetDeviceCaps( DC, TECHNOLOGY ) ) ;
                end ;
            2 : S := 'HORZSIZE = ' + inttostr( GetDeviceCaps( DC, HORZSIZE ) ) ;
            3 : S := 'VERTSIZE = ' + inttostr( GetDeviceCaps( DC, VERTSIZE ) ) ;
            4 : S := 'HORZRES = ' + inttostr( GetDeviceCaps( DC, HORZRES ) ) ;
            5 : S := 'VERTRES = ' + inttostr( GetDeviceCaps( DC, VERTRES ) ) ;
            6 : S := 'LOGPIXELSX = ' + inttostr( GetDeviceCaps( DC, LOGPIXELSX ) ) ;
            7 : S := 'LOGPIXELSY = ' + inttostr( GetDeviceCaps( DC, LOGPIXELSY ) ) ;
            8 : S := 'BITSPIXEL = ' + inttostr( GetDeviceCaps( DC, BITSPIXEL ) ) ;
            9 : S := 'PLANES = ' + inttostr( GetDeviceCaps( DC, PLANES ) ) ;
            10 : S := 'NUMBRUSHES = ' + inttostr( GetDeviceCaps( DC, NUMBRUSHES ) ) ;
            11 : S := 'NUMPENS = ' + inttostr( GetDeviceCaps( DC, NUMPENS ) ) ;
            12 : S := 'NUMFONTS = ' + inttostr( GetDeviceCaps( DC, NUMFONTS ) ) ;
            13 : S := 'NUMCOLORS = ' + inttostr( GetDeviceCaps( DC, NUMCOLORS ) ) ;
            14 : S := 'ASPECTX = ' + inttostr( GetDeviceCaps( DC, ASPECTX ) ) ;
            15 : S := 'ASPECTY = ' + inttostr( GetDeviceCaps( DC, ASPECTY ) ) ;
            16 : S := 'ASPECTXY = ' + inttostr( GetDeviceCaps( DC, ASPECTXY ) ) ;
            17 : S := 'PDEVICESIZE = ' + inttostr( GetDeviceCaps( DC, PDEVICESIZE ) ) ;
            18 : S := 'CLIPCAPS = ' + inttostr( GetDeviceCaps( DC, CLIPCAPS ) ) ;
            19 : S := 'SIZEPALETTE = ' + inttostr( GetDeviceCaps( DC, SIZEPALETTE ) ) ;
            20 : S := 'NUMRESERVED = ' + inttostr( GetDeviceCaps( DC, NUMRESERVED ) ) ;
            21 : S := 'COLORRES = ' + inttostr( GetDeviceCaps( DC, COLORRES ) ) ;
            22 : S := 'PHYSICALWIDTH = ' + inttostr( GetDeviceCaps( DC, PHYSICALWIDTH ) ) ;
            23 : S := 'PHYSICALHEIGHT = ' + inttostr( GetDeviceCaps( DC, PHYSICALHEIGHT ) ) ;
            24 : S := 'PHYSICALOFFSETX = ' + inttostr( GetDeviceCaps( DC, PHYSICALOFFSETX ) ) ;
            25 : S := 'PHYSICALOFFSETY = ' + inttostr( GetDeviceCaps( DC, PHYSICALOFFSETY ) ) ;
            26 : S := 'VREFRESH = ' + inttostr( GetDeviceCaps( DC, VREFRESH ) ) ; // NT only
            27 : S := 'DESKTOPHORZRES = ' + inttostr( GetDeviceCaps( DC, DESKTOPHORZRES ) ) ; // NT only
            28 : S := 'DESKTOPVERTRES = ' + inttostr( GetDeviceCaps( DC, DESKTOPVERTRES ) ) ; // NT only
            29 : S := 'BLTALIGNMENT = ' + inttostr( GetDeviceCaps( DC, BLTALIGNMENT ) ) ; // NT only
            30 : S := 'RASTERCAPS = ' + RasterCaps_To_String( GetDeviceCaps( DC, RASTERCAPS ) ) ;
            31 : S := 'CURVECAPS = ' + CurveCaps_To_String( GetDeviceCaps( DC, CURVECAPS ) ) ;
            32 : S := 'LINECAPS = ' + LineCaps_To_String( GetDeviceCaps( DC, LINECAPS ) ) ;
            33 : S := 'POLYGONALCAPS = ' + PolygonCaps_To_String( GetDeviceCaps( DC, POLYGONALCAPS ) ) ;
            34 : S := 'TEXTCAPS = ' + TextCaps_To_String( GetDeviceCaps( DC, TEXTCAPS ) ) ;
        end ;
        Result.Title := PChar( S ) ;
    finally
        DeleteDC( DC ) ;
    end ;
end ;


function TDisplay_Debugger.Count : longint ;

begin
    Count := 35 ;
end ;



{$IFNDEF DELPHI7UP}
const faSymLink = $40 ;
{$ENDIF}

function Attributes_To_String( Attr : integer ) : string ;

begin
    Result := Add_Flag( Attr, faDirectory, '', 'directory' ) ;
{$WARNINGS OFF}
    Result := Add_Flag( Attr, faReadOnly, Result, 'read-only' ) ;
    Result := Add_Flag( Attr, faHidden, Result, 'hidden' ) ;
    Result := Add_Flag( Attr, faSysFile, Result, 'system' ) ;
    Result := Add_Flag( Attr, faVolumeID, Result, 'label' ) ;
    Result := Add_Flag( Attr, faArchive, Result, 'archive' ) ;
    Result := Add_Flag( Attr, faSymLink, Result, 'link' ) ;
{$WARNINGS ON}
end ;


type TFile_Debugger = class( TText_Debugger )
                          public // API...
                              Name : string ;
                              Attr : integer ;

                              function Child( Index : longint ) : PDebug_Interface ;
                                  override ;
                              function Count : longint ; override ;
                      end ;

function TFile_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Dummy : integer ;
    S : string ;
    L : DWord ;
    H : Cardinal ;
    Handle : THandle ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'Attributes = ' + Attributes_To_String( Attr ) ;
        1 : begin
                Handle := CreateFile( PChar( Name ), GENERIC_READ,
                    FILE_SHARE_DELETE or FILE_SHARE_READ or FILE_SHARE_WRITE,
                    nil, OPEN_EXISTING, 0, 0 ) ;
                Dummy := GetLastError ;
                if( Dummy <> 0 ) then
                begin
                    S := 'Size = *error* ' + inttostr( Dummy ) ;
                end else
                begin
                    L := GetFileSize( Handle, @H ) ;
                    H := ( H shl 32 ) or L ;
                    S := 'Size = ' + inttostr( H ) ;
                    closeHandle( Handle ) ;
                end ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TFile_Debugger.Count : longint ;

begin
    Count := 2 ;
end ;



type TDirectory_Debugger = class( TText_Debugger )
                               public // Constructors and destructors...
                                   destructor Destroy ; override ;

                               private // Instance data...
                                   List : TStringList ;

                               public // API...
                                   Path : string ;

                                   function Child( Index : longint ) : PDebug_Interface ;
                                       override ;
                                   function Count : longint ; override ;
                           end ;

// Constructors and destructors...

destructor TDirectory_Debugger.Destroy ;

begin
    List.Free ;

    inherited Destroy ;
end ;


// API...

function TDirectory_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Attr : integer ;
    S : string ;

begin
    S := List[ Index ] ;
    Attr := integer( List.Objects[ Index ] ) ;
    if( ( Attr and faDirectory ) <> 0 ) then
    begin
        Result := TDirectory_Debugger.Create ;
        TDirectory_Debugger( Result ).Path := Path + S + '\' ;
        S := S + ' <directory>' ;
    end else
    begin
        Result := TFile_Debugger.Create ;
        TFile_Debugger( Result ).Attr := Attr ;
        TFile_Debugger( Result ).Name := Path + S ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDirectory_Debugger.Count : longint ;

var F : TSearchRec ;
    S : string ;

begin
    if( List = nil ) then
    begin
        List := TStringList.Create ;
        FindFirst( Path + '*.*', faAnyFile, F ) ;
        try
            while( True ) do
            begin
                S := F.Name ;
                if( ( S <> '.' ) and ( S <> '..' ) ) then
                begin
                    List.AddObject( S, pointer( F.Attr ) ) ;
                end ;
                if( FindNext( F ) <> 0 ) then
                begin
                    break ;
                end ;
            end ;
        except
        end ;
        List.Sort ;
        FindClose( F ) ;
    end ;
    Result := List.Count ;
end ;


type TDrive_Debugger = class( TText_Debugger )
                           public // API...
                               Drive : integer ;

                               function Child( Index : longint ) : PDebug_Interface ;
                                   override ;
                               function Count : longint ; override ;
                       end ;

function TDrive_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Saved : integer ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    Saved := SetErrorMode( SEM_FAILCRITICALERRORS ) ;
    case Index of
        0 : begin
                GetDir( Drive, S ) ;
                S := 'Current directory = ' + S ;
            end ;
        1 : begin
                Result.Free ;
                Result := TDirectory_Debugger.Create ;
                TDirectory_Debugger( Result ).Path := chr( Drive + 64 ) + ':\' ;
                S := 'File system' ;
            end ;
        2 : S := 'Free space = ' + inttostr( DiskFree( Drive ) ) ;
        3 : S := 'Size = ' + inttostr( DiskSize( Drive ) ) ;
        4 : begin
                S := chr( Drive + 64 ) + ':\' ;
                case GetDriveType( PChar( S ) ) of
                    DRIVE_REMOVABLE : S := 'Type = removable' ;
                    DRIVE_FIXED : S := 'Type = fixed' ;
                    DRIVE_REMOTE : S := 'Type = remote (network)' ;
                    DRIVE_CDROM : S := 'Type = CDROM' ;
                    DRIVE_RAMDISK : S := 'Type = RAM Disk' ;
                    else S := 'Type = ' + inttostr( GetDriveType( PChar( S ) ) ) ;
                end ;
            end ;
    end ;
    SetErrorMode( Saved ) ;
    Result.Title := PChar( S ) ;
end ;


function TDrive_Debugger.Count : longint ;

begin
    Result := 4 ;
end ;



type TMouse_Device_Debugger = class( TText_Debugger )
                                  public // API...
                                      function Child( Index : longint ) : PDebug_Interface ;
                                          override ;
                                      function Count : longint ; override ;
                              end ;

function TMouse_Device_Debugger.Child( Index : longint ) : PDebug_Interface ;

var A : array[ 0..2 ] of integer ;
    I : integer ;
    S : string ;
    X : TOSVersionInfo ;

begin
    X.dwOSVersionInfoSize := sizeof( X ) ;
    GetVersionEx( X ) ;
    Result := TText_Debugger.Create ;
    case Index of
        0..2 : begin
                   SystemParametersInfo( SPI_GETMOUSE, 0, @A, 0 ) ;
                   S := 'Parameter ' + inttostr( Index + 1 ) + ' = ' + inttostr( A[ Index ] ) ;
               end ;
        3 : begin
                SystemParametersInfo( SPI_GETMOUSEHOVERHEIGHT, 0, @I, 0 ) ;
                S := 'Hover height = ' + inttostr( I ) ;
            end ;
        4 : begin
                SystemParametersInfo( SPI_GETMOUSEHOVERTIME, 0, @A, 0 ) ;
                S := 'Hover time = ' + inttostr( I ) + ' ms' ;
            end ;
        5 : if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then
            begin
                SystemParametersInfo( SPI_GETMOUSETRAILS, 0, @I, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Mouse trails = ' + inttostr( I ) ;
            end else
            begin
                I := 0 ;
                SystemParametersInfo( SPI_GETWHEELSCROLLLINES, 0, @I, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Wheel scroll lines = ' + inttostr( I ) ;
            end ;
        6 : begin
                SystemParametersInfo( SPI_GETMOUSEHOVERWIDTH, 0, @I, 0 ) ;
                S := 'Hover width = ' + inttostr( I ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TMouse_Device_Debugger.Count : longint ;

var X : TOSVersionInfo ;

begin
    X.dwOSVersionInfoSize := sizeof( X ) ;
    GetVersionEx( X ) ;
    if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then
    begin
        Result := 6 ;
    end else
    begin
        Result := 7 ;
    end ;
end ;



type TDrive_List_Debugger = class( TText_Debugger )
                                 public // API...
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TDrive_List_Debugger.Child( Index : longint ) : PDebug_Interface ;

var D, Loop : integer ;
    S : string ;

begin
    D := 26 ;
    for Loop := 1 to 26 do
    begin
        S := chr( Loop + 64 ) + ':\' ;
        if( GetDriveType( PChar( S ) ) > 1 ) then
        begin
            if( Index = 0 ) then
            begin
                D := Loop ;
                break ;
            end ;
            dec( Index ) ;
        end ;
    end ;
    Result := TDrive_Debugger.Create ;
    TDrive_Debugger( Result ).Drive := D ;
    S := chr( 64 + D ) + ':' ;
    Result.Title := PChar( S ) ;
end ;


function TDrive_List_Debugger.Count : longint ;

var Loop : integer ;
    S : string ;

begin
    Result := 0 ;
    for Loop := 1 to 26 do
    begin
        S := chr( Loop + 64 ) + ':\' ;
        if( GetDriveType( PChar( S ) ) > 1 ) then
        begin
            inc( Result ) ;
        end ;
    end ;
end ;


type TAsync_Key_Debugger = class( TText_Debugger )
                               public // API...
                                   function Child( Index : longint ) : PDebug_Interface ;
                                       override ;
                                   function Count : longint ; override ;
                           end ;

function TAsync_Key_Debugger.Child( Index : longint ) : PDebug_Interface ;

var I : integer ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    I := GetAsyncKeyState( Index ) ;
    setlength( S, 256 ) ;
    setlength( S, GetKeyNameText( Index, @S[ 1 ], 256 ) ) ;
    S := inttostr( Index ) + ' "' + S + '"' ;
    if( I < 0 ) then
    begin
        S := S + ' <key is down>' ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TAsync_Key_Debugger.Count : longint ;

begin
    Result := 256 ;
end ;


type TKey_Debugger = class( TText_Debugger )
                         public // API...
                             function Child( Index : longint ) : PDebug_Interface ;
                                 override ;
                             function Count : longint ; override ;
                     end ;

function TKey_Debugger.Child( Index : longint ) : PDebug_Interface ;

var I : integer ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    I := GetKeyState( Index ) ;
    setlength( S, 256 ) ;
    setlength( S, GetKeyNameText( Index, @S[ 1 ], 256 ) ) ;
    S := inttostr( Index ) + ' "' + S + '"' ;
    if( ( I < 0 ) or ( ( I and 1 ) <> 0 ) ) then
    begin
        S := S + ' <' ;
        if( I < 0 ) then
        begin
            S := S + 'down' ;
        end ;
        if( ( I and 1 ) <> 0 ) then
        begin
            if( I < 0 ) then
            begin
                S := S + ', ' ;
            end ;
            S := S + 'toggled' ;
        end ;
        S := S + '>' ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TKey_Debugger.Count : longint ;

begin
    Result := 256 ;
end ;


type TKeyboard_Device_Debugger = class( TText_Debugger )
                                     public // API...
                                         function Child( Index : longint ) : PDebug_Interface ;
                                             override ;
                                         function Count : longint ; override ;
                                 end ;

function TKeyboard_Device_Debugger.Child( Index : longint ) : PDebug_Interface ;

var B : boolean ;
    I : integer ;
    S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TAsync_Key_Debugger.Create ;
                S := 'Async Key States' ;
            end ;
        1 : begin
                Result := TKey_Debugger.Create ;
                S := 'Key States' ;
            end ;
        2 : begin
                Result := TText_Debugger.Create ;
                setlength( S, 512 ) ;
                GetKeyboardLayoutName( @S[ 1 ] ) ;
                S := string( PChar( S ) ) ; // Limit to null-terminator
                S := 'Layout = "' + S + '"' ;
            end ;
        3 : begin
                SystemParametersInfo( SPI_GETKEYBOARDPREF, 0, @B, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Preferred = ' + Boolean_To_String( B ) ;
            end ;
        4 : begin
                I := 0 ;
                SystemParametersInfo( SPI_GETKEYBOARDDELAY, 0, @I, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Repeat-delay = ' + inttostr( I ) ;
            end ;
        5 : begin
                I := 0 ;
                SystemParametersInfo( SPI_GETKEYBOARDSPEED, 0, @I, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Speed = ' + inttostr( I ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TKeyboard_Device_Debugger.Count : longint ;

begin
    Result := 6 ;
end ;


type TDevice_List_Debugger = class( TText_Debugger )
                                 public // API...
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TDevice_List_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TDisplay_Debugger.Create ;
                S := 'Display' ;
            end ;
        1 : begin
                Result := TKeyboard_Device_Debugger.Create ;
                S := 'Keyboard' ;
            end ;
        2 : begin
                Result := TMouse_Device_Debugger.Create ;
                S := 'Mouse' ;
            end ;
        3 : begin
                Result := TDrive_List_Debugger.Create ;
                S := 'Storage devices' ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDevice_List_Debugger.Count : longint ;

begin
    Result := 4 ;
end ;


function Point_To_String( P : TPoint ) : string ;

begin
    Result := 'x: ' + inttostr( P.X ) + ', y: ' + inttostr( P.Y ) ;
end ;


function Rect_To_String( R : TRect ) : string ;

begin
    Result := 'left: ' + inttostr( R.Left ) + ', top: ' + inttostr( R.Top ) +
        ', right: ' + inttostr( R.Right ) + ', bottom: ' + inttostr( R.Bottom ) ;
end ;


function Window_Description( Handle : THandle ) : string ;

var Len : integer ;
    S : string ;

begin
    if( Handle = 0 ) then
    begin
        Result := '' ;
        exit ;
    end ;
    Len := GetWindowTextLength( Handle ) ;
    setlength( S, Len ) ;
    GetWindowText( Handle, PChar( S ), Len ) ;
    if( not IsWindowVisible( Handle ) ) then
    begin
        S := S + ' <invisible>' ;
    end ;
    Result := S ;
end ;


function CB_ENumChildWindows( Handle : THandle ; Context : integer ) : boolean ;
    stdcall ;

var List : TInteger_List ;

begin
    List := TInteger_List( Context ) ;
    List.Add( Handle ) ;
    Result := True ;
end ;


var _EnumSystemLocales_List : TStringList = nil ;

function EnumSystemLocales_List : TStringList ;

begin
    if( _EnumSystemLocales_List = nil ) then
    begin
        _EnumSystemLocales_List := TStringList.Create ;
    end ;
    Result := _EnumSystemLocales_List ;
end ;


function CB_EnumSystemLocales( Name : PChar ) : boolean ;

begin
    EnumSystemLocales_List.Add( Name ) ;
    Result := True ;
end ;


//var ENum_Props_List : TStringList = nil ;

function CB_ENumProps( Handle : THandle ; Name : PChar ; Data : THandle ;
    Context : integer ) : boolean ; stdcall ;

var List : TStringList ;

begin
    List := TStringList( Context ) ;
    List.AddObject( string( Name ), pointer( Data ) ) ;
    Result := True ;
end ;


var EnumDateFormats_List : TStringList = nil ;

function CB_EnumDateFormats( Value : PChar ) : boolean ;

begin
    EnumDateFormats_List.Add( string( Value ) ) ;
    Result := True ;
end ;


type TLogFont_Debugger = class( TText_Debugger )
                                 public // API...
                                     LogFont : TLogFont ;

                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TLogFont_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := 'Charset = ' + inttostr( LogFont.lfCharSet ) ;
                case LogFont.lfCharSet of
                    ANSI_CHARSET : S := S + ' (ANSI)' ;
                    DEFAULT_CHARSET : S := S + ' (Default)' ;
                    SYMBOL_CHARSET : S := S + ' (Symbol)' ;
                    SHIFTJIS_CHARSET : S := S + ' (ShiftJis)' ;
                    GB2312_CHARSET : S := S + ' (GB2312)' ;
                    HANGEUL_CHARSET : S := S + ' (Hangeul)' ;
                    CHINESEBIG5_CHARSET : S := S + ' (ChineseBig5)' ;
                    OEM_CHARSET : S := S + ' (OEM)' ;
                    JOHAB_CHARSET : S := S + ' (Johab)' ;
                    HEBREW_CHARSET : S := S + ' (Hebrew)' ;
                    ARABIC_CHARSET : S := S + ' (Arabic)' ;
                    GREEK_CHARSET : S := S + ' (Greek)' ;
                    TURKISH_CHARSET : S := S + ' (Turkish)' ;
                    THAI_CHARSET : S := S + ' (Thai)' ;
                    EASTEUROPE_CHARSET : S := S + ' (EastEurope)' ;
                    RUSSIAN_CHARSET : S := S + ' (Russian)' ;
                    MAC_CHARSET : S := S + ' (Mac)' ;
                    BALTIC_CHARSET : S := S + ' (Baltic)' ;
                    VIETNAMESE_CHARSET : S := S + ' (Vietnamese)' ;
                end ; // case
            end ; // 0
        1 : begin
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_DEFAULT_PRECIS, '', 'Default' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_CHARACTER_PRECIS, S, 'Character' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_STROKE_PRECIS, S, 'Stroke' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_MASK, S, 'Mask' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_EMBEDDED, S, 'Embedded' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_LH_ANGLES, S, 'LH angles' ) ;
                S := Add_Flag( LogFont.lfClipPrecision, CLIP_TT_ALWAYS, S, 'TT always' ) ;
                if( length( S ) > 0 ) then
                begin
                    S := ', (' + S + ')' ;
                end ;
                S := 'Clip precision = ' + inttostr( LogFont.lfClipPrecision ) + S ;
            end ;
        2 : S := 'Escapement = ' + inttostr( LogFont.lfEscapement ) ;
        3 : S := 'Face = "' + LogFont.lfFaceName + '"' ;
        4 : S := 'Height = ' + inttostr( LogFont.lfHeight ) ;
        5 : S := 'Italic = ' + Boolean_To_String( LogFont.lfItalic <> 0 ) ;
        6 : S := 'Orientation = ' + inttostr( LogFont.lfOrientation ) ;
        7 : begin
                S := 'Output precision = ' + inttostr( LogFont.lfOutPrecision ) ;
                case LogFont.lfOutPrecision of
                    OUT_CHARACTER_PRECIS : S := S + ' (character)' ;
                    OUT_DEFAULT_PRECIS : S := S + ' (default)' ;
                    OUT_DEVICE_PRECIS : S := S + ' (device)' ;
                    OUT_OUTLINE_PRECIS : S := S + ' (outline)' ;
                    OUT_RASTER_PRECIS : S := S + ' (raster)' ;
                    OUT_STRING_PRECIS : S := S + ' (string)' ;
                    OUT_STROKE_PRECIS : S := S + ' (stroke)' ;
                    OUT_TT_ONLY_PRECIS : S := S + ' (true type only)' ;
                    OUT_TT_PRECIS : S := S + ' (true type)' ;
                end ;
            end ;
        8 : begin
                case LogFont.lfPitchAndFamily and 3 of
                    DEFAULT_PITCH : S := 'Default, ' ;
                    FIXED_PITCH : S := 'Fixed, ' ;
                    VARIABLE_PITCH : S := 'Variable, ' ;
                end ;
                case LogFont.lfPitchAndFamily shr 2 of
                    FF_DECORATIVE : S := S + 'Decorative' ;
                    FF_DONTCARE : S := S + 'Don''t care' ;
                    FF_MODERN : S := S + 'Modern' ;
                    FF_ROMAN : S := S + 'Roman' ;
                    FF_SCRIPT : S := S + 'Script' ;
                    FF_SWISS : S := S + 'Swiss' ;
                end ;
                S := 'Pitch and family = ' + inttostr( LogFont.lfPitchAndFamily ) + ', ' + S ;
            end ;
        9 : begin
                S := 'Quality = ' + inttostr( LogFont.lfQuality ) ;
                case LogFont.lfQuality of
                    DEFAULT_QUALITY : S := S + ' (default)' ;
                    DRAFT_QUALITY : S := S + ' (draft)' ;
                    PROOF_QUALITY : S := S + ' (proof)' ;
                end ;
            end ;
        10 : S := 'Strikeout = ' + Boolean_To_String( LogFont.lfStrikeOut <> 0 ) ;
        11 : S := 'Underline = ' + Boolean_To_String( LogFont.lfUnderline <> 0 ) ;
        12 : begin
                 S := 'Weight = ' + inttostr( LogFont.lfWeight ) ;
                 case LogFont.lfWeight of
                     0..150 : S := S + ' (thin)' ;
                     151..250 : S := S + ' (ultralight)' ;
                     251..350 : S := S + ' (light)' ;
                     351..450 : S := S + ' (normal)' ;
                     451..550 : S := S + ' (medium)' ;
                     551..650 : S := S + ' (semibold)' ;
                     651..750 : S := S + ' (bold)' ;
                     751..850 : S := S + ' (ultrabold)' ;
                     else S := S + ' (heavy)' ;
                 end ;
             end ;
        13 : S := 'Width = ' + inttostr( LogFont.lfWidth ) ;
    end ;
end ;


function TLogFont_Debugger.Count : longint ;

begin
    Result := 14 ;
end ;


type TBitmap_Debugger = class( TText_Debugger )
                          public // API...
                              Handle : HBitmap ;

                              function Child( Index : longint ) : PDebug_Interface ;
                                  override ;
                              function Count : longint ; override ;
                      end ;

function TBitmap_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'xHotspot = ' ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TBitmap_Debugger.Count : longint ;

begin
    Result := 0 ;
end ;


type TIcon_Debugger = class( TText_Debugger )
                          public // API...
                              Icon : HIcon ;

                              function Child( Index : longint ) : PDebug_Interface ;
                                  override ;
                              function Count : longint ; override ;
                      end ;

function TIcon_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Info : TIconInfo ;
    S : string ;

begin
    GetIconInfo( Icon, Info ) ;
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'xHotspot = ' + inttostr( Info.xHotspot ) ;
        1 : S := 'yHotspot = ' + inttostr( Info.yHotspot ) ;
        2 : begin
                Result.Free ;
                Result := TBitmap_Debugger.Create ;
                TBitmap_Debugger( Result ).Handle := Info.hbmMask ;
                S := 'hbmMask = ' + inttostr( Info.hbmMask ) ;
            end ;
        3 : begin
                Result.Free ;
                Result := TBitmap_Debugger.Create ;
                TBitmap_Debugger( Result ).Handle := Info.hbmColor ;
                S := 'hbmColor = ' + inttostr( Info.hbmColor ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TIcon_Debugger.Count : longint ;

begin
    Result := 4 ;
end ;



type TIconMetrics_Debugger = class( TText_Debugger )
                                 public // API...
                                     Metrics : TIconMetrics ;

                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TIconMetrics_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'Horizontal spacing = ' + inttostr( Metrics.iHorzSpacing ) ;
        1 : S := 'Vertical spacing = ' + inttostr( Metrics.iVertSpacing ) ;
        2 : S := 'Title wrap = ' + Boolean_To_String( Metrics.iTitleWrap <> 0 ) ;
        3 : begin
                Result := TLogFont_Debugger.Create ;
                TLogFont_Debugger( Result ).LogFont := Metrics.lfFont ;
                S := 'Font' ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TIconMetrics_Debugger.Count : longint ;

begin
    Result := 4 ;
end ;



type TWindow_Debugger = class( TText_Debugger )
                            public // Constructors and destructors...
                                destructor Destroy ; override ;

                            private
                                Children : TInteger_List ;

                            public // API...
                                Handle : THandle ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

type TWindow_List_Debugger = class( TText_Debugger )
                                 public // API...
                                     List : TInteger_List ;

                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TWindow_List_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TWindow_Debugger.Create ;
    TWindow_Debugger( Result ).Handle := List[ Index ] ;
    Result.Title := PChar( Window_Description( List[ Index ] ) ) ;
end ;


function TWindow_List_Debugger.Count : longint ;

begin
    Result := List.Count ;
end ;



function WindowExStyle_To_String( X : cardinal ) : string ;

begin
    Result := Add_Flag( X, WS_EX_ACCEPTFILES, '', 'Accept files' ) ;
    Result := Add_Flag( X, WS_EX_APPWINDOW, Result, 'App window' ) ;
    Result := Add_Flag( X, WS_EX_CLIENTEDGE, Result, 'Sunken border' ) ;
    Result := Add_Flag( X, WS_EX_CONTEXTHELP, Result, 'Context help' ) ;
    Result := Add_Flag( X, WS_EX_CONTROLPARENT, Result, 'Control parent' ) ;
    Result := Add_Flag( X, WS_EX_DLGMODALFRAME, Result, 'Double border' ) ;
    Result := Add_Flag( X, WS_EX_LEFTSCROLLBAR, Result, 'Left scrollbar' ) ;
    Result := Add_Flag( X, WS_EX_MDICHILD, Result, 'MDI child' ) ;
    Result := Add_Flag( X, WS_EX_NOPARENTNOTIFY, Result, 'No parent notify' ) ;
    Result := Add_Flag( X, WS_EX_OVERLAPPEDWINDOW, Result, 'Overlapped window' ) ;
    Result := Add_Flag( X, WS_EX_PALETTEWINDOW, Result, 'Palette window' ) ;
    Result := Add_Flag( X, WS_EX_RIGHT, Result, 'Right-aligned' ) ;
    Result := Add_Flag( X, WS_EX_RTLREADING, Result, 'Right-to-left' ) ;
    Result := Add_Flag( X, WS_EX_STATICEDGE, Result, 'Static border' ) ;
    Result := Add_Flag( X, WS_EX_TOOLWINDOW, Result, 'Tool window' ) ;
    Result := Add_Flag( X, WS_EX_TOPMOST, Result, 'Topmost' ) ;
    Result := Add_Flag( X, WS_EX_TRANSPARENT, Result, 'Transparent' ) ;
    Result := Add_Flag( X, WS_EX_WINDOWEDGE, Result, 'Raised border' ) ;
end ;


function WindowStyle_To_String( X : cardinal ) : string ;

begin
    Result := Add_Flag( X, WS_BORDER, '', 'Border' ) ;
    Result := Add_Flag( X, WS_CAPTION, Result, 'Caption' ) ;
    Result := Add_Flag( X, WS_CHILD, Result, 'Child' ) ;
    Result := Add_Flag( X, WS_CLIPCHILDREN, Result, 'Clip children' ) ;
    Result := Add_Flag( X, WS_CLIPSIBLINGS, Result, 'Clip siblings' ) ;
    Result := Add_Flag( X, WS_DISABLED, Result, 'Disabled' ) ;
    Result := Add_Flag( X, WS_DLGFRAME, Result, 'Dialog frame' ) ;
    Result := Add_Flag( X, WS_GROUP, Result, 'Group start' ) ;
    Result := Add_Flag( X, WS_HSCROLL, Result, 'Hor. scrollbar' ) ;
    Result := Add_Flag( X, WS_ICONIC, Result, 'Minimized' ) ;
    Result := Add_Flag( X, WS_MAXIMIZE, Result, 'Maximized' ) ;
    Result := Add_Flag( X, WS_MAXIMIZEBOX, Result, 'Maximize button' ) ;
    Result := Add_Flag( X, WS_MINIMIZEBOX, Result, 'Minimize button' ) ;
    Result := Add_Flag( X, WS_OVERLAPPEDWINDOW, Result, 'Overlapped window' ) ;
{$WARNINGS OFF}
    Result := Add_Flag( X, integer( WS_POPUP ), Result, 'Popup' ) ;
    Result := Add_Flag( X, integer( WS_POPUPWINDOW ), Result, 'Popup window' ) ;
{$WARNINGS ON}
    Result := Add_Flag( X, WS_SYSMENU, Result, 'System menu' ) ;
    Result := Add_Flag( X, WS_TABSTOP, Result, 'Tab stop' ) ;
    Result := Add_Flag( X, WS_THICKFRAME, Result, 'Thick frame' ) ;
    Result := Add_Flag( X, WS_TILED, Result, 'Tiled' ) ;
    Result := Add_Flag( X, WS_TILEDWINDOW, Result, 'Tiled window' ) ;
    Result := Add_Flag( X, WS_VISIBLE, Result, 'Visible' ) ;
    Result := Add_Flag( X, WS_VSCROLL, Result, 'Vertical scrollbar' ) ;
end ;


type TMinimized_Metrics_Debugger = class( TText_Debugger )
                                       public // API...
                                           function Child( Index : longint ) : PDebug_Interface ;
                                               override ;
                                           function Count : longint ; override ;
                                   end ;

function TMinimized_Metrics_Debugger.Child( Index : longint ) : PDebug_Interface ;

var MM : TMinimizedMetrics ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    fillchar( MM, sizeof( MM ), 0 ) ;
    MM.cbSize := sizeof( MM ) ;
    SystemParametersInfo( SPI_GETMINIMIZEDMETRICS, sizeof( MM ), @MM, 0 ) ;
    case Index of
        0 : S := 'Width = ' + inttostr( MM.iWidth ) ;
        1 : S := 'HorzGap = ' + inttostr( MM.iHorzGap ) ;
        2 : S := 'VertGap = ' + inttostr( MM.iVertGap ) ;
        3 : begin
                case MM.iArrange and 3 of
                    0 : S := 'Bottomleft' ;
                    1 : S := 'Bottomright' ;
                    2 : S := 'Topleft' ;
                    3 : S := 'Topright' ;
                end ;
                case ( MM.iArrange shr 2 ) and 1 of
                    0 : S := S + 'left/right' ;
                    1 : S := S + 'up/down' ;
                end ;
                S := 'Arrange = ' + inttostr( MM.iArrange ) + ', ' + S ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TMinimized_Metrics_Debugger.Count : longint ;

begin
    Count := 4 ;
end ;



function RGB_To_String( X : integer ) : string ;

begin
    Result := 'A: ' + inttostr( X shr 24 ) + ', R: ' + inttostr( X and $FF ) +
        ', G: ' + inttostr( ( X shr 8 ) and $FF ) + ', B: ' +
        inttostr( ( X shr 16 ) and $FF ) ;
    //~~~ Add in color name
end ;


type TDeviceCaps_Debugger = class( TText_Debugger )
                                public // API...
                                    DC : HDC ;

                                    function Child( Index : longint ) : PDebug_Interface ;
                                        override ;
                                    function Count : longint ; override ;
                            end ;

function TDeviceCaps_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := 'ASPECTX = ' + inttostr( GetDeviceCaps( DC, ASPECTX ) ) ;
            end ;
        1 : begin
                S := 'ASPECTY = ' + inttostr( GetDeviceCaps( DC, ASPECTY ) ) ;
            end ;
        2 : begin
                S := 'ASPECTXY = ' + inttostr( GetDeviceCaps( DC, ASPECTXY ) ) ;
            end ;
        3 : begin
                S := 'BITSPIXEL = ' + inttostr( GetDeviceCaps( DC, BITSPIXEL ) ) ;
            end ;
        4 : begin
                S := 'COLORRES = ' + inttostr( GetDeviceCaps( DC, COLORRES ) ) ;
            end ;
        5 : begin
                S := 'HORZRES = ' + inttostr( GetDeviceCaps( DC, HORZRES ) ) ;
            end ;
        6 : begin
                S := 'HORZSIZE = ' + inttostr( GetDeviceCaps( DC, HORZSIZE ) ) ;
            end ;
        7 : begin
                S := 'LOGPIXELSX = ' + inttostr( GetDeviceCaps( DC, LOGPIXELSX ) ) ;
            end ;
        8 : begin
                S := 'LOGPIXELSY = ' + inttostr( GetDeviceCaps( DC, LOGPIXELSY ) ) ;
            end ;
        9 : begin
                S := 'NUMBRUSHES = ' + inttostr( GetDeviceCaps( DC, NUMBRUSHES ) ) ;
            end ;
        10 : begin
                S := 'NUMCOLORS = ' + inttostr( GetDeviceCaps( DC, NUMCOLORS ) ) ;
             end ;
        11 : begin
                S := 'NUMFONTS = ' + inttostr( GetDeviceCaps( DC, NUMFONTS ) ) ;
             end ;
        12 : begin
                S := 'NUMMARKERS = ' + inttostr( GetDeviceCaps( DC, NUMMARKERS ) ) ;
             end ;
        13 : begin
                S := 'NUMPENS = ' + inttostr( GetDeviceCaps( DC, NUMPENS ) ) ;
             end ;
        14 : begin
                S := 'NUMRESERVED = ' + inttostr( GetDeviceCaps( DC, NUMRESERVED ) ) ;
             end ;
        15 : begin
                S := 'PDEVICESIZE = ' + inttostr( GetDeviceCaps( DC, PDEVICESIZE ) ) ;
             end ;
        16 : begin
                S := 'PLANES = ' + inttostr( GetDeviceCaps( DC, PLANES ) ) ;
             end ;
        17 : begin
                S := 'SIZEPALETTE = ' + inttostr( GetDeviceCaps( DC, SIZEPALETTE ) ) ;
             end ;
        18 : begin
                S := 'VERTRES = ' + inttostr( GetDeviceCaps( DC, VERTRES ) ) ;
             end ;
        19 : begin
                S := 'VERTSIZE = ' + inttostr( GetDeviceCaps( DC, VERTSIZE ) ) ;
             end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDeviceCaps_Debugger.Count : longint ;

begin
    Count := 20 ;
end ;


type TDC_Debugger = class( TText_Debugger )
                        public // API...
                            DC : HDC ;

                            function Child( Index : longint ) : PDebug_Interface ;
                                override ;
                            function Count : longint ; override ;
                    end ;

function MapMode_To_String( X : integer ) : string ;

begin
    case X of
        MM_TEXT : Result := 'MM_TEXT' ;
        MM_LOMETRIC : Result := 'MM_LOMETRIC' ;
        MM_HIMETRIC : Result := 'MM_HIMETRIC' ;
        MM_LOENGLISH : Result := 'MM_LOENGLISH' ;
        MM_HIENGLISH : Result := 'MM_HIENGLISH' ;
        MM_TWIPS : Result := 'MM_TWIPS' ;
        MM_ISOTROPIC : Result := 'MM_ISOTROPIC' ;
        MM_ANISOTROPIC : Result := 'MM_ANISOTROPIC' ;
        else Result := inttostr( X ) ;
    end ;
end ;


function TDC_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Dummy : integer ;
    R : TRect ;
    S : string ;
    Size : TSize ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : case GetArcDirection( DC ) of
                AD_COUNTERCLOCKWISE : S := 'Arc direction = Counterclockwise' ;
                AD_CLOCKWISE : S := 'Arc direction = Clockwise' ;
                else S := 'Arc Direction = ' + inttostr( GetArcDirection( DC ) ) ;
            end ;
        1 : begin
                if( not GetAspectRatioFilterEx( DC, Size ) ) then
                begin
                    S := 'Aspect Ratio Filter = X: ' + inttostr( Size.cx ) +
                        ', Y: ' + inttostr( Size.cy ) ;
                end else
                begin
                    S := 'Aspect Ratio Filter = * error *' ;
                end ;
            end ;
        2 : S := 'Background color = ' + RGB_To_String( GetBkColor( DC ) ) ;
        3 : begin
                if( GetBkMode( DC ) = OPAQUE ) then
                begin
                    S := 'Opaque' ;
                end else
                begin
                    S := 'Transparent' ;
                end ;
                S := 'Background mode = ' + S ;
            end ;
        4 : begin
                case GetBoundsRect( DC, R, 0 ) of
                    0 : S := 'invalid DC' ;
                    DCB_DISABLE : S := Rect_To_String( R ) + ' (boundary accumulation is off)' ;
                    DCB_ENABLE : S := Rect_To_String( R ) + ' (boundary accumulation is off)' ;
                    DCB_RESET : S := 'empty' ;
                    else S := Rect_To_String( R ) ;
                end ;
                S := 'BoundRect = ' + S ;
            end ;
        5 : begin
                GetClipBox( DC, R ) ;
                S := 'ClipBox = ' + Rect_To_String( R ) ;
            end ;
        6 : begin
                S := 'DeviceCaps' ;
                Result.Free ;
                Result := TDeviceCaps_Debugger.Create ;
                TDeviceCaps_Debugger( Result ).DC := DC ;
            end ;
        7 : S := 'MapMode = ' + MapMode_To_String( GetMapMode( DC ) ) ;
        8 : begin
                Dummy := GetROP2( DC ) ;
                case Dummy of
                    R2_BLACK : S := 'Black' ;
                    R2_COPYPEN : S := 'CopyPen' ;
                    R2_MASKNOTPEN : S := 'MaskNotOpen' ;
                    R2_MASKPEN : S := 'MaskPen' ;
                    R2_MASKPENNOT : S := 'MaskPenNot' ;
                    R2_MERGENOTPEN : S := 'MergeNotPen' ;
                    R2_MERGEPEN : S := 'MergePen' ;
                    R2_MERGEPENNOT : S := 'MergePenNot' ;
                    R2_NOP : S := 'NOP' ;
                    R2_NOT : S := 'Not' ;
                    R2_NOTCOPYPEN : S := 'NotCopyPen' ;
                    R2_NOTMASKPEN : S := 'NotMaskPen' ;
                    R2_NOTMERGEPEN : S := 'NotMergePen' ;
                    R2_NOTXORPEN : S := 'NotXOrPen' ;
                    R2_WHITE : S := 'White' ;
                    R2_XORPEN : S := 'XOrPen' ;
                    else S := inttostr( Dummy ) ;
                end ;
                S := 'ROP2 = ' + S ;
            end ;
        9 : begin
                if( WindowFromDC( DC ) <> 0 ) then
                begin
                    Result.Free ;
                    Result := TWindow_Debugger.Create ;
                end ;
                S := 'Window = ' + Window_Description( WindowFromDC( DC ) ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDC_Debugger.Count : longint ;

begin
    Count := 10 ;
end ;



type TMenus_Debugger = class( TText_Debugger )
                            public // API...
                                Handle : THandle ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

type TMenu_Item_Debugger = class( TText_Debugger )
                            public // API...
                                Handle : THandle ;
                                Position : integer ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

type TMenu_Debugger = class( TText_Debugger )
                            public // API...
                                Handle : THandle ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

function TMenu_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TMenus_Debugger.Create ;
    TMenus_Debugger( Result ).Handle := Handle ;
    Result.Title := PChar( 'Children' ) ;
end ;


function TMenu_Debugger.Count : longint ;

begin
    Result := 1 ;
end ;



// TMenus_Debugger methods...

function TMenus_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TMenu_Item_Debugger.Create ;
    TMenu_Item_Debugger( Result ).Handle := GetSubMenu( Handle, Index ) ;
    TMenu_Item_Debugger( Result ).Position := Index ;
    Result.Title := PChar( inttostr( Index ) ) ;
end ;


function TMenus_Debugger.Count : longint ;

begin
    Result := GetMenuItemCount( Handle ) ;
end ;



// TMenu_Item_Debugger methods...

function TMenu_Item_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;
    Info : TMENUITEMINFO ;

begin
    Result := nil ;
    fillchar( Info, sizeof( Info ), 0 ) ;
    Info.cbSize := sizeof( Info ) ;
    Info.fMask := MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE ;
    case Index of
        0 : begin
                S := 'Children' ;
                Result := TMenus_Debugger.Create ;
                TMenus_Debugger( Result ).Handle := Handle ;
                //TODO: Should we use Info.hSubMenu instead?
            end ;
        1 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'Checked = ' +
                    Boolean_To_String( ( Info.fState and MFS_CHECKED ) <> 0 ) ;
            end ;
        2 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'Default = ' +
                    Boolean_To_String( ( Info.fState and MFS_DEFAULT ) <> 0 ) ;
            end ;
        4 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'Separator = ' +
                    Boolean_To_String( ( Info.fType and MFT_SEPARATOR ) <> 0 ) ;
            end ;
        5 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'Disabled = ' +
                    Boolean_To_String( ( Info.fState and MFS_DISABLED ) <> 0 ) ;
            end ;
        6 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'Highlighted = ' +
                    Boolean_To_String( ( Info.fState and MFS_HILITE ) <> 0 ) ;
            end ;
        7 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                S := 'ID = ' + inttostr( Info.wID ) ;
            end ;
        8 : begin
                Info.fType := MFT_BITMAP ;
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                if( Info.dwTypeData <> nil ) then
                begin
                    Result := TBitmap_Debugger.Create ;
                    TBitmap_Debugger( Result ).Handle := integer( Info.dwTypeData ) ;
                end else
                begin
                    Result := TText_Debugger.Create ;
                end ;
                S := 'Bitmap = ' + inttostr( integer( Info.dwTypeData ) ) ;
            end ;
        9 : begin
                GetMenuItemInfo( Handle, Position, True, Info ) ;
                Result := TText_Debugger.Create ;
                if( ( Info.fType and MFT_OWNERDRAW ) <> 0 ) then
                begin
                    S := 'Data = ' + inttostr( integer( Info.dwTypeData ) ) ;
                end else
                begin
                    S := 'Data = n/a' ;
                end ;
            end ;
        10 : begin
                 GetMenuItemInfo( Handle, Position, True, Info ) ;
                 Result := TText_Debugger.Create ;
                 S := 'Owner draw = ' +
                     Boolean_To_String( ( Info.fType and MFT_OWNERDRAW ) <> 0 ) ;
             end ;
        11 : begin
                 GetMenuItemInfo( Handle, Position, True, Info ) ;
                 Result := TText_Debugger.Create ;
                 S := 'Menubar break = ' +
                     Boolean_To_String( ( Info.fType and MFT_MENUBARBREAK ) <> 0 ) ;
             end ;
        12 : begin
                 Info.fType := MFT_RADIOCHECK ;
                 GetMenuItemInfo( Handle, Position, True, Info ) ;
                 if( Info.hbmpChecked <> 0 ) then
                 begin
                     Result := TBitmap_Debugger.Create ;
                     TBitmap_Debugger( Result ).Handle := Info.hbmpChecked ;
                     S := 'Radio check = ' + inttostr( Info.hbmpChecked ) ;
                 end else
                 begin
                     Result := TText_Debugger.Create ;
                     S := 'Radio check = 0' ;
                 end ;
             end ;
        13 : begin
                 GetMenuItemInfo( Handle, Position, True, Info ) ;
                 Result := TText_Debugger.Create ;
                 S := 'Owner draw = ' +
                     Boolean_To_String( ( Info.fType and MFT_RIGHTJUSTIFY ) <> 0 ) ;
             end ;
        14 : begin
                 Info.fType := MFT_STRING ;
                 GetMenuItemInfo( Handle, Position, True, Info ) ;
                 Result := TText_Debugger.Create ;
                 S := 'Text = "' +
                     String( Info.dwTypeData ) + '"' ;
             end ;
    end ;
    Result.Title := PChar( S ) ;
end ; // TMenu_Item_Debugger.Child


function TMenu_Item_Debugger.Count : longint ;

begin
    Result := 15 ;
end ;



function Window_Class_Name( Handle : THandle ) : string ;

begin
    setlength( Result, 256 ) ;
    setlength( Result, GetClassName( Handle, @Result[ 1 ], length( Result ) ) ) ;
end ;



type TCursor_Debugger = class( TText_Debugger )
                            public
                                Cursor : hCursor ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

function TCursor_Debugger.Child( Index : longint ) : PDebug_Interface ;

var P : TPoint ;
    Rect : TRect ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := 'Cursor = ' + inttostr( GetCursor ) ;
            end ;
        1 : begin
                GetCursorPos( P ) ;
                S := 'Position = ' + Point_To_String( P ) ;
            end ;
        2 : begin
                GetClipCursor( Rect ) ;
                S := 'Clipping = ' + Rect_To_String( Rect ) ;
            end ;
        3 : begin
                Result.Free ;
                Result := TIcon_Debugger.Create ;
                TIcon_Debugger( Result ).Icon := GetCursor ;
                S := 'Image' ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TCursor_Debugger.Count : longint ;

begin
    Count := 4 ;
end ;



type TWindow_Word_Data_Debugger = class( TText_Debugger )
                                      private
                                          _Count : integer ;

                                      public // API...
                                          Handle : THandle ;

                                          function Child( Index : longint ) : PDebug_Interface ;
                                              override ;
                                          function Count : longint ; override ;
                                  end ;

function TWindow_Word_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' +
        inttostr( GetWindowWord( Handle, Index * 2 ) ) ) ;
end ;


function TWindow_Word_Data_Debugger.Count : longint ;

begin
    Count := _Count div 2 ;
end ;


type TWindow_Long_Data_Debugger = class( TText_Debugger )
                                      private
                                          _Count : integer ;

                                      public // API...
                                          Handle : THandle ;

                                          function Child( Index : longint ) : PDebug_Interface ;
                                              override ;
                                          function Count : longint ; override ;
                                  end ;

function TWindow_Long_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' +
        inttostr( GetWindowLong( Handle, Index * 4 ) ) ) ;
end ;


function TWindow_Long_Data_Debugger.Count : longint ;

begin
    Count := _Count div 4 ;
end ;


type TWindow_Data_Debugger = class( TText_Debugger )
                                  public // API...
                                      Handle : THandle ;
                                      Name : string ;

                                      function Child( Index : longint ) : PDebug_Interface ;
                                          override ;
                                      function Count : longint ; override ;
                              end ;

function TWindow_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

var _Count : integer ;
    Info : TWNDCLASSEX ;
    S : string ;

begin
    fillchar( Info, sizeof( Info ), 0 ) ;
    Info.cbSize := sizeof( Info ) ;
    GetClassInfoEx( HInstance, PChar( Name ), Info ) ;
    _Count := Info.cbWndExtra ;
    Result := nil ;
    case Index of
        0 : begin
                Result := TWindow_Word_Data_Debugger.Create ;
                TWindow_Word_Data_Debugger( Result )._Count := _Count ;
                TWindow_Word_Data_Debugger( Result ).Handle := Handle ;
                S := 'Word' ;
            end ;
        1 : begin
                Result := TWindow_Long_Data_Debugger.Create ;
                TWindow_Long_Data_Debugger( Result )._Count := _Count ;
                TWindow_Long_Data_Debugger( Result ).Handle := Handle ;
                S := 'Long' ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TWindow_Data_Debugger.Count : longint ;

begin
    Count := 2 ;
end ;



type TClass_Word_Data_Debugger = class( TText_Debugger )
                                      private
                                          _Count : integer ;

                                      public // API...
                                          Handle : THandle ;

                                          function Child( Index : longint ) : PDebug_Interface ;
                                              override ;
                                          function Count : longint ; override ;
                                  end ;

function TClass_Word_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' +
        inttostr( GetClassWord( Handle, Index * 2 ) ) ) ;
end ;


function TClass_Word_Data_Debugger.Count : longint ;

begin
    Count := _Count div 2 ;
end ;


type TClass_Long_Data_Debugger = class( TText_Debugger )
                                      private
                                          _Count : integer ;

                                      public // API...
                                          Handle : THandle ;

                                          function Child( Index : longint ) : PDebug_Interface ;
                                              override ;
                                          function Count : longint ; override ;
                                  end ;

function TClass_Long_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' +
        inttostr( GetClassLong( Handle, Index * 4 ) ) ) ;
end ;


function TClass_Long_Data_Debugger.Count : longint ;

begin
    Count := _Count div 4 ;
end ;


type TClass_Data_Debugger = class( TText_Debugger )
                                  public // API...
                                      Handle : THandle ;
                                      Name : string ;

                                      function Child( Index : longint ) : PDebug_Interface ;
                                          override ;
                                      function Count : longint ; override ;
                              end ;

function TClass_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

var _Count : integer ;
    Info : TWNDCLASSEX ;
    S : string ;

begin
    fillchar( Info, sizeof( Info ), 0 ) ;
    Info.cbSize := sizeof( Info ) ;
    GetClassInfoEx( HInstance, PChar( Name ), Info ) ;
    _Count := Info.cbWndExtra ;
    Result := nil ;
    case Index of
        0 : begin
                Result := TClass_Word_Data_Debugger.Create ;
                TClass_Word_Data_Debugger( Result )._Count := _Count ;
                TClass_Word_Data_Debugger( Result ).Handle := Handle ;
                S := 'Word' ;
            end ;
        1 : begin
                Result := TClass_Long_Data_Debugger.Create ;
                TClass_Long_Data_Debugger( Result )._Count := _Count ;
                TClass_Long_Data_Debugger( Result ).Handle := Handle ;
                S := 'Long';
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TClass_Data_Debugger.Count : longint ;

begin
    Count := 2 ;
end ;



type TWindow_Class_Debugger = class( TText_Debugger )
                                  private
                                      Info : TWNDCLASSEX ;

                                  public // API...
                                      Handle : THandle ;
                                      Name : string ;

                                      function Child( Index : longint ) : PDebug_Interface ;
                                          override ;
                                      function Count : longint ; override ;
                              end ;

function TWindow_Class_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := Add_Flag( Info.Style, CS_BYTEALIGNCLIENT, '', 'byte-align client' ) ;
                S := Add_Flag( Info.Style, CS_BYTEALIGNWINDOW, S, 'byte-align window' ) ;
                S := Add_Flag( Info.Style, CS_CLASSDC, S, 'class DC' ) ;
                S := Add_Flag( Info.Style, CS_DBLCLKS, S, 'double-clicks' ) ;
                S := Add_Flag( Info.Style, CS_GLOBALCLASS, S, 'global' ) ;
                S := Add_Flag( Info.Style, CS_HREDRAW, S, 'horizontal redraw' ) ;
                S := Add_Flag( Info.Style, CS_NOCLOSE, S, 'no close' ) ;
                S := Add_Flag( Info.Style, CS_OWNDC, S, 'own DC' ) ;
                S := Add_Flag( Info.Style, CS_PARENTDC, S, 'parent DC' ) ;
                S := Add_Flag( Info.Style, CS_SAVEBITS, S, 'save bits' ) ;
                S := Add_Flag( Info.Style, CS_VREDRAW, S, 'vertical redraw' ) ;
                S := 'Style = ' + S ;
            end ;
        1 : S := 'WndProc = ' + inttostr( integer( @Info.lpfnWndProc ) ) ;
        2 : S := 'ClsExtra = ' + inttostr( Info.cbClsExtra ) ;
        3 : S := 'WndExtra = ' + inttostr( Info.cbWndExtra ) ;
        4 : S := 'Instance = ' + inttostr( Info.hInstance ) ;
        5 : begin
                if( Info.hIconSm <> 0 ) then
                begin
                    Result.Free ;
                    Result := TIcon_Debugger.Create ;
                    TIcon_Debugger( Result ).Icon := Info.hIconSm ;
                end ;
                S := 'IconSm = ' + inttostr( Info.hIconSm ) ;
            end ;
        6 : begin
                if( Info.hIconSm <> 0 ) then
                begin
                    Result.Free ;
                    Result := TCursor_Debugger.Create ;
                    TCursor_Debugger( Result ).Cursor := Info.hCursor ;
                end ;
                S := 'Cursor = ' + inttostr( Info.hCursor ) ;
            end ;
        7 : begin
                S := 'Background = ' + inttostr( Info.hbrBackground ) ; // TODO: Brush debugger?
            end ;
        8 : S := 'MenuName = "' + Info.lpszMenuName + '"' ;
        9 : begin
                if( Info.hIconSm <> 0 ) then
                begin
                    Result.Free ;
                    Result := TIcon_Debugger.Create ;
                    TIcon_Debugger( Result ).Icon := Info.hIconSm ;
                end ;
                S := 'IconSm = ' + inttostr( Info.hIconSm ) ;
            end ;
        10 : begin
                 Result.Free ;
                 Result := TClass_Data_Debugger.Create ;
                 TClass_Data_Debugger( Result ).Handle := Handle ;
                 S := 'Data' ;
             end ;
    end ;
    Result.Title := PChar( S ) ;
end ; // TWindow_Class_Debugger.Child


function TWindow_Class_Debugger.Count : longint ;

begin
    fillchar( Info, sizeof( Info ), 0 ) ;
    Info.cbSize := sizeof( Info ) ;
    GetClassInfoEx( HInstance, PChar( Name ), Info ) ;
    Result := 11 ;
end ;



type TWindow_Properties_Debugger = class( TText_Debugger )
                                       public // Constructors and destructors...
                                           destructor Destroy ; override ;

                                       private
                                           Children : TStringList ;

                                       public // API...
                                           Handle : THandle ;

                                           function Child( Index : longint ) : PDebug_Interface ;
                                               override ;
                                           function Count : longint ; override ;
                                   end ;

// Constructors and destructors...

destructor TWindow_Properties_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TWindow_Properties_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( Children[ Index ] ) ;
end ;


function TWindow_Properties_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Children := TStringList.Create ;
    end ;
    EnumPropsEx( Handle, @CB_ENumProps, integer( Children ) ) ;
    Count := Children.Count ;
end ;


type TThread_Debugger = class( TText_Debugger )
                            public // Constructors and destructors...
                                destructor Destroy ; override ;

                            private
                                Children : TInteger_List ;

                            public // API...
                                Handle : THandle ;

                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;

// Constructors and destructors...

destructor TThread_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TThread_Debugger.Child( Index : longint ) : PDebug_Interface ;

var KLH : THandle ;
    S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TWindow_List_Debugger.Create ;
                TWindow_List_Debugger( Result ).List := Children ;
                S := 'Created windows' ;
            end ;
        1 : begin
                Result := TText_Debugger.Create ;
                S := 'Desktop = ' + inttostr( GetThreadDesktop( Handle ) ) ;
            end ;
        2 : begin
                Result := TText_Debugger.Create ;
                if( Handle = GetCurrentThreadID ) then
                begin
                    S := inttostr( GdiGetBatchLimit ) ;
                end else
                begin
                    S := 'Unknown' ;
                end ;
                S := 'GDI Batch Limit = ' + S ;
            end ;
        3 : begin
                Result := TText_Debugger.Create ;
                KLH := GetKeyboardLayout( Handle ) ;
                S := 'Keyboard layout = ' + inttostr( KLH ) ;
            end ;
        4 : begin
                if( Handle = GetCurrentThreadID ) then
                begin
                    if( GetCapture = 0 ) then
                    begin
                        Result := TText_Debugger.Create ;
                    end else
                    begin
                        Result := TWindow_Debugger.Create ;
                        TWindow_Debugger( Result ).Handle := GetCapture ;
                    end ;
                    S := 'Mouse capture window = ' + Window_Description( GetCapture ) ;
                end else
                begin
                    S := 'Mouse capture window = unknown' ;
                end ;
            end ;
    end ; // case Index
    Result.Title := PChar( S ) ;
end ;


function TThread_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Children := TInteger_List.Create ;
    end ;
    EnumThreadWindows( Handle, @CB_ENumChildWindows, integer( Children ) ) ;
    Count := 5 ;
end ;



type TBringWindowToTop_Command = class( TText_Debugger )
                                      public // API...
                                          Handle : THandle ;

                                          procedure Activate ; override ;
                                  end ;

// TBringWindowToTop_Command methods...

procedure TBringWindowToTop_Command.Activate ;

begin
    BringWindowToTop( Handle ) ;
end ;



type TWindowPlacement_Debugger = class( TText_Debugger )
                                     public // API...
                                         Handle : THandle ;

                                         function Child( Index : longint ) : PDebug_Interface ;
                                             override ;
                                         function Count : longint ; override ;
                                 end ;

function TWindowPlacement_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;
    WP : TWindowPlacement ;

begin
    Result := TText_Debugger.Create ;
    fillchar( WP, sizeof( WP ), 0 ) ;
    WP.Length := sizeof( WP ) ;
    GetWindowPlacement( Handle, @WP ) ;
    case Index of
        0 : begin
                S := Add_Flag( WP.Flags, WPF_RESTORETOMAXIMIZED, '', 'Restore to maximized' ) ;
                S := Add_Flag( WP.Flags, WPF_SETMINPOSITION, S, 'Set to Min position' ) ;
                S := 'Flags = ' + S ;
            end ;
        1 : begin
                case WP.showCmd of
                    SW_HIDE : S := '(Hide)' ;
                    SW_MINIMIZE : S := '(Minimize)' ;
                    SW_RESTORE : S := '(Restore)' ;
                    SW_SHOW : S := '(Show)' ;
                    SW_SHOWMAXIMIZED : S := '(ShowMaximized)' ;
                    SW_SHOWMINIMIZED : S := '(ShowMinimized)' ;
                    SW_SHOWMINNOACTIVE : S := '(ShowMinNoActive)' ;
                    SW_SHOWNA : S := '(ShowNA)' ;
                    SW_SHOWNOACTIVATE : S := '(ShowNoActivate)' ;
                    SW_SHOWNORMAL : S := '(ShowNormal)' ;
                    else S := '' ;
                end ;
                S := 'ShowCmd = ' + inttostr( WP.showCmd ) + ' ' + S ;
            end ;
        2 : S := 'Min Position = ' + Point_To_String( WP.ptMinPosition ) ;
        3 : S := 'Max Position = ' + Point_To_String( WP.ptMaxPosition ) ;
        4 : S := 'Normal Position = ' + Rect_To_String( WP.rcNormalPosition ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TWindowPlacement_Debugger.Count : longint ;

begin
    Count := 5 ;
end ;



// TWindow_Debugger methods...

// Constructors and destructors...

destructor TWindow_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TWindow_Debugger.Child( Index : longint ) : PDebug_Interface ;

var DC : HDC ;
    P : integer ;
    R : TRect ;
    Rgn : HRGN ;
    S : string ;

begin
    Result := nil ;
    try
        case Index of
            0 : begin
                    Result := TBringWindowToTop_Command.Create ;
                    TBringWindowToTop_Command( Result ).Handle := Handle ;
                    S := '!> BringWindowToTop' ;
                end ;
            1 : begin
                    Result := TWindow_List_Debugger.Create ;
                    TWindow_List_Debugger( Result ).List := Children ;
                    S := 'Children' ;
                end ;
            2 : begin
                    S := Window_Class_Name( Handle ) ;
                    Result := TWindow_Class_Debugger.Create ;
                    TWindow_Class_Debugger( Result ).Name := S ;
                    TWindow_Class_Debugger( Result ).Handle := Handle ;
                    S := 'Class = "' + S + '"' ;
                end ;
            3 : begin
                    Result := TDC_Debugger.Create ;
                    DC := GetDC( Handle ) ;
                    TDC_Debugger( Result ).DC := DC ;
                    S := 'Client Device Context = ' + inttostr( DC ) ; //TODO: Need to release DC
                end ;
            4 : begin
                    GetClientRect( Handle, R ) ;
                    Result := TText_Debugger.Create ;
                    S := 'ClientRect = ' + Rect_To_String( R ) ;
                end ;
            5 : begin
                    Result := TText_Debugger.Create ;
                    S := 'Clipboard open = ' +
                        Boolean_To_String( Handle = GetOpenClipboardWindow ) ;
                end ;
            6 : begin
                    Result := TText_Debugger.Create ;
                    S := 'Clipboard owner = ' +
                        Boolean_To_String( Handle = GetClipboardOwner ) ;
                end ;
            7 : begin
                    Result := TText_Debugger.Create ;
                    S := 'ContextHelpID = ' + inttostr( GetWindowContextHelpId( Handle ) ) ;
                end ;
            8 : begin
                    S := Window_Class_Name( Handle ) ;
                    Result := TWindow_Class_Debugger.Create ;
                    TWindow_Data_Debugger( Result ).Name := S ;
                    TWindow_Data_Debugger( Result ).Handle := Handle ;
                    S := 'Data' ;
                end ;
            9 : begin
                    Result := TText_Debugger.Create ;
                    S := 'DlgProc = ' + inttostr( GetWindowLong( Handle, DWL_DLGPROC ) ) ;
                end ;
            10 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Enabled = ' +
                         Boolean_To_String( IsWindowEnabled( Handle ) ) ;
                 end ;
            11 : begin
                     Result := TText_Debugger.Create ;
                     S := 'ExStyle = ' +
                         WindowExStyle_To_String( GetWindowLong( Handle, GWL_EXSTYLE ) ) ;
                 end ;
            12 : begin
                     Result := TText_Debugger.Create ;
                     S := 'HInstance = ' +
                         inttostr( GetWindowLong( Handle, GWL_HINSTANCE ) ) ;
                 end ;
            13 : begin
                     Result := TText_Debugger.Create ;
                     S := 'ID = ' +
                         inttostr( GetWindowLong( Handle, GWL_ID ) ) ;
                 end ;
            14 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Maximized = ' + Boolean_To_String( IsZoomed( Handle ) ) ;
                 end ;
            15 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Minimized = ' + Boolean_To_String( IsIconic( Handle ) ) ;
                 end ;
            16 : begin
                     Result := TMenu_Debugger.Create ;
                     TMenu_Debugger( Result ).Handle := GetMenu( Handle ) ;
                     S := 'Menu = ' + inttostr( GetMenu( Handle ) ) ;
                 end ;
            17 : begin
                     if( GetWindow( Handle, GW_HWNDNEXT ) = 0 ) then
                     begin
                         Result := TText_Debugger.Create ;
                     end else
                     begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle :=
                             GetWindow( Handle, GW_HWNDNEXT ) ;
                     end ;
                     S := Window_Description( GetWindow( Handle, GW_HWNDNEXT ) ) ;
                     if( length( S ) > 0 ) then
                     begin
                         S := ' "' + S + '"' ;
                     end ;
                     S := 'Next Z-order = ' +
                         inttostr( GetWindow( Handle, GW_HWNDNEXT ) ) + S ;
                 end ;
            18 : begin
                     if( GetParent( Handle ) = 0 ) then
                     begin
                         Result := TText_Debugger.Create ;
                     end else
                     begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle := GetParent( Handle ) ;
                     end ;
                     S := Window_Description( GetParent( Handle ) ) ;
                     if( length( S ) > 0 ) then
                     begin
                         S := ' "' + S + '"' ;
                     end ;
                     S := 'Parent = ' + inttostr( GetParent( Handle ) ) + S ;
                 end ;
            19 : begin
                     if( GetWindow( Handle, GW_HWNDPREV ) = 0 ) then
                     begin
                         Result := TText_Debugger.Create ;
                     end else
                     begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle :=
                             GetWindow( Handle, GW_HWNDPREV ) ;
                     end ;
                     S := Window_Description( GetWindow( Handle, GW_HWNDPREV ) ) ;
                     if( length( S ) > 0 ) then
                     begin
                         S := ' "' + S + '"' ;
                     end ;
                     S := 'Previous Z-order = ' +
                         inttostr( GetWindow( Handle, GW_HWNDPREV ) ) + S ;
                 end ;
            20 : begin
                     GetWindowThreadProcessId( Handle, @P ) ;
                     Result := TText_Debugger.Create ;
                     S := 'Process ID = ' + inttostr( P ) ;
                 end ;
            21 : begin
                     Result := TWindow_Properties_Debugger.Create ;
                     TWindow_Properties_Debugger( Result ).Handle := Handle ;
                     S := 'Properties' ;
                 end ;
            22 : begin
                     GetWindowRect( Handle, R ) ;
                     Result := TText_Debugger.Create ;
                     S := 'Rect = ' + Rect_To_String( R ) ;
                 end ;
            23 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Style = ' +
                         WindowStyle_To_String( GetWindowLong( Handle, GWL_STYLE ) ) ;
                 end ;
            24 : begin
                     Result := TThread_Debugger.Create ;
                     TThread_Debugger( Result ).Handle :=
                         GetWindowThreadProcessId( Handle, nil ) ;
                     S := 'Thread Process ID = ' +
                         inttostr( GetWindowThreadProcessId( Handle, nil ) ) ;
                 end ;
            25 : begin
                     if( GetTopWindow( Handle ) = 0 ) then
                     begin
                         Result := TText_Debugger.Create ;
                     end else
                     begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle := GetTopWindow( Handle ) ;
                     end ;
                     S := Window_Description( GetTopWindow( Handle ) ) ;
                     if( length( S ) > 0 ) then
                     begin
                         S := ' "' + S + '"' ;
                     end ;
                     S := 'Topmost child = ' + inttostr( GetTopWindow( Handle ) ) + S ;
                 end ;
            26 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Unicode = ' +
                         Boolean_To_String( IsWindowUnicode( Handle ) ) ;
                 end ;
            27 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Userdata = ' +
                         inttostr( GetWindowLong( Handle, GWL_USERDATA ) ) ;
                 end ;
            28 : begin
                     Result := TText_Debugger.Create ;
                     S := 'Visible = ' +
                         Boolean_To_String( IsWindowVisible( Handle ) ) ;
                 end ;
            29 : begin
                     Result := TDC_Debugger.Create ;
                     DC := GetWindowDC( Handle ) ;
                     TDC_Debugger( Result ).DC := DC ;
                     S := 'Window Device Context = ' + inttostr( DC ) ; // TODO: Need to release DC
                 end ;
            30 : begin
                     Result := TText_Debugger.Create ;
                     GetUpdateRect( Handle, R, False ) ;
                     S := 'Update Rect = ' + Rect_To_String( R ) ;
                 end ;
            31 : begin
                     Result := TText_Debugger.Create ;
                     Rgn := CreateRectRgn( 1, 1, 1, 1 ) ;
                     case GetUpdateRgn( Handle, Rgn, False ) of
                         COMPLEXREGION : S := 'complex' ;
                         ERROR : S := 'error' ;
                         NULLREGION : S := 'empty' ;
                         SIMPLEREGION : S := 'simple' ;
                     end ;
                     S := 'Update Region = ' + S ; //TODO: Show region detail
                     DeleteObject( Rgn ) ;
                 end ;
            32 : begin
                     Result := TWindowPlacement_Debugger.Create ;
                     TWindowPlacement_Debugger( Result ).Handle := Handle ;
                     S := 'Window Placement' ;
                 end ;
            33 : begin
                     Result := TText_Debugger.Create ;
                     S := 'WndProc = ' +
                         inttostr( GetWindowLong( Handle, GWL_WNDPROC ) ) ;
                 end ;
        end ;
    except
        on E : Exception do
            begin
                Result := TText_Debugger.Create ;
                S := '* exception *' + E.Message ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ; // TWindow_Debugger.Child


function TWindow_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Children := TInteger_List.Create ;
    end ;
    EnumChildWindows( Handle, @CB_ENumChildWindows, integer( Children ) ) ;
    Result := 34 ;
end ;



type TDefault_System_Locale_Short_Date_Debugger = class( TText_Debugger )
                                                     public // Constructors and destructors...
                                                         destructor Destroy ; override ;

                                                     private
                                                         Children : TStringList ;

                                                     public // API...
                                                         function Child( Index : longint ) : PDebug_Interface ;
                                                             override ;
                                                         function Count : longint ; override ;
                                                 end ;

// Constructors and destructors...

destructor TDefault_System_Locale_Short_Date_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TDefault_System_Locale_Short_Date_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( Children[ Index ] ) ;
end ;


function TDefault_System_Locale_Short_Date_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    EnumDateFormats_List := Children ;
    EnumDateFormats( @CB_EnumDateFormats, LOCALE_SYSTEM_DEFAULT, DATE_SHORTDATE ) ;
    Count := Children.Count ;
    EnumDateFormats_List := nil ;
end ;



type TDefault_System_Locale_Long_Date_Debugger = class( TText_Debugger )
                                                     public // Constructors and destructors...
                                                         destructor Destroy ; override ;

                                                     private
                                                         Children : TStringList ;

                                                     public // API...
                                                         function Child( Index : longint ) : PDebug_Interface ;
                                                             override ;
                                                         function Count : longint ; override ;
                                                 end ;

// Constructors and destructors...

destructor TDefault_System_Locale_Long_Date_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TDefault_System_Locale_Long_Date_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( Children[ Index ] ) ;
end ;


function TDefault_System_Locale_Long_Date_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    EnumDateFormats_List := Children ;
    EnumDateFormats( @CB_EnumDateFormats, LOCALE_SYSTEM_DEFAULT, DATE_LONGDATE ) ;
    Count := Children.Count ;
    EnumDateFormats_List := nil ;
end ;



type TDefault_System_Locale_Debugger = class( TText_Debugger )
                                           public // API...
                                               function Child( Index : longint ) : PDebug_Interface ;
                                                   override ;
                                               function Count : longint ; override ;
                                       end ;

// API...

function TDefault_System_Locale_Debugger.Child( Index : longint ) : PDebug_Interface ;

    function Get_Locale_Info( ID : integer ; const Prefix : string ) : string ;

    var Size : integer ;

    begin
        Size := GetLocaleInfo( LOCALE_SYSTEM_DEFAULT, ID, nil, 0 ) - 1 ;
        setlength( Result, Size ) ;
        GetLocaleInfo( LOCALE_SYSTEM_DEFAULT, ID, PChar( Result ), Size ) ;
        Result := Prefix + ' = "' + Result + '"' ;
    end ;

var S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TDefault_System_Locale_Long_Date_Debugger.Create ;
                S := 'Long date formats' ;
            end ;
        1 : begin
                Result := TDefault_System_Locale_Short_Date_Debugger.Create ;
                S := 'Short date formats' ;
            end ;
        2 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_ILANGUAGE, 'Language ID' ) ;
            end ;
        3 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SLANGUAGE, 'Localized language name' ) ;
            end ;
        4 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SENGLANGUAGE, 'English name of language' ) ;
            end ;
        5 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SABBREVLANGNAME, 'Abbreviated language name' ) ;
            end ;
        6 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SNATIVELANGNAME, 'Native name of language' ) ;
            end ;
        7 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_ICOUNTRY, 'Country code' ) ;
            end ;
        8 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SCOUNTRY, 'Country name' ) ;
            end ;
        9 : begin
                Result := TText_Debugger.Create ;
                S := Get_Locale_Info( LOCALE_SENGCOUNTRY, 'English name of Country' ) ;
            end ;
        10 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVCTRYNAME, 'Abbreviated country name' ) ;
             end ;
        11 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SNATIVECTRYNAME, 'Native name of country' ) ;
             end ;
        12 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDEFAULTLANGUAGE, 'Default Language ID' ) ;
             end ;
        13 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDEFAULTCOUNTRY, 'Default country code' ) ;
             end ;
        14 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDEFAULTCODEPAGE, 'Default OEM code page' ) ;
             end ;
        15 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDEFAULTANSICODEPAGE, 'Default ANSI code page' ) ;
             end ;
        16 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDEFAULTMACCODEPAGE, 'Default Mac code page' ) ;
             end ;
        17 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IMEASURE, '' ) ;
                 if( S = ' = "0"' ) then
                 begin
                     S := 'Metric' ;
                 end else
                 begin
                     S := 'US (nonmetric)' ;
                 end ;
             end ;
        18 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDECIMAL, 'Decimal separator' ) ;
             end ;
        19 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_STHOUSAND, 'Thousand separator' ) ;
             end ;
        20 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SGROUPING, 'Digit grouping' ) ;
             end ;
        21 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDIGITS, 'Number of fractional digits' ) ;
             end ;
        22 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ILZERO, 'Leading zeros for decimal' ) ;
             end ;
        23 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_INEGNUMBER, 'Negative number mode' ) ;
             end ;
        24 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SNATIVEDIGITS, 'Native ASCII 0-9' ) ;
             end ;
        25 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SCURRENCY, 'Local monetary symbol' ) ;
             end ;
        26 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SINTLSYMBOL, 'International monetary symbol' ) ;
             end ;
        27 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONDECIMALSEP, 'Monetary decimal separator' ) ;
             end ;
        28 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHOUSANDSEP, 'Monetary thousand separator' ) ;
             end ;
        29 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONGROUPING, 'Monetary grouping' ) ;
             end ;
        30 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ICURRDIGITS, 'Number local monetary digits' ) ;
             end ;
        31 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IINTLCURRDIGITS, 'Number international monetary digits' ) ;
             end ;
        32 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ICURRENCY, 'Positive currency mode' ) ;
             end ;
        33 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_INEGCURR, 'Negative currency mode' ) ;
             end ;
        34 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDATE, 'Date separator' ) ;
             end ;
        35 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_STIME, 'Time separator' ) ;
             end ;
        36 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SSHORTDATE, 'Short date format string' ) ;
             end ;
        37 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SLONGDATE, 'Long date format string' ) ;
             end ;
        38 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_STIMEFORMAT, 'Time format string' ) ;
             end ;
        39 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDATE, 'Short date format ordering' ) ;
             end ;
        40 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ILDATE, 'Long date format ordering' ) ;
             end ;
        41 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ITIME, 'Time format specifier' ) ;
             end ;
        42 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ITIMEMARKPOSN, 'Time marker position' ) ;
             end ;
        43 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ICENTURY, 'Century format specifier (short date)' ) ;
             end ;
        44 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ITLZERO, 'Leading zeros in time field' ) ;
             end ;
        45 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IDAYLZERO, 'Leading zeros in day field (short date)' ) ;
             end ;
        46 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IMONLZERO, 'Leading zeros in month field (short date)' ) ;
             end ;
        47 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_S1159, 'AM designator' ) ;
             end ;
        48 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_S2359, 'PM designator' ) ;
             end ;
        49 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_ICALENDARTYPE, 'Type of calendar specifier' ) ;
             end ;
        50 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IOPTIONALCALENDAR, 'Additional calendar types specifier' ) ;
             end ;
        51 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IFIRSTDAYOFWEEK, 'First day of week specifier' ) ;
             end ;
        52 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IFIRSTWEEKOFYEAR, 'First week of year specifier' ) ;
             end ;
        53 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME1, 'Long name for Monday' ) ;
             end ;
        54 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME2, 'Long name for Tuesday' ) ;
             end ;
        55 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME3, 'Long name for Wednesday' ) ;
             end ;
        56 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME4, 'Long name for Thursday' ) ;
             end ;
        57 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME5, 'Long name for Friday' ) ;
             end ;
        58 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME6, 'Long name for Saturday' ) ;
             end ;
        59 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SDAYNAME7, 'Long name for Sunday' ) ;
             end ;
        60 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME1, 'Abbreviated for Monday' ) ;
             end ;
        61 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME2, 'Abbreviated for Tuesday' ) ;
             end ;
        62 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME3, 'Abbreviated for Wednesday' ) ;
             end ;
        63 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME4, 'Abbreviated for Thursday' ) ;
             end ;
        64 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME5, 'Abbreviated for Friday' ) ;
             end ;
        65 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME6, 'Abbreviated for Saturday' ) ;
             end ;
        66 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVDAYNAME7, 'Abbreviated for Sunday' ) ;
             end ;
        67 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME1, 'Long name for January' ) ;
             end ;
        68 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME2, 'Long name for February' ) ;
             end ;
        69 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME3, 'Long name for March' ) ;
             end ;
        70 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME4, 'Long name for April' ) ;
             end ;
        71 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME5, 'Long name for May' ) ;
             end ;
        72 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME6, 'Long name for June' ) ;
             end ;
        73 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME7, 'Long name for July' ) ;
             end ;
        74 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME8, 'Long name for August' ) ;
             end ;
        75 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME9, 'Long name for September' ) ;
             end ;
        76 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME10, 'Long name for October' ) ;
             end ;
        77 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME11, 'Long name for November' ) ;
             end ;
        78 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME12, 'Long name for December' ) ;
             end ;
        79 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SMONTHNAME13, 'Long name for 13th Moneth' ) ;
             end ;
        80 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME1, 'Abbreviated name for January' ) ;
             end ;
        81 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME2, 'Abbreviated name for February' ) ;
             end ;
        82 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME3, 'Abbreviated name for March' ) ;
             end ;
        83 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME4, 'Abbreviated name for April' ) ;
             end ;
        84 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME5, 'Abbreviated name for May' ) ;
             end ;
        85 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME6, 'Abbreviated name for June' ) ;
             end ;
        86 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME7, 'Abbreviated name for July' ) ;
             end ;
        87 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME8, 'Abbreviated name for August' ) ;
             end ;
        88 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME9, 'Abbreviated name for September' ) ;
             end ;
        89 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME10, 'Abbreviated name for October' ) ;
             end ;
        90 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME11, 'Abbreviated name for November' ) ;
             end ;
        91 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME12, 'Abbreviated name for December' ) ;
              end ;
        92 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SABBREVMONTHNAME13, 'Abbreviated name for 13th month' ) ;
             end ;
        93 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SPOSITIVESIGN, 'Positive sign' ) ;
             end ;
        94 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_SNEGATIVESIGN, 'Negative sign' ) ;
             end ;
        95 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IPOSSIGNPOSN, 'Positive sign position' ) ;
             end ;
        96 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_INEGSIGNPOSN, 'Negative sign position' ) ;
             end ;
        97 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IPOSSYMPRECEDES, 'Monetary symbol precedes positive amount' ) ;
             end ;
        98 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_IPOSSEPBYSPACE, 'Monetary symbol separated by space from positive amount' ) ;
             end ;
        99 : begin
                 Result := TText_Debugger.Create ;
                 S := Get_Locale_Info( LOCALE_INEGSYMPRECEDES, 'Monetary symbol precedes negative amount' ) ;
             end ;
        100 : begin
                  Result := TText_Debugger.Create ;
                  S := Get_Locale_Info( LOCALE_INEGSEPBYSPACE, 'Monetary symbol separated by space from negative amount' ) ;
              end ;
        101 : begin
                  Result := TText_Debugger.Create ;
                  S := Get_Locale_Info( LOCALE_FONTSIGNATURE, 'Font signature' ) ;
              end ;
        102 : begin
                  Result := TText_Debugger.Create ;
                  S := Get_Locale_Info( LOCALE_SISO639LANGNAME, 'ISO abbreviated language name' ) ;
              end ;
        103 : begin
                  Result := TText_Debugger.Create ;
                  S := Get_Locale_Info( LOCALE_SISO3166CTRYNAME, 'ISO abbreviated country name' ) ;
              end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDefault_System_Locale_Debugger.Count : longint ;

begin
    Count := 104 ;
end ;



type TDefault_User_Locale_Short_Data_Debugger = class( TText_Debugger )
                                                     public // Constructors and destructors...
                                                         destructor Destroy ; override ;

                                                     private
                                                         Children : TStringList ;

                                                     public // API...
                                                         function Child( Index : longint ) : PDebug_Interface ;
                                                             override ;
                                                         function Count : longint ; override ;
                                                 end ;

// Constructors and destructors...

destructor TDefault_User_Locale_Short_Data_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TDefault_User_Locale_Short_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( Children[ Index ] ) ;
end ;


function TDefault_User_Locale_Short_Data_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    EnumDateFormats_List := Children ;
    EnumDateFormats( @CB_EnumDateFormats, LOCALE_USER_DEFAULT, DATE_SHORTDATE ) ;
    Count := Children.Count ;
    EnumDateFormats_List := nil ;
end ;



type TDefault_User_Locale_Long_Data_Debugger = class( TText_Debugger )
                                                     public // Constructors and destructors...
                                                         destructor Destroy ; override ;

                                                     private
                                                         Children : TStringList ;

                                                     public // API...
                                                         function Child( Index : longint ) : PDebug_Interface ;
                                                             override ;
                                                         function Count : longint ; override ;
                                                 end ;

// Constructors and destructors...

destructor TDefault_User_Locale_Long_Data_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TDefault_User_Locale_Long_Data_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( Children[ Index ] ) ;
end ;


function TDefault_User_Locale_Long_Data_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    EnumDateFormats_List := Children ;
    EnumDateFormats( @CB_EnumDateFormats, LOCALE_USER_DEFAULT, DATE_LONGDATE ) ;
    Count := Children.Count ;
    EnumDateFormats_List := nil ;
end ;



type TDefault_User_Locale_Debugger = class( TText_Debugger )
                                         public // API...
                                             function Child( Index : longint ) : PDebug_Interface ;
                                                 override ;
                                             function Count : longint ; override ;
                                     end ;

// API...

function TDefault_User_Locale_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TDefault_User_Locale_Long_Data_Debugger.Create ;
                S := 'Long date formats' ;
            end ;
        1 : begin
                Result := TDefault_User_Locale_Short_Data_Debugger.Create ;
                S := 'Short date formats' ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDefault_User_Locale_Debugger.Count : longint ;

begin
    Count := 2 ;
end ;



type TInstalled_Locales_Debugger = class( TText_Debugger )
                                     public // Constructors and destructors...
                                         destructor Destroy ; override ;

                                     private
                                         Children : TStringList ;

                                     public // API...
                                         function Child( Index : longint ) : PDebug_Interface ;
                                             override ;
                                         function Count : longint ; override ;
                                 end ;

// Constructors and destructors...

destructor TInstalled_Locales_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TInstalled_Locales_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' + Children[ Index ] ) ;
end ;


function TInstalled_Locales_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    _EnumSystemLocales_List := Children ;
    EnumSystemLocales( @CB_EnumSystemLocales, LCID_INSTALLED ) ;
    Count := Children.Count ;
    _EnumSystemLocales_List := nil ;
end ;



type TSupported_Locales_Debugger = class( TText_Debugger )
                                     public // Constructors and destructors...
                                         destructor Destroy ; override ;

                                     private
                                         Children : TStringList ;

                                     public // API...
                                         function Child( Index : longint ) : PDebug_Interface ;
                                             override ;
                                         function Count : longint ; override ;
                                 end ;

// Constructors and destructors...

destructor TSupported_Locales_Debugger.Destroy ;

begin
    Children.Free ;

    inherited Destroy ;
end ;


// API...

function TSupported_Locales_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ;
    Result.Title := PChar( inttostr( Index ) + ': ' + Children[ Index ] ) ;
end ;


function TSupported_Locales_Debugger.Count : longint ;

begin
    if( Children = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    _EnumSystemLocales_List := Children ;
    EnumSystemLocales( @CB_EnumSystemLocales, LCID_SUPPORTED ) ;
    Count := Children.Count ;
    _EnumSystemLocales_List := nil ;
end ;



type TLocales_Debugger = class( TText_Debugger )
                             public // API...
                                 function Child( Index : longint ) : PDebug_Interface ;
                                     override ;
                                 function Count : longint ; override ;
                         end ;

function TLocales_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                S := 'Installed Locales' ;
                Result := TInstalled_Locales_Debugger.Create ;
            end ;
        1 : begin
                S := 'Supported Locales' ;
                Result := TSupported_Locales_Debugger.Create ;
            end ;
        2 : begin
                S := 'Default System Locale' ;
                Result := TDefault_System_Locale_Debugger.Create ;
            end ;
        3 : begin
                S := 'Default User Locale' ;
                Result := TDefault_User_Locale_Debugger.Create ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TLocales_Debugger.Count : longint ;

begin
    Result := 4 ;
end ;



type TSoundSentry_Debugger = class( TText_Debugger )
                                 public // API...
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : integer ; override ;
                             end ;

function TSoundSentry_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;
    SS : TSoundSentry ;

begin
    fillchar( SS, sizeof( SS ), 0 ) ;
    SS.cbSize := sizeof( SS ) ;
    SystemParametersInfo( SPI_GETSOUNDSENTRY, sizeof( SS ), @SS, 0 ) ;
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                S := Add_Flag( SS.dwFlags, SSF_AVAILABLE, '', 'Available' ) ;
                S := Add_Flag( SS.dwFlags, SSF_SOUNDSENTRYON, S, 'On' ) ;
                S := 'Flags = ' + S ;
            end ;
        1 : case SS.iFSTextEffect of
                SSTF_BORDER : S := 'FSTextEffect = SSTF_BORDER' ;
                SSTF_CHARS : S := 'FSTextEffect = SSTF_CHARS' ;
                SSTF_DISPLAY : S := 'FSTextEffect = SSTF_DISPLAY' ;
                SSTF_NONE : S := 'FSTextEffect = SSTF_NONE' ;
            end ;
        2 : S := 'FSTextEffectMSec = ' + inttostr( SS.iFSTextEffectMSec ) ;
        3 : S := 'FSTextEffectColorBits = ' + RGB_To_String( SS.iFSTextEffectColorBits ) ;
        4 : case SS.iFSGrafEffect of
                SSGF_DISPLAY : S := 'FSGrafEffect = SSGF_DISPLAY' ;
                SSGF_NONE : S := 'FSGrafEffect = SSGF_NONE' ;
                else S := 'FSGrafEffect = ' + inttostr( SS.iFSGrafEffect ) ;
            end ;
        5 : S := 'FSGrafEffectMSec = ' + inttostr( SS.iFSGrafEffectMSec ) ;
        6 : S := 'FSGrafEffectColor = ' + RGB_To_String( SS.iFSGrafEffectColor ) ;
        7 : case SS.iWindowsEffect of
                SSWF_CUSTOM : S := 'WindowsEffect = SSWF_CUSTOM' ;
                SSWF_DISPLAY : S := 'WindowsEffect = SSWF_DISPLAY' ;
                SSWF_NONE : S := 'WindowsEffect = SSWF_NONE' ;
                SSWF_TITLE : S := 'WindowsEffect = SSWF_TITLE' ;
                SSWF_WINDOW : S := 'WindowsEffect = SSWF_WINDOW' ;
                else S := 'WindowsEffect = ' + inttostr( SS.iWindowsEffect ) ;
            end ;
        8 : S := 'WindowsEffectMSec = ' + inttostr( SS.iWindowsEffectMSec ) ;
        9 : if( SS.lpszWindowsEffectDLL = nil ) then
            begin
                S := 'Windows Effect DLL = ""' ;
            end else
            begin
                S := 'Windows Effect DLL = "' + SS.lpszWindowsEffectDLL + '"' ;
            end ;
        10 : S := 'WindowsEffectOrdinal = ' + inttostr( SS.iWindowsEffectOrdinal ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TSoundSentry_Debugger.Count : integer ;

begin
    Result := 11 ;
end ;


function Supported_Function( const DLL, Fun : string ) : boolean ;

var H : THandle ;

begin
    Result := False ;
    H := LoadLibrary( PChar( DLL ) ) ;
    if( H <> 0 ) then
    begin
        if( GetProcAddress( H, PChar( Fun ) ) <> nil ) then
        begin
            Result := True ;
        end ;
        FreeLibrary( H ) ;
    end ;
end ;



type TShell_UI_Component_Debugger = class( TText_Debugger )
                                 public // API...
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

type TGetDpiForShellUiComponent = function( Component : integer ) : cardinal ; stdcall ;

function TShell_UI_Component_Debugger.Child( Index : longint ) : PDebug_Interface ;

var GetDpiForShellUiComponent : TGetDpiForShellUiComponent ;
    H : THandle ;
    S : string ;

begin
    S := '' ;
    H := LoadLibrary( 'SHCore.dll' ) ;
    if( H <> 0 ) then
    begin
        GetDpiForShellUiComponent := TGetDpiForShellUiComponent( GetProcAddress( H, 'GetDpiForShellUiComponent' ) ) ;
        if( assigned( GetDpiForShellUiComponent ) ) then
        begin
            S := inttostr( GetDpiForShellUiComponent( Index ) ) ;
        end ;
        FreeLibrary( H ) ;
    end ;
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'Taskbars = ' + S ;
        1 : S := 'Notification Area = ' + S ;
        2 : S := 'Deskband = ' + S ;
        else S := inttostr( Index ) + ' = ' + S ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TShell_UI_Component_Debugger.Count : longint ;

begin
    Count := 3 ;
end ;



type TClipboard_Formats_Debugger = class( TText_Debugger )
                                       public // API...
                                           function Child( Index : longint ) : PDebug_Interface ;
                                               override ;
                                           function Count : longint ; override ;
                                   end ;

function TClipboard_Formats_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Context : integer ;
    S : string ;

begin
    Context := EnumClipboardFormats( 0 ) ;
    while( Index > 0 ) do
    begin
        Context := EnumClipboardFormats( Context ) ;
    end ;
    setlength( S, 256 ) ;
    setlength( S, GetClipboardFormatName( Context, @S[ 1 ], 256 ) ) ;
    Result := TText_Debugger.Create ;
    Result.Title := PChar( S ) ;
    //~~~show data (GetClipboardData)
end ;


function TClipboard_Formats_Debugger.Count : longint ;

begin
    Result := CountClipboardFormats ;
end ;



type TClipboard_Debugger = class( TText_Debugger )
                               public // API...
                                   function Child( Index : longint ) : PDebug_Interface ;
                                       override ;
                                   function Count : longint ; override ;
                           end ;

function TClipboard_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := nil ;
    case Index of
       0 : begin
               S := 'Formats' ;
               Result := TClipboard_Formats_Debugger.Create ;
           end ;
       1 : begin
               S := 'Opened by ' + inttostr( GetOpenClipboardWindow ) + ' ' +
                   Window_Description( GetOpenClipboardWindow ) ;
               Result := TWindow_Debugger.Create ;
               TWindow_Debugger( Result ).Handle := GetOpenClipboardWindow ;
           end ;
       2 : begin
               S := 'Owned by ' + inttostr( GetClipboardOwner ) + ' ' +
                   Window_Description( GetClipboardOwner ) ;
               Result := TWindow_Debugger.Create ;
               TWindow_Debugger( Result ).Handle := GetClipboardOwner ;
           end ;
       3 : begin
               S := 'Viewer ' + inttostr( GetClipboardViewer ) + ' ' +
                  Window_Description( GetClipboardViewer ) ;
               Result := TWindow_Debugger.Create ;
               TWindow_Debugger( Result ).Handle := GetClipboardViewer ;
           end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TClipboard_Debugger.Count : longint ;

begin
    Count := 4 ;
end ;



type TNonClient_Metrics_Debugger = class( TText_Debugger )
                                       public // API...
                                           function Child( Index : longint ) : PDebug_Interface ;
                                               override ;
                                           function Count : longint ; override ;
                                   end ;

function TNonClient_Metrics_Debugger.Child( Index : longint ) : PDebug_Interface ;

var NCM : TNonClientMetrics ;
    S : string ;

begin
    fillchar( NCM, sizeof( NCM ), 0 ) ;
    NCM.cbSize := sizeof( NCM ) ;
    SystemParametersInfo( SPI_GETNONCLIENTMETRICS, sizeof( NCM ), @NCM, 0 ) ;
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'BorderWidth = ' + inttostr( NCM.iBorderWidth ) ;
        1 : S := 'ScrollWidth = ' + inttostr( NCM.iScrollWidth ) ;
        2 : S := 'ScrollHeight = ' + inttostr( NCM.iScrollHeight ) ;
        3 : S := 'CaptionWidth = ' + inttostr( NCM.iCaptionWidth ) ;
        4 : S := 'CaptionHeight = ' + inttostr( NCM.iCaptionHeight ) ;
        5 : begin
                Result.Free ;
                Result := TLogFont_Debugger.Create ;
                TLogFont_Debugger( Result ).Logfont := NCM.lfCaptionFont ;
            end ;
        6 : S := 'SmCaptionWidth = ' + inttostr( NCM.iSmCaptionWidth ) ;
        7 : S := 'SmCaptionHeight = ' + inttostr( NCM.iSmCaptionHeight ) ;
        8 : begin
                Result.Free ;
                Result := TLogFont_Debugger.Create ;
                TLogFont_Debugger( Result ).Logfont := NCM.lfSmCaptionFont ;
            end ;
        9 : S := 'MenuWidth = ' + inttostr( NCM.iMenuWidth ) ;
        10 : S := 'MenuHeight = ' + inttostr( NCM.iMenuHeight ) ;
        11 : begin
                 Result.Free ;
                 Result := TLogFont_Debugger.Create ;
                 TLogFont_Debugger( Result ).Logfont := NCM.lfMenuFont ;
             end ;
        12 : begin
                 Result.Free ;
                 Result := TLogFont_Debugger.Create ;
                 TLogFont_Debugger( Result ).Logfont := NCM.lfStatusFont ;
             end ;
        13 : begin
                 Result.Free ;
                 Result := TLogFont_Debugger.Create ;
                 TLogFont_Debugger( Result ).Logfont := NCM.lfMessageFont ;
             end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TNonClient_Metrics_Debugger.Count : longint ;

begin
    Result := 14 ;
end ;



type TSerial_Keys_Debugger = class( TText_Debugger )
                                 public
                                     SK : TSerialKeys ;
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;


function TSerial_Keys_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : if( SK.lpszActivePort = nil ) then
            begin
                S := 'ActivePort = ""' ;
            end else
            begin
                S := 'ActivePort = "' + SK.lpszActivePort + '"' ;
            end ;
        1 : S := 'Baud rate = ' + inttostr( SK.iBaudRate ) ;
        2 : begin
                S := 'Port state = ' + inttostr( SK.iPortState ) ;
                case SK.iPortState of
                    0 : S := S + ' (input ignored)' ;
                    1 : S := S + ' (input is serial keys when no application has the port open)' ;
                    2 : S := S + ' (all input is serial keys commands)' ;
                end ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TSerial_Keys_Debugger.Count : longint ;

begin
    if( SK.lpszActivePort = nil ) then
    begin
        Result := 2 ;
    end else
    begin
        Result := 3 ;
    end ;
end ;


const SERKF_ACTIVE = 8 ;

type TSystem_Colors_Debugger = class( TText_Debugger )
                                   public
                                       function Child( Index : longint ) : PDebug_Interface ;
                                           override ;
                                       function Count : longint ; override ;
                               end ;

function TSystem_Colors_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Color : TColor ;
    S : string ;

begin
    Color := GetSysColor( Index ) ;
    Result := TText_Debugger.Create ;
    case Index of
        COLOR_3DDKSHADOW : S := '3D Dark Shadow' ;
        COLOR_BTNFACE : S := 'Button face' ;
        COLOR_BTNHIGHLIGHT : S := 'Button highlight' ;
        COLOR_3DLIGHT : S := '3D light' ;
        COLOR_BTNSHADOW : S := 'Button shadow' ;
        COLOR_ACTIVEBORDER : S := 'Active border' ;
        COLOR_ACTIVECAPTION : S := 'Active caption' ;
        COLOR_APPWORKSPACE : S := 'App workspace' ;
        COLOR_DESKTOP : S := 'Desktop' ;
        COLOR_BTNTEXT : S := 'Button text' ;
        COLOR_CAPTIONTEXT : S := 'Captino text' ;
        COLOR_GRAYTEXT : S := 'Grayed text' ;
        COLOR_HIGHLIGHT : S := 'Highlight' ;
        COLOR_HIGHLIGHTTEXT : S := 'Highlight text' ;
        COLOR_INACTIVEBORDER : S := 'Inactive border' ;
        COLOR_INACTIVECAPTION : S := 'Inactive caption' ;
        COLOR_INACTIVECAPTIONTEXT : S := 'Inactive caption text' ;
        COLOR_INFOBK : S := 'Tooltip background' ;
        COLOR_INFOTEXT : S := 'Tooltip text' ;
        COLOR_MENU : S := 'Menu' ;
        COLOR_MENUTEXT : S := 'Menu text' ;
        COLOR_SCROLLBAR : S := 'Scrollbar' ;
        COLOR_WINDOW : S := 'Window' ;
        COLOR_WINDOWFRAME : S := 'Window frame' ;
        COLOR_WINDOWTEXT : S := 'Window text' ;
        else S := inttostr( Index ) ;
    end ;
    S := S + ': ' + RGB_To_String( Color ) ;
    Result.Title := PChar( S ) ;
end ;


function TSystem_Colors_Debugger.Count : longint ;

begin
    Count := 31 ;
end ;


type TPage_File_Debugger = class( TText_Debugger )
                               private
                                   _Files : string ;

                               public
                                   function Child( Index : longint ) : PDebug_Interface ;
                                       override ;
                                   function Count : longint ; override ;
                           end ;

function TPage_File_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Dummy : integer ;
    S : string ;

begin
    if( _Files = '' ) then
    begin
        _Files := OS.Swap_File_List ;
    end ;
    S := _Files + ';' ;
    Dummy := pos( ';', S ) ;
    while( Index > 0 ) do
    begin
        dec( Index ) ;
        S := copy( S, Dummy + 1, length( S ) ) ;
        Dummy := pos( ';', S ) ;
    end ;
    S := copy( S, 1, Dummy - 1 ) ;
    Result := TFile_Debugger.Create ;
    TFile_Debugger( Result ).Title := PChar( S ) ;
    TFile_Debugger( Result ).Name := S ;
end ;


function TPage_File_Debugger.Count : longint ;

var Loop : integer ;

begin
    if( _Files = '' ) then
    begin
        _Files := OS.Swap_File_List ;
    end ;
    Result := 1 ;
    for Loop := 1 to length( _Files ) do
    begin
        if( _Files[ Loop ] = ';' ) then
        begin
            inc( Result ) ;
        end ;
    end ;
end ;


type TDesktop_Debugger = class( TText_Debugger )
                            public
                                function Child( Index : longint ) : PDebug_Interface ;
                                    override ;
                                function Count : longint ; override ;
                        end ;


function TDesktop_Debugger.Child( Index : longint ) : PDebug_Interface ;

var I : integer ;
    List : THandle ;
    S : string ;
    SHFILEINFO : TSHFILEINFO ;
    X, Y : integer ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                I := 0 ;
                SystemParametersInfo( SPI_GETGRIDGRANULARITY, 0, @I, 0 ) ;
                S := 'Desktop sizing grid granularity = ' + inttostr( I ) ;
            end ;
        1 : begin
                fillchar( SHFILEINFO, sizeof( SHFILEINFO ), 0 ) ;
                List := SHGetFileInfo( '', 0, SHFILEINFO, sizeof( SHFILEINFO ),
                    SHGFI_SHELLICONSIZE or SHGFI_SMALLICON ) ;
                ImageList_GetIconSize( List, X, Y ) ;
                S := 'Small icon size = x: ' + inttostr( X ) + ', y: ' + inttostr( Y ) ;
            end ;
        2 : begin
                fillchar( SHFILEINFO, sizeof( SHFILEINFO ), 0 ) ;
                List := SHGetFileInfo( '', 0, SHFILEINFO, sizeof( SHFILEINFO ),
                    SHGFI_SHELLICONSIZE ) ;
                ImageList_GetIconSize( List, X, Y ) ;
                S := 'Large icon size = x: ' + inttostr( X ) + ', y: ' + inttostr( Y ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TDesktop_Debugger.Count : longint ;

begin
    Result := 3 ;
end ;


type TStock_Brush_Debugger = class( TText_Debugger )
                                 public
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TStock_Brush_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'Black = ' + inttostr( GetStockObject( BLACK_BRUSH ) ) ;
        1 : S := 'Dark Grey = ' + inttostr( GetStockObject( DKGRAY_BRUSH ) ) ;
        2 : S := 'Grey = ' + inttostr( GetStockObject( GRAY_BRUSH ) ) ;
        3 : S := 'Light Grey = ' + inttostr( GetStockObject( LTGRAY_BRUSH ) ) ;
        4 : S := 'Null = ' + inttostr( GetStockObject( NULL_BRUSH ) ) ;
        5 : S := 'White = ' + inttostr( GetStockObject( WHITE_BRUSH ) ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TStock_Brush_Debugger.Count : longint ;

begin
    Count := 6 ;
end ;


type TStock_Font_Debugger = class( TText_Debugger )
                                public
                                    function Child( Index : longint ) : PDebug_Interface ;
                                        override ;
                                    function Count : longint ; override ;
                            end ;

function TStock_Font_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;
    X : TOSVersionInfo ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'ANSI fixed = ' + inttostr( GetStockObject( ANSI_FIXED_FONT ) ) ;
        1 : S := 'ANSI variable = ' + inttostr( GetStockObject( ANSI_VAR_FONT ) ) ;
        2 : begin
                X.dwOSVersionInfoSize := sizeof( X ) ;
                GetVersionEx( X ) ;
                if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then // Windows 95
                begin
                    S := 'Default GUI = ' + inttostr( GetStockObject( DEFAULT_GUI_FONT ) ) ;
                end else
                begin
                    S := 'Default device = ' + inttostr( GetStockObject( DEVICE_DEFAULT_FONT ) ) ;
                end ;
            end ;
        3 : S := 'OEM fixed = ' + inttostr( GetStockObject( OEM_FIXED_FONT ) ) ;
        4 : S := 'System = ' + inttostr( GetStockObject( SYSTEM_FONT ) ) ;
        5 : S := 'System fixed = ' + inttostr( GetStockObject( SYSTEM_FIXED_FONT ) ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TStock_Font_Debugger.Count : longint ;

begin
    Count := 6 ;
end ;


type TStock_Pen_Debugger = class( TText_Debugger )
                               public
                                   function Child( Index : longint ) : PDebug_Interface ;
                                       override ;
                                   function Count : longint ; override ;
                           end ;

function TStock_Pen_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : S := 'Black = ' + inttostr( GetStockObject( BLACK_PEN ) ) ;
        1 : S := 'Null = ' + inttostr( GetStockObject( NULL_PEN ) ) ;
        2 : S := 'White = ' + inttostr( GetStockObject( WHITE_PEN ) ) ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TStock_Pen_Debugger.Count : longint ;

begin
    Count := 3 ;
end ;


type TStock_Object_Debugger = class( TText_Debugger )
                                   public
                                       function Child( Index : longint ) : PDebug_Interface ;
                                           override ;
                                       function Count : longint ; override ;
                               end ;

function TStock_Object_Debugger.Child( Index : longint ) : PDebug_Interface ;

var S : string ;

begin
    Result := nil ;
    case Index of
        0 : begin
                Result := TStock_Brush_Debugger.Create ;
                S := 'Brushes' ;
            end ;
        1 : begin
                Result := TStock_Font_Debugger.Create ;
                S := 'Fonts' ;
            end ;
        2 : begin
                Result := TStock_Pen_Debugger.Create ;
                S := 'Pens' ;
            end ;
        3 : begin
                S := 'Default_Palette = ' + inttostr( GetStockObject( DEFAULT_PALETTE ) ) ;
            end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TStock_Object_Debugger.Count : longint ;

begin
    Count := 4 ;
end ;


type TMessage_Queue_Debugger = class( TText_Debugger )
                                   private
                                       Types : array[ 0..8 ] of boolean ;

                                   public
                                       function Child( Index : longint ) : PDebug_Interface ;
                                           override ;
                                       function Count : longint ; override ;
                               end ;


function TMessage_Queue_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Loop : integer ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    for Loop := 0 to 8 do
    begin
        if( Types[ Loop ] ) then
        begin
            dec( Index ) ;
            if( Index = 0 ) then
            begin
                case Loop of
                    0 : S := 'WM_HOTKEY is in the queue.' ;
                    1 : S := 'Input message is in the queue.' ;
                    2 : S := 'WM_KEYUP, WM_KEYDOWN, WM_SYSKEYUP, or WM_SYSKEYDOWN is in the queue.' ;
                    3 : S := 'Mouse-button message is in the queue.' ;
                    4 : S := 'WM_MOUSEMOVE message is in the queue.' ;
                    5 : S := 'WM_PAINT message is in the queue.' ;
                    6 : S := 'Posted message (other than those listed) is in the queue.' ;
                    7 : S := 'Message sent by another thread or application is in the queue.' ;
                    8 : S := 'WM_TIMER is in the queue.' ;
                end ;
                break ;
            end ;
        end ;
    end ;
    Result.Title := PChar( S ) ;
end ;


function TMessage_Queue_Debugger.Count : longint ;

begin
    fillchar( Types, sizeof( Types ), 0 ) ;
    Result := 0 ;
    if( GetQueueStatus( QS_HOTKEY ) <> 0 ) then
    begin
        Types[ 0 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_INPUT ) <> 0 ) then
    begin
        Types[ 1 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_KEY ) <> 0 ) then
    begin
        Types[ 2 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_MOUSEBUTTON ) <> 0 ) then
    begin
        Types[ 3 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_MOUSEMOVE ) <> 0 ) then
    begin
        Types[ 4 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_PAINT ) <> 0 ) then
    begin
        Types[ 5 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_POSTMESSAGE ) <> 0 ) then
    begin
        Types[ 6 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_SENDMESSAGE ) <> 0 ) then
    begin
        Types[ 7 ] := True ;
        inc( Result ) ;
    end ;
    if( GetQueueStatus( QS_TIMER ) <> 0 ) then
    begin
        Types[ 8 ] := True ;
        inc( Result ) ;
    end ;
end ; // TMessage_Queue_Debugger.Count



type TFonts_Debugger = class( TText_Debugger )
                           public
                               destructor Destroy ; override ;

                           private
                               Fonts : TStringList ;

                           public
                               function Child( Index : longint ) : PDebug_Interface ;
                                   override ;
                               function Count : longint ; override ;
                       end ;

destructor TFonts_Debugger.Destroy ;

begin
    Fonts.Free ;

    inherited Destroy ;
end ;


function TFonts_Debugger.Child( Index : longint ) : PDebug_Interface ;

begin
    Result := TText_Debugger.Create ; // ~~~
    Result.Title := PChar( Fonts[ Index ] ) ;
end ;


function CB_Enum_Fonts( LogFont : pLogFont ; TextMetric : pTextMetric ;
    Typ, Context : integer ) : integer ; stdcall ;

begin
    Result := -1 ; // Don't stop
    TStringList( Context ).Add( LogFont^.lfFaceName ) ;
end ;


function TFonts_Debugger.Count : longint ;

var DC : HDC ;

begin
    if( Fonts = nil ) then
    begin
        Fonts := TStringList.Create ;
        DC := GetDC( 0 ) ;
        try
            EnumFontFamilies( DC, nil, @CB_Enum_Fonts, integer( Fonts ) ) ;
        finally
            ReleaseDC( 0, DC ) ;
        end ;
        Fonts.Sort ;
    end ;
    Result := Fonts.Count ;
end ;



type TFile_System_Debugger = class( TText_Debugger )
                                 public
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TFile_System_Debugger.Child( Index : longint ) : PDebug_Interface ;

var Folder : integer ;
    PPIDL : PItemIDList ;
    S, S1 : string ;

begin
    Result := TText_Debugger.Create ;
    Folder := 0 ;
    case Index of
        0 : begin
                S := 'CSIDL_BITBUCKET' ;
                Folder := $000A ; // CSIDL_BITBUCKET
            end ;
        1 : begin
                S := 'CSIDL_CONTROLS' ;
                Folder := $0003 ; // CSIDL_CONTROLS
            end ;
        2 : begin
                S := 'CSIDL_DESKTOP' ;
                Folder := 0 ; // CSIDL_DESKTOP
            end ;
        3 : begin
                S := 'CSIDL_DESKTOPDIRECTORY' ;
                Folder := $0010 ; // CSIDL_DESKTOPDIRECTORY
            end ;
        4 : begin
                S := 'CSIDL_DRIVES' ;
                Folder := $0011 ; // CSIDL_DRIVES
            end ;
        5 : begin
                S := 'CSIDL_FONTS' ;
                Folder := $0014 ; // CSIDL_FONTS ;
            end ;
        6 : begin
                S := 'CSIDL_NETHOOD' ;
                Folder := $0013 ; // CSIDL_NETHOOD ;
            end ;
        7 : begin
                S := 'CSIDL_NETWORK' ;
                Folder := $0012 ; // CSIDL_NETWORK ;
            end ;
        8 : begin
                S := 'CSIDL_PERSONAL' ; // "My documents"
                Folder := CSIDL_PERSONAL ;
            end ;
        9 : begin
                S := 'CSIDL_PRINTERS' ;
                Folder := CSIDL_PRINTERS ;
            end ;
        10 : begin
                S := 'CSIDL_PROGRAMS' ;
                Folder := CSIDL_PROGRAMS ;
            end ;
        11 : begin
                 S := 'CSIDL_RECENT' ;
                 Folder := $0008 ; // CSIDL_RECENT
             end ;
        12 : begin
                 S := 'CSIDL_SENDTO' ;
                 Folder := $0009 ; // CSIDL_SENDTO
             end ;
        13 : begin
                 S := 'CSIDL_STARTMENU' ;
                 Folder := $0016 ; // CSIDL_STARTMENU
             end ;
        14 : begin
                 S := 'CSIDL_STARTUP' ;
                 Folder := $0007 ; // CSIDL_STARTUP ;
             end ;
        15 : begin
                 S := 'CSIDL_TEMPLATES' ;
                 Folder := $0015 ; // CSIDL_TEMPLATES
             end ;
    end ; // case Index
    SHGetSpecialFolderLocation( 0, Folder, PPIDL ) ;
    setlength( S1, MAX_PATH ) ;
    if( SHGetPathFromIDList( PPIDL, PChar( S1 ) ) ) then
    begin
        setlength( S1, strlen( PChar( S1 ) ) ) ;
    end else
    begin
        S1 := '' ;
    end ;
    Result.Title := PChar( S + ' = "' + S1 + '"' ) ;
end ; // TFile_System_Debugger.Child


function TFile_System_Debugger.Count : longint ;

begin
    Result := 16 ;
end ;



type TIcon_System_Debugger = class( TText_Debugger )
                                 public
                                     function Child( Index : longint ) : PDebug_Interface ;
                                         override ;
                                     function Count : longint ; override ;
                             end ;

function TIcon_System_Debugger.Child( Index : longint ) : PDebug_Interface ;

var B : boolean ;
    IM : TIconMetrics ;
    LF : TLogFont ;
    S : string ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : begin
                fillchar( IM, sizeof( IM ), 0 ) ;
                IM.cbSize := sizeof( IM ) ;
                SystemParametersInfo( SPI_GETICONMETRICS, sizeof( IM ), @IM, 0 ) ;
                Result := TIconMetrics_Debugger.Create ;
                TIconMetrics_Debugger( Result ).Metrics := IM ;
                S := 'Icon metrics' ;
            end ;
        1 : begin
                fillchar( LF, sizeof( LF ), 0 ) ;
                IM.cbSize := sizeof( LF ) ;
                SystemParametersInfo( SPI_GETICONTITLELOGFONT, sizeof( LF ), @LF, 0 ) ;
                Result := TIconMetrics_Debugger.Create ;
                TLogFont_Debugger( Result ).LogFont := LF ;
                S := 'Icon title font' ;
            end ;
        2 : begin
                SystemParametersInfo( SPI_GETICONTITLEWRAP, 0, @B, 0 ) ;
                Result := TText_Debugger.Create ;
                S := 'Icon title wrapping = ' + Boolean_To_String( B ) ;
            end ;
    end ; // case Index
    Result.Title := PChar( S ) ;
end ; // TIcon_System_Debugger.Child


function TIcon_System_Debugger.Count : longint ;

begin
    Result := 3 ;
end ;



// TMS_Windows_Debugger methods...

// API...

function TMS_Windows_Debugger.Has_Children : boolean ;

begin
    Result := True ;
end ;


type TGetProcessDPIAwareness = function( hProcess : THandle ; var value : integer ) : HResult ; stdcall ;

type TTokenUser = packed record
                      User : TSidAndAttributes ;
                  end ;
     PTokenUser = ^TTokenUser ;

procedure TMS_Windows_Debugger.Build_Children ;

var AI : TAnimationInfo ;
    ATO : TAccessTimeout ;
    B : boolean ;
    D : DWord ;
    Dummy, Index : integer ;
    FK : TFilterKeys ;
    GetProcessDPIAwareness : TGetProcessDPIAwareness ;
    H : THandle ;
    HC : THighContrast ;
    hToken : THandle ;
    I : integer ;
    I64 : int64 ;
    MK : TMouseKeys ;
    P : TPoint ;
    Ptr : Pointer ;
    Rect : TRect ;
    Result : TCOM_Debug_Interface ;
    S : string ;
    SID : PSID ;
    SK : TStickyKeys ;
    SerialK : TSerialKeys ;
    TIC : TTokenInformationClass ;
    TK : TToggleKeys ;
    X : TOSVersionInfo ;
    SIDp : pointer ;
    SNU : SID_Name_Use ;
    SIDs : array[ 0..260 ] of char ;
    S2 : string ;
    D1 : DWord ;
    SIDA : PSIDIdentifierAuthority ;
    PC : PChar ;

begin
    X.dwOSVersionInfoSize := sizeof( X ) ;
    GetVersionEx( X ) ;

    for Index := 0 to 53 do
    begin
        try
            case Index of
                0 : begin
                        fillchar( ATO, sizeof( ATO ), 0 ) ;
                        ATO.cbSize := sizeof( ATO ) ;
                        SystemParametersInfo( SPI_GETACCESSTIMEOUT, sizeof( ATO ), @ATO, 0 ) ;
                        S := Add_Flag( ATO.dwFlags, ATF_ONOFFFEEDBACK, '', 'On/off feedback' ) ;
                        S := Add_Flag( ATO.dwFlags, 4, S, 'Available' ) ;
                        S := Add_Flag( ATO.dwFlags, ATF_TIMEOUTON, S, 'Timeout on' ) ;
                        Result := TText_Debugger.Create ;
                        S := 'Access Timeout = ' + inttostr( ATO.iTimeOutMSec ) + ', ' + S ;
                    end ;
                1 : begin
                        Result := TWindow_Debugger.Create ;
                        TWindow_Debugger( Result ).Handle := GetActiveWindow ;
                        S := Window_Description( GetActiveWindow ) ;
                        if( length( S ) > 0 ) then
                        begin
                            S := ' "' + S + '"' ;
                        end ;
                        S := 'Active window = ' + inttostr( GetActiveWindow ) + S ;
                    end ;
                2 : begin
                        fillchar( AI, sizeof( AI ), 0 ) ;
                        AI.cbSize := sizeof( AI ) ;
                        SystemParametersInfo( SPI_GETANIMATION, sizeof( AI ), @AI, 0 ) ;
                        Result := TText_Debugger.Create ;
                        S := 'Animation enabled = ' +
                            Boolean_To_String( AI.iMinAnimate <> 0 ) ;
                    end ;
                3 : begin
                        Result := TText_Debugger.Create ;
                        SystemParametersInfo( SPI_GETBORDER, 0, @I, 0 ) ;
                        S := 'Border multiplier factor = ' + inttostr( I ) ;
                    end ;
                4 : begin
                        Result := TText_Debugger.Create ;
                        S := 'Caret Blink Time = ' + inttostr( GetCaretBlinkTime ) ;
                    end ;
                5 : begin
                        Result := TText_Debugger.Create ;
                        GetCaretPos( P ) ;
                        S := 'Caret Position = ' + Point_To_string( P ) ;
                    end ;
                6 : begin
                        Result := TClipboard_Debugger.Create ;
                        S := 'Clipboard' ;
                    end ;
                7 : begin
                        Setlength( S, 127 ) ;
                        D := 127 ;
                        GetComputerName( PChar( S ), D ) ;
                        setlength( S, D - 1 ) ;
                        S := 'Computer Name = "' + S + '"' ;
                        Result := TText_Debugger.Create ;
                    end ;
                8 : begin
                        Result := TText_Debugger.Create ;
                        S := 'Current directory = ' + GetCurrentDir ;
                    end ;
                9 : begin
                        Result := TThread_Debugger.Create ;
                        TThread_Debugger( Result ).Handle := GetCurrentThreadID ;
                        S := 'Current thread = ' + inttostr( GetCurrentThreadID ) ;
                    end ;
                10 : begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle := GetFocus ;
                         S := 'Currently focused window = ' +
                             Window_Description( GetFocus ) ;
                     end ;
                11 : begin
                         Result := TCursor_Debugger.Create ;
                         TCursor_Debugger( Result ).Cursor := GetCursor ;
                         S := 'Cursor = ' + inttostr( GetCursor ) ;
                     end ;
                12 : begin
                        Result := TText_Debugger.Create ;
                        S := 'Date = ' + DateToStr( Now ) ;
                    end ;
                13 : begin
                         I := 0 ;
                         SystemParametersInfo( SPI_GETDEFAULTINPUTLANG, 0, @I, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Default input language keyboard layout handle = ' + inttostr( I ) ;
                     end ;
                14 : begin
                         Result := TDesktop_Debugger.Create ;
                         S := 'Desktop' ;
                     end ;
                15 : begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle := GetDesktopWindow ;
                         S := Window_Description( GetDesktopWindow ) ;
                         if( length( S ) > 0 ) then
                         begin
                             S := ' "' + S + '"' ;
                         end ;
                         S := 'Desktop window = ' + inttostr( GetDesktopWindow ) + S ;
                     end ;
                16 : begin
                         Result := TDevice_List_Debugger.Create ;
                         S := 'Devices' ;
                     end ;
                17 : begin
                         Result := TDisplay_Debugger.Create ;
                         S := 'Display' ;
                     end ;
                18 : begin
                         Setlength( S, 127 ) ;
                         D := 127 ;
                         GetUserName( PChar( S ), D ) ;
                         setlength( S, D - 1 ) ;

                         D := GetSIDLengthRequired( 1 ) ;
                         SIDp := AllocMem( D ) ;
                         D1 := 127 ;
                         setlength( S2, 127 ) ;
                         SNU := SID_NAME_USE( 0 ) ; // Cast is required for FPC
                         LookupAccountName( nil, PChar( S ), SIDp, D,
                             PChar( S2 ), D1, SNU ) ;
                         setlength( S2, D1 - 1 ) ;
                         S := 'Domain = "' + Edit( S2, 4 ) + '"' ;
                         Result := TText_Debugger.Create ;
                     end ;
                19 : begin
                         Result := TText_Debugger.Create ;
                         S := 'Double-click time = ' + inttostr( GetDoubleClickTime ) + ' ms' ;
                     end ;
                20 : begin
                         SystemParametersInfo( SPI_GETDRAGFULLWINDOWS, 0, @B, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Drag full windows = ' + Boolean_To_String( B ) ;
                     end ;
                21 : begin
                         Result := TFile_System_Debugger.Create ;
                         S := 'File_System' ;
                     end ;
                22 : begin
                         fillchar( FK, sizeof( FK ), 0 ) ;
                         FK.cbSize := sizeof( FK ) ;
                         SystemParametersInfo( SPI_GETFILTERKEYS, sizeof( FK ), @FK, 0 ) ;
                         S := Add_Flag( FK.dwFlags, FKF_AVAILABLE, '', 'Available' ) ;
                         S := Add_Flag( FK.dwFlags, FKF_CLICKON, S, 'Click on' ) ;
                         S := Add_Flag( FK.dwFlags, FKF_FILTERKEYSON, S, 'Filterkeys on' ) ;
                         S := Add_Flag( FK.dwFlags, FKF_HOTKEYACTIVE, S, 'Hotkey active' ) ;
                         S := Add_Flag( FK.dwFlags, FKF_HOTKEYSOUND, S, 'Hotkey sound' ) ;
                         if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then
                         begin
                             S := Add_Flag( FK.dwFlags, FKF_CONFIRMHOTKEY, S, 'Confirm hotkey' ) ;
                             S := Add_Flag( FK.dwFlags, FKF_INDICATOR, S, 'Indicator' ) ;
                         end ;
                         Result := TText_Debugger.Create ;
                         S := 'Filterkeys = ' + S ;
                     end ;
                23 : begin
                         SystemParametersInfo( SPI_GETFONTSMOOTHING, 0, @B, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Font smoothing = ' + Boolean_To_String( B ) ;
                     end ;
                24 : begin
                         S := 'Fonts' ;
                         Result := TFonts_Debugger.Create ;
                     end ;
                25 : begin
                         Result := TWindow_Debugger.Create ;
                         TWindow_Debugger( Result ).Handle := GetForegroundWindow ;
                         S := Window_Description( GetForegroundWindow ) ;
                         if( length( S ) > 0 ) then
                         begin
                             S := ' "' + S + '"' ;
                         end ;
                         S := 'Foreground window = ' +
                             inttostr( GetForegroundWindow ) + S ;
                     end ;
                26 : begin
                         Result := TText_Debugger.Create ;
                         S := 'DialogBaseUnits = ' + inttostr( GetDialogBaseUnits ) ;
                     end ;
                27 : begin
                         Result := TText_Debugger.Create ;
                         if( QueryPerformanceCounter( I64 ) ) then
                         begin
                             S := 'High-resolution performance counter = ' +
                                 inttostr( I ) ;
                         end else
                         begin
                             S := 'High-resolution performance counter = n/a' ;
                         end ;
                     end ;
                28 : begin
                         Result := TText_Debugger.Create ;
                         if( QueryPerformanceFrequency( I64 ) ) then
                         begin
                             S := 'High-resolution performance counter frequency = ' +
                                 inttostr( I ) ;
                         end else
                         begin
                             S := 'High-resolution performance counter frequency = n/a' ;
                         end ;
                     end ;
                29 : begin
                         Result := TIcon_System_Debugger.Create ;
                         S := 'Icons' ;
                     end ;
                30 : begin
                         S := 'Locales' ;
                         Result := TLocales_Debugger.Create ;
                     end ;
                31 : begin
                         S := 'Message queue' ;
                         Result := TMessage_Queue_Debugger.Create ;
                     end ;
                32 : begin
                         Result := TMinimized_Metrics_Debugger.Create ;
                         S := 'Minimized metrics' ;
                     end ;
                33 : begin
                         fillchar( MK, sizeof( MK ), 0 ) ;
                         MK.cbSize := sizeof( MK ) ;
                         SystemParametersInfo( SPI_GETMOUSEKEYS, sizeof( MK ), @MK, 0 ) ;
                         S := Add_Flag( MK.dwFlags, MKF_AVAILABLE, '', 'Available' ) ;
                         S := Add_Flag( MK.dwFlags, MKF_HOTKEYACTIVE, S, 'Hot key' ) ;
                         S := Add_Flag( MK.dwFlags, MKF_HOTKEYSOUND, S, 'Sound' ) ;
                         S := Add_Flag( MK.dwFlags, MKF_MOUSEKEYSON, S, 'Active' ) ;
                         if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then
                         begin
                             S := Add_Flag( MK.dwFlags, MKF_CONFIRMHOTKEY, S, 'Confirm' ) ;
                             S := Add_Flag( MK.dwFlags, MKF_INDICATOR, S, 'Indicator' ) ;
                             S := Add_Flag( MK.dwFlags, MKF_MODIFIERS, S, 'Modifiers' ) ;
                             S := Add_Flag( MK.dwFlags, MKF_REPLACENUMBERS, S, 'Replace numbers' ) ;
                         end ;
                         Result := TText_Debugger.Create ;
                         S := 'Mouse Keys: ' + s ;
                     end ;
                34 : begin
                         Result := TNonClient_Metrics_Debugger.Create ;
                         S := 'Non-client metrics' ;
                     end ;
                35 : begin
                         Result := TPage_File_Debugger.Create ;
                         S := 'Page Files' ;
                     end ;
                36 : begin
                         SystemParametersInfo( SPI_GETMENUDROPALIGNMENT, 0, @B, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Popup menus left-aligned = ' + Boolean_To_String( B ) ;
                     end ;
                37 : begin
                         Result := TRegistry_Debugger.Create ;
                         S := 'Registry' ;
                     end ;
                38 : begin
                         SystemParametersInfo( SPI_GETSCREENSAVEACTIVE, 0, @B, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Screen saver enabled = ' + Boolean_To_String( B ) ;
                     end ;
                39 : begin
                         I := 0 ;
                         SystemParametersInfo( SPI_GETSCREENSAVETIMEOUT, 0, @I, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Screen saver timeout = ' + inttostr( I ) ;
                     end ;
                40 : begin
                         B := OpenThreadToken( GetCurrentThread, TOKEN_QUERY, True, hToken ) ;
                         if( not B ) then
                         begin
                             if( GetLastError = ERROR_NO_TOKEN ) then
                             begin
                                 B := OpenProcessToken( GetCurrentProcess, TOKEN_QUERY, hToken ) ;
                             end ;
                         end ;
                         if( B ) then
                         begin
                             // Get length of token...
                             TIC := TokenUser ;
                             Ptr := nil ;
                             D := 0 ;
                             GetTokenInformation( hToken, TIC, Ptr, D, D1 ) ;

                             // Now get access token...
                             Ptr := Allocmem( D1 ) ;
                             D := D1 ;
                             GetTokenInformation( hToken, TIC, Ptr, D, D1 ) ;

                             // Format SID...
                             fillchar( SIDs, sizeof( SIDs ), 0 ) ;
                             PC := PChar( @SIDs ) ;
                             SID := PTokenUser( Ptr )^.User.SID ;
                             SIDA := GetSidIdentifierAuthority( SID ) ;
                             D1 := GetSidSubAuthorityCount( SID )^ ;
                             D := ( 15 + 12 + ( 12 * D1 ) + 1 ) * Sizeof( Char ) ;
                             Strfmt( PC, 'S-1-', [] ) ; // SID Revision 1
                             if( ( SIDA.Value[ 0 ] <> 0 ) or ( SIDA.Value[ 1 ] <> 0 ) ) then
                             begin
                                 Strfmt( PC + strlen( PC ),
                                     '0x%.2x%.2x%.2x%.2x%.2x%.2x',
                                     [ SIDA.Value[ 0 ], SIDA.Value[ 1 ], SIDA.Value[ 2 ], SIDA.Value[ 3 ], SIDA.Value[ 4 ], SIDA.Value[ 5 ] ] ) ;
                             end else
                             begin
                                 Strfmt( PC + strlen( PC ),
                                     '%u',
                                     [ SIDA.Value[ 2 ] shl 24 or SIDA.Value[ 3 ] shl 16 or SIDA.Value[ 4 ] shl 8 or SIDA.Value[ 5 ] ] ) ;
                             end ;
                             for D := 0 to D1 - 1 do
                             begin
                                 Strfmt( PC + strlen( PC ),
                                     '-%u', [ GetSidSubAuthority( SID, D )^ ] ) ;
                             end ;
                             S := string( PC ) ;
                             freemem( Ptr ) ;
                         end else
                         begin
                             S := 'not available' ;
                         end ;
                         S := 'SID = ' + S ;
                         Result := TText_Debugger.Create ;
                     end ;
                41 : begin
                         Result := TSoundSentry_Debugger.Create ;
                         S := 'Sound Sentry' ;
                     end ;
                42 : begin
                         fillchar( SK, sizeof( SK ), 0 ) ;
                         SK.cbSize := sizeof( SK ) ;
                         SystemParametersInfo( SPI_GETSTICKYKEYS, sizeof( SK ), @SK, 0 ) ;
                         S := Add_Flag( SK.dwFlags, SKF_AUDIBLEFEEDBACK, '', 'Audible feedback' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_AVAILABLE, S, 'Available' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_CONFIRMHOTKEY, S, 'Confirm hotkey' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_HOTKEYACTIVE, S, 'Hotkey active' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_HOTKEYSOUND, S, 'Hotkey sound' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_INDICATOR, S, 'Visual indicator' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_STICKYKEYSON, S, 'On' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_TRISTATE, S, 'Tristate' ) ;
                         S := Add_Flag( SK.dwFlags, SKF_TWOKEYSOFF, S, 'Two keys off' ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Sticky keys = ' + inttostr( SK.dwFlags ) + ', ' + S ;
                     end ;
                43 : begin
                         S := 'Stock objects' ;
                         Result := TStock_Object_Debugger.Create ;
                     end ;
                44 : begin
                         Result := TSystem_Colors_Debugger.Create ;
                         S := 'System Colors' ;
                     end ;
                45 : begin
                         Result := TText_Debugger.Create ;
                         setlength( S, 256 ) ;
                         Dummy := GetSystemDirectory( PChar( S ), 256 ) ;
                         setlength( S, Dummy ) ;
                         S := 'System directory = ' + S ;
                     end ;
                46 : begin
                         Result := TSystem_Metrics_Debugger.Create ;
                         S := 'System Metrics' ;
                     end ;
                47 : begin
                         fillchar( TK, sizeof( TK ), 0 ) ;
                         TK.cbSize := sizeof( TK ) ;
                         SystemParametersInfo( SPI_GETTOGGLEKEYS, sizeof( TK ), @TK, 0 ) ;
                         S := Add_Flag( TK.dwFlags, TKF_AVAILABLE, '', 'Available' ) ;
                         S := Add_Flag( TK.dwFlags, TKF_CONFIRMHOTKEY, S, 'Confirm' ) ;
                         S := Add_Flag( TK.dwFlags, TKF_HOTKEYACTIVE, S, 'Hot key active' ) ;
                         S := Add_Flag( TK.dwFlags, TKF_HOTKEYSOUND, S, 'Hot key sound' ) ;
                         S := Add_Flag( TK.dwFlags, TKF_TOGGLEKEYSON, S, 'On' ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Toggle Keys = ' + inttostr( ATO.iTimeOutMSec ) + ', ' + S ;
                     end ;
                48 : begin
                         Setlength( S, 127 ) ;
                         D := 127 ;
                         GetUserName( PChar( S ), D ) ;
                         setlength( S, D - 1 ) ;
                         S := 'User = "' + S + '"' ;
                         Result := TText_Debugger.Create ;
                     end ;
                49 : begin
                         Result := TText_Debugger.Create ;
                         SystemParametersInfo( SPI_GETBEEP, 0, @B, 0 ) ;
                         S := 'Warning beep on = ' + Boolean_To_String( B ) ;
                     end ;
                50 : begin
                         Result := TText_Debugger.Create ;
                         setlength( S, 256 ) ;
                         Dummy := GetWindowsDirectory( PChar( S ), 256 ) ;
                         setlength( S, Dummy ) ;
                         S := 'Windows directory = ' + S ;
                     end ;
                51 : begin
                         I := 0 ;
                         SystemParametersInfo( SPI_GETWORKAREA, 0, @Rect, 0 ) ;
                         Result := TText_Debugger.Create ;
                         S := 'Working area = ' + Rect_To_String( Rect ) ;
                     end ;
                52 : begin
                         if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then
                         begin
                             fillchar( HC, sizeof( HC ), 0 ) ;
                             HC.cbSize := sizeof( HC ) ;
                             SystemParametersInfo( SPI_GETHIGHCONTRAST, sizeof( HC ), @HC, 0 ) ;
                             Result := TText_Debugger.Create ;
                             S := Add_Flag( TK.dwFlags, HCF_AVAILABLE, '', 'Available' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_CONFIRMHOTKEY, S, 'Confirm' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_HIGHCONTRASTON, S, 'On' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_HOTKEYACTIVE, S, 'Hotkey active' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_HOTKEYAVAILABLE, S, 'Hotkey available' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_HOTKEYSOUND, S, 'Hotkey sound' ) ;
                             S := Add_Flag( TK.dwFlags, HCF_INDICATOR, S, 'Indicator' ) ;
                             S := 'High contrast = ' + S ;
                         end else
                         begin
                             I := 0 ;
                             SystemParametersInfo( SPI_GETSNAPTODEFBUTTON, 0, @B, 0 ) ;
                             Result := TText_Debugger.Create ;
                             S := 'Snap to default button = ' + Boolean_To_String( B ) ;
                          end ;
                     end ;
                53 : begin
                         if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then // Windows 95
                         begin
                             SystemParametersInfo( SPI_GETMOUSETRAILS, 0, @I, 0 ) ;
                             Result := TText_Debugger.Create ;
                             S := 'Mouse trails = ' + inttostr( I ) ;
                         end else
                         begin
                             I := 0 ;
                             SystemParametersInfo( SPI_GETWHEELSCROLLLINES, 0, @I, 0 ) ;
                             Result := TText_Debugger.Create ;
                             S := 'Wheel scroll lines = ' + inttostr( I ) ;
                          end ;
                     end ;
        //NetQueryDisplayInformation
            end ; // case Index of
        except
            on E : Exception do
            begin
                S := '* ' + E.Message + ' *' ;
                Result := TText_Debugger.Create ;
            end ;
        end ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;
    end ; // for Index := 0 to Count - 1 do

    Result := TSystem_Parameters_Debugger.Create ;
    Result.Title := 'System Parameters' ;
    _Children.AddObject( 'System Parameters', Result ) ;

    if( X.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS ) then // Windows 95
    begin
        SystemParametersInfo( SPI_GETSCREENREADER, 0, @B, 0 ) ;
        Result := TText_Debugger.Create ;
        S := 'Screen reader = ' + Boolean_To_String( B ) ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;

        fillchar( SerialK, sizeof( SerialK ), 0 ) ;
        SerialK.cbSize := sizeof( SerialK ) ;
        SystemParametersInfo( SPI_GETSERIALKEYS, sizeof( SerialK ), @SerialK, 0 ) ;
        Result := TSerial_Keys_Debugger.Create ;
        TSerial_Keys_Debugger( Result ).SK := SerialK ;
        S := Add_Flag( SerialK.dwFlags, SERKF_ACTIVE, '', 'Active' ) ;
        S := Add_Flag( SerialK.dwFlags, SERKF_AVAILABLE, S, 'Available' ) ;
        S := Add_Flag( SerialK.dwFlags, SERKF_INDICATOR, S, 'Indicator' ) ;
        S := Add_Flag( SerialK.dwFlags, SERKF_SERIALKEYSON, S, 'On' ) ;
        S := 'Serial keys = ' + S ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;

        S := 'Windows Plus! installed = ' +
            Boolean_To_String( SystemParametersInfo( SPI_GETWINDOWSEXTENSION, 1, nil, 0 ) ) ;
        Result := TText_Debugger.Create ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;
    end ;
    if( Supported_Function( 'SHCore.dll', 'GetDpiForShellUiComponent' ) ) then
    begin
        Result := TShell_UI_Component_Debugger.Create ;
        S := 'Get DPI For Shell UI Component = * unknown *' ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;
    end ;
    if( Supported_Function( 'SHCore.dll', 'GetProcessDpiAwareness' ) ) then
    begin
        S := 'Process DPI Awareness = * unknown *' ;
        H := LoadLibrary( 'SHCore.dll' ) ;
        if( H <> 0 ) then
        begin
            GetProcessDPIAwareness := GetProcAddress( H, 'GetProcessDpiAwareness' ) ;
            if( assigned( GetProcessDPIAwareness ) ) then
            begin
                if( GetProcessDPIAwareness( 0, Dummy ) = 0 ) then
                begin
                    S := 'Process DPI Awareness = ' ;
                    case Dummy of
                        0 : S := S + 'DPI Unaware' ;
                        1 : S := S + 'System DPI Aware' ;
                        2 : S := S + 'Per Monitor DPI Aware' ;
                        else S := S + inttostr( Dummy ) ;
                    end ;
                end ;
            end ;
            FreeLibrary( H ) ;
        end ;
        Result := TText_Debugger.Create ;
        Result.Title := PChar( S ) ;
        _Children.AddObject( S, Result ) ;
    end ;

    inherited Build_Children ;
end ; // TMS_Windows_Debugger.Build_Children



end.
