{$IFNDEF FPC}
  {$E+}
  {$N+}
{$ENDIF}
{
        Program Name : OS_DOS
        Package Name : Subroutine Library - Layer 3
        Purpose      : DOS Operating system interface
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2007 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit OS_DOS ;

interface

{$I EDEFINES.INC}

uses { Subroutine Library - Layer 0... }
     Collect,
     _Common, { TObject }
     _Types, { int64 }
     _UE, { PSUnified_Exception }

     { Subroutine Library - Layer 1... }
     _DevMan, { PSDevice_Manager }
     _OS, { TOperating_System }
     _TPM ; { TSText_Presentation_Manager }

type TSDOS_Text_Presentation_Manager = object( TSText_Presentation_Manager )
                                           public { Constructors and destructors... }
                                               constructor Init ;
                                               destructor Done ; virtual ;

                                           private
                                               _Numeric_Text_Presentation_Manager :
                                                   PSNumeric_Text_Presentation_Manager ;
                                               _Monetary_Text_Presentation_Manager :
                                                   PSMonetary_Text_Presentation_Manager ;
                                               _Date_Text_Presentation_Manager :
                                                   PSDate_Text_Presentation_Manager ;

                                           public { API... }
                                               function Numeric_Text_Presentation_Manager :
                                                   PSNumeric_Text_Presentation_Manager ;
                                                   virtual ;

                                               function Monetary_Text_Presentation_Manager :
                                                   PSMonetary_Text_Presentation_Manager ;
                                                   virtual ;

                                               function Date_Text_Presentation_Manager :
                                                   PSDate_Text_Presentation_Manager ;
                                                   virtual ;
                                       end ;
     PSDOS_Text_Presentation_Manager = ^TSDOS_Text_Presentation_Manager ;

type TOS_DOS = object( TSOperating_System )
                   public { Constructors and destructors... }
                       constructor Init ;
                       destructor Done ; virtual ;

                   private
                       Internal_Clipboard : string ;
                       _Device_Manager : PSDevice_Manager ;
                       _Text_Presentation_Manager : PSDOS_Text_Presentation_Manager ;

                       File_Mapping : TCollection ;

                   public { API... }
                       function Name : string ; virtual ;
                       function Version : string ; virtual ;

                       function Application_Path : string ; virtual ;
                       procedure Beep( Typ : integer ) ; virtual ;
                       function Command_Line : string ; virtual ;
                       function Get_Char( Wait : boolean ;
                           var Valid : boolean ; Handle : integer ) : char ;
                           virtual ;
                       procedure Get_Date( var Year, Month, Day : word ) ;
                           virtual ;
                       function Get_Symbol( const X : string ) : string ;
                           virtual ;
                       procedure Get_Time( var Hour, Minute, Second, Hundreths, NS : word ) ;
                           virtual ;
                       function Last_Error : PSUnified_Exception ;
                           virtual ;
                       function Error_Text( Code : longint ;
                           const Language : string ) : string ; virtual ;
                       function Pending_Input( Handle : integer ) : boolean ; virtual ;
                       function PID : longint ; virtual ;
                       procedure Print( T : string ) ; virtual ;
                       function Current_Directory : string ; virtual ;
                       function Spawn( const Path, Parameters, Default_Dir : string ;
                           Flags : integer ) : boolean ; virtual ;
                       function Switch_Delimiters( Index : integer ) : string ;
                           virtual ;
                       function System_Path : string ; virtual ;
                       function Text_Presentation_Manager : PSText_Presentation_Manager ;
                           virtual ;
                       function Thread_ID : longint ; virtual ;
                       function Timestamp : int64 ; virtual ;
                       function Get_Clipboard( Path : string ) : string ;
                           virtual ;
                       procedure Set_Clipboard( Path, Value : string ) ;
                           virtual ;
                       function Max_File_Size : int64 ; virtual ;
                       function Device_Manager : PSDevice_Manager ;
                           virtual ;
                       function Memory_Map_File( _Name : string ;
                           Writeable : boolean ; Start, Size : int64 ;
                           Flag : boolean ) : integer ; virtual ;
                       procedure Unmap_File_From_Memory( Handle : integer ) ;
                           virtual ;
                       function Memory_Mapped_File_Base( Handle : integer ) : pointer ;
                           virtual ;
                       function Show_Message( Handle : longint ; Msg, Title : string ;
                           Options : integer ) : integer ; virtual ;
                       function Application_Data_Path( Default_Root, Path : string ;
                           Flags : integer ) : string ; virtual ;
                       function Create_Path( Path : string ) : integer ; virtual ;
               end ; { TOS_DOS }
     POS_DOS = ^TOS_DOS ;

implementation

uses { Borland... }
{$IFNDEF FPC_OR_DELPHI}
     WinDOS, { GetDate }
{$ENDIF}
{$IFNDEF WINDOWS}
     DOS, { GetDate }
{$ENDIF}
     { Subroutine Library - Layer 0... }
     ASCIIDef,
     L10n, { Text_* }
     _TSRs, { pPSP }

     { Subroutine Library - Layer 1... }
     Compatib, { GetCurDir }
     CommonUt, { Edit }
     UE, { Create_MSDOS_UE }

     { Subroutine Library - Layer 3... }
     Filestd ; { Clear_IOResult }


var _PID : longint ;

{$IFNDEF WINDOWS}
type TRegisters = Registers ;
{$ENDIF}


{ TOS_Device_Manager... }

type TOS_Device_Manager = object( TSDevice_Manager )
                              public { API... }
                                 function Device_Info( Device : string ) : PSDevice_Info ;
                                     virtual ;
                                 { Returns an information object for the
                                   requested device. }

                                  function Disk_Clusters( Drive : string ) : int64 ;
                                      virtual ;
                                  { Returns number of clusters on specified disk
                                    drive.  Returns -1 if an error or unknown. }

                                  function Disk_Sector_Size( Drive : string ) : longint ;
                                      virtual ;
                                  { Returns size of disk sectors, in bytes.
                                    Returns -1 if an error or unknown. }

                                  function File_System( Device : string ) : string ;
                                      virtual ;

                                  function File_System_Flags( Device : string ) : longint ;
                                      virtual ;

                                  function Fixed_Device( Device : string ) : boolean ;
                                      virtual ;
                                  { Returns true if the passed device is fixed
                                    (non-removable) }

                                  function Max_Filename_Length( Device : string ) : longint ;
                                      virtual ;

                                  function Media_ID( Device : string ) : longint ;
                                      virtual ;
                                  { Return type of media in/on device.  -1 if
                                    error or unknown. }

                                  function Valid_Device( Device : string ) : boolean ;
                                      virtual ;
                                  { Returns true if the passed device is valid. }

                                  function Device_Name( Clas, Index : integer ) : string ;
                                      virtual ;

                                  function Current_Path( Device : string ) : string ;
                                      virtual ;
                          end ;
     POS_Device_Manager = ^TOS_Device_Manager ;


{ API... }

function TOS_Device_Manager.Device_Info( Device : string ) : PSDevice_Info ;

begin
    Device_Info := nil ;
    if( not Valid_Device( Device ) ) then
    begin
        exit ;
    end ;
    {~~~}
end ;


function TOS_Device_Manager.Disk_Clusters( Drive : string ) : int64 ;

var Reg : TRegisters ;

begin
    if(
        ( length( Drive ) = 0 )
        or
        ( ( length( Drive ) > 1 ) and ( Drive[ 2 ] <> ':' ) )
      ) then
    begin
        Disk_Clusters := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Drive[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Disk_Clusters := -1 ;
        exit ;
    end ;
    Disk_Clusters := Reg.DX ;
end ;


function TOS_Device_Manager.Disk_Sector_Size( Drive : string ) : longint ;

var Reg : Tregisters ;

begin
    if(
        ( length( Drive ) = 0 )
        or
        ( ( length( Drive ) > 1 ) and ( Drive[ 2 ] <> ':' ) )
      ) then
    begin
        Disk_Sector_Size := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Drive[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Disk_Sector_Size := -1 ;
        exit ;
    end ;
    Disk_Sector_Size := Reg.CX ;
end ;


function TOS_Device_Manager.File_System( Device : string ) : string ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        File_System := '' ;
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Reg.Flags and 64 ) <> 0 ) then
    begin
        File_System := '' ;
        exit ;
    end ;
    S[ 0 ] := char( strlen( PChar( @S[ 1 ] ) ) ) ;
    File_System := S ;
end ;


function TOS_Device_Manager.File_System_Flags( Device : string ) : longint ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    File_System_Flags := 0 ;
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Reg.Flags and 64 ) <> 0 ) then
    begin
        exit ;
    end ;
    File_System_Flags := Reg.Bx ;
end ;


function TOS_Device_Manager.Fixed_Device( Device : string ) : boolean ;

var Reg : Tregisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Fixed_Device := False ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL = 255 ) then
    begin
        Fixed_Device := False ;
        exit ;
    end ;
    Fixed_Device := ( Mem[ Reg.Ds : Reg.Bx ] = $F8 ) ;
end ;


function TOS_Device_Manager.Max_Filename_Length( Device : string ) : longint ;

var Reg : Tregisters ;
    S : string[ 255 ] ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Max_Filename_Length := -1 ;
        exit ;
    end ;
    Reg.Ax := $71A0 ;
    Reg.ES := seg( S ) ;
    Reg.DI := ofs( S ) + 1 ;
    Reg.Cx := 254 ;
    Reg.DS := seg( Device ) ;
    Reg.Dx := ofs( Device ) + 1 ;
    Msdos( Reg ) ; { Get fs information }
    if( ( Reg.Flags and 64 ) <> 0 ) then
    begin
        Max_Filename_Length := -1 ;
        exit ;
    end ;
    Max_Filename_Length := Reg.Cx ;
end ;


function TOS_Device_Manager.Media_ID( Device : string ) : longint ;

var Reg : TRegisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Media_ID := -1 ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    if( Reg.AL <> 255 ) then
    begin
        Media_ID := -1 ;
        exit ;
    end ;
    Media_ID := Mem[ Reg.Ds : Reg.Bx ] ;
end ;


function TOS_Device_Manager.Valid_Device( Device : string ) : boolean ;

var Reg : TRegisters ;

begin
    if(
        ( length( Device ) = 0 )
        or
        ( ( length( Device ) > 1 ) and ( Device[ 2 ] <> ':' ) )
      ) then
    begin
        Valid_Device := False ;
        exit ;
    end ;
    Reg.Ax := $1C00 ;
    Reg.Dx := Ord( Device[ 1 ] ) - 64 ;
    Msdos( Reg ) ; { Get drive information }
    Valid_Device := ( Reg.AL <> 255 ) ;
end ;


function TOS_Device_Manager.Device_Name( Clas, Index : integer ) : string ;

var Loop : integer ;

begin
    Device_Name := '' ;
    if( ( Clas = DC_Unknown ) or ( Clas = DC_Disk ) ) then
    begin
        for Loop := 1 to 26 do
        begin
            if( Valid_Device( chr( Loop + 64 ) + ':' ) ) then
            begin
                dec( Index ) ;
                if( Index < 0 ) then
                begin
                    Device_Name := chr( Loop + 64 ) + ':' ;
                    exit ;
                end ;
            end ;
        end ;
    end ;
end ;


function TOS_Device_Manager.Current_Path( Device : string ) : string ;

begin
    Current_Path := '' ;
    if( Valid_Device( Device ) ) then
    begin
        GetCurDir( Result, ord( Device[ 1 ] ) - 64 ) ;
    end ;
end ;



{ TOS_DOS methods... }

{ Constructors and destructors... }

constructor TOS_DOS.Init ;

begin
    inherited Init ;

{$IFDEF DYNAMIC_STRINGS}
    Initialize( Internal_Clipboard ) ;
{$ENDIF}
    _Device_Manager := nil ;
    Internal_Clipboard := '' ;
    File_Mapping.Init ;
end ;


destructor TOS_DOS.Done ;

begin
    if( _Device_Manager <> nil ) then
    begin
        dispose( _Device_Manager, Done ) ;
        _Device_Manager := nil ;
    end ;
    if( _Text_Presentation_Manager <> nil ) then
    begin
        dispose( _Text_Presentation_Manager, Done ) ;
        _Text_Presentation_Manager := nil ;
    end ;
    File_Mapping.Done ;

    inherited Done ;
end ;


{ API... }

function TOS_DOS.Name : string ; { O/S name }

begin
    Name := 'MSDOS' ;
end ;


function TOS_DOS.Version : string ; { O/S version }

var S1, S2 : string ;

begin
    str( lo( DOSVersion ), S1 ) ;
    str( hi( DOSVersion ), S2 ) ;
    Version := S1 + '.' + S2 ;
end ;


function TOS_DOS.Application_Path : string ;

var Dummy : integer ;
    S : string[ 255 ] ;

begin
    S := Paramstr( 0 ) ;
    Dummy := length( S ) ;
    while( S[ Dummy ] <> '\' ) do
    begin
        dec( Dummy ) ;
    end ;
    S := copy( S, 1, Dummy ) ;
    if( length( S ) = 0 ) then
    begin
        GetCurDir( @S[ 1 ], 0 ) ;
        S[ 0 ] := chr( strlen( PChar( @S[ 1 ] ) ) ) ;
    end ;
    Application_Path := S ;
end ;


procedure TOS_DOS.Beep( Typ : integer ) ;

var R : TRegisters ;

begin
    R.Ax := $0200 ; { Do it this way so it beeps even if output is redirected }
    R.Dx := 7 ;
    Msdos( R ) ;
end ;


function TOS_DOS.Command_Line : string ;

var P : Pointer ;

begin
    P := PTR( PrefixSeg, 0 ) ;
    Command_Line := pPSP( P )^.Command_Tail ;
end ;


function TOS_DOS.Get_Char( Wait : boolean ; var Valid : boolean ;
    Handle : integer ) : char ;

var C : Char ;
    Dummy : integer ;
    Reg : Tregisters ;

label Loop ;

begin
Loop:
    with Reg do
    begin
        Ax := $0600 ;
        Dx := $00FF ;
        Msdos( Reg ) ; { Get character }
        Get_Char := Chr( Lo( Ax ) ) ; { ...and return it }
        Valid := ( ( Flags and 64 ) = 0 ) ;
        if( not Valid ) then { No input }
        begin
            if( Wait ) then { Wait for input }
            begin
                goto Loop ;
            end else
            begin
                Get_Char := NUL ; { Return nul }
            end ;
        end ;
    end ;
end ;


procedure TOS_DOS.Get_Date( var Year, Month, Day : word ) ;

var DOW : word ;

begin
    GetDate( Year, Month, Day, DOW ) ;
end ;


function TOS_DOS.Get_Symbol( const X : string ) : string ;

{$IFDEF WINDOWS}
Type Env_Array = array[ 1..255 ] of Byte ;
     Env_Pointer = ^Env_Array ;

var A : integer ;
    Ep : Env_Pointer ;
    Env : string ;
    Z : array[ 0..255 ] of char ;
{$ENDIF}

begin
{$IFDEF WINDOWS}
    StrPCopy( Z, X ) ;
    Ep := Env_Pointer( Getenvvar( Z ) ) ;
    Env := '' ;
    if( Ep <> nil ) then {If variable is defined}
    begin
        A := 1 ;
        while( Ep^[ A ] <> 0 ) do {Until end of variable}
        begin
            Env := Env + chr( Ep^[ A ] ) ;
            inc( A ) ;
        end ;
    end ;
    Get_Symbol := Env ;
{$ELSE}
    Get_Symbol := GetEnv( X ) ;
{$ENDIF}
end ;


procedure TOS_DOS.Get_Time( var Hour, Minute, Second, Hundreths, NS : word ) ;

begin
    GetTime( Hour, Minute, Second, Hundreths ) ;
    NS := 0 ;
end ;


function TOS_DOS.Last_Error : PSUnified_Exception ;

var Reg : TRegisters ;

begin
    UEC.Facility := 1 ; { MSDOS }
    if( Lo( Dosversion ) < 3 ) then
    begin
        Result := Create_MSDOS_UE(  ) ;
        UEC.Code := 31 ; { No extended error function until version 3 }
    end else
    begin
        Reg.Ax := $5900 ;
        Reg.Bx := 0 ;
        Msdos( Reg ) ;
        UEC.Code := Reg.Ax ;
    end ;
end ;


function TOS_DOS.Error_Text( Code : longint ; const Language : string ) : string ;

begin
    Error_Text := ERT( Code ) ;
end ;


function TOS_DOS.PID : longint ;

begin
    PID := _PID ;
end ;


function TOS_DOS.Pending_Input( Handle : integer ) : boolean ;

var R : TRegisters ;

begin
    R.Ax := $0B00 ; { Check input status }
    Msdos( R ) ;
    if( Lo( R.Ax ) <> 0 ) then
    begin
        Pending_Input := True ;
        exit ; { Exit if input waiting }
    end ;
    Pending_Input := False ;
end ;


var Temp_GetCurDir : string[ 255 ] ;

function TOS_DOS.Current_Directory : string ;

var X : array[ 0..255 ] of char ;

begin
{$IFDEF FPC}
    Temp_GetCurDir := GetCurrentDir ;
{$ELSE}
    GetCurrentDirectory( 127, @X ) ;
    Temp_GetCurDir := strpas( @X ) ;
{$ENDIF}
    Result := Temp_GetCurDir ;
end ;


procedure TOS_DOS.Print( T : string ) ;

var R : TRegisters ;
    Loop : integer ;

begin
    R.Ax := $0500 ; { Printer output }
    for Loop := 1 to length( T ) do
    begin
        R.Dx := ord( T[ Loop ] ) ;
        MSDOS( R ) ;
    end ;
end ;


function TOS_DOS.Spawn( const Path, Parameters, Default_Dir : string ;
    Flags : integer ) : boolean ;

begin
    Swapvectors ;
    Exec( Path, Parameters ) ;
    Swapvectors ;
end ;


function TOS_DOS.Switch_Delimiters( Index : integer ) : string ;

var Reg : TRegisters ;

begin
    Reg.Ax := $3700 ;
    Msdos( Reg ) ; { Get switchar }
    Switch_Delimiters := Chr( Reg.Dx ) ;
end ;


function TOS_DOS.Thread_ID : longint ;

begin
    Thread_ID := 0 ;
end ;


function TOS_DOS.Timestamp : int64 ;

begin
    Timestamp := 0 ;
end ;


function TOS_DOS.System_Path : string ;

var D : byte ;
    Reg : TRegisters ;

begin
    System_Path := '' ;
    if( CheckDOS and ( lo( DOSVersion ) >= 4 ) ) then
    begin
        Reg.Ax := $5200 ;
        Reg.Bx := 0 ;
        Reg.Es := 0 ;
        Msdos( Reg ) ; { Get list of lists }
        D := Mem[ Reg.Es : Reg.Bx + $43 ] ;
        System_Path := chr( D + 64 ) + ':\' ;
    end ;
end ;


function TOS_DOS.Text_Presentation_Manager : PSText_Presentation_Manager ;

begin
    if( _Text_Presentation_Manager = nil ) then
    begin
        _Text_Presentation_Manager := new( PSDOS_Text_Presentation_Manager, Init ) ;
    end ;
    Text_Presentation_Manager := _Text_Presentation_Manager ;
end ;


function TOS_DOS.Get_Clipboard( Path : string ) : string ;

var F : file ;
    S : string ;

begin
    if( Path = '' ) then
    begin
        Get_Clipboard := Internal_Clipboard ;
        exit ;
    end ;
    Path := Edit( Path, -1 ) ;
    S := Get_Symbol( 'Clipboard' ) ;
    if( length( S ) = 0 ) then
    begin
        S := Get_Symbol( 'BootDisk' ) ;
        if( length( S ) = 0 ) then
        begin
            S := System_Path ;
        end ;
    end ;
    if( length( S ) = 0 ) then
    begin
        S := 'C:\Clipboard\' ;
    end else
    if( copy( S, length( S ), 1 ) <> '\'  ) then
    begin
        S := S + '\' ;
    end ;
    S := S + Path ;
    assign( F, S ) ;
    {$I-}
    reset( F, 1 ) ;
    {$I+}
    if( IOResult = 0 ) then
    begin
        S[ 0 ] := chr( filesize( F ) ) ;
        {$I-}
        blockread( F, S[ 1 ], length( S ) ) ;
        {$I+}
        if( IOResult = 0 ) then
        begin
            Get_Clipboard := S ;
        end ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
        exit ;
    end ;

    Get_Clipboard := '' ;
end ;


procedure TOS_DOS.Set_Clipboard( Path, Value : string ) ;

var F : text ;
    S : string ;

begin
    if( Path = '' ) then
    begin
        Internal_Clipboard := Value ;
        exit ;
    end ;
    S := Get_Symbol( 'Clipboard' ) ;
    if( length( S ) = 0 ) then
    begin
        S := Get_Symbol( 'BootDisk' ) ;
        if( length( S ) = 0 ) then
        begin
            S := System_Path ;
        end ;
    end ;
    if( length( S ) = 0 ) then
    begin
        S := 'C:\Clipboard\' ;
    end else
    if( copy( S, length( S ), 1 ) <> '\'  ) then
    begin
        S := S + '\' ;
    end ;
    S := S + Path ;
    assign( F, S ) ;
    {$I-}
    rewrite( F ) ;
    {$I+}
    if( IOResult = 0 ) then
    begin
        {$I-}
        write( F, Value ) ;
        {$I+}
        Clear_IOResult ;
        {$I-}
        close( F ) ;
        {$I+}
        Clear_IOResult ;
    end ;
end ;


function TOS_DOS.Max_File_Size : int64 ;

begin
    Max_File_Size := $3FFFFFFF ;
end ;


function TOS_DOS.Device_Manager : PSDevice_Manager ;

begin
    if( _Device_Manager = nil ) then
    begin
        _Device_Manager := new( POS_Device_Manager, Init ) ;
    end ;
    Device_Manager := _Device_Manager ;
end ;


type TFile_Mapping_Info = record
                              Base : pointer ;
                              Size, Start : integer ;
                              Writeable : boolean ;
                              Name : string ;
                          end ;
     PFile_Mapping_Info = ^TFile_Mapping_Info ;


function TOS_DOS.Memory_Map_File( _Name : string ; Writeable : boolean ;
    Start, Size : int64 ; Flag : boolean ) : integer ;

var Dummy : integer ;
    FM : PFile_Mapping_Info ;
    Saved : integer ;

begin
    // Open file...
    Saved := FileMode ;
    if( Writeable ) then
    begin
        FileMode := 2 ;
    end else
    begin
        FileMode := 0 ;
    end ;
    assign( F, _Name ) ;
    {$I-}
    reset( F, 1 ) ;
    {$I+}
    Dummy := IOResult ;
    FileMode := Saved ;
    if( Dummy <> 0 ) then
    begin
        Memory_Map_File := 0 ;
        exit ;
    end ;
    if( Size = 0 ) then
    begin
        Size := filesize( F ) - Start ;
    end ;

    // Allocate memory...
    new( FM ) ;
    FM^.Size := Size ;
    FM^.Start := Start ;
    FM^.Writeable := Writeable ;
    FM^.Name := _Name ;
    FM^.Base := Allocmem( Size ) ;
    if( FM^.Base = nil ) then
    begin
        dispose( FM ) ;
        Memory_Map_File := 0 ;
        exit ;
    end ;

    // Copy file to memory...
    seek( F, Start ) ;
    blockread( F, FM^.Base^, Size ) ;
    {$I-}
    closefile( F ) ;
    {$I+}
    Dummy := IOResult ;

    // Finsih up...
    File_Mapping.Insert( Info ) ;
    Result := File_Mapping_List.Count ;
end ; { TOS_DOS.Memory_Map_File }


procedure TOS_DOS.Unmap_File_From_Memory( Handle : integer ) ;

var F : file ;
    FM : PFile_Mapping_Info ;

begin
    if( ( Handle > 0 ) and ( Handle <= File_Mapping.Count ) ) then
    begin
        FM := PFile_Mapping_Info( File_Mapping.At( Handle - 1 ) ) ;
        if( FM^.Writeable ) then
        begin
            assign( F, FM^.Name ) ;
            reset( F, 1 ) ;
            seek( F, FM^.Start ) ;
            blockwrite( F, FM^.Base^, FM^.Size ) ;
            {$I-}
            closefile( F ) ;
            {$I+}
            Dummy := IOResult ;
        end ;
        Freemem( FM^.Base, FM^.Size ) ;
        File_Mapping.Delete( Handle - 1 ) ;
        dispose( FM ) ;
    end ;
end ;


function TOS_DOS.Memory_Mapped_File_Base( Handle : integer ) : pointer ;

var FM : PFile_Mapping_Info ;

begin
    if( ( Handle > 0 ) and ( Handle <= File_Mapping.Count ) ) then
    begin
        FM := PFile_Mapping_Info( File_Mapping.At( Handle - 1 ) ) ;
        Memory_Mapped_File_Base := FM^.Base ;
    end else
    begin
        Memory_Mapped_File_Base := nil ;
    end ;
end ;


function TOS_DOS.Show_Message( Handle : longint ; Msg, Title : string ;
    Options : integer ) : integer ;

var S : string ;

    function Check( X : string ) : integer ;

    begin
        Result := 0 ;
        if( copy( X, 1, length( S ) ) = S ) then
        begin
            case X[ 1 ] of
                Text_Keystroke_Abort[ 1 ] : Result := MBR_Abort ;
                Text_Keystroke_Cancel[ 1 ] : Result := MBR_Cancel ;
                Text_Keystroke_Ignore[ 1 ] : Result := MBR_Ignore ;
                Text_Keystroke_No[ 1 ] : Result := MBR_No ;
                Text_Keystroke_OK[ 1 ] : Result := MBR_OK ;
                Text_Keystroke_Retry[ 1 ] : Result := MBR_Retry ;
                Text_Keystroke_Yes[ 1 ] : Result := MBR_Yes ;
            end ;
        end ;
    end ;

begin
    Result := 0 ;
    while( Result = 0 ) do
    begin
        write( Msg ) ;
        case ( Options and 7 ) of
            MBOF_OK : write( ' <', Text_Prompt_OK, '>: ' ) ;
            MBOF_OKCancel : write( ' <', Text_Prompt_OK, '>, ', Text_Prompt_Cancel, ': ' ) ;
            MBOF_AbortRetryIgnore : write( ' <', Text_Prompt_Abort, '>, ', Text_Prompt_Retry, ', ', Text_Prompt_Ignore, ': ' ) ;
            MBOF_YesNoCancel : write( ' <', Text_Prompt_Yes, '>, ', Text_Prompt_No, ', ', Text_Prompt_Cancel, ': ' ) ;
            MBOF_YesNo : write( ' <', Text_Prompt_Yes, '>, ', Text_Prompt_No, ': ' ) ;
            MBOF_RetryCancel : write( ' <', Text_Prompt_Retry, '>, ', Text_Prompt_Cancel, ': ' ) ;
        end ;
        readln( S ) ;
        S := Edit( S, -1 ) ;
        if( S = '' ) then
        begin
            case ( Options and 7 ) of
                MBOF_OK : S := 'OK' ;
                MBOF_OKCancel : S := 'OK' ;
                MBOF_AbortRetryCancel : S := 'ABORT' ;
                MBOF_YesNoCancel : S := 'YES' ;
                MBOF_YesNo : S := 'YES' ;
                MBOF_RetryCancel : S := 'RETRY' ;
            end ;
        end ;
        case ( Options and 7 ) of
            MBOF_OK : Result := Check( 'OK' ) ;
            MBOF_OKCancel : Result := Check( 'OK' ) + Check( 'CANCEL' ) ;
            MBOF_AbortRetryCancel : Result := Check( 'ABORT' ) + Check( 'CANCEL' ) + Check( 'RETRY' ) ;
            MBOF_YesNoCancel : Result := Check( 'YES' ) + Check( 'NO' ) + Check( 'CANCEL' ) ;
            MBOF_YesNo : Result := Check( 'YES' ) + Check( 'NO' ) ;
            MBOF_RetryCancel : Result := Check( 'RETRY' ) + Check( 'CANCEL' ) ;
        end ;
    end ; { while( Result = 0 ) }
end ; { TOS_DOS.Show_Message }


function TOS_DOS.Application_Data_Path( Default_Root, Path : string ; Flags : integer ) : string ;

var Dummy : integer ;
    S : string ;

begin
    S := paramstr( 0 ) ;
    if( S = '' ) then
    begin
        Application_Data_Path := inherited Application_Data_Path( Default_Root, Path ) ;
        { Default path }
        exit ;
    end ;
    Dummy := length( S ) ;
    while( ( Dummy > 0 ) and ( S[ Dummy ] <> '\' ) and ( S[ Dummy ] <> ':' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    S := copy( S, 1, Dummy ) ;
    if( ( S < 4 ) and ( copy( S, 2, 1 ) = ':' ) ) then { Only a device }
    begin
        if( copy( S, length( S ), 1 ) = '\' ) then
        begin
            S := copy( S, 1, length( S ) - 1 ) ; { Trim backslash}
        end ;
        Application_Data_Path :=
            S + inherited Application_Data_Path( Default_Root, Path ) ;
        exit ;
    end ;
    if( Path <> '' ) then
    begin
        if( copy( Path, 1, 1 ) <> '\' ) then
        begin
            S := S + '\' ;
        end ;
        S := S + Path ;
    end ;
    if( copy( S, length( S ), 1 ) <> '\' ) then
    begin
        S := S + '\' ;
    end ;
    Application_Data_Path := S ;
end ; { TOS_DOS.Application_Data_Path }


function TOS_DOS.Create_Path( Path : string ) ;

var Dummy : integer ;
    So_Far : string ;

begin
    So_Far := '' ;
    while( Path <> '' ) do
    begin
        Dummy := pos( '\', Path + '\' ) ;
        So_Far := So_Far + copy( Path, 1, Dummy - 1 ) ;
        Path := copy( Path, Dummy + 1, length( Path ) ) ;
        if( copy( So_Far, length( So_Far ), 1 ) <> ':' ) then
        begin
            {$I-}
            MkDir( So_Far ) ;
            {$I+}
            Dummy := IOResult ;
            if( ( Dummy <> 0 ) and ( Dummy <> 183 ) ) then // Not okay and not already exists
            begin
                Create_Path := Dummy ;
                exit ;
            end ;
        end ;
        So_Far := So_Far + '\' ;
    end ;
end ;



{ TSDOS_Text_Presentation_Manager methods... }

{ Constructors and destructors... }

constructor TSDOS_Text_Presentation_Manager.Init ;

begin
    inherited Init ;

    _Numeric_Text_Presentation_Manager := nil ;
    _Monetary_Text_Presentation_Manager := nil ;
    _Date_Text_Presentation_Manager := nil ;
end ;


destructor TSDOS_Text_Presentation_Manager.Done ;

begin
    if( _Numeric_Text_Presentation_Manager <> nil ) then
    begin
        dispose( _Numeric_Text_Presentation_Manager, Done ) ;
        _Numeric_Text_Presentation_Manager := nil ;
    end ;
    if( _Monetary_Text_Presentation_Manager <> nil ) then
    begin
        dispose( _Monetary_Text_Presentation_Manager, Done ) ;
        _Monetary_Text_Presentation_Manager := nil ;
    end ;
    if( _Date_Text_Presentation_Manager <> nil ) then
    begin
        dispose( _Date_Text_Presentation_Manager, Done ) ;
        _Date_Text_Presentation_Manager := nil ;
    end ;

    inherited Done ;
end ;


{ API... }

function TSDOS_Text_Presentation_Manager.Numeric_Text_Presentation_Manager : PSNumeric_Text_Presentation_Manager ;

begin
    if( _Numeric_Text_Presentation_Manager = nil ) then
    begin
        _Numeric_Text_Presentation_Manager := new( PSNumeric_Text_Presentation_Manager, Init ) ;
    end ;
    Numeric_Text_Presentation_Manager := _Numeric_Text_Presentation_Manager ;
end ;


function TSDOS_Text_Presentation_Manager.Monetary_Text_Presentation_Manager : PSMonetary_Text_Presentation_Manager ;

begin
    if( _Monetary_Text_Presentation_Manager = nil ) then
    begin
        _Monetary_Text_Presentation_Manager := new( PSMonetary_Text_Presentation_Manager, Init ) ;
    end ;
    Monetary_Text_Presentation_Manager := _Monetary_Text_Presentation_Manager ;
end ;


function TSDOS_Text_Presentation_Manager.Date_Text_Presentation_Manager : PSDate_Text_Presentation_Manager ;

begin
    if( _Date_Text_Presentation_Manager = nil ) then
    begin
        _Date_Text_Presentation_Manager := new( PSDate_Text_Presentation_Manager, Init ) ;
    end ;
    Date_Text_Presentation_Manager := _Date_Text_Presentation_Manager ;
end ;



procedure Unit_Setup ;

var Reg : Tregisters ;

begin
    { We must get the PID at startup because, for TSRS, the get-PSP function
      returns the foreground PID, not our PID.  At startup, get-PSP returns our
      PID, so we save it here so it is available later. }
    Reg.AX := $6200 ;
    MSDOS( Reg ) ;
    _PID := Reg.BX ;
end ;


begin
    Unit_Setup ;
end.
