{
        Program Name : OutEdit
        Package Name : Subroutine Library - Layer 4
        Purpose      : TPanel for input/output
        Institution  : Conroy & Conroy Co.
        Date Written : 12-Jan-1999
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1999 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          Outline displayer and editor.
}

unit OutEdit ;

interface

uses // Borland...
     Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, Buttons, ComCtrls, ExtCtrls,

     // Subroutine Library - Layer 4
     TDialogs ; // TDialog

type TOutline_Editor = class( TTreeView )
                           public // Constructors and destructors...
                               constructor Create( AOwner : TComponent ) ;
                                   override ;

                           private // Instance data...
                               Drag_Child : TTreeNode ;
                               Editing : boolean ;

                           protected // Internal event handlers...
                               procedure DragOver( Sender, Source : TObject ;
                                   X, Y : Integer ; State : TDragState ;
                                   var Accept : Boolean ) ;
                               procedure DragDrop( Sender, Source : TObject ;
                                   X, Y : Integer ) ;
                               procedure MouseDown( Sender : TObject ;
                                   Button : TMouseButton ;
                                   Shift : TShiftState ; X, Y : Integer ) ;
                               procedure KeyDown( Sender : TObject ;
                                   var Key : Word ; Shift : TShiftState ) ;
                               procedure CB_Editing( Sender : TObject ;
                                   Node : TTreeNode ;
                                   var AllowEdit : Boolean ) ;
                               procedure CB_Edited( Sender : TObject ;
                                   Node : TTreeNode ; var S : string ) ;

                           public // API...
                               On_Content_Change : TNotifyEvent ;

                               procedure Add ;
                               procedure Move_Node( Target, Drag_Child : TTreeNode ) ;
                               procedure Move_Up ;
                               procedure Move_Down ;
                               procedure Move_In ;
                               procedure Move_Out ;
                               function Node_Is_Terminal( Node : TTreeNode ) : boolean ;
                                   virtual ;
                               function Get_New_Node( Default_Title : string ) : TTreeNode ;
                                   virtual ;
                       end ; // TOutline_Editor

type
  TForm1 = class(TDialog)
    Panel1: TPanel;
    New_Folder: TSpeedButton;
    Delete_Button: TSpeedButton;
    procedure New_FolderClick(Sender: TObject);
    procedure Delete_ButtonClick(Sender: TObject);
    procedure TreeView1Change(Sender: TObject; Node: TTreeNode);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);

  private // Instance data...
    TreeView1 : TOutline_Editor ;

  public // API...
    procedure CB_Change( Sender : TObject ) ;
  end ;

var Form1: TForm1;

implementation

uses { Subroutine Library - Layer 0... }
     L10N ; // Text_*

{$R *.dfm}

function Is_Node_Child( Parent, Child : TTreeNode ) : boolean ;

begin
    Result := False ; // Assume note
    while( Child <> nil ) do
    begin
        Child := Child.Parent ;
        if( Child = Parent ) then
        begin
            Result := True ;
            exit ;
        end ;
    end ;
end ;


// TOutline_Editor methods...

// Constructors and destructors...

constructor TOutline_Editor.Create( AOwner : TComponent ) ;

begin
    inherited Create( AOwner ) ;

    DragMode := dmAutomatic ;
    OnDragOver := DragOver ;
    OnDragDrop := DragDrop ;
    OnMouseDown := MouseDown ;
    OnKeyDown := KeyDown ;
    OnEditing := CB_Editing ;
    OnEdited := CB_Edited ;
end ;


// Internal event handlers...

procedure TOutline_Editor.DragOver(Sender, Source: TObject; X, Y: Integer;
    State: TDragState; var Accept: Boolean) ;

var Node : TTreeNode ;

begin
    Node := GetNodeAt( X, Y ) ;
    Accept := ( Source = self ) and ( Source = Sender ) and not Is_Node_Child( Drag_Child, Node ) ;
end ;


procedure TOutline_Editor.DragDrop(Sender, Source: TObject; X, Y: Integer) ;

var Node : TTreeNode ;

begin
    Node := GetNodeAt( X, Y ) ;
    if( Node_Is_Terminal( Node ) ) then
    begin
        Node := Node.Parent ;
    end ;
    Items.AddChildObject( Node, Drag_Child.Text, Drag_Child.Data ) ;
    Move_Node( Node, Drag_Child ) ;
    Drag_Child.Free ;
    Drag_Child := nil ;
end ;


procedure TOutline_Editor.MouseDown(Sender: TObject; Button: TMouseButton;
    Shift: TShiftState; X, Y: Integer) ;

begin
    Drag_Child := GetNodeAt( X, Y ) ;
end ;


procedure TOutline_Editor.KeyDown(Sender: TObject; var Key: Word;
    Shift: TShiftState) ;

var Node, Target : TTreeNode ;

begin
    if( Editing ) then
    begin
        if( Key = VK_Return ) then
        begin
            Editing := False ;
            Key := 0 ;
        end ;
        exit ;
    end ;

    if( Key = VK_Insert ) then
    begin
        Node := Selected ;
        Target := Get_New_Node( Text_New ) ;
        Node := Items.InsertObject( Node, Target.Text, Target.Data ) ;
        Target.Free ;
        Node.EditText ;
        Selected := Node ;
        Key := 0 ;
        if( Assigned( On_Content_Change ) ) then
        begin
            On_Content_Change( self ) ;
        end ;
        exit ;
    end ;
    Node := Selected ;
    if( ( Key = VK_Delete ) or ( Key = VK_Back ) ) then
    begin
        if( Node <> nil ) then
        begin
            Key := 0 ;
            Node.Free ;
        end ;
        if( Assigned( On_Content_Change ) ) then
        begin
            On_Content_Change( self ) ;
        end ;
        exit ;
    end ;
    if( ( Key = VK_Up ) and ( ssCtrl in Shift ) ) then
    begin
        Move_Up ;
        Key := 0 ;
        exit ;
    end ;
    if( ( Key = VK_Down ) and ( ssCtrl in Shift ) ) then
    begin
        Move_Down ;
        Key := 0 ;
        exit ;
    end ;
    if( ( Key = VK_Right ) and ( ssCtrl in Shift ) ) then // Move into next branch
    begin
        Move_In ;
        Key := 0 ;
        exit ;
    end ;
    if( ( Key = VK_Left ) and ( ssCtrl in Shift ) ) then // Move into parent
    begin
        Move_Out ;
        Key := 0 ;
        exit ;
    end ;
end ; // TOutline_Editor.KeyDown


procedure TOutline_Editor.Add ;

var Child, Node : TTreenode ;
    First : boolean ; // True if this is the first node in the tree

begin
    Node := Selected ;
    First := ( Node = nil ) ;
    if( not First ) then
    begin
        if( Node_Is_Terminal( Node ) ) then
        begin
            Node := Node.Parent ;
        end ;
    end ;
    Child := Get_New_Node( Text_New ) ;
    Node := Items.AddChildObject( Node, Child.Text, Child.Data ) ;
    Child.Free ;
    Selected := Node ;
    if( not First ) then
    begin
        if( Node.Parent <> nil ) then
        begin
            Node.Parent.Expand( False ) ;
        end ;
    end ;
    Node.EditText ;
    if( Assigned( On_Content_Change ) ) then
    begin
        On_Content_Change( self ) ;
    end ;
end ;


procedure TOutline_Editor.Move_Node( Target, Drag_Child : TTreeNode ) ;

var Node, New_Node : TTreeNode ;

begin
    Node := Drag_Child.GetFirstChild ;
    while( Node <> nil ) do
    begin
        New_Node := Items.AddChildObject( Target, Drag_Child.Text, Drag_Child.Data ) ;
        Move_Node( New_Node, Node ) ;
        Node := Drag_Child.GetNextChild( Node ) ;
    end ;
end ;


procedure TOutline_Editor.Move_Up ;

var Node, Target : TTreeNode ;

begin
    Node := Selected ;
    Target := Node.getPrevSibling ;
    if( Target = nil ) then
    begin
        Target := Node.Parent ;
        if( Target = nil ) then
        begin
            exit ;
        end ;
    end ;
    Items.BeginUpdate ;
    Selected := Items.InsertObject( Target, Node.Text, Node.Data ) ;
    Move_Node( Node, Selected ) ;
    Node.Free ;
    Items.EndUpdate ;
    if( Assigned( On_Content_Change ) ) then
    begin
        On_Content_Change( self ) ;
    end ;
end ;


procedure TOutline_Editor.Move_Down ;

var Node, Target : TTreeNode ;

begin
    Node := Selected ;
    Target := Node.getNextSibling ;
    if( Target = nil ) then // Last node at this level
    begin
        // Move to be following the parent node
        Target := Node.Parent ;
        if( Target = nil ) then
        begin
            exit ; // No parent - no place to go
        end ;
        if( Target.GetNextSibling = nil ) then // Parent was the last node at it's level
        begin
            Items.BeginUpdate ;
            Selected := Items.AddObject( Target, Node.Text, Node.Data ) ;
            Move_Node( Node, Selected ) ; // Add to end of parent level
            Node.Free ;
            Items.EndUpdate ;
            exit ;
        end ;
        Target := Target.GetNextSibling ;
    end ;
    if( Target.getNextSibling = nil ) then // Make this the last node at this level
    begin
        Items.BeginUpdate ;
        Selected := Items.AddObject( Target, Node.Text, Node.Data ) ;
        Move_Node( Node, Selected ) ; // Add to end of parent level
        Node.Free ;
        Items.EndUpdate ;
        exit ;
    end ;
    Target := Target.getNextSibling ;
    Items.BeginUpdate ;
    Selected := Items.InsertObject( Target, Node.Text, Node.Data ) ;
    Move_Node( Node, Selected ) ;
    Node.Free ;
    Items.EndUpdate ;
    if( Assigned( On_Content_Change ) ) then
    begin
        On_Content_Change( self ) ;
    end ;
end ;


procedure TOutline_Editor.Move_In ;

var Node, Target : TTreeNode ;

begin
    Node := Selected ;
    Target := Node.getNextSibling ;
    if( Target = nil ) then
    begin
        exit ;
    end ;
    Items.BeginUpdate ;
    Selected := Items.InsertObject( Target.GetFirstChild, Node.Text, Node.Data ) ;
    Move_Node( Node, Selected ) ;
    Node.Free ;
    Items.EndUpdate ;
    if( Assigned( On_Content_Change ) ) then
    begin
        On_Content_Change( self ) ;
    end ;
end ;


procedure TOutline_Editor.Move_Out ;

var Node, Target : TTreeNode ;

begin
    Node := Selected ;
    Target := Node.Parent ;
    if( Target = nil ) then
    begin
        exit ;
    end ;
    Items.BeginUpdate ;
    Selected := Items.InsertObject( Target, Node.Text, Node.Data ) ;
    Move_Node( Node, Selected ) ;
    Node.Free ;
    Items.EndUpdate ;
    if( Assigned( On_Content_Change ) ) then
    begin
        On_Content_Change( self ) ;
    end ;
end ;


function TOutline_Editor.Node_Is_Terminal( Node : TTreeNode ) : boolean ;

begin
    Result := False ;
end ;


function TOutline_Editor.Get_New_Node( Default_Title : string ) : TTreeNode ;

begin
    Result := TTreeNode.Create( Items ) ;
    Result.Text := Default_Title ;
end ;


procedure TOutline_Editor.CB_Editing(Sender: TObject; Node: TTreeNode;
    var AllowEdit: Boolean) ;

begin
    Editing := True ;
end ;


procedure TOutline_Editor.CB_Edited(Sender: TObject; Node: TTreeNode;
    var S: String) ;

begin
    Editing := False ;
end ;



// TForm1...

// API...

procedure TForm1.CB_Change( Sender : TObject ) ;

begin
    Delete_Button.Enabled := ( TreeView1.Selected <> nil ) ;
end ;


procedure TForm1.TreeView1Change(Sender: TObject; Node: TTreeNode);

begin
    Delete_Button.Enabled := ( TreeView1.Selected <> nil ) ;
end ;


procedure TForm1.New_FolderClick(Sender: TObject);

begin
    TreeView1.Add ;
end ;


procedure TForm1.Delete_ButtonClick(Sender: TObject);

var Node : TTreeNode ;

begin
    Node := TreeView1.Selected ;
    if( Node <> nil ) then
    begin
        Node.Free ;
    end ;
    CB_Change( nil ) ;
end ;


procedure TForm1.FormCreate(Sender: TObject);

begin
    TreeView1 := TOutline_Editor.Create( self ) ;
    TreeView1.Parent := self ;
    TreeView1.Align := alClient ;
    TreeView1.OnAddition := TreeView1Change ;
    TreeView1.OnDeletion := TreeView1Change ;
    TreeView1.On_Content_Change := CB_Change ;
end ;


procedure TForm1.FormShow( Sender : TObject ) ;

begin
    Delete_Button.Caption := Text_Button_Caption_Delete ;
    New_Folder.Caption := Text_Button_Caption_New_Folder ;
end ;


end.
