<?php
/*
        Program Name : Parse
        Package Name : Subroutine Library - Layer 3
        Purpose      : Parsing functions
        Institution  : Conroy & Conroy Co.
        Date Written : 03-Nov-1991
        Written By   : Alan Conroy
        Version      : 2.1B

        Copyright (C) 1991-2008 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON                   
 
        30-Nov-2002     EAC      Initialize strings for Delphi.
        07-Mar-2006     RBJ      Cleaned up memory leaks.
        10-Jul-2007     EAC      Added Parse_String.
        11-Jun-2008              Added Parse_Position.
        22-Oct-2008              Added TStrings_Parser.

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit provides various functions for parsing strings.

*/

require_once('EDEFINES.INC');
require_once('_ASCII.php');
require_once('_ue.php');

/*
uses // Borland...
     Collect, // TInteger_Collection 
     CommBase, // TBase_Object 
*/

// Error constants... 
DEFINE('ParseErr_Success',0);
DEFINE('ParseErr_Ambiguous_Match',-1);

DEFINE('ParserErr_Success',0);
DEFINE('ParserErr_Unexpected_End_Of_File',1);
DEFINE('ParserErr_Unterminated_Literal',2);
DEFINE('ParserErr_No_Source',3);

$Parse_Error == 0 ;

uses // Subroutine Library - Layer 1... 
     CommonUt, // Edit 
     _FiP, // FSOM_Input 
     Instrs;

function get_switch_delimiters($n)
{
	switch($n)
	{
		case 0:
		{
			return '&';
		}
		case 1:
		{
			return '/';
		}
		case 2:
		{
			return '-';
		}
	}
}


function get_command_line()
{
	return $_SERVER['REQUEST_URI'];
}


// API... 

// Returns true if partial string matches full string, with a minimum of Len characters.
function Min_Match( $Full, $Partial, $Len )

{
    global $Parse_Error;
    $Parse_Error = ParseErr_Success ;
    if( copy( $Full, 1, strlen( $Partial ) ) == $Partial )
    {
        if( strlen( $Partial ) >= $Len )
        {
            return True ;
        }
        $Parse_Error = ParseErr_Ambiguous_Match ;
    }
    return False ;
} // Min_Match


/*
  This function parses the command line passed to the current program,
  normalizing it based on the DOS switchar parameter, and recognizing both "/"
  and "-" as switch indicators.  Characters within quotes remain unchanged. 
*/
function Parse_Command_Line() // Return parsed command line 
{
    global $Parse_Error ;

    $Parse_Error = ParseErr_Success ;
    $Command = Edit(get_command_line(), 8 ) ; // Get command line 
    $Command1 = get_switch_delimiters(0) ;
    if( ( $Command1 != '"' ) && ( $Command1 != "\'" ) && ( $Command1 != '/' ) && ( $Command1 != '-' ) )
    {
        $Quote = '' ;
        $Dummy1 = 1 ;
        while( $Dummy1 <= strlen( $Command ) ) // Process entire string 
        {
            $c = $Command[ Dummy1 ];
            if( $Quote != '' ) // if within a quote 
            {
            	if( $c == $Quote )
            	{
            		$Quote = '';
            	}
			}
            elseif( ( $c == '"' ) || ( $c == "\'" ) )
			{
            	$Quote = $c ; // Check for quotes 
            }
            if( $Quote == '' )
            {
                if( $c == $Command1 )
                {
                    $Command[ Dummy1 ] = '/' ;
                }
            }
            $Dummy1++ ; // Move to next character 
        }
    }
    return $Command ; // Return line 
} // Parse_Command_Line()


// Parse a command line into input and output specifications 
function Parse_IO( const Command : string ;
    var Input_Spec, Output_Spec : string )

var Dummy : integer ;
    Space_Count, Equal_Count : integer ;
    Space_Pos, Equal_Pos : integer ;
    Quote_Type : char ;
{
    Parse_Error = ParseErr_Success ;
    Space_Count = 0 ;
    Equal_Count = 0 ;
    Equal_Pos = 0 ;
    Space_Pos = 0 ;
    Quote_Type = ' ' ;
    for Dummy = 1 to strlen( Command ) do
    {
        if( Quote_Type != ' ' ) // Inside a quote 
        {
            if( Command[ Dummy ] == Quote_Type )
            {
                Quote_Type = ' ' ; // Done with quote 
            }
        }
        else
        {
            if( ( Command[ Dummy ] == '"' ) or ( Command[ Dummy ] == #39 ) )
            {
                Quote_Type = Command[ Dummy ] ;
            }
            else
            {
                if( Command[ Dummy ] == ' ' )
                {
                    inc( Space_Count ) ;
                    Space_Pos = Dummy ;
                }
                if( Command[ Dummy ] == '=' )
                {
                    inc( Equal_Count ) ;
                    Equal_Pos = Dummy ;
                }
            }
        }
    } // for Dummy = 1 to length( Command ) 
    if( ( Equal_Count == 0 ) && ( Space_Count == 0 ) ) // Only input present 
    {
        Input_Spec = Command ;
        Output_Spec = '' ;
        exit ;
    }
    if(
        (
          ( Equal_Count > 0 )
          &&
          ( Space_Count > 0 )
        )
        ||
        ( Equal_Count > 1 )
        ||
        ( Space_Count > 1 )
      )
    {
        Input_Spec = '' ;
        Output_Spec = '' ; // Invalid specification 
        exit ;
    }
    if( Space_Count == 1 )
    {
        Input_Spec = copy( Command, 1, Space_Pos - 1 ) ;
        Output_Spec = copy( Command, Space_Pos + 1, 254 ) ;
    }
    else
    {
        Output_Spec = copy( Command, 1, Equal_Pos - 1 ) ;
        Input_Spec = copy( Command, Equal_Pos + 1, 254 ) ;
    }
} // Parse_IO 


function Parse_Parameter( Separator : string ; var Command : string ) : string ;
// Return next parameter 

var Dummy : Integer ;
    Quote : Char ;
{
    Parse_Error = ParseErr_Success ;
    Quote = ' ' ; // No quote yet 
    if( strlen( Separator ) == 0 )
    {
        Separator = ' ' ; // Default to space 
    }
    for Dummy = 1 to strlen( Command ) do
    {
        if( Quote != ' ' ) // if within a quote 
           
                if( Command[ Dummy ] == Quote ) Quote = ' ' else
            else // Not within a quote
                if( ( Command[ Dummy ] == '"' ) or ( Command[ Dummy ] == Chr( 39 ) ) )
                    Quote = Command[ Dummy ] ; // Check for quotes 
        if( Command[ Dummy ] == Separator[ 1 ] ) if( Quote == ' ' )
        {
            Parse_Parameter = Copy( Command, 1, Dummy - 1 ) ; // Return parameter 
            Command = Copy( Command, Dummy + 1, length( Command ) ) ; // Strip from command line 
            Exit ; // Done 
        }
    }
    // At this point, the entire command line is a parameter 
    Parse_Parameter = Command ; // Return next parameter 
    Command = '' ; // Strip from command line 
} // Parse_Parameter 


// Find position of Separator in Command.  Returns 0 if Separator not found. 
function Parse_Position( Separator : string ; const Command : string )

var Dummy : Integer ;
    Quote : Char ;
{
    Parse_Error = ParseErr_Success ;
    Quote = ' ' ; // No quote yet 
    if( Length( Separator ) == 0 )
    {
        Separator = ' ' ; // Default to space 
    }
    for Dummy = 1 to Length( Command ) do
    {
        if( Quote != ' ' ) // if within a quote 
        {
            if( Command[ Dummy ] == Quote )
            {
                Quote = ' ' ;
            }
        }
        else // Not within a quote 
        if(
            ( Command[ Dummy ] == Separator[ 1 ] )
            &&
            ( copy( Command, Dummy, length( Separator ) ) == Separator )
          )
        {
            Parse_Position = Dummy ; // Return position 
            Exit ; // Done 
        }
        else
        if( ( Command[ Dummy ] == '"' ) or ( Command[ Dummy ] == Chr( 39 ) ) )
        {
            Quote = Command[ Dummy ] ; // Check for quotes
        }
    } // for Dummy = 1 to Length( Command ) 

    Parse_Position = 0 ; // Not found 
} // Parse_Position 


// Parse a quoted string, returning unquoted version 
function Parse_Quotes( const S : string )

var Dummy : integer ;
    Ch, Quote : char ;
    Res : string ;
{
    Res = '' ;
    Quote = ' ' ;
    for Dummy = 1 to length( S ) do
    {
        Ch = S[ Dummy ] ;
        if( Quote == ' ' ) // Not within a quote 
        {
            if( ( Ch == #39 ) or ( Ch == '"' ) )
            {
                Quote = Ch ; // Starting a quote 
            }
            else
            {
                Res = Res + Ch ;
            }
        }
        else
        {
            if( Ch == Quote ) // End of current quote 
            {
                Quote = ' ' ;
            }
            else
            {
                Res = Res + Ch ;
            }
        }
    }
    Parse_Quotes = Res ;
} // Parse_Quotes 


function Parse_String( const S : string ) : string ;
/* Parse a string literal into an actual string value.  String literals are
  delimited by single (') or double (") quotes.  Outside of quotes, special
  codes can indicate any ASCII value.  Special codes start with pound signs (#)
  which are followed by a number or ASCII code.  For instance: #0 or #NUL. */

var C : char ;
    Loop : integer ;
    Res : string ;
    Quote : char ;
    Special : string ; // Special code(s) 

    function Add_Special( C : char ) ; // Accumulate special code(s).  C == #0 means force termination of special code. 

    var Dummy, X : integer ;

    {
        if( C == #0 )
        {
            if( copy( Special, 1, 1 ) == '#' )
            {
                Special = copy( Special, 2, length( Special ) ) ;
                val( Special, X, Dummy ) ;
                if( Dummy == 0 )
                {
                    Res = Res + chr( X ) ;
                }
                else
                {
                    for Dummy = 0 to 31 do
                    {
                        if( Special == ASCII_Codes( Dummy ) )
                        {
                            Res = Res + chr( Dummy  ) ;
                            break ;
                        }
                    }
                }
            }
            Special = '' ;
        }
        else
        {
            if( C == '#' )
            {
                Add_Special( #0 ) ; // Terminate any accumulated specials 
            }
            Special = Special + C ;
        }
    }

{ // Parse_String 
    Res = '' ;
    Quote = #0 ;
    for Loop = 1 to length( S ) do
    {
        C = S[ Loop ] ;
        if( Quote == #0 ) // Not inside a quote 
        {
            if( ( C == '"' ) or ( C == #39 ) )
            {
                Quote = C ;
                Add_Special( #0 ) ;
            }
            else
            {
                Add_Special( C ) ;
            }
        }
        else
        { // Inside a quote 
            Add_Special( #0 ) ;
            if( C == Quote ) // Ending quote 
            {
                Quote = #0 ;
            }
            else
            {
                Res = Res + C ;
            }
        }
    }
    Parse_String = Res ;
} // Parse_String 


function Parse_Switch( Switch, Alternate : string ;
    var Command, Parameter : string ) : Integer ;
/*
  Parse a switch in a command line:
  Switch is a list of switches, separated by commas,
  Alternate is a list of switches, separated by commas,
  Command is the command line to search for the switches in,
  Parameter is the parameter following the switch (if any).

  Function returns:
        0 == No matching switches found
        1 == Switch matching one in Switch was found
        2 == Switch matching one in Alternate was found

  In the case of multiple switches, only the last occurance is parsed, but all
  occurances are removed from the command line.  Note that switches are passed
  without slashes.  Switch lists may optionally contain vertical bars (|) to
  indicate the minimum abbreviation allowed for the switch. 
*/

label Try_Again ;

var Switches : array[ 0..15 ] of string ;
    Switch_Min : array[ 0..15 ] of Integer ;
    Alternates : array[ 0..15 ] of string ;
    Alt_Min : array[ 0..15 ] of Integer ;
    Dummy, Dummy1, Dummy2 : Integer ; //Work variables
    First_Time : Boolean ; //True if first switch match
    Location : Integer ; //Location of switch in Work
    Quote : Char ;
    Work, Work1 : string ;

{
    Parse_Error = ParseErr_Success ;

    // The following code breaks the multiple switches in Switch and Alternate
      into two arrays of switches to make the following processing easier 
    for Dummy = 0 to 15 do
    {
        Switches[ Dummy ] = '' ;
        Alternates[ Dummy ] = '' ;
    }
    Dummy1 = -1 ;
    while( Length( Switch ) > 0 )
    {
        Dummy = Pos( ',', Switch + ',' ) ; // Find end of next switch 
        Dummy1 = Dummy1 + 1 ;
        if( Dummy1 == 16 )
        {
            Switch = '' ; // 16 is the maximum 
        }
        else
        {
            Switches[ Dummy1 ] = Copy( Switch, 1, Dummy - 1 ) ; // Get switch 
            Switch = Copy( Switch, Dummy + 1, length( Switch ) ) ; // Strip from parameter 
            Dummy2 = Pos( '|', Switches[ Dummy1 ] ) ; // Abbreviation specification? 
            if( Dummy2 > 1 ) // Abbreviation specified 
            {
                Switch_Min[ Dummy1 ] = Dummy2 - 1 ;
                Switches[ Dummy1 ] = Copy( Switches[ Dummy1 ], 1, Dummy2 - 1 ) +
                    Copy( Switches[ Dummy1 ], Dummy2 + 1, length( Switches[ Dummy1 ] ) ) ;
            end else
            {
                Switch_Min[ Dummy1 ] = Length( Switches[ Dummy1 ] ) ;
            }
        }
    }
    Dummy1 = -1 ;
    while( Length( Alternate ) > 0 ) do
    {
        Dummy = Pos( ',', Alternate + ',' ) ; // Find end of next switch 
        Dummy1 = Dummy1 + 1 ;
        if( Dummy1 == 16 )
        {
            Alternate = '' ; // 16 is the maximum 
        }
        else
        {
            Alternates[ Dummy1 ] = Copy( Alternate, 1, Dummy - 1 ) ; // Get switch 
            Alternate = Copy( Alternate, Dummy + 1, length( Alternate ) ) ; // Strip from parameter 
            Dummy2 = Pos( '|', Alternates[ Dummy1 ] ) ;
            // Abbreviation specification? 
            if( Dummy2 > 1 ) // Abbreviation specified 
            {
                Alt_Min[ Dummy1 ] = Dummy2 - 1 ;
                Alternates[ Dummy1 ] = Copy( Alternates[ Dummy1 ], 1, Dummy2 - 1 ) +
                    Copy( Alternates[ Dummy1 ], Dummy2 + 1, length( Alternates[ Dummy1 ] ) ) ;
            }
            else
            {
                Alt_Min[ Dummy1 ] = Length( Alternates[ Dummy1 ] ) ;
            }
        }
    }

    Work = Command ; // Make a local copy of the command line 
    Parameter = '' ; // Thus, no parameter 
    Parse_Switch = 0 ; // No match 
    First_Time = True ; // First time 
Try_Again:
    if( Length( Work ) == 0 )
    {
        Exit ; // No more switches in command line 
    }
    Dummy = Pos( '"', Work + '"' ) ; // Find location of quote 
    Dummy1 = Pos( Chr( 39 ), Work + Chr( 39 ) ) ; // Try alternate quote 
    if( Dummy1 < Dummy )
    {
        Dummy = Dummy1 ; // The first quote in the line 
    }
    if( Dummy == 1 ) // Command line starts with a literal 
    {
        Dummy1 = Instr( Dummy + 1, Work + Work[ 1 ], Work[ 1 ] ) ;
        // Find ending quote 
        Work = Copy( Work, Dummy1 + 1, length( Work ) ) ; // Strip literal from work string 
        goto Try_Again ; // Go back and try it again 
    }

    // At this point, Work contains the rightmost n characters in Command and
      Dummy is the location of the first quote in Work. 

    Dummy1 = Length( Work ) + 1 ;
    for Dummy2 = 0 to 15 do // Try all possible switches 
    {
        if( Length( Switches[ Dummy2 ] ) > 0 )
        {
            Location = Pos( '/' + Copy( Switches[ Dummy2 ], 1, Switch_Min[ Dummy2 ] ),
                uppercase( Work ) + '/' + Switches[ Dummy2 ] ) ; // Find one 
            if( Dummy1 > Location )
            {
                Dummy1 = Location ; // Choose earliest one 
                Work1 = Switches[ Dummy2 ] ;
                Parse_Switch = 1 ;
            }
        }
        if( Length( Alternates[ Dummy2 ] ) > 0 )
        {
            Location = Pos( '/' + Copy( Alternates[ Dummy2 ], 1, Alt_Min[ Dummy2 ] ),
                uppercase( Work ) + '/' + Alternates[ Dummy2 ] ) ; // Find one 
            if( Dummy1 > Location ) // Choose earliest one 
            {
                Dummy1 = Location ;
                Work1 = Alternates[ Dummy2 ] ;
                Parse_Switch = 2 ;
            }
        }
    }

    // At this point, Dummy1 contains the location of the first occurance 

    if( Dummy1 > Length( Work ) ) // No matching switches found 
    {
        if( First_Time )
        {
            Parse_Switch = 0 ; // No match 
        }
        Exit ;
    }
    if( Dummy1 > Dummy ) // Switch {s after a quote 
    {
        Quote = Work[ Dummy ] ;
        Dummy = Pos( Quote, Work + Quote ) ; // Find end of literal 
        Work = Copy( Work, Dummy + 1, length( Work ) ) ; // Strip literal 
        if( First_Time )
        {
            Parse_Switch = 0 ; // No match 
        }
        goto Try_Again ;
    }
    // Found a possible switch location at Dummy1 
    Dummy1 = Dummy1 + 1 ; // Skip slash 
    Dummy2 = 0 ; // Offset from switch {ning 
    while(
           ( Dummy1 + Dummy2 <= Length( Work ) )
           and
           ( Work[ Dummy1 + Dummy2 ] != ':' )
           and
           ( Work[ Dummy1 + Dummy2 ] != '=' )
           and
           ( Work[ Dummy1 + Dummy2 ] != '/' )
           and
           ( Work[ Dummy1 + Dummy2 ] != ',' )
           and
           ( Work[ Dummy1 + Dummy2 ] != ' ' )
           and
           ( Work[ Dummy1 + Dummy2 ] != '	' )
         )
    {
        Dummy2 = Dummy2 + 1 ; // Move to next character 
    }
    // At this point, Dummy1 is the start of the switch and Dummy2 is the size 
    if( Copy( uppercase( Work ), Dummy1, Dummy2 ) != Copy( Work1, 1, Dummy2 ) )
    // Incomplete match 
    {
        Work = Copy( Work, Dummy1 + Dummy2, length( Work ) ) ; // Strip switch 
        if( First_Time )
        {
            Parse_Switch = 0 ; // Not a match 
        }
        goto Try_Again ; // Try again 
    }
    First_Time = False ;

    // At this point, we have a match - now grab any parameters 
    Command = Copy( Command, 1, Length( Command ) - Length( Work ) + Dummy1 - 2 ) +
        Copy( Work, Dummy1 + Dummy2, length( Work ) ) ; // Strip switch from command line 
    Work = Copy( Work, Dummy1 + Dummy2, length( Work ) ) ; // and from work variable 
    if(
        ( Length( Work ) == 0 )
        ||
        (
          ( Work[ 1 ] != ':' )
          &&
          ( Work[ 1 ] != '=' )
        )
      ) // No parameters 
    {
        Parameter = '' ;
        goto Try_Again ;
    }
    Command = Copy( Command, 1, Length( Command ) - Length( Work ) ) + Copy( Work, 2, length( Work ) ) ;
    // Remove separator 
    Work = Copy( Work, 2, length( Work ) ) ; // Ditto 
    if( ( Work[ 1 ] == '"' ) or ( Work[ 1 ] == Chr( 39 ) ) ) // Quoted literal 
    {
        Dummy = Instr( 2, Work + Work[ 1 ], Work[ 1 ] ) ; // Find end of quote 
    }
    else
    {
        Dummy = Pos( '/', Work + '/' ) - 1 ; // Find end of parameter 
        if( Pos( ' ', Work + ' ' ) - 1 < Dummy )
        {
            Dummy = Pos( ' ', Work + ' ' ) - 1 ; // Find end of parameter 
        }
        if( Pos( ' ', Work + ' ' ) - 1 < Dummy )
        {
            Dummy = Pos( '	', Work + '	' ) - 1 ; // Find end of parameter 
        }
    }
    Parameter = Copy( Work, 1, Dummy ) ; // Get the parameter 
    if(
        ( copy( Parameter, 1, 1 ) == '"' )
        ||
        ( copy( Parameter, 1, 1 ) == #39 )
      ) // Quoted 
    {
        if( copy( Parameter, 1, 1 ) == copy( Parameter, length( Parameter ), 1  ) )
        {
            Parameter = copy( Parameter, 1, length( Parameter ) - 1 ) ; // Remove terminating quote 
        }
        Parameter = copy( Parameter, 2, length( Parameter ) ) ; // Remove starting quote 
    }
    Command = Copy( Command, 1, Length( Command ) - Length( Work ) ) +
        Copy( Work, Dummy + 1, length( Work ) ) ; // Strip parameter from command line 
    Work = Copy( Work, Dummy + 1, length( Work ) ) ; // Remove parameter 
    goto Try_Again ; // Now see if a later switch over-rides this occurance 
} // Parse_Switch 


// Return position of first switch in Command (0 if none) 
function Switch_Present( const Command : string )

var Dummy : Integer ;
    Quote : Char ;

{
    Parse_Error = ParseErr_Success ;
    Switch_Present = 0 ;
    Quote = ' ' ;
    Dummy = 1 ;
    while( Dummy <= Length( Command ) ) do //Process entire string
    {
        if( Quote != ' ' ) //if within a quote
        {
            if( Command[ Dummy ] == Quote )
            {
                Quote = ' ' ;
            }
        }
		else // Not within a quote 
        {
            if( ( Command[ Dummy ] == '"' ) or ( Command[ Dummy ] == Chr( 39 ) ) )
            {
                Quote = Command[ Dummy ] ; // Check for quotes 
            }
        }
        if( Quote == ' ' ) if( Command[ Dummy ] == '/' )
        {
            Switch_Present = Dummy ; // Switch found 
            Exit ;
        }
        Dummy = Dummy + 1 ; // Move to next character 
    }
} // Switch_Present 


// Returns path of executable 
function Program_Path()

var Dummy : integer ;
    S : string ;

{
    S = paramstr( 0 ) ;
    Dummy = length( S ) ;
    while( ( Dummy > 0 ) and ( S[ Dummy ] != '\\' ) )
    {
        dec( Dummy ) ;
    }
    Program_Path = copy( S, 1, Dummy )  ;
}



class TParser = object( TBase_Object )
{
        // Instance data... 
        private $Command = '' ; // Current input line 
        private $_Current_Line = 0 ; // Current line number 
        private $_Leading_Whitespace = FALSE ; // True if current line had leading whitespace 
        private $_Quotes = '"\'' ;

        private $Returned_Tokens = 0 ;
        // Number of returned tokens waiting for reads 

        private $Token_Error = 0 ;
        // Token error returned by Token routine 

        private $Token_Stack = new TInteger_Collection() ;
        // Stack of token lines 

        private $Tokens = new TString_Collection() ; // Special tokens 

        // API... 
        public $Non_Terminals = '$_ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' ;
        public $Token_Line = 0 ; // Line number of current token 
        public $Case_Sensitive = FALSE ; // True if case-sensitive outside of quotes 

        // Internal utility routines... 

        function Get_Token() // Return next token from source 

        var Aa : integer ; // Changed from byte to support lines > 255 bytes in length 
            Bb : Char ;
            C : PASCII_String ;
            E : integer ;
            Quote : Boolean ;
            Quote_Type : Char ;
            Temp : PString ;

        Label Next, Token_}

        {
            Token_Error = ParserErr_Success ;
            Quote_Type = ' ' ;

        Next:
            if( Command.Length == 0 ) // No more text in buffer 
            {
                if( not Read_Next_Line )
                {
                    Get_Token = new( PASCII_String, Init ) ;
                    exit ;
                }
                Inc( _Current_Line ) ; // Increment line count 
                Token_Line = Get_Current_Line ; // Token line is now same as current line 

                _Leading_Whitespace = Command.Class_Of( Command.At( 1 ) ) == CC_Whitespace ;
                E = 8700 ;
                if( Case_Sensitive )
                {
                    E = E - 32 ;  // No uppercase 
                }
                Temp = Command.Edit( E ) ; // Format line 
                Command.Assign( Temp ) ;
                dispose( Temp, Done ) ;
                if( Command.Length == 0 )
                {
                    goto Next ; // Ignore blank lines 
                }
            }
            Quote = False ;
            C = new( PASCII_String, Init ) ;
            for Aa = 1 to Command.Length do
            {
                Bb = char( Command.Upcase( Command.At( Aa ) ) ) ; // Get current character 
                if Quote // Within a quoted token 
                {
                    if( Bb == Quote_Type ) // End of quoted token 
                    {
                        Temp = Command.Copy( 1, Aa ) ;
                        C^.Assign( Temp ) ;
                        dispose( Temp, Done ) ;
                        Temp = Command.Copy( Aa + 1, Command.Length - Aa ) ;
                        Command.Assign( Temp ) ;
                        dispose( Temp, Done ) ;
                        goto Token_}
                    }
                }
                elseif( Aa == 1 ) // First character of token 
                {
                    if( pos( BB, _Quotes ) > 0 ) // if a quote 
                    {
                        Quote = True ; // Now within quotes 
                        Quote_Type = Bb // Save quote type 
                    }
                    else
                    {
                        if( Non_Terminals.Instr( 1, Temp_String( Bb ) ) == 0 )
                        {
                            C^.Set_String( char( Command.At( 1 ) ) ) ;
                            Temp = Command.Copy( 2, Command.Length - 1 ) ;
                            Command.Assign( Temp ) ;
                            dispose( Temp, Done ) ;
                            goto Token_}
                        }
                    }
                }
				else // After first character 
                {
                    if( Non_Terminals.Instr( 1, Temp_String( Bb ) ) == 0 )
                    {
                        Temp = Command.Copy( 1, Aa - 1 ) ;
                        C^.Assign( Temp ) ;
                        dispose( Temp, Done ) ;
                        Temp = Command.Copy( Aa, Command.Length - Aa + 1 ) ;
                        Command.Assign( Temp ) ;
                        dispose( Temp, Done ) ;
                        goto Token_}
                    }
                }
            } // for Aa = 1 to Command.Length
            C^.Assign( @Command ) ; // Token runs to end of line 
            Command.Assign( nil ) ;
            if( Quote )
            {
                Token_Error = ParserErr_Unterminated_Literal ;
            }
        Token_End:
            if( pos( copy( C^.Text, 1, 1 ), _Quotes ) == 0 )
            {
                Temp = C^.Edit( -1 ) ; // Format token 
                C^.Assign( Temp ) ;
                dispose( Temp, Done ) ;
            }
            if( C^.Length == 0 )
            {
                // need to dispose of PASCII_String 
                dispose( C, Done ) ;

                goto Next ; // Ignore spaces, etc 
            }
        //    Push_Token_Line ;
            if( Returned_Tokens > 0 )
            {
                Token_Line = Pop_Token_Line ;
                Returned_Tokens = Returned_Tokens - 1 ;
            }
            Get_Token = C ; // Return value 
        } // Get_Token 


        function Pop_Token_Line : Integer ; // Return previous token line 

        {
            if( Token_Stack.Count == 0 )
            {
                Pop_Token_Line = 0 ;
            
			}
			else
            {
                Pop_Token_Line = Token_Stack.At( Token_Stack.Count - 1 ) ;
                Token_Stack.AtDelete( Token_Stack.Count - 1 ) ;
            }
        }


        function Push_Token_Line ; // Push current token line 

        {
            Token_Stack.Insert( Token_Line ) ;
        }


        function Read_Next_Line : boolean ;

        {
            Command.Assign( nil ) ;
            Read_Next_Line = False ;
        }


        // Property handlers... 

        public function Get_Token_Line()
        {
            return $this->$Token_Line ;
        }


        public function Get_Current_Line()
        {
            return $this->$_Current_Line ;
        }


        // API... 

        public function Add_Special_Token( $S )
        {
            $this->$Tokens.Insert( $S ) ;
        }


        public function Grab_Line()
        {
        var E : integer ;
            Temp : PString ;

            if( Token_Stack.Count > 0 )
            {
                Token_Line = Token_Stack.At( Token_Stack.Count - 1 ) ;
            }
            Returned_Tokens = 0 ;
            while( Token_Stack.Count > 0 ) do
            {
                Token_Stack.AtDelete( Token_Stack.Count - 1 ) ;
            }
            if( Command.Length == 0 ) // No more text in buffer 
            {
        Next:
                if( Read_Next_Line )
                {
                    Inc( _Current_Line ) ; // Increment line count 
                    Token_Line = Get_Current_Line ; // Token line is now same as current line 

                    _Leading_Whitespace = Command.Class_Of( Command.At( 1 ) ) == CC_Whitespace ;
                    E = 8700 ;
                    if( Case_Sensitive )
                    {
                        E = E - 32 ;  // No uppercase 
                    }
                    Temp = Command.Edit( E ) ; // Format line 
                    Command.Assign( Temp ) ;
                    dispose( Temp, Done ) ;
                    if( Command.Length == 0 )
                    {
                        goto Next ; // Ignore blank lines 
                    }
                }
            }
            Temp = new( PASCII_String, Init ) ;
            Temp^.Assign( @Command ) ;
            Grab_Line = Temp ;
            Command.Assign( nil ) ;
        } // Grab_Line 


        public function _Last_Error()
        {
            _Last_Error = Token_Error ;
        }


        public function Token()
        {
        var Dummy : integer ;
            S, T : PString ; // Accumulated token 
            Temp : PString ;

            T = Get_Token ; //NOTE: Handle quoted data differently (ie skip the following)
            if( ( Tokens.Count > 0 ) and ( T^.Length > 0 ) )
            {
                // A list of special tokens is present and we have a token 
                for Dummy = 0 to Tokens.Count - 1 do
                {
                    S = Temp_String( Tokens.At( Dummy ) ) ;
                    if( T^.At( 1 ) == S^.At( 1 ) ) // A potential match 
                    {
                        Temp = Command.Copy( 1, S^.Length - 1 ) ;
                        if( T^.Text + Temp^.Text == S^.Text ) // Found a match 
                        {
                            dispose( Temp, Done ) ;
                            while( T^.Length < S^.Length ) do
                            {
                                T^.Append( Get_Token ) ;
                            }
                            Token = T ;
                            exit ;
                        }
                        dispose( Temp, Done )
                    }
                }
            }
            Token = T ; // Return result 
        } // Token 


        public function Put_Token( Value : PString )

        var Ch : integer ;
            Temp : PASCII_String ;

        {
            if( Value^.Length > 0 )
            {
                Inc( Returned_Tokens ) ;
                Temp = new( PASCII_String, Init ) ;
                Temp^.Assign( @Command ) ;
                Ch = Command.At( 1 ) ;
                if(
                    ( Command.Length > 0 )
                    and
                    ( Non_Terminals.Instr( 1, Temp_String( chr( Ch ) ) ) > 0 )
                  )
                {
                    Command.Assign( Value ) ;
                    Command.Append_Char( 32 ) ;
                    Command.Append( Temp ) ;
                end else
                {
                    Command.Assign( Value ) ;
                    Command.Append( Temp ) ;
                }
                dispose( Temp, Done ) ;
                Push_Token_Line ;
        //        Token_Line = Token_Stack.At( Token_Stack.Count - 1 ) ;
            }
        } // Put_Token 


        public function Leading_Whitespace()
        {
            return $this->$_Leading_Whitespace ;
        }


        public function Set_Quotes( $Value )
        {
            $this->$_Quotes = $Value ;
        }


        public function Line_Text()
        {
        var C : PASCII_String ;
            E : integer ;
            Temp : PString ;

            if( Command.Length == 0 ) // No more text in buffer 
            {
                if( not Read_Next_Line )
                {
                    Line_Text = new( PASCII_String, Init ) ;
                    exit ;
                }
                Inc( _Current_Line ) ; // Increment line count 
                Token_Line = Get_Current_Line ; // Token line is now same as current line 

                _Leading_Whitespace = Command.Class_Of( Command.At( 1 ) ) == CC_Whitespace ;
                E = 8700 ; // Remove leading/trailing spaces 
                if( Case_Sensitive )
                {
                    E = E - 32 ;  // No uppercase 
                }
                Temp = Command.Edit( E ) ; // Format line 
                Command.Assign( Temp ) ;
                dispose( Temp, Done ) ;
            }
            C = new( PASCII_String, Init ) ;
            C^.Assign( @Command ) ;
            Line_Text = C ;
        }


        public function Token_EOL()
        {
            return ( strlen( $this->$Command ) == 0 ) ;
        }
} // TParser



class TFile_Parser extends TParser
{ 
        // Instance data...
        private $_Source : PSFile ; // Input source


        // Internal utility routines... 

        function Read_Next_Line()

        var EF : boolean ;
            Res : PSUnified_Exception ;
            S : string ;

        {
            if( ( _Source == nil ) and ( Command.Length == 0 ) )
            {
                Token_Error = ParserErr_No_Source ;
                Read_Next_Line = False ;
                exit ;
            }

            Ef = False ;
        //
            if( Get_Source^.EOF )
            {
                EF = True ;
            }

            Get_Source^.Readln( S ) ;
            Command.Set_String( S ) ;
            Res = Get_Source^.IO_Error ;
            if( ( Res != nil ) or EF ) // EOF 
            {
                Command.Assign( nil ) ;
                Read_Next_Line = False ;
                Token_Error = ParserErr_Unexpected_End_Of_File ; // Unexpected end of file 
                Exit ;
            }
            Read_Next_Line = True ;
        } // Read_Next_Line 


        function Get_Source()

        {
            return $this->$_Source ;
        }


        function Set_Source( Value : PSFile ) ;

        {
            $this->$_Source = Value ;
            $this->$Command = '' ;
        }
}


class TString_Parser extends TParser
{
        // Instance data...
        private $_Source : TASCII_String ; // Input source
        private $_Supplied = False ; // True if the source has already been supplied


        // Constructors and destructors...

        function __construct Init ;

        {
            inherited Init ;

            _Source.Init ;
            _Supplied = False ;
        }


        // Internal utility routines...

        function Read_Next_Line()
        {
            if( _Supplied )
            {
                Command.Assign( nil ) ;
                Read_Next_Line = False ;
            }
            else
            {
                _Supplied = True ;
                Command.Assign( @_Source ) ;
                Read_Next_Line = True ;
            }
        }


        // Property handlers... 

        function Get_Source()
        {
            return $this->$_Source ;
        }


        function Set_Source( $Value )
        {
            $this->$_Source = $Value ;
            $this->$Command = $Value ;
            $this->$_Supplied = True ;
        }
}



class TStrings_Parser extends TParser
{
        // Instance data...
        private $_Source : TStringList ;
        private $Position = 0 ;

        // Internal utility routines... 

        function Read_Next_Line()
        {
            if( Position < _Source.Count )
            {
                Command.Assign( Temp_String( _Source[ Position ] ) ) ;
                inc( Position ) ;
                Read_Next_Line = True ;
            }
                else
            {
                Read_Next_Line = False ;
            }
        }


        // API... 

        function Get_Source()

        {
            return $this->$_Source ;
        }


        function Set_Source( Value : TStringList )

        {
            $this->$_Source = Value ;
            $this->$Position = 0 ;
        }
}



class TFiles_Parser extends TFile_Parser
{
        // Instance data... 
        private $_Files : Collect.TCollection ;

        // API... 

        function Close_File()

        var F : PSFile ;
        {
            F = _Files.At( _Files.Count - 1 ) ;
            dispose( F, Done ) ;
            _Files.AtDelete( _Files.Count - 1 ) ;
            if( _Files.Count == 0 )
            {
                F = nil ;
            end else
            {
                F = _Files.At( _Files.Count - 1 ) ;
            }
            Set_Source( F ) ;
        }


        function Open_File( const Name : string ; F : PSFile )

        var UEC : PSUnified_Exception ;
        {
            if( F == nil )
            {
                F = _FiP.Open( Name, FSOM_Input, UEC ) ;
            }
            _Files.Insert( F ) ;
            Set_Source( F ) ;
        }
}
?>
