{
        Program Name : RPG_Character
        Package Name :
        Purpose      : Generic RPG character object
        Institution  : Conroy & Conroy Co.
        Date Written : 01-Aug-1999
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit defines the TCharacter and related classes.
}

unit RPG_Character ;

interface

uses // Borland...
     Windows, // HBitmap
     Classes, // TList

     // C&C Subroutine Library - Layer 0...
     COMInter, // TCommon_COM_Interface
     _RPG_Character ; // TCharacter

type TGeneric_Character = class( TCharacter )
         private // Instance data...
             _Bitmap : HBitmap ;
             _Gender : TGender ;
             _Name : Ansistring ;
             _Characteristics : TList ;
             Temp : Ansistring ;

         public
             function Get_Bitmap : HBitmap ; override ;
             procedure Set_Bitmap( Value : HBitmap ) ; override ;
             function Get_Object : TCharacter_Object ; override ;
             procedure Set_Object( Value : TCharacter_Object ) ; override ;

             procedure Assign( Instance : TCharacter ) ; override ;
             function Name_Matches( X : PAnsiChar ) : boolean ; override ;
             function Serialize : PAnsiChar ; override ;
             procedure Deserialize( Buffer : PAnsiChar ) ; override ;
             function Get_Gender : TGender ; override ;
             procedure Set_Gender( Value : TGender ) ; override ;
             function Get_Name : PAnsiChar ; override ;
             procedure Set_Name( Value : PAnsiChar ) ; override ;
             function Description : PAnsiChar ; override ;
             function Status : PAnsiChar ; override ;
             function Item_Manager : TItem_Manager ; override ;
             function Effect_Manager : TEffect_Manager ; override ;

             function Get_Characteristic_By_Name( Name : PAnsiCHar ) : TCharacteristic ;
                 override ;
             function Get_Characteristic( Index : integer ) : TCharacteristic ;
                 override ;
             function Add_Characteristic( Value : TCharacteristic ) : integer ;
                 override ;
             function Set_Characteristic( Index : integer ; Value : PAnsiChar ;
                 Length : integer ) : integer ; override ;
             function Set_Characteristic_List( Index : integer ;
                 List : TString_List ) : integer ; override ;
     end ; // TGeneric_Character


type TGeneric_Characteristic = class( TCharacteristic )
                                   private // Instance data...
                                       _Name : AnsiString ;
                                       _Value : AnsiString ;
                                       _List : TString_List ;
                                       _Index : integer ;
                                       _Flags : cardinal ;
                                       Temp : AnsiString ;

                                   public // API...
                                       function Name : PAnsiChar ; override ;
                                       function Get_Value : PAnsiChar ;
                                           override ;
                                       procedure Set_Value( Value : PAnsiChar ) ;
                                           override ;
                                       function Length : integer ; override ;
                                       function Flags : cardinal ; override ;
                                       function Get_List : TString_List ;
                                           override ;
                                       procedure Set_List( Value : TString_List ) ;
                                           override ;
                                       function Get_Index : integer ; override ;
                                       procedure Set_Index( Value : integer ) ;
                                           override ;
                                       function Serialize : PAnsiChar ;
                                           override ;
                                       procedure Deserialize( Buffer : PAnsiChar ) ;
                                           override ;

                                       procedure Set_Name( S : Ansistring ) ;
                                           virtual ;
                                       procedure Set_Flags( F : cardinal ) ;
                                           virtual ;
                               end ; // TGeneric_Characteristic

     TGeneric_Effect = class( TEffect )
                           private
                               _Name : AnsiString ;
                               Temp : AnsiString ;

                           public
                               function Get_Name : PAnsiChar ; override ;
                               procedure Set_Name( Value : PAnsiChar ) ;
                                   override ;
                               function Serialize : PAnsiChar ; override ;
                               procedure Deserialize( Buffer : PAnsiChar ) ;
                                   override ;
                               function Advance( Time_Index : int64 ;
                                   C : TCharacter ) : boolean ; override ;
                       end ;

     TGeneric_Effect_Manager = class( TEffect_Manager )
                                   protected
                                       _Effects : TStringList ;
                                       Temp : Ansistring ;

                                   protected
                                       function Effects : TStringList ;

                                   public
                                       procedure Clear ; override ;
                                       function Count : longint ; override ;
                                       function Get_Effect( Index : integer ) : TEffect ;
                                           override ;
                                       procedure Set_Effect( Index : integer ; Value : TEffect ) ;
                                           override ;
                                       function Add_Effect( Effect : TEffect ) : longint ;
                                           override ;
                                       function Add_Name( Name : PAnsiChar ) : longint ;
                                           override ;
                                       procedure Advance( Target : TCharacter ) ;
                                           override ;
                                       procedure Remove( Name : PAnsiChar ) ;
                                           override ;
                                       function Serialize : PAnsiChar ;
                                           override ;
                                       procedure Deserialize( Buffer : PAnsiChar ) ;
                                           override ;
                                       function Get_Effect_Library : TString_List ;
                                           override ;
                                       function Create_Effect( Name : PAnsiChar ) : TEffect ;
                                           override ;
                               end ;

     TGeneric_Item = class( TItem )
                         protected
                             _Name : Ansistring ;
                             Temp : Ansistring ;

                         public
                             function Get_Name : PAnsiChar ; override ;
                             procedure Set_Name( Value : PAnsiChar ) ; override ;
                             function Serialize : PAnsiChar ; override ;
                             procedure Deserialize( Buffer : PAnsiChar ) ;
                                 override ;
     end ;

     TGeneric_Item_Manager = class( TItem_Manager )
                                 protected
                                     _Items : TStringList ;
                                     Temp : Ansistring ;

                                 protected
                                     function Items : TStringList ;

                                 public
                                     procedure Clear ; override ;
                                     function Count : longint ; override ;
                                     function Get_Item( Index : integer ) : TItem ;
                                         override ;
                                     procedure Set_Item( Index : integer ; Value : TItem ) ;
                                         override ;
                                     function Add( Item : TItem ) : longint ; overload ;
                                         override ;
                                     function Add( Name : PAnsiChar ) : longint ; overload ;
                                         override ;
                                     procedure Remove( Name : PAnsiChar ) ;
                                         override ;
                                     function Serialize : PAnsiChar ; override ;
                                     procedure Deserialize( Buffer : PAnsiChar ) ;
                                         override ;
                                     function Create_Item( Name : PAnsiChar ) : TItem ;
                                         override ;
                             end ;


implementation

uses // Pascal...
     Sysutils, // inttostr

     // C&C...
     HTML ; // TXML_Parser

// TGeneric_Character methods...

function TGeneric_Character.Get_Bitmap : HBitmap ;

begin
    Result := _Bitmap ;
end ;


procedure TGeneric_Character.Set_Bitmap( Value : HBitmap ) ;

begin
    _Bitmap := Value ;
end ;


function TGeneric_Character.Get_Object : TCharacter_Object ;

begin
    Result := nil ;
end ;


procedure TGeneric_Character.Set_Object( Value : TCharacter_Object ) ;

begin
end ;


procedure TGeneric_Character.Assign( Instance : TCharacter ) ;

begin
end ;


function TGeneric_Character.Name_Matches( X : PAnsiChar ) : boolean ;

begin
end ;


function TGeneric_Character.Serialize : PAnsiChar ;

var I : integer ;

begin
    Temp := '<name>' + _Name + '<name/>' ;
    if( _Characteristics <> nil ) then
    begin
        for I := 0 to _Characteristics.Count - 1 do
        begin
            Temp := Temp + '<char>' + TCharacteristic( _Characteristics[ I ] ).Serialize + '<char/>' ;
        end ;
    end ;
    Temp := Temp + '<effects>' + Effect_Manager.Serialize + '<effects/>' ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Character.Deserialize( Buffer : PAnsiChar ) ;

begin
    //~~~
end ;


function TGeneric_Character.Get_Gender : TGender ;

begin
    Result := _Gender ;
end ;


procedure TGeneric_Character.Set_Gender( Value : TGender ) ;

begin
    _Gender := Value ;
end ;


function TGeneric_Character.Get_Name : PAnsiChar ;

begin
    Result := PAnsiChar( _Name ) ;
end ;


procedure TGeneric_Character.Set_Name( Value : PAnsiChar ) ;

begin
    _Name := Value ;
end ;


function TGeneric_Character.Description : PAnsiChar ;

begin
    Result := nil ;
end ;


function TGeneric_Character.Status : PAnsiChar ;

begin
    Result := nil ;
end ;


function TGeneric_Character.Item_Manager : TItem_Manager ;

begin
    Result := nil ;
end ;


function TGeneric_Character.Effect_Manager : TEffect_Manager ;

begin
    Result := TGeneric_Effect_Manager.Create ;
end ;


function TGeneric_Character.Get_Characteristic_By_Name( Name : PAnsiCHar ) : TCharacteristic ;

var I : integer ;

begin
    Result := nil ;
    if( _Characteristics = nil ) then
    begin
        exit ;
    end ;
    for I := 0 to _Characteristics.Count - 1 do
    begin
        Result := TCharacteristic( _Characteristics[ I ] ) ;
        if( Result.Name = Name ) then
        begin
            exit ;
        end ;
    end ;
    Result := nil ;
end ;


function TGeneric_Character.Get_Characteristic( Index : integer ) : TCharacteristic ;

begin
    Result := nil ;
    if( _Characteristics = nil ) then
    begin
        exit ;
    end ;
    if( ( Index < 0 ) or ( Index > _Characteristics.Count - 1 ) ) then
    begin
        exit ;
    end ;
    Result := TCharacteristic( _Characteristics[ Index ] ) ;
end ;


function TGeneric_Character.Add_Characteristic( Value : TCharacteristic ) : integer ;

begin
    if( Value = nil ) then
    begin
        Result := -3 ;
        exit ;
    end ;
    if( _Characteristics = nil ) then
    begin
        _Characteristics := TList.Create ;
    end ;
    if( Get_Characteristic_By_Name( Value.Name ) <> nil ) then // Already exists
    begin
        Result := -2 ;
        exit ;
    end ;
    _Characteristics.Add( Value ) ;
    Result := 0 ;
end ;


function TGeneric_Character.Set_Characteristic( Index : integer ; Value : PAnsiChar ;
   Length : integer ) : integer ;

var C : TCharacteristic ;

begin
    Result := -1 ;
    if( _Characteristics = nil ) then
    begin
        exit ;
    end ;
    C := Get_Characteristic( Index ) ;
    if( C = nil ) then
    begin
        exit ;
    end ;
    C.Value := Value ;
    Result := 0 ;
end ;


function TGeneric_Character.Set_Characteristic_List( Index : integer ;
   List : TString_List ) : integer ;

var C : TCharacteristic ;

begin
    Result := -1 ;
    if( _Characteristics = nil ) then
    begin
        exit ;
    end ;
    if( ( Index < 0 ) or ( Index > _Characteristics.Count - 1 ) ) then
    begin
        exit ;
    end ;
    C := Get_Characteristic( Index ) ;
    if( C = nil ) then
    begin
        exit ;
    end ;
    C.List := List ;
    Result := 0 ;
end ;



// TGeneric_Characteristic methods...

// API...

function TGeneric_Characteristic.Name : PAnsiChar ;

begin
    Result := PAnsiChar( _Name ) ;
end ;


function TGeneric_Characteristic.Get_Value : PAnsiChar ;

begin
    Result := PAnsiChar( _Value ) ;
end ;


procedure TGeneric_Characteristic.Set_Value( Value : PAnsiChar ) ;

begin
    _Value := Value ;
end ;


function TGeneric_Characteristic.Length : integer ;

begin
    Result := system.length( _Value ) ;
end ;


function TGeneric_Characteristic.Flags : cardinal ;

begin
    Result := _Flags ;
end ;


function TGeneric_Characteristic.Get_List : TString_List ;

begin
    Result := _List ;
end ;


procedure TGeneric_Characteristic.Set_List( Value : TString_List ) ;

var I : integer ;

begin
    if( _List = nil ) then
    begin
        _List := TString_List.Create ;
    end ;
    _List.Clear ;
    if( Value = nil ) then
    begin
        exit ;
    end ;
    for I := 0 to Value.Count - 1 do
    begin
        _List.Add( Value[ I ] ) ;
    end ;
end ;


function TGeneric_Characteristic.Get_Index : integer ;

begin
    Result := _Index ;
end ;


procedure TGeneric_Characteristic.Set_Index( Value : integer ) ;

begin
    _Index := Value ;
end ;


procedure TGeneric_Characteristic.Set_Name( S : Ansistring ) ;

begin
    _Name := S ;
end ;


procedure TGeneric_Characteristic.Set_Flags( F : cardinal ) ;

begin
    _Flags := F ;
end ;


function TGeneric_Characteristic.Serialize : PAnsiChar ;

var I : integer ;
    S : Ansistring ;

begin
    S := '';
    if( _List <> nil ) then
    begin
        for I := 0 to _List.Count - 1 do
        begin
            S := S + '<item>' + _List[ I ] + '<item/>' ;
        end ;
        S := '<list>' + S + '<list/>' ;
    end ;
    Temp := '<name>' + _Name + '<name/>' +
        '<value>' + _Value + '<value/>' +
        '<index>' + inttostr( _Index ) + '<index/>' +
        '<flags>' + inttostr( _Flags ) + '<flags/>' + S ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Characteristic.Deserialize( Buffer : PAnsiChar ) ;

var Parser, List_Parser : TXML_Parser ;
    S : Ansistring ;

begin
    S := AnsiString( Buffer ) ;
    if( List <> nil ) then
    begin
        List.Free ;
    end ;
    Parser := TXML_Parser.Create ;
    try
        Parser.Source := Buffer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Get_Token ;
            if( copy( S, 1, 1 ) = '<' ) then
            begin
                S := lowercase( S ) ;
                if( S = '<name>' ) then
                begin
                    _Name := Parser.Get_Section( 'name' ) ;
                end else
                if( S = '<value>' ) then
                begin
                    _Value := Parser.Get_Section( 'value' ) ;
                end else
                if( S = '<index>' ) then
                begin
                    _Index := strtoint( Parser.Get_Section( 'index' ) ) ;
                end else
                if( S = '<flags>' ) then
                begin
                    _Flags := strtoint( Parser.Get_Section( 'flags' ) ) ;
                end else
                if( S = '<list>' ) then
                begin
                    _List := TString_List.Create ;
                    S := Parser.Get_Section( 'list' ) ;
                    List_Parser := TXML_Parser.Create ;
                    try
                        List_Parser.Source := PAnsiChar( S ) ;
                        while( not List_Parser.EOF ) do
                        begin
                            S := List_Parser.Get_Token ;
                            if( S = '<item>' ) then
                            begin
                                S := List_Parser.Get_Section( 'item' ) ;
                                List.Add( S ) ;
                            end ;
                        end ;
                    finally
                        List_Parser.Free ;
                    end ;
                end ; // if( S = '<list>' )
            end ; // if( copy( S, 1, 1 ) = '<' )
        end ; // while( not Parser.EOF )
    finally
        Parser.Free ;
    end ;
end ; // TGeneric_Characteristic.Deserialize



// TGeneric_Effect_Manager methods...

procedure TGeneric_Effect_Manager.Clear ;

var Index : integer ;

begin
    if( _Effects = nil ) then
    begin
        exit ;
    end ;
    for Index := 0 to Effects.Count - 1 do
    begin
        TEffect( Effects.Objects[ Index ] ).Free ;
    end ;
    Effects.Clear ;
end ;


function TGeneric_Effect_Manager.Effects : TStringList ;

begin
    if( _Effects = nil ) then
    begin
        _Effects := TStringList.Create ;
    end ;
    Result := _Effects ;
end ;


function TGeneric_Effect_Manager.Count : longint ;

begin
    Result := Effects.Count ;
end ;


function TGeneric_Effect_Manager.Get_Effect( Index : integer ) : TEffect ;

begin
    if( ( Index < 0 ) or ( Index >= Count ) ) then
    begin
        Result := nil ;
        exit ;
    end ;
    Result := TEffect( Effects.Objects[ Index ] ) ;
end ;


procedure TGeneric_Effect_Manager.Set_Effect( Index : integer ; Value : TEffect ) ;

begin
    if( ( Index < 0 ) or ( Index >= Count ) ) then
    begin
        exit ;
    end ;
end ;


function TGeneric_Effect_Manager.Add_Effect( Effect : TEffect ) : longint ;

begin
    if( Effect <> nil ) then
    begin
        Effects.AddObject( Effect.Name, Effect ) ;
    end ;
end ;


function TGeneric_Effect_Manager.Add_Name( Name : PAnsiChar ) : longint ;

var E : TGeneric_Effect ;

begin
    if( ( Name <> nil ) and ( Name <> '' ) ) then
    begin
        E := TGeneric_Effect.Create ;
        E.Name := Name ;
        Effects.AddObject( Name, E ) ;
    end ;
end ;


procedure TGeneric_Effect_Manager.Advance( Target : TCharacter ) ;

begin
    //~~~
end ;


procedure TGeneric_Effect_Manager.Remove( Name : PAnsiChar ) ;

var Effect : TEffect ;
    Index : integer ;

begin
    Index := Effects.Indexof( Name ) ;
    if( Index <> -1 ) then
    begin
        Effect := TEffect( Effects.Objects[ Index ] ) ;
        Effects.Delete( Index ) ;
        if( Effect <> nil ) then
        begin
            Effect.Detach ;
        end ;
    end ;
end ;


function TGeneric_Effect_Manager.Serialize : PAnsiChar ;

var I : integer ;

begin
    Temp := '' ;
    for I := 0 to Effects.Count - 1 do
    begin
        Temp := Temp + '<effect>' + TEffect( Effects.Objects[ I ] ).Serialize + '<effect/>' ;
    end ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Effect_Manager.Deserialize( Buffer : PAnsiChar ) ;

var E : TGeneric_Effect ;
    Parser : TXML_Parser ;
    S : Ansistring ;

begin
    Clear ;
    Parser := TXML_Parser.Create ;
    try
        Parser.Source := Buffer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Get_Token ;
            if( S = '<effect>' ) then
            begin
                E := TGeneric_Effect.Create ;
                E.Deserialize( Parser.Get_Section( 'effect' ) ) ;
                Effects.AddObject( E.Name, E ) ;
            end ;
        end ;
    finally
        Parser.Free ;
    end ;
end ;


function TGeneric_Effect_Manager.Get_Effect_Library : TString_List ;

begin
    Result := nil ; // No effects library
end ;


function TGeneric_Effect_Manager.Create_Effect( Name : PAnsiChar ) : TEffect ;

begin
    Result := TGeneric_Effect.Create ;
    TGeneric_Effect( Result )._Name := Name ;
end ;


// TGeneric_Effect methods...

function TGeneric_Effect.Get_Name : PAnsiChar ;

begin
    Result := PAnsiChar( _Name ) ;
end ;


procedure TGeneric_Effect.Set_Name( Value : PAnsiChar ) ;

begin
    if( ( Value <> nil ) and ( Value <> '' ) ) then
    begin
        _Name := Value ;
    end ;
end ;


function TGeneric_Effect.Serialize : PAnsiChar ;

begin
    Temp := '<name>' + _Name + '<name/>' ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Effect.Deserialize( Buffer : PAnsiChar ) ;

var Parser : TXML_Parser ;
    S : Ansistring ;

begin
    Parser := TXML_Parser.Create ;
    try
        Parser.Source := Buffer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Get_Token ;
            if( S = '<name>' ) then
            begin
                Name := Parser.Get_Section( 'name' ) ;
            end ;
        end ;
    finally
        Parser.Free ;
    end ;
end ;


function TGeneric_Effect.Advance( Time_Index : int64 ; C : TCharacter ) : boolean ;

begin
    // This routine intentionally left blank
    Result := False ;
end ;



// TGeneric_Item methods...

function TGeneric_Item.Get_Name : PAnsiChar ;

begin
    Result := PAnsiChar( _Name ) ;
end ;


procedure TGeneric_Item.Set_Name( Value : PAnsiChar ) ;

begin
    _Name := Value ;
end ;


function TGeneric_Item.Serialize : PAnsiChar ;

begin
    Temp := '<name>' + _Name + '<name/>' ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Item.Deserialize( Buffer : PAnsiChar ) ;

var Parser : TXML_Parser ;
    S : Ansistring ;

begin
    Parser := TXML_Parser.Create ;
    try
        Parser.Source := Buffer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Get_Token ;
            if( S = '<name>' ) then
            begin
                Name := Parser.Get_Section( 'name' ) ;
            end ;
        end ;
    finally
        Parser.Free ;
    end ;
end ;



// TGeneric_Item_Manager methods.

function TGeneric_Item_Manager.Items : TStringList ;

begin
    if( _Items = nil ) then
    begin
        _Items := TStringList.Create ;
    end ;
    Result := _Items ;
end ;


procedure TGeneric_Item_Manager.Clear ;

begin
    Items.Clear ;
end ;


function TGeneric_Item_Manager.Count : longint ;

begin
    Result := Items.Count ;
end ;


function TGeneric_Item_Manager.Get_Item( Index : integer ) : TItem ;

begin
    Result := nil ;
    if( ( Index < 0 ) or ( Index >= Items.Count ) ) then
    begin
        exit ;
    end ;
    Result := TItem( Items.Objects[ Index ] ) ;
end ;


procedure TGeneric_Item_Manager.Set_Item( Index : integer ; Value : TItem ) ;

begin
    if( ( Index < 0 ) or ( Index >= Items.Count ) ) then
    begin
        exit ;
    end ;
    if( Items.Objects[ Index ] <> Value ) then
    begin
        Items.Objects[ Index ].Free ;
    end ;
    Items.Objects[ Index ] := Value ;
    Items[ Index ] := Value.Name ;
end ;


function TGeneric_Item_Manager.Add( Item : TItem ) : longint ;

begin
    Result := Items.Count ;
    Items.AddObject( Item.Name, Item ) ;
end ;


function TGeneric_Item_Manager.Add( Name : PAnsiChar ) : longint ;

var I : TItem ;

begin
    Result := Items.Count ;
    I := Create_Item( Name ) ;
    Items.AddObject( Name, I ) ;
end ;


procedure TGeneric_Item_Manager.Remove( Name : PAnsiChar ) ;

var I : integer ;

begin
    I := Items.IndexOf( Name ) ;
    if( I <> -1 ) then
    begin
        Items.Objects[ I ].Free ;
        Items.Objects[ I ] := nil ;
    end ;
end ;


function TGeneric_Item_Manager.Serialize : PAnsiChar ;

var I : integer ;

begin
    Temp := '' ;
    for I := 0 to Items.Count - 1 do
    begin
        Temp := Temp + '<item>' + TItem( Items.Objects[ I ] ).Serialize + '<item/>' ;
    end ;
    Result := PAnsiChar( Temp ) ;
end ;


procedure TGeneric_Item_Manager.Deserialize( Buffer : PAnsiChar ) ;

var Dummy : integer ;
    E : TItem ;
    Parser : TXML_Parser ;
    S : Ansistring ;

begin
    Clear ;
    Parser := TXML_Parser.Create ;
    try
        Parser.Source := Buffer ;
        while( not Parser.EOF ) do
        begin
            S := Parser.Get_Token ;
            if( S = '<item>' ) then
            begin
                E := Create_Item( '' ) ;
                E.Deserialize( Parser.Get_Section( 'item' ) ) ;
                Items.AddObject( E.Name, E ) ;
            end ;
        end ;
    finally
        Parser.Free ;
    end ;
end ;


function TGeneric_Item_Manager.Create_Item( Name : PAnsiChar ) : TItem ;

begin
    Result := TGeneric_Item.Create ;
    Result.Name := Name ;
end ;



end.
