{
        File Name    : StoreLis
        Package Name : Subroutine Library - Layer 4
        Purpose      : Store List
        Institution  : Conroy & Conroy Co.
        Date Written : 29-Oct-2005
        Written By   : Alan Conroy
        Version      : 1.0

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains a list class where the data is stored on a
        managed store.
}

unit StoreLis ;

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

{$I EDEFINES.INC}

interface

uses // C&C Subroutine Library - Layer 0...
     _StoreLis, // TStore_List_64
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}

     // C&C Subroutine Library - Layer 1...
     _Stores, // TStore_Address

     // C&C Subroutine Library - Layer 2...
     Allocation_Cluster_Manager, // TCOM_Allocation_Cluster_Manager64
     MStores, // PSManaged_Store

     // C&C Subroutine Library - Layer 4...
     BigFH ; // TBig_FH

type TList_Record = packed record // Header for TStore_List64
                         Capacity : int64 ; // Maximum number of entries allocated
                         Count : int64 ; // Number of entries used
                         Delta : int64 ; // Minimum capacity extension size ( < 1 is = 1 ) in entries
                         Version : int64 ; // Version of TStore_List64 used to create the list
                         List : int64 ; // Pointer to first chunk of list
                         Reserved : int64 ; // Reserved for future use
                    end ;

     TList_Chunk_Record = packed record // Header for TStore_List64
                               Size : int64 ; // Entries in this chunk
                               Next : int64 ; // Pointer to next chunk
                               Extension : int64 ; // Reserved for future use
                               // All data after this is the list data
                          end ;

     TStore_List64 = class( TStore_List_64 )
                         public // Constructors and destructors...
                             constructor Create ;

                         private // Instance data...
                             _Address : int64 ; // Address of header
                             _Store : PSBig_FH ;
                             Header : TList_Record ;

                         private // Cached data...
                             Cached_Address : int64 ; // Address of last chunk (0 if nothing cached)
                             Cached_Size : int64 ; // Size of last chunk
                             Cached_Offset : int64 ; // First offset in last chunk

                         protected // Property handlers...
                             function Get_Item( Index : int64 ) : int64 ;
                                 override ;
                             function Get_Address : int64 ;
                                 override ;
                             function Get_Capacity : int64 ;
                                 override ;
                             function Get_Count : int64 ;
                                 override ;
                             function Get_Delta : int64 ;
                                 override ;
                             function Get_Store : PSBig_FH ;
                                 override ;
                             procedure Set_Item( Index, Value : int64 ) ;
                                 override ;
                             procedure Set_Address( Value : int64 ) ;
                                 override ;
                             procedure Set_Capacity( Value : int64 ) ;
                                 override ;
                             procedure Set_Count( Value : int64 ) ;
                                 override ;
                             procedure Set_Delta( Value : int64 ) ;
                                 override ;
                             procedure Set_Store( Value : PSBig_FH ) ;
                                 override ;

                         protected // Internal utility routines...
                             function Address_Of( Index : int64 ) : int64 ;

                         public // API...
                             function Add( Value : int64 ) : int64 ;
                                 override ;
                             function Add_Insert( Value : int64 ) : int64 ;
                                 override ;
                             function IndexOf( Value : int64 ) : int64 ;
                                 override ;
                             procedure Update_Header ;
                                 override ;
                     end ; // TStore_List64

type TStandard_COM_Store_List = class( TCOM_Store_List )
                              private // Instance data...
                                  Cached_Address : int64 ; // Address of last chunk (0 if nothing cached)
                                  _Address : TStore_Address ; // Address of header
                                  _Store : TManaged_Store ;
                                  Header : TList_Record ;
                                  Cached_Offset : int64 ; // First offset in last chunk
                                  Cached_Size : int64 ; // Size of last chunk

                              protected // Internal utility routines...
                                  function Address_Of( Index : int64 ) : int64 ;

                              public { API... }
                                  function Get_Address : longint ;
                                      override ; stdcall ;
                                  procedure Set_Address( Value : longint ) ;
                                      override ; stdcall ;
                                  function Add( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  function Get_Count : longint ;
                                      override ; stdcall ;
                                  procedure Set_Count( Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Item( Index : longint ) : longint ;
                                      override ; stdcall ;
                                  procedure Set_Item( Index, Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Capacity : longint ;
                                      override ; stdcall ;
                                  procedure Set_Capacity( Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Delta : longint ;
                                      override ; stdcall ;
                                  procedure Set_Delta( Value : longint ) ;
                                      override ; stdcall ;

                                  function Get_Store : TManaged_Store ;
                                      override ; stdcall ;
                                  procedure Set_Store( Value : TManaged_Store ) ;
                                      override ; stdcall ;

                              public // API...
                                  function Add_Insert( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  function IndexOf( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  procedure Update_Header ;
                                      override ; stdcall ;
                          end ; // TStandard_COM_Store_List

type TStandard_COM_Store64_List = class( TCOM_Store64_List )
                              public // Constructors and destructors...
                                  constructor Create ;
                                  destructor Destroy ; override ;

                              private // Instance data...
                                  _ACM : TCOM_Allocation_Cluster_Manager64 ;
                                  _Address : TStore_Address64 ; // Address of header
                                  _Store : TCOM_Managed_Store64 ;
                                  Header : TList_Record ;
                                  _Buffer : PChar ;
                                  _Current_Buffer : TStore_Address64 ; // Current location of Buffer

                              protected { Internal utility routines... }
                                  function Index_To_Offset( Index : longint ) : longint ;
                                  procedure Read( P : TStore_Address64 ) ;
                                  procedure Write( P : TStore_Address64 ) ;

                              public { API... }
                                  function Get_Address : TStore_Address64 ;
                                      override ; stdcall ;
                                  procedure Set_Address( Value : TStore_Address64 ) ;
                                      override ; stdcall ;
                                  function Add( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  function Get_Count : longint ;
                                      override ; stdcall ;
                                  procedure Set_Count( Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Item( Index : longint ) : longint ;
                                      override ; stdcall ;
                                  procedure Set_Item( Index, Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Capacity : longint ;
                                      override ; stdcall ;
                                  procedure Set_Capacity( Value : longint ) ;
                                      override ; stdcall ;
                                  function Get_Delta : longint ;
                                      override ; stdcall ;
                                  procedure Set_Delta( Value : longint ) ;
                                      override ; stdcall ;

                                  function Get_Store : TCOM_Managed_Store64 ;
                                      override ; stdcall ;
                                  procedure Set_Store( Value : TCOM_Managed_Store64 ) ;
                                      override ; stdcall ;

                              public // API...
                                  function Is_Class( Name : PChar ) : boolean ;
                                      override ; stdcall ;
                                  function Add_Insert( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  function IndexOf( Value : longint ) : longint ;
                                      override ; stdcall ;
                                  procedure Update_Header ;
                                      override ; stdcall ;
                          end ; // TStandard_COM_Store64_List

{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}

const Store_List64_Err_Invalid_List_Index = 1 ;

implementation

uses // Pascal...
     Sysutils, // lowercase
{$IFDEF CC_Regression_Test}
     Heaps, // THeap64
{$ENDIF}

     // C&C Subroutine Library - Layer 0...
     L10n, { Text_* }
     _Types, // Subroutine_Library_Version
     _UE, // tUnified_Exception

     // C&C Subroutine Library - Layer 1...
     UE ; // TUnified_Exception

{ TStore_List64 is a class that wraps a list in a big file heap.  In the heap,
  the list has a header (see List_Record for the header image).  The List's data
  is organized in a single-linked list of "chunks".  Each chunk contains a size,
  pointer, and the data for that chunk, like so:
    Size : int64                    Size of this chunk, in elements (bytes div 8)
    Next : int64                    Pointer to next chunk (0 if this is the last chunk)
    Delta : int64                   Minimum capacity extension size ( < 1 is = 1 )
    Data : array[ size ] of int64   Data for this chunk

  An example list:
      Header: Capacity = 10
              Count = 9
              Delta = 0
              Version = 0
              List ---------> Size = 7
                              Next -----------> Size = 3
                              Item[ 0 ]         Next = 0
                              Item[ 1 ]         Item[ 7 ]
                              Item[ 2 ]         Item[ 8 ]
                              Item[ 3 ]         Item[ 9 ] (unused)
                              Item[ 4 ]
                              Item[ 5 ]
                              Item[ 6 ]
}

type TStore_List64_Exception = class( TStandard_Error_Interface )
                                   private
                                       _Code : longint ;
                                       Temp : string ;

                                   public
                                       function Get_Facility : longint ;
                                           override ;
                                       function Get_Facility_Version : longint ;
                                           override ;
                                       function Version : longint ; override ;
                                       function Severity : longint ; override ;
                                       function Error_Text( var Size, Typ : longint ) : PChar ;
                                           override ;
                                       function Get_Error : longint ; override ;
                                       function Get_Previous : tUnified_Exception ;
                                           override ;
                                       procedure Terminate ; override ;
                               end ;

function TStore_List64_Exception.Get_Facility : longint ; stdcall ;

begin
    Result := 130 ;
end ;


function TStore_List64_Exception.Get_Facility_Version : longint ; stdcall ;

begin
    Result := Subroutine_Library_Version ;
end ;


function TStore_List64_Exception.Version : longint ; stdcall ;

begin
    Result := Subroutine_Library_Version ;
end ;


function TStore_List64_Exception.Severity : longint ; stdcall ;

begin
    Result := UE_Error ;
end ;


function TStore_List64_Exception.Error_Text( var Size, Typ : longint ) : PChar ; stdcall ;

begin
    case _Code of
        Store_List64_Err_Invalid_List_Index : Temp := Text_Error_Invalid_List_Index ;
    end ;
    Size := length( Temp ) ;
    Typ := 0 ;
    Result := PChar( Temp ) ;
end ;


function TStore_List64_Exception.Get_Error : longint ; stdcall ;

begin
    Result := _Code ;
end ;


function TStore_List64_Exception.Get_Previous : tUnified_Exception ; stdcall ;

begin
    Result := nil ;
end ;


procedure TStore_List64_Exception.Terminate ; stdcall ;

begin

end ;



function Create_Store_List64_Exception( C : longint ) : TUnified_Exception ;

begin
    Result := TStore_List64_Exception.Create ;
    TStore_List64_Exception( Result )._Code := C ;
end ;


// TStore_List64 methods...

// Constructors and destructors...

constructor TStore_List64.Create ;

begin
    inherited Create ;

    Header.Delta := 4 ;
end ;


// Property handlers...

function TStore_List64.Get_Item( Index : int64 ) : int64 ; stdcall ;

var This_Address : int64 ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
        Result := -1 ;
        exit ;
    end ;
    This_Address := Address_Of( Index ) ;
    _Store^.Read( This_Address, sizeof( Result ), Result ) ;
end ;


function TStore_List64.Get_Address : int64 ; stdcall ;

begin
    Result := _Address ;
end ;


function TStore_List64.Get_Capacity : int64 ; stdcall ;

begin
    Result := Header.Capacity ;
end ;


function TStore_List64.Get_Count : int64 ; stdcall ;

begin
    Result := Header.Count ;
end ;


function TStore_List64.Get_Delta : int64 ; stdcall ;

begin
    Result := Header.Delta ;
end ;


function TStore_List64.Get_Store : PSBig_FH ; stdcall ;

begin
    Result := _Store ;
end ;


procedure TStore_List64.Set_Item( Index, Value : int64 ) ; stdcall ;

var This_Address : int64 ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
    end ;
    This_Address := Address_Of( Index ) ;
    _Store^.Write( This_Address, sizeof( Value ), Value ) ;
end ;


procedure TStore_List64.Set_Address( Value : int64 ) ; stdcall ;

begin
    if( _Address <> Value ) then
    begin
        _Address := Value ;
        if( ( Value <> 0 ) and ( _Store <> nil ) ) then
        begin
            _Store^.Read( _Address, sizeof( Header ), Header ) ;
        end else
        begin
            fillchar( Header, sizeof( Header ), 0 ) ;
        end ;
        Cached_Address := 0 ;
    end ;
end ;


procedure TStore_List64.Set_Capacity( Value : int64 ) ; stdcall ;

var Current : int64 ;
    Hdr : TList_Chunk_Record ;

begin
    if( Value > Header.Capacity ) then // For now, we don't shrink the list size
    begin
        // Update capacity in header...
        if( ( Header.Delta > 1 ) and ( Value < Header.Capacity + Header.Delta ) ) then
        begin
            Value := Header.Capacity + Header.Delta ;
        end ;
        Header.Capacity := Value ;

        if( Cached_Address <> 0 ) then
        begin
            Current := Cached_Address ;
            Value := Value - Cached_Offset ;
        end else
        begin
            if( Header.List = 0 ) then // Nothing allocated yet
            begin
                Header.List := _Store^.Allocate( Value * sizeof( int64 ) + sizeof( Hdr ) ) ;
                _Store^.Fill( Header.List, Value * sizeof( int64 ) + sizeof( Hdr ), 0 ) ;
                fillchar( Hdr, sizeof( Hdr ), 0 ) ;
                Hdr.Size := Value ;
                _Store^.Write( Header.List, sizeof( Hdr ), Hdr ) ;
                Update_Header ;
                exit ;
            end ;
            Current := Header.List ;
        end ;
        while( Value >= 0 ) do
        begin
            _Store^.Read( Current, sizeof( Hdr ), Hdr ) ;
            Value := Value - Hdr.Size ;
            if( Hdr.Next = 0 ) then // Found end of current list
            begin
                Current := _Store^.Allocate( Value * sizeof( int64 ) + sizeof( Hdr ) ) ;
                Hdr.Next := Current ;
                _Store^.Write( Hdr.Next, sizeof( Hdr ), Hdr ) ;
                fillchar( Hdr, sizeof( Hdr ), 0 ) ;
                _Store^.Fill( Hdr.Next, Value * sizeof( int64 ) + sizeof( Hdr ), 0 ) ;
                Hdr.Size := Value ;
                _Store^.Write( Current, sizeof( Hdr ), Hdr ) ;
                break ;
            end ;
            Current := Hdr.Next ; // Follow pointer to next chunk
        end ; // while( Value >= 0 )
    end ; // if( Value > _Capacity )
end ; // TStore_List64.Set_Capacity


procedure TStore_List64.Set_Count( Value : int64 ) ; stdcall ;

begin
    if( Value <> Header.Count ) then
    begin
        if( Value > Header.Capacity ) then
        begin
            Header.Capacity := Value ;
        end ;
        Header.Count := Value ;
        Update_Header ;
    end ;
end ;


procedure TStore_List64.Set_Delta( Value : int64 ) ; stdcall ;

begin
    if( Value <> Header.Delta ) then
    begin
        Header.Delta := Value ;
        Update_Header ;
    end ;
end ;


procedure TStore_List64.Set_Store( Value : PSBig_FH ) ; stdcall ;

begin
    if( Value <> _Store ) then
    begin
        if( Value <> nil ) then
        begin
            Value^.Attach ;
        end ;
        if( _Store <> nil ) then
        begin
            _Store^.Detach ;
        end ;
        _Store := Value ;
        Cached_Address := 0 ;
        _Address := 0 ;
    end ;
end ;


// Internal utility routines...

function TStore_List64.Address_Of( Index : int64 ) : int64 ;

var Current : int64 ;
    Hdr : TList_Chunk_Record ;
    Start_Index : integer ;

begin
    Result := 0 ;
    if( Cached_Address <> 0 ) then
    begin
        Current := Cached_Address ;
        Index := Index - Cached_Offset ;
        Start_Index := Cached_Offset ;
    end else
    begin
        Current := Header.List ;
        Start_Index := 0 ;
    end ;
    while( Index >= 0 ) do
    begin
        _Store^.Read( Current, sizeof( Hdr ), Hdr ) ;
        if( Index < Hdr.Size ) then // Found the chunk in question
        begin
            Result := Current + Index * sizeof( int64 ) + sizeof( Hdr ) ;
            Cached_Address := Current ;
            Cached_Size := Hdr.Size ;
            Cached_Offset := Start_Index ;
            exit ;
        end ;
        Index := Index - Hdr.Size ;
        Start_Index := Start_Index + Hdr.Size ;
        _Store^.Read( Current, sizeof( Hdr ), Hdr ) ;
        if( Hdr.Next = 0 ) then // Should not happen
        begin
            exit ; { UNTESTED }
        end ;
        Current := Hdr.Next ; // Follow pointer to next chunk
    end ; // while( Index >= 0 )
end ; // TStore_List64.Address_Of


procedure TStore_List64.Update_Header ; stdcall ;

begin
    if( ( _Address = 0 ) and ( _Store <> nil ) ) then
    begin
        _Address := _Store^.Allocate( sizeof( Header ) ) ;
    end ;
    if( _Address <> 0 ) then { UNTESTED: else only happens when store is full }
    begin
        _Store^.Write( _Address, sizeof( Header ), Header ) ;
    end ;
end ;


// API...

function TStore_List64.Add( Value : int64 ) : int64 ; stdcall ;

begin
    Result := Header.Count ;
    Header.Count := Header.Count + 1 ;
    Set_Item( Result, Value ) ;
end ;


// Find first 0 entry and set to Value, else add to end.
function TStore_List64.Add_Insert( Value : int64 ) : int64 ; stdcall ;

begin
    Result := IndexOf( 0 ) ;
    if( Result = -1 ) then
    begin
        Result := Add( Value ) ;
    end else
    begin
        Items[ Result ] := Value ;
    end ;
end ;


function TStore_List64.IndexOf( Value : int64 ) : int64 ; stdcall ;

var Index : integer ;

begin
    for Index := 0 to Header.Count - 1 do
    begin
        if( Items[ Index ] = Value ) then
        begin
            Result := Index ;
            exit ;
        end ;
    end ;
    Result := -1 ;
end ;



// TStandard_COM_Store_List methods...

// Internal utility routines...

function TStandard_COM_Store_List.Address_Of( Index : int64 ) : int64 ;

var Current : int64 ;
    Hdr : TList_Chunk_Record ;
    Start_Index : integer ;

begin
    Result := 0 ;
    if( Cached_Address <> 0 ) then
    begin
        Current := Cached_Address ;
        Index := Index - Cached_Offset ;
        Start_Index := Cached_Offset ;
    end else
    begin
        Current := Header.List ;
        Start_Index := 0 ;
    end ;
    while( Index >= 0 ) do
    begin
        _Store.Read( Current, sizeof( Hdr ), Hdr ) ;
        if( Index < Hdr.Size ) then // Found the chunk in question
        begin
            Result := Current + Index * sizeof( int64 ) + sizeof( Hdr ) ;
            Cached_Address := Current ;
            Cached_Size := Hdr.Size ;
            Cached_Offset := Start_Index ;
            exit ;
        end ;
        Index := Index - Hdr.Size ;
        Start_Index := Start_Index + Hdr.Size ;
        _Store.Read( Current, sizeof( Hdr ), Hdr ) ;
        if( Hdr.Next = 0 ) then // Should not happen
        begin
            exit ; { UNTESTED }
        end ;
        Current := Hdr.Next ; // Follow pointer to next chunk
    end ; // while( Index >= 0 )
end ;


{ API... }

function TStandard_COM_Store_List.Get_Address : longint ;

begin
    Result := _Address ;
end ;


procedure TStandard_COM_Store_List.Set_Address( Value : longint ) ;

begin
    if( _Address <> Value ) then
    begin
        _Address := Value ;
        if( ( Value <> 0 ) and ( _Store <> nil ) ) then
        begin
            _Store.Read( _Address, sizeof( Header ), Header ) ;
        end else
        begin
            fillchar( Header, sizeof( Header ), 0 ) ;
        end ;
    end ;
end ;


function TStandard_COM_Store_List.Add( Value : longint ) : longint ;

begin
    Result := Header.Count ;
    Header.Count := Header.Count + 1 ;
    Set_Item( Result, Value ) ;
end ;


function TStandard_COM_Store_List.Get_Count : longint ;

begin
    Result := Header.Count ;
end ;


procedure TStandard_COM_Store_List.Set_Count( Value : longint ) ;

begin
    if( Value > Capacity ) then
    begin
        Set_Capacity( Value ) ;
    end ;
    Header.Count := Value ;
    Update_Header ;
end ;


function TStandard_COM_Store_List.Get_Item( Index : longint ) : longint ;

var This_Address : longint ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
        Result := -1 ;
        exit ;
    end ;
    This_Address := Address_Of( Index ) ;
    _Store.Read( This_Address, sizeof( Result ), Result ) ;
end ;


procedure TStandard_COM_Store_List.Set_Item( Index, Value : longint ) ;

var This_Address : longint ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
    end ;
    This_Address := Address_Of( Index ) ;
    _Store.Write( This_Address, sizeof( Value ), Value ) ;
end ;


function TStandard_COM_Store_List.Get_Capacity : longint ;

begin
    Result := Header.Capacity ;
end ;


procedure TStandard_COM_Store_List.Set_Capacity( Value : longint ) ;

var Current : longint ;
    Hdr : TList_Chunk_Record ;

begin
    if( Value > Header.Capacity ) then // For now, we don't shrink the list size
    begin
        // Update capacity in header...
        if( ( Header.Delta > 1 ) and ( Value < Header.Capacity + Header.Delta ) ) then
        begin
            Value := Header.Capacity + Header.Delta ;
        end ;
        Header.Capacity := Value ;

        if( Cached_Address <> 0 ) then
        begin
            Current := Cached_Address ;
            Value := Value - Cached_Offset ;
        end else
        begin
            if( Header.List = 0 ) then // Nothing allocated yet
            begin
                Header.List := _Store.Allocate( Value * sizeof( longint ) + sizeof( Hdr ) ) ;
                _Store.Fill( Header.List, Value * sizeof( longint ) + sizeof( Hdr ), 0 ) ;
                fillchar( Hdr, sizeof( Hdr ), 0 ) ;
                Hdr.Size := Value ;
                _Store.Write( Header.List, sizeof( Hdr ), Hdr ) ;
                Update_Header ;
                exit ;
            end ;
            Current := Header.List ;
        end ;
        while( Value >= 0 ) do
        begin
            _Store.Read( Current, sizeof( Hdr ), Hdr ) ;
            Value := Value - Hdr.Size ;
            if( Hdr.Next = 0 ) then // Found end of current list
            begin
                Current := _Store.Allocate( Value * sizeof( Value ) + sizeof( Hdr ) ) ;
                Hdr.Next := Current ;
                _Store.Write( Hdr.Next, sizeof( Hdr ), Hdr ) ;
                fillchar( Hdr, sizeof( Hdr ), 0 ) ;
                _Store.Fill( Hdr.Next, Value * sizeof( Value ) + sizeof( Hdr ), 0 ) ;
                Hdr.Size := Value ;
                _Store.Write( Current, sizeof( Hdr ), Hdr ) ;
                break ;
            end ;
            Current := Hdr.Next ; // Follow pointer to next chunk
        end ; // while( Value >= 0 )
    end ; // if( Value > _Capacity )
end ;


function TStandard_COM_Store_List.Get_Delta : longint ;

begin
    Result := Header.Delta ;
end ;


procedure TStandard_COM_Store_List.Set_Delta( Value : longint ) ;

begin
    if( Value <> Header.Delta ) then
    begin
        Header.Delta := Value ;
        Update_Header ;
    end ;
end ;


function TStandard_COM_Store_List.Get_Store : TManaged_Store ;

begin
    Result := _Store ;
end ;


procedure TStandard_COM_Store_List.Set_Store( Value : TManaged_Store ) ;

begin
    if( Value <> _Store ) then
    begin
        if( Value <> nil ) then
        begin
            Value.Attach ;
        end ;
        if( _Store <> nil ) then
        begin
            _Store.Detach ;
        end ;
        _Store := Value ;
        Cached_Address := 0 ;
        _Address := 0 ;
    end ;
end ;


function TStandard_COM_Store_List.Add_Insert( Value : longint ) : longint ;

begin
    Result := IndexOf( 0 ) ;
    if( Result = -1 ) then
    begin
        Result := Add( Value ) ;
    end else
    begin
        Set_Item( Result, Value ) ;
    end ;
end ;


function TStandard_COM_Store_List.IndexOf( Value : longint ) : longint ;

var Index : integer ;

begin
    for Index := 0 to Count - 1 do
    begin
        if( Get_Item( Index ) = Value ) then
        begin
            Result := Index ;
            exit ;
        end ;
    end ;
    Result := -1 ;
end ;


procedure TStandard_COM_Store_List.Update_Header ;

begin
    if( ( _Address = 0 ) and ( _Store <> nil ) ) then
    begin
        _Address := _Store.Allocate( sizeof( Header ) ) ;
    end ;
    if( _Address <> 0 ) then { UNTESTED: else only happens when store is full }
    begin
        _Store.Write( _Address, sizeof( Header ), Header ) ;
    end ;
end ;



// TStandard_COM_Store64_List methods...

// Constructors and destructors...

constructor TStandard_COM_Store64_List.Create ;

begin
    inherited Create ;

    _ACM := TCOM_Allocation_Cluster_Manager64.Create ;
end ;


destructor TStandard_COM_Store64_List.Destroy ;

begin
    _ACM.Detach ;
    Set_Store( nil ) ;

    inherited Destroy ;
end ;


{ Internal utility routines... }

procedure TStandard_COM_Store64_List.Read( P : TStore_Address64 ) ;

begin
    if( _Current_Buffer = P ) then // Already have this chunk in the buffer
    begin
        exit ;
    end ;
    _Store.Read( P, Header.Delta * sizeof( longint ), _Buffer^ ) ;
    _Current_Buffer := P ;
end ;


procedure TStandard_COM_Store64_List.Write( P : TStore_Address64 ) ;

begin
    if( P = -1 ) then // Use last address
    begin
        P := _Current_Buffer ;
    end ;
    _Store.Write( P, Header.Delta * sizeof( longint ), _Buffer^ ) ;
    _Current_Buffer := P ;
end ;


{ API... }

function TStandard_COM_Store64_List.Get_Address : TStore_Address64 ;

begin
    Result := _Address ;
end ;


procedure TStandard_COM_Store64_List.Set_Address( Value : TStore_Address64 ) ;

begin
    if( _Address <> Value ) then
    begin
        _Address := Value ;
        if( ( Value <> 0 ) and ( _Store <> nil ) ) then
        begin
            _Store.Read( _Address, sizeof( Header ), Header ) ;
        end else
        begin
            fillchar( Header, sizeof( Header ), 0 ) ;
        end ;
        _ACM.Set_Clustersize( Header.Delta * sizeof( longint ) ) ;
        _ACM.Set_Root( Header.List ) ;
    end ;
end ;


function TStandard_COM_Store64_List.Add( Value : longint ) : longint ;

var P : TStore_Address64 ;
    S : TStore_Address64 ;

begin
    // Make sure enough data is allocated...
    S := Header.Count + 1 ;
    S := S * sizeof( Value ) ; // Offset in data...
    if( Header.Count >= Header.Capacity ) then // Need a larger amount of data
    begin
        Set_Capacity( Header.Capacity + Header.Delta ) ;
        _ACM.Set_Size( S ) ;
        if( Header.Count >= Header.Capacity ) then // Couldn't expand
        begin
            Result := -1 ; { Untested }
            exit ;
        end ;
    end ;

    // Update header...
    Result := Header.Count ; // This will be the offset we added it at
    inc( Header.Count ) ;
    Update_Header ;

    // Write value...
    P := _ACM.Offset_To_Pointer( S - sizeof( Value ) ) ;
    if( P = 0 ) then
    begin
        Result := -1 ; { Untested.  Indicates a bug or a full store }
        exit ;
    end ;
    Read( P ) ;
    S := Result - ( ( Result div Header.Delta ) * Header.Delta ) ; // Offset in this buffer
    move( Value, _Buffer[ S * sizeof( Value ) ], sizeof( Value ) ) ;
    Write( P ) ;
end ; // TStandard_COM_Store64_List.Add


function TStandard_COM_Store64_List.Get_Count : longint ;

begin
    Result := Header.Count ;
end ;


procedure TStandard_COM_Store64_List.Set_Count( Value : longint ) ;

begin
    Set_Last_Error( nil ) ;
    if( Value < 0 ) then
    begin
        exit ;
    end ;
    if( Value > Capacity ) then
    begin
        Set_Capacity( Value ) ;
    end ;
    if( Value > Capacity ) then // Couldn't expand
    begin
        exit ;
    end ;
    while( Header.Count < Value ) do
    begin
        Add( 0 ) ;
        if( Last_Error <> nil ) then
        begin
            exit ;
        end ;
    end ;
    Header.Count := Value ;
    Update_Header ;
end ;


function TStandard_COM_Store64_List.Index_To_Offset( Index : longint ) : longint ;

var P : TStore_Address64 ;
    S : int64 ;

begin
    S := Index ;
    S := S * sizeof( Result ) ; // Offset in data...
    P := _ACM.Offset_To_Pointer( S ) ;
    Read( P ) ;
    Result := Index - ( ( Index div Header.Delta ) * Header.Delta ) ; // Longint offset in this buffer
    Result := Result * sizeof( longint ) ; // Byte offset in the buffer
end ;


function TStandard_COM_Store64_List.Get_Item( Index : longint ) : longint ;

var S : TStore_Address64 ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
        Result := -1 ;
        exit ;
    end ;

    S := Index_To_Offset( Index ) ;
    move( _Buffer[ S ], Result, sizeof( Result ) ) ;
end ;


procedure TStandard_COM_Store64_List.Set_Item( Index, Value : longint ) ;

var S : TStore_Address64 ;

begin
    if( ( Index < 0 ) or ( Index >= Header.Count ) ) then
    begin
        Set_Last_Error( Create_Store_List64_Exception( Store_List64_Err_Invalid_List_Index ) ) ;
        exit ;
    end ;

    S := Index_To_Offset( Index ) ;
    move( Value, _Buffer[ S ], sizeof( Value ) ) ;
    Write( -1 ) ;
end ;


function TStandard_COM_Store64_List.Get_Capacity : longint ;

begin
    Result := Header.Capacity ;
end ;


procedure TStandard_COM_Store64_List.Set_Capacity( Value : longint ) ;

begin
    if( Value < 0 ) then
    begin
        exit ; // Invalid value
    end ;
    if( _Store = nil ) then
    begin
        exit ;
    end ;
    if( Header.Delta = 0 ) then
    begin
        _ACM.Set_Clustersize( 16 ) ;
        Header.Delta := _ACM.Get_Clustersize div sizeof( longint ) ; // Default
    end ;
    if( Value mod Header.Delta <> 0 ) then
    begin
        Value := Value + Header.Delta ;
    end ;
    Value := ( Value div Header.Delta ) * Header.Delta ; // New size
    _ACM.Set_Size( Value * sizeof( longint ) ) ;
    Header.Capacity := _ACM.Get_Size div sizeof( longint ) ;
    Update_Header ;
end ;


function TStandard_COM_Store64_List.Get_Delta : longint ;

begin
    Result := Header.Delta ;
end ;


procedure TStandard_COM_Store64_List.Set_Delta( Value : longint ) ;

begin
    if( Header.List <> 0 ) then // Already have things allocated
    begin
        exit ; // Cannot change delta
    end ;
    if( Value <> Header.Delta ) then
    begin
        _ACM.Set_Clustersize( Value * sizeof( longint ) ) ;
        Header.Delta := _ACM.Get_Clustersize div sizeof( longint ) ;
        Update_Header ;
        Reallocmem( _Buffer, _ACM.Get_Clustersize ) ;
    end ;
end ;


function TStandard_COM_Store64_List.Get_Store : TCOM_Managed_Store64 ;

begin
    Result := _Store ;
end ;


procedure TStandard_COM_Store64_List.Set_Store( Value : TCOM_Managed_Store64 ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Store <> nil ) then
    begin
        _Store.Detach ;
    end ;
    _Store := Value ;
    _Address := 0 ;
    _Current_Buffer := 0 ;
    if( _Store <> nil ) then
    begin
        _ACM.Set_Store( _Store ) ;
        Header.Delta := _Store.Min_Storage div sizeof( longint ) ;
        _ACM.Set_Clustersize( Header.Delta * sizeof( longint ) ) ;
        Reallocmem( _Buffer, _ACM.Get_Clustersize ) ;
    end ;
end ;


function TStandard_COM_Store64_List.Is_Class( Name : PChar ) : boolean ; stdcall ;

var S : string ;

begin
    S := Name ;
    Result := lowercase( S ) = 'tstandard_com_store64_list' ;
end ;


function TStandard_COM_Store64_List.Add_Insert( Value : longint ) : longint ;

begin
    Result := IndexOf( 0 ) ;
    if( Result = -1 ) then
    begin
        Result := Add( Value ) ;
    end else
    begin
        Set_Item( Result, Value ) ;
    end ;
end ;


function TStandard_COM_Store64_List.IndexOf( Value : longint ) : longint ;

var Index : integer ;

begin
    for Index := 0 to Count - 1 do
    begin
        if( Get_Item( Index ) = Value ) then
        begin
            Result := Index ;
            exit ;
        end ;
    end ;
    Result := -1 ;
end ;


procedure TStandard_COM_Store64_List.Update_Header ;

begin
    if( _Store = nil ) then
    begin
        exit ;
    end ;
    if( _Address = 0 ) then
    begin
        _Address := _Store.Allocate( sizeof( Header ) ) ;
    end ;
    if( _Address <> 0 ) then { UNTESTED: else only happens when store is full }
    begin
        if( Header.List = 0 ) then
        begin
            Header.List := _ACM.Get_Root ;
        end ;
        _Store.Write( _Address, sizeof( Header ), Header ) ;
        if( Header.List <> 0 ) then
        begin
            _ACM.Set_Root( Header.List ) ;
        end ;
    end ;
end ;



{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_COM_Store64_List( T : TTest_Interface ) ;

var Address : int64 ;
    Test : TStandard_COM_Store64_List ;
    Store : TCOM_Heap64 ;

begin
    Store := TCOM_Heap64.Create ;
    Store.Attach ;
    Test := TStandard_COM_Store64_List.Create ;
    if( not Test.Is_Class( 'tstandard_com_store64_list' ) ) then
    begin
        Error ;
    end ;
    Test.Set_Count( 1 ) ;
    if( Test.Get_Count <> 0 ) then
    begin
        Error ;
    end ;
    Test.Set_Capacity( -1 ) ;
    if( Test.Capacity <> 0 ) then
    begin
        Error ;
    end ;

    Test.Set_Store( Store ) ;

    if( Test.Add( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    Test.Set_Capacity( 7 ) ;
    Test.Set_Count( -1 ) ;
    if( Test.Count <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Add( 5 ) <> 1 ) then
    begin
	    Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    Test.Add_Insert( 3 ) ;
    if( Test.Count <> 3 ) then
    begin
        Error ;
    end ;
    if( Test[ 2 ] <> 3 ) then
    begin
        Error ;
    end ;
    Test.Set_Count( 7 ) ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 3 ] <> 0 ) then
    begin
        Error ;
    end ;
    Test[ 2 ] := 10 ;
    if( Test[ 2 ] <> 10 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 6 ] <> 0 ) then
    begin
        Error ;
    end ;
    Test.Add_Insert( 9 ) ;
    if( Test[ 3 ] <> 9 ) then
    begin
        Error ;
    end ;
    if( Test.Indexof( 9 ) <> 3 ) then
    begin
        Error ;
    end ;
    if( Test.Indexof( 12345678 ) <> -1 ) then
    begin
        Error ;
    end ;
    Test.Set_Delta( 8 ) ;
    if( Test.Delta <> 4 ) then
    begin
        Error ;
    end ;
    Test.Set_Delta( 8 ) ;
    Test.Set_Count( Test.Count ) ;
    Address := Test.Get_Address ;
    Test.Set_Store( Test.Get_Store ) ;
    Test.Set_Count( Test.Get_Count - 1 ) ;
    Test.Set_Capacity( 16 ) ;
    Test.Free ;
    Test := TStandard_COM_Store64_List.Create ;
    Test.Set_Store( Store ) ;
    Test.Set_Address( Address ) ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 5 ] <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Delta <> 4 ) then
    begin
        Error ;
    end ;
    Test.Set_Store( nil ) ;
    Test.set_Store( Store ) ;
    Test.Set_Address( Address ) ;
    Test.Set_Address( 0 ) ;
    Test.Set_Address( 0 ) ;
    Test.Capacity := 1 ;
    Test[ -1 ] := 0 ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;

    Test.Free ;
    Store.Detach ;
end ;


procedure Test_Store_List64( T : TTest_Interface ) ;

(*
var Address : int64 ;
    Test : TStore_List64 ;
    Store : TSBig_FH ;
    _Result : PSUnified_Exception ;
*)

begin
(*
    Store.Init( 16, 0, 'C:\LB\QC\StoreList.tst', _Result, nil ) ;
    Store.Attach ;
    Test := TStore_List64.Create ;
    Test.Store := @Store ;

    if( Test.Add( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    Test.Capacity := 7 ;
    if( Test.Add( 5 ) <> 1 ) then
    begin
	Error ;
    end ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    Test.Count := 7 ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 6 ] <> 0 ) then
    begin
        Error ;
    end ;
    Test.Delta := 8 ;
    Test.Delta := 8 ;
    Test.Count := Test.Count ;
    Test.Store := Test.Store ;
    Address := Test.Address ;
    Test.Count := Test.Count - 1 ;
    Test.Capacity := 16 ;
    Test.Free ;
    Test := TStore_List64.Create ;
    Test.Store := @Store ;
    Test.Address := Address ;
    if( Test[ 0 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( Test[ 1 ] <> 5 ) then
    begin
        Error ;
    end ;
    if( Test[ 5 ] <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Delta <> 8 ) then
    begin
        Error ;
    end ;
    Test.Store := nil ;
    Test.Store := @Store ;
    Test.Address := Address ;
    Test.Address := 0 ;
    Test.Address := 0 ;
    Test.Capacity := 1 ;
    try
        Test[ -1 ] := 0 ;
        Error ;
    except
    end ;
    try
        Test[ -1 ] ;
        Error ;
    except
    end ;

    Test.Free ;
    Store.Done ;
*)
end ; // Test_Store_List64


procedure Test_Unit( T : TTest_Interface ) ;

begin
    Test_COM_Store64_List( T ) ;
    Test_Store_List64( T ) ;
end ;
{$ENDIF}



end.
