unit TextEditor ;

interface

uses { Borland... }
     Classes, { TStringList }

     { LB...}
     CommBase, { TBase_Delphi_Object }
     Standard ; { TInteger_List }

type TText_Editor_Observer = class( TBase_Delphi_Object )
                                 public { Property handlers... }
                                     procedure Line_Read( Count : integer ) ;
                                         virtual ; stdcall ; abstract ;
                             end ;

     TText_Editor = class( TBase_Delphi_Object )
                        public { Property handlers... }
                            function Get_Dirty : boolean ;
                                virtual ; stdcall ; abstract ;
                            procedure Set_Dirty( Value : boolean ) ;
                                virtual ; stdcall ; abstract ;
                            function Column_Count : integer ;
                                virtual ; stdcall ; abstract ;
                            function Row_Count : integer ;
                                virtual ; stdcall ; abstract ;

                        public { API... }
                            procedure Open_File( const Nam : string ;
                                Logger : TText_Editor_Observer ) ;
                                virtual ; stdcall ; abstract ;
                            procedure SaveToFile( const Nam : string ) ;
                                virtual ; stdcall ; abstract ;
                            procedure Clear ; virtual ; stdcall ; abstract ;
                            function Get_Row( Row : integer ) : string ;
                                virtual ; stdcall ; abstract ;
                            function Get_Text_Range( Start_Row, Start_Column, End_Row, End_Column : integer ) : string ;
                                virtual ; stdcall ; abstract ;

                        public { Properties... }
                            property Dirty : boolean
                                read Get_Dirty
                                write Set_Dirty ;
                    end ;

type TSimple_Text_Editor = class( TText_Editor )
                               public { Constructors and destructors... }
                                   constructor Create ;
                                   destructor Destroy ; override ;

                               private { Instance data... }
                                   C_Count : integer ;
                                   Lines : TStringList ;
                                   Offsets : TInteger64_List ;
                                   Default_Line_End : string ;
                                   _Dirty : boolean ;
                                   F : file ;
                                   File_Name : string ;

                               public { Property handlers... }
                                   function Get_Dirty : boolean ; override ;
                                   procedure Set_Dirty( Value : boolean ) ;
                                       override ;
                                   function Column_Count : integer ; override ;
                                   function Row_Count : integer ; override ;

                               public { API... }
                                   procedure Open_File( const Nam : string ;
                                       Logger : TText_Editor_Observer ) ; override  ;
                                   procedure SaveToFile( const Nam : string ) ;
                                       override ;
                                   procedure Clear ; override ;
                                   function Get_Row( Row : integer ) : string ;
                                       override ;
                                   function Get_Text_Range( Start_Row, Start_Column, End_Row, End_Column : integer ) : string ;
                                       override ;
                           end ; { TSimple_Text_Editor }

implementation

uses { Borland... }
     SysUtils, { fmOpenRead }

     { LB... }
     ASCIIDef, // LF
     CommonUt ; { Edit }


{ TText_Editor methods... }

{ Constructors and destructors... }

constructor TSimple_Text_Editor.Create ;

begin
    inherited Create ;

    Lines := TStringList.Create ;
    Offsets := TInteger64_List.Create ;
end ;


destructor TSimple_Text_Editor.Destroy ;

begin
    Lines.Free ;
    Lines := nil ;
    Offsets.Free ;
    Offsets := nil ;
    {$I-}
    closefile( F ) ;
    {$I+}
    IOResult ;
    File_Name := '' ;

    inherited Destroy ;
end ;


{ Property handlers... }

function TSimple_Text_Editor.Get_Dirty : boolean ;

begin
    Result := _Dirty ;
end ;


procedure TSimple_Text_Editor.Set_Dirty( Value : boolean ) ;

begin
   _Dirty := Value ;
end ;


function TSimple_Text_Editor.Column_Count : integer ;

begin
    Result := C_Count ;
end ;


function TSimple_Text_Editor.Row_Count : integer ;

begin
    Result := Offsets.Count ;
end ;


{ API... }

procedure TSimple_Text_Editor.Open_File( const Nam : string ;
    Logger : TText_Editor_Observer ) ;

var Last_Offset : int64 ;

    procedure _Add( const S : string ) ;

    begin
        Offsets.Add( Last_Offset + 1 ) ;
        if( Logger <> nil ) then
        begin
            Logger.Line_Read( Offsets.Count ) ;
        end ;
        Last_Offset := Last_Offset + length( S ) ;
        if( length( S ) > C_Count ) then
        begin
            C_Count := length( S ) ;
        end ;
    end ;

var Buffer : array[ 0..255 ] of byte ;
    Count, Dummy : integer ;
    Pos_CR, Pos_LF : integer ;
    S : string ;
    Saved : integer ;
    This_Line : string ;

begin
    { Setup... }
    Default_Line_End := CRLF ;
    This_Line := '' ;
    File_Name := Edit( Nam, -1 ) ;
    Last_Offset := -1 ;

    { Open the file... }
    assignfile( F, Nam ) ;
    Saved := FileMode ;
    FileMode := fmOpenRead ;
    reset( F, 1 ) ;
    FileMode := Saved ;
    C_Count := 0 ;
    Offsets.Clear ;
    Lines.Clear ;

    { Read the file in and deblock it into lines }
    Count := 255 ; // Make loop execute at least once
    while( Count > 0 ) do
    begin
        blockread( F, Buffer, sizeof( Buffer ), Count ) ;
        setlength( S, Count ) ;
        move( Buffer[ 0 ], PChar( S )[ 0 ], Count ) ;
        This_Line := This_Line + S ;
        Pos_CR := pos( CR, This_Line ) ;
        Pos_LF := pos( LF, This_Line ) ;
        if( ( ( Pos_CR = Pos_LF + 1 ) and ( Pos_LF > 0 ) ) or ( Pos_CR = 0 ) ) then
        begin
            Pos_CR := Pos_LF ;
        end ;
        while( ( Pos_CR > 0 ) and ( Pos_CR < length( This_Line ) ) ) do // Ignore if at very end of current line buffer
        begin
            if(
                ( pos( This_Line[ Pos_CR + 1 ], CRLF ) > 0 )
                and
                ( This_Line[ Pos_CR ] <> This_Line[ Pos_CR + 1 ] )
              ) then
            begin
                _Add( copy( This_Line, 1, Pos_CR + 1 ) ) ;
                This_Line := copy( This_Line, Pos_CR + 2, length( This_Line ) ) ;
            end else
            begin
                _Add( copy( This_Line, 1, Pos_CR ) ) ;
                This_Line := copy( This_Line, Pos_CR + 1, length( This_Line ) ) ;
            end ;

            // Find end of next line
            Pos_CR := pos( CR, This_Line ) ;
            Pos_LF := pos( LF, This_Line ) ;
            if( ( ( Pos_CR = Pos_LF + 1 ) and ( Pos_LF > 0 ) ) or ( Pos_CR = 0 ) ) then
            begin
                Pos_CR := Pos_LF ;
            end ;
        end ; // while
    end ; // while( Count > 0 )
    if( This_Line <> '' ) then
    begin
        _Add( This_Line ) ;
    end ;
    if( Logger <> nil ) then
    begin
        Logger.Line_Read( -Offsets.Count ) ;
    end ;
    if( Offsets.Count > 0 ) then
    begin
        S := Get_Row( 0 ) ;
        Dummy := length( S ) ;
        while( ( Dummy > 0 ) and ( pos( S[ Dummy ], CRLF ) > 0 ) ) do
        begin
            dec( Dummy ) ;
        end ;
        Default_Line_End := copy( S, Dummy + 1, length( S ) ) ;
    end ;
end ; // TSimple_Text_Editor.Open_File


procedure TSimple_Text_Editor.SaveToFile( const Nam : string ) ;

var F : textfile ;
    Loop : integer ;

begin
    if( Edit( Nam, -1 ) = File_Name ) then // Writing to same file
    begin
        //~~~
    end ;

    assignfile( F, Nam ) ;
    rewrite( F ) ;
    try
        for Loop := 0 to Offsets.Count - 1 do
        begin
            write( F, Get_Row( Loop ) ) ;
        end ;
    finally
        closefile( F ) ;
    end ;
end ;


procedure TSimple_Text_Editor.Clear ;

begin
    Lines.Clear ;
    Offsets.Clear ;
    {$I-}
    closefile( F ) ;
    {$I+}
    IOResult ;
    File_Name := '' ;
end ;


function TSimple_Text_Editor.Get_Row( Row : integer ) : string ;

var Count : integer ;
    Dummy : integer ;

begin
    if( ( Row < 0 ) or ( Row >= Offsets.Count ) ) then
    begin
        Result := '' ;
        exit ;
    end ;

    {$I-}
    seek( F, Offsets[ Row ] ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
        Result := ERT( Dummy ) ;
        exit ;
    end ;
    if( Row = Offsets.Count - 1 ) then
    begin
        Count := filesize( F ) - Offsets[ Row ] ;
    end else
    begin
        Count := Offsets[ Row + 1 ] - Offsets[ Row ] ;
    end ;
    setlength( Result, Count ) ;
    {$I-}
    blockread( F, PChar( Result )[ 0 ], Count ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
        Result := ERT( Dummy ) ;
        exit ;
    end ;
end ; { TSimple_Text_Editor.Get_Row }


procedure Normalize_Range( var Start_Row, Start_Column, End_Row, End_Column : integer ) ;

var Temp : integer ;

begin
    if( Start_Row > End_Row ) then
    begin
        Temp := End_Row ;
        End_Row := Start_Row ;
        Start_Row := Temp ;
        Temp := End_Column ;
        End_Column := Start_Column ;
        Start_Column := Temp ;
        exit ;
    end ;
    if( Start_Row = End_Row ) then
    begin
        if( Start_Column > End_Column ) then
        begin
            Temp := End_Column ;
            End_Column := Start_Column ;
            Start_Column := Temp ;
        end ;
    end ;
end ;


function TSimple_Text_Editor.Get_Text_Range( Start_Row, Start_Column, End_Row, End_Column : integer ) : string ;

var I : integer ;
    S : string ;

begin
    Result := '' ;
    Normalize_Range( Start_Row, Start_Column, End_Row, End_Column ) ;
    if( Start_Row < 0 ) then
    begin
        Start_Row := 0 ;
        Start_Column := 0 ;
    end ;
    if( End_Row >= Row_Count ) then
    begin
        End_Row := Row_Count - 1 ;
        End_Column := Column_Count ;
    end ;
    if( Start_Column < 0 ) then
    begin
        Start_Column := 0 ;
    end ;
    if( End_Column < 0 ) then
    begin
        End_Column := 0 ;
    end ;
    if( Start_Row = End_Row ) then // Part of a line
    begin
        S := Get_Row( Start_Row ) ;
        Result := copy( S, Start_Column + 1, End_Column - Start_Column ) ;
        exit ;
    end ;
    for I := Start_Row to End_Row do
    begin
        S := Get_Row( I ) ;
        if( I = Start_Row ) then
        begin
            Result := Result + copy( S, Start_Column + 1, length( S ) ) ;
        end else
        if( I = End_Row ) then
        begin
            Result := Result + copy( S, 1, End_Column ) ;
        end else
        begin
            Result := Result + S ;
        end ;
    end ;
end ; // TSimple_Text_Editor.Get_Text_Range(


end.
