{
        Program Name : UOS_File
        Package Name : Subroutine Library Layer 2
        Purpose      : UOS File class and File Header structure
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Oct-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Native file class for UOS
}

unit UOS_File ;

interface

uses // Pascal...
     Classes, // TList

     // C&C Subroutine Library - Layer 0...
     COMInter, // TCommon_COM_Interface
     _Stores, // TStore_Address64
     TypeDefs, // Max_Memory
{$IFDEF CC_Regression_Test}
     _TestInt, // TTest_Interface
{$ENDIF}

     // C&C Subroutine Library - Layer 2...
     _UOS_File_System, // TData_Stream
     Allocation_Cluster_Manager, // TCOM_Allocation_Cluster_Manager64
     MStores, // TCOM_Managed_Store64
     UOS_MStore ; // TUOS_Managed_Store

{$I EDEFINES.INC}

type TDirty_Notice = procedure( Sender : TObject ) of object ;

type TNative_File_ACM = class( TCOM_Allocation_Cluster_Manager64 )
                            public // Instance data...
                                Index : longint ; // Index of data stream name/pointer
                                Name : longint ;
                        end ;

type TUOS_Native_File = class( TUOS_File )
                             public // Constructors and destructors...
                                 constructor Create ;
                                 destructor Destroy ; override ;

                             private // Instance data...
                                 _Buffer : PChar ; // Data stream buffer
                                 _Buffer_Size : longint ; // Size of allocated buffer
                                 _Contiguous : boolean ;
                                 _Data_Stream_ACM : TNative_File_ACM ;
                                 _Max_Stream : longint ;
                                 _Store : TCOM_Managed_Store64 ;
                                 _Streams : TList ; // List of TNative_File_ACMs for accessed streams
                                 _Dirty : boolean ;
                                 _Directory : TUOS_Native_File ; // File where header is stored
                                 _Directory_Index : cardinal ; // Index in Directory
                                 _Directory_Cluster : int64 ; // (First) cluster of the folder file

                             protected // Internal utility routines...
                                 function Data_Stream_ACM : TNative_File_ACM ;

                                 // Load Buffer with buffer for specified stream, and create/return ACM for it
                                 function Find_Data_Stream( Index : longint ) : TNative_File_ACM ;

                                 // Return index of stream with given name...
                                 function IndexOf( Name : int64 ) : longint ;

                                 function Offset_To_Pointer( ACM : TNative_File_ACM ;
                                     Position : TStore_Address64 ) : TStore_Address64 ;

                                 function _Read( ACM : TNative_File_ACM ;
                                     Position : TStore_Address64 ;
                                     Length : TStore_Size64 ; Buff : pointer ) : TStore_Size64 ;
                                 function _Write( ACM : TNative_File_ACM ;
                                     Position : TStore_Address64 ;
                                     Length : TStore_Size64 ; Buff : pointer ) : TStore_Size64 ;
                                 procedure Set_Size( ACM : TNative_File_ACM ;
                                     Size : TStore_Size64 ) ;
                                 function Get_Size( ACM : TNative_File_ACM ) : TStore_Size64 ;

                                 function Buffer : PChar ;

                                 function Streams : TList ;

                                 procedure Update_Header ; virtual ;

                             public // API (overrides)...
                                 function Is_Class( Name : PChar ) : boolean ;
                                     override ;

                                 function Create_Stream( Name : int64 ) : longint ;
                                     override ;

                                 procedure Delete_Stream( Name : int64 ; Index : longint ) ;
                                     override ;

                                 function Max_Stream : longint ; override ;

                                 function Stream_Name( Index : longint ) : int64 ;
                                     override ;

                                 function Read( Stream : longint ;
                                     Position : TStore_Address64 ;
                                     Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                                     override ;

                                 function Write( Stream : longint ;
                                     Position : TStore_Address64 ;
                                     Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;
                                     override ;

                				         function Get_Stream_Size( Stream : longint ) : TStore_Size64 ;
                                     override ;

                		             procedure Set_Stream_Size( Stream : longint ;
                                     Value : TStore_Size64 ) ; override ;

                                 function Get_Contiguous : boolean ; override ;

                                 procedure Set_Contiguous( Value : boolean ) ;
                                     override ;

                                 function Get_File_Size : int64 ;
                                     override ;

                                 procedure Set_File_Size( Value : int64 ) ;
                                     override ;

                                 function Read_Only : boolean ; override ;

                                 function Write_Only : boolean ; override ;

                                 function XSpaceAvail : int64 ; override ;

                             public // API...
                                 Header : TUOS_File_Header ;
                                 On_Dirty : TDirty_Notice ;
                                 On_Destroy : TDirty_Notice ;

                                 function Get_Store : TCOM_Managed_Store64 ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}
                                 procedure Set_Store( Value : TCOM_Managed_Store64 ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Get_Dirty : boolean ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}
                                 procedure Set_Dirty( Value : boolean ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Get_Directory : TUOS_Native_File ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 procedure Set_Directory( Value : TUOS_Native_File ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Get_Directory_Cluster : int64 ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 procedure Set_Directory_Cluster( Value : int64 ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Get_Directory_Index : cardinal ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 procedure Set_Directory_Index( Value : cardinal ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Stream_Pointer( Index : longint ) : TStore_Address64 ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Get_Allocation_Cluster( Stream, Index : longint ) : TStore_Address64 ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 procedure Set_Allocation_Cluster( Stream, Index : longint ;
                                     Value : TStore_Address64 ) ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}
                                 function Get_ACM_Cluster( Stream : longint ;
                                     P : int64 ) : int64 ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                                 function Rebuild : boolean ;
                                     virtual ;
{$IFDEF Use_STDCALL}
                                     stdcall ;
{$ENDIF}

                             public // Properties...
                                 property Directory : TUOS_Native_File
                                     read Get_Directory
                                     write Set_Directory ;
                                 property Directory_Cluster : int64
                                     read Get_Directory_Cluster
                                     write Set_Directory_Cluster ;
                                 property Directory_Index : cardinal
                                     read Get_Directory_Index
                                     write Set_Directory_Index ;
                                 property Dirty : boolean
                                     read Get_Dirty
                                     write Set_Dirty ;
                                 property Store : TCOM_Managed_Store64
                                     read Get_Store
                                     write Set_Store ;
                         end ; // TUOS_Native_File

const UOS_File_Error_Invalid_Stream_Index = 1 ;
      UOS_File_Error_Duplicate_Name = 2 ;
      UOS_File_Error_Read_Failure = 3 ;
      UOS_File_Error_Write_Failure = 4 ;
      UOS_File_Error_No_Such_Stream = 5 ;
      UOS_File_Error_Invalid_Operation = 6 ;
      UOS_File_Error_Read_Past_End = 7 ;
      UOS_File_Error_Data_Structure_Error = 8 ; // Corrupted file structure
      UOS_File_Error_IO_Error = 9 ; // General I/O failure
      UOS_File_Error_Out_Of_Space = 10 ;


{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;
{$ENDIF}


implementation

uses // Pascal...
     sysutils,

     // C&C Subroutine Library - Layer 0...
     _UE, // TUnified_Exception
{$IFDEF CC_Regression_Test}
     Forms,
     dialogs,
     
     commonut,
     Heaps, // TCOM_Heap64
     RAMDisk, // TRAM_Disk64
{$ENDIF}
{$IFDEF CC_Track_Resources}
     ResMon,
{$ENDIF}

     // C&C Subroutine Library - Layer 1...
     UE ; // TStandard_Error_Interface

function ERT( C : longint ) : string ;

begin
    case C of
        UOS_File_Error_Invalid_Stream_Index : Result := 'Invalid stream index' ;
        UOS_File_Error_Duplicate_Name : Result := 'Duplicate Name' ;
        UOS_File_Error_Read_Failure : Result := 'Read failure' ;
        UOS_File_Error_Write_Failure : Result := 'Write failure' ;
        UOS_File_Error_No_Such_Stream : Result := 'No such stream' ;
        UOS_File_Error_Invalid_Operation : Result := 'Invalid operation' ;
        UOS_File_Error_Read_Past_End : Result := 'Read past end' ;
        UOS_File_Error_Data_Structure_Error : Result := 'Data structure error' ;
        UOS_File_Error_IO_Error : Result := 'I/O Error' ;
        UOS_File_Error_Out_Of_Space : Result := 'Out of space' ;
        else Result := 'Unknown error' ;
    end ;
end ;


type tUOS_File_Unified_Exception = class( TStandard_Error_Interface )
                                                        private
                                                            _Code : longint ;
                                                            _Text : string ;
                                                            _Previous : tUnified_Exception ;

                                                        public
                                                            function Get_Facility : longint ; override ;
                                                            function Get_Facility_Version : longint ;
                                                                override ;
                                                            function Version : longint ; override ;
                                                            function Severity : longint ; override ;
                                                            function Error_Text( var Size, Typ : longint ) : PChar ;
                                                                override ;
                                                            function Get_Error : longint ; override ;
                                                            function Get_Previous : tUnified_Exception ;
                                                                override ;
                                                            procedure Terminate ; override ;
                                                    end ;


function tUOS_File_Unified_Exception.Get_Facility : longint ;

begin
    Result := 134 ;
end ;


function tUOS_File_Unified_Exception.Get_Facility_Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_File_Unified_Exception.Version : longint ;

begin
    Result := 10 ;
end ;


function tUOS_File_Unified_Exception.Severity : longint ;

begin
    Result := UE_Error ;
end ;


function tUOS_File_Unified_Exception.Error_Text( var Size, Typ : longint ) : PChar ;

begin
    _Text := ERT( _Code ) ;
    Size := length( _Text ) ;
    Typ := 0 ;
    Result := PChar( _Text ) ;
end ;


function tUOS_File_Unified_Exception.Get_Error : longint ;

begin
    Result := _Code ;
end ;


function tUOS_File_Unified_Exception.Get_Previous : tUnified_Exception ;

begin
    Result := _Previous ;
end ;


procedure tUOS_File_Unified_Exception.Terminate ;

begin
    Free ;
end ;



function Create_Exception( C : longint ; E : TUnified_Exception ) : tUOS_File_Unified_Exception ;

begin
    Result := tUOS_File_Unified_Exception.Create ;
    Result._Code := C ;
    Result._Previous := E ;
end ;



// TUOS_Native_File methods...

// Constructors and destructors...

constructor TUOS_Native_File.Create ;

begin
    inherited Create ;

    _Max_Stream := -1 ;
end ;


destructor TUOS_Native_File.Destroy ;

var Loop : integer ;

begin
    if( assigned( On_Destroy ) ) then
    begin
        On_Destroy( self ) ;
    end ;
    if( Directory <> nil ) then
    begin
        Directory := nil ;
    end ;
    if( _Streams <> nil ) then
    begin
        for Loop := 0 to _Streams.Count - 1 do
        begin
            if( _Streams[ Loop ] <> nil ) then
            begin
                TCOM_Allocation_Cluster_Manager64( _Streams[ Loop ] ).Detach ;
                _Streams[ Loop ] := nil ;
            end ;
        end ;
        _Streams.Free ;
        _Streams := nil ;
    end ;
    if( _Data_Stream_ACM <> nil ) then
    begin
        _Data_Stream_ACM.Detach ;
        _Data_Stream_ACM := nil ;
    end ;
    Store := nil ;
{$IFDEF CC_Track_Resources}
    Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
    freemem( _Buffer ) ;
    _Buffer := nil ;

    inherited Destroy ;
end ; // TUOS_Native_File.Destroy


// Internal utility routines...

function TUOS_Native_File.Data_Stream_ACM : TNative_File_ACM ;

var C : integer ;

begin
    if( _Data_Stream_ACM = nil ) then
    begin
        _Data_Stream_ACM := TNative_File_ACM.Create ;
        _Data_Stream_ACM.Index := -1 ;
        _Data_Stream_ACM.Set_Store( _Store ) ;
        _Data_Stream_ACM.Set_Root( Header.Data_Stream ) ;
        C := 128 ;
        if( _Store.Min_Storage > C ) then
        begin
            C := _Store.Min_Storage ;
        end ;
        _Data_Stream_ACM.Set_Clustersize( C ) ; // Minimum clustersize for streams data
    end ;
    Result := _Data_Stream_ACM ;
end ;


type TData_Stream_Array = array[ 0..Max_Memory div sizeof( TData_Stream ) ] of TData_Stream ;
     PData_Stream_Array = ^TData_Stream_Array ;

// Load Buffer with buffer for specified stream, and create/return ACM for it

function TUOS_Native_File.Find_Data_Stream( Index : longint ) : TNative_File_ACM ;

var DS : TData_Stream ;
    I : TStore_Address64 ;
    Original : longint ;
    _File_Size : int64 ;

begin
    // Setup...
    Result := nil ; // Assume failure
    if( ( Index < 0 ) or ( Index > Max_Stream ) ) then
    begin
        exit ;
    end ;
    if( ( Streams.Count > Index ) and ( Streams[ Index ] <> nil ) ) then
    begin
        Result := TNative_File_ACM( Streams[ Index ] ) ;
        exit ;
    end ;
    Original := Index ;
    I := _Store.Min_Storage ;
    if( I < sizeof( TData_Stream ) * 4 ) then {Else untested}
    begin
        I := sizeof( TData_Stream ) * 4 ;
    end ;

    // Extend stream list...
    while( Streams.Count <= Index ) do
    begin
        Streams.Add( nil ) ;
    end ;

    if( Index <= Max_Header_Data_Stream ) then // Data in the header
    begin
        if( ( Index > 0 ) and ( Header.Streams[ Index ].Name = 0 ) ) then
        begin
            exit ; // Not a valid stream
        end ;
        Result := TNative_File_ACM.Create ;
        Result.Name := Header.Streams[ Index ].Name ;
        Result.Index := Index ;
        Result.Set_Store( _Store ) ;
        Result.Set_Clustersize( I ) ;
        if( Index = 0 ) then // Data stream
        begin
            _File_Size := Header.Size - ( Max_Header_Cluster + 1 ) * Header.Clustersize ;
            if( _File_Size < 0 ) then
            begin
                _File_Size := 0 ;
            end ;
            Result.Set_Root_And_Size( Header.Streams[ Index ].Pointer, _File_Size ) ;
        end else
        begin
            Result.Set_Root( Header.Streams[ Index ].Pointer ) ;
        end ;
        Streams[ Index ] := Result ;
        exit ;
    end ;

    Index := Index - Max_Header_Data_Stream - 1 ;
{$IFDEF CC_Regression_Test}
    if( Index * sizeof( TData_Stream ) > Data_Stream_ACM.Get_Size ) then
    begin
        exit ; // Shouldn't get here past above checks ( Index > Max_Stream )
    end ;
{$ENDIF}
    _Read( Data_Stream_ACM, Index * sizeof( TData_Stream ), sizeof( DS ), @DS ) ;
    Result := TNative_File_ACM.Create ;
    Result.Name := DS.Name ;
    Result.Index := Original ;
    Result.Set_Store( _Store ) ;
    Result.Set_Clustersize( I ) ;
    Result.Set_Root( DS.Pointer ) ;
    Streams[ Original ] := Result ;
end ; // TUOS_Native_File.Find_Data_Stream


function TUOS_Native_File.IndexOf( Name : int64 ) : longint ;

var B : array[ 0..31 ] of TData_Stream ; // 512 bytes (max)
    Loop : integer ;
    P : TStore_Address64 ;

begin
    // Setup...
    Result := -1 ; // Assume not found...
    Set_Last_Error( nil ) ;

    // Check header...
    for Loop := 0 to Max_Header_Data_Stream do
    begin
        if( Name = Header.Streams[ Loop ].Name ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;

    // Check data stream chain...
    P := Header.Data_Stream ;
    if( P = 0 ) then
    begin
        exit ; // No more streams in file
    end ;

    P := 0 ;
    Result := Max_Header_Data_Stream + 1 ;
    while( P < Data_Stream_ACM.Get_Size ) do
    begin
        Loop := _Read( Data_Stream_ACM, P, sizeof( B ), @B ) div sizeof( TData_Stream ) - 1 ;
        if( Last_Error <> nil ) then
        begin
            exit ; { Untested }
        end ;
        while( Loop >= 0 ) do
        begin
            if( Name = B[ Loop ].Name ) then
            begin
                Result := Result + Loop ;
                exit ;
            end ;
            dec( Loop ) ;
        end ;
        P := P + sizeof( B ) ;
        Result := Result + sizeof( B ) div sizeof( TData_Stream ) ;
    end ;

    Result := -1 ; // Not found...
end ; // TUOS_Native_File.IndexOf


function TUOS_Native_File.Offset_To_Pointer( ACM : TNative_File_ACM ;
    Position : TStore_Address64 ) : TStore_Address64 ;

var H : TStore_Address64 ;

begin
    if( ACM = Streams[ 0 ] ) then // Data stream
    begin
        H := Position div ACM.Get_Clustersize ; // Cluster index
        if( Contiguous ) then
        begin
            Result := Header.Clusters[ 0 ] + H * ACM.Get_Clustersize ;
            exit ;
        end else
        begin
            if( H <= Max_Header_Cluster ) then
            begin
                Result := Header.Clusters[ H ] ;
                exit ;
            end ;
            Position := Position - ( Max_Header_Cluster + 1 ) * ACM.Get_Clustersize ;
        end ;
    end ;
    Result := ACM.Offset_To_Pointer( Position ) ;
end ;


procedure TUOS_Native_File.Update_Header ;

begin
    Dirty := True ;
    if( assigned( On_Dirty ) ) then
    begin
        On_Dirty( self ) ;
    end else
    if( Directory <> nil ) then
    begin
        if( Directory.Write( 0, Directory_Index * sizeof( Header ), sizeof( Header ), Header ) = sizeof( Header ) ) then
        begin
            Dirty := False ;
        end ;
    end ;
end ;


function TUOS_Native_File.Get_Size( ACM : TNative_File_ACM ) : TStore_Size64 ;

begin
    if( ACM = Streams[ 0 ] ) then // Data stream
    begin
        Result := Header.Size ;
    end else
    begin
        Result := ACM.Get_Size ;
    end ;
end ;


procedure TUOS_Native_File.Set_Size( ACM : TNative_File_ACM ;
    Size : TStore_Size64 ) ;

var C : cardinal ;
    Old_Header : TUOS_File_Header ;
    H, P : TStore_Address64 ;
    I : longint ;
    Requested_Size : TStore_Address64 ;

begin
    Set_Last_Error( nil ) ;
    if( ACM = Streams[ 0 ] ) then // Data stream
    begin
        C := ACM.Get_Clustersize ;
        if( Size = 0 ) then
        begin
            H := -1 ;
        end else
        begin
            H := ( Size - 1 ) div C ; // Cluster index of last cluster of new size
        end ;
        Size := ( H + 1 ) * C ;
        if( Size = Header.Size ) then // No change
        begin
            exit ;
        end ;
        Requested_Size := Size ;
        if( Contiguous ) then // Changing the size of a contiguous store
        begin
            if( Header.Size = 0 ) then // First allocation
            begin
                P := _Store.Allocate( Size ) ;
            end else
            if( Size = 0 ) then
            begin
                _Store.Deallocate( Header.Clusters[ 0 ], Header.Size ) ;
                P := 0 ;
            end else
            begin
                P := _Store.Reallocate( Header.Clusters[ 0 ], Header.Size, Size ) ;
            end ;
            if( ( P = 0 ) and ( Size > 0 ) ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_Error_Out_Of_Space, _Store.Last_Error ) ) ;
                exit ;
            end ;
            Header.Clusters[ 0 ] := P ;
            if( Header.EOF > Requested_Size ) then
            begin
                Header.EOF := Requested_Size ;
            end ;
            Header.Size := Size ;
            Update_Header ;
            exit ;
        end ; // if( Contiguous )

        // Allocate clusters in the header (if expanding)...
        if( H > Max_Header_Cluster ) then
        begin
            H := Max_Header_Cluster ;
        end ;
        for I := 0 to H do
        begin
            if( Header.Clusters[ I ] = 0 ) then
            begin
                Header.Clusters[ I ] := _Store.Allocate( C ) ;
                if( Header.Clusters[ I ] = 0 ) then
                begin
                    Set_Last_Error( Create_Exception( UOS_File_Error_Out_Of_Space, _Store.Last_Error ) ) ; { Untested }
                    exit ;
                end ;
                Dirty := True ;
            end ; // if( Header.Clusters[ I ] = 0 )
        end ; // for I := 0 to H
        if( Dirty ) then
        begin
            Update_Header ;
        end ;

        // Deallocate clusters in the header (if truncating the file)...
        if( H < Max_Header_Cluster ) then // New size is smaller than what the header maps
        begin
            Old_Header := Header ;
            for I := H + 1 to Max_Header_Cluster do
            begin
                Header.Clusters[ I ] := 0 ;
            end ; // for I := H + 1 to Max_Header_Cluster
            Update_Header ;
            for I := H + 1 to Max_Header_Cluster do
            begin
                if( Old_Header.Clusters[ I ] <> 0 ) then
                begin
                    _Store.Deallocate( Old_Header.Clusters[ I ], C ) ;
                end ; // if( Old_Header.Clusters[ I ] = 0 )
            end ;
        end ; // if( H < Max_Header_Cluster )
        Size := Size - ( Max_Header_Cluster + 1 ) * C ;
        if( Size < 0 ) then
        begin
            Size := 0 ;
        end ;
    end ; // if( ACM = Streams[ 0 ] )

    ACM.Set_Size( Size ) ;
    if( ACM = Streams[ 0 ] ) then // Data stream
    begin
        Header.Size := 0 ;
        for I := 0 to Max_Header_Cluster do
        begin
            if( Header.Clusters[ I ] <> 0 ) then
            begin
                Header.Size := Header.Size + C ;
            end ;
        end ;
        Header.Size := Header.Size + ACM.Get_Size ;
        if( Header.EOF > Requested_Size ) then
        begin
            Header.EOF := Requested_Size ; // Make sure logical EOF isn't past physical EOF
        end ;
        Update_Header ;
    end ; // if( ACM = Streams[ 0 ] )
    if( ACM.Index <= Max_Header_Data_Stream ) then
    begin
        if( Header.Streams[ ACM.Index ].Pointer <> ACM.Get_Root ) then
        begin
            Header.Streams[ ACM.Index ].Pointer := ACM.Get_Root ;
            Update_Header ;
        end ;
    end else
    begin
        //~~~ Update stream header (which exists in streams pointed to by Data_Stream)
    end ;
end ; // TUOS_Native_File.Set_Size


function TUOS_Native_File.Read_Only : boolean ;

begin
    Result := ( ( Header.Flags and FAF_READONLY ) <> 0 ) or Store.Read_Only ;
end ;


function TUOS_Native_File.Write_Only : boolean ;

begin
    Result := Store.Write_Only
end ;


function TUOS_Native_File.XSpaceAvail : int64 ;

begin
    Result := Store.XSpaceAvail ;
end ;


function TUOS_Native_File._Read( ACM : TNative_File_ACM ;
    Position : TStore_Address64 ; Length : TStore_Size64 ;
    Buff : pointer ) : TStore_Size64 ;

var Clustersize : TStore_Size64 ;
    Offset, P : TStore_Address64 ;
    This_Length : longint ;

begin
    // Setup...
    Result := 0 ;
    if( ( ACM = nil ) or ( Length < 1 ) ) then
    begin
        exit ;
    end ;
    if( Position >= Get_Size( ACM ) ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Read_Past_End, nil ) ) ;
        exit ;
    end ;
    if( Position + Length - 1 >= Get_Size( ACM ) ) then
    begin
        Length := Get_Size( ACM ) - Position ;
    end ;
    Clustersize := ACM.Get_Clustersize ;
    if(
        ( Clustersize < _Store.Min_Storage ) // Too small
        or
        ( Clustersize > _Store.Max_Storage ) // Too large
        or
        ( ( Clustersize mod _Store.Min_Storage ) <> 0 ) // Not an exact multiple of store clustersize
      ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Data_Structure_Error, nil ) ) ; { Untested }
        exit ;
    end ;
    if( ( Header.Flags and FAF_CONTIGUOUS ) <> 0 ) then
    begin
        P := Header.Clusters[ 0 ] + Position ;
        P := P and ( Clustersize - 1 ) ;
    end else
    begin
        P := Offset_To_Pointer( ACM, Position ) ;
        if( P = 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Data_Structure_Error, nil ) ) ; { Untested }
            exit ;
        end ;
    end ;
    Result := Length ;

    // Convert from position to offset from start of cluster
    Offset := Position - ( Position div Clustersize ) * Clustersize ;
    while( Length > 0 ) do
    begin
        This_Length := Length ;
        if( This_Length > Clustersize - Offset ) then
        begin
            This_Length := Clustersize - Offset ;
        end ;  //todo:if offset=0 and this_length>=clustersize then read directly to Buff rather than intermediate Buffer buffer
        _Store.Read( P, Clustersize, Buffer^ ) ;
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Read_Failure, _Store.Last_Error ) ) ; { Untested }
            exit ;
        end ;
        move( Buffer[ Offset ], Buff^, This_Length ) ;
        Length := Length - This_Length ;
        if( Length = 0 ) then
        begin
            break ;
        end ;
        Offset := 0 ; // Start at beginning of next cluster
        Position := Position + This_Length ;
        if( ( Header.Flags and FAF_CONTIGUOUS ) <> 0 ) then
        begin
            P := P + Clustersize ;
        end else
        begin
            P := Offset_To_Pointer( ACM, Position ) ;
            if( P = 0 ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_Error_Data_Structure_Error, nil ) ) ; { Untested }
                exit ;
            end ;
        end ;
        Buff := PAnsichar( Buff ) + This_Length ; // Move forward in destination buffer
    end ; // while( Length > 0 )
end ; // TUOS_Native_File._Read


function TUOS_Native_File._Write( ACM : TNative_File_ACM ;
    Position : TStore_Address64 ; Length : TStore_Size64 ;
    Buff : pointer ) : TStore_Size64 ;

var Clustersize : TStore_Size64 ;
    DS : TData_Stream ;
    New_Size : TStore_Size64 ;
    Offset, Old, P : TStore_Address64 ;
    This_Length : longint ;

begin
    // Setup...
    Result := 0 ;
    Set_Last_Error( nil ) ;
    if( ( ACM = nil ) or ( Length < 1 ) ) then
    begin
        exit ;
    end ;
    Old := ACM.Get_Root ;
    if( Position < 0 ) then
    begin
        Position := 0 ;
    end ;
    New_Size := Position + Length ;
    if( Position + Length > Get_Size( ACM ) ) then
    begin
        Set_Size( ACM, Position + Length ) ;
        if( Position + Length > Get_Size( ACM ) ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
            exit ;
        end ;
    end ;
    P := Offset_To_Pointer( ACM, Position ) ;
    if( P = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Data_Structure_Error, nil ) ) ; { Untested }
        exit ;
    end ;
    Result := Length ;

    // Convert from position to offset from start of cluster
    Clustersize := ACM.Get_Clustersize ;
    Offset := Position - ( Position div Clustersize ) * Clustersize ;
    while( Length > 0 ) do
    begin
        This_Length := Length ;
        if( This_Length > Clustersize - Offset ) then
        begin
            This_Length := Clustersize - Offset ;
        end ;
        if( ( Offset > 0 ) or ( This_Length < Clustersize ) ) then // Not a whole cluster
        begin
            _Store.Read( P, Clustersize, Buffer^ ) ; // Read in cluster for modification
            if( _Store.Last_Error <> nil ) then
            begin
                Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
                exit ;
            end ;
        end ;
        move( Buff^, Buffer[ Offset ], This_Length ) ;
        _Store.Write( P, Clustersize, Buffer^ ) ;
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
            exit ;
        end ;
        Length := Length - This_Length ;
        if( Length = 0 ) then
        begin
            break ;
        end ;
        Offset := 0 ; // Start at beginning of next cluster
        Position := Position + This_Length ;
        P := Offset_To_Pointer( ACM, Position ) ;
        if( P = 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Data_Structure_Error, nil ) ) ;
            exit ;
        end ;
        Buff := PAnsiChar( Buff ) + This_Length ; // Move forward in destination buffer
    end ; // while( Length > 0 )
    if( Old <> ACM.Get_Root ) then
    begin
        if( ACM.Index = -1 ) then
        begin
            Header.Data_Stream := ACM.Get_Root ;
            Update_Header ;
        end else
        if( ACM.Index <= Max_Header_Data_Stream ) then
        begin
            Header.Streams[ ACM.Index ].Pointer := ACM.Get_Root ;
            Update_Header ;
        end else
        begin
            DS.Pointer := ACM.Get_Root ;
            _Write( Data_Stream_ACM, ( ACM.Index - Max_Header_Data_Stream - 1 ) * sizeof( TData_Stream ) + sizeof( DS.Name ), sizeof( DS.Pointer ), @DS.Pointer ) ;
        end ;
    end ;
    if( ACM.Index = 0 ) then
    begin
        if( New_Size > Header.EOF ) then
        begin
            Header.EOF := New_Size ;
            Update_Header ;
        end ;
    end ; // if( ACM.Index = 0 )
end ; // TUOS_Native_File._Write


// API...

function TUOS_Native_File.Is_Class( Name : PChar ) : boolean ;

begin
    Result := lowercase( Name ) = 'tuos_native_file' ;
end ;


function TUOS_Native_File.Get_Store : TCOM_Managed_Store64 ;

begin
    Result := _Store ;
end ;


procedure TUOS_Native_File.Set_Store( Value : TCOM_Managed_Store64 ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Store <> nil ) then
    begin
        _Store.Detach ;
    end ;
    _Store := Value ;
end ;


function TUOS_Native_File.Get_Dirty : boolean ;

begin
    Result := _Dirty ;
end ;


procedure TUOS_Native_File.Set_Dirty( Value : boolean ) ;

begin
    _Dirty := Value ;
end ;


function TUOS_Native_File.Get_Directory : TUOS_Native_File ;

begin
    Result := _Directory ;
end ;


procedure TUOS_Native_File.Set_Directory( Value : TUOS_Native_File ) ;

begin
    if( Value <> nil ) then
    begin
        Value.Attach ;
    end ;
    if( _Directory <> nil ) then
    begin
        _Directory.Detach ;
    end ;
    _Directory := Value ;
end ;


function TUOS_Native_File.Get_Directory_Cluster : int64 ;

begin
    Result := _Directory_Cluster ;
end ;


procedure TUOS_Native_File.Set_Directory_Cluster( Value : int64 ) ;

begin
    _Directory_Cluster := Value ;
end ;


function TUOS_Native_File.Get_Directory_Index : cardinal ;

begin
    Result := _Directory_Index ;
end ;


procedure TUOS_Native_File.Set_Directory_Index( Value : cardinal ) ;

begin
    _Directory_Index := Value ;
end ;


function TUOS_Native_File.Buffer : PChar ;

var I : int64 ;

begin
    if( _Buffer = nil ) then
    begin
        I := Data_Stream_ACM.Get_Clustersize ;
        if( I < sizeof( TData_Stream ) * 4 ) then
        begin
            I := sizeof( TData_Stream ) * 4 ;
        end ;
        if( I < Header.Clustersize ) then
        begin
            I := Header.Clustersize ;
        end ;
        if( I < Store.Min_Storage ) then
        begin
            I := Store.Min_Storage ;
        end ;
{$IFDEF CC_Track_Resources}
        Remove_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        Reallocmem( _Buffer, I ) ;
{$IFDEF CC_Track_Resources}
        Add_Resource( RT_Memory, integer( _Buffer ) ) ;
{$ENDIF}
        _Buffer_Size := I ;
    end ;
    Result := _Buffer ;
end ;


function TUOS_Native_File.Streams : TList ;

begin
    if( _Streams = nil ) then
    begin
        _Streams := TList.Create ;
    end ;
    Result := _Streams ;
end ;


function TUOS_Native_File.Create_Stream( Name : int64 ) : longint ;

var DS : TData_Stream ;
    Loop : integer ;
    S, P : TStore_Address64 ;

begin
    Result := -1 ;
    if( Name = 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Duplicate_Name, nil ) ) ;
        exit ;
    end ;
    if( Indexof( Name ) <> -1 ) then // Already exists
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Duplicate_Name, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    for Loop := 1 to Max_Header_Data_Stream do
    begin
        if( Header.Streams[ Loop ].Name = 0 ) then
        begin
            Header.Streams[ Loop ].Name := Name ;
            Update_Header ;
            Result := Loop ;
            if( Result > _Max_Stream ) then
            begin
                _Max_Stream := Result ;
            end ;
            exit ;
        end ;
    end ;

    // If we get here, we have no room in the file header for a new stream...
    Result := Max_Header_Data_Stream + 1 ;
    DS.Name := Name ;
    DS.Pointer := 0 ;
    if( Data_Stream_ACM.Get_Size = 0 ) then // First data stream after header
    begin
        fillchar( Buffer^, _Buffer_Size, 0 ) ;
        move( DS, Buffer^, sizeof( DS ) ) ;
        _Write( Data_Stream_ACM, 0, Data_Stream_ACM.Clustersize, Buffer ) ;
        _Max_Stream := Max_Header_Data_Stream + 1 ;
        Header.Data_Stream := Data_Stream_ACM.Get_Root ;
        Update_Header ;
        exit ;
    end ;

    P := 0 ;
    while( true ) do
    begin
        fillchar( Buffer^, _Buffer_Size, 0 ) ;
        S  := _Read( Data_Stream_ACM, P, Data_Stream_ACM.Clustersize, Buffer ) ;
        if( S = 0 ) then // Need to write at end of data stream
        begin
            Set_Last_Error( nil ) ;
            move( DS, Buffer^, sizeof( DS ) ) ;
            _Write( Data_Stream_ACM, P, Data_Stream_ACM.Clustersize, Buffer ) ;
            if( Result > _Max_Stream ) then
            begin
                _Max_Stream := Result ;
            end ;
            exit ;
        end ;
        for Loop := 0 to S div sizeof( TData_Stream ) - 1 do
        begin
            if( PData_Stream_Array( Buffer )[ Loop ].Name = 0 ) then
            begin
                Result := Result + Loop ;
                if( _Max_Stream < Result ) then
                begin
                    _Max_Stream := Result ;
                end ;
                PData_Stream_Array( Buffer )[ Loop ].Name := Name ;
                PData_Stream_Array( Buffer )[ Loop ].Pointer := 0 ;
                _Write( Data_Stream_ACM, P, _Buffer_Size, Buffer ) ;
                if( Result > _Max_Stream ) then
                begin
                    _Max_Stream := Result ;
                end ;
                exit ;
            end ;
        end ;
        Result := Result + S div sizeof( TData_Stream ) ;
        P := P + S ;
    end ; // while( true )
end ; // TUOS_Native_File.Create_Stream


function TUOS_Native_File.Max_Stream : longint ;

var Loop : integer ;

begin
    if( _Max_Stream < 0 ) then // Need to figure out the max stream
    begin
        if( Header.Data_Stream = 0 ) then {else Untested}
        begin
            _Max_Stream := 0 ;
            for Loop := Max_Header_Data_Stream downto 1 do
            begin
                if( Header.Streams[ Loop ].Name <> 0 ) then
                begin
                    _Max_Stream := Loop ; {Untested}
                    break ;
                end ;
            end ;
            Result := _Max_Stream ;
            exit ;
        end ;
        _Max_Stream := Max_Header_Data_Stream + Data_Stream_ACM.Get_Size div sizeof( TData_Stream ) ;
    end ;

    Result := _Max_Stream ;
end ;


function TUOS_Native_File.Stream_Name( Index : longint ) : int64 ;

var ACM : TNative_File_ACM ;

begin
    ACM := Find_Data_Stream( Index ) ;
    if( ACM = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := ACM.Name ;
    end ;
end ;


function TUOS_Native_File.Stream_Pointer( Index : longint ) : TStore_Address64 ;

var ACM : TNative_File_ACM ;

begin
    ACM := Find_Data_Stream( Index ) ;
    if( ACM = nil ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := ACM.Get_Root ;
    end ;
end ;


// I/O...

function TUOS_Native_File.Read( Stream : longint ; Position : TStore_Address64 ;
    Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;

var ACM : TNative_File_ACM ;

begin
    Result := 0 ;
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Read_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;

    Result := _Read( ACM, Position, Length, @Buff ) ;
end ;


function TUOS_Native_File.Write( Stream : longint ; Position : TStore_Address64 ;
    Length : TStore_Size64 ; var Buff ) : TStore_Size64 ;

var ACM : TNative_File_ACM ;

begin
    Result := 0 ;
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;

    Result := _Write( ACM, Position, Length, @Buff ) ;
end ; // TUOS_Native_File.Write


function TUOS_Native_File.Get_Stream_Size( Stream : longint ) : TStore_Size64 ;

var ACM : TNative_File_ACM ;

begin
    Result := 0 ;
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    if( Stream = 0 ) then
    begin
        Result := Header.Size ;
        exit ;
    end ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        exit ;
    end ;
    Result := Get_Size( ACM ) ;
    if( ACM.Last_Error <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_IO_Error, _Store.Last_Error ) ) ; { Untested }
    end ;
end ;


procedure TUOS_Native_File.Set_Stream_Size( Stream : longint ; Value : TStore_Size64 ) ;

var ACM : TNative_File_ACM ;

begin
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;
    Set_Size( ACM, Value ) ;
    if( ACM.Last_Error <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_IO_Error, ACM.Last_Error ) ) ; { Untested }
    end ;
end ; // TUOS_Native_File.Set_Stream_Size


function TUOS_Native_File.Get_Allocation_Cluster( Stream, Index : longint ) : TStore_Address64 ;

var ACM : TNative_File_ACM ;

begin
    Result := -1 ;
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    if( Stream = 0 ) then
    begin
        if( Contiguous ) then
        begin
            Result := Header.Clusters[ 0 ] ;
            exit ;
        end ;
        if( Index <= Max_Header_Cluster ) then
        begin
            Result := Header.Clusters[ Index ] ;
            exit ;
        end ;
        Index := Index - Max_Header_Cluster - 1 ;
    end ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;
    Result := ACM.Get_Allocation_Cluster( Index ) ;
    if( ACM.Last_Error <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_IO_Error, _Store.Last_Error ) ) ; { Untested }
    end ;
end ; // TUOS_Native_File.Get_Allocation_Cluster


procedure TUOS_Native_File.Set_Allocation_Cluster( Stream, Index : longint ;
    Value : TStore_Address64 ) ;

var ACM : TNative_File_ACM ;

begin
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;
    if( Stream = 0 ) then
    begin
        if( Contiguous ) then
        begin
            Header.Clusters[ 0 ] := Value ;
            exit ;
        end ;
        if( Index <= Max_Header_Cluster ) then
        begin
            Header.Clusters[ Index ] := Value ;
            Update_Header ;
            exit ;
        end ;
        Index := Index - Max_Header_Cluster - 1 ;
    end ;
    ACM.Set_Allocation_Cluster( Index, Value ) ;
    if( ACM.Last_Error <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_IO_Error, _Store.Last_Error ) ) ; { Untested }
    end ;
end ; // TUOS_Native_File.Set_Allocation_Cluster


function TUOS_Native_File.Get_ACM_Cluster( Stream : longint ; P : int64 ) : int64 ;

var ACM : TNative_File_ACM ;

begin
    Result := 0 ; // Assume failure
    if( Stream < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    Set_Last_Error( nil ) ;
    ACM := Find_Data_Stream( Stream ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;
    Result := ACM.Next_ACM_Cluster( P ) ;
end ; // TUOS_Native_File.Get_ACM_Cluster


function TUOS_Native_File.Rebuild : boolean ;

    function Valid_Pointer( P : int64 ) : boolean ;

    begin
        Result := True ; // Assume it is good
        if( P = 0 ) then // Nil is always valid
        begin
            exit ;
        end ;
        if(
            ( P < Store.Min_Storage ) // Too small
            or
            ( P > Store.Max_Storage )
            or
            ( ( P mod Store.Min_Storage ) <> 0 )
          ) then
        begin
            Result := False ;
        end ;
    end ;


    procedure Set_Corrupted ;

    begin
        Result := True ; // For now, just return true
    end ;


var ACM : TCOM_Allocation_Cluster_Manager64 ;
    I, Stream : longint ;
    Cluster, P : int64 ;

begin
    Result := False ;
    for Stream := 0 to Max_Stream do
    begin
        // Allocate the clusters for the stream data...
        if( ( Stream = 0 ) and ( ( Header.Flags and FAF_Contiguous ) <> 0 ) ) then
        begin
            Cluster := 0 ;
            P := Get_Allocation_Cluster( Stream, Cluster ) ;
            if( not Valid_Pointer( P ) ) then
            begin
                Set_Corrupted ;
            end ;
            // If contiguous, allocate the whole file
            if( not Store.Allocate_At( P, Header.Size ) ) then
            begin
                // Cluster(s) already allocated.  Something has corrupted the file structure
                Result := True ;
            end ;
        end else
        begin
            if( Stream = 0 ) then // Data stream
            begin
                // Allocate header clusters...
                for I := 0 to Max_Header_Cluster do
                begin
                    P := Header.Clusters[ I ] ;
                    if( P <> 0 ) then
                    begin
                        if( not Valid_Pointer( P ) ) then
                        begin
                            Set_Corrupted ;
                        end ;
                        if( not Store.Allocate_At( P, Header.Clustersize ) ) then
                        begin
                            Result := True ;
                        end ;
                    end ;
                end ;
            end ; // if( Stream = 0 )

            // Rebuild ACM...
            ACM := Find_Data_Stream( Stream ) ;
            if( ACM <> nil ) then
            begin
                Result := Result or ACM.Rebuild ;
            end ;
        end ;
    end ; //  for Stream := 0 to F.Max_Stream
end ; // TUOS_Native_File.Rebuild


function TUOS_Native_File.Get_Contiguous : boolean ;

begin
    Result := _Contiguous ;
end ;


procedure TUOS_Native_File.Set_Contiguous( Value : boolean ) ;

var ACM : TNative_File_ACM ;
    Index : longint ;
    C, S : TStore_Size64 ;
    New, P : TStore_Address64 ;

begin
    if( Header.Size = 0 ) then
    begin
        exit ;
    end ;
    if( _Contiguous <> Value ) then
    begin
        // Setup...
        ACM := Find_Data_Stream( 0 ) ;
        S := Header.Size ;
        C := ACM.Get_Clustersize ;
        Index := 0 ;

        // Convert...
        if( Value ) then
        begin
            // Convert to contiguous...
            P := Store.Allocate( S ) ;
            if( P = 0 ) then
            begin
                exit ; { Untested }
            end ;

            // Copy the current data to the new contiguous area...
            New := P ;
            while( S > 0 ) do
            begin
                S := S - C ;
                Store.Copy( Get_Allocation_Cluster( 0, Index ), P, C ) ;
                if( Store.Last_Error <> nil ) then
                begin
                    Set_Last_Error( Create_Exception( UOS_File_Error_Read_Failure, nil ) ) ;
                    exit ;
                end ;
                P := P + C ;
                inc( Index ) ;
            end ;

            // Deallocate the old data and assign pointer to new data...
            _Contiguous := Value ;
            Set_Size( ACM, Header.Size ) ;
            Header.Clusters[ 0 ] := New ;
            for Index := 1 to Max_Header_Cluster do
            begin
                Header.Clusters[ Index ] := 0 ;
            end ;
            Header.Flags := Header.Flags or FAF_CONTIGUOUS ;
        end else
        begin
            // Convert from contiguous...
            P := Header.Clusters[ 0 ] ; // First offset of contiguous area
            _Contiguous := Value ;
            ACM.Prepare_Size( S - Max_Header_Cluster * C ) ; // Make room for required number of cluster pointers
            while( S > 0 ) do
            begin
                S := S - C ;
                Set_Allocation_Cluster( 0, Index, P ) ;
                P := P + C ;
                inc( Index ) ;
            end ;
            Header.Flags := Header.Flags and ( not FAF_CONTIGUOUS ) ;
        end ;
        Update_Header ;
        if( Value ) then
        begin
            ACM.Set_Size( 0 ) ;
        end ;
    end ; // if( _Contiguous <> Value )
end ; // TUOS_Native_File.Set_Contiguous


function TUOS_Native_File.Get_File_Size : int64 ;

begin
    Result := Header.Size ;
end ;


procedure TUOS_Native_File.Set_File_Size( Value : int64 ) ;

begin
    Set_Stream_Size( 0, Value ) ;
end ;


procedure TUOS_Native_File.Delete_Stream( Name : int64 ; Index : longint ) ;

var ACM : TNative_File_ACM ;
    DS : TData_Stream ;

begin
    // Sanity checks...
    if( Index < 0 ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Stream_Index, nil ) ) ;
        exit ;
    end ;
    if( ( Name = 0 ) and ( Index = 0 ) ) then // Cannot delete file data via Delete_Stream
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_Invalid_Operation, nil ) ) ;
        exit ;
    end ;
    if( Name > 0 ) then
    begin
        Index := IndexOf( Name ) ;
        if( Index < 0 ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
            exit ;
        end ;
    end ;

    // Get ACM...
    ACM := Find_Data_Stream( Index ) ;
    if( ACM = nil ) then
    begin
        if( _Store.Last_Error <> nil ) then
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_Write_Failure, _Store.Last_Error ) ) ; { Untested }
        end else
        begin
            Set_Last_Error( Create_Exception( UOS_File_Error_No_Such_Stream, nil ) ) ;
        end ;
        exit ;
    end ;

    // Remove stream from file...
    if( Index <= Max_Header_Data_Stream ) then
    begin
        Header.Streams[ Index ].Name := 0 ;
        Header.Streams[ Index ].Pointer := 0 ;
        Update_Header ;
    end else
    begin
        DS.Name := 0 ;
        DS.Pointer := 0 ;
        _Write( Data_Stream_ACM, ( Index - Max_Header_Data_Stream - 1 ) * sizeof( DS ), sizeof( DS ), @DS ) ;
        if( Last_Error <> nil ) then
        begin
            exit ; { Untested }
        end ;
    end ;

    // Delete in-memory pointer to ACM...
    _Streams[ Index ] := nil ;

    Set_Size( ACM, 0 ) ;
    if( ACM.Last_Error <> nil ) then
    begin
        Set_Last_Error( Create_Exception( UOS_File_Error_IO_Error, _Store.Last_Error ) ) ; { Untested }
    end ;
end ; // TUOS_Native_File.Delete_Stream


{$IFDEF CC_Regression_Test}
procedure Test_Unit( T : TTest_Interface ) ;

var Store : TRAM_Disk64 ; // We have to use this rather than TCOM_Heap64 otherwise we can't deal with contiguous files (due to Delphi heap prefixes on allocated memory)
    Test : TUOS_Native_File ;

    procedure Error ;

    var S, T : integer ;

    begin
        if( Test.Last_Error <> nil ) then
        begin
            ShowMessage( Test.Last_Error.Error_Text( S, T ) ) ;
        end ;
        halt ;
    end ;


    procedure Verify( Stream : longint ; Name : string ) ;

    var Buffer, Buffer1 : array[ 0..512 ] of char ; // 513 bytes
        Count : cardinal ;
        F : file ;
        P : int64 ;

    begin
        // Verify...
        assignfile( F, 'e:\work\reference\' + Name ) ;
        reset( F, 1 ) ;
        P := 0 ;
        while( not eof( F ) ) do
        begin
            fillchar( Buffer, sizeof( Buffer ), 0 ) ;
            blockread( F, Buffer, sizeof( Buffer ), Count ) ;
            if( Count > 0 ) then
            begin
                if( Test.Read( Stream, P, sizeof( Buffer ), Buffer1 ) <> sizeof( Buffer ) ) then
                begin
                    Error ;
                end ;
                dec( Count ) ;
                while( Count < 4294967295 ) do
                begin
                    if( Buffer[ Count ] <> Buffer1[ Count ] ) then
                    begin
                        Error ;
                    end ;
                    dec( Count ) ;
                end ;
            end ;
            P := P + sizeof( Buffer ) ;
        end ; // while( not eof( F ) )

        closefile( F ) ;
    end ; // Test_Unit.Verify


    procedure Write( Stream : longint ; Name : string ; Check : boolean = True ) ;

    var Buffer : array[ 0..512 ] of char ; // 513 bytes
        Count : cardinal ;
        F : file ;
        P : int64 ;
        S, T : integer ;

    begin
        // Write...
        assignfile( F, 'e:\work\reference\' + Name ) ;
        reset( F, 1 ) ;
        P := 0 ;
        while( not eof( F ) ) do
        begin
            fillchar( Buffer, sizeof( Buffer ), 0 ) ;
            blockread( F, Buffer, sizeof( Buffer ), Count ) ;
            if( Test.Write( Stream, P, sizeof( Buffer ), Buffer ) <> sizeof( Buffer ) ) then
            begin
                if( Check and ( Test.Last_Error <> nil ) ) then
                begin
                    ShowMessage( Test.Last_Error.Error_Text( S, T ) ) ;
                    Error ;
                end ;
                exit ;
            end ;
            P := P + sizeof( Buffer ) ;
        end ;
        closefile( F ) ;

        Verify( Stream, Name ) ;
    end ;


var FL : text ;

    function Log( const S : string ) : boolean ;

    begin
        append( FL ) ;
        writeln( FL, S ) ;
        closefile( FL ) ;
        Result := T.Log_Operation( S ) ;
    end ;

var Buffer : array[ 0..512 ] of char ;

    procedure Validate( Loop : integer ) ;

    var S : integer ;

    begin
        S := sizeof( Buffer ) ;
        if( S > Test.Get_Stream_Size( Loop ) ) then
        begin
            S := Test.Get_Stream_Size( Loop ) ;
        end ;
        if( Test.Read( Loop, 0, S, Buffer ) <> S ) then
        begin
            Test.Read( Loop, 0, S, Buffer ) ;
            Error ;
        end ;
        if( Test.Last_Error <> nil ) then
        begin
            Error ;
        end ;
        S := S - 1 ;
        while( S >= 0 ) do
        begin
            if( Buffer[ S ] <> chr( Loop ) ) then
            begin
                Error ;
            end ;
            dec( S ) ;
        end ;
    end ;


var Loop : integer ;
    MStore : TUOS_Managed_Store ;
    S : int64 ;
    Sizes, Streams : TInteger_List ;
    UEC : TUnified_Exception ;

begin
(*
    Test := TUOS_Native_File.Create ;
    Test.Free ;
    Test := TUOS_Native_File.Create ;
    Store := TRAM_Disk64.Create ;
    Store.Set_Max_Storage( 5 * 1024 * 1024, UEC ) ; // 5 Mb
    Store.Attach ;
    MStore := TUOS_Managed_Store.Create ;
    MStore.Store := Store ;

    // General tests...
    if( Test.Get_Store <> nil ) then
    begin
        Error ;
    end ;
    Test.Set_Store( MStore ) ;
    if( Test.Get_Store <> MStore ) then
    begin
        Error ;
    end ;
    Test.Set_Store( MStore ) ;
    if( Test.Max_Stream <> 0 ) then
    begin
        Error ;
    end ;
    Test.Get_Stream_Size( 1 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 0 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Pointer( 0 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Pointer( 1 ) <> 0 ) then
    begin
        Error ;
    end ;

    // Data I/O...
    Write( 0, 'timeline phones numbers.txt' ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Write( 1, 'timeline phones numbers.txt', False ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 16 ) <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Max_Stream <> 1 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 1 ) <> 16 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Pointer( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 16 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 16 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Get_Stream_Size( 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    Write( 1, 'records.txt' ) ;
    Verify( 0, 'timeline phones numbers.txt' ) ;

    if( Test.Create_Stream( 2 ) <> 2 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 3 ) <> 3 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 4 ) <> 4 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 5 ) <> 5 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 6 ) <> 6 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Write( 2, 'quotes.txt' ) ;
    Write( 3, 'plants.txt' ) ;
    Write( 4, 'names.txt' ) ;
    Write( 5, 'phone.lis' ) ;
    Write( 6, 'games.txt' ) ;
    if( Test.Max_Stream < 6 ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 7 ) <> 7 ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( 5, 0 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 9 ) <> 5 ) then
    begin
        Error ;
    end ;
    if( Test.Max_Stream < 7 ) then
    begin
        Error ;
    end ;
    if( Test.Read( 11, 0, 1, Buffer ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Read( -1, 0, 1, Buffer ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Write( -1, 0, 1, Buffer ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Set_Stream_Size( 5, 16384 ) ;
    if( Test.Get_Stream_Size( 5 ) <> 16384 ) then
    begin
        exit ;
    end ;
    if( Test.Indexof( 6 ) <> 6 ) then
    begin
        Error ;
    end ;
    if( Test.Indexof( 8 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test.Create_Stream( 0 ) <> -1 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    if( Test.Get_Stream_Size( 0 ) = 0 ) then
    begin
        Error ;
    end ;
    Test.Set_Stream_Size( 11, 1 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( -1, -1 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( 0, 0 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( 11, 0 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( 0, 7 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 0 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 6 ) <> 6 ) then
    begin
        Error ;
    end ;
    if( Test.Stream_Name( 8 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test._Write( nil, 0, 0, @Buffer ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test._Read( nil, 0, 0, @Buffer ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Get_Stream_Size( -1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    Test.Set_Stream_Size( -1, 10 ) ;
    if( Test.Last_Error = nil ) then
    begin
        Error ;
    end ;
    for Loop := 20 to 64 do
    begin
        if( Test.Create_Stream( Loop ) <= 0 ) then
        begin
            Error ;
        end ;
        if( Test.Last_Error <> nil ) then
        begin
            Error ;
        end ;
    end ;
    Test.Set_Stream_Size( 0, 16 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Test.Set_Stream_Size( 0, 0 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    Test.Delete_Stream( 0, 2 ) ;
    if( Test.Last_Error <> nil ) then
    begin
        Error ;
    end ;
    if( Test.Find_Data_Stream( 2 ) <> nil ) then
    begin
        Error ;
    end ;

    Test.Free ;

    Test := TUOS_Native_File.Create ;
    Test.Set_Store( MStore ) ;
    Test.Header.Clustersize := 1024 ;
    Test.Buffer ;

    Test.Free ;

    Test := TUOS_Native_File.Create ;
    Test.Set_Store( MStore ) ;
    Test.Contiguous := True ;
    Test.Contiguous := False ;
    Test.Set_Stream_Size( 0, 512 ) ;
    Test.Contiguous := True ;
    Test.Set_Stream_Size( 0, 1024 ) ;
    Test.Contiguous := False ;
    Test.Set_Stream_Size( 0, 0 ) ;
    Write( 0, 'timeline phones numbers.txt' ) ;
    Test.Contiguous := True ;
    Test.Contiguous := True ;
    Test.Set_Stream_Size( 0, 0 ) ;
    Write( 0, 'timeline phones numbers.txt' ) ;
    Test.Free ;

    MStore.Free ;
*)

    // Stress test...
//    randomize ;
    Test := TUOS_Native_File.Create ;
    Store := TRAM_Disk64.Create ;
    Store.Set_Max_Storage( 655360 * 1024, UEC ) ; // 671 Mb
    Store.Attach ;
    MStore := TUOS_Managed_Store.Create ;
    MStore.Store := Store ;
    Test.Set_Store( MStore ) ;
    Streams := TInteger_List.Create ;
    Sizes := TInteger_List.Create ;
    for Loop := 0 to 99 do
    begin
        Streams.Add( -1 ) ; // Stream index for name (ie Streams[ Name ] = file data stream index)
        Sizes.Add( 0 ) ; // Size for stream (matches up against Streams[] )
    end ;
    Streams[ 0 ] := 0 ;
    assignfile( FL, 'k:\test.txt' ) ;
    rewrite( FL ) ;
    closefile( FL ) ;
    while( true ) do
    begin
        Loop := random( 100 ) ;
        case random( 10 ) of
            0 : // Create stream
                begin
                    if( Loop = 0 ) then // Name 0 is not used
                    begin
                        continue ;
                    end ;
                    if( Streams[ Loop ] < 0 ) then // Not already created
                    begin
                        if( Log( 'C' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        S := Test.Create_Stream( Loop ) ;
                        if( S <= 0 ) then
                        begin
                            Test.Create_Stream( Loop ) ;
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                        Streams[ Loop ] := S ;

                        if( Test.Get_Stream_Size( S ) <> Sizes[ Loop ] ) then
                        begin
                            Test.Get_Stream_Size( S ) ;
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Test.Get_Stream_Size( S ) ;
                            Error ;
                        end ;
                    end ;
                end ;
            1 : // Delete stream
                begin
                    if( Loop = 0 ) then // Name 0 is not used
                    begin
                        continue ;
                    end ;
                    if( Streams[ Loop ] > 0 ) then
                    begin
                        if( Log( 'D' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        Test.Delete_Stream( 0, Streams[ Loop ] ) ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Test.Delete_Stream( 0, Streams[ Loop ] ) ;
                            Error ;
                        end ;
                        Streams[ Loop ] := -1 ;
                        Sizes[ Loop ] := 0 ;
                    end ;
                end ;
            2 : // Read stream
                begin
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        if( Log( 'R' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        S := sizeof( Buffer ) ;
                        if( S > Test.Get_Stream_Size( Streams[ Loop ] ) ) then
                        begin
                            S := Test.Get_Stream_Size( Streams[ Loop ] ) ;
                        end ;
                        if( Test.Read( Streams[ Loop ], 0, S, Buffer ) <> S ) then
                        begin
                            Test.Read( Streams[ Loop ], 0, S, Buffer ) ;
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                        S := S - 1 ;
                        while( S >= 0 ) do
                        begin
                            if( Buffer[ S ] <> chr( Loop ) ) then
                            begin
                                Error ;
                            end ;
                            dec( S ) ;
                        end ;
                    end ;
                end ;
            3 : // Write stream
                begin
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        if( Log( 'W' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        fillchar( Buffer, sizeof( Buffer ), Loop ) ;
                        if( Test.Write( Streams[ Loop ], 0, sizeof( Buffer ), Buffer ) <> sizeof( Buffer ) ) then
                        begin
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                        S := sizeof( Buffer ) ;
                        if( S > Test.Get_Stream_Size( Streams[ Loop ] ) ) then
                        begin
                            Test.Get_Stream_Size( Streams[ Loop ] ) ;
                            Error ;
                        end ;
                        Sizes[ Loop ] := Test.Get_Stream_Size( Streams[ Loop ] ) ;
                    end ;
                end ;
            4 : // Max_Stream
                begin
                    if( Log( 'M' ) ) then
                    begin
                        exit ;
                    end ;
                    Test.Max_Stream ;
                    if( Test.Last_Error <> nil ) then
                    begin
                        Error ;
                    end ;
                end ;
            5 : // Stream_Name
                begin
                    if( Loop = 0 ) then
                    begin
                        continue ;
                    end ;
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        if( Log( 'N' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        if( Test.Stream_Name( Streams[ Loop ] ) <> Loop ) then
                        begin
                            Test.Stream_Name( Streams[ Loop ] ) ;
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                    end ;
                end ;
            6 : // Stream_Pointer
                begin
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        if( Log( 'P' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        Test.Stream_Pointer( Streams[ Loop ] ) ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                    end ;
                end ;
            7 : // Get_Stream_Size
                begin
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        if( Log( 'G' + inttostr( Loop ) ) ) then
                        begin
                            exit ;
                        end ;
                        if( Test.Get_Stream_Size( Streams[ Loop ] ) <> Sizes[ Loop ] ) then
                        begin
                            Test.Get_Stream_Size( Streams[ Loop ] ) ;
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Test.Get_Stream_Size( Streams[ Loop ] ) ;
                            Error ;
                        end ;
                    end ;
                end ;
            8 : // Set_Stream_Size
                begin
                    if( Streams[ Loop ] >= 0 ) then
                    begin
                        S := random( 256 ) * 64 ;
                        if( Log( 'S' + inttostr( Loop ) + ' ' + inttostr( S ) ) ) then
                        begin
                            exit ;
                        end ;
                        Test.Set_Stream_Size( Streams[ Loop ], S ) ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Test.Set_Stream_Size( Streams[ Loop ], S ) ;
                            Error ;
                        end ;
                        Sizes[ Loop ] := S ;
                        if( Sizes[ Loop ] > Test.Get_Stream_Size( Streams[ Loop ] ) ) then
                        begin
                            Error ;
                        end ;
                        fillchar( Buffer, sizeof( Buffer ), Loop ) ;
                        if( Test.Write( Streams[ Loop ], 0, sizeof( Buffer ), Buffer ) <> sizeof( Buffer ) ) then
                        begin
                            Error ;
                        end ;
                        if( Test.Last_Error <> nil ) then
                        begin
                            Error ;
                        end ;
                        S := sizeof( Buffer ) ;
                        if( S > Test.Get_Stream_Size( Streams[ Loop ] ) ) then
                        begin
                            Test.Get_Stream_Size( Streams[ Loop ] ) ;
                            Error ;
                        end ;
                        Sizes[ Loop ] := Test.Get_Stream_Size( Streams[ Loop ] ) ;
                    end ;
                end ;
            9 : // Change continuity
                begin
                    if( Log( 'X' ) ) then
                    begin
                        exit ;
                    end ;
                    Validate( 0 ) ;
                    Test.Contiguous := not Test.Contiguous ;
                    Validate( 0 ) ;
                end ;
        end ; // case random( 9 )
    end ; // while( true )
    Test.Free ;
    MStore.Detach ;
end ;
{$ENDIF}

end.
