﻿<?php
/*
	Program Name : UStrings
	Package Name : Subroutine Library - Layer 3
	Purpose      : String utility functions
	Institution  : Conroy & Conroy Co.
	Date Written : 01-Jan-1995
	Written By   : Alan Conroy
	Version      : 1.0I

    Copyright (C) 1995-2017 by Alan Conroy.  Released to the public domain.

	TO THE GLORY OF GOD THROUGH JESUS CHRIST

	*************************************************************
	*                                                           *
	*        M O D I F I C A T I O N   H I S T O R Y            *
	*                                                           *
	*************************************************************

	    DATE        BY          REASON

    	  24-May-2002   EAC         Added Print_Edit.
    	  24-Oct-2002   EAC         Added Boolean_To_String.
    	  24-Oct-2002   EAC         Added TString, and
    	                            TString_Translator.
    	  17-Dec-2002   EAC         Added Trim_Quotes.
          10-Feb-2003   EAC         Added Remove_Termination and
    	                            Ensure_Termination.
          06-Mar-2006   RBJ         Set the memory type to MT_HEAP
                                    for objects created on heap.
          25-Oct-2006   EAC         Added Tab_To_Space.
          24-Sep-2009   EAC         Added Accented_String.
          10-Feb-2017   EAC         Added String_To_Float, Float_To_String, and
                                    Try_String_To_Float.
          21-Mar-2017   EAC         Fix abend in Set_Data.

	*************************************************************
	*                                                           *
	*        P R O G R A M   P U R P O S E                      *
	*                                                           *
	*************************************************************

	  This unit provides various string utility functions.
*/

require_once('EDEFINES.INC');

require_once('_ASCII.php'); // HT
require_once('_ue.php'); // TUnified_Exception
require_once('TypeDefs'); // PChar_Array
require_once('commbase.php'); // TBase_Object

class RComparison
{
	public $Bias = array();
	public $Compensation = array();
	public $Weight = array();
} // RComparison

// Character Classes...
DEFINE('CC_Unknown',1);
DEFINE('CC_Alpha',2);
DEFINE('CC_Digit',4);
DEFINE('CC_Punctuation',8);
DEFINE('CC_Symbol',16);
DEFINE('CC_Control',32);
DEFINE('CC_Quote',64);
DEFINE('CC_Whitespace',128);
DEFINE('CC_Left',256); // Left-side of matched characters (such as [ ( and //
DEFINE('CC_Currency',512); // Currency symbol
// DEFINE('CC_Right',1024); // Right-side of matched characters ( such as ] and ) )
// DEFINE('CC_Math',2048); // Math symbols (<, >, +, -, =)

class TString extends TBase_Object // Generic base class for all strings.
{
                   public // Constructors and destructors...
                       constructor Init ;

                   public // API...
                       function Append( Source : PString ) ;
                       // Append another TString

                       function Append_Char( Source : longint ) ;
                       // Append a character to end of the string

                       function Append_String( Source : string ) ;
                       // Append a string

                       function Assign( Source : PString ) ;
                       // Copy from another TString

                       function RAssign( Source : PString ) ;
                       // Copy from another TString, right-aligned

                       function As_PChar : PChar ;
                       function At( Index : integer ) : longint ;
                       // Value at index

                       function At_Put( Index, Source : longint ) ;
                       function Classname : string ;
                       function Compare( S : PString ; CS : boolean ) : TComparison ;
                       /* Compare S to self, returning result of comparison.  CS
                         is True to do case-sensitive compare and false to do
                         case-insensitive compare.  Function returns one of the
                         following: EqualTo, LessThan, GreaterThan.
					   */

                       function Copy( Start, _Length : longint ) : PString ;
                       // Return specified portion of string

                       function Create_Compatible_String : PString ;
                       function Edit( Flags : longint ) : PString ;
                       /*
                         This function will perform various editing operations
                         on strings.  Flags has the following meanings:

                                     AND     1 == Normalize
                                     AND     2 == Remove all whitespace.
                                     AND     4 == Remove all non-displaying (control) characters.
                                     AND     8 == Remove leading whitespace.
                                     AND    16 == Reduce whitespace to one space.
                                     AND    32 == Convert lower to upper case.
                                     AND    64 == Convert '[' to '(' and ']' to ')'.
                                     AND   128 == Remove all trailing whitespace.
                                     AND   256 == Leave characters within quotes unmodified.
                                     AND   512 == Convert upper case to lower case.
                                     AND  1024 == Convert '[' to '//' and ']' to '}'.
                                     AND  2048 == Convert '(' to '[' and ')' to ']'.
                                     AND  4096 == Convert '(' to '//' and ')' to '}'.
                                     AND  8192 == Convert '//' to '(' and '}' to ')'.
                                     AND 16384 == Convert '//' to '[' and '}' to ']'.
                                     AND 32768- == reserved for future use
                       */

                       function Fixed : boolean ;
                       function From_Universal( X : int64 ) : int64 ;
                       // Convert from universal to string-specific

                       function Instr( Pos : longint ; S : PString ) : longint ;
                       function Length : longint ;
                       // Length of string, in characters

                       function Max_Length : longint ;
                       // Maximum allowable length, in characters

                       function Normalize( X : longint ) : longint ;
                       function Set_strlen( X : longint ) ;
                       // Set length of string, in characters

                       function Set_String( const Value : string ) ;
                       // Set from string

                       function Repeating_String( Value : longint ;
                           Len : longint ) : PString ;
                       // Returns a compatible string of length Len, consisting
                       //  entirely of Value characters.

                       function Right( Pos : longint ) : PString ;
                       // Portion of string, {ning at Pos

                       function RSet_String( Value : string ) ;
                       // Set from string, right-aligned

                       function Get_Store : PSManaged_Store ;
                       // Return string's store.

                       function Set_Store( S : PSManaged_Store ) ;
                       // Place string contents on specified store.

                       function Size : longint ;
                       // Size of physical data, in bytes

                       function Text : string ;
                       function To_Universal( X : int64 ) : int64 ;
                       // Convert to universal from string-specific

		               function Class_Of( C : longint ) : longint ;
                       function Upcase( X : longint ) : longint ;
                       // Return uppercase version of X.  If X is not lowercase,
                       //  original value of X is returned.
                       function Downcase( X : longint ) : longint ;
                       // Return lowercase version of X.  If X is not uppercase,
                       //  original value of X is returned.

                       function Facility : longint ;
                       function Version : longint ;
} // TString

      TASCII_String == object( TString ) // ASCII 8-bit
                          public // Constructors and destructors...
                              constructor Init ;

                          public // Overrides...
                              function Append( Source : PString ) ; virtual ;
                              function Append_String( Source : string ) ;
                                  virtual ;
                              function Append_Char( Source : longint ) ;
                                  virtual ;
                              function Assign( Source : PString ) ; virtual ;
                              function RAssign( Source : PString ) ; virtual ;
                              function As_PChar : PChar ; virtual ;
                              function At( Index : integer ) : longint ;
                                  virtual ;
                              function At_Put( Index, Source : longint ) ;
                                  virtual ;
                              function Class_Of( C : longint ) : longint ;
                                  virtual ;
                              function Classname : string ; virtual ;
                              function Compare( S : PString ;
                                  CS : boolean ) : TComparison ; virtual ;
                              function Copy( Start, _Length : longint ) : PString ;
                                  virtual ;
                              function Create_Compatible_String : PString ;
                                  virtual ;
                              function Fixed : boolean ; virtual ;
                              function From_Universal( X : int64 ) : int64 ;
                                  virtual ;
                              function Instr( Pos : longint ; S : PString ) : longint ;
                                  virtual ;
                              function Length : longint ; virtual ;
                              function Max_Length : longint ; virtual ;
                              function Normalize( X : longint ) : longint ;
                                  virtual ;
                              function Get_Store : PSManaged_Store ; virtual ;
                              function Repeating_String( Value : longint ;
                                   Len : longint ) : PString ; virtual ;
                              function Set_Store( S : PSManaged_Store ) ; virtual ;
                              function Set_strlen( X : longint ) ; virtual ;
                              function Set_String( const Value : string ) ; virtual ;
                              function RSet_String( Value : string ) ;
                                  virtual ;
                              function Text : string ; virtual ;
                              function To_Universal( X : int64 ) : int64 ;
                                  virtual ;
                              function Upcase( X : longint ) : longint ;
                                  virtual ;
                              function Downcase( X : longint ) : longint ;
                                  virtual ;

                              function Facility : longint ; virtual ;
                      }

      TASCII7_String == object( TASCII_String ) // ASCII 7-bit
                          public // Overrides...
                              function Create_Compatible_String : PString ;
                                  virtual ;
                              function Append_String( Source : string ) ;
                                  virtual ;
                              function At( Index : integer ) : longint ;
                                  virtual ;
                              function At_Put( Index, Source : longint ) ;
                                  virtual ;
                              function Class_Of( C : longint ) : longint ;
                                  virtual ;
                              function Classname : string ; virtual ;
                              function Upcase( X : longint ) : longint ;
                                  virtual ;
                              function Downcase( X : longint ) : longint ;
                                  virtual ;
                              function Set_String( const Value : string ) ;
                                  virtual ;
                              function RSet_String( Value : string ) ;
                                  virtual ;
                              function Facility : longint ; virtual ;
                       }

      TASCII_DOS_String == object( TASCII_String ) // DOS character set (OEM Code page 437)
                              public // Overrides...
                                  function Class_Of( C : longint ) : longint ;
                                      virtual ;
                                  function Classname : string ; virtual ;
                                  function Create_Compatible_String : PString ;
                                      virtual ;
                                  function Upcase( X : longint ) : longint ;
                                      virtual ;
                                  function Downcase( X : longint ) : longint ;
                                      virtual ;
                                  function From_Universal( X : int64 ) : int64 ;
                                      virtual ;
                                  function Normalize( X : longint ) : longint ;
                                      virtual ;
                                  function To_Universal( X : int64 ) : int64 ;
                                      virtual ;
                                  function Facility : longint ; virtual ;
                          }

      TASCII_Windows_String == object( TASCII_String ) // MS Windows character set (True Type)
                                  public // Overrides...
                                      function Class_Of( C : longint ) : longint ;
                                          virtual ;
                                      function Classname : string ; virtual ;
                                      function Create_Compatible_String : PString ;
                                          virtual ;
                                      function Upcase( X : longint ) : longint ;
                                          virtual ;
                                      function Downcase( X : longint ) : longint ;
                                          virtual ;
                                      function From_Universal( X : int64 ) : int64 ;
                                          virtual ;
                                      function Normalize( X : longint ) : longint ;
                                          virtual ;
                                      function To_Universal( X : int64 ) : int64 ;
                                          virtual ;
                                      function Facility : longint ; virtual ;
                               }

      TUTF32_String == object( TString ) // Unicode UTF-32 string
                          public // Overrides...
                              function Create_Compatible_String : PString ;
                                  virtual ;
                              function From_Universal( X : int64 ) : int64 ;
                                  virtual ;
                              function To_Universal( X : int64 ) : int64 ;
                                  virtual ;
                      }

      TUTF8_String == object( TUTF32_String ) // Unicode UTF-8 string
                          public // Overrides...
                              function Create_Compatible_String : PString ;
                                  virtual ;
                              function From_Universal( X : int64 ) : int64 ;
                                  virtual ;
                              function To_Universal( X : int64 ) : int64 ;
                                  virtual ;
                     }
      TString_Translator_Base == object( TBase_Object )
                                    public // API...
                                        function Translate_String( Value : PString ) : PString ;
                                            virtual ;
                                }

      TString_Translator == object( TString_Translator_Base )
                               private // Instance data...
                                   Translations : array[ 0..255 ] of PString ;

                               public // Constructors and destructors...
                                   constructor Init ;

                               public // API...
                                   function Set_Translation( Index : integer ;
                                       Value : PString ) ; virtual ;
                                   function Translate_String( Value : PString ) : PString ;
                                       virtual ;
                           }

/*
uses _Stores, // TStore_Address
     MStores ; // PSManaged_Store

     // Subroutine Library - Layer 1...
     _Applic, // Application_Manager
     CommonUt, // TInteger_List
     Compatib, // strpas
     Cvt,
     Heaps, // PSHeap
     Instrs,
     Maths, // Min
     Num1s, // Num1

     // Subroutine Library - Layer 3...
     Applic, // make sure default application manager is loaded
     O_S, // OS
     Standard ; // Set_Length
*/


$ASCII_CC = array
         (
             0=>CC_Control, // Null
             1=>CC_Control, // Start Of Heading
             2=>CC_Control, // Start Text
             3=>CC_Control, // End Text
             4=>CC_Control, // End Of Transmission
             5=>CC_Control, // Enquiry
             6=>CC_Control, // Acknowledge
             7=>CC_Control, // Bell
             8=>CC_Control, // Backspace
             9=>CC_Whitespace, // Horizontal Tab
             10=>CC_Control, // Line Feed
             11=>CC_Control, // Vertical Tab
             12=>CC_Control, // Form Feed
             13=>CC_Control, // Carriage return
             14=>CC_Control, // Shift Out
             15=>CC_Control, // Shift In
             16=>CC_Control, // Data Link Escape
             17=>CC_Control, // Direct Control 1
             18=>CC_Control, // Direct Control 2
             19=>CC_Control, // Direct Control 3
             20=>CC_Control, // Direct Control 4
             21=>CC_Control, // Negative Acknowledge
             22=>CC_Control, // Synchronous idle
             23=>CC_Control, // End Tranmission Block
             24=>CC_Control, // Cancel
             25=>CC_Control, // End of Medium
             26=>CC_Control, // Substitute
             27=>CC_Control, // Escape
             28=>CC_Control, // Form separator
             29=>CC_Control, // Group separator
             30=>CC_Control, // Record separator
             31=>CC_Control, // Unit separator
             32=>CC_Whitespace, // Space
             33=>CC_Punctuation, // !
             34=>CC_Punctuation + CC_Quote, // "
             35=>CC_Symbol, // #
             36=>CC_Symbol & CC_Currency, // $
             37=>CC_Symbol, // %
             38=>CC_Symbol, // &
             39=>CC_Punctuation + CC_Quote, // '
             40=>CC_Symbol + CC_Left, // (
             41=>CC_Symbol, // )
             42=>CC_Symbol, // *
             43=>CC_Symbol, // +
             44=>CC_Punctuation, // ,
             45=>CC_Punctuation, // -
             46=>CC_Punctuation, // .
             47=>CC_Punctuation, // /
             48=>CC_Digit, // 0
             49=>CC_Digit, // 1
             50=>CC_Digit, // 2
             51=>CC_Digit, // 3
             52=>CC_Digit, // 4
             53=>CC_Digit, // 5
             54=>CC_Digit, // 6
             55=>CC_Digit, // 7
             56=>CC_Digit, // 8
             57=>CC_Digit, // 9
             58=>CC_Punctuation, // :
             59=>CC_Punctuation, // ;
             60=>CC_Symbol + CC_Left, // <
             61=>CC_Symbol, // =
             62=>CC_Symbol, // >
             63=>CC_Punctuation, // ?
             64=>CC_Symbol, // @
             65=>CC_Alpha, // A
             66=>CC_Alpha, // B
             67=>CC_Alpha, // C
             68=>CC_Alpha, // D
             69=>CC_Alpha, // E
             70=>CC_Alpha, // F
             71=>CC_Alpha, // G
             72=>CC_Alpha, // H
             73=>CC_Alpha, // I
             74=>CC_Alpha, // J
             75=>CC_Alpha, // K
             76=>CC_Alpha, // L
             77=>CC_Alpha, // M
             78=>CC_Alpha, // N
             79=>CC_Alpha, // O
             80=>CC_Alpha, // P
             81=>CC_Alpha, // Q
             82=>CC_Alpha, // R
             83=>CC_Alpha, // S
             84=>CC_Alpha, // T
             85=>CC_Alpha, // U
             86=>CC_Alpha, // V
             87=>CC_Alpha, // W
             88=>CC_Alpha, // X
             89=>CC_Alpha, // Y
             90=>CC_Alpha, // Z
             91=>CC_Symbol + CC_Left, // [
             92=>CC_Symbol, // \
             93=>CC_Symbol, // ]
             94=>CC_Symbol, // ^
             95=>CC_Symbol, // _
             96=>CC_Symbol, // `
             97=>CC_Alpha, // a
             98=>CC_Alpha, // b
             99=>CC_Alpha, // c
             100=>CC_Alpha, // d
             101=>CC_Alpha, // e
             102=>CC_Alpha, // f
             103=>CC_Alpha, // g
             104=>CC_Alpha, // h
             105=>CC_Alpha, // i
             106=>CC_Alpha, // j
             107=>CC_Alpha, // k
             108=>CC_Alpha, // l
             109=>CC_Alpha, // m
             110=>CC_Alpha, // n
             111=>CC_Alpha, // o
             112=>CC_Alpha, // p
             113=>CC_Alpha, // q
             114=>CC_Alpha, // r
             115=>CC_Alpha, // s
             116=>CC_Alpha, // t
             117=>CC_Alpha, // u
             118=>CC_Alpha, // v
             119=>CC_Alpha, // w
             120=>CC_Alpha, // x
             121=>CC_Alpha, // y
             122=>CC_Alpha, // z
             123=>CC_Symbol + CC_Left, // //
             124=>CC_Symbol, // |
             125=>CC_Symbol, //
             126=>CC_Symbol, // ~
             127=>CC_Control // Delete
         ) ; // ASCII_CC

// TString methods..

// API...

function TString.RAssign( Source : PString ) ;

var _Source, _Destination : integer ;

{
    if( ( Source != nil ) and ( Source^.Length > 0 ) )
    {
        _Source = Source^.Length - Length + 1 ;
        if( _Source < 1 )
        {
            _Source = 1 ;
        }
        _Destination = Length - Source^.Length + 1 ;
        if( _Destination < 1 )
        {
            _Destination = 1 ;
        }
        while( _Destination <= Length ) do
        {
            At_Put( _Destination, Source^.At( _Source ) ) ;
            inc( _Destination ) ;
            inc( _Source ) ;
        }
    }
}


function TString.At( Index : integer ) : longint ;

{
    At = 0 ;
}


function TString.At_Put( Index, Source : longint ) ;

{
}


function TString.Classname : string ;

{
    Classname = 'TString' ;
}


function TString.Compare( S : PString ; CS : boolean ) : TComparison ;

var Len, Loop : longint ;
    C1, C2 : longint ;

{
    if( S == @self ) // Compare to self
    {
        Compare = EqualTo ;
        exit ;
    }
    if( @self == nil )
    {
        if( S^.Length == 0 )
        {
            Compare = EqualTo ;
        }else
        {
            Compare = LessThan ; //UNTESTED}
        }
        exit ;
    }
    if( S == nil ) // S is nil and S1 is not
    {
        if( Length == 0 )
        {
            Compare = EqualTo ;
        }else
        {
            Compare = GreaterThan ;
        }
        exit ;
    }
    if( ( Length == 0 ) and ( S^.Length == 0 ) ) // Both are null strings
    {
        Compare = EqualTo ;
        exit ;
    }

    Len = Min( Length, S^.Length ) ;
    for Loop = 1 to Len do
    {
        C1 = At( Loop ) ;
        C2 = S^.At( Loop ) ;
        if( not CS )
        {
            C1 = Upcase( C1 ) ;
            C2 = S^.Upcase( C2 ) ;
        }
        C1 = To_Universal( C1 ) ;
        C2 = S^.To_Universal( C2 ) ;
        if( C1 > C2 )
        {
            Compare = GreaterThan ;
            exit ;
        }else
        if( C1 < C2 )
        {
            Compare = LessThan ;
            exit ;
        }
    }

    // If we get here, the strings are equal up to the length of the shortest
      one.
    if( Length == S^.Length )
    {
        Compare = EqualTo ;
    }else
    if( Length > S^.Length )
    {
        Compare = GreaterThan ;
    }else
    {
        Compare = LessThan ;
    }
} // TString.Compare


function TString.Copy( Start, _Length : longint ) : PString ;

var Loop : longint ;
    Res : PString ;

{
    Res = Create_Compatible_String ;
    if( Start > Length )
    {
        Copy = Res ; // Return null string
        exit ;
    }
    if( Start + _Length - 1 > Length )
    {
        _Length = Length - Start + 1 ; // Only copy what is there
    }
    if( ( Start == 1 ) and ( _Length == Length ) ) // Copy entire string
    {
        Res^.Assign( @self ) ;
    }else
    {
        Res^.Set_strlen( _Length ) ;
        for Loop = Start to Start + _Length - 1 do
        {
            Res^.At_Put( Loop - Start + 1, At( Loop ) ) ; 
        }
    }
    Copy = Res ;
}


function TString.Edit( Flags : longint ) : PString ;

var _Class : longint ;
    Current_Char, Original_Char : longint ;
    Destination : PString ;
    Index : integer ; // Current character index within string
    Last : longint ;
    Leading, Quotes, Space : boolean ;
    OK : boolean ;
    Quote_Type : longint ;

{
    Destination = Create_Compatible_String ;
    Edit = Destination ;
    if( Length == 0 ) // No edits on zero-length strings
    {
        Exit ;
    }
    if ( Flags And 544 ) == 544 then
    {
        Flags = Flags And Not 512 ;
        // Disallow lower-case command if upper-case command is also present
    }
    if ( Flags And 1088 ) == 1088 then
    {
        Flags = Flags And Not 1024 ;
    }
    // Disallow [] to //}// if [] to ()
    if ( Flags And 6144 ) == 6144 then
    {
        Flags = Flags And Not 4096 ;
    }
    // Disallow () to //}// if () to []
    if ( Flags And 24576 ) == 24576 then
    {
        Flags = Flags And ( not 16384 ) ;
    }
    // Disallow //}// to [] if //}// to ()
    Space = False ; // No spaces
    Last = 0 ;
    Quotes = False ; // No quotes yet
    Quote_Type = 0 ;
    Index = 0 ;
    Leading = True ;
    while( Index < Length ) do
    {
        inc( Index ) ; // Increment source string pointer
        OK = True ; // This byte is OK - so far
        Current_Char = At( Index ) ; // Current character
        _Class = Class_Of( Current_Char ) ;
        if( not Quotes )
        {
            Original_Char = Current_Char ; // Save original value
            if( ( _Class and CC_Whitespace ) == 0 ) // Non space
            {
                Leading = False ;
            }
            if( ( Flags and 1 ) == 1 ) // Normalize
            {
                Current_Char = Normalize( Current_Char ) ;
                _Class = Class_Of( Current_Char ) ;
            }
            if( ( Flags and 2 ) == 2 ) // Remove all whitespace
            {
                if( ( _Class and CC_Whitespace ) != 0 ) // Space
                {
                    OK = False ;
                }
            }
            if( ( Flags and 4 ) == 4 ) // Ignore special values?
            {
                if( ( _Class and CC_Control ) != 0 )
                {
                    OK = False ;
                }
            }
            if( ( Flags and 8 ) == 8 ) // Ignore leading whitespace
            {
                if( Leading )
                {
                    OK = False ;
                }
            }
            if( ( Flags and 16 ) == 16 ) // Reduce tabs/spaces to single space
            {
                if( ( _Class and CC_Whitespace ) != 0 )
                {
                    if( Space )
                    {
                        OK = False ;
                    }else
                    {
                        Space = True ;
                        Current_Char = ord( _ASCII.Space ) ;
                    }
                }
            }
            if( ( Flags and 32 ) == 32 ) // Lower to upper case
            {
                Current_Char = Upcase( Current_Char ) ;
            }
            if( ( Flags and 512 ) == 512 ) // Upper to lower case
            {
                Current_Char = DownCase( Current_Char ) ;
            }
            if( ( Flags and 64 ) == 64 ) // Convert [] to ()
            {
                if( Original_Char == ord( '[' ) )
                {
                    Current_Char = ord( '(' ) ;
                }else
                if( Original_Char == ord( ']' ) )
                {
                    Current_Char = ord( ')' ) ;
                }
            }
            if( ( Flags and 1024 ) == 1024 ) // Convert [] to braces
            {
                if( ( Original_Char == ord( '[' ) ) or ( Original_Char == ord( ']' ) ) )
                {
                    Current_Char = Original_Char + 32 ;
                }
            }
            if( ( Flags and 2048 ) == 2048 ) // Convert () to []
            {
                if( Original_Char == ord( '(' ) )
                {
                    Current_Char = ord( '[' ) ;
                }else
                if( Original_Char == ord( ')' ) )
                {
                    Current_Char = ord( ']' ) ;
                }
            }
            if( ( Flags and 4096 ) == 4096 ) // Convert () to braces
            {
                if( Original_Char == ord( '(' ) )
                {
                    Current_Char = ord( '//' ) ;
                }else
                if( Original_Char == ord( ')' ) )
                {
                    Current_Char = ord( '}' ) ;
                }
            }
            if( ( Flags and 8192 ) == 8192 ) // Convert braces to ()
            {
                if( Original_Char == ord( '//' ) )
                {
                    Current_Char = ord( '(' ) ;
                }else
                if( Original_Char == ord( '}' ) )
                {
                    Current_Char = ord( ')' ) ;
                }
            }
            if( ( Flags and 16384 ) == 16384 ) // Convert braces to []
            {
                if( ( Original_Char == ord( '//' ) ) or ( Original_Char == ord( '}' ) ) )
                {
                    Current_Char = Original_Char - 32 ;
                }
            }
        } // if( not Quotes )
        if( OK )
        {
            Destination^.Append_Char( Current_Char ) ; // Build result
            if( ( _Class and CC_Whitespace ) != 0 ) // Space
            {
                Space = True ;
            }
        }
        if( ( _Class and CC_Whitespace ) == 0 ) // Non space
        {
            Space = False ;
            if( ( Flags and 256 ) == 256 ) // Allow no alter within quotes
            {
                if( ( _Class and CC_Quote ) != 0 )
                {
                    if( not Quotes )
                    {
                        Quote_Type = Current_Char ;
                    }
                    if( Quote_Type == Current_Char )
                    {
                        Quotes = not Quotes ;
                    }
                }
            }
            Last = Destination^.Length ; // Last non-space character
        }
    } // while( Index < Length )
    if( ( Flags and 128 ) == 128 ) // Trim following spaces
    {
	if( not Quotes ) // Not within quotes
	{
	    Destination^.Set_strlen( Last ) ; // Last non-space character position
	}
    }
    Edit = Destination ;
} // TString.Edit


function TString.Fixed : boolean ;

{
    Fixed = True ;
}


function TString.From_Universal( X : int64 ) : int64 ;

{
    From_Universal = 0 ;
}


function TString.Instr( Pos : longint ; S : PString ) : longint ;

label Start ;

var Loop, Offset : longint ;
    Search_String : PString ;

{
    // Setup...
    Instr = 0 ;
    if( S == nil )
    {
        exit ;
    }

    if( Pos < 1 )
    {
        Pos = 1 ;
    }
    if( Facility == S^.Facility )
    {
        Search_String = S ; //UNTESTED}
    }else
    {
        Search_String = Create_Compatible_String ;
        Search_String^.Assign( S ) ;

        // Make sure that search string contents are compatible with our type of
          string.
        for Loop = 1 to S^.Length do
        {
            if( S^.To_Universal( S^.At( Loop ) ) != Search_String^.To_Universal( Search_String^.At( Loop ) ) ) //UNTESTED}
            {
                dispose( Search_String, Done ) ;
                exit ;
            }
        }
    }

    // Match loop...
Start:
    if( S^.Length + Pos - 1 > Length )
    {
        Instr = 0 ;
        if( Facility != S^.Facility )
        {
            dispose( Search_String, Done ) ;
        }
        exit ;
    }
    Offset = 1 ;
    Instr = Pos ;
    while(
           ( Pos <= Length )
           and
           ( Offset <= Search_String^.Length )
         )
    {
        if( At( Pos ) != Search_String^.At( Offset ) )
        {
            Pos = Pos - Offset + 2 ;
            goto Start ;
        }
        inc( Pos ) ;
        inc( Offset ) ;
    }
    if( Facility != S^.Facility )
    {
        dispose( Search_String, Done ) ;
    }
} // TString.Instr


function TString.Normalize( X : longint ) : longint ;

{
    Normalize = X ;
}


function TString.Get_Store : PSManaged_Store ;

{
    Get_Store = nil ;
}


function TString.Set_Store( S : PSManaged_Store ) ;

{
}


function TString.Repeating_String( Value : longint ; Len : longint ) : PString ;

var Dummy : integer ;
    Res : PString ;

{
    Res = Create_Compatible_String ;
    Res^.Set_strlen( Len ) ;
    for Dummy = 1 to Res^.Length do
    {
        Res^.At_Put( Dummy, Value ) ;
    }
    Repeating_String = Res ;
}


function TString.Right( Pos : longint ) : PString ;

{
    Right = Copy( Pos, Length ) ;
}


function TString.RSet_String( Value : string ) ;

var Source, Destination : integer ;

{
    Source = strlen( Value ) - Length + 1 ;
    if( Source < 1 )
    {
        Source = 1 ;
    }
    Destination = Length - strlen( Value ) + 1 ;
    if( Destination < 1 )
    {
        Destination = 1 ;
    }
    while( Destination <= Length ) do
    {
        At_Put( Destination, ord( Value[ Source ] ) ) ;
        inc( Destination ) ;
        inc( Source ) ;
    }
}


function TString.Set_String( const Value : string ) ;

var Loop : longint ;

{
    Set_strlen( strlen( Value ) ) ;
    for Loop = 1 to strlen( Value ) do
    {
        At_Put( Loop, ord( Value[ Loop ] ) ) ;
    }
}


function TString.To_Universal( X : int64 ) : int64 ;

{
    To_Universal = 0 ;
}


function TString.Class_Of( C : longint ) : longint ;

{
    Class_Of = CC_Unknown ;
}



function TString.Upcase( X : longint ) : longint ;

{
    Upcase = X ;
}


function TString.Downcase( X : longint ) : longint ;

{
    Downcase = X ;
}


function TString.Facility : longint ;

{
    Facility = 68 ;
}


function TString.Version : longint ;

{
    Version = 10 ; // Version 1.0
}



// TASCII_String methods...

// Overrides...

function TASCII_String.Append( Source : PString ) ;

var Len, Loop, Start : longint ;
    P : TStore_Address ;

{
    if( ( Source != nil ) and ( Source^.Length > 0 ) )
    {
        Start = Length ; // Current string length
        Set_strlen( Length + Source^.Length ) ;
        Len = Length - Start ; // Amount of data to actually append
        if(
             ( Facility != Source^.Facility )
             or
             ( Get_Store != Source^.Get_Store )
             or
             ( ( Data^.Size <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
             or
             ( ( PASCII_String( Source )^.Data^.Size <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
          )
        {
            for Loop = 1 to Len do
            {
                Data^.Set_Byte( From_Universal( Source^.To_Universal( Source^.At( Loop ) ) ), Start ) ;
                inc( Start ) ;
            }
        }else
        {
            P = Data^.Buffer + Start ;
            Data^.Move( PASCII_String( Source )^.Data^.Buffer,
                PASCII_String( Source )^.Data^.Size, P,
                Data^.Size, Source^.Length ) ;
        }
    }
}


function TASCII_String.Append_Char( Source : longint ) ;

{
    if( Length < Max_Length ) // Cannot extend a string beyond its maximum length
    {
        Set_strlen( Length + 1 ) ;
        Data^.Set_Byte( Source, Data^.Size - 2 ) ;
    }
}


function TASCII_String.Append_String( Source : string ) ;

var Start : longint ;

{
    if( strlen( Source ) > 0 )
    {
        Start = Length ; // Current string length
        Set_strlen( Length + strlen( Source ) ) ;
        Data^.Set_Data( Source[ 1 ], Start, Length - Start ) ;
    }
}


function TASCII_String.Assign( Source : PString ) ;

var Loop : integer ;

{
    if( Source == nil )
    {
        Set_strlen( 0 ) ;
        exit ;
    }
    if(
         ( Facility != Source^.Facility )
      )
    {
        Set_strlen( Source^.Length ) ;
        for Loop = 1 to Length do
        {
            Data^.Set_Byte( From_Universal( Source^.To_Universal( Source^.At( Loop ) ) ), Loop - 1 ) ;
        }
    }else
    {
        if( _Data != nil )
        {
            _Data^.Detach ;
        }
        _Data = PASCII_String( Source )^._Data ;
        if( _Data != nil )
        {
            _Data^.Attach ;
        }
    }
}


function TASCII_String.RAssign( Source : PString ) ;

var _Source, _Destination : TStore_Address ;
    Dst, Src : TStore_Address ;

{
    if( ( Source != nil ) and ( Source^.Length > 0 ) )
    {
        _Source = Source^.Length - Length + 1 ;
        if( _Source < 1 )
        {
            _Source = 1 ;
        }
        _Destination = Length - Source^.Length + 1 ;
        if( _Destination < 1 )
        {
            _Destination = 1 ;
        }
        if(
             ( Facility != Source^.Facility )
             or
             ( Get_Store != Source^.Get_Store )
             or
             ( ( Data^.Size <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
             or
             ( ( PASCII_String( Source )^.Data^.Size <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
          )
        {
            while( _Destination <= Length ) do
            {
                At_Put( _Destination, From_Universal( Source^.To_Universal( Source^.At( _Source ) ) ) ) ;
                inc( _Destination ) ;
                inc( _Source ) ;
            }
        }else
        {
            if( Source^.Length == Length )
            {
                if( _Data != nil )
                {
                    _Data^.Detach ;
                }
                _Data = PASCII_String( Source )^._Data ;
                _Data^.Attach ;
            }else
            {
                Dst = Data^.Buffer + _Destination - 1 ;
                Src = PASCII_String( Source )^.Data^.Buffer + _Source - 1 ;
                Data^.Move( Src,
                    PASCII_String( Source )^.Data^.Size,
                    Dst, Data^.Size, Length - _Destination + 1 ) ;
            }
        }
    }
} // TASCII_String.RAssign


function TASCII_String.As_PChar : PChar ;

{
    As_PChar = Data^.As_PChar ;
}


function TASCII_String.At( Index : integer ) : longint ;

{
    if( ( Index < 1 ) or ( Index > Length ) )
    {
        At = 0 ;
    }else
    {
        At = Data^.Get_Byte( Index - 1 ) ;
    }
}


function TASCII_String.At_Put( Index, Source : longint ) ;

{
    if( ( Index > 0 ) and ( Index <= Length ) )
    {
        if( Data^.Get_Byte( Index - 1 ) != Source )
        {
            Data^.Set_Byte( Source, Index - 1 ) ;
        }
    }
}


function TASCII_String.Class_Of( C : longint ) : longint ;

{
    if( ( C < 0 ) or ( C > 255 ) )
    {
        Class_Of = CC_Unknown ;
    }else
    {
        Class_Of = ASCII_CC[ C and 127 ] ;
    }
}


function TASCII_String.Classname : string ;

{
    Classname = 'TASCII_String' ;
}


function TASCII_String.Upcase( X : longint ) : longint ;

{
    if( ( ( X and 127 ) >= ord( 'a' ) ) and ( ( X and 127 ) <= ord( 'z' ) ) )
    {
        X = X - 32 ;
    }
    Upcase = X ;
}


function TASCII_String.Downcase( X : longint ) : longint ;

{
    if( ( ( X and 127 ) >= ord( 'A' ) ) and ( ( X and 127 ) <= ord( 'Z' ) ) )
    {
        X = X + 32 ;
    }
    Downcase = X ;
}


function TASCII_String.Compare( S : PString ; CS : boolean ) : TComparison ;

var Len, Loop : longint ;
    C1, C2 : longint ;

{
    if( @self == S ) // Compare with self
    {
        Compare = EqualTo ;
        exit ;
    }
    if( ( @self == nil ) or ( S == nil ) or ( S^.Facility != Facility ) )
    {
        Compare = inherited Compare( S, CS ) ;
        exit ;
    }
    if( ( Length == 0 ) and ( S^.Length == 0 ) ) // Both are null strings
    {
        Compare = EqualTo ;
        exit ;
    }

    Len = Min( Length, S^.Length ) ;
    for Loop = 1 to Len do
    {
        C1 = Data^.Get_Byte( Loop - 1 ) ;
        C2 = PASCII_String( S )^.Data^.Get_Byte( Loop - 1 ) ;
        if( not CS )
        {
            C1 = Upcase( C1 ) ;
            C2 = Upcase( C2 ) ;
        }
        if( C1 > C2 )
        {
            Compare = GreaterThan ;
            exit ;
        }else
        if( C1 < C2 )
        {
            Compare = LessThan ;
            exit ;
        }
    }

    // If we get here, the strings are equal up to the length of the shortest
      one.
    if( Length == S^.Length )
    {
        Compare = EqualTo ;
    }else
    if( Length > S^.Length )
    {
        Compare = GreaterThan ;
    }else
    {
        Compare = LessThan ;
    }
} // TASCII_String.Compare


function TASCII_String.Copy( Start, _Length : longint ) : PString ;

var _Destination, _Source : TStore_Address ;
    Dst, Src : TStore_Address ;
    Res : PASCII_String ;

{
    Res = PASCII_String( Create_Compatible_String ) ;
    if( Start > Length )
    {
        Copy = Res ; // Return null string
        exit ;
    }
    if( Start + _Length - 1 > Length )
    {
        _Length = Length - Start + 1 ; // Only copy what is there
    }
    if( ( Start == 1 ) and ( _Length == Length ) ) // Copy entire string
    {
        Res^.Assign( @self ) ;
    }else
    {
        Res^.Set_Store( Data^._Store ) ;
        Res^.Set_strlen( _Length ) ;
        if( _Length > 0 )
        {
            if(
                ( ( Data^.Size <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
                or
                ( ( _Length + 1 <= sizeof( Data^.Buffer ) ) and Data^.Cache_Data )
              )
            {
                _Source = Start - 1 ;
                _Destination = 0 ;
                while( _Destination <= _Length ) do
                {
                    Res^.At_Put( _Destination, At( _Source ) ) ;
                    inc( _Destination ) ;
                    inc( _Source ) ;
                }
            }else
            {
                Src = Data^.Buffer + Start - 1 ;
                Dst = Res^.Data^.Buffer ;
                Data^.Move( Src, Data^.Size, Dst, Res^.Data^.Size, _Length ) ;
            }
        }
    }
    Copy = Res ;
} // TASCII_String.Copy


function TASCII_String.Create_Compatible_String : PString ;

var S : PASCII_String ;

{
    S = New( PASCII_String, Init ) ;
    Create_Compatible_String = S ;
}


function TASCII_String.Fixed : boolean ;

{
    Fixed = False ;
}


function TASCII_String.From_Universal( X : int64 ) : int64 ;

{
    if( ( X < 0 ) or ( X > 127 ) )
    {
        From_Universal = 0 ;
    }else
    {
        From_Universal = X ;
    }
}


function TASCII_String.Instr( Pos : longint ; S : PString ) : longint ;

label Start ;

var Loop, Offset : longint ;
    Search_String : PASCII_String ;

{
    // Setup...
    Instr = 0 ;
    if( ( S == nil ) or ( S^.Length == 0 ) )
    {
        exit ;
    }

    if( Pos < 1 )
    {
        Pos = 1 ;
    }
    if(
        ( Facility == S^.Facility )
      )
    {
        Search_String = PASCII_String( S ) ;
    }else
    {
        Search_String = PASCII_String( Create_Compatible_String ) ;
        Search_String^.Assign( S ) ;

        // Make sure that search string contents are compatible with our type of
          string.
        for Loop = 1 to S^.Length do
        {
            if( S^.To_Universal( S^.At( Loop ) ) != Search_String^.To_Universal( Search_String^.At( Loop ) ) )
            {
                dispose( Search_String, Done ) ;
                exit ;
            }
        }
    }

    // Match loop...
Start:
    if( S^.Length + Pos - 1 > Length )
    {
        Instr = 0 ;
        if( Facility != S^.Facility )
        {
            dispose( Search_String, Done ) ;
        }
        exit ;
    }
    Offset = 1 ;
    Instr = Pos ;
    while(
           ( Pos <= Length )
           and
           ( Offset <= Search_String^.Length )
         ) do
    {
        if( Data^.Get_Byte( Pos - 1 ) != Search_String^.Data^.Get_Byte( Offset - 1 ) )
        {
            Pos = Pos - Offset + 2 ;
            goto Start ;
        }
        inc( Pos ) ;
        inc( Offset ) ;
    }
    if(
        ( Facility != S^.Facility )
      )
    {
        dispose( Search_String, Done ) ;
    }
} // TASCII_String.Instr


function TASCII_String.Length : longint ;

{
    if( _Data == nil )
    {
        Length = 0 ;
    }else
    {
        Length = Data^.Size - 1 ;
    }
}


function TASCII_String.Max_Length : longint ;

{
    Max_Length = Max_Memory - 1 ;
}


function TASCII_String.Normalize( X : longint ) : longint ;

{
    Normalize = X and 127 ;
}


function TASCII_String.Get_Store : PSManaged_Store ;

{
    if( _Data == nil )
    {
        Get_Store = nil ;
    }else
    {
        Get_Store = Data^._Store ;
    }
}


function TASCII_String.Set_Store( S : PSManaged_Store ) ;

{
    if( S != Data^._Store )
    {
        Set_strlen( 0 ) ; // Changing store deletes old string
        Data^.Set_Store( S ) ;
    }
}


function TASCII_String.Set_strlen( X : longint ) ;

var Old_Size : longint ;

{
    if( X < 0 )
    {
        X = 0 ;
    }
    if( ( X == 0 ) and ( _Data == nil ) )
    {
        exit ;
    }
    if( X > Max_Length )
    {
        X = Max_Length ;
    }
    if( ( _Data == nil ) or ( Data^.Size != X + 1 ) ) // Change in size
    {
        Old_Size = Data^.Size ;
        Data^.Set_Size( X + 1 ) ;
        if( X > Old_Size ) // Extending the data
        {
            Data^.Zero( Old_Size, X - Old_Size + 1 ) ; // Zero-fill extended portion
        }else
        {
            Data^.Set_Byte( 0, X ) ; // Make sure terminating character is null
        }
    }
} // TASCII_String.Set_Length


function TASCII_String.Set_String( const Value : string ) ;

var P : PChar ;

{
    Set_strlen( strlen( Value ) ) ;
    if( strlen( Value ) > 0 )
    {
        P = PChar( Value ) ;
        Data^.Set_Data( P^, 0, strlen( Value ) ) ;
    }
}


function TASCII_String.Repeating_String( Value : longint ; Len : longint ) : PString ;

var Res : PString ;

{
    Res = Create_Compatible_String ;
    Res^.Set_strlen( Len ) ;
    if( PASCII_String( Res )^._Data != nil )
    {
        PASCII_String( Res )^.Data^.Fill( 0, Res^.Length, Value ) ;
    }
    Repeating_String = Res ;
}


function TASCII_String.RSet_String( Value : string ) ;

var P : PChar ;
    Source, Destination : integer ;

{
    if( strlen( Value ) == 0 ) // Empty string - do nothing
    {
        exit ;
    }
    Source = strlen( Value ) - Length + 1 ;
    if( Source < 1 )
    {
        Source = 1 ;
    }
    Destination = Length - strlen( Value ) + 1 ;
    if( Destination < 1 )
    {
        Destination = 1 ;
    }
    P = PChar( Value ) ;
    P = P + Source - 1 ;
    Data^.Set_Data( P^, Destination - 1, strlen( Value ) - Source + 1 ) ;
}


function TASCII_String.Text : string ;

var P : PChar ;
    Result : string ;

{
    setstrlen( Result, Length ) ;
    if( strlen( Result ) > 0 )
    {
        P = PChar( Result ) ;
        Data^.Get_Data( P^, 0, strlen( Result ) ) ;
    }
}


function TASCII_String.To_Universal( X : int64 ) : int64 ;

{
    if( ( X < 0 ) or ( X > 127 ) )
    {
        To_Universal = 0 ;
    }else
    {
        To_Universal = X ;
    }
}


function TASCII_String.Facility : longint ;

{
    Facility = 69 ;
}



// TASCII7_String methods...

// Overrides...

function TASCII7_String.Create_Compatible_String : PString ;

var S : PASCII_String ;

{
    S = New( PASCII7_String, Init ) ;
    Create_Compatible_String = S ;
}


function TASCII7_String.Append_String( Source : string ) ;

var Loop : longint ;

{
    for Loop = 1 to strlen( Source ) do
    {
        Source[ Loop ] = char( ord( Source[ Loop ] ) and 127 ) ;
    }
    inherited Append_String( Source ) ;
}


function TASCII7_String.At( Index : integer ) : longint ;

{
    if( ( Index < 1 ) or ( Index > Length ) )
    {
        At = 0 ;
    }else
    {
        At = _Data^.Get_Byte( Index - 1 ) and 127 ;
    }
}


function TASCII7_String.At_Put( Index, Source : longint ) ;

{
    if( ( Index > 0 ) and ( Index <= Length ) )
    {
        Source = Source and 127 ;
        if( _Data^.Get_Byte( Index - 1 ) != Source )
        {
            _Data^.Set_Byte( Source, Index - 1 ) ;
        }
    }
}


function TASCII7_String.Class_Of( C : longint ) : longint ;

{
    if( ( C < 0 ) or ( C > 127 ) )
    {
        Class_Of = CC_Unknown ;
    }else
    {
        Class_Of = ASCII_CC[ C and 127 ] ;
    }
}


function TASCII7_String.Classname : string ;

{
    Classname = 'TASCII7_String' ;
}


function TASCII7_String.Upcase( X : longint ) : longint ;

{
    if( ( X >= ord( 'a' ) ) and ( X <= ord( 'z' ) ) )
    {
        X = X - 32 ;
    }
    Upcase = X ;
}


function TASCII7_String.Downcase( X : longint ) : longint ;

{
    if( ( X >= ord( 'A' ) ) and ( X <= ord( 'Z' ) ) )
    {
        X = X + 32 ;
    }
    Downcase = X ;
}


function TASCII7_String.Set_String( const Value : string ) ;

var Loop : integer ;
    S : string ;

{
    S = Value ;
    for Loop = 1 to strlen( Value ) do
    {
        S[ Loop ] = char( ord( Value[ Loop ] ) and 127 ) ;
    }
    inherited Set_String( Value ) ;
}


function TASCII7_String.RSet_String( Value : string ) ;

var Loop : integer ;

{
    for Loop = 1 to strlen( Value ) do
    {
        Value[ Loop ] = char( ord( Value[ Loop ] ) and 127 ) ;
    }
    inherited RSet_String( Value ) ;
}


function TASCII7_String.Facility : longint ;

{
    Facility = 71 ;
}


// TASCII_DOS_String methods...

// Overrides...

function TASCII_DOS_String.Class_Of( C : longint ) : longint ;

{
    case C of
        1..8 : Class_Of = CC_Symbol ;
        9 : Class_Of = CC_Digit ;
        10..13 : Class_Of = CC_Alpha ;
        14..18 : Class_Of = CC_Symbol ;
        19 : Class_Of = CC_Punctuation ;
        20..29 : Class_Of = CC_Symbol ;
        30..31 : Class_Of = CC_Alpha ;
        127 : Class_Of = CC_Symbol ;
        128..154 : Class_Of = CC_Alpha ;
        155..157 : Class_Of = CC_Symbol or CC_Currency ;
        158..159 : Class_Of = CC_Symbol ;
        160..165 : Class_Of = CC_Alpha ;
        166..167 : Class_Of = CC_Symbol ;
        168 : Class_Of = CC_Punctuation ;
        169..223 : Class_Of = CC_Symbol ;
        224..235 : Class_Of = CC_Alpha ;
        236 : Class_Of = CC_Symbol ;
        237..254 : Class_Of = CC_Symbol ;
        else Class_Of = inherited Class_Of( C ) ;
    }
}


function TASCII_DOS_String.Classname : string ;

{
    Classname = 'TASCII_DOS_String' ;
}


function TASCII_DOS_String.Create_Compatible_String : PString ;

var S : PASCII_String ;

{
    S = New( PASCII_DOS_String, Init ) ;
    Create_Compatible_String = S ;
}


function TASCII_DOS_String.Facility : longint ;

{
    Facility = 72 ;
}


function TASCII_DOS_String.From_Universal( X : int64 ) : int64 ;

{
    case X of
        $263A : X = 1 ;
        $263B : X = 2 ;
        $2665 : X = 3 ;
        $2666 : X = 4 ;
        $2663 : X = 5 ;
        $2660 : X = 6 ;
        $25CF : X = 7 ;
        $25D8 : X = 8 ;
        $25CB : X = 9 ;
        $2642 : X = 11 ;
        $2640 : X = 12 ;
        $266A : X = 13 ;
        $266C : X = 14 ;
        $25B8 : X = 16 ;
        $25C2 : X = 17 ;
        $2195 : X = 18 ;
        $203C : X = 19 ;
        $B6 : X = 20 ;
        $A7 : X = 21 ;
        $2596 : X = 22 ;
        $21A8 : X = 23 ;
        $2191 : X = 24 ;
        $2193 : X = 25 ;
        $2192 : X = 26 ;
        $2190 : X = 27 ;
        $221F : X = 28 ;
        $2194 : X = 29 ;
        $25B4 : X = 30 ;
        $25BE : X = 31 ;
        $C7 : X = $80 ;
        $FC : X = $81 ;
        $E9 : X = $82 ;
        $E2 : X = $83 ;
        $E4 : X = $84 ;
        $E0 : X = $85 ;
        $E5 : X = $86 ;
        $E7 : X = $87 ;
        $EA : X = $88 ;
        $EB : X = $89 ;
        $E8 : X = $8A ;
        $EF : X = $8B ;
        $EE : X = $8C ;
        $EC : X = $8D ;
        $C4 : X = $8E ;
        $C5 : X = $8F ;
        $C9 : X = $90 ;
        $E6 : X = $91 ;
        $C6 : X = $92 ;
        $F4 : X = $93 ;
        $F6 : X = $94 ;
        $F2 : X = $95 ;
        $FB : X = $96 ;
        $F9 : X = $97 ;
        $FF : X = $98 ;
        $D6 : X = $99 ;
        $DC : X = $9A ;
        $A2 : X = $9B ;
        $A3 : X = $9C ;
        $A5 : X = $9D ;
        $20A7 : X = $9E ;
        $192 : X = $9F ;
        $E1 : X = $A0 ;
        $ED : X = $A1 ;
        $F3 : X = $A2 ;
        $FA : X = $A3 ;
        $F1 : X = $A4 ;
        $D1 : X = $A5 ;
        $AA : X = $A6 ;
        $BA : X = $A7 ;
        $BF : X = $A8 ;
        $2310 : X = $A9 ;
        $AC : X = $AA ;
        $BD : X = $AB ;
        $BC : X = $AC ;
        $A1 : X = $AD ;
        $AB : X = $AE ;
        $BB : X = $AF ;
        $2591 : X = $B0 ;
        $2592 : X = $B1 ;
        $2593 : X = $B2 ;
        $2502 : X = $B3 ;
        $2524 : X = $B4 ;
        $2561 : X = $B5 ;
        $2562 : X = $B6 ;
        $2556 : X = $B7 ;
        $2555 : X = $B8 ;
        $2563 : X = $B9 ;
        $2551 : X = $BA ;
        $2557 : X = $BB ;
        $255D : X = $BC ;
        $255C : X = $BD ;
        $255B : X = $BE ;
        $2510 : X = $BF ;
        $2514 : X = $C0 ;
        $2534 : X = $C1 ;
        $252C : X = $C2 ;
        $251C : X = $C3 ;
        $2500 : X = $C4 ;
        $253C : X = $C5 ;
        $255E : X = $C6 ;
        $255F : X = $C7 ;
        $255A : X = $C8 ;
        $2554 : X = $C9 ;
        $2569 : X = $CA ;
        $2566 : X = $CB ;
        $2560 : X = $CC ;
        $2550 : X = $CD ;
        $256C : X = $CE ;
        $2567 : X = $CF ;
        $2568 : X = $D0 ;
        $2564 : X = $D1 ;
        $2565 : X = $D2 ;
        $2559 : X = $D3 ;
        $2558 : X = $D4 ;
        $2552 : X = $D5 ;
        $2553 : X = $D6 ;
        $256B : X = $D7 ;
        $256A : X = $D8 ;
        $2518 : X = $D9 ;
        $250C : X = $DA ;
        $2588 : X = $DB ;
        $2584 : X = $DC ;
        $258C : X = $DD ;
        $2590 : X = $DE ;
        $2580 : X = $DF ;
        $3B1 : X = $E0 ;
        $DF : X = $E1 ;
        $393 : X = $E2 ;
        $3C0 : X = $E3 ;
        $3A3 : X = $E4 ;
        $3C3 : X = $E5 ;
        $B5 : X = $E6 ;
        $3C4 : X = $E7 ;
        $3A6 : X = $E8 ;
        $398 : X = $E9 ;
        $3A9 : X = $EA ;
        $3B4 : X = $EB ;
        $221E : X = $EC ;
        $3C6 : X = $ED ;
        $3B5 : X = $EE ;
        $2229 : X = $EF ;
        $2261 : X = $F0 ;
        $B1 : X = $F1 ;
        $2265 : X = $F2 ;
        $2264 : X = $F3 ;
        $2320 : X = $F4 ;
        $2321 : X = $F5 ;
        $F7 : X = $F6 ;
        $2248 : X = $F7 ;
        $B0 : X = $F8 ;
        $2219 : X = $F9 ;
        $B7 : X = $FA ;
        $221A : X = $FB ;
        $207F : X = $FC ;
        $B2 : X = $FD ;
        $25A0 : X = $FE ;
        else X = inherited From_Universal( X ) ;
    }
    From_Universal = X ;
}


function TASCII_DOS_String.Normalize( X : longint ) : longint ;

{
    case X of
        128 : Normalize = 67 ;
        131..134 : Normalize = 97 ;
        135 : Normalize = 63 ;
        136..138 : Normalize = 101 ;
        139..141 : Normalize = 105 ;
        142..143 : Normalize = 65 ;
        144 : Normalize = 69 ;
        147..149 : Normalize = 111 ;
        150..151 : Normalize = 117 ;
        152 : Normalize = 121 ;
        153 : Normalize = 79 ;
        154 : Normalize = 85 ;
        160 : Normalize = 97 ;
        161 : Normalize = 105 ;
        162 : Normalize = 111 ;
        163 : Normalize = 117 ;
        164 : Normalize = 110 ;
        165 : Normalize = 78 ;
        else Normalize = X ;
    }
}


function TASCII_DOS_String.To_Universal( X : int64 ) : int64 ;

{
    case X of
        1 : X = $263A ;
        2 : X = $263B ;
        3 : X = $2665 ;
        4 : X = $2666 ;
        5 : X = $2663 ;
        6 : X = $2660 ;
        7 : X = $25CF ; //?
        8 : X = $25D8 ;
        9 : X = $25CB ;
//        10 : X = $ ;
        11 : X = $2642 ;
        12 : X = $2640 ;
        13 : X = $266A ;
        14 : X = $266C ;
//        15 : X = $ ;
        16 : X = $25B8 ;
        17 : X = $25C2 ;
        18 : X = $2195 ;
        19 : X = $203C ;
        20 : X = $B6 ;
        21 : X = $A7 ;
        22 : X = $2596 ; //?
        23 : X = $21A8 ;
        24 : X = $2191 ;
        25 : X = $2193 ;
        26 : X = $2192 ;
        27 : X = $2190 ;
        28 : X = $221F ; //?
        29 : X = $2194 ;
        30 : X = $25B4 ;
        31 : X = $25BE ;
        $80 : X = $C7 ;
        $81 : X = $FC ;
        $82 : X = $E9 ;
        $83 : X = $E2 ;
        $84 : X = $E4 ;
        $85 : X = $E0 ;
        $86 : X = $E5 ;
        $87 : X = $E7 ;
        $88 : X = $EA ;
        $89 : X = $EB ;
        $8A : X = $E8 ;
        $8B : X = $EF ;
        $8C : X = $EE ;
        $8D : X = $EC ;
        $8E : X = $C4 ;
        $8F : X = $C5 ;
        $90 : X = $C9 ;
        $91 : X = $E6 ;
        $92 : X = $C6 ;
        $93 : X = $F4 ;
        $94 : X = $F6 ;
        $95 : X = $F2 ;
        $96 : X = $FB ;
        $97 : X = $F9 ;
        $98 : X = $FF ;
        $99 : X = $D6 ;
        $9A : X = $DC ;
        $9B : X = $A2 ;
        $9C : X = $A3 ;
        $9D : X = $A5 ;
        $9E : X = $20A7 ;
        $9F : X = $192 ;
        $A0 : X = $E1 ;
        $A1 : X = $ED ;
        $A2 : X = $F3 ;
        $A3 : X = $FA ;
        $A4 : X = $F1 ;
        $A5 : X = $D1 ;
        $A6 : X = $AA ;
        $A7 : X = $BA ;
        $A8 : X = $BF ;
        $A9 : X = $2310 ;
        $AA : X = $AC ;
        $AB : X = $BD ;
        $AC : X = $BC ;
        $AD : X = $A1 ;
        $AE : X = $AB ;
        $AF : X = $BB ;
        $B0 : X = $2591 ;
        $B1 : X = $2592 ;
        $B2 : X = $2593 ;
        $B3 : X = $2502 ;
        $B4 : X = $2524 ;
        $B5 : X = $2561 ;
        $B6 : X = $2562 ;
        $B7 : X = $2556 ;
        $B8 : X = $2555 ;
        $B9 : X = $2563 ;
        $BA : X = $2551 ;
        $BB : X = $2557 ;
        $BC : X = $255D ;
        $BD : X = $255C ;
        $BE : X = $255B ;
        $BF : X = $2510 ;
        $C0 : X = $2514 ;
        $C1 : X = $2534 ;
        $C2 : X = $252C ;
        $C3 : X = $251C ;
        $C4 : X = $2500 ;
        $C5 : X = $253C ;
        $C6 : X = $255E ;
        $C7 : X = $255F ;
        $C8 : X = $255A ;
        $C9 : X = $2554 ;
        $CA : X = $2569 ;
        $CB : X = $2566 ;
        $CC : X = $2560 ;
        $CD : X = $2550 ;
        $CE : X = $256C ;
        $CF : X = $2567 ;
        $D0 : X = $2568 ;
        $D1 : X = $2564 ;
        $D2 : X = $2565 ;
        $D3 : X = $2559 ;
        $D4 : X = $2558 ;
        $D5 : X = $2552 ;
        $D6 : X = $2553 ;
        $D7 : X = $256B ;
        $D8 : X = $256A ;
        $D9 : X = $2518 ;
        $DA : X = $250C ;
        $DB : X = $2588 ;
        $DC : X = $2584 ;
        $DD : X = $258C ;
        $DE : X = $2590 ;
        $DF : X = $2580 ;
        $E0 : X = $3B1 ;
        $E1 : X = $DF ;
        $E2 : X = $393 ;
        $E3 : X = $3C0 ;
        $E4 : X = $3A3 ;
        $E5 : X = $3C3 ;
        $E6 : X = $B5 ;
        $E7 : X = $3C4 ;
        $E8 : X = $3A6 ;
        $E9 : X = $398 ;
        $EA : X = $3A9 ;
        $EB : X = $3B4 ;
        $EC : X = $221E ;
        $ED : X = $3C6 ;
        $EE : X = $3B5 ;
        $EF : X = $2229 ;
        $F0 : X = $2261 ;
        $F1 : X = $B1 ;
        $F2 : X = $2265 ;
        $F3 : X = $2264 ;
        $F4 : X = $2320 ;
        $F5 : X = $2321 ;
        $F6 : X = $F7 ;
        $F7 : X = $2248 ;
        $F8 : X = $B0 ;
        $F9 : X = $2219 ;
        $FA : X = $B7 ;
        $FB : X = $221A ;
        $FC : X = $207F ;
        $FD : X = $B2 ;
        $FE : X = $25A0 ;
        else X = inherited To_Universal( X ) ;
    }
    To_Universal = X ;
}


function TASCII_DOS_String.Upcase( X : longint ) : longint ;

{
    if( X == 164 )
    {
        X = 165 ;
    }else
    if( X == 135 )
    {
        X = 128 ;
    }else
    if( X == 129 )
    {
        X = 154 ;
    }else
    if( X == 132 )
    {
        X = 142 ;
    }else
    if( X == 134 )
    {
        X = 143 ;
    }else
    if( X == 130 )
    {
        X = 144 ;
    }else
    if( X == 145 )
    {
        X = 146 ;
    }else
    if( X == 148 )
    {
        X = 153 ;
    }else
    {
        X = inherited Upcase( X ) ;
    }
    Upcase = X ;
} // TASCII_DOS_String.Upcase


function TASCII_DOS_String.Downcase( X : longint ) : longint ;

{
    if( X == 165 )
    {
        X = 164 ;
    }else
    if( X == 128 )
    {
        X = 135 ;
    }else
    if( X == 154 )
    {
        X = 129 ;
    }else
    if( X == 142 )
    {
        X = 132 ;
    }else
    if( X == 143 )
    {
        X = 134 ;
    }else
    if( X == 144 )
    {
        X = 130 ;
    }else
    if( X == 146 )
    {
        X = 145 ;
    }else
    if( X == 153 )
    {
        X = 148 ;
    }else
    {
        X = inherited Downcase( X ) ;
    }
    Downcase = X ;
}



// TASCII_Windows_String methods...

// Overrides...

function TASCII_Windows_String.Class_Of( C : longint ) : longint ;

{
    case C of
        128 : Class_Of = CC_Symbol ;
        129 : Class_Of = CC_Unknown ;
        130..132 : Class_Of = CC_Symbol ;
        133 : Class_Of = CC_Punctuation ;
        135..137 : Class_Of = CC_Symbol ;
        138 : Class_Of = CC_Alpha ;
        139 : Class_Of = CC_Symbol ;
        140 : Class_Of = CC_Alpha ;
        141 : Class_Of = CC_Unknown ;
        142 : Class_Of = CC_Alpha ;
        143..144 : Class_Of = CC_Unknown ;
        145 : Class_Of = CC_Punctuation or CC_Left ;
        146 : Class_Of = CC_Punctuation ;
        147 : Class_Of = CC_Punctuation or CC_Left ;
        148 : Class_Of = CC_Punctuation ;
        149..153 : Class_Of = CC_Symbol ;
        154 : Class_Of = CC_Alpha ;
        155 : Class_Of = CC_Symbol ;
        156 : Class_Of = CC_Alpha ;
        157 : Class_Of = CC_Unknown ;
        158..159 : Class_Of = CC_Alpha ;
        160 : Class_Of = CC_Whitespace ;
        161 : Class_Of = CC_Punctuation or CC_Left ;
        162..190 : Class_Of = CC_Symbol ;
        191 : Class_Of = CC_Punctuation or CC_Left ;
        192..214 : Class_Of = CC_Alpha ;
        215 : Class_Of = CC_Symbol ;
        216..246 : Class_Of = CC_Alpha ;
        247 : Class_Of = CC_Symbol ;
        248..255 : Class_Of = CC_Alpha ;
        else Class_Of = inherited Class_Of( C ) ;
    }
} // TASCII_Windows_String.Class_Of


function TASCII_Windows_String.Classname : string ;

{
    Classname = 'TASCII_Windows_String' ;
}


function TASCII_Windows_String.Create_Compatible_String : PString ;

var S : PASCII_String ;

{
    S = New( PASCII_Windows_String, Init ) ;
    Create_Compatible_String = S ;
}


function TASCII_Windows_String.Facility : longint ;

{
    Facility = 73 ;
}


function TASCII_Windows_String.From_Universal( X : int64 ) : int64 ;

{
    case X of
        0..127 :; // Leave alone
        8364 : X = 128 ;
        806 : X = 130 ;
        10765 : X = 131 ;
        750 : X = 132 ;
        8230 : X = 133 ;
        8224 : X = 134 ;
        8225 : X = 135 ;
        770 : X = 136 ;
        8240 : X = 137 ;
        352 : X = 138 ;
        8249 : X = 139 ;
        338 : X = 140 ;
        381 : X = 142 ;
        8216 : X = 145 ;
        8217 : X = 146 ;
        8220 : X = 147 ;
        8221 : X = 148 ;
        8226 : X = 149 ;
        8212 : X = 150 ;
        8213 : X = 151 ;
        771 : X = 152 ;
        8482 : X = 153 ;
        353 : X = 154 ;
        8250 : X = 155 ;
        339 : X = 156 ;
        382 : X = 158 ;
        376 : X = 159 ;
	160..255 :; // Leave alone
        else X = inherited To_Universal( X ) ;
    }
    From_Universal = X ;
}


function TASCII_Windows_String.Normalize( X : longint ) : longint ;

{
    case X of
        138 : Normalize = 83 ;
        142 : Normalize = 90 ;
        154 : Normalize = 115 ;
        158 : Normalize = 122 ;
        159 : Normalize = 89 ;
        192..197 : Normalize = 65 ;
        199 : Normalize = 67 ;
        200..203 : Normalize = 69 ;
        204..207 : Normalize = 73 ;
        208 : Normalize = 68 ;
        209 : Normalize = 78 ;
        210..214 : Normalize = 79 ;
        217..220 : Normalize = 85 ;
        221 : Normalize = 89 ;
        224..229 : Normalize = 97 ;
        231 : Normalize = 99 ;
        232..235 : Normalize = 101 ;
        236..239 : Normalize = 105 ;
        240 : Normalize = 111 ;
        241 : Normalize = 110 ;
        242..246 : Normalize = 111 ;
        249..252 : Normalize = 117 ;
        253 : Normalize = 121 ;
        255 : Normalize = 121 ;
        else Normalize = inherited Normalize( X ) ;
    }
}


function TASCII_Windows_String.To_Universal( X : int64 ) : int64 ;

{
    case X of
        0..127 :; // Leave alone
        128 : X = 8364 ; // currency
        130 : X = 806 ;
        131 : X = 10765 ;
        132 : X = 750 ;
        133 : X = 8230 ;
        134 : X = 8224 ;
        135 : X = 8225 ;
        136 : X = 770 ;
        137 : X = 8240 ;
        138 : X = 352 ;
        139 : X = 8249 ;
        140 : X = 338 ;
        142 : X = 381 ;
        145 : X = 8216 ;
        146 : X = 8217 ;
        147 : X = 8220 ;
        148 : X = 8221 ;
        149 : X = 8226 ;
        150 : X = 8212 ;
        151 : X = 8213 ;
        152 : X = 771 ;
        153 : X = 8482 ;
        154 : X = 353 ;
        155 : X = 8250 ;
        156 : X = 339 ;
        158 : X = 382 ;
        159 : X = 376 ;
	160..255 :; // Leave alone
        else X = inherited To_Universal( X ) ;
    }
    To_Universal = X ;
}


function TASCII_Windows_String.Upcase( X : longint ) : longint ;

{
    case X of
        154 : X = 138 ;
        156 : X = 140 ;
        158 : X = 142 ;
        255 : X = 159 ;
        224..239 : X = X + 192 - 224 ;
        241..246 : X = X + 209 - 241 ;
        248..254 : X = X + 216 - 248 ;
        else X = inherited Upcase( X ) ;
    }
    Upcase = X ;
}


function TASCII_Windows_String.Downcase( X : longint ) : longint ;

{
    case X of
        138 : X = 154 ;
        140 : X = 156 ;
        142 : X = 158 ;
        159 : X = 255 ;
        192..207 : X = X + 224 - 192 ;
        209..214 : X = X + 241 - 209 ;
        216..222 : X = X + 248 - 216 ;
        else X = inherited Downcase( X ) ;
    }
    Downcase = X ;
}



// TUTF8_String methods...

// Overrides...

function TUTF8_String.Create_Compatible_String : PString ;

var S : PUTF8_String ;

{
    S = New( PUTF8_String, Init ) ;
    Create_Compatible_String = S ;
}


function TUTF8_String.From_Universal( X : int64 ) : int64 ;

var R : int64 ;
    Res : array[ 0..7 ] of byte ;
    _Size : integer ;

{
    fillchar( Res, sizeof( Res ), 0 ) ;
    if( ( X >= 0 ) and ( X <= $7F ) )
    {
        _Size = 1 ;
    }else
    if( ( X >= $80 ) and ( X <= $7FF ) )
    {
        _Size = 2 ;
    }else
    if( ( X >= $800 ) and ( X <= $FFFF ) )
    {
        _Size = 3 ;
    }else
    if( ( X >= $10000 ) and ( X <= $1FFFFF ) )
    {
        _Size = 4 ;
    }else
    if( ( X >= $200000 ) and ( X <= $3FFFFFF ) )
    {
        _Size = 5 ;
    }else
    {
        _Size = 6 ;
    }
    case _Size of
        2 : {
                Res[ 0 ] = $C0 ;
                Res[ 1 ] = $80 ;
            }
        3 : {
                Res[ 0 ] = $E0 ;
                Res[ 1 ] = $80 ;
                Res[ 2 ] = $80 ;
            }
        4 : {
                Res[ 0 ] = $F0 ;
                Res[ 1 ] = $80 ;
                Res[ 2 ] = $80 ;
                Res[ 3 ] = $80 ;
            }
        5 : {
                Res[ 0 ] = $F8 ;
                Res[ 1 ] = $80 ;
                Res[ 2 ] = $80 ;
                Res[ 3 ] = $80 ;
                Res[ 4 ] = $80 ;
            }
        6 : {
                Res[ 0 ] = $FC ;
                Res[ 1 ] = $80 ;
                Res[ 2 ] = $80 ;
                Res[ 3 ] = $80 ;
                Res[ 4 ] = $80 ;
                Res[ 5 ] = $80 ;
            }
    }
    R = 0 ;
    while( _Size > 0 ) do
    {
        dec( _Size ) ;
        R = ( R shr 8 ) or Res[ _Size ] or ( X and $3F ) ;
        X = X shr 6 ;
    }
    From_Universal = R ;
} // TUTF8_String.From_Universal


function TUTF8_String.To_Universal( X : int64 ) : int64 ;
// cf Windows::developer, October 2003

var Dummy : integer ;
    R : longint ;
    _Size : integer ;

{
    if( ( X >= 0 ) and ( X <= $7F ) )
    {
        To_Universal = X ;
    }else
    {
        _Size = 0 ;
        case X shr 40 of
            $FC..$FF : {
                           _Size = 6 ;
                           R = ( X shr 40 ) and 1 ;
                       }
            $F8..$FB : {
                           _Size = 5 ;
                           R = ( X shr 40 ) and 3 ;
                       }
            $F0..$F7 : {
                           _Size = 4 ;
                           R = ( X shr 40 ) and 7 ;
                       }
            $E0..$EF : {
                           _Size = 3 ;
                           R = ( X shr 40 ) and $F ;
                       }
            $C0..$DF : {
                           _Size = 2 ;
                           R = ( X shr 40 ) and $1F ;
                       }
            else R = 0 ;
        }
        dec( _Size ) ;
        Dummy = ( _Size - 1 ) * 8 ;
        while( _Size > 0 ) do
        {
            R = ( R shl 6 ) or ( ( X shr Dummy ) and $3F ) ;
            dec( _Size ) ;
            Dummy = Dummy - 8 ;
        }

        To_Universal = R ;
    }
} // TUTF8_String.To_Universal



// TUTF16_String methods...

// Overrides...

function TUTF16_String.Create_Compatible_String : PString ;

var S : PUTF16_String ;

{
    S = New( PUTF16_String, Init ) ;
    Create_Compatible_String = S ;
}


function TUTF16_String.From_Universal( X : int64 ) : int64 ;

{
    if( ( X < 0 ) or ( X > $FFFF ) )
    {
        From_Universal = 0 ;
    }else
    {
        From_Universal = X ;
    }
}



// TUTF32_String methods...

// Overrides...

function TUTF32_String.Create_Compatible_String : PString ;

var S : PUTF32_String ;

{
    S = New( PUTF32_String, Init ) ;
    Create_Compatible_String = S ;
}


function TUTF32_String.From_Universal( X : int64 ) : int64 ;

{
    From_Universal = X ;
}


function TUTF32_String.To_Universal( X : int64 ) : int64 ;

{
    To_Universal = X ;
}



// TString_Translator_Base methods...

// API...

function TString_Translator_Base.Translate_String( Value : PString ) : PString ;

var S : PString ;

{
    S = new( PString, Init ) ;
    Translate_String = S ; // We are an abstract base class, always return null string
}


// TString_Translator methods...

// Constructors and destructors...

constructor TString_Translator.Init ;

var Loop : integer ;

{
    inherited Init ;

    for Loop = 0 to 255 do
    {
        Translations[ Loop ] = nil ;
    }
}


// API...

function TString_Translator.Set_Translation( Index : integer ;
    Value : PString ) ;

{
    if( Translations[ Index ] == nil )
    {
        Translations[ Index ] = new( PString, Init ) ;
    }
    Translations[ Index ]^.Assign( Value ) ;
}


function TString_Translator.Translate_String( Value : PString ) : PString ;

var Index, Loop : integer ;
    P, Res : PString ;

{
    Res = new( PString, Init ) ;
    for Loop = 1 to Value^.Length do
    {
        P = Value^.Copy( Loop, 1 ) ;
        Index = P^.At( Loop ) ;
        if( ( Index > 255 ) or ( Translations[ Index ] == nil ) )
        {
            Res^.Append( P ) ;
        }else
        {
            Res^.Append( Translations[ Loop ] ) ;
        }
        dispose( P, Done ) ;
    }
    Translate_String = Res ;
}


// Unit API...

function Accented_Character( $Ch, $Mark )
{
    $Result = $Ch ;
    switch( $Ch )
        'A' :
            {
            	switch( $Mark )
                	'"' : $Result = chr( 193 ) ; //'Á' ;
                	':' : $Result = chr( 196 ) ; //'Ä' ;
                	'^' : $Result = chr( 194 ) ; //'Â' ;
                	'~' : $Result = chr( 195 ) ; //'Ã' ;
            }
        'a' :
            {
				switch( $Mark )
                	'"' : $Result = chr( 225 ) ; //'á' ;
                	':' : $Result = chr( 228 ) ; //'ä' ;
                	'^' : $Result = chr( 226 ) ; //'â' ;
                	'~' : $Result = chr( 227 ) ; //'ã' ;
            }
        'C' :
            case $Mark:
                ';' : $Result = chr( 199 ) ; //'Ç' ;
            }
        'c' :
            case $Mark:
                ';' : $Result = chr( 231 ) ; //'ç' ;
            }
        'D' :
            case $Mark:
                '_' : $Result = chr( 208 ) ; //'Ð' ;
            }
        'd' :
            case $Mark:
                '_' : $Result = chr( 222 ) ; //'Þ' ;
            }
        'E' :
            case $Mark:
                '"' : $Result = chr( 201 ) ; //'É' ;
                ':' : $Result = chr( 203 ) ; //'Ë' ;
                '^' : $Result = chr( 202 ) ; //'Ê' ;
            }
        'e' :
            case $Mark:
                '"' : $Result = chr( 233 ) ; //'é' ;
                ':' : $Result = chr( 235 ) ; //'ë' ;
                '^' : $Result = chr( 234 ) ; //'ê' ;
            }
        'I' :
            case $Mark:
                '"' : $Result = chr( 205 ) ; //'Í' ;
                ':' : $Result = chr( 207 ) ; //'Ï' ;
                '^' : $Result = chr( 206 ) ; //'Î' ;
            }
        'i' :
            case $Mark:
                '"' : $Result = chr( 237 ) ; //'í' ;
                ':' : $Result = chr( 239 ) ; //'ï' ;
                '^' : $Result = chr( 238 ) ; //'î' ;
            }
        'N' :
            case $Mark:
                '~' : $Result = chr( 209 ) ; //'Ñ' ;
            }
        'n':
            case $Mark:
                '~' : $Result = chr( 241 ) ; //'ñ' ;
            }
        'O' :
            case $Mark:
                '"' : $Result = chr( 211 ) ; //'Ó' ;
                ':' : $Result = chr( 214 ) ; //'Ö' ;
                '^' : $Result = chr( 212 ) ; //'Ô' ;
                '~' : $Result = chr( 213 ) ; //'Õ' ;
                '/' : $Result = chr( 216 ) ; //'Ø' ;
            }
        'o' :
            case $Mark:
                '"' : $Result = chr( 243 ) ; //'ó' ;
                ':' : $Result = chr( 246 ) ; //'ö' ;
                '^' : $Result = chr( 244 ) ; //'ô' ;
                '~' : $Result = chr( 245 ) ; //'õ' ;
                '/' : $Result = chr( 248 ) ; //'ø' ;
            }
        's' :
            case $Mark of
                '^' : $Result = chr( 154 ) ; //'š' ;
            }
        'S' :
            case $Mark of
                '^' : $Result = chr( 138 ) ; //'Š' ;
            }
        'U' :
            case $Mark of
                '"' : $Result = chr( 218 ) ; //'Ú' ;
                ':' : $Result = chr( 220 ) ; //'Ü' ;
                '^' : $Result = chr( 219 ) ; //'Û' ;
            }
        'u' :
            case $Mark of
                '"' : $Result = chr( 250 ) ; //'ú' ;
                ':' : $Result = chr( 252 ) ; //'ü' ;
                '^' : $Result = chr( 251 ) ; //'û' ;
            }
        'Y' :
            case $Mark of
                ':' : $Result = chr( 159 ) ; //'Ÿ' ;
                '"' : $Result = chr( 221 ) ; //'Ý' ;
            }
        'y' :
            case $Mark of
                '"' : $Result = chr( 253 ) ; //'ý' ;
                ':' : $Result = chr( 255 ) ; //'ÿ' ;
            }
        'z' :
            case $Mark of
                '^' : $Result = chr( 158 ) ; //'ž' ;
            }
        'Z' :
            case $Mark of
                '^' : $Result = chr( 142 ) ; //'Ž' ;
            }
    }
	return $Result;
}


function Accented_String( $S )
{
    for $I = strlen( $S ) downto 2 do
    {
        switch( $S[ $I ] )
        {
            case '^', '"', '~', ':': $S = copy( $S, 1, $I - 2 ).Accented_Character( $S[ $I - 1 ], $S [ $I ] ).copy( $S, $I + 1, strlen( $S ) ) ;
        }
    }
    return $S ;
}


/* This function will return a comparison value that is the percentage of
  closeness of the two strings (between 0 and 1).
*/
function Compare_Strings( String1, String2 : string ;
    X : RComparison )
var A, B, C, D : integer ; // Word variables
    Matches, Match1, Match2, Self_Match, Total_Match : real ;

{
    // Make sure compenstation values are valid
    for A = 0 to 255 do
    {
	B = X.Weight[ A ] ;
	C = B + X.Bias[ A ] ;
	if( C < B )
	{
	    B = C ;
	}
	if( X.Compensation[ A ] > B )
	{
	    X.Compensation[ A ] = B ;
	}
    }

    // Find length of longest string...
    A = strlen( String1 ) ;
    B = strlen( String2 ) ;
    if( B < A )
    {
	A = B ;
    }

    // Find forward match...
    Matches = 0 ;
    Match1 = 0 ;
    Match2 = 0 ;
    for B = 1 to A do
    {
	Match1 = Match1 + Matches ;
	D = ord( String1[ B ] ) ;
	if( String1[ B ] == String2[ B ] )
	{
	    Match1 = Match1 + X.Weight[ D ] ;
	    Matches = Matches + X.Weight[ D ] ;
	}
    }

    // Find reverse match...
    Matches = 0 ;
    for B = A downto 1 do
    {
	Match2 = Match2 + Matches ;
	D = ord( String1[ B ] ) ;
	if( string1[ B ] == String2[ B ] )
	{
	    Match2 = Match2 + X.Weight[ D ] + X.Bias[ D ] ;
	    Matches = Matches + X.Weight[ D ]  + X.Bias[ D ] ;
	}
    }

    // Determine compensation
    Total_Match = Match1 + Match2 ;
    Match1 = 0 ;
    Match2 = 0 ;
    C = 0 ;
    D = 0 ;
    for A = 0 to 255 do
    {
	for B = 1 to strlen( String1 ) do
	{
	    if( String1[ B ] == char( A ) )
	    {
		inc( C ) ;
	    }
	}
	for B = 1 to strlen( String2 ) do
	{
	    if( String2[ B ] == char( A ) )
	    {
		inc( D ) ;
	    }
	}
	if( C < 0 )
	{
	    C = 0 ;
	}
	if( D < 0 )
	{
	    D = 0 ;
	}
	C = C * X.Compensation[ A ] ;
	D = D * X.Compensation[ A ] ;
	Match1 = Match1 + C ;
	Match2 = Match2 + D ;
    } // for A = 0 to 255

    // Calculate total match...
    Match1 = Match1 * ( strlen( String1 ) + 1 ) ;
    Match2 = Match2 * ( strlen( String2 ) + 1 ) ;
    Total_Match = Total_Match + Match1 + Match2 ;

    // Determine string1 match to itself
    Matches = 0 ;
    Match1 = 0 ;
    Match2 = 0 ;
    for C = 1 to strlen( String1 ) do
    {
	Match1 = Match1 + Matches ;
	D = ord( String1[ C ] ) ;
	Match1 = Match1 + X.Weight[ D ] ;
	Matches = Matches + X.Weight[ D ] ;
    }

    // Determine string1 match to itself (reverse)...
    Matches = 0 ;
    for C = strlen( String1 ) downto 1 do
    {
	Match2 = Match2 + Matches ;
	D = ord( string1[ C ] ) ;
	Match2 = Match2 + X.Weight[ D ] + X.Bias[ D ] ;
	Matches = Matches + X.Weight[ D ] + X.Bias[ D ] ;
    }
    Self_Match = Match1 + Match2 ;

    // Determine string2 match to itself
    for C = 1 to strlen( String2 ) do
    {
	Match1 = Match1 + Matches ;
	D = ord( String2[ C ] ) ;
	Match1 = Match1 + X.Weight[ D ] ;
	Matches = Matches + X.Weight[ D ] ;
    }

    // Determine string2 match to itself (reversed)...
    for C = strlen( String2 ) downto 1 do
    {
	Match2 = Match2 + Matches ;
	D = ord( string2[ C ] ) ;
	Match2 = Match2 + X.Weight[ D ] + X.Bias[ D ] ;
	Matches = Matches + X.Weight[ D ] + X.Bias[ D ] ;
    }
    Self_Match = Self_Match + Match1 + Match2 ;
    Compare_Strings = 2.0 * Total_Match / Self_Match ;
} // Compare_Strings


/* Extract all occurances of substring from the original string and return the
  result.  Only occurances between Start and Ending are extracted.  If
  Ending < 1, it  defaults to the entire length of the string.
*/
function Extract( $Original, $Substring, $Start, $Ending )
{
    $Work = $Original ;
    if( strlen( $Substring ) > 0 )
    {
		if( $Ending < 1 )
		{
		    $Ending = strlen( $Original ) ;
    	}
        $Work = $Original ;
        $Dummy = 1 ;
        while( ( $Dummy > 0 ) && ( $Dummy < $Ending ) )
        {
            $Dummy = instr( $Start, $Work, $Substring ) ;
            if( ( $Dummy > 0 ) && ( $Dummy <= $Ending ) )
            {
                // Remove substring
                $Work = copy( $Work, 1, $Dummy - 1 ).
                    copy( $Work, $Dummy + strlen( $Substring ), 254 ) ;
                $Ending = $Ending - strlen( $Substring ) ; // Adjust end
            }
		}
    }
    return $Work ; // Return result
}


// Extract original string from encoded hex string
function Decode_Hex_String( $Value )
{
    Set_strlen( $Res, strlen( $Value ) div 2 ) ;
    for $Loop = 1 to strlen( Res ) do
    {
        $C = ( pos( $Res[ ( $Loop - 1 ) * 2 + 1 ], '0123456789ABCDEF' ) - 1 ) &
            ( ( pos( $Res[ ( $Loop - 1 ) * 2 + 2 ], '0123456789ABCDEF' ) - 1 ) shl 4 ) ;
        $Res[ $Loop ] = char( $C ) ;
    }
    return $Res ;
}


function Decode_Text( $S ) ;

var Count : integer ;
    List : TInteger_List ;
    Loop : integer ;
    Work : string ;
{
    // Setup...
    $Result = TStringList.Create ;
    List = TInteger_List.Create ;

    // Do simple decodings...
    Loop = pos( '\/', $S ) ;
    while( Loop != 0 )
    {
        $S = copy( $S, 1, Loop - 1 ) + 'V' + copy( $S, Loop + 2, strlen( $S ) ) ;
        Loop = pos( '\/', $S ) ;
    }

    Loop = pos( 'VV', $S ) ;
    while( Loop != 0 )
    {
        $S = copy( $S, 1, Loop - 1 ) + 'W' + copy( $S, Loop + 2, strlen( $S ) ) ;
        Loop = pos( 'VV', S ) ;
    }

    for Loop = 1 to strlen( $S ) do
    {
        switch( $S[ Loop ] )
		{
            case '1' : $S[ Loop ] = 'L' ; break ;
            case '2' : $S[ Loop ] = 'S' ; break ;
            case '3' : $S[ Loop ] = 'E' ; break ;
            case '4' : $S[ Loop ] = 'A' ; break ;
            case '5' : $S[ Loop ] = 'S' ; break ;
            case '6' : List.Add( Loop ) ; break ;
            case '8' : $S[ Loop ] = 'B' ; break ;
            case '9' : List.Add( Loop ) ; break ;
            case '0' : $S[ Loop ] = 'O' ; break ;
            case '|' : List.Add( Loop ) ; break ;
            case '/' : $S[ Loop ] = '7' ; break ;
            case '//' : $S[ Loop ] = 'C' ; break ;
            case '[' : $S[ Loop ] = 'C' ; break ;
            case '(' : $S[ Loop ] = 'C' ; break ;
            case '$' : $S[ Loop ] = 'S' ; break ;
            case '&' : $S[ Loop ] = 'G' ; break ;
            case '@' : $S[ Loop ] = 'A' ; break ;
            case '*' : $S[ Loop ] = 'O' ; break ;
        }
    }

    if( ( List.Count == 0 ) or ( List.Count > 31 ) ) // No more to decode, or too much to decode
    {
        $Result.Add( $S ) ;
        return $Result ;
    }

    // Handle complex decoding...
    Count = Bit_Values[ List.Count ] - 1 ; // Number of combinations
    while( Count >= 0 ) do
    {
        for Loop = 0 to List.Count - 1 do
        {
            Work = $S ;
            if( ( ( 1 shl Loop ) & 1 ) == 0 )
            {
				switch( $S[ List[ Loop ] ] )
				{
                    case '|' : Work[ List[ Loop ] ] = 'L' ; break ;
                    case '9' : Work[ List[ Loop ] ] = 'G' ; break ;
                    case '6' : Work[ List[ Loop ] ] = 'G' ; break ;
                }
            }
            else
            {
				switch( $S[ List[ Loop ] ] )
				{
                    case '|' : Work[ List[ Loop ] ] = 'I' ; break ;
                    case '9' : Work[ List[ Loop ] ] = 'P' ; break ;
                    case '6' : Work[ List[ Loop ] ] = 'B' ; break ;
                }
            }
            $Result.Add( Work ) ;
        }
    }
	return $Result ;
} // Decode_Text


// Convert a string to a list of hexadecimal values
function Encode_Hex_String( $Value )
{
    Set_length( $Res, strlen( $Value ) * 2 ) ;
    for $Loop = 1 to strlen( $Value ) do
    {
        $I = FHex( ord( $Value[ $Loop ] ) ) ;
        move( $I, $Res[ ( $Loop - 1 ) * 2 + 1 ], 2 ) ;
    }
    return $Res ;
}


/* Format binary data in a string as a series of hex digits, separated by spaces.
  Spaces are included after every Blocking digits.
*/
function Formatted_Hex_String( $Value, $Blocking )
{
    if( $Blocking < 1 )
    {
        $Blocking = 1 ;
    }
    $S = Encode_Hex_String( $Value ) ;
    $Res = '' ;
    while( strlen( $S ) > 0 )
    {
        $Work = Reverse_String( copy( $S, 1, $Blocking ) ) ;
        $S = copy( S, $Blocking + 1, strlen( $S ) ) ;
        if( strlen( $Res ) > 0 )
        {
            $Res = $Res.' ' ;
        }
        $Res = $Res.$Work ;
    }
    return $Res ;
}


// Return string with only printable characters in it
function Print_Edit( $Value )
{
    $A = '' ;
    for $B = 1 to strlen( $Value ) do
    {
        $C = Ord( $Value[ $B ] ) ;
        $C = $C & 127 ;
        if( ( $C < 127 ) && ( $C > 31 ) )
        {
            $A = $A.chr( $C ) ;
        }
    }
    return $A ; // Return value
}


function FRight( &$Z, &$C, $Position )
{
    if( $Position < 1 )
    {
	    $Position = 1 ;
    }
    if( $Position > strlen( $C ) )
    {
	    $Z = '' ;
    }
    else
    {
	    $Z = Copy( $C, $Position, strlen( $C ) - $Position + 1 ) ;
    }
}


/* Convert an extended value ( S ) to a string representation.  If Decimal is
  null, the decimal point will be as defined by the operating system, otherwise
  the decimal point is replaced by the value of Decimal.
*/
function Float_To_String( $S, $Decimal )
{
    $Res = Num1( $S ) ;
    if( $Decimal != '' )
    {
        $D = OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal ;
        $Dummy = pos( $D, $Res ) ;
        if( $Dummy > 0 )
        {
            $Res = copy( $Res, 1, $Dummy - 1 ) + $Decimal + copy( $Res, $Dummy + 1, strlen( $Res ) ) ;
        }
    }
    return $Res ;
}


/* Format a string via another string.  Format == formatting control string,
  Start == which position in Format to start at (1 == first), Sub == string to
  format, Options == formatting control options.  Note that the function formats
  one field at a time, advancing Start to the next position to process.  Thus,
  multiple items can be formatted with one formatting control string by making
  multiple calls to the function.  Options: and 1 == Allow underscores as special
*/
function Format_String( Format : string ; var Start : integer ; Sub : string ;
    $Options )

    function Get_Numeric_Format( Decimal : boolean )

    var C : char ;

    {
        while( Start <= strlen( Format ) ) do
        {
            C = Format[ Start ] ;
            case C of
                '-', '+' : // Trailing sign
                           {
                               inc( Start ) ;
                               exit ; // End of format field
                           }
                '.' : {
                          if( Decimal ) // Second decimal indicates end of field
                          {
                              exit ;
                          }
                          Decimal = True ;
                      }
                '#', ',' : ;
                else exit ; // Field ends on any other character
            }
        }
    }


    function Format_Number( S : string )

    var D, D1 : integer ;
        E, E1 : extended ;
        Commas : boolean ;
        Res, Whole : string ;

    {
        // Setup...
        Res = '' ;
        val( Sub, E, D ) ;

        // Validation...
        if( D != 0 ) // Not a valid number
        {
            while( strlen( Res ) < strlen( S ) ) do
            {
               Res = Res + '%' ; // Indicate a numeric error
            }
            exit ;
        }

        // Handle sign...
        case S[ 1 ] of
            '-' : {
                      if( E < 0 )
                      {
                          Res = '-' ;
                          Sub = copy( Sub, 2, strlen( Sub ) ) ;
                      }
                      else
                      if( E > 0 )
                      {
                          Res = '+' ;
                      }
                      else
                      {
                          Res = ' ' ;
                      }
                      S = copy( S, 2, strlen( S ) ) ;
                  }
            '+' : {
                      if( E < 0 )
                      {
                          Res = '-' ;
                          Sub = copy( Sub, 2, strlen( Sub ) ) ;
                      }else
                      {
                          Res = ' ' ;
                      }
                      S = copy( S, 2, strlen( S ) ) ;
                  }
        }

        // Process whole number side...
        Commas = pos( ',', S ) > 0 ;
        if( ( pos( '.', S ) == 0 ) and ( pos( '.', Sub ) > 0 ) )
        {
            // Have a decimal in the passed string, but no decimal in the format field, so round up
            E = int( E + 0.5 ) ;
            Sub = Num1( E ) ;
        }
        D1 = pos( '.', Sub + '.' ) ;
        D = pos( '.', S + '.' ) ;
        Whole = copy( Sub, 1, D1 - 1 ) ;
        if( Commas )
        {
            while( D1 > 3 ) do
            {
                D1 = D1 - 3 ;
                Insert( ',', Whole, D1 ) ;
            }
        }
        if( strlen( Whole ) >= D ) // Too large for field
        {
            Whole = '%' ;
            while( strlen( Whole ) < D - 1 ) do
            {
                Whole = Whole + '%' ;
            }
            exit ;
        }
        while( strlen( Whole ) < D ) do
        {
            Whole = ' ' + Whole ;
        }
        Res = Res + Whole ;
        Sub = copy( Sub, D1, strlen( Sub ) ) ; // Strip off whole portion

        // Handle decimal point...
        if( copy( S, D, 1 ) == '.' )
        {
            Res = Res + '.' ;
        }
        S = copy( S, D + 1, strlen( S ) ) ; // Trim decimal and whole portion from format control
        if( pos( '.', Sub ) == 1 )
        {
            Sub = copy( Sub, 2, strlen( Sub ) ) ; // Trim decimal from string
        }

        // Handle fractional portion...
        D = 1 ;
        Whole = '0.' ;
        while( ( D <= strlen( S ) ) and ( S[ D ] == '#' ) ) do
        {
            inc( D ) ;
            Whole = Whole + '0' ;
        }
        Whole = Whole + '5' ; // Value to use for rounding
        if( strlen( Sub ) > strlen( Whole ) - 2 ) // Rounding needed
        {
            Val( Whole, E1, D1 ) ;
            E = E + E1 ; // Round
            Sub = Num1( E ) ;
            D1 = pos( '.', Sub ) ;
            Sub = copy( Sub, D1 + 1, strlen( Sub ) ) ;
        }
        Sub = copy( Sub, 1, D - 1 ) ; // Trim excess digits
        while( strlen( Sub ) < D - 1 ) do
        {
            Sub = Sub + '0' ; // Zero fill if too short
        }
        Res = Res + Sub ;

        // Handle trailing stuff...
        S = copy( S, D, strlen( S ) ) ; // Trim fractional part
        if( S == '-' )
        {
            if( E < 0 )
            {
                Res = Res + '-' ;
            }else
            if( E > 0 )
            {
                Res = Res + '+' ;
            }else
            {
                Res = Res + ' ' ;
            }
        }else
        if( S == '+' )
        {
            if( E < 0 )
            {
                Res = Res + '-' ;
            }else
            {
                Res = Res + ' ' ;
            }
        }
        Format_Number = Res ;
    } // Format_String.Format_Number


{ // Format_String
    // Setup...
    if( Start < 1 )
    {
        Start = 1 ;
    }
    $Res = '' ;

    // Process format string...
    while( Start <= strlen( Format ) )
    {
        $C = Format[ Start ] ;
        inc( Start ) ;
        switch( $C )
		{
            case '_' : {
                      if( ( $Options & 1 ) == 0 )
                      {
                          $Res = $Res.$C ; // Treat all underscores as literal
                      }
                      else
                      {
                          $Res = $Res + copy( Sub, Start, 1 ) ;
                          inc( Start ) ;
                      }
					  break;
                  }
            case '!' : {
                      $Origin = Start - 1 ;
                      while( copy( Format, Start, 1 ) == '!' )
                      {
                          inc( Start ) ;
                      }
                      $S = copy( Sub, 1, Start - $Origin ) ;
                      while( strlen( $S ) < Start - $Origin )
                      {
                          $S = $S.' ' ; // Space-fill string to fill rest of range
                      }
                      $Res = $Res.$S ;
                      exit ;
                  }
            case '\\' : {
                      $Origin = Start - 1 ;
                      while(
                             ( copy( Format, Start, 1 ) != ' ' )
                             &&
                             ( copy( Format, Start, 1 ) != '\\' )
                           )
                      {
                          inc( Start ) ;
                      }
                      $S = copy( Sub, 1, Start - $Origin ) ;
                      while( strlen( $S ) < Start - $Origin )
                      {
                          $S = $S.' ' ; // Space-fill string to fill rest of range
                      }
                      $Res = $Res.$S ;
                      exit ;
                  }
            case '#' : {
                      $S = Get_Numeric_Format( False ) ;
                      $Res = $Res.Format_Number( $S ) ;
                      exit ;
                  }
            case '.' : {
                      $S = Get_Numeric_Format( True ) ;
                      $Res = $Res.Format_Number( $S ) ;
                      exit ;
                  }
            case '-' : {
                      $S = Get_Numeric_Format( False ) ;
                      $Res = $Res.Format_Number( $S ) ;
                      exit ;
                  }
            case '+' : {
                      $S = Get_Numeric_Format( False ) ;
                      $Res = $Res.Format_Number( $S ) ;
                      exit ;
                  }
			default: $Res = $Res + $C ;
        }
        return $Res ;
    }
} // Format_String


function Repeating_String( $Value, $Len )
{
    Set_strlen( $S, $Len ) ;
    fillchar( $S[ 1 ], strlen( $S ), ord( $Value ) ) ;
    return $S ;
}


function Reverse_String( $Value )
{
    $L = 1 ;
    $H = strlen( $Value ) ;
    while( $L < $H )
    {
        $LC = $Value[ $L ] ;
        $HC = $Value[ $H ] ;
        $Value[ $L ] = $HC ;
        $Value[ $H ] = $LC ;
        inc( $L ) ;
        dec( $H ) ;
    }
    return $Value ;
}


function Right( $C, $Position )
{
    Fright( $A, $C, $Position ) ;
    return $A ;
}


function Replace( $A, $Substring, $Substitute, $Options )
{
    if( $Substring == $Substitute ) // Nop
    {
        return $A ;
    }
    $Dummy = pos( $Substring, $A ) ;
    while( $Dummy > 0 )
    {
        $A = copy( $A, 1, $Dummy - 1 ).$Substitute.copy( $A, $Dummy + strlen( $Substring ), strlen( A ) ) ;
        $Dummy = $Dummy + strlen( $Substitute ) ;
        $Dummy = instr( $Dummy, $A, $Substring ) ;
    }
    return $A ;
}


function String_To_Boolean( $S )
{
    if( strlen( $S ) == 0 )
    {
	    return False ;
    }
    $S = upcase( $S[ 1 ] ) ;
    if( ( $S == 'T' ) || ( $S == '1' ) || ( $S == 'Y' ) || ( $S == 'X' ) )
    {
	    return True ;
    }
    return False ;
}


function String_To_Float( $S )
{
    $Dummy = pos( '.', $S ) ;
    if( $Dummy > 0 )
    {
        $S = copy( $S, 1, $Dummy - 1 ).
            OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal.
            copy( $S, $Dummy + 1, strlen( $S ) ) ;
    }
    val( $S, $Result, $Dummy ) ;
	return $Result;
}


/* Translate A based on B.  Returns a string which is A, but with each character
  converted to characters in B based on ASCII values.  For example, if A[ 1 ]
  is "A", then A[ 1 ] is set to B[ 65 ].
*/
function Substitute( $A, $B )
{
    for $Dummy = 1 to strlen( $A ) do
    {
        $I = ord( $A[ $Dummy ] ) + 1 ;
		if( $I <= strlen( $B ) )
        {
            $A[ $Dummy ] = $B[ $I ] ;
        }
    }
    return $A ;
} // Substitute


function Tab_To_Space( $S, $Tab_Stops )
{
    for $Loop = 1 to strlen( $S ) do
    {
        if( $S[ $Loop ] == HT )
        {
            $S[ $Loop ] = ' ' ;
            $Dummy = $Loop + 1 ;
            while( ( $Dummy mod $Tab_Stops ) != 0 )
            {
                insert( ' ', $S, $Dummy ) ;
                inc( $Dummy ) ;
            }
        }
    }
    return $S ;
}


function Remove_Termination( $S, $T )
{
    if( copy( $S, strlen( $S ) - strlen( $T ) + 1, strlen( $T ) ) == $T )
    {
        return copy( $S, 1, strlen( $S ) - strlen( $T ) ) ;
    }else
    {
        return $S ;
    }
}


function Ensure_Termination( $S, $T )
{
    if( copy( $S, strlen( $S ) - strlen( $T ) + 1, strlen( $T ) ) == $T )
    {
        return $S ;
    }
    else
    {
        return $S.$T ;
    }
}


function Trim_Quotes( $S )
{
    if( ( copy( $S, 1, 1 ) == '"' ) or ( copy( $S, 1, 1 ) == #39 ) ) // A quoted string
    {
        if( copy( $S, strlen( $S ), 1 ) == copy( $S, 1, 1 ) )
        {
            $S = copy( $S, 1, strlen( $S ) - 1 ) ; // Trim trailing quote
        }
        $S = copy( $S, 2, strlen( $S ) ) ; // Trim leading quote
    }
    return $S ;
}


function Try_String_To_Float( $S ; &$E )
{
    $Dummy = pos( '.', $S ) ;
    if( $Dummy > 0 )
    {
        $S = copy( $S, 1, $Dummy - 1 ).
            OS^.Text_Presentation_Manager^.Numeric_Text_Presentation_Manager^.Decimal.
            copy( $S, $Dummy + 1, strlen( $S ) ) ;
    }
    val( $S, E, $Dummy ) ;
    return ( $Dummy == 0 ) ;
}


function BCD( $C )
{
    switch( $C )
	{
        case '-' : return 13 ; // other valid == 11
        case '+' : return 12 ; // other valid == 10,14,15
        case '0'..'9' : return ord( $C ) - 48 ;
        else return 15 ;
    }
}


function InitCap( $S )
{
    $S = Edit( $S, 512 ) ;
    $S[ 1 ] = upcase( $S[ 1 ] ) ;
    for $Loop = 2 to strlen( $S ) do
    {
        if( pos( $S[ $Loop - 1 ], ' (/-";:!.,?&+=|//[' ) > 0 )
        {
            $S[ $Loop ] = upcase( $S[ $Loop ] ) ;
        }
    }
    return $S ;
}


/* Return position of substring Str in Haystack, starting at position.  Options
  indicate the following:
        and 256 == ignore text within quotes (")
  Direction indicates left-to-right (1) or right-to-left (-1) scanning of Haystack.
*/
function Position( Start : integer ; const Haystack, Str : string ;
    Options : integer == 0 ; Direction : integer == 1 )

var Quote : char ;
    Last : integer ;
{
    Quote = ' ' ; // Not inside a quote
    if( Direction > 0 )
    {
        Last = strlen( Haystack ) - strlen( Str ) + 1
    }
    else
    {
        Last = strlen( Haystack ) ;
    }
    while( ( Start > 0 ) and ( Start <= Last ) )
    {
        if( Haystack[ Start ] == '"' )
        {
            if( ( Options & 256 ) != 0 ) // Ignore within quotes
            {
                if( Quote == '"' )
                {
                    Quote = ' ' ; // No longer within quote
                }
                else
                {
                    Quote = '"' ;
                }
            }
        }
        if( ( Quote == ' ' ) && ( copy( Haystack, Start, strlen( Str ) ) == Str ) )
        {
            return Start ; // Found
        }
        Start = Start + Direction ;
    } // while
    return 0;
} // Position


function String_To_BCD( $S )
{
    // Setup
    $_Res = '' ;

    // Convert
    for $Dummy = 1 to strlen( $S ) do
    {
        $_Res = $_Res.chr( BCD( $S[ $Dummy ] ) ) ;
    }
    return $_Res ;
}


function String_To_Packed_BCD( $S )
{
    // Setup...
    $_Res = '' ;
    $Current = 0 ;

    // Make sure string is even bytes in length...
    if( odd( strlen( $S ) ) )
    {
        if( ( copy( $S, 1, 1 ) == '-' ) || ( copy( $S, 1, 1 ) == '+' ) )
        {
            $S = copy( $S, 1, 1 ).'0'.copy( $S, 2, strlen( $S ) ) ;
        }
        else
        {
            $S = '0'.$S ;
        }
    }

    // Convert
    for $Dummy = 1 to strlen( $S ) do
    {
        if( odd( $Dummy ) )
        {
            $Current = BCD( $S[ $Dummy ] ) shl 4 ;
        }
        else
        {
            $Current = $Current & BCD( $S[ $Dummy ] ) ;
            $_Res = $_Res.chr( $Current ) ;
        }
    }
    return $_Res ;
} // String_To_Packed_BCD
?>
