{
        Program Name : Unicode
        Package Name : Subroutine Library
        Purpose      : Unicode support routines
        Institution  : Conroy & Conroy Co.
        Date Written : 2-Jan-2016
        Written By   : Alan Conroy
        Version      : 1.0

    	Copyright (C) 2016 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Unicode support...
}

unit Unicode ;

interface

type TUnicode_String = class
{$IFDEF CC_Track_Resources}
                           public // Constructors and destructors...
                               constructor Create ;
                               destructor Destroy ; override ;
{$ENDIF}

                           private // Instance data...
                               Has_Asterisk : boolean ;
                               Contents : array[ 0..384 ] of cardinal ;

                           protected // Property handlers...
                               // Return length of our contents...
                               function Get_Length : integer ;
                               procedure Set_Length( Value : integer ) ;

                           public // API...
                               function As_String : string ;

                               // Assign our contents from a UTF8 string...
                               procedure Assign_From_String( const S : string ;
                                   Format : integer ) ;

                               { Return true if our substring matches the passed
                                 substring.  W_Start indicates the start
                                 position and _Length is the length of the
                                 substring.  Name_Start indicates the start in the
                                 compared string }
                               function Compare( Wildcard_Start, _Length : integer ;
                                   Match : TUnicode_String ;
                                   Match_Start : integer ) : boolean ;

                               // Create...
                               function Copy( Start, Len : integer ) : TUnicode_String ;

                               // Return true if our contents are equal to the match
                               function Equal( Match : TUnicode_String ) : boolean ;

                               // Insert character at given position
                               procedure Insert( Position : integer ;
                                   Value : cardinal ) ;

                               // Convert our characters to lowercase...
                               procedure Lowercase ;

                               // Position of substring...
                               function Pos( const Value : string ;
                                   Start : integer = 1 ) : integer ; overload ;

                               function Pos( const Value : TUnicode_String ;
                                   Start : integer = 1 ) : integer ; overload ;

                               // Return rightmost instance of Value
                               function RPos( Value : char ) : integer ;

                           public // Properties...
                               property Length : integer
                                   read Get_Length
                                   write Set_Length ;
                       end ; // TUnicode_String


{ Determine if Name matches the wildcard specification in Spec.  The strings
  are treated as Format, as follows:
      1 = ASCII
      2 = Unicode-16
      3 = Unicode UTF8
      4 = Unicode-32
}
function WC_Match( Spec, Name : string ; Format : integer ) : boolean ;
function Create_WC_Context( Spec : string ; Format : integer ;
    var Extension : TUnicode_String ) : TUnicode_String ;
function Context_WC_Match( _Spec, _Spec_Extension : TUnicode_String ;
    Name : string ; Format : integer ) : boolean ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}


implementation

uses CommonUt // Edit
     ,_Unicode // Foldings
{$IFDEF CC_Track_Resources}
     ,ResMon
{$ENDIF}
{$IFDEF CC_Regression_Test}
     ,sysutils
     ,strutils
     ,windows
     ,dialogs
{$ENDIF}
     ;

function Wildcard_Match( Wildcard, Name : TUnicode_String ) : boolean ;

var Dummy, Dummy1 : integer ;
    Wildcard_Start, Wildcard_End, Name_Start{, Name_End} : integer ;
    S : TUnicode_String ;

begin
    // Quick check and setup...
    if( Wildcard.As_String = '*' ) then
    begin
        Result := True ;
        exit ;
    end ;

    // Non-wildcard check...
    if( not Wildcard.Has_Asterisk ) then
    begin
        Result := Wildcard.Equal( Name ) ;
        exit ;
    end ;
    Result := False ; // Assume failure

    // Do wildcard match...
    Wildcard_Start := 1 ;
    Name_Start := 1 ;
    Wildcard_End := Wildcard.Length ;
//    Name_End := Name.Length ;

    // Check prefix before first wildcard...
    Dummy := Wildcard.Pos( '*' ) ;
    if( Dummy > Wildcard_Start ) then // Something before the asterisk
    begin
        if( not Wildcard.Compare( Wildcard_Start, Dummy - Wildcard_Start, Name,
          Name_Start ) ) then
        begin
            exit ;
        end ;
        Wildcard_Start := Dummy ;
        Name_Start := Dummy ;
    end ;

    // Check suffix after last wildcard...
    Dummy := Wildcard.RPos( '*' ) ;
    if( Dummy < Wildcard.Length ) then
    begin
        if( not Wildcard.Compare( Dummy + 1, Wildcard.Length - Dummy, Name,
          Name.Length - ( Wildcard.Length - Dummy ) + 1 ) ) then
        begin
            exit ;
        end ;
        Wildcard_End := Dummy ;
//        Name_End := Name.Length - ( Wildcard.Length - Dummy ) ;
    end ;

    // Check for remaining matches, left-to-right...
    while( Wildcard_Start <= Wildcard_End ) do
    begin
        if( Wildcard_Start >= Wildcard_End ) then
        begin
            break ; // All that's left in the wildcard spec is an asterisk - we match
        end ;
        Dummy := Wildcard.Pos( '*', Wildcard_Start + 1 ) ;
        S := Wildcard.Copy( Wildcard_Start + 1, Dummy - Wildcard_Start - 1 ) ;
        try
            Dummy1 := Name.Pos( S, Name_Start ) ;
        finally
            S.Free ;
        end ;
        if( Dummy1 = 0 ) then
        begin
            exit ; // Not found
        end ;

        // Move past wildcard and matching characters...
        Name_Start := Dummy1 + Dummy - Wildcard_Start - 1 ;
        Wildcard_Start := Dummy ;
    end ; // while

    Result := True ;
end ; // Wildcard_Match



// TUnicode_String methods...

{$IFDEF CC_Track_Resources}
// Constructors and destructors...

constructor TUnicode_String.Create ;

begin
    inherited Create ;

    Add_Resource( RT_Object, integer( self ) ) ;
end ;


destructor TUnicode_String.Destroy ;

begin
    Remove_Resource( RT_Object, integer( self ) ) ;

    inherited Destroy ;
end ;
{$ENDIF}


// API...

function TUnicode_String.As_String : string ;

var Dummy, Loop : integer ;

begin
    System.setlength( Result, Length ) ;
    for Loop := 1 to Length do
    begin
        Dummy := Contents[ Loop ] ;
        if( Dummy > 127 ) then
        begin
            Dummy := Dummy or 128 ;
        end ;
        Result[ Loop ] := chr( Dummy ) ;
    end ;
end ;


procedure TUnicode_String.Assign_From_String( const S : string ;
    Format : integer ) ;

var Index, Size, Mask : integer ;
    Value : cardinal ;

begin
    Index := 1 ; // Index in Spec
    Contents[ 0 ] := 0 ;
    if( Format = 3 ) then // UTF8
    begin
        while( Index <= system.length( S ) ) do
        begin
            Value := 0 ;
            if( S[ Index ] > #$FC ) then
            begin
                Size := 6 ;
                Mask := 1 ;
            end else
            if( S[ Index ] > #$F8 ) then
            begin
                Size := 5 ;
                Mask := 3 ;
            end else
            if( S[ Index ] > #$F0 ) then
            begin
                Size := 4 ;
                Mask := 7 ;
            end else
            if( S[ Index ] > #$E0 ) then
            begin
                Size := 3 ;
                Mask := $F ;
            end else
            if( S[ Index ] > #$C0 ) then
            begin
                Size := 2 ;
                Mask := $1F ;
            end else
            begin
                Size := 1 ;
                Mask := $7F ;
            end ;
            while( Size > 0 ) do
            begin
                dec( Size ) ;
                Value := Value or ( ord( S[ Index ] ) and Mask ) ;
                if( Size > 0 ) then
                begin
                    Value := Value shl 6 ;
                end ;
                Mask := $3F ;
                inc( Index ) ;
            end ;
            inc( Contents[ 0 ] ) ;
            Contents[ Contents[ 0 ] ] := Value ;
        end ; // while( Index < system.length( S ) )
    end else
    begin
        Value := 0 ;
        Index := 1 ; // Index in S
        while( Index <= system.length( S ) ) do
        begin
            move( PChar( S )[ Index - 1 ], Value, Format ) ;
            Index := Index + Format ;
            inc( Contents[ 0 ] ) ;
            Contents[ Contents[ 0 ] ] := Value ;
        end ;
    end ; // if( Format = 3 )
end ; // TUnicode_String.Assign_From_String


function TUnicode_String.Get_Length : integer ;

begin
    Result := Contents[ 0 ] ;
end ;


procedure TUnicode_String.Set_Length( Value : Integer ) ;

begin
    Contents[ 0 ] := Value ;
end ;


function lowcase( V : cardinal ; var _Folding_Index : integer ) : cardinal ;

var L, H, Index : integer ;

begin
    Result := V ;
    _Folding_Index := -1 ; // Indicates no translation
    if( ( V < $41 ) or ( V > $118BF ) ) then // Not within range of our table
    begin
        exit ;
    end ;
    L := 0 ;
    H := high( Foldings ) ;
    while( L < H ) do
    begin
        Index := L + ( ( H - L ) div 2 ) ;
        if( V = Foldings[ Index, 0 ] ) then
        begin
            if( Foldings[ Index, 2 ] > 0 ) then // Multiple outputs
            begin
                Result := 0 ;
                _Folding_Index := Index ;
                exit ;
            end ;
            Result := Foldings[ Index, 1 ] ;
            exit ;
        end else
        if( V < Foldings[ Index, 0 ] ) then
        begin
            H := Index ;
        end else
        begin
            L := Index + 1 ;
        end ;
        if( L >= H ) then
        begin
            exit ;
        end ;
    end ; // while( L < H )
end ; // lowcase


function TUnicode_String.Copy( Start, Len : integer ) : TUnicode_String ;

begin
    // Setup...
    Result := TUnicode_String.Create ;
    if( Start > Length ) then
    begin
        exit ;
    end ;
    if( Start < 1 ) then
    begin
        Start := 1 ;
    end ;
    if( Start + Len - 1 > Length ) then
    begin
        Len := Length - Start + 1 ;
    end ;

    Result.Length := Len ;
    move( Contents[ Start ], Result.Contents[ 1 ], Len * sizeof( cardinal ) ) ;
end ; // TUnicode_String.Copy


function TUnicode_String.Compare( Wildcard_Start, _Length : integer ;
    Match : TUnicode_String ; Match_Start : integer ) : boolean ;

var Loop : integer ;

begin
    // Setup...
    Result := False ;
    if( _Length < 1 ) then
    begin
        exit ;
    end ;
    if( Wildcard_Start < 1 ) then
    begin
        Wildcard_Start := 1 ;
    end ;
    if( Match_Start < 1 ) then
    begin
        Match_Start := 1 ;
    end ;
    if( Wildcard_Start + _Length > Length + 1 ) then
    begin
        _Length := Length - Wildcard_Start + 1 ;
    end ;
    if( Match_Start + _Length > Match.Length + 1 ) then
    begin
        _Length := Match.Length - Match_Start + 1 ;
    end ;

    // Do comparison...
    for Loop := 0 to _Length - 1 do
    begin
        if( Contents[ Wildcard_Start + Loop ] <>
            Match.Contents[ Match_Start + Loop ] ) then
        begin
            if(
                ( Contents[ Wildcard_Start + Loop ] <> ord( '?' ) )
                and
                ( Match.Contents[ Wildcard_Start + Loop ] <> ord( '?' ) )
              ) then
            begin
                exit ;
            end ;
        end ;
    end ;
    Result := True ;
end ; // TUnicode_String.Compare


function TUnicode_String.Equal( Match : TUnicode_String ) : boolean ;

var Loop : integer ;

begin
    Result := False ;
    if( Length <> Match.Length ) then
    begin
        exit ;
    end ;
    for Loop := 1 to Length do
    begin
        if( ( Contents[ Loop ] <> Match.Contents[ Loop ] ) ) then
        begin
            if(
                ( Contents[ Loop ] <> ord( '?' ) )
                and
                ( Match.Contents[ Loop ] <> ord( '?' ) )
              ) then
            begin
                exit ;
            end ;
        end ;
    end ;
    Result := True ;
end ; // TUnicode_String.Equal


procedure TUnicode_String.Insert( Position : integer ; Value : cardinal ) ;

begin
    move( Contents[ Position ], Contents[ Position + 1 ], sizeof( Contents ) - sizeof( cardinal ) * ( Position + 1 ) ) ;
    Contents[ Position ] := Value ;
    inc( Contents[ 0 ] ) ;
end ;


procedure TUnicode_String.Lowercase ;

var Dummy, V : integer ;
    _Folding_Index : integer ;

begin
    Dummy := 1 ;
    while( Dummy <= Length ) do
    begin
        V := lowcase( Contents[ Dummy ], _Folding_Index ) ;
        if( ( V = 0 ) and ( _Folding_Index >= 0 ) ) then
        begin
            Contents[ Dummy ] := Foldings[ _Folding_Index, 1 ] ;
            for V := 2 to 3 do
            begin
                if( Foldings[ _Folding_Index, V ] <> 0 ) then
                begin
                    Insert( Dummy + 1, Foldings[ _Folding_Index, V ] ) ;
                    inc( Dummy ) ;
                end ;
            end ;
        end else
        begin
            Contents[ Dummy ] := V ;
        end ;
        inc( Dummy ) ;
    end ;
end ; // TUnicode_String.Lowercase


function TUnicode_String.Pos( const Value : TUnicode_String ;
    Start : integer = 1 ) : integer ;

var Dummy, Dummy1 : integer ;
    Found : boolean ;

begin
    Result := 0 ;
    if( Start > Length ) then
    begin
        exit ;
    end ;
    if( Value.Length > Length - Start + 1 ) then
    begin
        exit ; // Substring is longer than our contents
    end ;
    for Dummy := Start to Length - Value.Length + 1 do
    begin
        Found := True ;
        for Dummy1 := 1 to Value.Length do
        begin
            if(
                ( Value.Contents[ Dummy1 ] <> Contents[ Dummy1 + Dummy - 1 ] )
                and
                ( Contents[ Dummy1 + Dummy - 1 ] <> ord( '?' ) )
                and
                ( Value.Contents[ Dummy1 ] <> ord( '?' ) )
              ) then
            begin
                Found := False ;
                break ;
            end ;
        end ; // for Dummy1
        if( Found ) then
        begin
            Result := Dummy ;
            exit ;
        end ;
    end ; // for Dummy
end ; // TUnicode_String.Pos


function TUnicode_String.Pos( const Value : string ;
    Start : integer = 1 ) : integer ;

var Dummy, Dummy1 : integer ;
    Found : boolean ;

begin
    Result := 0 ;
    if( Start > Length ) then
    begin
        exit ;
    end ;
    if( System.Length( Value ) > Length - Start + 1 ) then
    begin
        exit ; // Substring is longer than our contents
    end ;
    for Dummy := Start to Length - system.length( Value ) + 1 do
    begin
        Found := True ;
        for Dummy1 := 1 to system.length( Value ) do
        begin
            if( ord( Value[ Dummy1 ] ) <> Contents[ Dummy1 + Dummy - 1 ] ) then
            begin
                Found := False ;
                break ;
            end ;
        end ; // for Dummy1
        if( Found ) then
        begin
            Result := Dummy ;
            exit ;
        end ;
    end ; // for Dummy
end ; // TUnicode_String.Pos


function TUnicode_String.RPos( Value : char ) : integer ;

var Loop, V : cardinal ;

begin
    V := ord( Value ) ;
    for Loop := Length downto 1 do
    begin
        if( Contents[ Loop ] = V ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
    Result := 0 ;
end ;


function Context_WC_Match( _Spec, _Spec_Extension : TUnicode_String ;
    Name : string ; Format : integer ) : boolean ;

var Dummy : integer ;
    _Name, _Name_Extension : TUnicode_String ;

begin
    // Setup...
    Result := False ; // Assume no match
    if( ( Format < 1 ) or ( Format > 4 ) ) then
    begin
        Format := 1 ;
    end ;
    Name := Edit( Name, 16 ) ;

    // Create the Unicode string instances...
    _Name := TUnicode_String.Create ;
    try
        _Name.Assign_From_String( Name, Format ) ;

        // Normalize to lowercase...
        _Name.Lowercase ;

        // Parse out the extensions...
        Dummy := _Name.RPos( '.' ) ;
        if( Dummy = 0 ) then
        begin
            Dummy := _Name.Length + 1 ;
        end ;
        _Name_Extension := _Name.Copy( Dummy + 1, 1024 ) ;
        try
            _Name.Length := Dummy - 1 ;
            _Name.Has_Asterisk := ( _Name.Pos( '*' ) > 0 ) ;
            _Name_Extension.Has_Asterisk := ( _Name_Extension.Pos( '*' ) > 0 ) ;

            // Check for extension...
            if( not Wildcard_Match( _Spec_Extension, _Name_Extension ) ) then
            begin
                exit ; // Extension doesn't match - exit with false
            end ;
            Result := Wildcard_Match( _Spec, _Name ) ;
        finally
            _Name_Extension.Free ;
        end ;
    finally
        _Name.Free ;
    end ;
end ; // Context_WC_Match


function Create_WC_Context( Spec : string ; Format : integer ;
    var Extension : TUnicode_String ) : TUnicode_String ;

var Dummy : integer ;

begin
    // Setup...
    if( ( Format < 1 ) or ( Format > 4 ) ) then
    begin
        Format := 1 ;
    end ;
    Spec := Edit( Spec, 16 ) ;

    // Pre-normalize the specification...
    Dummy := pos( '**', Spec ) ;
    while( Dummy > 0 ) do
    begin
        delete( Spec, Dummy, 1 ) ;
        Dummy := pos( '**', Spec ) ;
    end ;
    Dummy := pos( '*?', Spec ) ;
    while( Dummy > 0 ) do
    begin
        Spec := copy( Spec, 1, Dummy - 1 ) + '?*' +
            copy( Spec, Dummy + 2, length( Spec ) ) ;
        Dummy := pos( '*?', Spec ) ;
    end ;
    if( pos( '.', Spec ) = 0 ) then // No explicit extension
    begin
        Spec := Spec + '.*' ;
    end ;

    // Create the Unicode string instances...
    Result := TUnicode_String.Create ;
    Result.Assign_From_String( Spec, Format ) ;

    // Normalize to lowercase...
    Result.Lowercase ;

    // Parse out the extensions...
    Dummy := Result.RPos( '.' ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := Result.Length + 1 ;
    end ;
    Extension := Result.Copy( Dummy + 1, 1024 ) ;
    Result.Length := Dummy - 1 ;
    Result.Has_Asterisk := ( Result.Pos( '*' ) > 0 ) ;
    Extension.Has_Asterisk := ( Extension.Pos( '*' ) > 0 ) ;
end ; // Create_WC_Context


function WC_Match( Spec, Name : string ; Format : integer ) : boolean ;

var _Spec, _Spec_Extension : TUnicode_String ;
    S : string ;

begin
    _Spec := Create_WC_Context( Spec, Format, _Spec_Extension ) ;
    try
        S := _Spec.As_String ;
        if( ( S = '*' ) or ( S = '*.*' ) ) then
        begin
            Result := True ; // Wildcard matches anything/everything
            exit ;
        end ;
        Result := Context_WC_Match( _Spec, _Spec_Extension, Name, Format ) ;
    finally
        _Spec.Free ;
        _Spec_Extension.Free ;
    end ;
end ; // WC_Match


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    halt ;
end ;


function RPos( const Contents : string ; Value : char ) : integer ;

var Loop, V : cardinal ;

begin
    V := ord( Value ) ;
    for Loop := Length( Contents ) downto 1 do
    begin
        if( Contents[ Loop ] = Value ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
    Result := 0 ;
end ;


function _Pos( const Contents : string ; const Value : string ;
    Start : integer = 1 ) : integer ;

var Dummy, Dummy1 : integer ;
    Found : boolean ;

begin
    Result := 0 ;
    if( Start > Length( Contents ) ) then
    begin
        exit ;
    end ;
    if( length( Value ) > Length( Contents ) - Start + 1 ) then
    begin
        exit ; // Substring is longer than our contents
    end ;
    for Dummy := Start to Length( Contents ) - length( Value ) + 1 do
    begin
        Found := True ;
        for Dummy1 := 1 to length( Value ) do
        begin
            if(
                ( Value[ Dummy1 ] <> Contents[ Dummy1 + Dummy - 1 ] )
                and
                ( Contents[ Dummy1 + Dummy - 1 ] <> '?' )
                and
                ( Value[ Dummy1 ] <> '?' )
              ) then
            begin
                Found := False ;
                break ;
            end ;
        end ; // for Dummy1
        if( Found ) then
        begin
            Result := Dummy ;
            exit ;
        end ;
    end ; // for Dummy
end ; // TUnicode_String.Pos


function _Equal( const Contents, Match : string ) : boolean ;

var Loop : integer ;

begin
    Result := False ;
    if( Length( Contents ) <> length( Match ) ) then
    begin
        exit ;
    end ;
    for Loop := 1 to Length( Contents ) do
    begin
        if( ( Contents[ Loop ] <> Match[ Loop ] ) ) then
        begin
            if( ( Contents[ Loop ] <> '?' ) and ( Match[ Loop ] <> '?' ) ) then
            begin
                exit ;
            end ;
        end ;
    end ;
    Result := True ;
end ; // TUnicode_String.Equal


function _Compare( const Contents : string ; Wildcard_Start, _Length : integer ;
    const Match : string ; Name_Start : integer ; Wildcard : boolean ) : boolean ;

var Loop : integer ;

begin
    // Setup...
    Result := False ;
    if( _Length < 1 ) then
    begin
        exit ;
    end ;
    if( Wildcard_Start < 1 ) then
    begin
        Wildcard_Start := 1 ;
    end ;
    if( Name_Start < 1 ) then
    begin
        Name_Start := 1 ;
    end ;
    if( Wildcard_Start + _Length > Length( Contents ) + 1 ) then
    begin
        _Length := Length( Contents ) - Wildcard_Start + 1 ;
    end ;

    // Do comparison...
    for Loop := 0 to _Length - 1 do
    begin
        if( Name_Start + Loop > length( Match ) ) then
        begin
            exit ;
        end ;
        if( Contents[ Wildcard_Start + Loop ] <> Match[ Name_Start + Loop ] ) then
        begin
            if( ( not Wildcard ) or ( Contents[ Wildcard_Start + Loop ] <> '?' ) ) then
            begin
                exit ;
            end ;
        end ;
    end ;
    Result := True ;
end ; // _Compare


function _Wildcard_Match( const Wildcard, Name : string ;
    Wildcard_has_asterisk : boolean ) : boolean ;

var Dummy, Dummy1 : integer ;
    Wildcard_Start, Wildcard_End, Name_Start, Name_End : integer ;
    S : string ;

begin
    // Quick check and setup...
    if( Wildcard = '*' ) then
    begin
        Result := True ;
        exit ;
    end ;

    // Non-wildcard check...
    if( not Wildcard_Has_Asterisk ) then
    begin
        Result := _Equal( Wildcard, Name ) ;
        exit ;
    end ;
    Result := False ; // Assume failure

    // Do wildcard match...
    Wildcard_Start := 1 ;
    Name_Start := 1 ;
    Wildcard_End := length( Wildcard ) ;
    Name_End := length( Name ) ;

    // Check prefix before first wildcard...
    Dummy := pos( '*', Wildcard ) ;
    if( Dummy > Wildcard_Start ) then // Something before the asterisk
    begin
        if( not _Compare( Wildcard, Wildcard_Start, Dummy - Wildcard_Start, Name, Name_Start, True ) ) then
        begin
            exit ;
        end ;
        Wildcard_Start := Dummy ;
        Name_Start := Dummy ;
    end ; // if( Dummy > Wildcard_Start )

    // Check suffix after last wildcard...
    Dummy := RPos( Wildcard, '*' ) ;
    if( Dummy < length( Wildcard ) ) then
    begin
        if( not _Compare( Wildcard, Dummy + 1, length( Wildcard ) - Dummy, Name,
	  length( Name ) - ( length( Wildcard ) - Dummy ) + 1, True ) ) then
        begin
            exit ;
        end ;
        Wildcard_End := Dummy ;
        Name_End := length( Name ) - ( length( Wildcard ) - Dummy ) ;
    end ; // if( Dummy < length( Wildcard ) )

    // Check for remaining matches, left-to-right...
    while( Wildcard_Start <= Wildcard_End ) do
    begin
        if( Wildcard_Start >= Wildcard_End ) then
        begin
            break ; // All that's left in the wildcard spec is an asterisk - we match
        end ;
        Dummy := PosEx( '*', Wildcard, Wildcard_Start + 1 ) ;
        S := Copy( Wildcard, Wildcard_Start + 1, Dummy - Wildcard_Start - 1 ) ;
        Dummy1 := _Pos( Name, S, Name_Start ) ;
        if( Dummy1 = 0 ) then
        begin
            exit ; // Not found
        end ;

        // Move past wildcard and matching characters...
        Name_Start := Dummy1 + Dummy - Wildcard_Start - 1 ;
        Wildcard_Start := Dummy ; // Move past wildcard and matching characters
    end ; // while

    Result := True ;
end ; // _Wildcard_Match


function _WC_Match( Spec, Name : string ; Format : integer ) : boolean ;

var Dummy : integer ;
    _Spec, _Name, _Spec_Extension, _Name_Extension : string ;
    _Name_Has_Asterisk, _Name_Extension_Has_Asterisk : boolean ;

begin
    // Setup...
    Result := False ; // Assume no match
    Spec := Edit( Spec, 16 ) ;
    Name := Edit( Name, 16 ) ;

    // Pre-normalize the specification...
    Dummy := pos( '**', Spec ) ;
    while( Dummy > 0 ) do
    begin
        delete( Spec, Dummy, 1 ) ;
        Dummy := pos( '**', Spec ) ;
    end ;
    if( ( Spec = '*' ) or ( Spec = '*.*' ) ) then
    begin
        Result := True ; // Wildcard matches anything/everything
        exit ;
    end ;
    Dummy := pos( '*?', Spec ) ;
    while( Dummy > 0 ) do
    begin
        Spec := copy( Spec, 1, Dummy - 1 ) + '?*' + copy( Spec, Dummy + 2, length( Spec ) ) ;
        Dummy := pos( '*?', Spec ) ;
    end ;
    if( pos( '.', Spec ) = 0 ) then // No explicit extension
    begin
        Spec := Spec + '.*' ;
    end ;

    _Spec := Spec ;
    _Name := Name ;

    // Normalize to lowercase...
    _Spec := lowercase( _Spec ) ;
    _Name := Lowercase( _Name ) ;

    // Parse out the extensions...
    Dummy := RPos( _Spec, '.' ) ;
    _Spec_Extension := Copy( _Spec, Dummy + 1, 1024 ) ;
    set_length( _Spec, Dummy - 1 ) ;
    Dummy := RPos( Name, '.' ) ;
    if( Dummy = 0 ) then
    begin
        Dummy := length( _Name ) + 1 ;
    end ;
    _Name_Extension := Copy( Name, Dummy + 1, 1024 ) ;
    setlength( _Name, Dummy - 1 ) ;
    _Name_Has_Asterisk := pos( '*', _Name ) > 0 ;
    _Name_Extension_Has_Asterisk := ( pos( '*', _Name_Extension ) > 0 ) ;

    // Check for extension...
    if( not _Wildcard_Match( _Spec_Extension, _Name_Extension, pos( '*', _Spec_Extension ) > 0 ) ) then
    begin
        exit ; // Extension doesn't match - exit with false
    end ;
    Result := _Wildcard_Match( _Spec, _Name, pos( '*', _Spec ) > 0 ) ;
end ; // _WC_Match



procedure Test_TUnicode_String ;

var S : string[ 255 ] ;
    X, Y : TUnicode_String ;

begin
    X := TUnicode_String.Create ;
    if( X.As_String <> '' ) then
    begin
        Error ;
    end ;
    if( X.Length <> 0 ) then
    begin
        Error ;
    end ;
    X.Assign_From_String( 'abcdef', 1 ) ;
    if( X.As_String <> 'abcdef' ) then
    begin
        Error ;
    end ;
    if( X.Length <> 6 ) then
    begin
        Error ;
    end ;
    X.Length := 3 ;
    if( X.Length <> 3 ) then
    begin
        Error ;
    end ;
    if( X.As_String <> 'abc' ) then
    begin
        Error ;
    end ;
    X.Assign_From_String( 'ABCDEF!b', 1 ) ;
    X.Lowercase ;
    if( X.As_String <> 'abcdef!b' ) then
    begin
        Error ;
    end ;
    Y := X.Copy( 2, 3 ) ;
    if( Y.As_String <> 'bcd' ) then
    begin
        Error ;
    end ;
    if( X.Equal( Y ) ) then
    begin
        Error ;
    end ;
    Y.Assign_From_String( X.As_String, 1 ) ;
    if( not X.Equal( Y ) ) then
    begin
        Error ;
    end ;
    if( X.RPos( 'X' ) <> 0 ) then
    begin
        Error ;
    end ;
    if( X.RPos( 'b' ) <> 8 ) then
    begin
        Error ;
    end ;
    if( X.Pos( 'X' ) <> 0 ) then
    begin
        Error ;
    end ;
    if( X.Pos( 'b' ) <> 2 ) then
    begin
        Error ;
    end ;
    if( X.Pos( 'b', 10 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( X.Pos( 'bbbbbbbbbbb' ) <> 0 ) then
    begin
        Error ;
    end ;
    if( X.Pos( X ) <> 1 ) then
    begin
        Error ;
    end ;
    if( not X.Compare( 1, 5, Y, 1 ) ) then
    begin
        Error ;
    end ;
    if( not X.Compare( 1, 5, Y, 1 ) ) then
    begin
        Error ;
    end ;
    Y.Assign_From_String( 'a?cde?!b', 1 ) ;
    if( not X.Compare( 1, 8, Y, 1 ) ) then
    begin
        Error ;
    end ;
    Y.Assign_From_String( 'a?cde?!x', 1 ) ;
    if( X.Compare( 1, 8, Y, 1 ) ) then
    begin
        Error ;
    end ;
    Y.Free ;
    Y := X.Copy( 100, 2 ) ;
    if( Y.Length <> 0 ) then
    begin
        Error ;
    end ;
    Y.Free ;
    Y := X.Copy( -1, 100 ) ;
    if( Y.As_String <> 'abcdef!b' ) then
    begin
        Error ;
    end ;
    S := #1#0#0#0#$DF#0#0#0 ;
    X.Assign_From_String( S, 4 ) ;
    if( X.Length <> 2 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> $DF ) then
    begin
        Error ;
    end ;
    X.Lowercase ;
    if( X.Length <> 3 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> $73 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 3 ] <> $73 ) then
    begin
        Error ;
    end ;
    S := #1#0#$DF#0 ;
    X.Assign_From_String( S, 2 ) ;
    if( X.Length <> 2 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> 1 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> $DF ) then
    begin
        Error ;
    end ;
    S := #$90#3#1#0 ;
    X.Assign_From_String( S, 2 ) ;
    if( X.Length <> 2 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> $390 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> 1 ) then
    begin
        Error ;
    end ;
    X.Lowercase ;
    if( X.Length <> 4 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> $3B9 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> $308 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 3 ] <> $301 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 4 ] <> 1 ) then
    begin
        Error ;
    end ;
    S := X.As_String ;
    if( S <> #185#136#129#1 ) then
    begin
        Error ;
    end ;
    if( X.Compare( 1, -1, Y, 1 ) ) then
    begin
        Error ;
    end ;
    if( X.Compare( -1, 5, Y, -1 ) ) then
    begin
        Error ;
    end ;
    Y.Length := X.Length ;
    if( X.Equal( Y ) ) then
    begin
        Error ;
    end ;
    Y.Assign_From_String( '??????????', 1 ) ;
    Y.Length := X.Length ;
    if( not X.Equal( Y ) ) then
    begin
        Error ;
    end ;
    if( X.Pos( Y, 100 ) <> 0 ) then
    begin
        Error ;
    end ;
    Y.Assign_From_String( '1234567890', 1 ) ;
    if( X.Pos( Y, 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    Y.Length := 3 ;
    if( X.Pos( Y, 1 ) <> 0 ) then
    begin
        Error ;
    end ;
    if( X.Pos( #1 ) <> 4 ) then
    begin
        Error ;
    end ;
    S := ' ' + #$C1#33 + #$E1#1#34 + #$F1#1#1#35 + #$F9#1#1#1#36 + #$FD#1#1#1#1#37 ;
    X.Assign_From_String( S, 3 ) ;
    if( X.Length <> 6 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 1 ] <> 32 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 2 ] <> 33 or $40 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 3 ] <> 34 or $1040 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 4 ] <> 35 or $41040 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 5 ] <> 36 or $1041040 ) then
    begin
        Error ;
    end ;
    if( X.Contents[ 6 ] <> 37 or $41041040 ) then
    begin
        Error ;
    end ;
    X.Assign_From_String( 'abc'#1, 1 ) ;
    Y.Assign_From_String( '?'#1, 1 ) ;
    if( X.Pos( Y ) <> 3 ) then
    begin
        Error ;
    end ;
    X.Free ;
    Y.Free ;
end ;


procedure Test_Native_String ;

begin
    if( _WC_Match( 'A', 'b', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( 'A', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '*', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '**', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( 'Ab', 'Abb', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( 'A?', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '?b', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '?b.*', 'ab.cd', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( 'b?b.*', 'ab.cd', 0 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( 'a*b', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( 'aa*bb', 'aabb', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( 'a*b*', 'bc', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( 'a*b', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( 'a*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( 'bc.x', 'bc.txt', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '*?b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '*a*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( '*a*b*', 'ac', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( '*b*', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not _WC_Match( '*a?c*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( _WC_Match( '*a?c*', 'ac', 1 ) ) then
    begin
        Error ;
    end ;
end ; // Test_Native_String


procedure Test_Wildcard ;

begin
    if( WC_Match( 'A', 'b', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( 'A', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '*', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '**', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( 'Ab', 'Abb', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( 'A?', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '?b', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '?b.*', 'ab.cd', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( 'b?b.*', 'ab.cd', 0 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( 'a*b', 'ab', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( 'aa*bb', 'aabb', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( 'a*b*', 'bc', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( 'a*b', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( 'a*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( 'bc.x', 'bc.txt', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '*?b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '*a*b*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( '*a*b*', 'ac', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( '*b*', 'a', 1 ) ) then
    begin
        Error ;
    end ;
    if( not WC_Match( '*a?c*', 'abc', 1 ) ) then
    begin
        Error ;
    end ;
    if( WC_Match( '*a?c*', 'ac', 1 ) ) then
    begin
        Error ;
    end ;
end ;


(*
var Loop : integer ;
    _NS, _US, NS, US : cardinal ;
*)

procedure Test_Unit ;

begin
    Test_TUnicode_String ;

    Test_Native_String ;

    Test_Wildcard ;

(*
_NS := gettickcount ;
for Loop := 1 to 100000 do
begin
  Test_Native_String ;
end ;
NS := gettickcount - _NS ;
_US := gettickcount ;
for Loop := 1 to 100000 do
begin
  Test_Wildcard ;
end ;
US := gettickcount - _US ;
showmessage(inttostr(NS));
showmessage(inttostr(US));
*)
end ;
{$ENDIF}


end.

