{
        Program Name : VCL_Std
        Package Name : Subroutine Libraray - Layer 1
        Purpose      : LCL/VCL support
        Institution  : Conroy & Conroy Co.
        Date Written : 25-June-2002
        Written By   : Alan Conroy
        Version      : 1.0A

        Copyright (C) 2002-2008 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

          23-Aug-2006   EAC         Added Gradient_Fill and
                                    Bigradient_Fill.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit contains LCL/VCL-specific support routines.
}

unit VCL_Std ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses // Borland...
     Windows, // TMsg
     Buttons, // TSpeedButton
     Controls, // TWinControl
     Graphics, // TFont
     StdCtrls, // TLabel
     Classes, // TComponent
     Messages, // TMessage
{$IFDEF FPC}
     LCLType, // TMsg
{$ENDIF}

     // Subroutine Library - Layer 0...
     _DebugIn, // PSDebug_Interface

     // Subroutine Library - Layer 1...
     DebugInt ; // TText_Debugger

{ Adjust all children of passed control so that they are evenly spaced across
  the control.  Note: Only horizontal spacing is supported currently. }
procedure Adjust_Button_Panel( C : TWinControl ) ;

// Does a dual gradient fill of a rectangle on a canvas.
procedure Bigradient_Fill( Canvas : TCanvas ; const Rect : TRect ;
    Source, Intermediate, Target : TColor ) ;

procedure Clone_Component( Source, Destination : TComponent ) ;

procedure Clone_Control( Source, Destination : TControl ) ;

procedure Clone_Speed_Button( Source, Destination : TSpeedButton ) ;

// Returns True if the passed control has any children.
function Control_Has_Children( Control : TControl ) : boolean ;

function Convert_Message( const Msg : TMessage ) : Windows.TMsg ;

function Convert_Msg( const Msg : Windows.TMsg ) : TMessage ;

procedure Copy_ListBox_To_Clipboard( LB : TListBox ) ;

procedure Delete_All_Children( Control : TWinControl ) ;

function Get_Sort_Image_List( Owner : TComponent ) : TImageList ;

// Does a gradient fill of a rectangle on a canvas.
procedure Gradient_Fill( Canvas : TCanvas ; const Rect : TRect ;
    Source, Target : TColor ) ;

// Return minimum height required of label to show its text
function Min_Label_Height( L : TLabel ) : integer ;

function Object_Description( O : TObject ) : string ;

function Object_Iterator( O : TObject ) : TCOM_Debug_Interface ;

procedure Remove_Duplicates( L : TStringList ) ;

// Reset pop-up hints.
procedure Reset_Hint ;

function String_To_Color( X : string ) : TColor ;

// Return virtual key code associated with string (0 if no match).
function String_To_VK( X : string ) : integer ;

function VCL_Debugger : PDebug_Interface ;

// Return string indicative of passed virtual key code
function VK_To_String( Code : integer ) : string ;

{ Write contents of Text to the specified handle using the specified font.  X, Y
  indicate starting x/Y offset in Handle's control, and Max_X indicates the
  maximum width to write.  Returns the Maximum width minus the width of the
  output text (ie the amount of Max_X left after the write). }
function Write_Text( Handle : THandle ; Font : TFont ; X, Y, Max_X : integer ;
    Text : string ) : integer ;


type TComponent_Debug_Iterator = class( TText_Debugger )
                                     public // Property handlers...
                                         function Child( Index : integer ) : TCOM_Debug_Interface ;
                                             override ;

                                     public // API...
                                         Component : TComponent ;

                                         function Count : integer ; override ;
                                 end ;

type TControl_Debug_Iterator = class( TComponent_Debug_Iterator )
                                   public // Property handlers...
                                       function Child( Index : integer ) : TCOM_Debug_Interface ;
                                           override ;

                                   public // API...
                                       function Count : integer ; override ;
                               end ;

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses // Borland...
     Clipbrd, // Clipboard
     Forms, // Screen
     ExtCtrls, // TPanel
     SysUtils, // inttostr
     WinAPI, { cl* constants }

     // C&C Subroutine Library - Layer 0...
     _ASCII, // CR

     // C&C Subroutine Library - Layer 1...
     CommonUt ; // TInteger_List

// Debuggers...

// TComponent_Debug_Iterator...

// Property handlers...

function TComponent_Debug_Iterator.Child( Index : integer ) : TCOM_Debug_Interface ; stdcall ;

begin
    Result := TText_Debugger.Create ;
    case Index of
        0 : Result.Title := PAnsiChar( 'ComponentIndex = ' + inttostr( Component.ComponentIndex ) ) ;
        1 : Result.Title := 'Components' ;
        2 : Result.Title := 'ComponentState' ;
        3 : Result.Title := 'ComponentStyle' ;
        4 : Result.Title := PAnsiChar( 'DesignInfo = ' + inttostr( Component.DesignInfo ) ) ;
        5 : Result.Title := PAnsiChar( 'Name = "' + Component.Name + '"' ) ;
        6 : Result.Title := PAnsiChar( 'Owner = ' + Object_Description( Component.Owner ) ) ;
        7 : Result.Title := PAnsiChar( 'Tag = ' + inttostr( Component.Tag ) ) ;
        8 : Result.Title := PAnsiChar( 'VCLComObject = ' + Pointer_To_String( Component.VCLComObject ) ) ;
    end ; // case Index
end ;


// API...

function TComponent_Debug_Iterator.Count : integer ; stdcall ;

begin
    Result := 9 ;
end ;


function TControl_Debug_Iterator.Child( Index : integer ) : PDebug_Interface ; stdcall ;

var Control : TControl ;

begin
    Control := TControl( Component ) ;
    Result := TText_Debugger.Create ;
    case Index of
        0 : Result.Title := PAnsiChar( ' Parent = ' + Object_Description( Control.Parent ) ) ;
        1 : Result.Title := 'Action' ;//Control.Action
        2 : Result.Title := 'ActionLink' ;//TLabel( Control.ActionLink ) ;
        3 : Result.Title := 'Align' ;//Control.Align
        4 : Result.Title := 'Anchors' ;//Control.Anchors
        5 : Result.Title := PAnsiChar( 'AutoSize = ' + Boolean_To_String( TLabel( Control ).AutoSize ) ) ;
        6 : Result.Title := 'BiDiMode' ;//Control.BiDiMode
        7 : Result.Title := 'BoundsRect' ;//Control.BoundsRect
        8 : Result.Title := PAnsiChar( 'Caption = "' + TLabel( Control ).Caption + '"' ) ;
        9 : Result.Title := PAnsiChar( 'ClientHeight = ' + inttostr( Control.ClientHeight ) ) ;
        10 : Result.Title := 'ClientOrigin' ;//Control.ClientOrigin
        11 : Result.Title := 'ClientRect' ;//Control.ClientRect
        12 : Result.Title := PAnsiChar( 'ClientWidth = ' + inttostr( Control.ClientWidth ) ) ;
        13 : Result.Title := 'Color' ;//TLabel( Control ).Color ;
        14 : Result.Title := 'Constraints' ;//Control.Constraints
        15 : Result.Title := 'ControlState' ;//Control.ControlState
        16 : Result.Title := 'ControlStyle' ;//Control.ControlStyle
        17 : Result.Title := 'Cursor' ;//Control.Cursor
        18 : Result.Title := 'DesktopFont' ;//T?( Control ).DesktopFont ;
        19 : Result.Title := 'DockOrientation' ;//Control.DockOrientation
        20 : Result.Title := 'DragCursor' ;//TLabel( Control ).DragCursor ;
        21 : Result.Title := 'DragKind' ;//TLabel( Control ).DragKind ;
        22 : Result.Title := 'DragMode' ;//TLabel( Control ).DragMode ;
        23 : Result.Title := PAnsiChar( 'Enabled = ' + Boolean_To_String( Control.Enabled ) ) ;
        24 : Result.Title := PAnsiChar( 'Floating = ' + Boolean_To_String( Control.Floating ) ) ;
        25 : Result.Title := 'Font' ;//TLabel( Control ).Font ;
        26 : Result.Title := PAnsiChar( 'Height = ' + inttostr( Control.Height ) ) ;
        27 :
{$IFDEF DELPHI5UP}
             Result.Title := PAnsiChar( 'HelpContext = ' + inttostr( Control.HelpContext ) ) ;
{$ELSE}
             Result.Title := PAnsiChar( 'HelpContext = n/a' ) ;
{$ENDIF}
        28 :
{$IFDEF DELPHI5UP}
             Result.Title := PAnsichar( 'HelpKeyword = "' + Control.HelpKeyword + '"' ) ;
{$ELSE}
             Result.Title := PAnsichar( 'HelpKeyword = n/a' ) ;
{$ENDIF}
        29 : Result.Title := 'HelpType' ;//Control.HelpType
        30 : Result.Title := PAnsichar( 'Hint = "' + Control.Hint + '"' ) ;
        31 : Result.Title := 'HostDockSite' ;//Control.HostDockSite
        32 : Result.Title := PAnsichar( 'Left = ' + inttostr( Control.Left ) ) ;
        33 : Result.Title := PAnsichar( 'LRDockWidth = ' + inttostr( Control.LRDockWidth ) ) ;
        34 : Result.Title := 'MouseCapture' ;//+ Boolean_To_String( T?( Control ).MouseCapture ) ) ;
        35 : Result.Title := PAnsichar( 'ParentBiDiMode = ' + Boolean_To_String( TLabel( Control ).ParentBiDiMode ) ) ;
        36 : Result.Title := PAnsichar( 'ParentColor = ' + Boolean_To_String( TLabel( Control ).ParentColor ) ) ;
        37 : Result.Title := PAnsichar( 'ParentFont = ' + Boolean_To_String( Tlabel( Control ).ParentFont ) ) ;
        38 : Result.Title := PAnsichar( 'ParentShowHint = ' + Boolean_To_String( TLabel( Control ).ParentShowHint ) ) ;
        39 : Result.Title := PAnsichar( 'PopupMenu = ' + Object_Description( TEdit( Control ).PopupMenu ) ) ;
        40 : Result.Title := 'ScalingFlags' ;//Control.ScalingFlags
        41 : Result.Title := PAnsichar( 'ShowHint = ' + Boolean_To_String( Control.ShowHint ) ) ;
        42 : Result.Title := PAnsichar( 'TBDockHeight = ' + inttostr( Control.TBDockHeight ) ) ;
        43 : Result.Title := PAnsichar( 'Text = "' + TEdit( Control ).Text + '"' ) ;
        44 : Result.Title := PAnsichar( 'Top = ' + inttostr( Control.Top ) ) ;
        45 : Result.Title := PAnsichar( 'UndockHeight = ' + inttostr( Control.UndockHeight ) ) ;
        46 : Result.Title := PAnsichar( 'UndockWidth = ' + inttostr( Control.UndockWidth ) ) ;
        47 : Result.Title := PAnsichar( 'Visible = ' + Boolean_To_String( Control.Visible ) ) ;
        48 : Result.Title := PAnsichar( 'Width = ' + inttostr( Control.Width ) ) ;
        49 : Result.Title := 'WindowText' ;//+ T?( Control ).WindowText + '"' ) ;
        else
            begin
                Result.Free ;
                Result := inherited Child( Index - 50 ) ;
            end ;
    end ;
end ;


function TControl_Debug_Iterator.Count : integer ; stdcall ;

begin
    Result := 50 + inherited Count ;
end ;


type TChild_Controls_Debug_Iterator = class( TText_Debugger )
                                          protected // Property handlers...
                                              function Child( Index : integer ) : PDebug_Interface ;
                                                  override ;

                                          public // API...
                                              Control : TWinControl ;

                                              function Count : integer ; override ;
                                      end ;

// Property handlers...

function TChild_Controls_Debug_Iterator.Child( Index : integer ) : PDebug_Interface ; stdcall ;

begin
    Result := Object_Iterator( Control.Controls[ Index ] ) ;
    Result.Title := PAnsichar( inttostr( Index ) + ': ' + Object_Description( Control.Controls[ Index ] ) ) ;
end ;


// API...

function TChild_Controls_Debug_Iterator.Count : integer ; stdcall ;

begin
    Result := Control.ControlCount ;
end ;


type TWinControl_Debug_Iterator = class( TControl_Debug_Iterator )
                                   protected // Property handlers...
                                       function Child( Index : integer ) : PDebug_Interface ;
                                           override ;

                                   public // API...
                                       function Count : integer ; override ;
                               end ;

// Property handlers...

function TWinControl_Debug_Iterator.Child( Index : integer ) : PDebug_Interface ; stdcall ;

var Control : TWinControl ;

begin
    Control := TWinControl( Component ) ;
    Result := TText_Debugger.Create ;
    try
        case Index of
            0 : begin
                    Result.Free ;
                    Result := TChild_Controls_Debug_Iterator.Create ;
                    TChild_Controls_Debug_Iterator( Result ).Control := Control ;
                    Result.Title := ' Children' ;//Control.Controls
                end ;
            1 :
{$IFDEF FPC}
                Result.Title := ' ParentWindow' ;//
{$ELSE}
                Result.Title := PAnsichar( ' ParentWindow = ' + inttostr( Control.ParentWindow ) ) ;
{$ENDIF}
            2 :
{$IFDEF DELPHI5UP}
                Result.Title := PAnsichar( 'AlignDisabled = ' + Boolean_To_String( Control.AlignDisabled ) ) ;
{$ELSE}
                Result.Title := PAnsichar( 'AlignDisabled = n/a' ) ;
{$ENDIF}
            3 : Result.Title := 'BevelEdges' ;//
            4 : Result.Title := 'BevelInner' ;//
            5 : Result.Title := 'BevelKind' ;//
            6 : Result.Title := 'BevelOuter' ;//
            7 : Result.Title := PAnsichar( 'BevelWidth = ' + inttostr( TPanel( Control ).BevelWidth ) ) ;
            8 : Result.Title := PAnsichar( 'BorderWidth = ' + inttostr( TPanel( Control ).BorderWidth ) ) ;
            9 : Result.Title := 'Brush' ;//
            10 :
{$IFDEF FPC}
                Result.Title := 'Ctl3D' ;//
{$ELSE}
                Result.Title := PAnsichar( 'Ctl3D = ' + Boolean_To_String( TPanel( Control ).Ctl3D ) ) ;
{$ENDIF}
            11 : Result.Title := 'DefWndProc' ;//+ Pointer_To_String( T?( Control ).DefWndProc ) ) ;
            12 : Result.Title := 'DockClients' ;//
            13 : Result.Title := PAnsichar( 'DockSite = ' + Boolean_To_String( TPanel( Control ).DockSite ) ) ;
            14 : Result.Title := PAnsichar( 'DoubleBuffered = ' + Boolean_To_String( Control.DoubleBuffered ) ) ;
            15 : if( Control.HandleAllocated ) then
                 begin
                     Result.Title := PAnsichar( 'Handle = ' + inttostr( Control.Handle ) ) ;
                 end else
                 begin
                     Result.Title := 'Handle not allocated' ;
                 end ;
            16 : Result.Title := 'ImeMode' ;//
            17 :
{$IFDEF FPC}
                Result.Title := 'ImeName' ;//
{$ELSE}
                Result.Title := PAnsichar( 'ImeName = "' + TEdit( Control ).ImeName + '"' );
{$ENDIF}
            18 :
{$IFDEF FPC}
                Result.Title := 'ParentCtl3D' ;//
{$ELSE}
                Result.Title := PAnsichar( 'ParentCtl3D = ' + Boolean_To_String( TPanel( Control ).ParentCtl3D ) ) ;
{$ENDIF}
            19 : Result.Title := PAnsichar( 'Showing = ' + Boolean_To_String( Control.Showing ) ) ;
            20 : Result.Title := PAnsichar( 'TabOrder = ' + inttostr( Control.TabORder ) ) ;
            21 : Result.Title := PAnsichar( 'TabStop = ' + Boolean_To_String( Control.TabStop ) ) ;
            22 : Result.Title := PAnsichar( 'UseDockManager = ' + Boolean_To_String( TPanel( Control ).UseDockManager ) ) ;
            23 :
{$IFDEF DELPHI5UP}
                 Result.Title := PAnsichar( 'VisibleDockClientCount = ' + inttostr( Control.VisibleDockClientCount ) ) ;
{$ELSE}
                 Result.Title := PAnsichar( 'VisibleDockClientCount = n/a' ) ;
{$ENDIF}
            24 : Result.Title := 'WindowHandle' ;//+ inttostr( T?( Control ).WindowHandle ) ) ;
            else
                begin
                    Result.Free ;
                    Result := inherited Child( Index - 25 ) ;
                end ;
        end ;
    except
        on E : Exception do Result.Title := PAnsichar( ansistring( E.Message ) ) ;
    end ;
end ;


// API...

function TWinControl_Debug_Iterator.Count : integer ; stdcall ;

begin
    Result := 25 + inherited Count ;
end ;


type TApplication_Forms_Debug_Iterator = class( TText_Debugger )
                                             protected // Property handlers...
                                                 function Child( Index : integer ) : PDebug_Interface ;
                                                     override ;

                                             public // API...
                                                 function Count : integer ; override ;
                                         end ;

// Property handlers...

function TApplication_Forms_Debug_Iterator.Child( Index : integer ) : PDebug_Interface ; stdcall ;

begin
    Result := Object_Iterator( Screen.Forms[ Index ] ) ;
    Result.Title := PAnsichar( ansistring( inttostr( Index ) + ': ' + Object_Description( Screen.Forms[ Index ] ) ) ) ;
end ;


// API...

function TApplication_Forms_Debug_Iterator.Count : integer ; stdcall ;

begin
    Result := Screen.FormCount ;
end ;


type TVCL_Debugger = class( TText_Debugger )
                         public
                             function Count : integer ; override ;
                             function Child( Ordinal : integer ) : PDebug_Interface ;
                                 override ;
                     end ;

function TVCL_Debugger.Count : integer ; stdcall ;

begin
    Result := 1 ;
end ;


function TVCL_Debugger.Child( Ordinal : integer ) : PDebug_Interface ; stdcall ;

begin
    Result := TApplication_Forms_Debug_Iterator.Create ;
    Result.Set_Title( 'Forms' ) ;
end ;



// API...

procedure Adjust_Button_Panel( C : TWinControl ) ;

var Control : TControl ;
    Control_List : TList ;
    Position_List : TInteger_List ;
    Dummy, Loop : integer ;
    Width : integer ;
    X : integer ;

begin
    // Setup...
    Control_List := TList.Create ;
    Position_List := TInteger_List.Create ;
    try
        // Build list of children controls, in ascending left-to-right order
        for Loop := 0 to C.ControlCount - 1 do
        begin
            Control := C.Controls[ Loop ] ;
            if( Control.Visible ) then
            begin
                if(
                    ( Control_List.Count = 0 )
                    or
                    ( Position_List[ Position_List.Count - 1 ] < Control.Left )
                  ) then // First item or positionally at end
                begin
                    Control_List.Add( Control ) ;
                    Position_List.Add( Control.Left ) ;
                end else
                begin
                    // Insert in left-to-right order...
                    for Dummy := 0 to Control_List.Count - 1 do
                    begin
                        if( Position_List[ Dummy ] > Control.Left ) then
                        begin
                            Control_List.Insert( Dummy, Control ) ;
                            Position_List.Insert( Dummy, Control.Left ) ;
                            break ;
                        end ;
                    end ;
                end ;
            end ; // if( Control.Visible )
        end ; // for Loop := 0 to C.ControlCount - 1

        // Now logically break control into sections, one per child control and adjust...
        if( Control_List.Count = 0 ) then
        begin
            exit ;
        end ;
        Width := C.ClientWidth div Control_List.Count ;
        X := 0 ;
        for Loop := 0 to Control_List.Count - 1 do
        begin
            Control := TControl( Control_List[ Loop ] ) ;
            Dummy := Loop * Width + Width div 2 ; // Center of section
            Dummy := Dummy - Control.Width div 2 ; // Left of control
            if( X < Dummy ) then // Don't overlay previous control
            begin
                X := Dummy ;
            end ;
            Control.Left := X ;
            X := X + Control.Width ;
        end ; // for Loop := 0 to Control_List.Count - 1

    finally
        // Clean up...
        Control_List.Free ;
        Position_List.Free ;
    end ;
end ; // Adjust_Button_Panel


procedure Bigradient_Fill( Canvas : TCanvas ; const Rect : TRect ;
    Source, Intermediate, Target : TColor ) ;

var R : TRect ;

begin
    R := Rect ;
    R.Bottom := R.Top + ( ( R.Bottom - R.Top ) div 2 ) ;
    Gradient_Fill( Canvas, R, Source, Intermediate ) ;
    R.Top := R.Bottom ;
    R.Bottom := Rect.Bottom ;
    Gradient_Fill( Canvas, R, Intermediate, Target ) ;
end ;


procedure Clone_Component( Source, Destination : TComponent ) ;

begin
    if( ( Source = nil ) or ( Destination = nil ) ) then
    begin
        exit ;
    end ;
    try
//        Destination.ComObject := Source.ComObject ;
        Destination.ComponentIndex := Source.ComponentIndex ;
//        Destination.ComponentState := Source.ComponentState ;
//        Destination.ComponentStyle := Source.ComponentStyle ;
        Destination.DesignInfo := Source.DesignInfo ;
        Destination.Tag := Source.Tag ;
        Destination.VCLComObject := Source.VCLComObject ;
    except
    end ;
end ;


type TFake_Control = class( TControl ) ; // To get access to protected properties

procedure Clone_Control( Source, Destination : TControl ) ;

var _Source, _Destination : TFake_Control ;

begin
    if( ( Source = nil ) or ( Destination = nil ) ) then
    begin
        exit ;
    end ;
    Clone_Component( Source, Destination ) ;
    _Source := TFake_Control( Source ) ;
    _Destination := TFake_Control( Destination ) ;
    try
        Destination.Action := Source.Action ;
        Destination.Align := Source.Align ;
        Destination.Anchors := Source.Anchors ;
        Destination.BiDiMode := Source.BiDiMode ;
        Destination.BoundsRect := Source.BoundsRect ;
        Destination.ClientHeight := Source.ClientHeight ;
        Destination.ClientWidth := Source.ClientWidth ;
//        Destination.Color := Source.Color ;
        Destination.Constraints := Source.Constraints ;
        Destination.ControlState := Source.ControlState ;
        Destination.ControlStyle := Source.ControlStyle ;
        Destination.Cursor := Source.Cursor ;
//        Destination.DesktopFont := Source.DesktopFont ;
        Destination.DockOrientation := Source.DockOrientation ;
//        Destination.DragCursor := Source.DragCursor ;
//        Destination.DragKind := Source.DragKind ;
//        Destination.DragMode := Source.DragMode ;
        Destination.Enabled := Source.Enabled ;
        Destination.FloatingDockSiteClass := Source.FloatingDockSiteClass ;
        try
            _Destination.Font.Assign( _Source.Font ) ;
        except
        end ;
        Destination.Height := Source.Height ;
{$IFDEF CL_CONTROL_HELPCONTEXT}
        Destination.HelpContext := Source.HelpContext ;
        Destination.HelpKeyword := Source.HelpKeyword ;
        Destination.HelpType := Source.HelpType ;
{$ENDIF}
        Destination.Hint := Source.Hint ;
        Destination.HostDockSite := Source.HostDockSite ;
//        Destination.IsControl := Source.IsControl ;
        Destination.Left := Source.Left ;
        Destination.LRDockWidth := Source.LRDockWidth ;
//        Destination.MouseCapture := Source.MouseCapture ;
        Destination.Parent := Source.Parent ;
//        Destination.ParentBiDiMode := Source.ParentBiDiMode ;
        try
            _Destination.ParentColor := _Source.ParentColor ;
        except
        end ;
        try
            _Destination.ParentFont := _Source.ParentFont ;
        except
        end ;
        try
            _Destination.ParentShowHint := _Source.ParentShowHint ;
        except
        end ;
//        Destination.PopupMenu := Source.PopupMenu ;
//        Destination.ScalingFlags := Source.ScalingFlags ;
        Destination.ShowHint := Source.ShowHint ;
        Destination.TBDockHeight := Source.TBDockHeight ;
        try
            _Destination.Text := _Source.Text ;
        except
        end ;
        Destination.Top := Source.Top ;
        Destination.UndockHeight := Source.UndockHeight ;
        Destination.UndockWidth := Source.UndockWidth ;
        Destination.Visible := Source.Visible ;
        Destination.Width := Source.Width ;
//        Destination.WindowProc := Source.WindowProc ;
{$IFNDEF FPC}
        try
            _Destination.WindowText := _Source.WindowText ;
        except
        end ;
{$ENDIF}
//        Destination.OnCanResize := Source.OnCanResize ;
        try
            _Destination.OnClick := _Source.OnClick ;
        except
        end ;
//        Destination.OnConstrainedResize := Source.OnConstrainedResize ;
//        Destination.OnContextPopup := Source.OnContextPopup ;
        try
            _Destination.OnDblClick := _Source.OnDblClick ;
        except
        end ;
//        Destination.OnDragDrop := Source.OnDragDrop ;
//        Destination.OnDragOver := Source.OnDragOver ;
//        Destination.OnEndDock := Source.OnEndDock ;
//        Destination.OnEndDrag := Source.OnEndDrag ;
//        Destination.OnMouseDown := Source.OnMouseDown ;
//        Destination.OnMouseMove := Source.OnMouseMove ;
//        Destination.OnMouseUp := Source.OnMouseUp ;
//        Destination.OnMouseWheel := Source.OnMouseWheel ;
//        Destination.OnMouseWheelDown := Source.OnMouseWheelDown ;
//        Destination.OnMouseWheelUp := Source.OnMouseWheelUp ;
//        Destination.OnResize := Source.OnResize ;
//        Destination.OnStartDock := Source.OnStartDock ;
//        Destination.OnStartDrag := Source.OnStartDrag ;
    except
    end ;
end ; // Clone_Control


procedure Clone_Speed_Button( Source, Destination : TSpeedButton ) ;

begin
    if( ( Source = nil ) or ( Destination = nil ) ) then
    begin
        exit ;
    end ;
    Clone_Control( Source, Destination ) ;
    try
        Destination.Parent := Source.Parent ;
        Destination.AllowAllUp := Source.AllowAllUp ;
        Destination.Down := Source.Down ;
        Destination.Flat := Source.Flat ;
        Destination.Glyph := Source.Glyph ;
        Destination.GroupIndex := Source.GroupIndex ;
        Destination.Layout := Source.Layout ;
        Destination.Margin := Source.Margin ;
        Destination.NumGlyphs := Source.NumGlyphs ;
        Destination.Spacing := Source.Spacing ;
        Destination.Transparent := Source.Transparent ;
        Destination.OnClick := Source.OnClick ;
        Destination.OnDblClick := Source.OnDblClick ;
        Destination.OnMouseDown := Source.OnMouseDown ;
        Destination.OnMouseMove := Source.OnMouseMove ;
        Destination.OnMouseUp := Source.OnMouseUp ;
    except
    end ;
end ;


function Control_Has_Children( Control : TControl ) : boolean ;

begin
    if( Control = nil ) then
    begin
        Result := False ;
    end else
    if( Control is TWinControl ) then
    begin                
        Result := ( TWinControl( Control ).ControlCount > 0 ) ;
    end else
    begin
        Result := False ;
    end ;
end ;


function Convert_Message( const Msg : TMessage ) : Windows.TMsg ;

begin
    Result.Message := Msg.Msg ;
    Result.WParam := Msg.WParam ;
    Result.LParam := Msg.LParam ;
end ;


function Convert_Msg( const Msg : Windows.TMsg ) : TMessage ;

begin
    Result.Msg := Msg.Message ;
    Result.WParam := Msg.WParam ;
    Result.LParam := Msg.LParam ;
end ;


procedure Copy_ListBox_To_Clipboard( LB : TListBox ) ;

var Loop : integer ;
    S : string ;

begin
    S := '' ;
    for Loop := 0 to LB.Items.Count - 1 do
    begin
        S := S + LB.Items[ Loop ] + CR ;
    end ;
    Clipboard.AsText := S ;
end ;


procedure Delete_All_Children( Control : TWinControl ) ;

var Loop : integer ;

begin
    if( Control = nil ) then
    begin
        exit ;
    end ;
    for Loop := Control.ControlCount - 1 downto 0 do
    begin
        Control.Controls[ Loop ].Free ;
    end ;
end ;


function Get_Sort_Image_List( Owner : TComponent ) : TImageList ;

var Bitmap : TBitmap ;
    R : TRect ;

begin
    Result := TImageList.Create( Owner ) ;
    Bitmap := TBitmap.Create ;
    try
        Bitmap.Width := Result.Height ;
        Bitmap.Height := Bitmap.Width ;
        R := Rect( -1, -1, Bitmap.Width + 1, Bitmap.Width + 1 ) ;
        DrawFrameControl( Bitmap.Canvas.Handle, R, DFC_SCROLL, DFCS_SCROLLDOWN or DFCS_FLAT or DFCS_INACTIVE or DFCS_TRANSPARENT ) ;
        Result.ShareImages := False ;
        Result.Add( Bitmap, nil ) ;
        Bitmap.Handle := 0 ;
    finally
        Bitmap.Free ;
    end ;

    // Up-arrow image...
    Bitmap := TBitmap.Create ;
    try
        Bitmap.Width := Result.Height ;
        Bitmap.Height := Bitmap.Width ;
        R := Rect( -1, -1, Bitmap.Width + 1, Bitmap.Width + 1 ) ;
        DrawFrameControl( Bitmap.Canvas.Handle, R, DFC_SCROLL, DFCS_SCROLLUP or DFCS_FLAT or DFCS_INACTIVE or DFCS_TRANSPARENT ) ;
        Result.Add( Bitmap, nil ) ;
        Bitmap.Handle := 0 ;
    finally
        Bitmap.Free ;
    end ;

    // Blank image...
    Bitmap := TBitmap.Create ;
    try
        Bitmap.Width := Result.Height ;
        Bitmap.Height := Bitmap.Width ;
        R := Rect( -1, -1, Bitmap.Width + 1, Bitmap.Width + 1 ) ;
        Bitmap.Canvas.Brush.Color := clBtnFace ;
        Bitmap.Canvas.FillRect( R ) ;
        Result.Add( Bitmap, nil ) ;
        Bitmap.Handle := 0 ;
    finally
        Bitmap.Free ;
    end ;
end ; // Get_Sort_Image_List


procedure Gradient_Fill( Canvas : TCanvas ; const Rect : TRect ;
    Source, Target : TColor ) ;
// Note: Currently a vertical fill.

var Delta : TPoint ; // Deltas within rectangle
    DR, DG, DB : double ; // Deltas between source and target for each component
    Loop : integer ;
    RS, RT, GS, GT, BS, BT : byte ; // Individual RGB components
    Saved : TColor ;

begin
    // Setup...
    RS := Source and $FF ;
    GS := Source shr 8 ;
    BS := Source shr 16 ;
    RT := Target and $FF ;
    GT := Target shr 8 ;
    BT := Target shr 16 ;
    Delta.X := Rect.Right - Rect.Left ;
    Delta.Y := Rect.Bottom - Rect.Top ;
    DR := ( RT - RS ) / Delta.Y ;
    DG := ( GT - GS ) / Delta.Y ;
    DB := ( BT - BS ) / Delta.Y ;

    // Now do the fill...
    Saved := Canvas.Pen.Color ;
    try
        for Loop := Rect.Top to Rect.Bottom - 1 do
        begin
            Canvas.Pen.Color := RGB(
                                     RS + trunc( DR * ( Loop - Rect.Top ) ),
                                     GS + trunc( DG * ( Loop - Rect.Top ) ),
                                     BS + trunc( DB * ( Loop - Rect.Top ) )
                                   ) ;
            Canvas.MoveTo( Rect.Left, Loop ) ;
            Canvas.LineTo( Rect.Right, Loop ) ;
        end ;
    finally
        Canvas.Pen.Color := Saved ;
    end ;
end ; // Gradient_Fill


function Min_Label_Height( L : TLabel ) : integer ;

var C : char ;
    DC : HDC ;
    Dummy, Old, Temp : integer ;
    Height : integer ;
    Lines : integer ;
    S : string ;
    Size : TSize ;
    Width : integer ;

begin
    if( L = nil ) then
    begin
        Result := 0 ;
        exit ;
    end ;
    DC := L.Canvas.Handle ;
    Height := L.Canvas.TextHeight( L.Caption ) ;
    if( not L.WordWrap ) then
    begin
        Width := $7FFFFFFF ;
    end else
    begin
        Width := L.ClientWidth ;
    end ;
    Lines := 0 ;
    S := L.Caption ;
    while( ( length( S ) > 0 ) and ( copy( S, length( S ), 1 ) = ' ' ) ) do
    begin
        Setlength( S, length( S ) - 1 ) ;
    end ;

    while( length( S ) > 0 ) do
    begin
        Old := 0 ;
        for Dummy := 1 to length( S ) do
        begin
            C := S[ Dummy ] ;
            if( C = CR ) then // Physical EOL
            begin
                inc( Lines ) ;
                S := copy( S, Dummy + 1, length( S ) ) ;
                break ;
            end else
            if( C = ' ' ) then
            begin
                GetTextExtentPoint32( DC, PAnsichar( ansistring( S ) ), Dummy - 1, Size ) ;
                if( Size.CX > Width ) then
                begin
                    if( Old = 0 ) then // A single word that doesn't fit width
                    begin
                        Temp := Dummy ;
                        while( ( Temp <= length( S ) ) and ( S[ Temp ] = ' ' ) ) do
                        begin
                            inc( Temp ) ; // Skip past multiple spaces at end of line
                        end ;
                        inc( Lines ) ;
                        S := copy( S, Temp, length( S ) ) ;
                        break ;
                    end else
                    begin
                        inc( Lines ) ;
                        S := copy( S, Old + 1, length( S ) ) ;
                        break ;
                    end ;
                end else
                begin
                    Old := Dummy ; // Remember position of last space
                end ; // if( Size.CX > Width )
            end else
            if( Dummy = length( S ) ) then
            begin
                GetTextExtentPoint32( DC, PAnsichar( ansistring( S ) ), Dummy, Size ) ;
                if( ( Size.CX > Width ) and ( Old <> 0 ) ) then
                begin
                    inc( Lines ) ;
                end ;
                S := '' ;
                inc( Lines ) ;
            end ; // if( C = CR )
        end ; // for Dummy := 1 to length( S )
    end ; // while( length( S ) > 0 )
    if( length( Edit( S, 2 ) ) > 0 ) then // Something besides spaces left
    begin
        inc( Lines ) ; { UNTESTED: Not possible to get here? }
    end ;
    Result := Lines * Height ;
end ; // Min_Label_Height


function Object_Description( O : TObject ) : string ;

begin
    if( O = nil ) then
    begin
        Result := 'nil' ;
        exit ;
    end ;
    Result := '' ;
    if( O is TComponent ) then
    begin
        Result := TComponent( O ).Name + ' ' ;
    end ;
    Result := Result + '(' + O.Classname + ')' ;
    if( O is TControl ) then
    begin
        Result := Result + ' "' + TLabel( O ).Caption + '"' ;
    end ;
    if( O is TWinControl ) then
    begin
        if( not TWinControl( O ).Visible ) then
        begin
            Result := Result + ' <invisible>' ;
        end ;
    end ;
    Result := Result + ' ' + Pointer_To_String( O ) ;
end ;


function Object_Iterator( O : TObject ) : TCOM_Debug_Interface ;

begin
    if( O = nil ) then
    begin
        Result := TText_Debugger.Create ;
        exit ;
    end ;
    if( O is TWinControl ) then
    begin
        Result := TWinControl_Debug_Iterator.Create ;
        TWinControl_Debug_Iterator( Result ).Component := TComponent( O ) ;
    end else
    if( O is TControl ) then
    begin
        Result := TControl_Debug_Iterator.Create ;
        TControl_Debug_Iterator( Result ).Component := TComponent( O ) ;
    end else
    if( O is TComponent ) then
    begin
        Result := TComponent_Debug_Iterator.Create ;
        TComponent_Debug_Iterator( Result ).Component := TComponent( O ) ;
    end else
    begin
        Result := TText_Debugger.Create ;
    end ;
end ;


procedure Remove_Duplicates( L : TStringList ) ;

var Loop : integer ;

begin
    Loop := 1 ;
    while( Loop < L.Count ) do
    begin
        if( L[ Loop - 1 ] = L[ Loop ] ) then
        begin
            L.Delete( Loop ) ;
        end else
        begin
            inc( Loop ) ;
        end ;
    end ;
end ;


// Reset pop-up hints.
procedure Reset_Hint ;

begin
    Application.ShowHint := False ;
    Application.ShowHint := True ;
end ;


{$IFDEF FPC}
const clMenuHighlight = TColor(SYS_COLOR_BASE or COLOR_MENUHILIGHT) ;
const clMenuBar = TColor(SYS_COLOR_BASE or COLOR_MENUBAR) ;
const VK_ACCEPT = 30 ;
const VK_CONVERT = 28 ;
const VK_FINAL = 24 ;
const VK_HANGUL = 21 ;
const VK_HANJA = 25 ;
const VK_JUNJA = 23 ;
const VK_KANA = 21 ;
const VK_KANJI = 25 ;
const VK_MODECHANGE = 31 ;
const VK_NONCONVERT = 29 ;
{$ENDIF}

function String_To_Color( X : string ) : TColor ;

begin
    X := Edit( X, 2 or 32 ) ; // No spaces, uppercase
    if( X = 'CLSCROLLBAR' ) then
    begin
        Result := clScrollBar ;
    end else
    if( X = 'CLBACKGROUND' ) then
    begin
        Result := clBackground ;
    end else
    if( X = 'CLACTIVECAPTION' ) then
    begin
        Result := clActiveCaption ;
    end else
    if( X = 'CLINACTIVECAPTION' ) then
    begin
        Result := clInactiveCaption ;
    end else
    if( X = 'CLMENU' ) then
    begin
        Result := clMenu ;
    end else
    if( X = 'CLWINDOW' ) then
    begin
        Result := clWindow ;
    end else
    if( X = 'CLWINDOWFRAME' ) then
    begin
        Result := clWindowFrame ;
    end else
    if( X = 'CLMENUTEXT' ) then
    begin
        Result := clMenuText ;
    end else
    if( X = 'CLWINDOWTEXT' ) then
    begin
        Result := clWindowText ;
    end else
    if( X = 'CLCAPTIONTEXT' ) then
    begin
        Result := clCaptionText ;
    end else
    if( X = 'CLACTIVEBORDER' ) then
    begin
        Result := clActiveBorder ;
    end else
    if( X = 'CLINACTIVEBORDER' ) then
    begin
        Result := clInactiveBorder ;
    end else
    if( X = 'CLAPPWORKSPACE' ) then
    begin
        Result := clAppWorkSpace ;
    end else
    if( X = 'CLHIGHLIGHT' ) then
    begin
        Result := clHighlight ;
    end else
    if( X = 'CLHIGHLIGHTTEXT' ) then
    begin
        Result := clHighlightText ;
    end else
    if( X = 'CLBTNFACE' ) then
    begin
        Result := clBtnFace ;
    end else
    if( X = 'CLBTNSHADOW' ) then
    begin
        Result := clBtnShadow ;
    end else
    if( X = 'CLGRAYTEXT' ) then
    begin
        Result := clGrayText ;
    end else
    if( X = 'CLBTNTEXT' ) then
    begin
        Result := clBtnText ;
    end else
    if( X = 'CLINACTIVECAPTIONTEXT' ) then
    begin
        Result := clInactiveCaptionText ;
    end else
    if( X = 'CLBTNHIGHLIGHT' ) then
    begin
        Result := clBtnHighlight ;
    end else
    if( X = 'CL3DDKSHADOW' ) then
    begin
        Result := cl3DDkShadow ;
    end else
    if( X = 'CL3DLIGHT' ) then
    begin
        Result := cl3DLight ;
    end else
    if( X = 'CLINFOTEXT' ) then
    begin
        Result := clInfoText ;
    end else
    if( X = 'CLINFOBK' ) then
    begin
        Result := clInfoBk ;
    end else
    if( X = 'CLHOTLIGHT' ) then
    begin
        Result := clHotLight ;
    end else
    if( X = 'CLGRADIENTACTIVE_CAPTION' ) then
    begin
        Result := clGradientActiveCaption ;
    end else
    if( X = 'CLGRADIENTINACTIVE_CAPTION' ) then
    begin
        Result := clGradientInactiveCaption ;
    end else
    if( X = 'CLMENUHIGHLIGHT' ) then
    begin
        Result := clMenuHighlight ;
    end else
    if( X = 'CLMENUBAR' ) then
    begin
        Result := clMenuBar ;
    end else
    if( X = 'CLBLACK' ) then
    begin
        Result := clBlack ;
    end else
    if( X = 'CLMAROON' ) then
    begin
        Result := clMaroon ;
    end else
    if( X = 'CLGREEN' ) then
    begin
        Result := clGreen ;
    end else
    if( X = 'CLOLIVE' ) then
    begin
        Result := clOlive ;
    end else
    if( X = 'CLNAVY' ) then
    begin
        Result := clNavy ;
    end else
    if( X = 'CLPURPLE' ) then
    begin
        Result := clPurple ;
    end else
    if( X = 'CLTEAL' ) then
    begin
        Result := clTeal ;
    end else
    if( X = 'CLGRAY' ) then
    begin
        Result := clGray ;
    end else
    if( X = 'CLSILVER' ) then
    begin
        Result := clSilver ;
    end else
    if( X = 'CLRED' ) then
    begin
        Result := clRed ;
    end else
    if( X = 'CLLIME' ) then
    begin
        Result := clLime ;
    end else
    if( X = 'CLYELLOW' ) then
    begin
        Result := clYellow ;
    end else
    if( X = 'CLBLUE' ) then
    begin
        Result := clBlue ;
    end else
    if( X = 'CLFUCHSIA' ) then
    begin
        Result := clFuchsia ;
    end else
    if( X = 'CLAQUE' ) then
    begin
        Result := clAqua ;
    end else
    if( X = 'CLLTGRAY' ) then
    begin
        Result := clLtGray ;
    end else
    if( X = 'CLDKGRAY' ) then
    begin
        Result := clDkGray ;
    end else
    if( X = 'CLWHITE' ) then
    begin
        Result := clWhite ;
    end else
    if( X = 'CLMONEYGREEN' ) then
    begin
        Result := clMoneyGreen ;
    end else
    if( X = 'CLSKYBLUE' ) then
    begin
        Result := clSkyBlue ;
    end else
    if( X = 'CLCREAM' ) then
    begin
        Result := clCream ;
    end else
    if( X = 'CLMEDGRAY' ) then
    begin
        Result := clMedGray ;
    end else
    if( X = 'CLNONE' ) then
    begin
        Result := clNone ;
    end else
    begin
        Result := clDefault ;
    end ;
end ;


function String_To_VK( X : string ) : integer ;

begin
    X := Edit( X, 2 or 32 ) ; // No spaces, uppercase
    if( X = 'ACCEPT' ) then
    begin
        Result := VK_ACCEPT ;
    end else
    if( X = 'ADD' ) then
    begin
        Result := VK_ADD ;
    end else
    if( X = 'APPS' ) then
    begin
        Result := VK_APPS ;
    end else
    if( X = 'ATTN' ) then
    begin
        Result := VK_ATTN ;
    end else
    if( X = 'BACK' ) then
    begin
        Result := VK_BACK ;
    end else
    if( X = 'CANCEL' ) then
    begin
        Result := VK_CANCEL ;
    end else
    if( X = 'CAPITAL' ) then
    begin
        Result := VK_CAPITAL ;
    end else
    if( X = 'CLEAR' ) then
    begin
        Result := VK_CLEAR ;
    end else
    if( X = 'CONTROL' ) then
    begin
        Result := VK_CONTROL ;
    end else
    if( X = 'CONVERT' ) then
    begin
        Result := VK_CONVERT ;
    end else
    if( X = 'CRSEL' ) then
    begin
        Result := VK_CRSEL ;
    end else
    if( X = 'DECIMAL' ) then
    begin
        Result := VK_DECIMAL ;
    end else
    if( X = 'DELETE' ) then
    begin
        Result := VK_DELETE ;
    end else
    if( X = 'DIVIDE' ) then
    begin
        Result := VK_DIVIDE ;
    end else
    if( X = 'DOWN' ) then
    begin
        Result := VK_DOWN ;
    end else
    if( X = 'END' ) then
    begin
        Result := VK_END ;
    end else
    if( X = 'ENTER' ) then
    begin
        Result := VK_RETURN ;
    end else
    if( X = 'EREOF' ) then
    begin
        Result := VK_EREOF ;
    end else
    if( ( X = 'ESCAPE' ) or ( X = 'ESC' ) ) then
    begin
        Result := VK_ESCAPE ;
    end else
    if( X = 'EXSEL' ) then
    begin
        Result := VK_EXSEL ;
    end else
    if( X = 'EXECUTE' ) then
    begin
        Result := VK_EXECUTE ;
    end else
    if( X = 'F1' ) then
    begin
        Result := VK_F1 ;
    end else
    if( X = 'F2' ) then
    begin
        Result := VK_F2 ;
    end else
    if( X = 'F3' ) then
    begin
        Result := VK_F3 ;
    end else
    if( X = 'F4' ) then
    begin
        Result := VK_F4 ;
    end else
    if( X = 'F5' ) then
    begin
        Result := VK_F5 ;
    end else
    if( X = 'F6' ) then
    begin
        Result := VK_F6 ;
    end else
    if( X = 'F7' ) then
    begin
        Result := VK_F7 ;
    end else
    if( X = 'F8' ) then
    begin
        Result := VK_F8 ;
    end else
    if( X = 'F9' ) then
    begin
        Result := VK_F9 ;
    end else
    if( X = 'F10' ) then
    begin
        Result := VK_F10 ;
    end else
    if( X = 'F11' ) then
    begin
        Result := VK_F11 ;
    end else
    if( X = 'F12' ) then
    begin
        Result := VK_F12 ;
    end else
    if( X = 'F13' ) then
    begin
        Result := VK_F13 ;
    end else
    if( X = 'F14' ) then
    begin
        Result := VK_F14 ;
    end else
    if( X = 'F15' ) then
    begin
        Result := VK_F15 ;
    end else
    if( X = 'F16' ) then
    begin
        Result := VK_F16 ;
    end else
    if( X = 'F17' ) then
    begin
        Result := VK_F17 ;
    end else
    if( X = 'F18' ) then
    begin
        Result := VK_F18 ;
    end else
    if( X = 'F19' ) then
    begin
        Result := VK_F19 ;
    end else
    if( X = 'F20' ) then
    begin
        Result := VK_F20 ;
    end else
    if( X = 'F21' ) then
    begin
        Result := VK_F21 ;
    end else
    if( X = 'F22' ) then
    begin
        Result := VK_F22 ;
    end else
    if( X = 'F23' ) then
    begin
        Result := VK_F23 ;
    end else
    if( X = 'F24' ) then
    begin
        Result := VK_F24 ;
    end else
    if( X = 'FINAL' ) then
    begin
        Result := VK_FINAL ;
    end else
    if( X = 'HANGUL' ) then
    begin
        Result := VK_HANGUL ;
    end else
    if( X = 'HANJA' ) then
    begin
        Result := VK_HANJA ;
    end else
    if( X = 'HELP' ) then
    begin
        Result := VK_HELP ;
    end else
    if( X = 'HOME' ) then
    begin
        Result := VK_HOME ;
    end else
    if( X = 'INSERT' ) then
    begin
        Result := VK_INSERT ;
    end else
    if( X = 'JINJA' ) then
    begin
        Result := VK_JUNJA ;
    end else
    if( X = 'KANA' ) then
    begin
        Result := VK_KANA ;
    end else
    if( X = 'KANJI' ) then
    begin
        Result := VK_KANJI ;
    end else
    if( X = 'LCONTROL' ) then
    begin
        Result := VK_LCONTROL ;
    end else
    if( X = 'LEFT' ) then
    begin
        Result := VK_LEFT ;
    end else
    if( X = 'LMENU' ) then
    begin
        Result := VK_LMENU ;
    end else
    if( X = 'LMOUSE' ) then
    begin
        Result := VK_LBUTTON ;
    end else
    if( X = 'LSHIFT' ) then
    begin
        Result := VK_LSHIFT ;
    end else
    if( X = 'LWIN' ) then
    begin
        Result := VK_LWIN ;
    end else
    if( X = 'MENU' ) then
    begin
        Result := VK_MENU ;
    end else
    if( X = 'MMOUSE' ) then
    begin
        Result := VK_MBUTTON ;
    end else
    if( X = 'MODE' ) then
    begin
        Result := VK_MODECHANGE ;
    end else
    if( X = 'MULTIPLY' ) then
    begin
        Result := VK_MULTIPLY ;
    end else
    if( X = 'NONAME' ) then
    begin
        Result := VK_NONAME ;
    end else
    if( X = 'NONCONVERT' ) then
    begin
        Result := VK_NONCONVERT ;
    end else
    if( X = 'NUM0' ) then
    begin
        Result := VK_NUMPAD0 ;
    end else
    if( X = 'NUM1' ) then
    begin
        Result := VK_NUMPAD1 ;
    end else
    if( X = 'NUM2' ) then
    begin
        Result := VK_NUMPAD2 ;
    end else
    if( X = 'NUM3' ) then
    begin
        Result := VK_NUMPAD3 ;
    end else
    if( X = 'NUM4' ) then
    begin
        Result := VK_NUMPAD4 ;
    end else
    if( X = 'NUM5' ) then
    begin
        Result := VK_NUMPAD5 ;
    end else
    if( X = 'NUM6' ) then
    begin
        Result := VK_NUMPAD6 ;
    end else
    if( X = 'NUM7' ) then
    begin
        Result := VK_NUMPAD7 ;
    end else
    if( X = 'NUM8' ) then
    begin
        Result := VK_NUMPAD8 ;
    end else
    if( X = 'NUM9' ) then
    begin
        Result := VK_NUMPAD9 ;
    end else
    if( X = 'NUMLOCK' ) then
    begin
        Result := VK_NUMLOCK ;
    end else
    if( X = 'PA1' ) then
    begin
        Result := VK_PA1 ;
    end else
    if( X = 'PAGEDOWN' ) then
    begin
        Result := VK_NEXT ;
    end else
    if( X = 'PAGEUP' ) then
    begin
        Result := VK_PRIOR ;
    end else
    if( X = 'PAUSE' ) then
    begin
        Result := VK_PAUSE ;
    end else
    if( X = 'PLAY' ) then
    begin
        Result := VK_PLAY ;
    end else
    if( X = 'PRINT' ) then
    begin
        Result := VK_PRINT ;
    end else
    if( X = 'PROCESSKEY' ) then
    begin
        Result := VK_PROCESSKEY ;
    end else
    if( X = 'RCONTROL' ) then
    begin
        Result := VK_RCONTROL ;
    end else
    if( X = 'RIGHT' ) then
    begin
        Result := VK_RIGHT ;
    end else
    if( X = 'RMENU' ) then
    begin
        Result := VK_RMENU ;
    end else
    if( X = 'RMOUSE' ) then
    begin
        Result := VK_RBUTTON ;
    end else
    if( X = 'RSHIFT' ) then
    begin
        Result := VK_RSHIFT ;
    end else
    if( X = 'RWIN' ) then
    begin
        Result := VK_RWIN ;
    end else
    if( X = 'SCROLL' ) then
    begin
        Result := VK_SCROLL ;
    end else
    if( X = 'SELECT' ) then
    begin
        Result := VK_SELECT ;
    end else
    if( X = 'SEPARATOR' ) then
    begin
        Result := VK_SEPARATOR ;
    end else
    if( X = 'SHIFT' ) then
    begin
        Result := VK_SHIFT ;
    end else
    if( X = 'SNAPSHOT' ) then
    begin
        Result := VK_SNAPSHOT ;
    end else
    if( X = 'SPACE' ) then
    begin
        Result := VK_SPACE ;
    end else
    if( X = 'SUBTRACT' ) then
    begin
        Result := VK_SUBTRACT ;
    end else
    if( X = 'TAB' ) then
    begin
        Result := VK_TAB ;
    end else
    if( X = 'UP' ) then
    begin
        Result := VK_UP ;
    end else
    if( X = 'ZOOM' ) then
    begin
        Result := VK_ZOOM ;
    end else
    if( length( X ) = 0 ) then
    begin
        Result := 0 ;
    end else
    begin
        Result := ord( X[ 1 ] ) ;
    end ;
end ;


function VCL_Debugger : PDebug_Interface ;

begin
    Result := TVCL_Debugger.Create ;
end ;


function VK_To_String( Code : integer ) : string ;

begin
    case Code of
        VK_LBUTTON : Result := 'L Mouse' ;
        VK_RBUTTON : Result := 'R Mouse' ;
        VK_CANCEL : Result := 'Cancel' ;
        VK_MBUTTON : Result := 'M Mouse' ;
        VK_BACK : Result := 'Back' ;
        VK_TAB : Result := 'Tab' ;
        VK_CLEAR : Result := 'Clear' ;
        VK_RETURN : Result := 'Enter' ;
        VK_SHIFT : Result := 'Shift' ;
        VK_CONTROL : Result := 'Control' ;
        VK_MENU : Result := 'Menu' ;
        VK_PAUSE : Result := 'Pause' ;
        VK_CAPITAL : Result := 'Capital' ;
        VK_KANA : Result := 'Kana' ;
        VK_JUNJA : Result := 'Jinja' ;
        VK_FINAL : Result := 'Final' ;
        VK_HANJA : Result := 'Hanja' ;
        VK_CONVERT : Result := 'Convert' ;
        VK_NONCONVERT : Result := 'Nonconvert' ;
        VK_ACCEPT : Result := 'Accept' ;
        VK_MODECHANGE : Result := 'Mode' ;
        VK_ESCAPE : Result := 'Escape' ;
        VK_SPACE : Result := 'Space' ;
        VK_PRIOR : Result := 'Page Up' ;
        VK_NEXT : Result := 'Page Down' ;
        VK_END : Result := 'End' ;
        VK_HOME : Result := 'Home' ;
        VK_LEFT : Result := 'Left' ;
        VK_UP : Result := 'Up' ;
        VK_RIGHT : Result := 'Right' ;
        VK_DOWN : Result := 'Down' ;
        VK_SELECT : Result := 'Select' ;
        VK_PRINT : Result := 'Print' ;
        VK_EXECUTE : Result := 'Execute' ;
        VK_SNAPSHOT : Result := 'Snapshot' ;
        VK_INSERT : Result := 'Insert' ;
        VK_DELETE : Result := 'Delete' ;
        VK_HELP : Result := 'Help' ;
        VK_LWIN : Result := 'L Win' ;
        VK_RWIN : Result := 'R Win' ;
        VK_APPS : Result := 'Apps' ;
        VK_NUMPAD0 : Result := 'Num 0' ;
        VK_NUMPAD1 : Result := 'Num 1' ;
        VK_NUMPAD2 : Result := 'Num 2' ;
        VK_NUMPAD3 : Result := 'Num 3' ;
        VK_NUMPAD4 : Result := 'Num 4' ;
        VK_NUMPAD5 : Result := 'Num 5' ;
        VK_NUMPAD6 : Result := 'Num 6' ;
        VK_NUMPAD7 : Result := 'Num 7' ;
        VK_NUMPAD8 : Result := 'Num 8' ;
        VK_NUMPAD9 : Result := 'Num 9' ;
        VK_MULTIPLY : Result := 'Multiply' ;
        VK_ADD : Result := 'Add' ;
        VK_SEPARATOR : Result := 'Separator' ;
        VK_SUBTRACT : Result := 'Subtract' ;
        VK_DECIMAL : Result := 'Decimal' ;
        VK_DIVIDE : Result := 'Divide' ;
        VK_F1 : Result := 'F1' ;
        VK_F2 : Result := 'F2' ;
        VK_F3 : Result := 'F3' ;
        VK_F4 : Result := 'F4' ;
        VK_F5 : Result := 'F5' ;
        VK_F6 : Result := 'F6' ;
        VK_F7 : Result := 'F7' ;
        VK_F8 : Result := 'F8' ;
        VK_F9 : Result := 'F9' ;
        VK_F10 : Result := 'F10' ;
        VK_F11 : Result := 'F11' ;
        VK_F12 : Result := 'F12' ;
        VK_F13 : Result := 'F13' ;
        VK_F14 : Result := 'F14' ;
        VK_F15 : Result := 'F15' ;
        VK_F16 : Result := 'F16' ;
        VK_F17 : Result := 'F17' ;
        VK_F18 : Result := 'F18' ;
        VK_F19 : Result := 'F19' ;
        VK_F20 : Result := 'F20' ;
        VK_F21 : Result := 'F21' ;
        VK_F22 : Result := 'F22' ;
        VK_F23 : Result := 'F23' ;
        VK_F24 : Result := 'F24' ;
        VK_NUMLOCK : Result := 'Numlock' ;
        VK_SCROLL : Result := 'Scroll' ;
        VK_LSHIFT : Result := 'L Shift' ;
        VK_RSHIFT : Result := 'R Shift' ;
        VK_LCONTROL : Result := 'L Control' ;
        VK_RCONTROL : Result := 'R Control' ;
        VK_LMENU : Result := 'L Menu' ;
        VK_RMENU : Result := 'R Menu' ;
        VK_PROCESSKEY : Result := 'Processkey' ;
        VK_ATTN : Result := 'Attn' ;
        VK_CRSEL : Result := 'Crsel' ;
        VK_EXSEL : Result := 'Exsel' ;
        VK_EREOF : Result := 'ErEOF' ;
        VK_PLAY : Result := 'Play' ;
        VK_ZOOM : Result := 'Zoom' ;
        VK_NONAME : Result := 'Noname' ;
        VK_PA1 : Result := 'PA1' ;
        VK_OEM_CLEAR : Result := 'Clear' ;
        else Result := chr( Code ) ;
    end ; // case Code
end ; // VK_To_String


{$IFDEF VER140}
type tagSize = TSize ;
{$ENDIF}

function Write_Text( Handle : THandle ; Font : TFont ; X, Y, Max_X : integer ;
    Text : string ) : integer ;

    procedure Write_Segment( var X : integer ; Y : integer ; Text : string ) ;

        procedure Text_Out( const S : string ) ;

        var Size : tSize ;

        begin
            SetBKMode( Handle, TRANSPARENT ) ; { Something resets this, so change it back }
            Windows.TextOutA( Handle, X, Y, PAnsichar( ansistring( S ) ), length( S ) ) ;
            GetTextExtentPoint32( Handle, PAnsichar( ansistring( S ) ), length( S ), Size ) ;
            X := X + Size.cx ;
        end ;

    var Dummy : integer ;
        S : string ;

    begin
        Dummy := pos( '&', Text + '&' ) ;
        S := '' ;
        while( Dummy <= length( Text ) ) do
        begin
            S := S + copy( Text, 1, Dummy - 1 ) ;
            Text := copy( Text, Dummy + 1, length( Text ) ) ;
            if( copy( Text, 1, 1 ) = '&' ) then { "&&" means literal ampersand }
            begin
                Text := copy( Text, Dummy + 1, length( Text ) ) ;
                S := S + '&' ;
            end else
            begin
                { Underscore next letter }
                if( length( S ) > 0 ) then
                begin
                    Text_Out( S ) ;
                    S := '' ;
                end ;
                if( length( Text ) > 0 ) then
                begin
                    Windows.TextOutA( Handle, X, Y, PAnsichar( Ansistring( Text ) ), 1 ) ;
                    Text_Out( '_' ) ;
                    Text := copy( Text, 2, length( Text ) ) ; { Remove underscored character }
                end ;
            end ; { if( copy( Text, 1, 1 ) = '&' ) }
            Dummy := pos( '&', Text + '&' ) ;
        end ;
        if( length( S ) > 0 ) then
        begin
            Text_Out( S ) ;
        end ;
        if( length( Text ) > 0 ) then
        begin
            Text_Out( Text ) ;
        end ;
    end ; { Write_Text.Write_Segment }

var Dummy : integer ;
    S : string ;
    Saved : Thandle ;
    Saved_Color : TColor ;
    Size : tSize ;

begin { Write_Text }
    { Setup... }
    Saved_Color := GetTextColor( Handle ) ;
    Saved := SelectObject( Handle, Font.Handle ) ;
    try
        SetTextColor( Handle, ColorToRGB( Font.Color ) ) ;

        { Handle up to tab... }
        Dummy := pos( #9, Text + #9 ) ;
        S := copy( Text, 1, Dummy - 1 ) ;
        Text := copy( Text, Dummy, length( Text ) ) ;
        Write_Segment( X, Y, S ) ;

        if( length( Text ) > 0 ) then
        begin
            { Space for tab }
            GetTextExtentPoint32( Handle, #9, 1, Size ) ;
            X := X + Size.CX ;

            { Write text following tab... }
            Write_Segment( X, Y, Text ) ;
        end ;

    finally
        { Cleanup }
        SelectObject( Handle, Saved ) ;
        SetTextColor( Handle, Saved_Color ) ;
    end ;
    Result := Max_X - X ;
end ; { Write_Text }



{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


procedure Test_Convert_Message ;

var M : TMsg ;
    M1 : TMessage ;

begin
    M1.Msg := 1 ;
    M1.WParam := 0 ;
    M1.LParam := 2 ;
    M := Convert_Message( M1 ) ;
    if( M.Message <> 1 ) then
    begin
        Error ;
    end ;
    if( M.WParam <> 0 ) then
    begin
        Error ;
    end ;
    if( M.LParam <> 2 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Convert_Msg ;

var M : TMsg ;
    M1 : TMessage ;

begin
    M.Message := 1 ;
    M.WParam := 0 ;
    M.LParam := 2 ;
    M1 := Convert_Msg( M ) ;
    if( M1.Msg <> 1 ) then
    begin
        Error ;
    end ;
    if( M1.WParam <> 0 ) then
    begin
        Error ;
    end ;
    if( M1.LParam <> 2 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Get_Sort_Image_List ;

var List : TImageList ;

begin
    List := Get_Sort_Image_List( Application ) ;
    List.Free ;
end ;


procedure Test_Reset_Hint ;

begin
    Reset_Hint ;
    if( not Application.ShowHint ) then
    begin
        Error ;
    end ;
end ;


procedure Test_String_To_VK ;

    procedure Test( const X : string ; Value : integer ) ;

    begin
        if( String_To_VK( X ) <> Value ) then
        begin
            Error ;
        end ;
        if( Edit( VK_To_String( Value ), 2 or 32 ) <> X ) then
        begin
            Error ;
        end ;
    end ;

begin
    Test( 'LMOUSE', VK_LBUTTON ) ;
    Test( 'RMOUSE', VK_RBUTTON ) ;
    Test( 'CANCEL', VK_CANCEL ) ;
    Test( 'MMOUSE', VK_MBUTTON ) ;
    Test( 'BACK', VK_BACK ) ;
    Test( 'TAB', VK_TAB ) ;
    Test( 'CLEAR', VK_CLEAR ) ;
    Test( 'ENTER', VK_RETURN ) ;
    Test( 'SHIFT', VK_SHIFT ) ;
    Test( 'CONTROL', VK_CONTROL ) ;
    Test( 'MENU', VK_MENU ) ;
    Test( 'PAUSE', VK_PAUSE ) ;
    Test( 'CAPITAL', VK_CAPITAL ) ;
    Test( 'KANA', VK_KANA ) ;
    Test( 'JINJA', VK_JUNJA ) ;
    Test( 'FINAL', VK_FINAL ) ;
    Test( 'HANJA', VK_HANJA ) ;
    Test( 'CONVERT', VK_CONVERT ) ;
    Test( 'NONCONVERT', VK_NONCONVERT ) ;
    Test( 'ACCEPT', VK_ACCEPT ) ;
    Test( 'MODE', VK_MODECHANGE ) ;
    Test( 'ESCAPE', VK_ESCAPE ) ;
    Test( 'SPACE', VK_SPACE ) ;
    Test( 'PAGEUP', VK_PRIOR ) ;
    Test( 'PAGEDOWN', VK_NEXT ) ;
    Test( 'END', VK_END ) ;
    Test( 'HOME', VK_HOME ) ;
    Test( 'LEFT', VK_LEFT ) ;
    Test( 'UP', VK_UP ) ;
    Test( 'RIGHT', VK_RIGHT ) ;
    Test( 'DOWN', VK_DOWN ) ;
    Test( 'SELECT', VK_SELECT ) ;
    Test( 'PRINT', VK_PRINT ) ;
    Test( 'EXECUTE', VK_EXECUTE ) ;
    Test( 'SNAPSHOT', VK_SNAPSHOT ) ;
    Test( 'INSERT', VK_INSERT ) ;
    Test( 'DELETE', VK_DELETE ) ;
    Test( 'HELP', VK_HELP ) ;
    Test( 'LWIN', VK_LWIN ) ;
    Test( 'RWIN', VK_RWIN ) ;
    Test( 'APPS', VK_APPS ) ;
    Test( 'NUM0', VK_NUMPAD0 ) ;
    Test( 'NUM1', VK_NUMPAD1 ) ;
    Test( 'NUM2', VK_NUMPAD2 ) ;
    Test( 'NUM3', VK_NUMPAD3 ) ;
    Test( 'NUM4', VK_NUMPAD4 ) ;
    Test( 'NUM5', VK_NUMPAD5 ) ;
    Test( 'NUM6', VK_NUMPAD6 ) ;
    Test( 'NUM7', VK_NUMPAD7 ) ;
    Test( 'NUM8', VK_NUMPAD8 ) ;
    Test( 'NUM9', VK_NUMPAD9 ) ;
    Test( 'MULTIPLY', VK_MULTIPLY ) ;
    Test( 'ADD', VK_ADD ) ;
    Test( 'SEPARATOR', VK_SEPARATOR ) ;
    Test( 'SUBTRACT', VK_SUBTRACT ) ;
    Test( 'DECIMAL', VK_DECIMAL ) ;
    Test( 'DIVIDE', VK_DIVIDE ) ;
    Test( 'F1', VK_F1 ) ;
    Test( 'F2', VK_F2 ) ;
    Test( 'F3', VK_F3 ) ;
    Test( 'F4', VK_F4 ) ;
    Test( 'F5', VK_F5 ) ;
    Test( 'F6', VK_F6 ) ;
    Test( 'F7', VK_F7 ) ;
    Test( 'F8', VK_F8 ) ;
    Test( 'F9', VK_F9 ) ;
    Test( 'F10', VK_F10 ) ;
    Test( 'F11', VK_F11 ) ;
    Test( 'F12', VK_F12 ) ;
    Test( 'F13', VK_F13 ) ;
    Test( 'F14', VK_F14 ) ;
    Test( 'F15', VK_F15 ) ;
    Test( 'F16', VK_F16 ) ;
    Test( 'F17', VK_F17 ) ;
    Test( 'F18', VK_F18 ) ;
    Test( 'F19', VK_F19 ) ;
    Test( 'F20', VK_F20 ) ;
    Test( 'F21', VK_F21 ) ;
    Test( 'F22', VK_F22 ) ;
    Test( 'F23', VK_F23 ) ;
    Test( 'F24', VK_F24 ) ;
    Test( 'NUMLOCK', VK_NUMLOCK ) ;
    Test( 'SCROLL', VK_SCROLL ) ;
    Test( 'LSHIFT', VK_LSHIFT ) ;
    Test( 'RSHIFT', VK_RSHIFT ) ;
    Test( 'LCONTROL', VK_LCONTROL ) ;
    Test( 'RCONTROL', VK_RCONTROL ) ;
    Test( 'LMENU', VK_LMENU ) ;
    Test( 'RMENU', VK_RMENU ) ;
    Test( 'PROCESSKEY', VK_PROCESSKEY ) ;
    Test( 'ATTN', VK_ATTN ) ;
    Test( 'CRSEL', VK_CRSEL ) ;
    Test( 'EXSEL', VK_EXSEL ) ;
    Test( 'EREOF', VK_EREOF ) ;
    Test( 'PLAY', VK_PLAY ) ;
    Test( 'ZOOM', VK_ZOOM ) ;
    Test( 'NONAME', VK_NONAME ) ;
    Test( 'PA1', VK_PA1 ) ;

    if( String_To_VK( '' ) <> 0 ) then
    begin
        Error ;
    end ;
    if( String_To_VK( 'A' ) <> 65 ) then
    begin
        Error ;
    end ;
    if( String_To_VK( 'HANGUL' ) <> VK_HANGUL ) then // Same as KANA
    begin
        Error ;
    end ;
    if( String_To_VK( 'KANJI' ) <> VK_KANJI ) then // Same as HANJA
    begin
        Error ;
    end ;
    if( String_To_VK( 'ESC' ) <> VK_ESCAPE ) then // Same as ESCAPE
    begin
        Error ;
    end ;
end ;


var Button : TSpeedButton ;
    EditBox : TEdit ;
    Form : TForm ;
    _Label : TLabel ;
    Panel : TPanel ;

procedure Test_Adjust_Button_Panel ;

var Dummy : integer ;

begin
    Adjust_Button_Panel( EditBox ) ;
    EditBox.Visible := True ;
    EditBox.Width := Panel.ClientWidth div 4 ;
    Adjust_Button_Panel( Panel ) ;
    Dummy := Panel.ClientWidth div 2 - EditBox.Width div 2 ;
    if( ( EditBox.Left < Dummy - 1 ) or ( EditBox.Left > Dummy + 1 ) ) then
    begin
        Error ;
    end ;
    Button.Visible := False ;
    Adjust_Button_Panel( Panel ) ;
    Dummy := Panel.ClientWidth div 4 - EditBox.Width div 2 ;
    if( ( EditBox.Left < Dummy - 1 ) or ( EditBox.Left > Dummy + 1 ) ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Control_Has_Children ;

begin
    if( not Control_Has_Children( Form ) ) then
    begin
        Error ;
    end ;
    if( Control_Has_Children( nil ) ) then
    begin
        Error ;
    end ;
    if( Control_Has_Children( Button ) ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Delete_All_Children ;

begin
    Delete_All_Children( nil ) ;
    Delete_All_Children( Panel ) ;
    if( Control_Has_Children( Panel ) ) then
    begin
        Error ;
    end ;
    Delete_All_Children( Panel ) ;
end ;


procedure Test_Min_Label_Height ;

begin
    if( Min_Label_Height( nil ) <> 0 ) then
    begin
        Error ;
    end ;
    if( Min_Label_Height( _Label ) <> _Label.Height ) then
    begin
        Error ;
    end ;
    _Label.AutoSize := False ;
    _Label.Width := _Label.Canvas.TextWidth( 'T' ) + 2 ;
    _Label.Caption := 'TT' ;
    if( Min_Label_Height( _Label ) <> _Label.Height ) then
    begin
        Error ;
    end ;
    _Label.Caption := 'T T' ;
    if( Min_Label_Height( _Label ) <> _Label.Height ) then
    begin
        Error ;
    end ;
    _Label.WordWrap := True ;
    if( Min_Label_Height( _Label ) <> _Label.Height * 2 ) then
    begin
        Error ;
    end ;
    _Label.Caption := 'T T  ' ;
    if( Min_Label_Height( _Label ) <> _Label.Height * 2 ) then
    begin
        Error ;
    end ;
    _Label.Caption := 'TTTT  T' ;
    if( Min_Label_Height( _Label ) <> _Label.Height * 2 ) then
    begin
        Error ;
    end ;
    _Label.Caption := 'T'#13'T' ;
    if( Min_Label_Height( _Label ) <> _Label.Height * 2 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Write_Text ;

var Font : TFont ;

begin
    Font := TFont.Create ;
    if( Write_Text( _Label.Canvas.Handle, Font, 0, 0, 32767, 'T' ) <> 32767 - _Label.Canvas.TextWidth( 'T' ) ) then
    begin
        Error ;
    end ;
    if( Write_Text( _Label.Canvas.Handle, Font, 0, 0, 32767, 'T&' ) <> 32767 - _Label.Canvas.TextWidth( 'T' ) ) then
    begin
        Error ;
    end ;
    if( Write_Text( _Label.Canvas.Handle, Font, 0, 0, 32767, 'T&&' ) <> 32767 - _Label.Canvas.TextWidth( 'T&' ) ) then
    begin
        Error ;
    end ;
    if( Write_Text( _Label.Canvas.Handle, Font, 0, 0, 32767, '&T' ) <> 32767 - _Label.Canvas.TextWidth( '_' ) ) then
    begin
        Error ;
    end ;
    if( Write_Text( _Label.Canvas.Handle, Font, 0, 0, 32767, 'T'#9'T' ) > 32767 - _Label.Canvas.TextWidth( 'TT' ) ) then
    begin
        Error ;
    end ;
    Font.Free ;
end ;


type TTest_Object = class
                        public
                            procedure Test ;
                    end ;

procedure TTest_Object.Test ;

begin
end ;


procedure Test_Object_Description ;

var O : TTest_Object ;
    S : string ;

begin
    S := Object_Description( Application ) ;
    if( copy( S, 1, 16 ) <> ' (TApplication) ' ) then
    begin
        Error ;
    end ;
    if( Object_Description( nil ) <> 'nil' ) then
    begin
        Error ;
    end ;
    O := TTest_Object.Create ;
    S := Object_Description( O ) ;
    if( copy( S, 1, 14 ) <> '(TTest_Object)' ) then
    begin
        Error ;
    end ;
    O.Free ;
    S := Object_Description( EditBox ) ;
    if( copy( S, 1, 12 ) <> ' (TEdit) "" ' ) then
    begin
        Error ;
    end ;
    EditBox.Visible := False ;
    S := Object_Description( EditBox ) ;
    if( copy( S, 1, 24 ) <> ' (TEdit) "" <invisible> ' ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
    Test_Convert_Message ;
    Test_Convert_Msg ;
    Test_Get_Sort_Image_List ;
    Test_Reset_Hint ;
    Test_String_To_VK ;

    Form := TForm.Create( Application ) ;
    Panel := TPanel.Create( Form ) ;
    Panel.Parent := Form ;
    Button := TSpeedButton.Create( Form ) ;
    Button.Parent := Panel ;
    Button.Left := 0 ;
    EditBox := TEdit.Create( Form ) ;
    EditBox.Parent := Panel ;
    EditBox.Left := 2 ;
    _Label := TLabel.Create( Form ) ;
    _Label.Parent := Panel ;
    _Label.Caption := 'T' ;
    _Label.Left := 4 ;

    Test_Object_Description ;
    Test_Min_Label_Height ;
    Test_Adjust_Button_Panel ;
    Test_Control_Has_Children ;
    Test_Write_Text ;
    Test_Delete_All_Children ;

    Form.Free ;
end ;
{$ENDIF}



end.

