/*
        Program Name : TCommon_COM_Interface
        Package Name : Subroutine Library - Layer 0
        Purpose      : Abstract virtual common COM interface class
        Institution  : Conroy & Conroy Co.
        Date Written : 21-March-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2002-2015 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TBase_COM_Interface is an abstract base class definition for COM
        objects used in the Application Foundation.
*/

#ifndef _COMINTER_HPP
	#define _COMINTER_HPP

	#include "eDefines.hpp"
	#include "_DebugIn.hpp"
	#include "_Types.hpp"
	#include "_UE.hpp"

	#ifdef STDCALL_SUPPORTED

class TBase_COM_Interface
{
    public: // API...
        // Initialize the object.  Returns result of initialization attempt.
        virtual tUnified_Exception* __stdcall Initialize() = 0  ;

        /* Terminate the use of the object.  Returns
           result of termination.  This generally
           should not be used - instead, use the Detach method. */
        virtual tUnified_Exception* __stdcall Terminate() = 0 ;

        // Increments the reference count for the object.
        virtual void __stdcall Attach() = 0 ;

        /* Decrements the reference count for the
           object.  When count reaches 0, the object
           is destroyed.  Returns true if object is
           destroyed, false otherwise. */
        virtual bool __stdcall Detach() = 0 ;

        // Returns True if this class is the passed class name.
        virtual bool __stdcall Is_Class( char* Name ) = 0 ;

        // Version of the COM interface for this object.
        virtual int32 __stdcall Interface_Version() = 0 ;

        /* Facility ID for the facility represented
           by this object.  Returns -1 if no facility assigned. */
        virtual int32 __stdcall Facility_ID() = 0 ;

        /* Facility version.  Result only has
           meaning if Facility_ID doesn't return -1. */
        virtual int32 __stdcall Facility_Version() = 0 ;

        // Returns a debugging interface for the object.  Returns nil if not supported.
        virtual TDebug_Interface* __stdcall Debugger() = 0 ;

        // Returns last error.
        virtual tUnified_Exception* __stdcall Last_Error() = 0 ;

        /* Returns a pointer to an object with
           extended common COM methods for object.
           Always returns nil for now. */
        virtual void* __stdcall Extension() = 0 ;
} ; // TBase_COM_Interface

	#endif
#endif
