{
        Program Name : _Common
        Package Name : Subroutine Library - Layer 0
        Purpose      : Common base objects (TObject and TDelphi_Object)
        Institution  : Conroy & Conroy Co.
        Date Written : 19-Mar-2002
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2005 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

           DATE        BY          REASON                         

        1-Nov-2002     EAC         Removed pointless overrides of
                                   TObject methods in BP.  Added
                                   reference counting support.

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

}

unit _Common ;

interface

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$IFDEF BP}
uses Objects ; { TObject }
{$ENDIF}

type TMemory_Type = ( MT_Unknown, { Don't know }
                      MT_Heap, { Allocated on heap }
                      MT_Data, { "Global" data }
                      MT_Stack, { "Local" data }
                      MT_Code { executable }
                    ) ;

{$IFDEF DYNAMIC_OBJECTS}
type TDelphi_Object = class
                          public { API... }
                               procedure Attach ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                   stdcall ;
  {$ENDIF}
                                   abstract ;
                               procedure Detach ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                   stdcall ;
  {$ENDIF}
                                   abstract ;
                               function Is_Class( const Name : string ) : boolean ;
                                   virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                   stdcall ;
  {$ENDIF}
                                   abstract ;
                      end ;

type TObject = object
                   private { Instance data... }
  {$IFDEF CC_Debug}
                        Sanity : string[ 7 ] ;
  {$ENDIF}

                   public { Constructors and destructors... }
                        constructor Init ;
                        procedure Free ;
{$ELSE}
type TObject = object( Objects.TObject )
                   private { Instance data... }
  {$IFDEF CC_Debug}
                        Sanity : string[ 7 ] ;
  {$ENDIF}

                   public { Constructors and destructors... }
                        constructor Init ;
{$ENDIF}
                        destructor Done ; virtual ;

                   public { API... }
                        procedure Attach ; virtual ;
                        procedure Detach ; virtual ;
                        function Classname : string ; virtual ;
                        function Is_Class( const Name : string ) : boolean ;
                            virtual ;
                        function Get_Memory_Type : TMemory_Type ; virtual ;
               end ; { TObject }
     PObject = ^TObject ;

procedure Abstract_Error ;


implementation

{$IFDEF CC_Track_Resources}
  {$DEFINE FPC_OR_DELPHI_OR_CC_Track_Resources}
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
  {$DEFINE FPC_OR_DELPHI_OR_CC_Track_Resources}
{$ENDIF}

{$IFDEF FPC_OR_DELPHI_OR_CC_Track_Resources}
uses { Borland... }
{$ENDIF}
{$IFDEF CC_Track_Resources}
     ResMon
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
  {$IFDEF CC_Track_Resources}
     ,
  {$ENDIF}
     SysUtils { Exception }
{$ENDIF}
{$IFDEF FPC_OR_DELPHI_OR_CC_Track_Resources}
     ;
{$ENDIF}

{ Unit API... }

procedure Abstract_Error ;

begin
{$IFDEF EXCEPTIONS}
    raise Exception.Create( 'Call to abstract method' ) ;
{$ELSE}
    writeln( 'Call to abstract method' ) ;
    halt ;
{$ENDIF}
end ;


{ TObject methods... }

constructor TObject.Init ;

begin
{$IFDEF BP}
    inherited Init ;
{$ENDIF}
{$IFDEF CC_Track_Resources}
     Add_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}
{$IFDEF CC_Debug}
    Sanity := 'TObject' ;
{$ENDIF}
end ;


destructor TObject.Done ;

begin
{$IFDEF CC_Debug}
    Sanity := '       ' ;
{$ENDIF}
{$IFDEF CC_Track_Resources}
     Remove_Resource( RT_Object, integer( self ) ) ;
{$ENDIF}
end ;


{$IFDEF FPC_OR_DELPHI}
procedure TObject.Free ;

begin
    if( @self <> nil ) then
    begin
  {$IFDEF CC_DEBUG}
        if( Sanity <> 'TObject' ) then
        begin
            writeln( 'Invalid object' ) ;
        end ;
  {$ENDIF}
{$IFDEF FPC}
      if( Get_Memory_Type = MT_Heap ) then
      begin
          dispose( PObject( @self ), Done ) ;
      end else
      begin
          Done ;
      end ;
{$ELSE}
      Done ;
  {$IFDEF DELPHI}
      if( Get_Memory_Type = MT_Heap ) then
      begin
          dispose( PObject( @self ) ) ;
      end ;
  {$ENDIF}
{$ENDIF}
    end ;
end ;
{$ENDIF}


procedure TObject.Attach ;

begin
{$IFDEF CC_DEBUG}
    if( Sanity <> 'TObject' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Abstract_Error ;
end ;


procedure TObject.Detach ;

begin
{$IFDEF CC_Debug}
    if( Sanity <> 'TObject' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Abstract_Error ;
end ;


function TObject.Classname : string ;

begin
{$IFDEF CC_Debug}
    if( Sanity <> 'TObject' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Classname := 'TObject' ;
end ;


function TObject.Is_Class( const Name : string ) : boolean ;

begin
{$IFDEF CC_Debug}
    if( Sanity <> 'TObject' ) then
    begin
        writeln( 'Invalid object' ) ;
    end ;
{$ENDIF}
    Is_Class := ( Name = Classname ) ;
end ;


function TObject.Get_Memory_Type : TMemory_type ;

begin
    Result := MT_Unknown ;
end ;



end.

