{
        Program Name : _DebugMa
        Package Name : Subroutine Library - Layer 2
        Purpose      : Debug Manager pure virtual classes
        Institution  : Conroy & Conroy Co.
        Date Written : 15-August-2003
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2003 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The classes in this unit define the Debug Manager interface.
}

unit _DebugMa ;

interface

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

uses { Layer 0... }
     CommBase, { TBase_Object }
     ComInter { TCommon_COM_Interface }
{$IFNDEF MOBILE}
     ,_DebugIn { PDebug_Interface }
{$ENDIF}
     ,_Types { Ansistring }
     ;

{$IFDEF DYNAMIC_OBJECTS}
type TUnit_Debugger = class( TBase_Delphi_Object )
                          public { API... }
                              { Returns the name of the unit }
                              function Get_Name : Ansistring ;
                                  virtual ; abstract ;

                              { Creates and returns an instance of a unit debug
                                interface }
{$IFNDEF MOBILE}
                              function Get_Instance : PSDebug_Interface ;
                                  virtual ; abstract ;
{$ENDIF}
                      end ;

type TDebug_Manager = class( TBase_Delphi_Object )
                          public { Unit debugger support... }
                              function Get_Unit_Debugger( Index : longint ) : TUnit_Debugger ;
                                  virtual ; abstract ;

                              { Register a unit debugger }
                              procedure Register_Unit_Debugger( D : TUnit_Debugger ) ;
                                  virtual ; abstract ;

                              { Remove a registered unit debugger }
                              procedure Remove_Unit_Debugger( D : TUnit_Debugger ) ;
                                  virtual ; abstract ;

                              { Returns count of registered unit debuggers }
                              function Unit_Debugger_Count : longint ;
                                  virtual ; abstract ;
                      end ;

type TCOM_Unit_Debugger = class( TCommon_COM_Interface )
                              public { API... }
                                  { Returns the name of the unit }
                                  function Get_Name : PAnsiChar ;
                                      virtual ; stdcall ; abstract ;

                                  { Creates and returns an instance of a unit
                                    debug interface }
{$IFNDEF MOBILE}
                                  function Get_Instance : TDebug_Interface ;
                                      virtual ; stdcall ; abstract ;
{$ENDIF}
                          end ;

type TCOM_Debug_Manager = class( TCommon_COM_Interface )
                              public { Unit debugger support... }
                                  function Get_Unit_Debugger( Index : longint ) : TCOM_Unit_Debugger ;
                                      virtual ; stdcall ; abstract ;

                                  { Register a unit debugger }
                                  procedure Register_Unit_Debugger( D : TCOM_Unit_Debugger ) ;
                                      virtual ; stdcall ; abstract ;

                                  { Remove a registered unit debugger }
                                  procedure Remove_Unit_Debugger( D : TCOM_Unit_Debugger ) ;
                                      virtual ; stdcall ; abstract ;

                                  { Returns count of registered unit debuggers }
                                  function Unit_Debugger_Count : longint ;
                                      virtual ; stdcall ; abstract ;
                          end ;
{$ENDIF}

type TSUnit_Debugger = object( TBase_Object )
                          public { API... }
                              { Returns the name of the unit }
                              function Get_Name : Ansistring ; virtual ;

                              { Creates and returns an instance of a unit debug
                                interface }
{$IFNDEF MOBILE}
                              function Get_Instance : PSDebug_Interface ;
                                  virtual ;
{$ENDIF}
                      end ;
     PSUnit_Debugger = ^TSUnit_Debugger ;

type TSDebug_Manager = object( TBase_Object )
                          public { Unit debugger support... }
                              function Get_Unit_Debugger( Index : longint ) : PSUnit_Debugger ;
                                  virtual ;

                              { Register a unit debugger }
                              procedure Register_Unit_Debugger( D : PSUnit_Debugger ) ;
                                  virtual ;

                              { Remove a registered unit debugger }
                              procedure Remove_Unit_Debugger( D : PSUnit_Debugger ) ;
                                  virtual ;

                              { Returns count of registered unit debuggers }
                              function Unit_Debugger_Count : longint ;
                                  virtual ;
                      end ;
     PSDebug_Manager = ^TSDebug_Manager ;

procedure Register_Unit_Debugger( X : PSUnit_Debugger ) ;

var _Debug_Manager : PSDebug_Manager ;

implementation

uses { Layer 1... }
     _Applic, { Application_Manager }
     Collect ; { TCollection }

var Pending_Unit_Debuggers : PCollection
{$IFDEF VARIABLE_DECLARATION_ASSIGN}
    = nil
{$ENDIF}
    ;

procedure Register_Unit_Debugger( X : PSUnit_Debugger ) ;

var Loop : integer ;

begin
    if(
        ( Application_Manager = nil )
        or
        ( Application_Manager.Debug_Manager = nil )
      ) then
    begin
        if( Pending_Unit_Debuggers = nil ) then
        begin
            Pending_Unit_Debuggers := new( PCollection, Init( 0, 4 ) ) ;
        end ;
        Pending_Unit_Debuggers.Insert( X ) ;
    end else
    begin
        if( Pending_Unit_Debuggers <> nil ) then
        begin
            for Loop := 0 to Pending_Unit_Debuggers.Count - 1 do
            begin
                Application_Manager.Debug_Manager.Register_Unit_Debugger( PSUnit_Debugger( Pending_Unit_Debuggers.At( Loop ) ) ) ;
            end ;
            Pending_Unit_Debuggers.DeleteAll ;
        end ;
        if( X <> nil ) then
        begin
            Application_Manager.Debug_Manager.Register_Unit_Debugger( X ) ;
        end ;
    end ;
end ;


// TSUnit_Debugger methods...

function TSUnit_Debugger.Get_Name : Ansistring ;

begin
    Get_Name := '' ;
end ;


{$IFNDEF MOBILE}
function TSUnit_Debugger.Get_Instance : PSDebug_Interface ;

begin
    Get_Instance := nil ;
end ;
{$ENDIF}


function TSDebug_Manager.Get_Unit_Debugger( Index : longint ) : PSUnit_Debugger ;

begin
    Get_Unit_Debugger := nil ;
end ;


procedure TSDebug_Manager.Register_Unit_Debugger( D : PSUnit_Debugger ) ;

begin
    { Intentionally left blank }
end ;


procedure TSDebug_Manager.Remove_Unit_Debugger( D : PSUnit_Debugger ) ;

begin
    { Intentionally left blank }
end ;


function TSDebug_Manager.Unit_Debugger_Count : longint ;

begin
    Unit_Debugger_Count := 0 ;
end ;


begin
{$IFNDEF VARIABLE_DECLARATION_ASSIGN}
    Pending_Unit_Debuggers := nil ;
{$ENDIF}
end.
