{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
          Program Name : _FiP
          Package Name : Subroutine Library - Layer 1
          Purpose      : File Processor abstract class definition
          Institution  : Conroy & Conroy Co.
          Date Written : 30-Nov-88
          Written By   : Alan Conroy
          Version      : 2.0D

          TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2008 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the abstract File_Processor class and the classes
        for registering file system servers, etc.
}

unit _FiP ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I eDefines.inc}

{$IFDEF FPC_OR_DELPHI}
  {$DEFINE WINDOWS}
{$ENDIF}

uses { Layer 0... }
     CommBase, { TBase_Object }
     ComInter, { TCommon_COM_Object }
{$IFNDEF MOBILE}
     _DebugIn, { PDebug_Interface }
{$ENDIF}

     { Layer 1... }
     _Files, { PSFile }
     _FileSys, { TFind_Record }
     _Stores, { PSStore_Server }
     _Types, { int64 }
     _UE ; { PSUnified_Exception }


const { File Open modes... }
      FSOM_Default = 0 ; { Open file if there, create it if not. }
      FSOM_Append = 1 ; { Append mode (file must already exist) }
      FSOM_Input = 2 ; { File must already exist }
      FSOM_Output = 3 ; { File is created (deleted if existing) }
      FSOM_Text = 4 ; { File is text mode }
      { None of the following is currently supported... }
      FSOM_Read_Only = 8 ; { File cannot be written to }
      FSOM_Share_Read = 16 ; { File can be opened read-only by others }
      FSOM_Share_Write = 32 ; { File can be opened writable by others }
      { FS Open Mode 64+ are reserved for future use}

      { Filename query flags... }
      TFQF_Read_Only = 1 ; { Open as read-only }
      TFQF_Overwrite = 2 ; { Prompt for overwrite on existing files }
      TFQF_No_Validation = 4 ; { Do not validate filename }
      TFQF_Multiple_Files = 8 ; { Allow multiple filename to be returned }
      TFQF_Path_Must_Exist = 16 ; { Path must exist }
      TFQF_File_Must_Exist = 32 ; { File must exist }
      TFQF_Prompt_On_Creation = 64 ; { Prompt if file doesn't exist }
      TFQF_No_Read_Only = 128 ; { Don't allow read-only files }
      TFQF_No_Test = 256 ; { Disables checking for network file protection and inaccessibility of disk drives (when the user tries to save a file in a create-no-modify shared network directory) }
      TFQF_No_Dereference = 512 ; { Disable dereferencing of file links }
      TFQF_Show_All = 1024 ; { Show all files }
      {2048+ = reserved}

      { Filename query modes... }
type TFilename_Query_Mode = ( TFNQM_Open, TFNQM_Save ) ;

type TSFile_Processor = object( TBase_Object )
                          public { API... }
                            destructor Done ; virtual ;
                            function Lookup( Filename : Ansistring ;
                                var Find_Record : TFind_Record ) : Ansistring ;
                                virtual ;
                            function Open( const Name : Ansistring ; Mode : integer ;
                                var UEC : PSUnified_Exception ) : PSFile ; virtual ;
                            procedure Initialize( const Nam, ID : Ansistring ;
                                Flags : longint ; var UEC : PSUnified_Exception ) ; virtual ;
                            function DeviceSize( const Nam : Ansistring ) : int64 ;
                                virtual ;
                            function DeviceFree( const Nam : Ansistring ) : int64 ;
                                virtual ;
                            function MKDir( const Nam : Ansistring ) : longint ;
                                virtual ;
                            procedure Close_File_System( Name : Ansistring ) ;
                                virtual ;
                            function Open_File_System( Name : Ansistring ;
                                Flags : integer ;
                                var Res : PSUnified_Exception ) : PSFile_System ; virtual ;
{$IFNDEF MOBILE}
                            function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}
                            function File_System_Open( Nam : Ansistring ) : boolean ;
                                virtual ;
                            { Scan a file specification and fill the Info record.  Separator is the path
                              separator to use in parsing. }
                            procedure Filespec_String_Scan( const Name, Separator : Ansistring ;
                                var Info : TFSSS ) ; virtual ;
                            function Query_Filename( Default, Prompt : Ansistring ;
                                Mode : TFilename_Query_Mode ;
                                Flags : integer ) : Ansistring ; virtual ;
                            procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;
                                virtual ;
                            procedure Register_Store_Server( SS : PSStore_Server ) ;
                                virtual ;
                            function File_Size( Nam : Ansistring ) : int64 ;
                                virtual ;
                       end ; { TSFile_Processor }
     PSFile_Processor = ^TSFile_Processor ;

{$IFDEF FPC_OR_DELPHI}
type TFile_Processor = class( TBase_Delphi_Object )
                          public { API... }
                            function Lookup( Filename : Ansistring ;
                                var Find_Record : TFind_Record ) : Ansistring ;
                                virtual ; abstract ;
                            function Open( const Name : Ansistring ; Mode : integer ;
                                var UEC : TUnified_Exception ) : PSFile ; virtual ; abstract ;
                            procedure Initialize( const Nam, ID : Ansistring ;
                                Flags : longint ; var UEC : TUnified_Exception ) ;
                                virtual ; abstract ;
                            function DeviceSize( const Nam : Ansistring ) : int64 ;
                                virtual ; abstract ;
                            function DeviceFree( const Nam : Ansistring ) : int64 ;
                                virtual ; abstract ;
                            function MKDir( const Nam : Ansistring ) : longint ;
                                virtual ; abstract ;
                            procedure Close_File_System( Name : Ansistring ) ;
                                virtual ; abstract ;
                            function Open_File_System( Name : Ansistring ;
                                Flags : integer ;
                                var Res : TUnified_Exception ) : PSFile_System ;
                                virtual ; abstract ;
                            function File_System_Open( Nam : Ansistring ) : boolean ;
                                virtual ; abstract ;
                            { Scan a file specification and fill the Info record.  Separator is the path
                              separator to use in parsing. }
                            procedure Filespec_String_Scan( const Name, Separator : Ansistring ;
                                var Info : TFSSS ) ; virtual ; abstract ;
                            function Query_Filename( Default, Prompt : Ansistring ;
                                Mode : TFilename_Query_Mode ;
                                Flags : integer ) : Ansistring ; virtual ; abstract ;
                            procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;
                                virtual ; abstract ;
                            procedure Register_Store_Server( SS : PSStore_Server ) ;
                                virtual ; abstract ;
                            function File_Size( Nam : Ansistring ) : int64 ;
                                virtual ; abstract ;
                       end ; { TFile_Processor }

type TCOM_File_Processor = class( TCommon_COM_Interface )
                              public { API... }
                                function Lookup( Filename : PAnsiChar ;
                                    var Find_Record : TFind_Record ) : PAnsiChar ;
                                    virtual ; stdcall ; abstract ;
                                function Open( const Name : PAnsiChar ; Mode : integer ;
                                    var UEC : PSUnified_Exception ) : PSFile ;
                                    virtual ; stdcall ; abstract ;
                                procedure _Initialize( const Nam, ID : PAnsiChar ;
                                    Flags : longint ; var UEC : PSUnified_Exception ) ;
                                    virtual ; stdcall ; abstract ;
                                function DeviceSize( Nam : PAnsiChar ) : int64 ;
                                    virtual ; stdcall ; abstract ;
                                function DeviceFree( Nam : PAnsiChar ) : int64 ;
                                    virtual ; stdcall ; abstract ;
                                function MKDir( Nam : PAnsiChar ) : longint ;
                                    virtual ; stdcall ; abstract ;
                                procedure Close_File_System( Name : PAnsiChar ) ;
                                    virtual ; stdcall ; abstract ;
                                function Open_File_System( Name : PAnsiChar ;
                                    Flags : integer ;
                                    var Res : PSUnified_Exception ) : PSFile_System ;
                                    virtual ; stdcall ; abstract ;
                                function File_System_Open( Nam : PAnsiChar ) : boolean ;
                                    virtual ; stdcall ; abstract ;
                                { Scan a file specification and fill the Info record.  Separator is the path
                                  separator to use in parsing. }
                                procedure Filespec_String_Scan( Name, Separator : PAnsiChar ;
                                    var Info : TFSSS ) ; virtual ; stdcall ; abstract ;
                                function Query_Filename( Default, Prompt : PAnsiChar ;
                                    Mode : TFilename_Query_Mode ;
                                    Flags : integer ) : PAnsiChar ; virtual ; stdcall ; abstract ;
                                procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;
                                    virtual ; stdcall ; abstract ;
                                procedure Register_Store_Server( SS : PSStore_Server ) ;
                                    virtual ; stdcall ; abstract ;
                                function File_Size( Nam : PAnsiChar ) : int64 ;
                                    virtual ; stdcall ; abstract ;
                           end ; { TCOM_File_Processor }
{$ENDIF}

var File_Processor : PSFile_Processor
{$IFDEF FPC_OR_DELPHI}
 = nil
{$ENDIF}
 ;

{ Scan a file specification and fill the Info record.  Separator is the path
  separator to use in parsing. }
procedure Filespec_String_Scan( const Name, Separator : Ansistring ;
    var Info : TFSSS ) ;
procedure Simple_Filespec_String_Scan( const Name, Separator : Ansistring ;
    var Info : TFSSS ) ;
procedure Close_File_System( const Name : Ansistring ) ;
function Open_File_System( const Name : Ansistring ; Flags : integer ;
    var Res : PSUnified_Exception ) : PSFile_System ;
function Open( const Name : Ansistring ; Mode : integer ; var UEC : PSUnified_Exception ) : PSFile ;
function Lookup( const Filename : Ansistring ; var Find_Record : TFind_Record ) : Ansistring ;
function Query_Filename( const Default, Prompt : Ansistring ; Mode : TFilename_Query_Mode ;
    Flags : integer ) : Ansistring ;
procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;
procedure Register_Store_Server( SS : PSStore_Server ) ;
procedure Initialize_File_System( const Nam, ID : Ansistring ; Flags : longint ;
    var UEC : PSUnified_Exception ) ;
function DeviceSize( const Nam : Ansistring ) : int64 ;
function DeviceFree( const Nam : Ansistring ) : int64 ;
function MKDir( const Nam : Ansistring ) : longint ;


implementation

uses { Subroutine Library - Layer 0... }
     _Common ; { Abstract_Error }

{ Unit API... }

procedure Filespec_String_Scan( const Name, Separator : Ansistring ;
    var Info : TFSSS ) ;

begin
    File_Processor^.Filespec_String_Scan( Name, Separator, Info ) ;
end ;


procedure Close_File_System( const Name : Ansistring ) ;

begin
    File_Processor^.Close_File_System( Name ) ;
end ;


function Open_File_System( const Name : Ansistring ; Flags : integer ;
    var Res : PSUnified_Exception ) : PSFile_System ;

begin
    Open_File_System := File_Processor^.Open_File_System( Name, Flags, Res ) ;
end ;


function Open( const Name : Ansistring ; Mode : integer ;
    var UEC : PSUnified_Exception ) : PSFile ;

begin
    Open := File_Processor^.Open( Name, Mode, UEC ) ;
end ;


function Lookup( const Filename : Ansistring ; var Find_Record : TFind_Record ) : Ansistring ;

begin
    Lookup := File_Processor^.Lookup( Filename, Find_Record ) ;
end ;


function Query_Filename( const Default, Prompt : Ansistring ; Mode : TFilename_Query_Mode ;
    Flags : integer ) : Ansistring ;

begin
    Query_Filename := File_Processor^.Query_Filename( Default, Prompt, Mode, Flags ) ;
end ;


procedure Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;

begin
    File_Processor^.Register_File_System_Server( FSS ) ;
end ;


procedure Register_Store_Server( SS : PSStore_Server ) ;

begin
    File_Processor^.Register_Store_Server( SS ) ;
end ;


procedure Initialize_File_System( const Nam, ID : Ansistring ; Flags : longint ;
    var UEC : PSUnified_Exception ) ;

begin
    File_Processor^.Initialize( Nam, ID, Flags, UEC ) ;
end ;


function DeviceSize( const Nam : Ansistring ) : int64 ;

begin
    DeviceSize := File_Processor^.DeviceSize( Nam ) ;
end ;


function DeviceFree( const Nam : Ansistring ) : int64 ;

begin
    DeviceFree := File_Processor^.DeviceFree( Nam ) ;
end ;


function MKDir( const Nam : Ansistring ) : longint ;

begin
    MKDir := File_Processor^.MKDir( Nam ) ;
end ;



{ TSFile_Processor methods... }

{ Constructors and destructors... }

destructor TSFile_Processor.Done ;

begin
    inherited Done ;
    
    File_Processor := nil ;
end ;


{ API... }

function TSFile_Processor.Lookup( Filename : Ansistring ;
    var Find_Record : TFind_Record ) : Ansistring ;

begin
    Abstract_Error ;
end ;


function TSFile_Processor.Open_File_System( Name : Ansistring ;
    Flags : integer ; var Res : PSUnified_Exception ) : PSFile_System ;

begin
    fillchar( Res, sizeof( Res ), 0 ) ;
    Open_File_System := nil ;
    Abstract_Error ;
end ;


procedure TSFile_Processor.Close_File_System( Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


{$IFNDEF MOBILE}
function TSFile_Processor.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;
{$ENDIF}


function TSFile_Processor.File_System_Open( Nam : Ansistring ) : boolean ;

begin
    File_System_Open := False ;
    Abstract_Error ;
end ;


procedure TSFile_Processor.Filespec_String_Scan( const Name, Separator : Ansistring ;
    var Info : TFSSS ) ;

begin
    fillchar( Info, sizeof( Info ), 0 ) ;
    Abstract_Error ;
end ;


function TSFile_Processor.Query_Filename( Default, Prompt : Ansistring ;
    Mode : TFilename_Query_Mode ; Flags : integer ) : Ansistring ;

begin
    Query_Filename := '' ;
    Abstract_Error ;
end ;


procedure TSFile_Processor.Register_File_System_Server( FSS : PSDefault_File_System_Server ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile_Processor.Register_Store_Server( SS : PSStore_Server ) ;

begin
    Abstract_Error ;
end ;


function TSFile_Processor.Open( const Name : Ansistring ; Mode : integer ;
    var UEC : PSUnified_Exception ) : PSFile ;

begin
    Open := nil ;
    Abstract_Error ;
end ;


procedure TSFile_Processor.Initialize( const Nam, ID : Ansistring ; Flags : longint ;
    var UEC : PSUnified_Exception ) ;

begin
    fillchar( UEC, sizeof( UEC ), 0 ) ;
    Abstract_Error ;
end ;


function TSFile_Processor.DeviceSize( const Nam : Ansistring ) : int64 ;

begin
    DeviceSize := -1 ;
    Abstract_Error ;
end ;


function TSFile_Processor.DeviceFree( const Nam : Ansistring ) : int64 ;

begin
    DeviceFree := -1 ;
    Abstract_Error ;
end ;


function TSFile_Processor.MKDir( const Nam : Ansistring ) : longint ;

begin
    MKDir := -1 ;
    Abstract_Error ;
end ;


function TSFile_Processor.File_Size( Nam : Ansistring ) : int64 ;

begin
    File_Size := -1 ;
    Abstract_Error ;
end ;


procedure Terminate ;

begin
    if( File_Processor <> nil ) then
    begin
        dispose( File_Processor, Done ) ;
        File_Processor := nil ;
    end ;
end ;


{ Scan a file specification and fill the Info record.  Separator is the path
  separator to use in parsing.  This is a simplified version of what the file
  processor supports. }
procedure Simple_Filespec_String_Scan( const Name, Separator : Ansistring ;
    var Info : TFSSS ) ;

var Dummy : integer ;
    _Name : Ansistring ;
    L : integer ;

begin
    { Setup... }
    fillchar( Info, sizeof( Info ), 0 ) ;
    _Name := Name ;

    { Parse name }
    L := length( Separator ) ;
    if( copy( _Name, 1, 2 * L ) = Separator + Separator ) then
    begin
        Dummy := 2 * L + 1 ;
        while(
               ( Dummy <= length( _Name ) )
               and
               ( copy( _Name, Dummy, L ) <> Separator )
             ) do
        begin
            Dummy := Dummy + 1 ;
        end ;
        Dummy := Dummy - 1 ; { Device is //name }
    end else
    begin
        Dummy := pos( ':', _Name ) ;
    end ;
    if( Dummy > 0 ) then
    begin
        Info.Device := copy( _Name, 1, Dummy ) ;
        _Name := copy( _Name, Dummy + 1, length( _Name ) ) ;
    end ;

    Dummy := length( _Name ) ;
    if( L = 1 ) then { Faster code for 1-character separators }
    begin
        while( ( Dummy > 0 ) and ( _Name[ Dummy ] <> Separator ) ) do
        begin
            dec( Dummy ) ;
        end ;
    end else
    begin
        while(
               ( Dummy > 0 )
               and
               ( copy( _Name, Dummy, L ) <> Separator )
             ) do
        begin
            dec( Dummy ) ;
        end ;
    end ;
    Info.Filename := System.copy( _Name, Dummy + 1, length( _Name ) ) ;
    Info.Path := copy( _Name, 1, Dummy ) ;
    if( Separator <> '/' ) then
    begin
        Dummy := pos( '/', Info.Filename ) ;
        if( Dummy = 0 ) then
        begin
            Dummy := length( Info.Filename ) + 1 ;
        end ;
        Info.Switches := copy( Info.Filename, Dummy, length( Info.Filename ) ) ;
        Info.Filename := copy( Info.Filename, 1, Dummy - 1 ) ;
    end ;
    Dummy := length( Info.Filename ) ;
    while( ( Dummy > 0 ) and ( Info.Filename[ Dummy ] <> '.' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    if( Dummy = 0 ) then { No extension }
    begin
        Dummy := length( Info.Filename ) + 1 ;
        Info.Flags := Info.Flags or FSSS_Null_Extension ;
    end ;
    Info.Name := copy( Info.Filename, 1, Dummy - 1 ) ;
    Info.Extension := copy( Info.Filename, Dummy + 1, length( Info.Filename ) ) ;
end ;


{$IFNDEF FPC_OR_DELPHI}
begin
    { Initialize data... }
    File_Processor_Exists := False ;
{$ENDIF}

{$IFDEF FPC_OR_DELPHI}
initialization

finalization
    Terminate ;
{$ENDIF}
end.
