<?php
 /*       
        Program Name : _GenVal
        Package Name : Subroutine Library - Layer 0
        Purpose      : Generic Value class definitions.
        Institution  : Conroy & Conroy Co.
        Date Written : 14-February-2005
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2005-2021 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TGeneric_Value is a class that provides a generic value (ie variant).
        There are three intended uses for this (and associated) classes:
          1) RTTI support
          2) Compiler/debugger symbol support
          3) Uber-variant
*/

include_once "_Common.php";

// TData_Type_Family
const DTF_Unknown = 0 ; // Variant
const DTF_Undefined = 1 ; // Binary
const DTF_Logical = 2 ;
const DTF_Numeric = 3 ;
const DTF_Character = 4 ; // Character/string
const DTF_Pointer = 5 ; // pointer/method/procedure/function
const DTF_Composite = 6 ;
const DTF_Collection = 7 ;
const DTF_User_Defined = 8 ;
const DTF_Meta = 9 ; // Type
const DTF_Other = 10 ;

// TValue_Data_Type
const VDT_Empty = 0 ;
const VDT_Null = 1 ;
const VDT_Boolean = 2 ;
const VDT_Integer = 3 ;
const VDT_Real = 4 ;
const VDT_String = 5 ;
const VDT_Char = 6 ;
const VDT_Pointer = 7 ;
const VDT_Value = 8 ; // TGeneric_Value defining a type
const VDT_Code = 9 ; // Function or procedure
const VDT_Scalar = 10 ;
const VDT_User_Defined = 11 ;
const VDT_Unknown = 12 ;

// TCall_Mechanism
const CM_By_Value = 0 ;
const CM_By_Reference = 1 ;
const CM_By_Name = 2 ;

// TMethod_Class
const MC_Private = 0 ;
const MC_Protected = 1 ;
const MC_Public = 2 ;
const MC_Published = 3 ;
                   
// TSGeneric_Parameter methods...
class TSGeneric_Parameter
{
	function Typ()
	{
   		Abstract_Error() ;
    	return NULL;
	}


	function Output()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Constant()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Mechanism()
	{
	    Abstract_Error() ;
	    return CM_By_Value;
	}


	function _Default()
	{
	    Abstract_Error() ;
	    return NULL;
	}
} // TSGeneric_Parameter


// TSGeneric_Number methods...

class TSGeneric_Number
{
	function Fixed()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Signed()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Mantissa_Size()
	{
	    Abstract_Error() ;
	    return 0;
	} 


	function Exponent_Size()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Fractional_Size()
	{
	    Abstract_Error() ;
	    return 0;
	}
} // TSGeneric_Number


// TSGeneric_Character methods...

class TSGeneric_Character
{
	function Max_Length()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Fixed()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Length()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Logical_Length()
	{
	    Abstract_Error() ;
	    return 0;
	}
} // TSGeneric_Character



// TSGeneric_Collection methods...

class TSGeneric_Collection
{
	function Fixed()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Sparse()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Subscripts()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function High_Bound( $Subscript )
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Low_Bound( $Subscript ) 
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Get_Subscript( $Address ) 
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Row_Major()
	{
	    Abstract_Error() ;
	    return true;
	}
} // TSGeneric_Collection


// TSGeneric_Composite methods...

class TSGeneric_Composite
{
	function Ancestor( $index )
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Ancestor_Count()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Ancestor_Class( $Index )
	{
	    Abstract_Error() ;
	    return MC_Private;
	}


	function Count()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Get_Item( $Index ) 
	{
	    Abstract_Error() ;
	    return NULL;
	}
} // TSGeneric_Composite


// TSGeneric_Method methods...

class TSGeneric_Method
{
	function Is_Abstract()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Is_Constructor()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Is_Destructor()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Is_Virtual()
	{
	    Abstract_Error() ;
	    return false;
	}


	function _Static()
	{
	    Abstract_Error() ;
	    return false;
	}


	function C_Order()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Standard()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Execute( $Parameters ) 
	{
   	 Abstract_Error() ;
   	 return NULL;
	}


	function Compile_Time()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Count()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Max_Count()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Parameter( $Index )
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Target()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Method_Class()
	{
	    Abstract_Error();
		return MC_Public;
	}
} // TSGeneric_Method


// TSProperty_Information methods...

class TSProperty_Information
{
	
	function Getter()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Setter()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function _Default()
	{
	    Abstract_Error() ;
	    return false;
	}
}


// TSType_Information methods...

class TSType_Information
{
	function Is_Composite()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Is_Collection()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Memory_Type()
	{
	    Abstract_Error() ;
	    return MT_Unknown;
	}


	function Parameter()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Numeric_Information()
	{
   		Abstract_Error() ;
   		return NULL;
	}


	function Character_Information()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Composite_Information()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Collection_Information()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Method()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function _Property()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Address()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Range_Limited()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Low_Range()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function High_Range()
	{
	   	Abstract_Error() ;
	    return NULL;
	}


	function Value_Name( $Value )
	{
	    Abstract_Error() ;
		return "";
	}
} // TSType_Information


// TSGeneric_Value methods...

class TSGeneric_Value
{
	function Family() 
	{
	    Abstract_Error() ;
	    return DTF_Unknown;
	}


	function Data_Type()
	{
	    Abstract_Error() ;
	    return VDT_Unknown;
	}


	function User_Data_Type_Name()
	{
	    Abstract_Error() ;
	    return "";
	}


	function Name() // Variable, type, method, or parameter name
	{
	    Abstract_Error() ;
	    return "";
	}


	function Type_Information()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Constant()
	{
	    Abstract_Error() ;
	    return true;
	}


	function Size() // Size in bytes
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Total_Size()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Dereference()
	{
	    Abstract_Error() ;
		return NULL;
	}


	function Can_Convert( $Typ )
	{
	    Abstract_Error() ;
	    return false;
	}


	function Convert( $Class_Type, &$Loss )
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function _Clone()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Simple_Cast( $Typ )
	{
	    Abstract_Error() ;
	}


	function Cast( $Typ )
	{
	    Abstract_Error() ;
	}


	function Deserialize( $Value )
	{
	    Abstract_Error() ;
	}


	function Serialize()
	{
	    Abstract_Error() ;
	    return NULL;
	}


	function Set_Null()
	{
   		Abstract_Error() ;
	}


	function Get_Undefined( $Buffer, $Len )
	{
	    Abstract_Error() ;
	}


	function Set_Undefined( $Buffer, $Len )
	{
	    Abstract_Error() ;
	}


	function Get_Boolean()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Set_Boolean( $B )
	{
	    Abstract_Error() ;
	}


	function Get_ByteBool()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Set_ByteBool( $B )
	{
	    Abstract_Error() ;
	}


	function Get_WordBool()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Set_WordBool( $B )
	{
	    Abstract_Error() ;
	}


	function Get_LongBool()
	{
	    Abstract_Error() ;
	    return false;
	}


	function Set_LongBool( $B )
	{
	    Abstract_Error() ;
	}


	function Get_Char()
	{
	    Abstract_Error() ;
	    return '\0';
	}


	function Set_Char( $C )
	{
	    Abstract_Error() ;
	}


	function Get_String()
	{
	    Abstract_Error() ;
	    return "";
	}


	function Set_String( $S )
	{
	   	Abstract_Error() ;
	}


	function Get_PChar()
	{
	   	Abstract_Error() ;
	   	return NULL;
	}


	function Set_PChar( $P )
	{
	    Abstract_Error() ;
	}


	function Get_Integer8()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Integer8( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Integer16()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Integer16( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Integer32()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Integer32( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Integer64()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Integer64( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Unsigned_Integer8()
	{
	   	Abstract_Error() ;
	   	return 0;
	}


	function Set_Unsigned_Integer8( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Unsigned_Integer16()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Unsigned_Integer16( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Unsigned_Integer32()
	{
	    Abstract_Error() ;
	    return 0;
	}


	function Set_Unsigned_Integer32( $I )
	{
	    Abstract_Error() ;
	}


	function Get_Single()
	{
	   	Abstract_Error() ;
	    return 0.0;
	}


	function Set_Single( $S )
	{
	    Abstract_Error() ;
	}


	function Get_Double()
	{
	    Abstract_Error() ;
	    return 0.0;
	}


	function Set_Double( $D )
	{
	    Abstract_Error() ;
	}


	function Get_Extended()
	{
	    Abstract_Error() ;
	    return 0.0;
	}


	function Set_Extended( $E )
	{
	    Abstract_Error() ;
	}
} // TSGeneric_Value
