#        Program Name : _GenVal
#        Package Name : Subroutine Library - Layer 0
#        Purpose      : Generic Value class definitions.
#        Institution  : Conroy & Conroy Co.
#        Date Written : 14-February-2005
#        Written By   : Alan Conroy
#        Version      : 1.0
#
#        Copyright (C) 2005-2021 by Alan Conroy.  Released to the Public Domain.
#
#        TO THE GLORY OF GOD THROUGH JESUS CHRIST
#
#        *************************************************************
#        *                                                           *
#        *        M O D I F I C A T I O N   H I S T O R Y            *
#        *                                                           *
#        *************************************************************
#
#            DATE        BY          REASON
#
#        *************************************************************
#        *                                                           *
#        *        P R O G R A M   P U R P O S E                      *
#        *                                                           *
#        *************************************************************
#
#          TGeneric_Value is a class that provides a generic value (ie variant).
#        There are three intended uses for this (and associated) classes:
#          1) RTTI support
#          2) Compiler/debugger symbol support
#          3) Uber-variant

import _Common

# TData_Type_Family
DTF_Unknown = 0 # Variant
DTF_Undefined = 1 # Binary
DTF_Logical = 2
DTF_Numeric = 3
DTF_Character = 4 # Character/string
DTF_Pointer = 5 # pointer/method/procedure/function
DTF_Composite = 6
DTF_Collection = 7
DTF_User_Defined = 8
DTF_Meta = 9 # Type
DTF_Other = 10

# TValue_Data_Type
VDT_Empty = 0
VDT_Null = 1
VDT_Boolean = 2
VDT_Integer = 3
VDT_Real = 4
VDT_String = 5
VDT_Char = 6
VDT_Pointer = 7
VDT_Value = 8 # TGeneric_Value defining a type
VDT_Code = 9 # def or procedure
VDT_Scalar = 10
VDT_User_Defined = 11
VDT_Unknown = 12

# TCall_Mechanism
CM_By_Value = 0
CM_By_Reference = 1
CM_By_Name = 2

# TMethod_Class
MC_Private = 0
MC_Protected = 1
MC_Public = 2
MC_Published = 3
                   
# TSGeneric_Parameter methods...
class TSGeneric_Parameter:
    def Typ():
            Abstract_Error()
            return NULL


    def Output():
            Abstract_Error()
            return false


    def Constant():
            Abstract_Error()
            return true


    def Mechanism():
        Abstract_Error()
        return CM_By_Value


    def _Default():
        Abstract_Error()
        return NULL


# TSGeneric_Number methods...

class TSGeneric_Number:
    def Fixed():
        Abstract_Error()
        return true


    def Signed():
        Abstract_Error()
        return true


    def Mantissa_Size():
        Abstract_Error()
        return 0


    def Exponent_Size():
        Abstract_Error()
        return 0


    def Fractional_Size():
        Abstract_Error()
        return 0


# TSGeneric_Character methods...

class TSGeneric_Character:
    def Max_Length():
        Abstract_Error()
        return 0


    def Fixed():
        Abstract_Error()
        return true


    def Length():
        Abstract_Error()
        return 0


    def Logical_Length():
        Abstract_Error()
        return 0


# TSGeneric_Collection methods...

class TSGeneric_Collection:
    def Fixed():
        Abstract_Error()
        return true


    def Sparse():
        Abstract_Error()
        return false


    def Subscripts():
        Abstract_Error()
        return 0


    def High_Bound( Subscript ):
        Abstract_Error()
        return NULL


    def Low_Bound( Subscript ):
        Abstract_Error()
        return NULL


    def Get_Subscript( Address ):
        Abstract_Error()
        return NULL


    def Row_Major():
        Abstract_Error()
        return true


# TSGeneric_Composite methods...

class TSGeneric_Composite:
    def Ancestor( index ):
        Abstract_Error()
        return NULL


    def Ancestor_Count():
        Abstract_Error()
        return 0


    def Ancestor_Class( Index ):
        Abstract_Error()
        return MC_Private


    def Count():
        Abstract_Error()
        return 0


    def Get_Item( Index ):
        Abstract_Error()
        return NULL


# TSGeneric_Method methods...

class TSGeneric_Method:
    def Is_Abstract():
        Abstract_Error()
        return false


    def Is_Constructor():
        Abstract_Error()
        return false


    def Is_Destructor():
        Abstract_Error()
        return false


    def Is_Virtual():
        Abstract_Error()
        return false


    def _Static():
        Abstract_Error()
        return false            


    def C_Order():
        Abstract_Error()
        return false


    def Standard():
        Abstract_Error()
        return true


    def Execute( Parameters ):
            Abstract_Error()
            return NULL


    def Compile_Time():
        Abstract_Error()
        return false


    def Count():
        Abstract_Error()
        return 0


    def Max_Count():
        Abstract_Error()
        return 0


    def Parameter( Index ):
        Abstract_Error()
        return NULL


    def Target():
        Abstract_Error()
        return false


    def Method_Class():
        Abstract_Error()
        return MC_Public


# TSProperty_Information methods...

class TSProperty_Information:
    def Getter():
        Abstract_Error()
        return NULL;


    def Setter():
        Abstract_Error()
        return NULL;


    def _Default():
        Abstract_Error()
        return false;


# TSType_Information methods...

class TSType_Information:
    def Is_Composite():
        Abstract_Error()
        return false


    def Is_Collection():
        Abstract_Error()
        return false


    def Memory_Type():
        Abstract_Error()
        return MT_Unknown


    def Parameter():
        Abstract_Error()
        return NULL


    def Numeric_Information():
        Abstract_Error()
        return NULL


    def Character_Information():
        Abstract_Error()
        return NULL


    def Composite_Information():
        Abstract_Error()
        return NULL


    def Collection_Information():
        Abstract_Error()
        return NULL


    def Method():
        Abstract_Error()
        return NULL


    def _Property():
        Abstract_Error()
        return NULL


    def Address():
        Abstract_Error()
        return 0


    def Range_Limited():
        Abstract_Error()
        return false


    def Low_Range():
        Abstract_Error()
        return NULL


    def High_Range():
        Abstract_Error()
        return NULL


    def Value_Name( Value ):
        Abstract_Error()
        return ""


# TSGeneric_Value methods...

class TSGeneric_Value:
    def Family():
        Abstract_Error()
        return DTF_Unknown


    def Data_Type():
        Abstract_Error()
        return VDT_Unknown


    def User_Data_Type_Name():
        Abstract_Error()
        return ""


    def Name(): # Variable, type, method, or parameter name
        Abstract_Error()
        return ""


    def Type_Information():
        Abstract_Error()
        return NULL


    def Constant():
        Abstract_Error()
        return true


    def Size(): # Size in bytes
        Abstract_Error()
        return 0


    def Total_Size():
        Abstract_Error()
        return 0


    def Dereference():
        Abstract_Error()
        return NULL


    def Can_Convert( Typ ):
            Abstract_Error()
            return false


    def Convert( Class_Type, Loss ):
            Abstract_Error()
            return NULL


    def _Clone():
        Abstract_Error()
        return NULL


    def Simple_Cast( Typ ):
        Abstract_Error()


    def Cast( Typ ):
        Abstract_Error()


    def Deserialize( Value ):
        Abstract_Error()



    def Serialize():
        Abstract_Error()
        return NULL


    def Set_Null():
        Abstract_Error()


    def Get_Undefined( Buffer, Len ):
        Abstract_Error()


    def Set_Undefined( Buffer, Len ):
        Abstract_Error()


    def Get_Boolean():
        Abstract_Error()
        return false


    def Set_Boolean( B ):
        Abstract_Error()


    def Get_ByteBool():
        Abstract_Error()
        return false


    def Set_ByteBool( B ):
        Abstract_Error()


    def Get_WordBool():
        Abstract_Error()
        return false


    def Set_WordBool( B ):
        Abstract_Error()


    def Get_LongBool():
        Abstract_Error()
        return false


    def Set_LongBool( B ):
        Abstract_Error()


    def Get_Char():
        Abstract_Error()
        return '\0'


    def Set_Char( C ):
        Abstract_Error()


    def Get_String():
        Abstract_Error()
        return ""


    def Set_String( S ):
        Abstract_Error()



    def Get_PChar():
        Abstract_Error()
        return NULL


    def Set_PChar( P ):
        Abstract_Error()


    def Get_Integer8():
        Abstract_Error()
        return 0


    def Set_Integer8( I ):
        Abstract_Error()


    def Get_Integer16():
        Abstract_Error()
        return 0


    def Set_Integer16( I ):
        Abstract_Error()


    def Get_Integer32():
        Abstract_Error()
        return 0


    def Set_Integer32( I ):
        Abstract_Error()


    def Get_Integer64():
        Abstract_Error()
        return 0


    def Set_Integer64( I ):
        Abstract_Error()


    def Get_Unsigned_Integer8():
        Abstract_Error()
        return 0


    def Set_Unsigned_Integer8( I ):
        Abstract_Error()


    def Get_Unsigned_Integer16():
        Abstract_Error()
        return 0


    def Set_Unsigned_Integer16( I ):
        Abstract_Error()


    def Get_Unsigned_Integer32():
        Abstract_Error()
        return 0


    def Set_Unsigned_Integer32( I ):
        Abstract_Error()


    def Get_Single():
        Abstract_Error()
        return 0.0


    def Set_Single( S ):
        Abstract_Error()


    def Get_Double():
        Abstract_Error()
        return 0.0


    def Set_Double( D ):
        Abstract_Error()


    def Get_Extended():
        Abstract_Error()
        return 0.0


    def Set_Extended( E ):
        Abstract_Error()

