{
        Program Name : SCU
        Package Name : SCU
        Purpose      : SCU definition
        Institution  : Conroy & Conroy Co.
        Date Written : 16-November-2001
        Written By   : Alan Conroy
        Version      : 1.0

	      Copyright (C) 2001 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************
}

unit _SCU ;

interface

{$I EDEFINES.INC}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

uses _DebugIn, { PDebug_Interface }
     _UE, { PSUnified_Exception }
     _Loggers ; { PLogger }

const SCUErr_Facility = 22 ;
      SCUErr_Success = 0 ;
      SCUErr_Value_Not_Found = 1 ;
      SCUErr_Invalid_Size = 2 ;
      SCUErr_Exists = 3 ; { Key or value already exists }
      SCUErr_Alarm_Failure = 4 ; { Operation caused an alarm which user used to abort operation }
      SCUErr_Bad_Data = 5 ; { Bad data (used by SCU32.DLL) }

{$IFDEF DYNAMIC_OBJECTS}
type TDelphi_SCU_Object = class
                              public { API... }
                                   procedure Attach ; virtual ; stdcall ; abstract ;
                                   procedure Detach ; virtual ; stdcall ; abstract ;
                          end ;

type T_SCU = class( TDelphi_SCU_Object )
                public { API... }
                    { Instance support... }
                    function Initialize : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    procedure Terminate ; virtual ; stdcall ; abstract ;

                    function Init( Name : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Version : longint ; virtual ; stdcall ; abstract ;

                    function Write( Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Read( Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Read_Size( Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Read_Type( Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Exists( Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
                    function Key_Exists( Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;

                    function Lookup( Path : PAnsiChar ; DT : longint ) : PAnsiChar ;
                        virtual ; stdcall ; abstract ;

                    function Delete( Path : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_( Path : PAnsiChar ; DT : longint ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_And_Write( Path : PAnsiChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Rename( Old, New : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;

                    function Set_Value( User, Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Get_Value( User, Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Value_Size( User, Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Value_Type( User, Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Value_Exists( User, Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
                    function Delete_Value( User, Path : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_Value( User, Path : PAnsiChar ;
                        DT : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Verify : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Debugger : PDebug_Interface ;
                        virtual ; stdcall ; abstract ;

                    function Read_Value( User, Path : PAnsiChar ;
                        var DT, Size : longint ; var Buffer : PAnsiChar) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    { Reads the specified value.  If result indicates a success,
                      DT is the type of the data, Size is the actual data size,
                      and Buffer is a pointer to a buffer containing the data.
                      DT, Size, and Buffer are ignored on call, and set on
                      return.

                      This function serves as a single call to: 1) determine if
                      the value exists, 2) determine what type it is, and
                      3) read the value's data. }

                    function Init_And_Extend( Name : PAnsiChar ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes. }

{$IFDEF FPC_OR_DELPHI}
                    function Delete_With_Callback( Path : PAnsiChar ;
                        Logger : TCOM_Logger ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
{$ELSE}
                    function Delete_With_Callback( Path : PAnsiChar ;
                        Logger : PSLogger ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
{$ENDIF}

                    procedure Shutdown ; virtual ; stdcall ; abstract ;

                { V1.2 interface: }
                public { Contextual version of instance support... }
                    function Init_Context( Name : PAnsiChar ; Flags : longint ) : longint ;
                        virtual ; stdcall ; abstract ;
                    procedure Terminate_Context( Context : longint ) ;
                        virtual ; stdcall ; abstract ;
                    function Init_And_Extend_Context( Name : PAnsiChar ;
                        Size : longint ; Flags : longint ) : longint ;
                        virtual ; stdcall ; abstract ;
                    { Initializes access to the specified file.  If the file
                      doesn't exist, it is created and pre-extended to the
                      specified size, in bytes.  Flags and 1 = turn on caching }

                public { Contextual version of Low-level API... }
                    function Write_Context( Context : longint ; Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Read_Context( Context : longint ; Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Read_Size_Context( Context : longint ; Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Read_Type_Context( Context : longint ; Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Exists_Context( Context : longint ; Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
                    function Key_Exists_Context( Context : longint ; Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;

                    function Lookup_Context( Context : longint ; Path : PAnsiChar ; DT : longint ) : PAnsiChar ;
                        virtual ; stdcall ; abstract ;

                    function Delete_Context( Context : longint ; Path : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_Context( Context : longint ; Path : PAnsiChar ; DT : longint ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_And_Write_Context( Context : longint ; Path : PAnsiChar ; DT : longint ;
                        Data : pointer ; Size : longint ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Rename_Context( Context : longint ; Old, New : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;

                public { Contextual version of Standard API... }
                    function Set_Value_Context( Context : longint ; User, Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Get_Value_Context( Context : longint ; User, Path : PAnsiChar ; Data : pointer ;
                        Size : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Value_Size_Context( Context : longint ; User, Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Value_Type_Context( Context : longint ; User, Path : PAnsiChar ) : longint ;
                        virtual ; stdcall ; abstract ;
                    function Value_Exists_Context( Context : longint ; User, Path : PAnsiChar ) : boolean ;
                        virtual ; stdcall ; abstract ;
                    function Delete_Value_Context( Context : longint ; User, Path : PAnsiChar ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Create_Value_Context( Context : longint ; User, Path : PAnsiChar ;
                        DT : longint ) : PSUnified_Exception ; virtual ; stdcall ; abstract ;
                    function Verify_Context( Context : longint ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
                    function Debugger_Context( Context : longint ) : PDebug_Interface ;
                        virtual ; stdcall ; abstract ;

                    function Read_Value_Context( Context : longint ; User, Path : PAnsiChar ;
                        var DT, Size : longint ; var Buffer : PAnsiChar) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;

{$IFDEF FPC_OR_DELPHI}
                    function Delete_With_Callback_Context( Context : longint ;
                        Path : PAnsiChar ; Logger : TCOM_Logger ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
{$ELSE}
                    function Delete_With_Callback_Context( Context : longint ;
                        Path : PAnsiChar ; Logger : PSLogger ) : PSUnified_Exception ;
                        virtual ; stdcall ; abstract ;
{$ENDIF}
            end ; { T_SCU }
{$ENDIF}

const SCUDT_Undefined = 0 ;
      SCUDT_Key = 1 ;
      SCUDT_String = 2 ;
      SCUDT_Integer = 3 ; { 4-bit signed }
      SCUDT_Boolean = 4 ;
      SCUDT_Float = 5 ; { Double-precision }
      SCUDT_Binary = 6 ;
      SCUDT_All = -1 ; { All types }

implementation

end.
