/*
              Program Name : _TSRS
              Package Name : Subroutine Library - Layer 0
              Purpose      : Support for Windows and DOS structures
              Institution  : Conroy & Conroy Co.
              Date Written : 21-Jan-2020
              Written By   : Alan Conroy
              Version      : 2.1B

              Copyright (C) 1992-2021 by Alan Conroy. 

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          These strutures are for support of Windows and DOS applications.
*/

#pragma pack(1)

/* DOS structure definitions */
       /* VESA subfunction 0 buffer */
struct TVESA_0
{
                char Signature[4] ;
                unsigned char Major_Version ;
                unsigned char Minor_Version ;
                char* Manufacturer ;
                signed int Flags ;
                void* Code_Numbers ;
};

       /* VESA subfunction 0 buffer */
struct TVESA_1
{
                  unsigned short int Mode ;
                  unsigned char Flags[2] ;
                  unsigned short int Granularity ;
                  unsigned short int Size ;
                  unsigned short int Segments[2] ;
                  void* Routine ;
                  unsigned char Line_Bytes ;
                  unsigned short int X_Resolution ;
                  unsigned short int Y_Resolution ;
                  unsigned char Pixel_Width ;
                  unsigned char Pixel_Height ;
                  unsigned char Bitplanes ;
                  unsigned char Bits_Per_Pixel ;
                  unsigned char Blocks ;
                  unsigned char Model ;
                  unsigned char Block_Size ;
};

       /* Find packed record */
struct TDOS_Find_Record
{
                            unsigned char Reserved[21] ;
                            unsigned char Attributes ;
                            unsigned short int Time ;
                            unsigned short int Date ;
                            signed int Size ;
                            char Name[13] ;
};

       /* DPE (Disk Partition Entry) */
struct TDPE
{
                unsigned char Status ; /* 0 = inactive, 0x80 = boot partition */
                unsigned char Start_Head ;
                unsigned short int Start_Sector_Cylinder ; /* 0:4 = sector, 8:15,6:7 = cylinder */
                unsigned char Partition_Type ;
                /* 0 = entry not allocated
                  1 = Primary DOS (FAT12)
                  2 = XENIX
                  3 = XENIX
                  4 = Primary DOS (FAT16)
                  5 = Extended DOS (DOS 3.3)
                  6 = DOS 4.0 partition with >32 Mb
                  0xDB = Concurrent DOS
                */
                unsigned char End_Head ;
                unsigned short int End_Sector_Cylinder ; /* 0:4 = sector, 8:15,6:7 = cylinder */
                signed int Boot_Sector ;
                signed int Sectors ;
};

       /* DPS ( Disk Partition Sector) */
struct TDPS
{
                unsigned char Code[ 0x1BE ] ;
                TDPE Partitions[4] ;
                unsigned short int ID ; /* 0xAA55 (indicates a disk partition sector ) */
};

       /* EXE File Header */
struct TEXE_Header
{
                        unsigned short int ID ; /* 0x5A4D */               /*e_magic*/
                        signed int Length ;                  /*e_cblp, e_cp*/
                        unsigned short int Segment_Address_Count ;      /*e_crlc*/
                        unsigned short int Header_Size ;                /*e_cparhdr*/
                        unsigned short int Min_Paragraphs ;             /*e_minalloc*/
                        unsigned short int Max_Paragraphs ;              /*e_maxalloc*/
                        unsigned short int SS_Displacement ;            /*e_ss*/
                        unsigned short int Starting_SP ;                /*e_sp*/
                        unsigned short int Checksum ;                   /*e_csum*/
                        unsigned short int Starting_IP ;                /*e_ip*/
                        unsigned short int CS_Offset ;                  /*e_cs*/
                        unsigned short int RT_Address ;                 /*e_lfarlc*/
                        unsigned short int Overlay ;                    /*e_ovno*/
                        unsigned short int Reserved[4] ;  /*e_res*/
                        unsigned short int OEM_ID ;                     /*e_oemid*/
                        unsigned short int OEM_Information ;            /*e_oeminfo*/
                        unsigned short int Reserved1[10] ; /*e_res2*/
                        signed int Offset ; /*Pointer to NE_Header*/ /*e_lfanew*/
};

       /* NE Header */
struct TNE_Header
{
                      unsigned short int ID ; /*'NE'*/                         /*ne_magic*/
                      unsigned char Linker_Version ;                    /*ne_ver*/
                      unsigned char Linker_Revision ;                   /*ne_rev*/
                      unsigned short int Entry_Table_Offset ;                /*ne_enttab*/
                      unsigned short int Entry_Table_Length ;                /*ne_cbenttab*/
                      signed int Checksum ;                       /*ne_crc*/
                      unsigned short int Flags ;                             /*ne_flags*/
                      unsigned short int Automatic_DS_Number ;               /*ne_autodata*/
                      unsigned short int Local_Heap_Size ; /*initial*/         /*ne_heap*/
                      unsigned short int Stack_Size ; /*initial*/              /*ne_stack*/
                      signed int Starting_CSIP ;                  /*ne_csip*/
                      signed int Starting_SSSP ;                  /*ne_sssp*/
                      unsigned short int Segment_Table_Entries ;             /*ne_cseg*/
                      unsigned short int Module_Ref_Entries ;                /*ne_cmod*/
                      unsigned short int Nonresident_Name_Table_Size ;       /*ne_cbnrestab*/
                      unsigned short int Segment_Table_Offset ;              /*ne_segtab*/
                      unsigned short int Resource_Table_Offset ;             /*ne_rsrctab*/
                      unsigned short int Resident_Name_Table_Offset ;        /*ne_restab*/
                      unsigned short int Module_Ref_Table_Offset ;           /*ne_modtab*/
                      unsigned short int Imported_Name_Table_Offset ;        /*ne_imptab*/
                      signed int Nonresident_Name_Table_Offset ;  /*ne_nrestab*/
                      unsigned short int Movable_Entry_Points ;              /*ne_cmovent*/
                      unsigned short int Logical_Sector_Alignment_Shift ;    /*ne_align*/
                      unsigned short int Resource_Segment_Count ;            /*ne_cres*/
                      unsigned char Target_OS ;                         /*ne_exetyp*/
                      unsigned char Additional_Flags ;                  /*ne_flagsothers*/
                      unsigned short int Return_Thunks_Offset ;              /*ne_pretthunks*/
                      unsigned short int Segment_Ref_Bytes_Offset ;          /*ne_psegrefbytes*/
                      unsigned short int Reserved ;                          /*ne_swaparea*/
                      unsigned short int Expected_Windows_Version_Number ;   /*ne_expver*/
};

       /* PE Header */
struct TPE_Header_With_Optional
{
                                    signed int Signature ; /* "PE"#0#0 */
                                    unsigned short int Machine ;
                                    unsigned short int NumberOfSections ;
                                    signed int TimeDateStamp ;
                                    signed int PointerToSymbolTable ;
                                    signed int NumberOfSymbols ;
                                    unsigned short int SizeOfOptionalHeader ;
                                    unsigned short int Characteristics ;
                                    unsigned short int Magic ;
                                    unsigned char MajorLinkerVersion ;
                                    unsigned char MinorLinkerVersion ;
                                    signed int SizeOfCode ;
                                    signed int SizeOfInitializedData ;
                                    signed int SizeOfUninitializedData ;
                                    signed int AddressOfEntryPoint ;
                                    signed int BaseOfCode ;
                                    signed int BaseOfData ;
                                    signed int ImageBase ;
                                    signed int SectionAlignment ;
                                    signed int FileAlignment ;
                                    unsigned short int MajorOperatingSystemVersion ;
                                    unsigned short int MinorOperatingSystemVersion ;
                                    unsigned short int MajorImageVersion ;
                                    unsigned short int MinorImageVersion ;
                                    unsigned short int MajorSubsystemVersion ;
                                    unsigned short int MinorSubsystemVersion ;
                                    signed int Win32VersionValue ;
                                    signed int SizeOfImage ;
                                    signed int SizeOfHeaders ;
                                    signed int CheckSum ;
                                    unsigned short int Subsystem ;
                                    unsigned short int DllCharacteristics ;
                                    signed int SizeOfStackReserve ;
                                    signed int SizeOfStackCommit ;
                                    signed int SizeOfHeapReserve ;
                                    signed int SizeOfHeapCommit ;
                                    signed int LoaderFlags ;
                                    signed int NumberOfRvaAndSizes ;
};

struct TImage_Data_Directory
{
    signed int VirtualAddress ;
    signed int Size ;
};

struct TData_Directory
{
	TImage_Data_Directory Entry[ 1024 ] ;
};

struct TImage_Import_Descriptor
{
                                    signed int Characteristics ;
                                    signed int TimeDateStamp ;
                                    signed int ForwarderChain ;
                                    signed int Name ;
};

       /* FCB (File Control Block) for V1.n */
struct TDOS_FCBv1
{
                    unsigned char Extended ; /* FFH if extended, offset -7 */
                    unsigned char Filler[5] ;
                    unsigned char Extended_Attribute ; /* Offset -1 */
                    unsigned char Drive_Number ; /* Offset 0 */
                    char File_Name[11] ;
                    unsigned short int Current_Block ;
                    unsigned short int Record_Size ;
                    signed int File_Size ;
                    unsigned short int Date ;
                    unsigned short int Time ;
                    unsigned char Flags ;
                    unsigned short int Absolute_Cluster ;
                    unsigned short int Starting_Cluster ;
                    unsigned short int Relative_Cluster ;
                    unsigned char Filler1 ;
                    unsigned char Current_Record ;
                    signed int Record_Number ; 
};

       /* FCB (File Control Block) for V2.n */
struct TDOS_FCBv2
{
                    unsigned char Extended ; /*FFH if extended, offset -7*/
                    unsigned char Filler[5] ;
                    unsigned char Extended_Attribute ; /*Offset -1*/
                    unsigned char Drive_Number ; /*Offset 0*/
                    char File_Name[11] ;
                    unsigned short int Current_Block ;
                    unsigned short int Record_Size ;
                    signed int File_Size ;
                    unsigned short int Date ;
                    unsigned short int Time ;
                    unsigned char Flags ;
                    unsigned short int Starting_Cluster ;
                    unsigned char Filler1[5] ;
                    unsigned char Current_Record ;
                    signed int Record_Number ; 
};

       /* FCB (File Control Block) for V3.n, SHARE loaded, local file */
struct TDOS_FCB_Local
{
                    unsigned char Extended ; /*FFH if extended, offset -7*/
                    unsigned char Filler[5] ;
                    unsigned char Extended_Attribute ; /*Offset -1*/
                    unsigned char Drive_Number ; /*Offset 0*/
                    char File_Name[11] ;
                    unsigned short int Current_Block ;
                    unsigned short int Record_Size ;
                    signed int File_Size ;
                    unsigned short int Date ;
                    unsigned short int Time ;
                    unsigned char SFT_Number ;
                    unsigned char Flags ;
                    unsigned short int Starting_Cluster ;
                    unsigned short int SHARE_Offset ;
                    unsigned char Attributes ;
                    unsigned char Filler1 ;
                    unsigned char Current_Record ;
                    signed int Record_Number ; 
};

       /* FCB (File Control Block) for V3.n, SHARE loaded, remote file */
struct TDOS_FCB_Remote
{
                    unsigned char Extended ; /*FFH if extended, offset -7*/
                    unsigned char Filler[5] ;
                    unsigned char Extended_Attribute ; /*Offset -1*/
                    unsigned char Drive_Number ; /*Offset 0*/
                    char File_Name[11] ;
                    unsigned short int Current_Block ;
                    unsigned short int Record_Size ;
                    signed int File_Size ;
                    unsigned short int Date ;
                    unsigned short int Time ;
                    unsigned short int Directory_Sector ;
                    unsigned short int Relative_Cluster ;
                    unsigned short int Absolute_Cluster ;
                    unsigned char Current_Record ;
                    signed int Record_Number ; 
};

       /* FCB (File Control Block) for V3.n, SHARE not loaded */
struct TDOS_FCBv3
{
                    unsigned char Extended ; /*FFH if extended, offset -7*/
                    unsigned char Filler[5] ;
                    unsigned char Extended_Attribute ; /*Offset -1*/
                    unsigned char Drive_Number ; /*Offset 0*/
                    char File_Name[11] ;
                    unsigned short int Current_Block ;
                    unsigned short int Record_Size ;
                    signed int File_Size ;
                    unsigned short int Date ;
                    unsigned short int Time ;
                    unsigned char Flags ;
                    unsigned short int Starting_Cluster ;
                    unsigned short int Directory_Sector ;
                    unsigned char Directory_Entry ;
                    unsigned char Current_Record ;
                    signed int Record_Number ; 
};

       /* PSP (Program Segment Prefix */
struct tPSP
{
                    unsigned short int Int_20H ;
                    unsigned short int Size ;
                    unsigned char Reserved ;
                    unsigned char FAR_Jump[5] ;
                    signed int Int_22H ;
                    signed int Int_23H ;
                    signed int Int_24H ;
                    unsigned short int Parent ;
                    unsigned char Files[20] ;
                    unsigned short int Environment ;
                    signed int Ss_Sp ;
                    unsigned short int Max_Files ;
                    void* File_Table ;
                    void* Previous_PSP ;
                    unsigned char Filler1[20] ;
                    unsigned char Dispatch[3] ;
                    unsigned char Filler2[9] ;
                    unsigned char Fcb1[16] ;
                    unsigned char Fcb2[20] ;
                    char Command_Tail[128] ;
};

       /* DPB (Drive Parameter Block) For DOS V1.n */
struct TDOS_DPBv1
{
                    unsigned char Drive ;
                    unsigned char Unit_Number ;
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Largest_Sector ;
                    unsigned char Shift_Count ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Copies ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int First_Data_Sector ;
                    unsigned short int Max_Cluster ; 
};

       /* DPB (Drive Parameter Block) For DOS V2.n */
struct TDOS_DPBv2
{
                    unsigned char Drive ;
                    unsigned char Unit_Number ;
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Largest_Sector ;
                    unsigned char Shift_Count ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Copies ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int First_Data_Sector ;
                    unsigned short int Max_Cluster ;
                    unsigned char FAT_Sectors ;
                    unsigned short int Root_Sector ;
                    void* Driver ;
                    unsigned char Media_ID ;
                    unsigned char Rebuild ;
                    void* Next_Driver ;
                    unsigned short int Current_Directory_Cluster ;
                    char Current_Path[64] ;
};

       /* DPB (Drive Parameter Block) For DOS V3.n */
struct TDOS_DPBv3
{
                    unsigned char Drive ;
                    unsigned char Unit_Number ;
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Largest_Sector ;
                    unsigned char Shift_Count ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Copies ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int First_Data_Sector ;
                    unsigned short int Max_Cluster ;
                    unsigned char FAT_Sectors ;
                    unsigned short int Root_Sector ;
                    void* Driver ;
                    unsigned char Media_ID ;
                    unsigned char Rebuild ;
                    void* Next_Driver ;
                    unsigned short int Search_Start ;
                    unsigned short int Free_Clusters ; 
};

       /* DPB (Drive Parameter Block) For DOS V4.n */
struct TDOS_DPBv4
{
                    unsigned char Drive ;
                    unsigned char Unit_Number ;
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Largest_Sector ;
                    unsigned char Shift_Count ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Copies ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int First_Data_Sector ;
                    unsigned short int Max_Cluster ;
                    unsigned short int FAT_Sectors ;
                    unsigned short int Root_Sector ;
                    void* Driver ;
                    unsigned char Media_ID ;
                    unsigned char Rebuild ;
                    void* Next_Driver ;
                    unsigned short int Search_Start ;
                    unsigned short int Free_Clusters ; 
};

       /* BPB (BIOS Parameter Block), DOS V2.n and V3.n */
       /* As it appears on disks without 8 sectors/track, in sector 0 */
struct TBPBv2
{
                    unsigned char Jump[3] ; /* Jump instruction */
                    char System_ID[8] ; /* Manufacturer name/version */
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Sectors_Per_Cluster ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Count ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int Total_Sectors ;
                    unsigned char Media_ID ; /* Media descriptor:
                                                F0 = 3.5" FDD, 2 sides, 80 tracks, 18 sectors/track
                                                F8 = HDD
                                                F9 = 5.25" FDD, 2 sides, 80 tracks, 15 sectors/track
                                                     3.5" FDD, 2 sides, 80 tracks, 9 sectors/track
                                                FA = FDD, 1 side, 80 tracks, 8 sectors/track
                                                FB = FDD, 2 sides, 80 tracks, 8 sectors/track
                                                FC = FDD, 1 side, 40 tracks, 9 sectors/track
                                                FD = FDD, 2 sides, 40 tracks, 9 sectors/track
                                                FE = FDD, 1 side, 40 tracks, 8 sectors/track
                                                FF = FDD, 2 sides, 40 tracks, 8 sectors/track
                                        */
                    unsigned short int FAT_Sectors ;
                    unsigned short int Sectors_Per_Track ;
                    unsigned short int Heads ;
                    signed int Hidden_Sectors ;
                    /* Data after this is bootstrap code in FAT12/FAT16 */
                    /* FAT32 fields: */
                    signed int Total_Sectors_High ;
                    signed int FAT_Sectors_High ;
                    unsigned short int Flags ;
                    unsigned short int Version ;
                    signed int Root_Start_Cluster ;
                    unsigned short int FS_Info_Sec ;
                    unsigned short int Backup_Boot_Sec ;
                    unsigned short int Reserved ;
};

       /* BPB (BIOS Parameter Block), DOS V4.n */
struct TBPBv4
{
                    unsigned char Jump[3] ;
                    char System_ID[8] ;
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Sectors_Per_Cluster ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Count ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int Total_Sectors ;
                    unsigned char Media_ID ;
                    unsigned short int FAT_Sectors ;
                    signed int Extended_Total_Sectors ;
                    unsigned char Filler[6] ;
                    unsigned short int Cylinders ;
                    unsigned char Device_Type ;
                    unsigned short int Attributes ; 
};

       /* BPB (BIOS Parameter Block), Special - for DDT entries */
struct TBPB
{
                    unsigned short int Bytes_Per_Sector ;
                    unsigned char Sectors_Per_Cluster ;
                    unsigned short int Reserved_Sectors ;
                    unsigned char FAT_Count ;
                    unsigned short int Root_Directory_Entries ;
                    unsigned short int Total_Sectors ;
                    unsigned char Media_ID ;
                    unsigned short int FAT_Sectors ;
                    unsigned short int Sectors_Per_Track ;
                    unsigned short int Heads ;
                    unsigned short int Hidden_Sectors ; 
};

       /* DDT (Drive Data Table), DOS V3.3 */
struct TDOS_DDTv3
{
                    void* Next ;
                    unsigned char Unit_Number ;
                    unsigned char Drive_Number ;
                    TBPB Bios_Pb ;
                    unsigned char Filler ;
                    unsigned short int Open ;
                    char Volume_Label[11] ;
                    unsigned char Null ;
                    unsigned char Device_Type ;
                    unsigned short int Flags ;
                    unsigned short int Cylinders ;
                    TBPB Max_Capacity ;
                    unsigned char Filler1[3] ;
                    char File_System[9] ;
                    unsigned char Null1 ;
                    signed int Last_Access ;
};

       /* DDT (Drive Data Table), DOS V4.n */
struct TDOS_DDTv4
{
                    void* Next ;
                    unsigned char Unit_Number ;
                    unsigned char Drive_Number ;
                    TBPB Bios_Pb ;
                    unsigned char Filler[9] ;
                    unsigned char Device_Type ;
                    unsigned short int Flags ;
                    unsigned short int Cylinders ;
                    TBPB Max_Capacity ;
                    unsigned char Filler1[13] ;
                    signed int Last_Access ;
                    char Volume_Label[11] ;
                    unsigned char Null ;
                    signed int Serial_Number ;
                    char File_System[8] ;
                    unsigned char Null1 ;
};

       /* DPL (DOS Parameter List) */
union TDPL
{
	struct
	{
		unsigned short int Ax ;
		unsigned short int Bx ;
		unsigned short int Cx ;
		unsigned short int Dx ;
		unsigned short int Si ;
		unsigned short int Di ;
		unsigned short int Ds ;
		unsigned short int Es ;
		unsigned short int Filler ;
		unsigned short int Id ;
		unsigned short int Process ;
	};
	struct 
	{
		unsigned char Al ;
		unsigned char Ah ;
		unsigned char Bl ;
		unsigned char Bh ;
		unsigned char Cl ;
		unsigned char Ch ;
		unsigned char Dl ;
		unsigned char Dh ;
	};
};

       /* Disk Serial Number packed record */
struct TSN_Record
{
    unsigned short int Level ;
    signed int Serial_Number ;
    char Volume_Label[11] ;
    char File_System[8] ;
};

       /* KEYB internal data, DOS V3.3 */
struct TDOS_Keybv3
{
                    void* Int_09 ;
                    void* Int_2F ;
                    unsigned char Filler[6] ;
                    unsigned short int Flags ;
                    unsigned char Filler1[6] ;
                    char Country_ID[2] ;
                    unsigned short int Current_Code_Page ;
                    unsigned short int First ;
                    unsigned short int Ptr ;
                    unsigned char Filler2[2] ;
                    unsigned short int Xlate ;
                    unsigned short int Last ;
                    unsigned char Filler3[9] ;
};

       /* KEYB internal data, DOS V4.n */
struct TDOS_Keybv4
{
                    void* Int_09 ;
                    void* Int_2F ;
                    unsigned char Filler[6] ;
                    unsigned short int Flags ;
                    unsigned char Filler1[6] ;
                    char Country_ID[2] ;
                    unsigned short int Current_Code_Page ;
                    unsigned char Filler2[2] ;
                    unsigned short int First ;
                    unsigned short int Ptr ;
                    unsigned char Filler3[2] ;
                    unsigned short int Xlate ;
                    unsigned short int Last ;
                    unsigned char Filler4[9] ;
};

       /* EXEC Parameter Block */
struct TEPB
{
                    unsigned short int Environment ;
                    void* Command_Line ;
                    void* Fcb1 ;
                    void* Fcb2 ;
                    signed int Initial_SS_SP ;
                    signed int Entry_Pointer ;
};

       /* SDB (Search Data Block) */
struct TSDB
{
                    unsigned char Drive_Letter ;
                    char Search_Template[11] ;
                    unsigned char Search_Attributes ;
                    unsigned short int Directory_Entry ;
                    unsigned short int Parent_Cluster ;
                    unsigned char Filler[4] ;
                    unsigned char Found_Attribute ;
                    unsigned short int Time ;
                    unsigned short int Date ;
                    signed int Size ;
                    char Name[13] ;
};

       /* List of Lists (LOL), DOS V2.n */
struct TDOS_LOLv2
{
                    unsigned char Filler[14] ; /*Offset -16*/
                    unsigned short int MCB ; /*Offset -2*/
                    void* DPB ; /*Offset 0*/
                    void* SFT ;
                    void* CLOCK_Driver ;
                    void* CON_Driver ;
                    unsigned char Last_Drive ; /*Offset 10H*/
                    unsigned short int Max_Bytes_Per_Block ;
                    void* First_Disk_Buffer ;
                    unsigned char NUL_Driver_Header[18] ;
};

       /* List of Lists (LOL), DOS V3.0 */
struct TDOS_LOLv30
{
                    unsigned char Filler[4] ; /*Offset -16*/
                    unsigned short int Retry_Count ; /*Offset -12*/
                    unsigned short int Retry_Delay ; /*Offset -10*/
                    void* Current_Disk_Buffer ; /*Offset -8*/
                    unsigned short int CON_Input ; /*Offset -4*/
                    unsigned short int Mcb ; /*Offset -2*/
                    void* Dpb ; /*Offset 0*/
                    void* Sft ;
                    void* CLOCK_Driver ;
                    void* CON_Driver ;
                    unsigned char Device_Count ; /*Offset 10H*/
                    unsigned short int Max_Bytes_Per_Block ;
                    void* First_Disk_Buffer ;
                    void* Cds ;
                    unsigned char Last_Drive ;
                    void* STRING_Workspace ;
                    unsigned short int STRING_Size ;
                    void* Fcb ;
                    unsigned short int Protected_FCBs ;
                    unsigned char NUL_Driver_Header[18] ;
};

       /* List of Lists (LOL), DOS V3.1+ */
struct TDOS_LOLv3
{
                    unsigned char Filler[4] ; /*Offset -16*/
                    unsigned short int Retry_Count ; /*Offset -12*/
                    unsigned short int Retry_Delay ; /*Offset -10*/
                    void* Current_Disk_Buffer ; /*Offset -8*/
                    unsigned short int CON_Input ; /*Offset -4*/
                    unsigned short int Mcb ; /*Offset -2*/
                    void* Dpb ; /*Offset 0*/
                    void* Sft ;
                    void* CLOCK_Driver ;
                    void* CON_Driver ;
                    unsigned short int Max_Bytes_Per_Block ; /*Offset 10H*/
                    void* First_Disk_Buffer ;
                    void* Cds ;
                    void* Fcb ;
                    unsigned short int Protected_FCBs ;
                    unsigned char Device_Count ;
                    unsigned char Last_Drive ;
                    unsigned char NUL_Driver_Header[18] ;
                    unsigned char JOIN_Count ;
};

       /* List of Lists (LOL), DOS V4.n */
struct TDOS_LOLv4
{
                    unsigned char Filler[14] ; /*Offset -16*/
                    unsigned short int Mcb ; /*Offset -2*/
                    void* Dpb ; /*Offset 0*/
                    void* Sft ;
                    void* CLOCK_Driver ;
                    void* CON_Driver ;
                    unsigned short int Max_Bytes_Per_Block ; /*Offset 10H*/
                    void* Disk_Buffer ; /*12H*/
                    void* Cds ;         /*16H*/
                    void* Fcb ;         /*1AH*/
                    unsigned short int Protected_FCBS ; /*1EH*/
                    unsigned char Device_Count ;   /*20H*/
                    unsigned char Last_Drive ;     /*21H*/
                    unsigned char NUL_Driver_Header[18] ; /*22H*/
                    unsigned char JOIN_Count ;     /*34H*/
                    unsigned short int Program_Names ;  /*35H*/
                    void* IFS_Resident ; /*37H*/
                    void* IFS_Chain ;   /*3BH*/
                    unsigned short int Buffers_X ;      /*3FH*/
                    unsigned short int Buffers_Y ;      /*41H*/
                    unsigned char Boot_Drive ;     /*43H*/
                    unsigned char Filler1 ;
                    unsigned short int Extended_RAM_Size ;
};

       /* MCB (Memory Control Block) */
struct TDOS_MCB
{
                    unsigned char Block_Type ;
                    unsigned short int Owner ;
                    unsigned short int Size ;
                    unsigned char Filler[3] ;
                    char Name[8] ;
};

       /* SCB (Subsegment Control Block) */
struct TSCB
{
                    char Subsegment_Type ;
                    unsigned short int Start ;
                    unsigned short int Size ;
                    unsigned char Filler[3] ;
                    char Name[8] ;
};

       /* STACKS Code Segment */
struct TSCS
{
                    unsigned short int Filler ;
                    unsigned short int Stacks ;
                    unsigned short int SCB_Size ;
                    unsigned short int Stack_Size ;
                    void* Data_Segment ;
                    unsigned short int Scb ;
                    unsigned short int Last ;
                    unsigned short int Next ;
};

       /* SAE (STACKS Array Elements) */
struct TSAE
{
                    unsigned char Status ;
                    unsigned char Filler ;
                    unsigned short int Previous_SP ;
                    unsigned short int Previous_SS ;
                    unsigned short int Top ;
};

       /* Sharing packed record */
struct TDOS_Sharing_Record
{
                            unsigned char Flag ;
                            unsigned short int Size ;
                            unsigned char Checksum ;
                            unsigned short int Lock_Record ;
                            void* Sft ;
                            unsigned short int Sequence ;
                            char Pathname ; /* First byte of variable area */
};
       /* Lock packed record */
struct TDOS_Lock_Record
{
                            unsigned short int Next ;
                            signed int Start ;
                            signed int Ending ;
                            void* Sft ;
                            unsigned short int Owner ;
};

       /* SFT (System File Table), DOS V2.n */
union TDOS_SFTv2
{
    struct
	{
		unsigned char Handle_Count ;
        unsigned char Open_Mode ;
        unsigned char Attributes ;
        unsigned char Drive ;
        char Name[11] ;
        unsigned short int Filler[2] ;
        signed int Size ;
        unsigned short int Date ;
        unsigned short int Time ;
        unsigned char Device_Attribute ;
        void* Driver ;
        unsigned short int Absolute_Cluster ;
        unsigned short int Filler1 ;
        signed int Current_Position ;
    } ;
    struct
	{
		unsigned char Filler2[28] ;
        unsigned short int Start ;
        unsigned short int Relative_Cluster ;
    } ;
};

       /* SFT (System File Table), DOS V3.n */
struct TDOS_SFTv3
{
                    unsigned short int Handle_Count ;
                    unsigned short int Open_Mode ;
                    unsigned char Attributes ;
                    unsigned short int Device_Info ;
                    void* Driver ;
                    unsigned short int Start_Cluster ;
                    unsigned short int Time ;
                    unsigned short int Date ;
                    signed int Size ;
                    signed int Current_Position ;
                    unsigned short int Relative_Cluster ;
                    unsigned short int Absolute_Cluster ;
                    unsigned short int Dir_Sector ;
                    unsigned char Directory_Entry ;
                    char Fcb_Fn[11] ;
                    void* Previous_SFT ;
                    unsigned short int Owner ;
                    unsigned short int Sharing_Record ;
};

       /* SFT (System File Table), DOS V4.n */
union TDOS_SFTv4
{
    struct
    {
		unsigned short int Handle_Count ;
        unsigned short int Open_Mode ;
        unsigned char Attributes ;
        unsigned short int Device_Info ;
        void* Driver ;
        unsigned short int Start_Cluster ;
        unsigned short int Time ;
        unsigned short int Date ;
        signed int Size ;
        void* Redirifs ;
        unsigned char Filler[3] ;
        char Fcb_Fn[11] ;
        void* Previous_SFT ;
        unsigned short int Machine_Number ;
        unsigned short int Owner ;
        unsigned short int Sharing_Record ;
        unsigned short int Absolute_Cluster ;
        void* IFS_Driver ;
    } ;
    struct
	{
		unsigned char Filler1[25] ;
        unsigned short int Relative_Cluster ;
        unsigned short int Directory_Sector ;
        unsigned char Directory_Entry ;
    } ;
};

       /* CDS (Current Directory Structure), DOS 3.n */
union TDOS_CDSv3
{
    struct
	{		
		char Curr_Path[67] ;
        unsigned short int Flags ;
        void* Dpb ;
        void* Redirifs ;
        unsigned short int Parameter ;
        unsigned short int Root_OFS ;
    } ;
    struct
    {
	    unsigned char Filler[73] ;
        unsigned short int Start ;
        unsigned short int Filler0 ;
    } ;
};

       /* CDS (Current Directory Structure), DOS 4.n */
union TDOS_CDSv4
{
	struct
	{
        char Curr_Path[67] ;
        unsigned short int Flags ;
        void* Dpb ;
        void* Rediris ;
        unsigned short int Parameter ;
        unsigned short int Root_OFS ;
        unsigned char Filler1 ;
        void* IFS_Driver ;
        unsigned short int Filler2 ;
    } ;
    struct
    {
        unsigned char Filler[73] ;
        unsigned short int Start ;
        unsigned short int Filler0 ;
    } ;
};

       /* DOS Directory entry */
struct TDOS_Dir_Rec
{
                        char Name[11] ;
                        /* Special codes: first byte: 0 = Last directory entry
                                                     5 = First character of filename is 0xE5
                                                     2E = File applies to current dir
                                                     E5 = Deleted */
                        unsigned char Attributes ;
                        /* And 1 = read-only
                          And 2 = Hidden
                          And 4 = System
                          And 8 = Volume name/label
                          And 16 = directory
                          And 32 = Archive
                          And 64 = reserved
                        */
                        unsigned char Reserved[10] ;
                        unsigned short int Update_Time ; /* 0:4 = seconds in 2-second increments
                                               5:10 = Minute
                                               11:15 = Hour */
                        unsigned short int Update_Date ; /* 0:4 = Day
                                               5:8 = Month
                                               9:15 = Year - 1980 */
                        unsigned short int Start_Cluster ;
                        signed int Size ;
};

       /* FindFirst/Next data block for SDAs */
struct TSDB_Rec
{
                    unsigned char Drive_Letter ;
                    char Search_Template[11] ;
                    unsigned char Search_Attributes ;
                    unsigned short int Directory_Entry ;
                    unsigned short int Parent_Cluster ;
                    unsigned char Filler[4] ;
};

       /* SDA (Swappable DOS Area), DOS 3.n */
struct TSDAv3
{
                    unsigned char Critical_Error_Flag ;
                    unsigned char InDOS_Flag ;
                    unsigned char Error_Drive ;
                    unsigned char Error_Locus ;
                    unsigned short int Extended_Error_Code ;
                    unsigned char Suggested_Action ;
                    unsigned char Error_Class ;
                    void* Error_ES_DI ;
                    void* Current_DTA ;
                    unsigned short int Current_PSP ;
                    unsigned short int Stored_SP ;
                    unsigned short int Return_Code ;
                    unsigned char Current_Drive ;
                    unsigned char Break_Flag ;
                    /* Swap following only if in DOS */
                    unsigned short int Stored_AX ;
                    unsigned short int Sharing_PSP ;
                    unsigned short int Sharing_Number ;
                    unsigned short int First_MB ;
                    unsigned short int Best_MB ;
                    unsigned short int Last_MB ;
                    unsigned char Filler[10] ;
                    unsigned char Dd ;
                    unsigned char Mm ;
                    unsigned short int Yy_1980 ;
                    unsigned short int Dd_1980 ;
                    unsigned char Day_Of_Week ;
                    unsigned char Filler1[3] ;
                    unsigned char Driver_Request_Header[26] ;
                    void* Entry_Point ;
                    unsigned char Driver_Request_Header1[22] ;
                    unsigned char Driver_Request_Header2[22] ;
                    unsigned char PSP_Type ;
                    unsigned char Filler2[7] ;
                    unsigned char CLOCK_Record[6] ;
                    unsigned char Filler3[2] ;
                    char Fn1[128] ;
                    char Fn2[128] ;
                    TSDB_Rec Sdb ;
                    TDOS_Dir_Rec Found_File ;
                    TDOS_CDSv3 Drive_CDSCopy ;
                    char Fcb_Fn1[11] ;
                    unsigned char Filler4 ;
                    char Fcb_Fn2[11] ;
                    unsigned char Filler5[9] ;
                    unsigned char Search_Attributes ;
                    unsigned char FCB_Type ;
                    unsigned char Extended_Attributes ;
                    unsigned char Open_Mode ;
                    unsigned char Filler6[3] ;
                    unsigned char DOS_Flag ;
                    unsigned char Filler7[9] ;
                    unsigned char Termination_Type ;
                    unsigned char Filler8 ;
                    unsigned char Replace_Byte ;
                    void* Error_DPB ;
                    void* Stack_Frame ;
                    unsigned short int Stored_SP1 ;
                    void* DPB_Ptr ;
                    unsigned char Filler9[8] ;
                    unsigned char Media_ID ;
                    unsigned char Filler10[5] ;
                    void* Current_SFT ;
                    void* Drive_CDSPTR ;
                    void* Caller_FCB ;
                    unsigned short int Filler11 ;
                    unsigned short int Temp ;
                    void* Jft ;
                    unsigned short int Fn1_Csofs ;
                    unsigned short int Fn2_Csofs ;
                    unsigned char Filler12[18] ;
                    signed int Current_Offset ;
                    unsigned char Filler13[12] ;
                    signed int Appended ;
                    void* Disk_Buffer ;
                    void* Sft ;
                    unsigned short int Stored_BX ;
                    unsigned short int Stored_DS ;
                    unsigned short int Temp1 ;
                    void* Call_Frame ;
                    TSDB_Rec Ren_SrcFile ;
                    TDOS_Dir_Rec Ren_File ;
                    unsigned char Error_Stack[331] ;
                    unsigned char Disk_Stack[384] ;
                    unsigned char Char_Stack[384] ;
};

     /* SDA (Swappable DOS Area), DOS 3.3 */
struct TSDAv33
{
                    unsigned char Critical_Error_Flag ;
                    unsigned char InDOS_Flag ;
                    unsigned char Error_Drive ;
                    unsigned char Error_Locus ;
                    unsigned short int Extended_Error_Code ;
                    unsigned char Suggested_Action ;
                    unsigned char Error_Class ;
                    void* Error_ES_DI ;
                    void* Current_DTA ;
                    unsigned short int Current_PSP ;
                    unsigned short int Stored_SP ;
                    unsigned short int Return_Code ;
                    unsigned char Current_Drive ;
                    unsigned char Break_Flag ;
                    /*Swap following only if in DOS*/
                    unsigned short int Stored_AX ;
                    unsigned short int Sharing_PSP ;
                    unsigned short int Sharing_Number ;
                    unsigned short int First_MB ;
                    unsigned short int Best_MB ;
                    unsigned short int Last_MB ;
                    unsigned char Filler[10] ;
                    unsigned char Dd ;
                    unsigned char Mm ;
                    unsigned short int Yy_1980 ;
                    unsigned short int Dd_1980 ;
                    unsigned char Day_Of_Week ;
                    unsigned char Filler1[3] ;
                    unsigned char Driver_Request_Header[26] ;
                    void* Entry_Point ;
                    unsigned char Driver_Request_Header1[22] ;
                    unsigned char Driver_Request_Header2[22] ;
                    unsigned char PSP_Type ;
                    unsigned char Filler2[7] ;
                    unsigned char CLOCK_Record[6] ;
                    unsigned char Filler3[2] ;
                    char Fn1[128] ;
                    char Fn2[128] ;
                    TSDB_Rec Sdb ;
                    TDOS_Dir_Rec Found_File ;
                    TDOS_CDSv3 Drive_CDSCopy ;
                    char Fcb_Fn1[11] ;
                    unsigned char Filler4 ;
                    char Fcb_Fn2[11] ;
                    unsigned char Filler5[9] ;
                    unsigned char Search_Attributes ;
                    unsigned char FCB_Type ;
                    unsigned char Extended_Attributes ;
                    unsigned char Open_Mode ;
                    unsigned char Filler6[3] ;
                    unsigned char DOS_Flag ;
                    unsigned char Filler7[9] ;
                    unsigned char Termination_Type ;
                    unsigned char Filler8 ;
                    unsigned char Replace_Byte ;
                    void* Error_DPB ;
                    void* Stack_Frame ;
                    unsigned short int Stored_SP1 ;
                    void* DPB_Ptr ;
                    unsigned char Filler9[8] ;
                    unsigned char Media_ID ;
                    unsigned char Filler10[5] ;
                    void* Current_SFT ;
                    void* Drive_CDSPTR ;
                    void* Caller_FCB ;
                    unsigned short int Filler11 ;
                    unsigned short int Temp ;
                    void* Jft ;
                    unsigned short int Fn1_Csofs ;
                    unsigned short int Fn2_Csofs ;
                    unsigned char Filler12[18] ;
                    signed int Current_Offset ;
                    unsigned char Filler13[12] ;
                    signed int Appended ;
                    void* Disk_Buffer ;
                    void* Sft ;
                    unsigned short int Stored_BX ;
                    unsigned short int Stored_DS ;
                    unsigned short int Temp1 ;
                    void* Call_Frame ;
                    TSDB_Rec Ren_SrcFile ;
                    TDOS_Dir_Rec Ren_File ;
                    unsigned char Error_Stack[331] ;
                    unsigned char Disk_Stack[384] ;
                    unsigned char Char_Stack[384] ;
                    unsigned char Flag ;
                    unsigned char Drive ;
                    unsigned short int Filler14 ;
};

       /* SDA (Swappable DOS Area), DOS 4.n */
struct TSDAv4
{
                    unsigned char Critical_Error_Flag ;
                    unsigned char InDOS_Flag ;
                    unsigned char Error_Drive ;
                    unsigned char Error_Locus ;
                    unsigned short int Extended_Error_Code ;
                    unsigned char Suggested_Action ;
                    unsigned char Error_Class ;
                    void* Error_ES_DI ;
                    void* Current_DTA ;
                    unsigned short int Current_PSP ;
                    unsigned short int Stored_SP ;
                    unsigned short int Return_Code ;
                    unsigned char Current_Drive ;
                    unsigned char Break_Flag ;
                    unsigned char Filler[2] ;
                    /*Swap following only if in DOS*/
                    unsigned short int Stored_AX ;
                    unsigned short int Sharing_PSP ;
                    unsigned short int Sharing_Number ;
                    unsigned short int First_MB ;
                    unsigned short int Best_MB ;
                    unsigned short int Last_MB ;
                    unsigned char Filler1[10] ;
                    unsigned char Dd ;
                    unsigned char Mm ;
                    unsigned short int Yy_1980 ;
                    unsigned short int Dd_1980 ;
                    unsigned char Day_Of_Week ;
                    unsigned char Filler2[3] ;
                    unsigned char Driver_Request_Header[30] ;
                    void* Entry_Point ;
                    unsigned char Driver_Request_Header1[22] ;
                    unsigned char Driver_Request_Header2[30] ;
                    unsigned char Filler3[6] ;
                    unsigned char CLOCK_Record[6] ;
                    unsigned char Filler4[2] ;
                    char Fn1[128] ;
                    char Fn2[128] ;
                    TSDB_Rec Sdb ;
                    TDOS_Dir_Rec Found_File ;
                    TDOS_CDSv4 Drive_CDSCopy ;
                    char Fcb_Fn1[11] ;
                    unsigned char Filler5 ;
                    char Fcb_Fn2[11] ;
                    unsigned char Filler6[11] ;
                    unsigned char Search_Attributes ;
                    unsigned char Open_Mode ;
                    unsigned char Filler7[3] ;
                    unsigned char DOS_Flag ;
                    unsigned char Filler8[9] ;
                    unsigned char Termination_Type ;
                    unsigned char Filler9[3] ;
                    void* Error_DPB ;
                    void* Stack_Frame ;
                    unsigned short int Stored_SP1 ;
                    void* DPB_Pointer ;
                    unsigned short int Disk_Buffer ;
                    unsigned char Filler10[8] ;
                    unsigned char Media_ID ;
                    unsigned char Filler11 ;
                    void* Filler12 ;
                    void* Current_SFT ;
                    void* Drive_CDSPTR ;
                    void* Caller_FCB ;
                    unsigned char Filler13[4] ;
                    void* Jft ;
                    unsigned short int Fn1_Csofs ;
                    unsigned short int Fn2_Csofs ;
                    unsigned char Filler14[12] ;
                    unsigned short int Directory_Cluster ;
                    unsigned char Filler15[10] ;
                    signed int Offset ;
                    unsigned short int Filler16 ;
                    unsigned short int Partial_Bytes ;
                    unsigned short int Sectors ;
                    unsigned char Filler17[6] ;
                    signed int Appended_Bytes ;
                    void* Disk_Buffer1 ;
                    void* Sft1 ;
                    unsigned short int Store_BX ;
                    unsigned short int Store_DS ;
                    unsigned short int Temp ;
                    void* Previous_Call_Frame ;
                    unsigned char Filler18[5] ;
                    void* Filler19 ;
                    unsigned short int Special_Open_Action ;
                    unsigned short int SPop_Attr ;
                    unsigned short int SPop_Mode ;
                    unsigned char Filler20[9] ;
                    unsigned short int Save_DS ;
                    unsigned char Filler21[5] ;
                    void* Filename ;
                    void* Filler22 ;
                    unsigned short int Ssve_SS ;
                    unsigned short int Save_SP ;
                    unsigned char Stack_Switch ;
                    TSDB_Rec Ren_SrcFile ;
                    TDOS_Dir_Rec Ren_File ;
                    unsigned char Error_Stack[331] ;
                    unsigned char Disk_Stack[384] ;
                    unsigned char Char_Stack[384] ;
                    unsigned char AH_Flag ;
                    unsigned char Drive ;
                    unsigned char Flag ;
                    unsigned char Filler23[9] ;
};

struct TOfS /* Structure for manipulation of pointers */
{
    unsigned short int Offset ;
	unsigned short int Segment ; 
};

union XRegisters /* Extended Registers type */
{
    struct
	{
	    unsigned short int Bp ; 
		unsigned short int Es ;
		unsigned short int Ds ; 
		unsigned short int Di ;  
		unsigned short int Si ; 
		unsigned short int Dx ; 
		unsigned short int Cx ; 
		unsigned short int Bx ; 
		unsigned short int Ax ; 
		unsigned short int Ss ; 
		unsigned short int Sp ; 
		unsigned short int Cs ; 
		unsigned short int Ip ;
        unsigned short int Flags ;
	};
	struct
	{
        unsigned short int Fill1a ; 
		unsigned short int Fill1b ; 
		unsigned short int Fill2a ; 
		unsigned short int Fill2b ; 
		unsigned short int Fill3a ; 
		unsigned short int Fill3b ;
        unsigned short int Fill4a ; 
		unsigned short int Fill4b ; 
		unsigned short int Fill5a ; 
		unsigned short int Fill5b ;
		unsigned short int Dl ; 
		unsigned short int Dh ; 
		unsigned short int Cl ; 
		unsigned short int Ch ;
        unsigned char Bl ;
		unsigned short int Bh ; 
		unsigned short int Al ; 
		unsigned short int Ah ;
	};
};

       /* CTR (Clock Transfer packed record) */
struct TDOS_CTR
{
    unsigned short int Dd_1980 ;
    unsigned char Minutes ;
    unsigned char Hours ;
    unsigned char Hundreths ;
    unsigned char Seconds ;
};

       /* DDH (Device Driver Header) */
union TDOS_DDH
{
	struct
	{
        void* Next ;
        unsigned short int Attributes ;
        unsigned short int Exit_Point ;
        unsigned short int Entry_Point ;
        unsigned char Sub_Units ;
        unsigned char Filler[7] ;
        unsigned short int Reserved ;
        char Drive ;
        unsigned char Units ;
        unsigned char Signature[6] ;
    } ;
    struct
	{
		unsigned char Filler1[10] ;
        char Name[8] ;
    } ;
};

       /* DDB (DOS Disk Buffer), DOS V2.n */
struct TDDBv2
{
                    void* Next ;
                    unsigned char Drive ;
                    unsigned char Filler[3] ;
                    unsigned short int Sector ;
                    unsigned char Copies ;
                    unsigned char Offset ;
                    void* Dpb ;
                    unsigned char Start ;
};

       /* DDB (DOS Disk Buffer), DOS V3.n */
struct TDDBv3
{
                    void* Next ;
                    unsigned char Drive ;
                    unsigned char Flags ;
                    unsigned short int Sector ;
                    unsigned char Copies ;
                    unsigned char Offset ;
                    void* Dpb ;
                    unsigned short int Filler ;
                    unsigned char Start ; 
};

       /* DDB (DOS Disk Buffer), DOS V4.n */
struct TDDBv4
{
                    unsigned short int Next ;
                    unsigned short int Previous ;
                    unsigned char Drive ;
                    unsigned char Flags ;
                    signed int Sector ;
                    unsigned char Copies ;
                    unsigned short int Offset ;
                    void* Dpb ;
                    unsigned short int Count ;
                    unsigned char Filler ;
                    unsigned char Start ;
};

       /* DBHCH (Disk Buffer Hash Chain Head), DOS V4.n */
struct TDOS_DBHCH
{
                    unsigned short int Page ;
                    void* Least_Used ;
                    unsigned char Dirty_Buffers ;
                    unsigned char Filler ;
};

       /* DBI (Disk Buffer Info), DOS V4.0 */
struct TDOS_DBIv40
{
                    void* Dbhch ;
                    unsigned short int Chain_Count ;
                    void* Lookahead ;
                    unsigned char Flag ;
                    unsigned short int Handle ;
                    unsigned short int Page ;
                    unsigned short int Filler ;
                    unsigned short int Segment ;
                    unsigned short int Filler1 ;
                    unsigned short int Mapping[4] ;
};

       /* DBI (Disk Buffer Info), DOS V4.01 */
struct TDOS_DBIv41
{
                      void* Dbhch ;
                      unsigned short int Chain_Count ;
                      void* Lookahead ;
                      unsigned short int Lookahead_Count ;
                      unsigned char Flag ;
                      unsigned short int Segment ;
                      unsigned short int Page ;
                      unsigned short int Segment1 ;
                      unsigned short int Page1 ;
                      unsigned char Frames ;
                      unsigned short int Workspace ;
                      unsigned short int Handle ;
                      unsigned short int Page2 ;
                      unsigned short int Filler ;
                      unsigned short int Frame ;
                      unsigned short int Filler1 ;
                      unsigned char Flag1 ;
};

       /* Module link list entry, Windows NT/2000/XP */
struct TModule_Link_List_Entry
{
                                   TModule_Link_List_Entry *Next ;
                                   TModule_Link_List_Entry *Previous ;
                                   signed int Filler ;
                                   signed int Filler1 ;
                                   signed int Base_Address ; /* Base address of module */
};

       /* Module Link List, Windows NT/2000/XP */
struct TModule_Link_List
{
                             signed int Filler ;
                             signed int Filler1 ;
                             signed int Filler2 ;
                             void* List_Start ;
                             void* List_End ;
                             TModule_Link_List_Entry *Memory_List_Start ;
                             TModule_Link_List_Entry  *Memory_List_End ;
                             void* Initialization_Order_List_Start ;
                             void* Initialization_Order_List_End ;
};

       /* PEB (Process Environment Block), Windows NT/2000/XP */
struct TPEB
{
    signed int Environment ;
    signed int Filler ;
    signed int Filler1 ;
    TModule_Link_List *Module_Link_List ;
    signed int Process_Parameters ;
    signed int Subsystem_Data ;
    signed int Process_Heap ;
};

struct TPE_Header
{
    unsigned short int Machine ;
    unsigned short int NumberOfSections ;
    signed int TimeDateStamp ;
    signed int PointtoSymbolTable ;
    signed int NumberofSymbols ;
    unsigned short int SizeOfOptionalHeader ;
    unsigned short int Characteristics ;
};

struct TPE_Optional_Header
{
                              unsigned short int Magic ; // Always appears to be set to 0x010B.
                              unsigned char MajorLinkerVersion ;
                              unsigned char MinorLinkerVersion ;
                              unsigned int SizeOfCode ;
                              unsigned int SizeOfInitializedData ;
                              unsigned int SizeOfUninitializedData ;
                              unsigned int AddressOfEntryPoint ;
                              unsigned int BaseOfCode ;
                              unsigned int BaseOfData ;
                              unsigned int ImageBase ;
                              unsigned int SectionAlignment ;
                              unsigned int FileAlignment ;
                              unsigned short int MajorOperatingSystemVersion ;
                              unsigned short int MinorOperatingSystemVersion ;
                              unsigned short int MajorImageVersion ;
                              unsigned short int MinorImageVersion ;
                              unsigned short int MajorSubsystemVersion ;
                              unsigned short int MinorSubsystemVersion ;
                              unsigned int Reserved1 ;
                              unsigned int SizeOfImage ;
                              unsigned int SizeOfHeaders ;
                              unsigned int CheckSum ;
                              unsigned short int Subsystem ;
                              unsigned short int DllCharacteristics ;
                              unsigned int SizeOfStackReserve ;
                              unsigned int SizeOfStackCommit ;
                              unsigned int SizeOfHeapReserve ;
                              unsigned int SizeOfHeapCommit ;
                              unsigned int LoaderFlags ;
                              unsigned int NumberOfRvaAndSizes ;
};

struct TPE_Image_Section_Header
{
                                    char Name[8] ;
                                    union
                                    {									
                                    	signed int PhysicalAddress ;
                                        signed int VirtualSize ;
                                    } Misc ;
                                    signed int VirtualAddress ;
                                    signed int SizeOfRawData ;
                                    signed int PointerToRawData ;
                                    signed int PointerToRelocations ;
                                    signed int PointerToLinenumbers ;
                                    unsigned short int NumberOfRelocations ;
                                    unsigned short int NumberOfLinenumbers ;
                                    signed int Characteristics ;
};
