/*
        Program Name : _UOS_File_System
        Package Name : Subroutine Library Layer 2
        Purpose      : UOS File System class definition
        Institution  : Conroy & Conroy Co.
        Date Written : 6-Nov-2015
        Written By   : Alan Conroy
        Version      : 1.0

	    Copyright (C) 2015 by Alan Conroy.  All rights reserved.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Abstract UOS File System class definition.
*/

	 ComInter, // TCommon_COM_Interface
	 Heaps, // TCOM_Heap64
	 _Stores, // TStore_Address64
	 _UE, // TUnified_Exception

const Max_Header_Data_Stream = 4 ;
const Max_Header_Cluster = 4 ;

struct TData_Stream
{
	long long int Name ;
	long long int Pointer ;
} ;

struct TUOS_File_Header
{
	// Name and sizes...
	long int Name ;
	long long int Size ; // Size on disk
	long long int EOF ; // Logical size
	long long int Uncompressed_Size ;
	unsigned long int Clustersize ;
	unsigned long int Record_Size ;

	// Dates...
	long long int Creation ; // Creation date
	long long int Last_Modified ; // Last write date
	long long int Last_Backup ;
	long long int Last_Access ; // Last read/open date
	long long int Expiration ;

	// Security...
	unsigned long int Creator ;
	unsigned long int Owner ;
	long long int ACL ;

	// Misc...
	long long int Flags ;
	long int Version_Limit ;
	long long int Extension ;

	// Data streams...
	TData_Stream Streams[ Max_Header_Data_Stream + 1 ] ;
	long long int Data_Stream ; // Overflow pointer
	long long int Clusters[ Max_Header_Cluster + 1 ] ;

	// File-system data...
	long long int Parent ;
	long long int File_System ; // Reserved for file system
} ;

	  // File Attribute Flags...
const FAF_DSM_MASK = 3 ; // Mask for data security mode
const FAF_PLACED = 4 ; // Placed at a specific location
const FAF_CONTIGUOUS = 8 ; // Contiguous file data
const FAF_DELETED = 16 ; // Deleted file
const FAF_READONLY = 32 ; // Read-only data
const FAF_LINK = 64 ; // Link to another file
const FAF_SYSTEM = 128 ; // System file
const FAF_HIDDEN = 256 ; // Hidden file
const FAF_DIRECTORY = 512 ; // Directory file
const FAF_PERMANENT = 1024 ; // Non-deletable file
const FAF_COPYING = 2048 ; // In the process of copying


struct FS_Boot_Record // First 24 bytes of store header
	unsigned char Bootstrap1 [ 16 ] ;
	long long int Header ;
} ;

struct FS_Header // 204 bytes
	short int ID ; // 255
	unsigned char ID1 ; // 135
	unsigned char Version ; // UOSFS format version 10 = V1.0
	long long int AT_Offset ; // Current AT location
	long long int AT_Size ; // AT size
	long long int Flags ;
	unsigned long int Clustersize ; // Volume clustersize
	unsigned long int Folder_Clustersize ; // Default folder file clustersize
	long long int Root ; // Root folder header
	char Volume_Label [ 128 ] ;
	char Password [ 32 ] ;
} ;

const UOSFSF_Dirty = 1 ; // 0=clean, 1=dirty
const UOSFSF_Private = 2 ; // 0=private volume, 2=public volume

