/*{
        Program Name : _Cache
        Package Name : LB
        Purpose      : Cache class definition
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Jun-1994
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1994-2009 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

        *    DATE        BY          REASON                         *

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          This unit defines classes for handling caches.
*/

#ifndef _CACHEH
	#define _CACHEH

	#include "eDefines.h"

class TCOM_Cache : public TCommon_COM_Interface
{
      public: //API...
          virtual int __stdcall Read_Cache( void* Data,
              int Address, int Size ) = 0;
          //Reads up to Size bytes from the cache for the specified address.  Returns the number of bytes actually read (if 0 then the data was not cached).  Cached data is copied to Data.

          virtual void __stdcall Write_Cache( void* Data,
              int Address, int Size ) = 0;
          //Writes to cache up to Size bytes from Data.

          virtual void __stdcall Clear() = 0; //Clears cache

          virtual void __stdcall Set_Max_Size( int Size ) = 0;
          //Sets maximum cache size in bytes.  If this is more than is currently cached, some cache is thrown away.

          virtual int __stdcall Get_Max_Size() = 0;
          //Gets current maximum cache size in bytes.

          virtual int __stdcall Get_Size() = 0;
          //Returns amount of currently cached data, in bytes.

          virtual bool __stdcall Lock( int Address ) = 0;
          //Locks the data starting at the specified address into memory so that it is not thrown away.  Returns false if the specified data is not in the cache.

          virtual bool __stdcall Unlock( int Address ) = 0;
          //Unlocks the data starting at the specified address in memory.  Returns false if the specified data is not in the cache, or if it is not locked.

          virtual int __stdcall Get_Hits() = 0; //Total read hits
          virtual int __stdcall Get_Misses() = 0; //Read misses
          virtual int __stdcall Get_Writes() = 0;//Total writes
          virtual int __stdcall Get_Turns() = 0; //Number of times blocks were thrown away to make room for more
          virtual int __stdcall Get_Replacements() = 0; //Number of write hits
          virtual void __stdcall Set_Hits( int Value ) = 0; //Total read hits
          virtual void __stdcall Set_Misses( int Value ) = 0; //Read misses
          virtual void __stdcall Set_Writes( int Value ) = 0; //Total writes
          virtual void __stdcall Set_Turns( int Value ) = 0; //Number of times blocks were thrown away to make room for more
          virtual void __stdcall Set_Replacements( int Value ) = 0; //Number of write hits
} ; //TCOM_Cache


class TCOM_Cache64 : public TCommon_COM_Interface
{
    public: //API...
        virtual int __stdcall Read_Cache( void* Data, int64 Address, int64 Size ) = 0;
        //Reads up to Size bytes from the cache for the specified address.  Returns the number of bytes actually read (if 0 then the data was not cached).  Cached data is copied to Data.

        virtual void __stdcall Write_Cache( void* Data, int64 Address, int64 Size ) = 0;
        //Writes to cache up to Size bytes from Data.

        virtual void __stdcall Clear() = 0; //Clears cache

        virtual void __stdcall Set_Max_Size( int Size ) = 0;
        //Sets maximum cache size in bytes.  If this is more than is currently cached, some cache is thrown away.

        virtual int __stdcall Get_Max_Size() = 0;
        //Gets current maximum cache size in bytes.

        virtual int __stdcall Get_Size() = 0;
        //Returns amount of currently cached data, in bytes.

        virtual bool __stdcall Lock( int64 Address ) = 0;
        //Locks the data starting at the specified address into memory so that it is not thrown away.  Returns false if the specified data is not in the cache.

        virtual bool __stdcall Unlock( int64 Address ) = 0;
        //Unlocks the data starting at the specified address in memory.  Returns false if the specified data is not in the cache, or if it is not locked.

        virtual int __stdcall Get_Hits() = 0; //Total read hits

        virtual int __stdcall Get_Misses() = 0; //Read misses
        virtual int __stdcall Get_Writes() = 0; //Total writes
        virtual int __stdcall Get_Turns() = 0; //Number of times blocks were thrown away to make room for more
        virtual int __stdcall Get_Replacements() = 0; //Number of write hits
        virtual void __stdcall Set_Hits( int Value ) = 0;  //Total read hits
        virtual void __stdcall Set_Misses( int Value ) = 0;  //Read misses
        virtual void __stdcall Set_Writes( int Value ) = 0;  //Total writes
        virtual void __stdcall Set_Turns( int Value ) = 0; //Number of times blocks were thrown away to make room for more
        virtual void __stdcall Set_Replacements( int Value ) = 0; //Number of write hits
} ; //TCOM_Cache64
#endif

