{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
              Program Name : _Files
              Package Name : Subroutine Library - Layer 1
              Purpose      : Abstract File classes
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 2.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2015 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the TSFile class.
}

unit _Files ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I EDEFINES.INC}

{$IFDEF DELPHI}
  {$DEFINE WINDOWS}
{$ENDIF}

uses { Layer 0... }
{$IFNDEF MOBILE}
     _DebugIn, { PDebug_Interface }
{$ENDIF}
     TypeDefs, { TRecord_Size }
     _Types, { Ansistring }

     { Layer 1... }
     Stores, { TSBase_Store }
     _Stores, { TStore_Address }
     _UE, { PSUnified_Exception }

{$IFDEF WINDOWS}
  {$IFDEF FPC_OR_DELPHI}
     SysUtils ;
  {$ELSE}
     Windos,
     strings,
     Objects;
  {$ENDIF}

{$ELSE}
     { Borland... }
     Dos,
     Windos,
     Objects,
     Strings ;
{$ENDIF}

type TSFile = object( TSBase_Store )
                 public
                     Extendable : boolean ; { True if we can extend the file }
                     File_Mode : integer ; { Mode that the file is opened in }
                     IO_Error : PSUnified_Exception ;
                     Record_Size : TRecord_Size ; { Size of record, in bytes, for non-text files }

                 public { API... }
                     procedure Reset( const Name : Ansistring ) ; virtual ;
                     procedure Rewrite( const Name : Ansistring ) ; virtual ;
                     procedure Append( const Name : Ansistring ) ; virtual ;
                     procedure Close ; virtual ;
                     procedure Seek( Pos : TStore_Address ) ; virtual ;

                     function Read_Integer : TStore_Address ; virtual ;

                     function Read_Float : Extended ; virtual ;

                     procedure Read( var Buffer ) ; virtual ;
                     procedure Write( var Buffer ) ; virtual ;
                     procedure Readln( var Buffer : AnsiString ) ; virtual ;
                     procedure Writeln( Buffer : Ansistring ) ; virtual ;
                     procedure Blockread( var Buf ; Count : TRecord_Size ;
                         var Res : TRecord_Size ) ; virtual ;
                     procedure Blockwrite( var Buf ; Count : TRecord_Size ;
                         var Res : TRecord_Size ) ; virtual ;
                     function EOF : boolean ; virtual ;
                     function EOLn : boolean ; virtual ;
                     function FilePos : TStore_Size ; virtual ;
                     function Get_Size : TStore_Size ; virtual ;
                     procedure Set_Size( New_Size : TStore_Size ) ; virtual ;
                     procedure Erase( const Name : Ansistring ) ; virtual ;
                     procedure Flush ; virtual ;
                     function FileName : Ansistring ; virtual ;
                     function SeekEOF : Boolean ; virtual ;
                     function SeekEOLn : Boolean ; virtual ;
                     procedure GetFTime( var _Time : Longint ) ; virtual ;
                     procedure SetFTime( _Time : longint ) ; virtual ;
                     procedure SetTextBuf( var buf ; size : Word ) ;
                         virtual ;
                     procedure Set_Extend( State : boolean ) ; virtual ;
                     procedure Truncate( Size : TStore_Address ) ; virtual ;
                     procedure Initialize ; virtual ;
{$IFNDEF MOBILE}
                     function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}
                     function Get_Attributes : longint ; virtual ;
                     procedure Set_Attributes( Value : longint ) ; virtual ;
                     function Facility : longint ; virtual ;

                     function Version : longint ; virtual ;

                     function Facility_Name : PAnsiChar ; virtual ;
                     function Opened : boolean ; virtual ;
                     function Get_Time : longint ; virtual ;
                     procedure Set_Time( Value : longint ) ; virtual ;
{$IFDEF PROPERTIES}
                     property Attributes : longint
                                  read Get_Attributes
                                  write Set_Attributes ;
                     property Time : longint
                                  read Get_Time
                                  write Set_Time ;
{$ENDIF}
                     procedure GetFAttr( var Attr : Word ) ;
                         virtual ;
                     procedure SetFAttr( Attr : Word ) ; virtual ;
                     procedure Set_Error( X : PSUnified_Exception ) ; virtual ;
             end ;
     PSFile = ^TSFile ;

{$IFDEF DYNAMIC_OBJECTS}
type TCOM_File = class( TStandard_COM_Store )
                     public { API... }
                         procedure Reset( const Name : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Rewrite( const Name : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Append( const Name : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Close ;
                             virtual ; stdcall ; abstract ;
                         procedure Seek( Pos : TStore_Address ) ;
                             virtual ; stdcall ; abstract ;

                         function Read_Integer : TStore_Address ;
                             virtual ; stdcall ; abstract ;

                         function Read_Float : Extended ;
                             virtual ; stdcall ; abstract ;

                         procedure Read( var Buffer ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Write( var Buffer ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Readln( var Buffer : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Writeln( Buffer : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Blockread( var Buf ; Count : TRecord_Size ;
                             var Res : TRecord_Size ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Blockwrite( var Buf ; Count : TRecord_Size ;
                             var Res : TRecord_Size ) ;
                             virtual ; stdcall ; abstract ;
                         function EOF : boolean ; virtual ; stdcall ; abstract ;
                         function EOLn : boolean ;
                             virtual ; stdcall ; abstract ;
                         function FilePos : TStore_Size ;
                             virtual ; stdcall ; abstract ;
                         function Get_Size : TStore_Size ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Size( New_Size : TStore_Size ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Erase( const Name : PAnsiChar ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Flush ; virtual ; stdcall ; abstract ;
                         function FileName : PAnsiChar ;
                             virtual ; stdcall ; abstract ;
                         function SeekEOF : Boolean ;
                             virtual ; stdcall ; abstract ;
                         function SeekEOLn : Boolean ;
                             virtual ; stdcall ; abstract ;
                         procedure GetFTime( var _Time : Longint ) ;
                             virtual ; stdcall ; abstract ;
                         procedure SetFTime( _Time : longint ) ;
                             virtual ; stdcall ; abstract ;
                         procedure SetTextBuf( var buf ; size : Word ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Extend( State : boolean ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Truncate( Size : TStore_Address ) ;
                             virtual ; stdcall ; abstract ;
                         function Get_Attributes : longint ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Attributes( Value : longint ) ;
                             virtual ; stdcall ; abstract ;
                         function Facility : longint ;
                             virtual ; stdcall ; abstract ;

                         function Version : longint ;
                             virtual ; stdcall ; abstract ;

                         function Facility_Name : PAnsiChar ;
                             virtual ; stdcall ; abstract ;
                         function Opened : boolean ;
                             virtual ; stdcall ; abstract ;
                         function Get_Time : longint ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Time( Value : longint ) ;
                             virtual ; stdcall ; abstract ;
                         procedure GetFAttr( var Attr : Word ) ;
                             virtual ; stdcall ; abstract ;
                         procedure SetFAttr( Attr : Word ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Error( F, X : integer ) ;
                             virtual ; stdcall ; abstract ;
                         function Get_Extendable : boolean ;
                             virtual ; stdcall ; abstract ;
                         function Get_File_Mode : integer ;
                             virtual ; stdcall ; abstract ;
                         function Get_IO_Error : TUnified_Exception ;
                             virtual ; stdcall ; abstract ;
                         function Get_Record_Size : TRecord_Size ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Extendable( Value : boolean ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_File_Mode( Value : integer ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_IO_Error( Value : TUnified_Exception ) ;
                             virtual ; stdcall ; abstract ;
                         procedure Set_Record_Size( Value : TRecord_Size ) ;
                             virtual ; stdcall ; abstract ;

                         property Extendable : boolean
                             read Get_Extendable
                             write Set_Extendable ;
                         property File_Mode : integer
                             read Get_File_Mode
                             write Set_File_Mode ;
                         property IO_Error : TUnified_Exception
                             read Get_IO_Error
                             write Set_IO_Error ;
                         property Record_Size : TRecord_Size
                             read Get_Record_Size
                             write Set_Record_Size ;

                         property Attributes : longint
                             read Get_Attributes
                             write Set_Attributes ;
                         property Time : longint
                             read Get_Time
                             write Set_Time ;
                 end ;
{$ENDIF}

implementation

uses { Layer 0... }
     _Common ; { Abstract_Error } 

{ TSFile methods... }

{ API... }

procedure TSFile.Reset( const Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Rewrite( const Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Append( const Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Close ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Seek( Pos : TStore_Address ) ;

begin
    Abstract_Error ;
end ;


function TSFile.Read_Integer : TStore_Address ;

begin
    Read_Integer := 0 ;
    Abstract_Error ;
end ;


function TSFile.Read_Float : Extended ;

begin
    Read_Float := 0.0 ;
    Abstract_Error ;
end ;


procedure TSFile.Read( var Buffer ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Write( var Buffer ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Readln( var Buffer : AnsiString ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Writeln( Buffer : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Blockread( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

begin
    Res := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.Blockwrite( var Buf ; Count : TRecord_Size ;
    var Res : TRecord_Size ) ;

begin
    Res := 0 ;
    Abstract_Error ;
end ;


function TSFile.EOF : boolean ;

begin
    EOF := True ;
    Abstract_Error ;
end ;


function TSFile.EOLn : boolean ;

begin
    EOLn := True ;
    Abstract_Error ;
end ;


function TSFile.FilePos : TStore_Size ;

begin
    FilePos := 0 ;
    Abstract_Error ;
end ;


function TSFile.Get_Size : TStore_Size ;

begin
    Get_Size := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.Set_Size( New_Size : TStore_Size ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Erase( const Name : Ansistring ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Flush ;

begin
    Abstract_Error ;
end ;


function TSFile.FileName : Ansistring ;

begin
    FileName := '' ;
    Abstract_Error ;
end ;


function TSFile.SeekEOF : Boolean ;

begin
    SeekEOF := True ;
    Abstract_Error ;
end ;


function TSFile.SeekEOLn : Boolean ;

begin
    SeekEOLn := True ;
    Abstract_Error ;
end ;


procedure TSFile.GetFTime( var _Time : Longint ) ;

begin
    _Time := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.SetFTime( _Time : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.SetTextBuf( var buf ; size : Word ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Set_Extend( State : boolean ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Truncate( Size : TStore_Address ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Initialize ;

begin
    Abstract_Error ;
end ;


{$IFNDEF MOBILE}
function TSFile.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ;
end ;
{$ENDIF}


function TSFile.Get_Attributes : longint ;

begin
    Get_Attributes := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.Set_Attributes( Value : longint ) ;

begin
    Abstract_Error ;
end ;


function TSFile.Facility : longint ;

begin
    Facility := -1 ;
    Abstract_Error ;
end ;


function TSFile.Version : longint ;

begin
    Version := 0 ;
    Abstract_Error ;
end ;


function TSFile.Facility_Name : PAnsiChar ;

begin
    Facility_Name := '' ;
    Abstract_Error ;
end ;


function TSFile.Opened : boolean ;

begin
    Opened := False ;
    Abstract_Error ;
end ;


function TSFile.Get_Time : longint ;

begin
    Get_Time := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.Set_Time( Value : longint ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.GetFAttr( var Attr : Word ) ;

begin
    Attr := 0 ;
    Abstract_Error ;
end ;


procedure TSFile.SetFAttr( Attr : Word ) ;

begin
    Abstract_Error ;
end ;


procedure TSFile.Set_Error( X : PSUnified_Exception ) ;

begin
    Abstract_Error ;
end ;



end.
