/*
              Program Name : _FileSys
              Package Name : Subroutine Library - Layer 1
              Purpose      : Abstract File System rlass
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2008 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the abstract TSFile_System, and related, objects.
*/

#ifndef _FILESYSH
	#define _FILESYSH

	#include "_stores.h"
	
struct FS_Header_Information {
								 short int Header_Version ; // Version of FS_Header_Information
								 char ID[ 128 ] ; // Label for this file system device
								 TStore_Address Root ; // Pointer to root of file system/directories
								 short int Version ; // File system structure version
								 short int Default_Cluster_Size ; // Default cluster size for allocations (0 is invalid)
								 short int Flags ; // FSD_*
							 } ;

// Used to hold results of Filespec_String_Scan
struct TFSSS {
				 char Device[ 256 ] ; // Device, server, or protocol specification, including colon
				 char Path[ 256 ] ; // Path, including final separator
				 char Filename[ 256 ] ; // Full filename
				 char Name[ 256 ] ; // Name, minus (last) extension
				 char Extension[ 256 ] ; // (Last) extension (not including dot)
				 char Password[ 256 ] ; // Embedded password
				 char Switches[ 256 ] ; // Any switches at end of filename
				 int Flags ; // Result flags (see FSSS_*)
			 } ;

	  // File System File Flags...
const int FSFF_Directory = 1 ; // A directory file
const int FSFF_Journaling = 2 ; // A journalled file
const int FSFF_File_System = 4 ; // A file-system file
const int FSFF_Compressed = 8 ; // A compressed file
const int FSFF_Erase = 16 ; // Erase file upon deletion
const int FSFF_Statistics = 32 ; // Keep file access statistics
const int FSFF_Read_Verify = 64 ; // Verify read accesses
const int FSFF_Write_Verify = 128 ; // Verify write accesses
const int FSFF_Hidden = 256 ; // A hidden file
const int FSFF_Contiguous = 512 ; // File is contiguous on file system device
const int FSFF_Deleted = 1024 ;
// File is deleted but still open (marked for deletion)
const int FSFF_Protected = 2048 ; // for DOS compatibility only
const int FSFF_System = 4096 ; // for DOS compatibility only
const int FSFF_VolumeID = 8192 ; // for DOS compatibility only
const int FSFF_Archive = 16384 ; // for DOS compatibility only
// FS File Flags 32768+ are reserved for future use

// File System Flags...
const int FSF_Fixed_Size = 1 ; // The file system is on a fixed-size device
const int FSF_Read_Only = 2 ; // The file system is opened in read-only mode
const int FSF_NoOwnStore = 4 ; // The file system doesn't own the store - so don't close it when file system closes
      // 8-32768 = reserved

// Flags for TFSSS.Flags
const int FSSS_Null_Extension = 1 ; // Filename had no dot for an extension

typedef void* PSFile_System;
typedef void* PSWildcard;

struct TFind_Record {
                        int Initialized ; // 0 if this record was initialized
                        PSWildcard Context ; // Top-level Wildcard object
                        char Separator ; // File list separator (NULL means none)
                        char Specification[ 256 ] ; // Original specification
                        char File_Name[ 256 ] ; // Full name and path of found file
                        char File_List[ 256 ] ; // comma-delimited list of paths
                        PSFile_System FS ; // File system
                        char Wildcard_Context[ 256 ] ; // Used by FindFirst/FindNext
                        int Wildcard_Attributes ; // Used by FindFirst/FindNext
                        char Last_Name[ 256 ] ; // Used by FindFirst/FindNext
                        char Found_Name[ 256 ] ; // Name (not path) of found file
                        int Found_Attributes ; // Attributes of found file
                        char Data[ 512 ] ; // Context information - file system dependant
                    } ;
#endif
