{$IFNDEF FPC}
  {$N+}
  {$E+}
{$ENDIF}
{
        Program Name    : _Languag
        Package Name    : Subroutine Library - Layer 1
        Purpose         : Language-related data conversion routines
        Institution     : Conroy & Conroy Co.
        Date written    : 1-May-15
        Written by      : Alan Conroy
        Version         : 1.0

        Copyright (C) 2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

      ***************************************************************
      *                                                             *
      *      M O D I F I C A T I O N   H I S T O R Y                *
      *                                                             *
      ***************************************************************

          DATE        BY            REASON

      *******************************************************
      *                                                     *
      *          P R O G R A M   P U R P O S E              *
      *                                                     *
      *******************************************************

      This module contains routines for converting numeric data.
}

unit _Languag ;

{$I EDEFINES.INC}

interface

uses { Layer 0... }
     _ComInter, { TBase_COM_Interface }
     CommBase, { TBase_Delphi_Object }
     _Types ; { Ansistring }

const NTS_Arabic = 0 ;
      NTS_Roman = 1 ;
      NTS_English = 2 ;
      NTS_Arabic_Ordinal = 3 ;
      NTS_English_Ordinal = 4 ;
      NTS_Metric_Prefix = 5 ;
      NTS_Radix = 6 ;
      NTS_Alt_Radix = 7 ;
      NTS_Group0 = 8 ;
      NTS_Group1 = 9 ;
      NTS_Group2 = 10 ;
      NTS_Group3 = 11 ;
      NTS_Prefix = 12 ;
      NTS_Alt_Prefix = 13 ;
      NTS_Greek_Prefix = 14 ;
      NTS_Multiplier = 15 ;
      NTS_Math = 16 ;
      NTS_Positional = 17 ;
      NTS_2D = 18 ;
      NTS_3D = 19 ;
      NTS_Government = 20 ;
      NTS_Metric_Abbreviation = 21 ;
      NTS_Binary_Prefix = 22 ;
      NTS_Ordinal_Multiplier = 23 ;

type TSLanguage = object( TBase_Object )
                      public
                          { Converts a number to a specific numeric string. }
                          function Numeric_To_String( Numeric : double ;
                              Typ : integer ) : Ansistring ;
                              virtual ;

                          { Determines last position in S that is part of a
                            number in English.  Returns < 1 if S doesn't start
                            with a valid number in English. }
                          function Numeric_Extent( S : Ansistring ) : integer ;
                              virtual ;

                          function To_Numeric( S : Ansistring ;
                              var Numeric : double ) : boolean ;
                              virtual ;

                          { Converts a numeric string to a number and returns
                            the type of the value.  The return is True if the
                            passed numeric string is recognized or false
                            otherwise. }
                          function String_To_Numeric( S : Ansistring ;
                              var Numeric : double ;
                              var Typ : integer ) : boolean ;
                              virtual ;

                          { Converts a numeric string to a number and returns
                            the type of the value.  The return is True if the
                            passed numeric string is recognized or false
                            otherwise.  This differs from String_To_Numeric in
                            that it recognizes only strings containing valid
                            prefix and suffixes, even if the specific
                            combination is not a recognized English word.  For
                            example, "centgram" would be recognized as a
                            100-sided 2D object.  However, strings without both
                            prefix and suffix are not recognized (such as "one",
                            or "hex"). }
                          function Generic_String_To_Numeric( S : Ansistring ;
                              var Numeric : double ;
                              var Typ : integer ) : boolean ;
                              virtual ;
                  end ; { TSLanguage }
     PSLanguage = ^TSLanguage ;

{$IFDEF DYNAMIC_OBJECTS}
type TLanguage = class( TBase_Delphi_Object )
                     public
                         { Converts a number to a specific numeric string. }
                         function Numeric_To_String( Numeric : double ;
                             Typ : integer ) : Ansistring ; virtual ; abstract ;

                         { Determines last position in S that is part of a number in English.  Returns
                           < 1 if S doesn't start with a valid number in English. }
                         function Numeric_Extent( S : Ansistring ) : integer ;
                             virtual ; abstract ;

                         function To_Numeric( S : Ansistring ;
                             var Numeric : double ) : boolean ;
                             virtual ; abstract ;

                         { Converts a numeric string to a number and returns the type of the value.  The
                           return is True if the passed numeric string is recognized or false
                           otherwise. }
                         function String_To_Numeric( S : Ansistring ;
                             var Numeric : double ;
                             var Typ : integer ) : boolean ;
                             virtual ; abstract ;

                         { Converts a numeric string to a number and returns the type of the value.  The
                           return is True if the passed numeric string is recognized or false
                           otherwise.  This differs from String_To_Numeric in that it recognizes only
                           strings containing valid prefix and suffixes, even if the specific combination
                           is not a recognized English word.  For example, "centgram" would be recognized
                           as a 100-sided 2D object.  However, strings without both prefix and suffix
                           are not recognized (such as "one", or "hex"). }
                         function Generic_String_To_Numeric( S : Ansistring ;
                             var Numeric : double ;
                             var Typ : integer ) : boolean ;
                             virtual ; abstract ;
                 end ; { TLanguage }


type TCOM_Language = class( TBase_COM_Interface )
                         public
                             { Converts a number to a specific numeric string. }
                             function Numeric_To_String( Numeric : double ;
                                 Typ : integer ) : Ansistring ;
                                 virtual ; stdcall ; abstract ;

                             { Determines last position in S that is part of a number in English.  Returns
                               < 1 if S doesn't start with a valid number in English. }
                             function Numeric_Extent( S : Ansistring ) : integer ;
                                 virtual ; stdcall ; abstract ;

                             function To_Numeric( S : Ansistring ;
                                 var Numeric : double ) : boolean ;
                                 virtual ; stdcall ; abstract ;

                             { Converts a numeric string to a number and returns the type of the value.  The
                               return is True if the passed numeric string is recognized or false
                               otherwise. }
                             function String_To_Numeric( S : Ansistring ;
                                 var Numeric : double ;
                                 var Typ : integer ) : boolean ;
                                 virtual ; stdcall ; abstract ;

                             { Converts a numeric string to a number and returns the type of the value.  The
                               return is True if the passed numeric string is recognized or false
                               otherwise.  This differs from String_To_Numeric in that it recognizes only
                               strings containing valid prefix and suffixes, even if the specific combination
                               is not a recognized English word.  For example, "centgram" would be recognized
                               as a 100-sided 2D object.  However, strings without both prefix and suffix
                               are not recognized (such as "one", or "hex"). }
                             function Generic_String_To_Numeric( S : Ansistring ;
                                 var Numeric : double ;
                                 var Typ : integer ) : boolean ;
                                 virtual ; stdcall ; abstract ;
                     end ; { TCOM_Language }
{$ENDIF}

implementation

uses { Layer 0... }
     _Common ; { Abstract_Error }

{ TSLanguage methods... }

function TSLanguage.Numeric_To_String( Numeric : double ;
    Typ : integer ) : Ansistring ;

begin
    Result := '' ;
    Abstract_Error ;
end ;


function TSLanguage.Numeric_Extent( S : Ansistring ) : integer ;

begin
    Result := 0 ;
    Abstract_Error ;
end ;


function TSLanguage.To_Numeric( S : Ansistring ; var Numeric : double ) : boolean ;

begin
    Result := False ;
    Abstract_Error ;
end ;


function TSLanguage.String_To_Numeric( S : Ansistring ;
  var Numeric : double ; var Typ : integer ) : boolean ;

begin
    Result := False ;
    Abstract_Error ;
end ;


function TSLanguage.Generic_String_To_Numeric( S : Ansistring ;
  var Numeric : double ; var Typ : integer ) : boolean ;

begin
    Result := False ;
    Abstract_Error ;
end ;


end.
