{
        Program Name : _UE
        Package Name : Subroutine Library - Layer 0
        Purpose      : UE definitions
        Institution  : Conroy & Conroy Co.
        Date Written : 9-Feb-2013
        Last Modified:
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 2013-2021 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides definitions for UE (Unified Exceptions).
}

unit _UE ;

interface

{$I EDEFINES.INC}

uses { C&C Subroutine Library - layer 0 }
     _Common { TObject }
     ,_Types { Ansistring }
     ,_DebugIn { TDebug_Interface }
     ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

{$IFDEF DYNAMIC_OBJECTS}
type tUnified_Exception = class
                            public
                                function Initialize : TUnified_Exception ;
                                    virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                procedure Terminate ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                procedure Attach ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                procedure Detach ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Is_Class( Name : PAnsiChar ) : boolean ;
                                    virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Interface_Version : longint ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Get_Facility : longint ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Get_Facility_Version : longint ;
                                    virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Version : longint ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Debugger : TDebug_Interface ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Severity : longint ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Error_Text( var Size, Typ : longint ) : PAnsiChar ;
                                    virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Get_Error : longint ; virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                                function Get_Previous : tUnified_Exception ;
                                    virtual ;
  {$IFDEF STDCALL_SUPPORTED}
                                    stdcall ;
  {$ENDIF}
                                    abstract ;
                        end ; // tUnified_Exception
{$ENDIF}

type PSUnified_Exception = ^tSUnified_Exception ;
     tSUnified_Exception = object( TObject )
                             private { Instance data... }
                                 _Reference_Count : longint ;

                             public
                                 procedure Attach ; virtual ;
                                 procedure Detach ; virtual ;
                                 function Get_Facility : longint ; virtual ;
                                 function Get_Facility_Version : longint ;
                                     virtual ;
                                 function Version : longint ; virtual ;
                                 function Severity : longint ;
                                     virtual ;
                                 function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                     virtual ;
                                 function Get_Error : longint ; virtual ;
                                 function Get_Previous : PsUnified_Exception ;
                                     virtual ;
                                 procedure Terminate ; virtual ;
                                 procedure Free ; virtual ;
                                 function Facility : longint ; virtual ;
                                 function Facility_Version : longint ; virtual ;
                                 function Code : longint ; virtual ;
                         end ;

const UE_Success = 0 ;
const UE_Informational = 1 ;
const UE_Warning = 2 ;
const UE_Error = 3 ;
const UE_Fatal = 4 ;


{ Log a fatal error - for cases where a non-class function has an exception }
procedure Fatal_Error( E : PsUnified_Exception ) ;


implementation

{$IFDEF FPC_OR_DELPHI}
uses Sysutils ; { Exception }
{$ENDIF}


{ tError_Interface methods... }

procedure tSUnified_Exception.Attach ;

begin
    inc( _Reference_Count ) ;
end ;


procedure tSUnified_Exception.Detach ;

begin
    dec( _Reference_Count ) ;
    if( _Reference_Count <= 0 ) then
    begin
        if( Get_Memory_Type = MT_Heap ) then
        begin
            dispose( PSUnified_Exception( @self ), Done ) ;
        end ;
    end ;
end ;


function tSUnified_Exception.Get_Facility : longint ;

begin
    Get_Facility := -1 ;
    Abstract_Error ;
end ;


function tSUnified_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 0 ;
    Abstract_Error ;
end ;


function tSUnified_Exception.Version : longint ;

begin
    Version := 10 ; { Complies with 1.0 of the Unified Exception specification }
end ;


function tSUnified_Exception.Severity : longint ;

begin
    Severity := UE_Error ;
    Abstract_Error ;
end ;


function tSUnified_Exception.Error_Text( var Size, Typ : longint ) : Ansistring ;

begin
    Error_Text := '' ;
    Size := 0 ;
    Typ := 0 ;
    Abstract_Error ;
end ;


function tSUnified_Exception.Get_Error : longint ;

begin
    Get_Error := 0 ;
    Abstract_Error ;
end ;


function tSUnified_Exception.Get_Previous : PsUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure tSUnified_Exception.Terminate ;

begin
    { Does nothing.  Exists solely for overrides. }
end ;


procedure tSUnified_Exception.Free ;

begin
    Terminate ;
    inherited Free ;
end ;


function tSUnified_Exception.Facility : longint ;

begin
    Facility := Get_Facility ;
end ;


function tSUnified_Exception.Facility_Version : longint ;

begin
    Facility_Version := Get_Facility_Version ;
end ;


function tSUnified_Exception.Code : longint ;

begin
    Code := Get_Error ;
end ;


procedure Fatal_Error( E : PsUnified_Exception ) ;

var S : Ansistring ;
    _S, _T : longint ;

begin
    S := E^.Error_Text( _S, _T ) ;
{$IFDEF EXCEPTION}
    E.Free ;
    raise Exception.Create( S ) ;
{$ENDIF}
    writeln( S ) ;
    halt ;
end ;



end.

