{$N+}
{
        Program Name : CacheMan
        Package Name : Subroutine Library - Layer 1
        Purpose      : Cache Manager class
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Jun-1994
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1994-2008 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          The cache manager is used to manage multiple caches.

          LEVEL 2 CERTIFICATION: 5-Sep-2008 by EAC
}

unit CacheMan ;

interface

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

{$I EDEFINES.INC}

uses { Layer 0... }
     Collect, { TCollection }
{$IFDEF BP}
     Objects, { TCollection }
{$ENDIF}
     { Layer 1... }
     _CacheMa ; { TSCache_Manager }

{$IFDEF CC_Regression_Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses { Borland... }
{$IFDEF FPC_OR_Delphi}
     Classes, { TList }
{$ENDIF}
     { Layer 1... }
     Compatib ; { MaxAvail }

{$IFDEF CC_Regression_Test}
var MaxAvail : integer ;
{$ENDIF}

{$IFDEF DYNAMIC_OBJECTS}
type TEvent_Record = record
                         case boolean of
                             True : ( P1, P2 : pointer ) ;
                             False : ( E : TCache_Event ) ;
                     end ;

var R : TEvent_Record ;

type TEvent_List = class
                       private
                           List : TList ;

                       public
                           constructor Create ;
                           destructor Destroy ; override ;

                           function Get_Item( Index : integer ) : TCache_Event ;
                           procedure Set_Item( Index : integer ; Value : TCache_Event ) ;

                           procedure Add( E : TCache_Event ) ;
                           function Count : integer ;
                           procedure Delete( Index : integer ) ;
                           function IndexOf( E : TCache_Event ) : integer ;
                           property Items[ Index : integer ] : TCache_Event
                               read Get_Item
                               write Set_Item ; default ;
                   end ;

constructor TEvent_List.Create ;

begin
    inherited Create ;

    List := TList.Create ;
end ;


destructor TEvent_List.Destroy ;

begin
    List.Free ;

    inherited Destroy ;
end ;


function TEvent_List.Get_Item( Index : integer ) : TCache_Event ;

begin
    move( List.List[ Index * 2 ], R.P1, sizeof( R.P1 ) ) ;
    move( List.List[ Index * 2 + 1 ], R.P2, sizeof( R.P2 ) ) ;
    @Result := @R.E ;
end ;


procedure TEvent_List.Set_Item( Index : integer ; Value : TCache_Event ) ; {UNTESTED}

begin
    move( @Value, @R.E, sizeof( TCache_Event ) ) ;
    List[ Index * 2 ] := R.P1 ;
    List[ Index * 2 + 1 ] := R.P2 ;
end ;


procedure TEvent_List.Add( E : TCache_Event ) ;

begin
    move( @E, @R.E, sizeof( TCache_Event ) ) ;
    List.Add( R.P1 ) ;
    List.Add( R.P2 ) ;
end ;


function TEvent_List.Count : integer ;

begin
    Result := List.Count div 2 ;
end ;


procedure TEvent_List.Delete( Index : integer ) ;

begin
    List.Delete( Index * 2 + 1 ) ;
    List.Delete( Index * 2 ) ;
end ;


function TEvent_List.IndexOf( E : TCache_Event ) : integer ;

var Loop : integer ;

begin
    move( @E, @R.E, sizeof( TCache_Event ) ) ;
    Result := -1 ;
    for Loop := 0 to Count - 1 do
    begin
        if( ( List[ Loop * 2 ] = R.P1 ) and ( List[ Loop * 2 + 1 ] = R.P2 ) ) then
        begin
            Result := Loop ;
            exit ;
        end ;
    end ;
end ;
{$ENDIF}


type PCache_Record = ^TCache_Record ;
     TCache_Record = record
                         Routine : TCache_Handler ;
                         Priority : longint ;
                         Next : PCache_Record ;
                     end ;

type TStandard_Cache_Manager = object( TSCache_Manager )
                                   private { Instance data... }
                                       PCache_Handlers : PCache_Record ;
{$IFDEF DELEGATION}
                                       Events : TEvent_List ;
                                       Priorities : TList ;
{$ENDIF}

                                   public { API... }
                                       procedure Initialize ; virtual ;
                                       procedure Terminate ; virtual ;
                                       procedure Add_Handler( Routine : TCache_Handler ;
                                           Priority : longint ) ; virtual ;
                                       procedure Delete_Handler( Routine : TCache_Handler ) ;
                                           virtual ;
                                       procedure Free_Cache( Siz, Priority : longint ;
                                           Exclude : TCache_Handler ) ;
                                           virtual ;
{$IFDEF DELEGATION}
                                       procedure Add_Event( Routine : TCache_Event ;
                                           Priority : longint ) ; virtual ;
                                       procedure Delete_Event( Routine : TCache_Event ) ;
                                           virtual ;
                                       procedure Free_Event( Siz, Priority : longint ;
                                           Exclude : TCache_Event ) ;
                                           virtual ;
{$ENDIF}
                               end ;
     PStandard_Cache_Manager = ^TStandard_Cache_Manager ;

{ API... }

procedure TStandard_Cache_Manager.Initialize ;

begin
    PCache_Handlers := nil ;
{$IFDEF DELEGATION}
    Events := TEvent_List.Create ;
    Priorities := TList.Create ;
{$ENDIF}
end ;


procedure TStandard_Cache_Manager.Terminate ;

begin
{$IFDEF DELEGATION}
    Free_Event( -1, MaxInt, nil ) ;
    Events.Free ;
    Events := nil ;
    Priorities.Free ;
    Priorities := nil ;
{$ELSE}
    Free_Cache( -1, MaxInt, nil ) ;
{$ENDIF}

    if( @self = Cache_Manager ) then
    begin
        Cache_Manager := nil ;
    end ;
    Free ;
end ;


procedure TStandard_Cache_Manager.Add_Handler( Routine : TCache_Handler ;
    Priority : longint ) ;
{ Note: The cache list is kept in ascending order by priority to make our job
  easier later. }
var Current, Previous : PCache_Record ;

begin
    if( PCache_Handlers = nil ) then { If first one, then it is simple }
    begin
        New( PCache_Handlers ) ;
        PCache_Handlers^.Next := nil ;
        PCache_Handlers^.Priority := Priority ;
        PCache_Handlers^.Routine := Routine ;
        exit ;
    end ;

    { We must insert into list... }
    Current := PCache_Handlers ;
    Previous := nil ;
    while ( Current <> nil ) do { Until end of list }
    begin
        if( Current^.Priority > Priority ) then { Found insertion point }
        begin
            if( Previous = nil ) then { Insert at beginning of list }
            begin
                New( PCache_Handlers ) ;
                Previous := PCache_Handlers ;
            end else
            begin
                New( Previous^.Next ) ; { Do insertion }
                Previous := Previous^.Next ;
            end ;
            Previous^.Next := Current ;
            Previous^.Priority := Priority ;
            Previous^.Routine := Routine ;
            exit ;
        end ;
        Previous := Current ;
        Current := Current^.Next ; { Go to next list element }
    end ;

    { We hit the end of the list, add to end. }
    New( Current ) ;
    Previous^.Next := Current ;
    Current^.Next := nil ;
    Current^.Priority := Priority ;
    Current^.Routine := Routine ;
end ; { TStandard_Cache_Manager.Add_Handler }


procedure TStandard_Cache_Manager.Delete_Handler( Routine : TCache_Handler ) ;

var Current, Previous : PCache_Record ;

begin
    Current := PCache_Handlers ;
    Previous := nil ;
    while( Current <> nil ) do { Until end of list }
    begin
        if( @Current^.Routine = @Routine ) then { Found it }
        begin
            if( Previous = nil ) then { First one in list }
            begin
                PCache_Handlers := Current^.Next
            end else
            begin
                Previous^.Next := Current^.Next ;
            end ;
            Dispose( Current ) ; { Free up memory }
            exit ;
        end ;
        Previous := Current ;
        Current := Current^.Next ; { Try next entry }
    end ;
end ;


procedure TStandard_Cache_Manager.Free_Cache( Siz, Priority : longint ;
    Exclude : TCache_Handler ) ;

var Current : PCache_Record ;
    Dummy : longint ;

begin
    Current := PCache_Handlers ;
    while( Current <> nil ) do { Go through entire list }
    begin
        if( ( MaxAvail >= Siz ) and ( Siz >= 0 ) ) then
        begin { Enough memory is now available and we are not zeroing }
            Current := nil ; { Terminate loop }
        end else
        begin
            if(
                ( Current^.Priority <= Priority )
                and
                ( @Current^.Routine <> @Exclude )
              ) then
            begin
{$IFDEF FUNCTION_RESULT_REQUIRED}
                Dummy :=
{$ENDIF}
                Current^.Routine( Siz, 0 ) ;
            end ;
            Current := Current^.Next ; { Follow link to next one }
        end ;
    end ; { while( Current <> nil ) }

{$IFDEF DELEGATION}
    for Dummy := 0 to Events.Count - 1 do
    begin
        if( ( MaxAvail >= Siz ) and ( Siz >= 0 ) ) then
        begin { Enough memory is now available and we are not zeroing }
            break ; { Terminate loop }
        end else
        begin
            if( integer( Priorities[ Dummy ] ) <= Priority ) then
            begin
                TCache_Event( Events[ Dummy ] )( Siz, 0 ) ;
            end ;
        end ;
    end ; { for Dummy := 0 to Events.Count - 1 }
{$ENDIF}
end ;


{$IFDEF DELEGATION}
procedure TStandard_Cache_Manager.Add_Event( Routine : TCache_Event ;
    Priority : longint ) ;

begin
    Events.Add( Routine ) ;
    Priorities.Add( pointer( Priority ) ) ;
end ;


procedure TStandard_Cache_Manager.Delete_Event( Routine : TCache_Event ) ;

var Index : integer ;

begin
    Index := Events.IndexOf( Routine ) ;
    if( Index <> -1 ) then
    begin
        Events.Delete( Index ) ;
        Priorities.Delete( Index ) ;
    end ;
end ;


procedure TStandard_Cache_Manager.Free_Event( Siz, Priority : longint ;
    Exclude : TCache_Event ) ;

var Current : PCache_Record ;
    Dummy : longint ;
    E : TCache_Event ;

begin
    Current := PCache_Handlers ;
    while( Current <> nil ) do { Go through entire list }
    begin
        if( ( MaxAvail >= Siz ) and ( Siz >= 0 ) ) then
        begin { Enough memory is now available and we are not zeroing }
            Current := nil ; { Terminate loop }
        end else
        begin
            if( Current^.Priority <= Priority ) then
            begin
                Current^.Routine( Siz, 0 ) ;
            end ;
            Current := Current^.Next ; { Follow link to next one }
        end ;
    end ; { while( Current <> nil ) }

    for Dummy := 0 to Events.Count - 1 do
    begin
        if( ( MaxAvail >= Siz ) and ( Siz >= 0 ) ) then
        begin { Enough memory is now available and we are not zeroing }
            break ; { Terminate loop }
        end else
        begin
            if(
                ( integer( Priorities[ Dummy ] ) <= Priority )
                and
                ( @Events[ Dummy ] <> @Exclude )
              ) then
            begin
                @E := @Events[ Dummy ] ;
                E( Siz, 0 ) ;
            end ;
        end ;
    end ; { for Dummy := 0 to Events.Count - 1 }
end ; { TStandard_Cache_Manager.Free_Event }
{$ENDIF}


{$IFDEF CC_Regression_Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


var Called : boolean = false ;
    Called1 : boolean = false ;
    Called2 : boolean = false ;
    Called3 : boolean = false ;

function Test_Cache_Handler( Siz, Code : longint ) : longint ;

begin
    Called := True ;
    Test_Cache_Handler := 0 ;
end ;


function Test_Cache_Handler1( Siz, Code : longint ) : longint ;

begin
    Called1 := True ;
    Test_Cache_Handler1 := 0 ;
end ;


function Test_Cache_Handler2( Siz, Code : longint ) : longint ;

begin
    Called2 := True ;
    Test_Cache_Handler2 := 0 ;
end ;


function Test_Cache_Handler3( Siz, Code : longint ) : longint ;

begin
    Called3 := True ;
    Test_Cache_Handler3 := 0 ;
end ;


procedure Test_Add_Handler ;

begin
    Cache_Manager^.Add_Handler( @Test_Cache_Handler, 3 ) ;
    if( PStandard_Cache_Manager( Cache_Manager )^.PCache_Handlers = nil ) then
    begin
        Error ;
    end ;
    Cache_Manager^.Add_Handler( @Test_Cache_Handler1, 1 ) ;
    Cache_Manager^.Add_Handler( @Test_Cache_Handler2, 2 ) ;
    Cache_Manager^.Add_Handler( @Test_Cache_Handler3, 4 ) ;
    if( @PStandard_Cache_Manager( Cache_Manager )^.PCache_Handlers^.Routine <> @Test_Cache_Handler1 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Delete_Handler ;

begin
    Cache_Manager^.Delete_Handler( @Test_Cache_Handler1 ) ;
    Cache_Manager^.Delete_Handler( @Test_Cache_Handler ) ;
    Cache_Manager^.Delete_Handler( @Test_Cache_Handler2 ) ;
    Cache_Manager^.Delete_Handler( @Test_Cache_Handler3 ) ;
    if( PStandard_Cache_Manager( Cache_Manager )^.PCache_Handlers <> nil ) then
    begin
        Error ;
    end ;
    Cache_Manager^.Delete_Handler( @Test_Cache_Handler2 ) ;
    Cache_Manager^.Free_Cache( 1, 0, nil ) ;
end ;


procedure Test_Free_Cache ;

begin
    Cache_Manager^.Free_Cache( 1, 0, nil ) ;
    Cache_Manager^.Free_Cache( -1, 10, nil ) ;
    if( not ( Called and Called1 and Called2 and Called3 ) ) then
    begin
        Error ;
    end ;
end ;


  {$IFDEF DELEGATION}
type TTest = class
                 public
                     function E1( Size, Code : longint ) : longint ;
                     function E2( Size, Code : longint ) : longint ;
                     function E3( Size, Code : longint ) : longint ;
                     function E( Size, Code : longint ) : longint ;
             end ;

function TTest.E1( Size, Code : longint ) : longint ;

begin
    Called1 := True ;
    Result := 0 ;
end ;


function TTest.E2( Size, Code : longint ) : longint ;

begin
    Called2 := True ;
    Result := 0 ;
end ;


function TTest.E3( Size, Code : longint ) : longint ;

begin
    Called3 := True ;
    Result := 0 ;
end ;


function TTest.E( Size, Code : longint ) : longint ;

begin
    Called := True ;
    Result := 0 ;
end ;
  {$ENDIF}


procedure Test_Unit ;

  {$IFDEF DELEGATION}
var Test : TTest ;
  {$ENDIF}

begin
    MaxAvail := $7FFFFFFF ;
    Test_Add_Handler ;
    Test_Free_Cache ;
    Test_Delete_Handler ;

    MaxAvail := 0 ;
    Cache_Manager^.Add_Handler( @Test_Cache_Handler, 3 ) ;
    Cache_Manager^.Free_Cache( 1, 0, nil ) ;

  {$IFDEF DELEGATION}
    Test := TTest.Create ;
    MaxAvail := $7FFFFFFF ;
    Called := False ;
    Called1 := False ;
    Called2 := False ;
    Called3 := False ;

    if( PStandard_Cache_Manager( Cache_Manager )^.Events = nil ) then
    begin
        Error ;
    end ;
    Cache_Manager^.Add_Event( Test.E, 3 ) ;
    if( PStandard_Cache_Manager( Cache_Manager )^.Events.Count = 0 ) then
    begin
        Error ;
    end ;
    Cache_Manager^.Add_Event( Test.E1, 1 ) ;
    Cache_Manager^.Add_Event( Test.E2, 2 ) ;
    Cache_Manager^.Add_Event( Test.E3, 4 ) ;
    if( PStandard_Cache_Manager( Cache_Manager )^.Events.Count <> 4 ) then
    begin
        Error ;
    end ;

    Cache_Manager^.Free_Event( 1, 0, nil ) ;
    Cache_Manager^.Free_Event( -1, 10, nil ) ;
    if( not ( Called and Called1 and Called2 and Called3 ) ) then
    begin
        Error ;
    end ;

    Cache_Manager^.Delete_Event( Test.E1 ) ;
    Cache_Manager^.Delete_Event( Test.E ) ;
    Cache_Manager^.Delete_Event( Test.E2 ) ;
    Cache_Manager^.Delete_Event( Test.E3 ) ;
    if( PStandard_Cache_Manager( Cache_Manager )^.Events.Count > 0 ) then
    begin
        Error ;
    end ;
    Cache_Manager^.Delete_Event( Test.E2 ) ;
    Cache_Manager^.Free_Event( 1, 0, nil ) ;


    MaxAvail := 0 ;
    Cache_Manager^.Add_Event( Test.E, 3 ) ;
    Cache_Manager^.Free_Event( 1, 0, nil ) ;


    MaxAvail := $7FFFFFFF ;
    Called := False ;
    Called1 := False ;
    Called2 := False ;
    Called3 := False ;

    Cache_Manager^.Add_Event( Test.E, 3 ) ;
    Cache_Manager^.Add_Event( Test.E1, 1 ) ;
    Cache_Manager^.Add_Event( Test.E2, 2 ) ;
    Cache_Manager^.Add_Event( Test.E3, 4 ) ;
    Cache_Manager^.Free_Cache( 1, 0, nil ) ;
    Cache_Manager^.Free_Cache( -1, 10, nil ) ;
    if( not ( Called and Called1 and Called2 and Called3 ) ) then
    begin
        Error ;
    end ;


    MaxAvail := 0 ;
    Cache_Manager^.Add_Event( Test.E, 3 ) ;
    Cache_Manager^.Free_Cache( 1, 0, nil ) ;

    Test.Free ;
  {$ENDIF}

    Cache_Manager^.Terminate ;
    Cache_Manager := new( PStandard_Cache_Manager, Init ) ;
    Cache_Manager^.Initialize ;
end ;
{$ENDIF}


begin
    Cache_Manager := new( PStandard_Cache_Manager, Init ) ;
    Cache_Manager^.Initialize ;
end.
