{
        Program Name : DISK
        Package Name : Subroutine Library - Layer 1
        Purpose      : Absolute disk read/write for MSDOS
        Institution  : Conroy & Conroy Co.
        Date Written : 01-Sep-87
        Written By   : Alan Conroy
        Version      : 2.1

        Copyright (C) 1987-2015 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

}

unit Disk ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$IFNDEF FPC}
uses _Types ; { SmallInt }
{$ENDIF}

{$I EDEFINES.INC}

procedure Disk_Read( Unit_Number, Sector, Count : SmallInt ; const Buf ;
    var Result, Flags : SmallInt ) ;
procedure Disk_Write( Unit_Number, Sector, Count : SmallInt ; const Buf ;
    var Result, Flags : SmallInt ) ;


implementation

procedure Disk_Read( Unit_Number, Sector, Count : SmallInt ; const Buf ;
    var Result, Flags : SmallInt ) ;
{ This is an absolute disk read procedure.  Data passed: Unit_Number, Sector,
  Count. Buffer to use is Buf.  Returns result in Result. }

var R : integer ;
{$IFDEF ASM}
  {$IFDEF _32BIT}
    F : integer ;
  {$ENDIF}
{$ENDIF}

begin
{$IFDEF ASM}
  {$IFDEF FPC}
    {$ASMMODE intel}
  {$ENDIF}
  {$IFDEF _32BIT}
    F := Flags ;
    asm
	      PUSH DS { Save DS and BP }
        PUSH BP
        MOV  AX, Unit_Number
        LDS  BX, Buf
        MOV  CX, Count
        MOV  DX, Sector
        INT  $25 { Absolute disk read }
        POP  DX  { Old flags }
        POP  BP  { Restore BP }
	      PUSHF    { Save returned flags }
        LES  DI, R
        CLD
        STOSW   { Store result }
        LES  DI, F
	      POP  AX { Get result flags }
        STOSW   { Store result }
        POP  DS { Restore DS }
    end ;
    Flags := F ;
  {$ENDIF}
{$ELSE}
  {$IFDEF ASM_INLINE}
    inline(
           $1E /            { PUSH DS ; Save DS and BP }
           $55 /            { PUSH BP }
           $8B / $86 / Unit_Number / { MOV  AX, [BP].Unit_Number }
           $C5 / $9E / Buf / { LDS  BX, [BP].Buf }
           $8B / $8E / Count / { MOV  CX, [BP].Count }
           $8B / $96 / Sector / { MOV  DX, [BP].Sector }
           $CD / $25 /      { INT  25H ; Absolute disk read }
           $5A /            { POP  DX  ; Old flags }
           $5D /            { POP  BP  ; Restore BP }
	         $9C /	    { PUSHF    ; Save returned flags }
           $C4 / $BE / R / { LES  DI, [BP].Result }
           $FC /            { CLD }
           $AB /            { STOSW   ; Store result }
           $C4 / $BE / Flags /  { LES  DI, [BP].Flags }
	         $58 /	    { POP  AX ; Get result flags }
           $AB /            { STOSW   ; Store result }
           $1F              { POP  DS ; Restore DS }
          )
    {$ENDIF}
{$ENDIF}
    Result := R ;
end ;


procedure Disk_Write( Unit_Number, Sector, Count : SmallInt ; const Buf ;
    var Result, Flags : SmallInt ) ;
{ This is an absolute disk write procedure.  Data passed: Unit, Sector, Count.
  Buffer to use is Buf.  Returns result in Result. }

var R : integer ;
{$IFDEF ASM}
  {$IFDEF _32BIT}
    F : integer ;
  {$ENDIF}
{$ENDIF}

begin
{$IFDEF ASM}
  {$IFDEF _32BIT}
    F := Flags ;
    asm
        PUSH DS { Save DS and BP }
        PUSH BP
        MOV  AX, Unit_Number
        LDS  BX, Buf
        MOV  CX, Count
        MOV  DX, Sector
        INT  26H { Absolute disk write }
        POP  DX  { Old flags }
        POP  BP  { Restore BP }
	      PUSHF    { Save returned flags }
        LES  DI, R
        CLD
        STOSW   { Store result }
        LES  DI, F
	      POP  AX { Get result flags }
        STOSW   { Store result }
        POP  DS { Restore DS }
    end ;
    Flags := F ;
  {$ENDIF}
{$ELSE}
  {$IFDEF ASM_INLINE}
    inline(
           $1E /            { PUSH DS ; Save DS and BP }
           $55 /            { PUSH BP }
           $8B / $86 / Unit_Number / { MOV  AX, [BP].Unit_Number }
           $C5 / $9E / Buf / { LDS  BX, [BP].Buf }
           $8B / $8E / Count / { MOV  CX, [BP].Count }
           $8B / $96 / Sector / { MOV  DX, [BP].Sector }
           $CD / $26 /        { INT  26H ; Absolute disk write }
           $5A /            { POP  DX  ; Old flags }
           $5D /            { POP  BP  ; Restore BP }
	         $9C /	    { PUSHF    ; Save returned flags }
           $C4 / $BE / R / { LES  DI, [BP].Result }
           $FC /            { CLD }
           $Ab /            { STOSW   ; Store result }
           $C4 / $BE / Flags /  { LES  DI, [BP].Flags }
	         $58 /	    { POP  AX ; Get result flags }
           $AB /            { STOSW   ; Store result }
           $1F              { POP  DS ; Restore DS }
          )
  {$ENDIF}
{$ENDIF}
    Result := R ;
end ;


end.
