{$IFNDEF FPC}
  {$N+}
{$ENDIF}
{
              Program Name : FileSys
              Package Name : Subroutine Library - Layer 3
              Purpose      : File System routines
              Institution  : Conroy & Conroy Co.
              Date Written : 30-Nov-88
              Written By   : Alan Conroy
              Version      : 1.0

              TO THE GLORY OF GOD THROUGH JESUS CHRIST

        Copyright (C) 1988-2008 by Conroy & Conroy.  Released to the public domain.

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

            DATE      BY             REASON


        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

          This unit defines the abstract TSFile_System_Server class.
}

unit FileSys ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

{$I EDEFINES.INC}

{$IFDEF FPC_OR_DELPHI}
{$DEFINE WINDOWS}
{$ENDIF}

uses { Subroutine Library - Layer 0... }
{$IFDEF WINDOWS}
     Collect, { TCollection }
  {$IFDEF FPC_OR_DELPHI}
     SysUtils,
  {$ELSE}
     Windos,
     strings,
     Objects,
  {$ENDIF}
  {$IFNDEF MOBILE}
     _DebugIn, { PDebug_Interface }
  {$ENDIF}
     TypeDefs, { TRecord_Size }
     _Types, { Ansistring }
     _UE, { TUnified_Exception }

     { Subroutine Library - Layer 1... }
     _FileSys, { TSFile_System }
     _Stores ; { TStore_Address }


type FileRec = TFileRec ;

{$ELSE}
     { Borland... }
     Dos,
     Windos,
     Objects,
     Strings,

     { Subroutine Library - Layer 0... }
     Collect, { TCollection }
     _UE ; { TUnified_Exception }
{$ENDIF}

const File_SystemErr_Facility = 59 ;
      File_SystemErr_Success = 0 ;
      File_SystemErr_Invalid_Handle = 1 ;
      File_SystemErr_FileNotOpen = 2 ;
      File_SystemErr_Unsupported_Operation = 3 ;
      File_SystemErr_EndOfFile = 4 ;
      File_SystemErr_IllegalOperationType = 5 ;
      File_SystemErr_Internal_Corruption = 6 ;
      File_SystemErr_Data_Format_Error = 7 ;
      File_SystemErr_LineTooLong = 8 ;
      File_SystemErr_Invalid_Record_Size = 9 ;
      File_SystemErr_FileName = 10 ;
      File_SystemErr_InvalidDevice = 11 ;
      File_SystemErr_ProtectionViolation = 12 ;
      File_SystemErr_Device_In_Use = 13 ;
      File_SystemErr_FileNotFound = 14 ;
      File_SystemErr_NotSupported = 15 ;

type TSFile_System_Server = object( TSDefault_File_System_Server )
                               public { API... }
                                   function Get_File_System_For_Store( Store : PSStore ;
                                       Own : boolean ) : PSFile_System ;
                                       virtual ;
                                   { Creates and returns a file system for the
                                     passed store.  If there is no known file
                                     system on the store, nil is returned. }

                                   function Get_File_System( const Nam : Ansistring ) : PSFile_System ;
                                       virtual ;
                                   { Creates and returns a file system for the
                                     passed file specification.  If there is no
                                     known file system on the store, nil is
                                     returned. }

                                   procedure Initialize( const Nam, ID : Ansistring ;
                                       Flags : integer ; var UEC : PSUnified_Exception ) ;
                                       virtual ;
                                   { Initializes a file system on the specified
                                     device, with the specified ID and flags.
                                     Returns status in UEC. }
                           end ; { TSFile_System_Server }
     PSFile_System_Server = ^TSFile_System_Server ;

function Find_File_In_Path( const Nam : Ansistring ; Env : Ansistring ) : Ansistring ;

function Normalize_Hierarchy( Name, Separator, Self_Reference, Parent_Reference,
    Prefix : Ansistring ) : Ansistring ; { UNTESTED }
function Create_File_System_Exception( Code : longint ) : PSUnified_Exception ;
function File_SystemErr_Text( _Code : integer ) : Ansistring ;

var DOS_FS : TSFile_System ;

implementation

uses { Borland... }
{$IFDEF MSWINDOWS}
     WinAPI,
{$ENDIF}

     { Subroutine Library - Layer 0... }
     L10n, { Text_* }

     { Subroutine Library - Layer 1... }
     _Applic, { Get_Serial_Number }
     Compatib, { FindFirst }
     CommonUt, { Clear_IOResult }
     _FIP, { Filespec_String_Scan }
     UE, { Create_PSMSDOS_UE }
{$IFDEF MSWINDOWS}
     Win_UE, // Create_PSWindows_UE
{$ENDIF}
     XLate, { WC_Match }

     { Subroutine Library - Layer 3... }
     GIO, { Log_Error }
     O_S ; { OS }

function File_SystemErr_Text( _Code : integer ) : Ansistring ;

var S : Ansistring ;

begin
    case _Code of
        File_SystemErr_Success: S := 'Success' ;
        File_SystemErr_Invalid_Handle: S := 'Invalid handle' ;
        File_SystemErr_FileNotOpen: S := 'File not open' ;
        File_SystemErr_Unsupported_Operation: S := 'Unsupported operation' ;
        File_SystemErr_EndOfFile: S := 'End of file' ;
        File_SystemErr_IllegalOperationType: S := 'Illegal operation type' ;
        File_SystemErr_Internal_Corruption: S := 'Internal corruption' ;
        File_SystemErr_Data_Format_Error: S := 'Data format error' ;
        File_SystemErr_LineTooLong: S := 'Line too long' ;
        File_SystemErr_Invalid_Record_Size: S := 'Invalid record size' ;
        File_SystemErr_FileName: S := 'Filename' ;
        File_SystemErr_InvalidDevice: S := 'Invalid device' ;
        File_SystemErr_ProtectionViolation: S := 'Protection violation' ;
        File_SystemErr_Device_In_Use: S := 'Device in use' ;
        File_SystemErr_FileNotFound: S := 'File not found' ;
    end ;
    File_SystemErr_Text := S ;
end ;


type TSFile_System_Exception = object( TSUnified_Exception )
                                  private
                                      _Code : longint ;

                                  public
                                      function Get_Facility : longint ; virtual ;
                                      function Get_Facility_Version : longint ;
                                          virtual ;
                                      function Version : longint ; virtual ;
                                      function Severity : longint ; virtual ;
                                      function Error_Text( var Size, Typ : longint ) : Ansistring ;
                                          virtual ;
                                      function Get_Error : longint ; virtual ;
                                      function Get_Previous : PsUnified_Exception ;
                                          virtual ;
                                      procedure Terminate ; virtual ;
                              end ;
     PSFile_System_Exception = ^TSFile_System_Exception ;

function TSFile_System_Exception.Get_Facility : longint ;

begin
    Get_Facility := File_SystemErr_Facility ;
end ;


function TSFile_System_Exception.Get_Facility_Version : longint ;

begin
    Get_Facility_Version := 10 ;
end ;


function TSFile_System_Exception.Version : longint ;

begin
    Version := Subroutine_Library_Version ;
end ;


function TSFile_System_Exception.Severity : longint ;

begin
    case _Code of
        File_SystemErr_Success: Severity := UE_Success ;
        else Severity := UE_Error ;
    end ;
end ;


function TSFile_System_Exception.Error_Text( var Size, Typ : longint ) : Ansistring ;

var S : Ansistring ;

begin
    S := File_SystemErr_Text( _Code ) ;
    Size := length( S ) ;
    Typ := 0 ;
    Error_Text := S ;
end ;


function TSFile_System_Exception.Get_Error : longint ;

begin
    Get_Error := _Code ;
end ;


function TSFile_System_Exception.Get_Previous : PsUnified_Exception ;

begin
    Get_Previous := nil ;
end ;


procedure TSFile_System_Exception.Terminate ;

begin
    Free ;
end ;



function Create_File_System_Exception( Code : longint ) : PSUnified_Exception ;

var E : PSFile_System_Exception ;

begin
    E := new( PSFile_System_Exception, Init ) ;
    PSFile_System_Exception( E )^._Code := Code ;
    Create_File_System_Exception := E ;
    E.Attach ;
end ;



function Convert_File_Mode( File_Mode : integer ) : integer ;

var X : integer ;

begin
    X := 0 ;
    if( ( File_Mode and FSOM_Read_Only ) = 0 ) then
    begin
        X := 2 ;
    end ;
(*
    case ( File_Mode and ( FSOM_Share_Write or FSOM_Share_Read ) ) of
        0 : X := X or $10 ;
        FSOM_Share_Write: X := X or $30 ;
        FSOM_Share_Read: X := X or $20 ;
        else X := X or $40 ;
    end ;
*)
    Convert_File_Mode := X ;
end ;


type TFile_Record = record
                        _File : file ;
                    end ;
     PFile_Record = ^TFile_Record ;

     TFile_System = object( TSFile_System )
                        private { Instance data... }
{$IFDEF CC_Debug}
                            TFile_System_Sanity : string[ 15 ] ;
{$ENDIF}
                            Serial_Number : longint ;

{$IFDEF FPC_OR_DELPHI}
                        protected { Instance data... }
{$ELSE}
                        public
{$ENDIF}
                            File_List : TCollection ;

                        public { Constructors and destructors... }
                            constructor Init ;
                            destructor Done ; virtual ;

                        private { Internal utility routines... }
                            function Get_File_Rec( Handle : integer ) : PFile_Record ;

{$IFDEF FPC_OR_DELPHI}
                        protected { Instance support... }
{$ELSE}
                        public { These routines should NOT be called from outside this unit! }
{$ENDIF}
                            procedure Filespec_String_Scan( const Name, _Separator : Ansistring ;
                                var Info : TFSSS ) ; virtual ;
                            procedure Set_Error( X : PSUnified_Exception ) ;
                                virtual ;

                        public { API... }
                            { Increment reference count to file system. }
                            procedure Attach ; virtual ;

                            { Decrement reference count to file system. }
                            procedure Detach ; virtual ;

                            { Close an open file via handle. }
                            procedure Close( Handle : integer ) ; virtual ;

                            { Close an open find record. }
                            procedure Close_Find( var Find_Record : TFind_Record ) ;
                                virtual ;

                            { Create file and open with specified mode.  If
                              file already exists, it is deleted and re-created.
                              Returns the file handle (0 if an error). }
                            function Create( const Name : Ansistring ;
                                Mode : integer ) : integer ; virtual ;

{$IFNDEF MOBILE}
                            { Returns a debugger interface to the file system }
                            function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}

                            { Deletes the specified file. }
                            procedure Erase( const Name : Ansistring ) ; virtual ;

                            { Returns name associated with specified file. }
                            function File_Name( Handle : integer ) : Ansistring ;
                                 virtual ;

                            function First_Directory( const Name : Ansistring ) : integer ;
                                virtual ;

                            { Flush any buffered output for specified file. }
                            procedure Flush( Handle : integer ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_Attributes( Handle : integer ) : integer ;
                                 virtual ;

                            { Returns size of specified file, in bytes. }
                            function Get_Size( Handle : integer ) : TStore_Address ;
                                 virtual ;

                            function Get_Time( Handle, Typ : integer ) : Longint ;
                                 virtual ;

                            { Normalize the passed path: (device:)\path }
                            function Normalize_Path( Name : Ansistring ) : Ansistring ;
                                virtual ;

                            { Open existing file with specified mode.  Returns
                              the file handle (0 if an error). }
                            function Open( const Name : Ansistring ;
                                Mode : integer ) : integer ; virtual ;

                            { Read from specified file at specified position
                              into Buffer, up to Len bytes.  Actual bytes read
                              is returned in Count.  Count is ignored on call. }
                            procedure Read( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ; virtual ;

                            { Renames Old file to New file name. }
                            procedure Rename( const Old, New : Ansistring ) ;
                                 virtual ;

                            procedure Set_Attributes( Handle : integer ;
                                Attr : longint ) ; virtual ;

                            procedure Set_Time( Handle : integer ;
                                Typ, Time : Longint ) ; virtual ;

                            { Sets the size of the specified file to the
                              specified size, in bytes. }
                            procedure Set_Size( Handle : integer ;
                                New_Size : TStore_Address ) ; virtual ;

                            { Write to specified file at specified position from
                              Buffer, up to Len bytes.  Actual bytes written is
                              returned in Count.  Count is ignored on call. }
                            procedure Write( Handle : integer ;
                                Position : TStore_Address ; Len : TStore_Address ;
                                var Buffer ; var Count : TStore_Address ) ;
                                virtual ;

                            { Returns last error code. }
                            function Last_Error : PSUnified_Exception ;
                                virtual ;

                            function DeviceSize( Nam : Ansistring ) : TStore_Address ;
                                virtual ;

                            function DeviceFree( Nam : Ansistring ) : TStore_Address ;
                                virtual ;

                            function MkDir( Nam : Ansistring ) : longint ;
                                virtual ;

                            procedure Get_Header( const _Path : Ansistring ;
                                var Header : FS_Header_Information ;
                                 var Res : PSUnified_Exception ) ; virtual ;

                            procedure Set_Header( const _Path : Ansistring ;
                                var Header : FS_Header_Information ;
                                var Res : PSUnified_Exception ) ; virtual ;

                            { Returns the attributes of the specified file. }
                            function Get_File_Attributes( const Name : Ansistring ) : longint ;
                                 virtual ;

                            { Returns the attributes of the specified file. }
                            procedure Set_File_Attributes( const Name : Ansistring ;
                                Value : longint ) ; virtual ;

                            function FindFirst( Nam : Ansistring ; Attributes : longint ;
                                var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            function FindNext( var Find_Record : TFind_Record ) : longint ;
                                virtual ;

                            procedure Initialize( Nam, ID : Ansistring ;
                                Flags : longint ; var UEC : PSUnified_Exception ) ; virtual ;

                            procedure Initialize_Store( S : PSStore ; ID : Ansistring ;
                                Flags : longint ; var UEC : PSUnified_Exception ) ; virtual ;

                            procedure Verify( Name : Ansistring ; var Res : PSUnified_Exception ) ;
                                virtual ;

                            function Path : Ansistring ; virtual ;

                            { What character or string is used to separate
                              paths from each other and the filename.  If the
                              file system doesn't support paths, this returns a
                              null string. }
                            function Separator : Ansistring ; virtual ;

                            function Facility : longint ; virtual ;

                            function Version : longint ; virtual ;

                            { Locks the file system for synchronous I/O.
                              Returns True if this process was the previously
                              locking process. }
                            function Lock( Nam : Ansistring ) : boolean ; virtual ;

                            { Unlocks the file system. }
                            procedure Unlock( Nam : Ansistring ) ; virtual ;
                     end ; { TFile_System }
     PFile_System = ^TFile_System ;

     PWildcard = ^TWildcard ;
     TWildcard = object( TSWildcard )
                     private { Instance data... }
{$IFDEF CC_Debug}
                         TWildcard_Sanity : string[ 10 ] ;
{$ENDIF}
                         Child : PWildcard ; { Child who is looking through sub-directories }
                         Parent : PWildcard ; { Parent wildcard object }
                         Current_Name : Ansistring ; { Name of current file }
                         Current_Spec : Ansistring ; { Current lookup specification }
                         Remaining_Spec : Ansistring ; { Part of lookup specification that applies to directory levels below ours }
                         Remaining_Path : Ansistring ; { Remaining_Spec minus "\...\" }
                         Doing_Directories : boolean ; { True while searching subdirectories for \...\ }
                         Find_Record : TFind_Record ; { Our current context }
                         Serial_Number : integer ;

                     public { API... }
                         File_System : PFile_System ; { File system we are doing lookup in }
                         Case_Sensitive : boolean ; { True if filenames are case-sensitive }

                     public { Constructors and destructors... }
                         constructor Init( FS : PFile_System ;
                             _Parent : PWildcard ) ;
                         constructor Create( FS : PFile_System ) ;
                         destructor Done ; virtual ;

                     public { API... }
                         function Lookup( const Name, Path : Ansistring ;
                             var Context : TFind_Record ) : Ansistring ; virtual ;
                 end ; { TWildcard }


{$IFDEF FPC_OR_DELPHI}
type TWin32_File_System = object( TFile_System )
                              private { Instance data... }
                                  Handles : TInteger_Collection ;
                                  Names : TString_Collection ;

                              protected { Instance support... }
                                  procedure _Set_Last_Error ;

                              public { Constructors and destructors... }
                                  constructor Init ;
                                  destructor Done ; virtual ;

                              public { API... }
                                  { Close an open file via handle. }
                                  procedure Close( Handle : integer ) ; virtual ;

                                  { Create file and open with specified mode.  If
                                    file already exists, it is deleted and re-created.
                                    Returns the file handle (0 if an error). }
                                  function Create( const Name : Ansistring ;
                                      Mode : integer ) : integer ; virtual ;

{$IFNDEF MOBILE}
                                  { Returns a debugger interface to the file system }
                                  function Debugger : PSDebug_Interface ; virtual ;
{$ENDIF}

                                  { Returns name associated with specified file. }
                                  function File_Name( Handle : integer ) : Ansistring ;
                                       virtual ;

                                  { Flush any buffered output for specified file. }
                                  procedure Flush( Handle : integer ) ; virtual ;

                                  { Returns size of specified file, in bytes. }
                                  function Get_Size( Handle : integer ) : TStore_Address ;
                                       virtual ;

                                  function Get_Time( Handle, Typ : integer ) : Longint ;
                                       virtual ;

                                  { Open existing file with specified mode.  Returns
                                    the file handle (0 if an error). }
                                  function Open( const Name : Ansistring ;
                                      Mode : integer ) : integer ; virtual ;

                                  { Read from specified file at specified position
                                    into Buffer, up to Len bytes.  Actual bytes read
                                    is returned in Count.  Count is ignored on call. }
                                  procedure Read( Handle : integer ;
                                      Position : TStore_Address ; Len : TStore_Address ;
                                      var Buffer ; var Count : TStore_Address ) ; virtual ;

                                  procedure Set_Time( Handle : integer ;
                                      Typ, Time : Longint ) ; virtual ;

                                  { Sets the size of the specified file to the
                                    specified size, in bytes. }
                                  procedure Set_Size( Handle : integer ;
                                      New_Size : TStore_Address ) ; virtual ;

                                  { Write to specified file at specified position from
                                    Buffer, up to Len bytes.  Actual bytes written is
                                    returned in Count.  Count is ignored on call. }
                                  procedure Write( Handle : integer ;
                                      Position : TStore_Address ; Len : TStore_Address ;
                                      var Buffer ; var Count : TStore_Address ) ;
                                      virtual ;

                                  { Returns the attributes of the specified file. }
                                  function Get_File_Attributes( const Name : Ansistring ) : longint ;
                                       virtual ;

                                  { Returns the attributes of the specified file. }
                                  procedure Set_File_Attributes( const Name : Ansistring ;
                                      Value : longint ) ; virtual ;

                                  function Facility : longint ; virtual ;

                                  function Version : longint ; virtual ;
                          end ; { TWin32_File_System }
{$ENDIF}

type PFS_Record = ^TFS_Record ;
     TFS_Record = record
                        Store_Name : Ansistring ; { Name of store }
                        FS : PFile_System ; { File system object controlling this file system file }
                        Users : longint ; { Reference count }
                        Next : PFS_Record ; { Next item in list }
                  end ;

{ This is global because when defined locally it causes stack/memory corruption.
  Apparently the compiler inserts some code to close the context when the
  variable goes out of scope.  By making it global, this close operation doesn't
  happen.  Certainly, the problem goes away when this is gloabl instead of
  local. }
var FR : TSearchRec ;

{ Returns True if Name matches Wildcard }
function WC_Path_Match( Wildcard, Name : Ansistring ; Case_Sensitive : boolean ) : boolean ;

var Count, Dummy : integer ;
    Wildcard_Position, Name_Position : integer ;

begin
    if( not Case_Sensitive ) then
    begin
        Wildcard := uppercase( Wildcard ) ;
        Name := uppercase( Name ) ;
    end ;

    { Setup... }
    if( copy( Wildcard, 1, 1 ) <> '\' ) then
    begin
        Wildcard := '\' + Wildcard ;
    end ;
    if( copy( Name, 1, 1 ) <> '\' ) then
    begin
        Name := '\' + Name ;
    end ;
    Dummy := pos( '...', Wildcard ) ;
    if( copy( Wildcard, 1, Dummy - 1 ) <> copy( Name, 1, Dummy - 1 ) ) then
    begin
        WC_Path_Match := False ;
        exit ;
    end ;
    if( Dummy > 0 ) then { \...\ specified }
    begin
        Wildcard := copy( Wildcard, Dummy + 3, length( Wildcard ) ) ;
        Count := 0 ;
        for Dummy := 1 to length( Wildcard ) do
        begin
            if( Wildcard[ Dummy ] = '\' ) then
            begin
                inc( Count ) ;
            end ;
        end ;
        Dummy := length( Name ) ;
        while( ( Dummy > 0 ) and ( ( Count > 1 ) or ( pos( Name[ Dummy ], ':\' ) = 0 ) ) ) do
        begin
            if( Name[ Dummy ] = '\' ) then
            begin
                dec( Count ) ;
            end ;
            dec( Dummy ) ;
        end ;
        Name := copy( Name, Dummy, length( Name ) ) ;
    end ;

    while( length( Name ) > 0 ) do
    begin
        Wildcard_Position := length( Wildcard ) ;
        while(
               ( Wildcard_Position > 0 )
               and
               ( Wildcard[ Wildcard_Position ] <> '\' )
             ) do
        begin
            dec( Wildcard_Position ) ;
        end ;
        Name_Position := length( Name ) ;
        while(
               ( Name_Position > 0 )
               and
               ( Name[ Name_Position ] <> '\' )
             ) do
        begin
            dec( Name_Position ) ;
        end ;

        if( not WC_Match( copy( Wildcard, Wildcard_Position, length( Wildcard ) ),
            copy( Name, Name_Position, length( Name ) ) ) ) then
        begin
            WC_Path_Match := False ;
            exit ;
        end ;
        Wildcard := copy( Wildcard, 1, Wildcard_Position - 1 ) ;
        Name := copy( Name, 1, Name_Position - 1 ) ;
    end ; { while( length( Name ) > 0 ) }
    WC_Path_Match := ( length( Wildcard ) = 0 ) ;
end ; { WC_Path_Match }



{ TWildcard methods... }

{ Constructors and destructors... }

constructor TWildcard.Init( FS : PFile_System ; _Parent : PWildcard ) ;

begin
    inherited Init ;

    Serial_Number := Get_Serial_Number ;
    Child := nil ;
    File_System := FS ;
    Doing_Directories := False ;
{$IFDEF FPC_OR_DELPHI}
    Initialize( Current_Name ) ;
    Initialize( Current_Spec ) ;
    Initialize( Remaining_Spec ) ;
    Initialize( Remaining_Path ) ;
{$ELSE}
    Current_Name := '' ;
    Current_Spec := '' ;
    Remaining_Spec := '' ;
    Remaining_Path := '' ;
{$ENDIF}
    Case_Sensitive := False ; { Default to non-case-sensitive }
    Parent := _Parent ;
    if( FS <> nil ) then
    begin
        FS^.Attach ;
    end ;
    Initialize_Find_Record( Find_Record ) ;
{$IFDEF CC_Debug}
    TWildcard_Sanity := 'TWildcard' ;
{$ENDIF}
end ;


constructor TWildcard.Create( FS : PFile_System ) ;

begin
    inherited Init ;

    Init( FS, nil ) ;
end ;


destructor TWildcard.Done ;

begin
    if( Child <> nil ) then
    begin
        dispose( Child, Done ) ;
        Child := nil ;
    end ;

    if( File_System <> nil ) then
    begin
        File_System^.Detach ;
    end ;
    Uninitialize_Find_Record( Find_Record ) ;

{$IFDEF CC_Debug}
    TWildcard_Sanity := ' ' ;
{$ENDIF}

    inherited Done ;
end ;


{ API... }

function TWildcard.Lookup( const Name, Path : Ansistring ;
    var Context : TFind_Record ) : Ansistring ;

    function _Lookup( Name : Ansistring ) : Ansistring ;

    var A : integer ;
        Dummy{, Dummy1} : integer ;
        _Path, Work : Ansistring ;
        FSSS : TFSSS ;

    label Loop ;

    begin
        if( length( Name ) > 0 ) then { Starting new lookup }
        begin
            if( Child <> nil ) then
            begin
                dispose( Child, Done ) ;
                Child := nil ;
            end ;

            Name := File_System^.Normalize_Path( Name ) ;
            if( copy( Name, 1, 1 ) = '\' ) then
            begin
                Name := copy( Name, 2, length( Name ) ) ;
            end ;

            Dummy := File_System^.First_Directory( Name ) ;
            Current_Spec := copy( Name, 1, Dummy - 1 ) ;
            Remaining_Spec := copy( Name, Dummy + 1, length( Name ) ) ;
            Current_Name := '' ;
            Doing_Directories :=
                ( Current_Spec = '...' ) or ( length( Remaining_Spec ) > 0 ) ;
            if( Current_Spec = '...' ) then
            begin
                Remaining_Spec := '\...\' + Remaining_Spec ; { Make sure child does ellipse search }
            end ;
            if( copy( Remaining_Spec, 1, 5 ) = '\...\' ) then
            begin
                Remaining_Path := copy( Remaining_Spec, 5, length( Remaining_Spec ) ) ;
            end else
            begin
                Remaining_Path := Remaining_Spec ;
            end ;
            Dummy := length( Remaining_Path ) ;
            while( ( Dummy > 0 ) and ( Remaining_Path[ Dummy ] <> '\' ) ) do
            begin
                dec( Dummy ) ;
            end ;
            Remaining_Path := copy( Remaining_Path, 1, Dummy ) ;
        end ; { if( length( Name ) > 0 ) }

        if( Child <> nil ) then
        begin
            Work := Child^.Lookup( '', Path + '\' + Current_Name, Context ) ; { Have child lookup next match }
            if( length( Work ) > 0 ) then
            begin
                Context.File_Name := Current_Name + '\' + Work ;
                Dummy := pos( '\', Current_Spec ) ;
                if( Dummy > 0 ) then
                begin
                    Context.File_Name :=
                        copy( Current_Spec, 1, Dummy ) + Context.File_Name ;
                end ;
                _Lookup := Context.File_Name ;
                exit ;
            end ;

            { Child found nothing - done with it... }
            dispose( Child, Done ) ;
            Child := nil ;
        end ;
        { (Child will always be nil at this point) }

        { Find next match at our level... }
Loop:
        if(
            ( pos( '*', Current_Spec ) + pos( '?', Current_Spec ) + pos( '...', Current_Spec ) > 0 )
            or
            ( length( Current_Name ) = 0 )
          ) then
        begin
            if( Doing_Directories ) then
            begin
                A := FSFF_Directory ; { Look only in directories }
            end else
            begin
(*                A := not FSFF_Directory ; { Look for everything but directories } *)
                A := -1 ; { Look for everything }
            end ;
            if( length( Current_Name ) = 0 ) then { Our first lookup }
            begin
                if( Current_Spec = '...' ) then
                begin
                    Work := '*' ;
                end else
                begin
                    Work := Current_Spec ;
                end ;
                Filespec_String_Scan( Work, '\', FSSS ) ;
                _Path := FSSS.Device ;
                if( length( Path ) > 0 ) then
                begin
                    Dummy := File_System^.FindFirst( Path + '\' + Work, A, Find_Record ) ;
                end else
                begin
                    Dummy := File_System^.FindFirst( Work, A, Find_Record ) ;
                end ;
                if( Dummy <> 0 ) then
                begin
                    _Lookup := '' ;
                    exit ;
                end ;
            end else
            begin
                if( File_System^.FindNext( Find_Record ) <> 0 ) then { No more found }
                begin
                    _Lookup := '' ;
                    exit ;
                end ;
            end ;
            Current_Name := Find_Record.Found_Name ;
        end else
        begin
            _Lookup := '' ;
            exit ;
        end ;
        if( ( Find_Record.Found_Attributes and FSFF_Directory ) = 0 ) then { Not a directory }
        begin
            if( Doing_Directories ) then
            begin
                goto Loop ;
            end ;
        end ;

        { We found a match for our current spec.  See if sub-directories to search }
        if(
            Doing_Directories
            and
            ( length( Remaining_Spec ) > 0 )
            and
            ( Current_Name <> '.' )
            and
            ( Current_Name <> '..' ) { No iterating on relative directories (would cause an infinite loop) }
          ) then
        begin
            Child := new( PWildcard, Init( File_System, @self ) ) ;
            Work := Child^.Lookup( Remaining_Spec, Path + _Path + '\' + Current_Name, Context ) ;
            if( length( Work ) > 0 ) then
            begin
                Work := Current_Name + '\' + Work ;
                Context.File_Name := Work ;
                Dummy := pos( ':', Current_Spec ) ;
                if(
                    ( Dummy > 0 )
                    and
                    ( copy( Current_Spec, Dummy + 1, 1 ) = '\' )
                  ) then
                begin
                    Dummy := Dummy + 1 ;
                end ;
                _Lookup := copy( Current_Spec, 1, Dummy ) + Context.File_Name ;
                exit ;
            end ;
            dispose( Child, Done ) ;
            Child := nil ;
            goto Loop ;
        end ; { if }
        _Lookup := Current_Name ;
{$IFDEF CC_Debug}
        if( Context.Sanity <> 'TFind_Record' ) then
        begin
            halt ;
        end ;
{$ENDIF}
{        Context.Data := Find_Record.Data ;}
        Context.Found_Attributes := Find_Record.Found_Attributes ;
        Context.Wildcard_Attributes := Find_Record.Wildcard_Attributes ;
        Context.Found_Name := Find_Record.Found_Name ;
        Context.File_Name := Current_Name ;
        Context.Last_Name := Current_Name ;
        Context.Wildcard_Context := Find_Record.Wildcard_Context ;
    end ; { TWildcard.Lookup._Lookup }

var Res, Work : Ansistring ;

label Try_Again ;

begin
{$IFDEF CC_Debug}
    if( Context.Sanity <> 'TFind_Record' ) then
    begin
        halt ;
    end ;
{$ENDIF}
    Res := _Lookup( Name ) ;
Try_Again:
    if( ( length( Res ) = 0 ) and Doing_Directories ) then
    begin
        if( ( Current_Spec = '...' ) or ( length( Remaining_Spec ) = 0 ) ) then
        begin
            Doing_Directories := False ;
            Current_Name := '' ; { Start search over for non-directories }
            Res := _Lookup( '' ) ;
        end ;
    end ;
    if(
        ( Parent = nil ) { We are the top level }
        and
        ( length( Res ) > 0 ) { ...and something was found }
        and
        ( pos( '...', Current_Spec + Remaining_Spec ) > 0 ) { Using ... construct }
      ) then
    begin
        if( Current_Spec = '...' ) then
        begin
            Work := Remaining_Spec ;
        end else
        begin
            Work := Current_Spec + '\' + Remaining_Spec ;
        end ;
        if( not WC_Path_Match( Work, Res, Case_Sensitive ) ) then
        begin
            Res := _Lookup( '' ) ;
            goto Try_Again ;
        end ;
    end ;
    Lookup := Res ;
end ; { TWildcard.Lookup }


{$IFDEF FPC_OR_DELPHI}
function Convert_DOS_Attributes( Attr : longint ) : longint ;

var A : longint ;

begin
    A := 0 ;
  {$IFDEF FPC_OR_DELPHI}
    {$WARNINGS OFF}
  {$ENDIF}
    if( ( Attr and faReadOnly ) <> 0 ) then
    begin
        A := A or FSFF_Protected ;
    end ;
    if( ( Attr and faHidden ) <> 0 ) then
    begin
        A := A or FSFF_Hidden ;
    end ;
    if( ( Attr and faSysFile ) <> 0 ) then
    begin
        A := A or FSFF_System ;
    end ;
    if( ( Attr and faVolumeID ) <> 0 ) then
    begin
        A := A or FSFF_VolumeID ;
    end ;
    if( ( Attr and faDirectory ) <> 0 ) then
    begin
        A := A or FSFF_Directory ;
    end ;
    if( ( Attr and faArchive ) <> 0 ) then
    begin
        A := A or FSFF_Archive ;
    end ;
  {$IFDEF FPC_OR_DELPHI}
    {$WARNINGS ON}
  {$ENDIF}
    Convert_DOS_Attributes := A ;
end ; { Convert_DOS_Attributes }


function Convert_Attributes( Value : longint ) : longint ;

var Attr : integer ;

begin
    Attr := 0 ;
  {$IFDEF FPC_OR_DELPHI}
    {$WARNINGS OFF}
  {$ENDIF}
    if( ( Value and FSFF_Protected ) <> 0 ) then
    begin
        Attr := Attr or faReadOnly ;
    end ;
    if( ( Value and FSFF_Hidden ) <> 0 ) then
    begin
        Attr := Attr or faHidden ;
    end ;
    if( ( Value and FSFF_System ) <> 0 ) then
    begin
        Attr := Attr or faSysFile ;
    end ;
    if( ( Value and FSFF_VolumeID ) <> 0 ) then
    begin
        Attr := Attr or faVolumeID ;
    end ;
    if( ( Value and FSFF_Directory ) <> 0 ) then
    begin
        Attr := Attr or faDirectory ;
    end ;
    if( ( Value and FSFF_Archive ) <> 0 ) then
    begin
        Attr := Attr or faArchive ;
    end ;
  {$IFDEF FPC_OR_DELPHI}
    {$WARNINGS ON}
  {$ENDIF}
    Convert_Attributes := Attr ;
end ; { Convert_Attributes }
{$ENDIF}


{ TFile_System methods... }

{ Constructors and destructors... }

constructor TFile_System.Init ;

begin
    inherited Init ;
    
    File_List.Init( 0, 16 ) ;
    Serial_Number := Get_Serial_Number ;
{$IFDEF CC_Debug}
    TFile_System_Sanity := 'TFile_System' ;
{$ENDIF}
end ;


destructor TFile_System.Done ;

var Loop : integer ;

begin
    { Close all files... }
    for Loop := File_List.Count - 1 downto 0 do
    begin
        Close( Loop + 1 ) ;
    end ;
    File_List.Done ;
{$IFDEF CC_Debug}
    TFile_System_Sanity := ' ' ;
{$ENDIF}

    inherited Done ;
end ;


{ Internal utility routines... }

procedure TFile_System.Filespec_String_Scan( const Name, _Separator : Ansistring ;
    var Info : TFSSS ) ;

begin
    { This routine left intentionally blank.
      We do no additional processing other than what the file processor does. }
end ;


function TFile_System.Get_File_Rec( Handle : integer ) : PFile_Record ;

var File_Rec : PFile_Record ;

begin
    { Check for errors... }
    dec( Handle ) ; { Handles are zero-based }
    if( ( Handle < 0 ) or ( Handle >= File_List.Count ) ) then
    begin
        Get_File_Rec := nil ;
        Set_Error( Create_File_System_Exception( File_SystemErr_Invalid_Handle ) ) ;
        exit ;
    end ;
    File_Rec := PFile_Record( File_List.At( Handle ) ) ;
    Get_File_Rec := File_Rec ;
    if( File_Rec = nil ) then
    begin
        Set_Error( Create_File_System_Exception( File_SystemErr_FileNotOpen ) ) ;
        exit ;
    end ;
end ;


procedure TFile_System.Set_Error( X : PSUnified_Exception ) ;

begin
    if( IO_Error <> nil ) then
    begin
        IO_Error.Detach ;
    end ;
    IO_Error := X ;
    if( X <> nil ) then
    begin
        IO_Error.Attach ;
    end ;
end ;


{ API... }

{ Increment reference count to file system. }
procedure TFile_System.Attach ;

begin
    { Because this is an abstract object, we do nothing here. }
end ;


{ Decrement reference count to file system. }
procedure TFile_System.Detach ;

begin
    { Because this is an abstract object, we do nothing here. }
end ;


{ Close an open file via handle. }
procedure TFile_System.Close( Handle : integer ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Close the file... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.Close( File_Rec^._File ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    if( IO_Error = nil ) then
    begin
        dispose( PFile_Record( File_List.At( Handle - 1 ) ) ) ;
        File_List.AtPut( Handle - 1, nil ) ;
    end ;

    { Reclaim unused handles at the end of the file list }
    while(
           ( File_List.Count > 0 )
           and
           ( File_List.At( File_List.Count - 1 ) = nil )
         ) do
    begin
        File_List.AtDelete( File_List.Count - 1 ) ;
    end ;
end ;


{ Close an open find record. }
procedure TFile_System.Close_Find( var Find_Record : TFind_Record ) ;

begin
    if( @self = @DOS_FS ) then
    begin
        if( Find_Record.Data[ 0 ] <> 0 ) then
        begin
            Find_Record.Data[ 0 ] := 0 ;
            move( Find_Record.Data[ 1 ], FR, sizeof( FR ) ) ;
            try
                IOResult ; { Clear any pending errors }
                {$I-}
                SysUtils.FindClose( FR ) ;
                {$I+}
                IOResult ; { Ignore any error on clone }
            except
            end ;
        end ;
    end ;
end ;


{ Create file and open with specified mode.  If file already exists, it is
  deleted and re-created.  Returns the file handle (0 if an error). }
function TFile_System.Create( const Name : Ansistring ; Mode : integer ) : integer ;

var F : file ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Name ) ;
    {$I-}
    System.Rewrite( F ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    if( IO_Error <> nil ) then
    begin
        Create := 0 ; { Creation failure }
    end else
    begin
        System.close( F ) ;
        Create := Open( Name, Mode ) ;
    end ;
end ;


{$IFNDEF MOBILE}
{ Returns a debugger interface to the file system. }
function TFile_System.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;
{$ENDIF}


{ Deletes the specified file. }
procedure TFile_System.Erase( const Name : Ansistring ) ;

var F : File ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Name ) ;
    {$I-}
    System.Erase( F ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
end ;


{ Returns name associated with specified file. }
function TFile_System.File_Name( Handle : integer ) : Ansistring ;

var File_Rec : PFile_Record ;
    Temp : FileRec ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    move( File_Rec^, Temp, sizeof( Temp ) ) ;
    File_Name := strpas( PAnsiChar( @Temp.Name ) ) ;
end ;


function TFile_System.First_Directory( const Name : Ansistring ) : integer ;

var Dummy1 : integer ;

begin
    Dummy1 := pos( ':', Name ) ;
    if( Dummy1 > 0 ) then { Device included }
    begin
        First_Directory := Dummy1 + 1 +
            pos( '\', copy( Name, Dummy1 + 2, length( Name ) ) + '\' ) ;
    end else
    begin
        First_Directory := pos( '\', Name + '\' ) ;
    end ;
end ;


{ Flush any buffered output for specified file. }
procedure TFile_System.Flush( Handle : integer ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;
end ;


{ Returns the attributes of the specified file. }
function TFile_System.Get_Attributes( Handle : integer ) : integer ;

var File_Rec : PFile_Record ;

begin
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Attributes := 0 ;
        exit ;
    end ;

    Get_Attributes := Get_File_Attributes( File_Name( Handle ) ) ;
end ;


{ Returns size of specified file, in bytes. }
function TFile_System.Get_Size( Handle : integer ) : TStore_Address ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Size := -1 ;
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    Get_Size := System.Filesize( File_Rec^._File ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
end ;


function TFile_System.Get_Time( Handle, Typ : integer ) : Longint ;

var File_Rec : PFile_Record ;
    Time : longint ;

begin
    { Setup... }
    Time := 0 ;
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        Get_Time := Time ;
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
    {$I-}
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS OFF}
    Time := FileAge( File_Name( Handle ) ) ;
  {$WARNINGS ON}
{$ELSE}
  {$IFDEF WINDOWS}
    Windos.Getftime( File_Rec^._File, Time ) ;
  {$ELSE}
    Dos.Getftime( File_Rec^._File, Time ) ;
  {$ENDIF}
{$ENDIF}
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    Get_Time := Time ;
end ;


function TFile_System.Normalize_Path( Name : Ansistring ) : Ansistring ;

begin
    Normalize_Path := Normalize_Hierarchy( Name, '\', '.', '..', ':' ) ;
end ;


{ Open existing file with specified mode.  Returns the file handle (0 if an
  error). }
function TFile_System.Open( const Name : Ansistring ; Mode : integer ) : integer ;

var Save : integer ;
    File_Rec : PFile_Record ;

begin
    new( File_Rec ) ;
    Save := FileMode ;
    FileMode := Convert_File_Mode( Mode ) ;
    Clear_IOResult ; { Clear any pending errors }
    assign( File_Rec^._File, Name ) ;
    {$I-}
    System.Reset( File_Rec^._File, 1 ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    FileMode := Save ;
    if( IO_Error <> nil ) then
    begin
        Open := 0 ; { Open failure }
        dispose( File_Rec ) ;
    end else
    begin
        File_List.Insert( File_Rec ) ;
        Open := File_List.Count ;
    end ;
end ;


{ Read from specified file at specified position into Buffer, up to Len bytes.
  Actual bytes read is returned in Count.  Count is ignored on call. }
procedure TFile_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var Dummy : integer ;
    File_Rec : PFile_Record ;
{$IFDEF _64BIT}
    R : integer ;
{$ELSE}
    R : TRecord_Size ;
{$ENDIF}

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Position... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.Seek( File_Rec^._File, Position ) ;
    {$I+}
    Dummy := IOResult ;
    if( Dummy <> 0 ) then
    begin
        Set_Error( Create_PSMSDOS_UE( Dummy, '' ) ) ;
        Count := 0 ;
        exit ;
    end ;

    { Read the data... }
    {$I-}
    System.BlockRead( File_Rec^._File, Buffer, Len, R ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    Count := R ;
end ; { TFile_System.Read }


{ Renames Old file to New file name. }
procedure TFile_System.Rename( const Old, New : Ansistring ) ;

var F : File ;

begin
    Clear_IOResult ; { Clear any pending errors }
    assign( F, Old ) ;
    {$I-}
    System.Rename( F, New ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
end ;


procedure TFile_System.Set_Attributes( Handle : integer ; Attr : longint ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Set_File_Attributes( File_Name( Handle ), Attr ) ;
end ;


procedure TFile_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

var File_Rec : PFile_Record ;

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    Clear_IOResult ; { Clear any pending errors }
{$IFDEF FPC_OR_DELPHI}
(*
    Set_Error( 0, FileSetDate( File_Rec^.TF, Time ) ) ;
*)
{$ELSE}
    {$I-}
  {$IFDEF WINDOWS}
        Windos.Setftime( File_Rec^._File, Time ) ;
  {$ELSE}
        Dos.Setftime( File_Rec^._File, Time ) ;
  {$ENDIF}
        {$I+}
        Set_Error( 2, IOResult ) ;
{$ENDIF}
end ;


{ Sets the size of the specified file to the specified size, in bytes. }
procedure TFile_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

var File_Rec : PFile_Record ;

{$IFDEF FPC_OR_DELPHI}
var Buffer : array[ 0..32767 ] of char ;
    R : TStore_Address ;
    Size : int64 ;
{$ELSE}

label Finish ;

var Reg : TRegisters ;
    _Handle : integer ;
    Name : Ansistring ;
{$ENDIF}

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

{$IFDEF FPC_OR_DELPHI}
    Size := Get_Size( Handle ) ;
    if( New_Size > Size ) then { Expand file size }
    begin
        fillchar( Buffer, sizeof( Buffer ), 0 ) ;
        while( New_Size > Size ) do
        begin
            if( New_Size - Size > 32768 ) then
            begin
                Write( Handle, Size, 32768, Buffer, R ) ;
                if( R <> 32768 ) then
                begin
                    exit ;
                end ;
            end else
            begin
                Write( Handle, Size, New_Size - Size, Buffer, R ) ;
                if( R <> New_Size - Size ) then
                begin
                    exit ;
                end ;
            end ;
            Size := Size + R ;
        end ;
        exit ;
    end ;

    Seek( File_Rec^._File, New_Size ) ;
    Truncate( File_Rec^._File ) ;
{$ELSE}
    Name := File_Name( Handle ) + #0 ;
    Reg.AX := $3D02 ;
    Reg.DS := seg( Name ) ;
    Reg.DX := ofs( Name[ 1 ] ) ;
    intr( $21, Reg ) ; { Open file }
    _Handle := Reg.AX ;
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
        Set_Error( 1, Reg.AX ) ;
        exit ;
    end ;

    Reg.AX := $4200 ;
    Reg.BX := _Handle ;
    Reg.CX := New_Size shr 16 ;
    Reg.DX := New_Size and $FFFF ;
    intr( $21, Reg ) ; { Set position }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
        Set_Error( 1, Reg.AX ) ;
        goto Finish ;
    end ;

    Reg.AX := $4000 ;
    Reg.BX := _Handle ;
    Reg.CX := 0 ;
    Reg.DS := DSeg ;
    Reg.DX := ofs( Reg ) ;
    intr( $21, Reg ) ; { Truncate/expand }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
        Set_Error( 1, Reg.AX ) ;
    end ;

Finish:
    Reg.AX := $3E00 ;
    Reg.BX := _Handle ;
    intr( $21, Reg ) ; { Close file }
    if( ( Reg.Flags and 1 ) = 1 ) then
    begin
        if( IO_Error.Code <> 0 ) then
        begin
            Set_Error( 1, Reg.AX ) ;
        end ;
    end ;
{$ENDIF}
end ; { TFile_System.Set_Size }


{ Write to specified file at specified position from Buffer, up to Len bytes.
  Actual bytes written is returned in Count.  Count is ignored on call. }
procedure TFile_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

var File_Rec : PFile_Record ;
{$IFDEF _64BIT}
    R : integer ;
{$ELSE}
    R : TRecord_Size ;
{$ENDIF}

begin
    { Setup... }
    File_Rec := Get_File_Rec( Handle ) ;
    if( File_Rec = nil ) then
    begin
        exit ;
    end ;

    { Position... }
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.Seek( File_Rec^._File, Position ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    if( IO_Error <> nil ) then
    begin
        Count := 0 ;
        exit ;
    end ;

    { Write the data... }
    {$I-}
    System.BlockWrite( File_Rec^._File, Buffer, Len, R ) ;
    {$I+}
    Set_Error( Create_PSMSDOS_UE( IOResult, '' ) ) ;
    Count := R ;
end ;


{ Returns last error code. }
function TFile_System.Last_Error : PSUnified_Exception ;

begin
    Last_Error := IO_Error ;
end ;


function TFile_System.DeviceSize( Nam : Ansistring ) : TStore_Address ;

begin
{$IFDEF DOS_OR_WINDOWS}
    if( ( length( Nam ) < 2 ) or ( Nam[ 2 ] <> ':' ) ) then
    begin
        DeviceSize := DiskSize( 0 ) ;
        exit ;
    end ;
    DeviceSize := DiskSize( ord( Nam[ 1 ] ) - 64 ) ;
{$ELSE}
    DeviceSize := -1 ;
{$ENDIF}
end ;


function TFile_System.DeviceFree( Nam : Ansistring ) : TStore_Address ;

begin
    if( ( length( Nam ) < 2 ) or ( Nam[ 2 ] <> ':' ) ) then
    begin
        DeviceFree := -1 ;
        exit ;
    end ;
{$IFDEF DOS_OR_WINDOWS}
    DeviceFree := DiskFree( ord( Nam[ 1 ] ) - 64 ) ;
{$ELSE}
    DeviceFree := -1 ;
{$ENDIF}
end ;


function TFile_System.MkDir( Nam : Ansistring ) : longint ;

var Dummy : integer ;

begin
    Clear_IOResult ; { Clear any pending errors }
    {$I-}
    System.MkDir( Nam ) ;
    {$I+}
    Dummy := IOResult ;
    Set_Error( Create_PSMSDOS_UE( Dummy, '' ) ) ;
    MkDir := Dummy ;
end ;


procedure TFile_System.Get_Header( const _Path : Ansistring ;
    var Header : FS_Header_Information ; var Res : PSUnified_Exception ) ;

begin
    Set_Error( nil ) ;
    {TODO}
    Res := IO_Error ;
end ;


procedure TFile_System.Set_Header( const _Path : Ansistring ;
    var Header : FS_Header_Information ; var Res : PSUnified_Exception ) ;

begin
    Set_Error( nil ) ;
    {TODO}
    Res := IO_Error ;
end ;


{ Returns the attributes of the specified file. }
function TFile_System.Get_File_Attributes( const Name : Ansistring ) : longint ;

var Attr : integer ;

begin
    Clear_IOResult ; { Clear any pending errors }
{$IFDEF DOS_OR_WINDOWS}
    Attr := GetFileAttributes( Name ) ;
    Get_File_Attributes := Convert_DOS_Attributes( Attr ) ;
{$ELSE}
    Get_File_Attributes := 0 ;
{$ENDIF}
end ; { TFile_System.Get_File_Attributes }


{ Returns the attributes of the specified file. }
procedure TFile_System.Set_File_Attributes( const Name : Ansistring ;
    Value : longint ) ;

var Attr : integer ;

begin
    Attr := Convert_Attributes( Value ) ;
    Clear_IOResult ; { Clear any pending errors }
{$IFDEF MSWINDOWS}
  {$IFDEF FPC_OR_DELPHI}
    if( SetFileAttributes( Name, Attr ) ) then
    begin
        Set_Error( nil ) ;
    end else
    begin
        Set_Error( Create_PSWindows_UE( GetLastError, '' ) ) ;
    end ;
  {$ELSE}
    assign( F, Name ) ;
    {$I-}
    {$IFDEF WINDOWS}
    Windos.Setfattr( F, Attr ) ;
    {$ELSE}
    Dos.Setfattr( F, Attr ) ;
    {$ENDIF}
    {$I+}
    Set_Error( 2, IOResult ) ;
  {$ENDIF}
{$ELSE}
    //TODO:Do nothing for Linux, for now...
{$ENDIF}
end ;


procedure Invalid_Compile ;

begin
    Log_Error( nil, Text_Error_Filesys_Unit_Was_Not_Compiled_Correctly, -1 ) ;
    halt ;
end ;


procedure Copy_Find_Info( var Find_Record : TFind_Record ;
    const FR : TSearchRec ) ;

var Dummy : integer ;

begin
    if( sizeof( FR ) > sizeof( Find_Record.Data ) - 1 ) then
    begin
        Invalid_Compile ;
    end ;
    move( FR, Find_Record.Data[ 1 ], sizeof( FR ) ) ;
    Find_Record.Last_Name := FR.Name ;
    Dummy := length( FR.Name ) ;
    while( ( Dummy > 0 ) and ( FR.Name[ Dummy ] <> '\' ) ) do
    begin
        dec( Dummy ) ;
    end ;
    Find_Record.Found_Name := copy( FR.Name, Dummy + 1, length( FR.Name ) ) ;
    Find_Record.Found_Attributes := Convert_DOS_Attributes( FR.Attr ) ;
end ; { Copy_Find_Info }


function TFile_System.FindFirst( Nam : Ansistring ; Attributes : longint ;
    var Find_Record : TFind_Record ) : longint ;

var Attr : integer ;
    Dummy, I : integer ;
    F : file ;

begin
    if( Find_Record.Initialized <> 0 ) then
    begin
        Initialize_Find_Record( Find_Record ) ;
    end ;
    Attr := Convert_Attributes( Attributes ) ;
    I := Compatib.FindFirst( Nam, Attr, FR ) ;
    Set_Error( Create_PSMSDOS_UE( I, '' ) ) ;
    Find_Record.Data[ 0 ] := 1 ;
    Copy_Find_Info( Find_Record, FR ) ;
    if(
        ( I = 18 ) { No more files }
        and
        ( pos( '*', Nam ) + pos( '?', Nam ) = 0 ) { No wildcard }
        and
        Compatib.FileExists( Nam )
      ) then
    begin
        Set_Error( nil ) ;
        Dummy := length( Nam ) ;
        while( ( Dummy > 0 ) and ( pos( Nam[ Dummy ], ':\' ) = 0 ) ) do
        begin
            dec( Dummy ) ;
        end ;
        Find_Record.Found_Name := copy( Nam, Dummy + 1, length( Nam ) ) ;
        Find_Record.Found_Attributes := 0 ;
        Clear_IOResult ; { Clear any pending errors }
        assign( F, Nam ) ;
        {$I-}
        reset( F ) ;
        {$I+}
        if( IOResult = 5 ) then { Access violation }
        begin
            Find_Record.Found_Attributes := FSFF_Directory ;
        end else
        begin
            System.close( F ) ;
        end ;
    end ;
    FindFirst := I ;
end ; { TFile_System.FindFirst }


function TFile_System.FindNext( var Find_Record : TFind_Record ) : longint ;

var I : integer ;

begin
    move( Find_Record.Data[ 1 ], FR, sizeof( FR ) ) ;
    I := Compatib.FindNext( FR ) ;
    FindNext := I ;
    Set_Error( Create_PSMSDOS_UE( I, '' ) ) ;
    Copy_Find_Info( Find_Record, FR ) ;
end ;


procedure TFile_System.Initialize( Nam, ID : Ansistring ; Flags : longint ;
    var UEC : PSUnified_Exception ) ;

begin
    UEC := Create_File_System_Exception( File_SystemErr_Unsupported_Operation ) ;
    if( UEC <> nil ) then
    begin
        UEC.Attach ;
    end ;
    Set_Error( UEC ) ;
end ;


procedure TFile_System.Initialize_Store( S : PSStore ; ID : Ansistring ;
    Flags : longint ; var UEC : PSUnified_Exception ) ;

begin
    UEC := Create_File_System_Exception( File_SystemErr_Unsupported_Operation ) ;
    if( UEC <> nil ) then
    begin
        UEC.Attach ;
    end ;
    Set_Error( UEC ) ;
end ;


procedure TFile_System.Verify( Name : Ansistring ; var Res : PSUnified_Exception ) ;

begin
    Res := Create_File_System_Exception( File_SystemErr_Unsupported_Operation ) ;
    if( Res <> nil ) then
    begin
        Res.Attach ;
    end ;
    Set_Error( Res ) ;
end ;


function TFile_System.Separator : Ansistring ;

begin
    Separator := '\' ; { This is the default }
end ;


function TFile_System.Path : Ansistring ;

begin
    Path := '' ;
end ;


function TFile_System.Facility : longint ;

begin
    Facility := 1 ; { DOS }
end ;


function TFile_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;


function TFile_System.Lock( Nam : Ansistring ) : boolean ;

begin
    Lock := False ;
end ;


procedure TFile_System.Unlock( Nam : Ansistring ) ;

begin
    { This routine left intentionally blank }
end ;


var F : Text ;

function Find_Logical( const X : Ansistring ) : Ansistring ;
{ Try to resolve a logical to a valid path or device }

var Env : Ansistring ;

begin
    Env := OS^.Get_Symbol( X ) ; { Get logical definition }
    if( length( Env ) = 0 ) then { Not defined }
    begin
        if( length( X ) = 1 ) then { A valid device }
        begin
            Find_Logical := X ;
        end else
        begin
            Find_Logical := '' ;
        end ;
        exit ;
    end ;

    Env := Find_File_In_Path( 'X.X', Env ) ;
    Env := copy( Env, 1, length( Env ) - 3 ) ; { Subtract off "X.X" }
    Find_Logical := Env ;
end ; { Find_Logical }


const Max_Find_File_Nest = 8 ;

var Nested : integer ;

function Find_File_In_Path( const Nam : Ansistring ; Env : Ansistring ) : Ansistring ;

var Dummy, Index : integer ;
    This_Path, XX : Ansistring ;

begin
    if( Nested > Max_Find_File_Nest ) then
    begin
        Find_File_In_Path := '' ; { Nested too deep }
        exit ;
    end ;

    while( length( Env ) > 0 ) do { Process all items in list }
    begin
        { Extract next path in list...}
        Index := pos( ';', Env ) ; { End of next item }
        if( Index = 0 ) then
        begin
            Index := length( Env ) + 1 ;
        end ;
        This_Path := copy( Env, 1, Index - 1 ) ;
        Env := copy( Env, Index + 1, 254 ) ;

        { See if we can find a file... }
        if( length( This_Path ) > 0 ) then { Non-null path }
        begin
            if( This_Path[ 1 ] = '_' ) then { Use literal device }
            begin
                Clear_IOResult ; { Clear any pending errors }
                assign( F, copy( This_Path, 2, length( This_Path ) ) ) ;
                {$I-}
                reset( F ) ;
                {$I+}
                Dummy := IOResult ;
                if( Dummy = 0 ) then
                begin
                    {$I-}
                    close( F ) ;
                    {$I+}
                    Clear_IOResult ;
                    XX := copy( This_Path, 2, length( This_Path ) ) ;
                    if( XX[ length( XX ) ] = '\' ) then
                    begin
                        Find_File_In_Path := XX + Nam ;
                    end else
                    begin
                        Find_File_In_Path := XX + '\' + Nam ;
                    end ;
                    exit ;
                end ; {if( IOResult = 0 ) then}
            end ; {if( B[ 1 ] = '_' ) then}
            inc( Nested ) ;
            XX := Find_Logical( This_Path ) ; { Recursively translate the logical }
            dec( Nested ) ;
            if( length( XX ) = 0 ) then
            begin
                XX := This_Path ;
            end ;
            if( length( XX ) > 0 ) then
            begin
                Clear_IOResult ; { Clear any pending errors }
                if( XX[ length( XX ) ] = '\' ) then
                begin
                    assign( F, XX + Nam ) ;
                end else
                begin
                    assign( F, XX + '\' + Nam ) ;
                end ;
                {$I-}
                reset( F ) ;
                {$I+}
                Dummy := IOResult ;
                if( Dummy = 0 ) then
                begin
                    {$I-}
                    close( F ) ;
                    {$I+}
                    Clear_IOResult ;
                end else
                if( Dummy = 3 ) then { Path not found or invalid }
                begin
                    XX := '' ;
                end ;
            end ; { if( length( XX ) > 0 ) }
            if( length( XX ) <> 0 ) then { Found a valid path }
            begin
                if( XX[ length( XX ) ] = '\' ) then
                begin
                    Find_File_In_Path := XX + Nam ;
                end else
                begin
                    Find_File_In_Path := XX + '\' + Nam ;
                end ;
                exit ;
            end ;
        end ;
    end ;
    Find_File_In_Path := '' ; { Nothing valid was found }
end ; { Find_File_In_Path }


function Find_File( X : Ansistring ; Literal : boolean ) : Ansistring ;

    function Combine( const Path, Work : Ansistring ) : Ansistring ;

    begin
        if( ( copy( Path, length( Path ), 1 ) = '\' ) and ( copy( Work, 1, 1 ) = '\' ) ) then
        begin
            Combine := Path + copy( Work, 2, length( Work ) ) ;
        end else
        begin
            Combine := Path + Work ;
        end ;
    end ;

label Skip ;

var A, Dummy : Integer ;
    B, Env, Path, Work : Ansistring ;
    F : Text ;
    Saved : integer ;

Begin
    { First, see if we are nested more than eight (8) deep.  If so, we
      automatically fail.  This prevents any infinite loop due to recursive
      symbol definitions. }
    inc( Nested ) ; { Nested one more level }
    if( Nested = 8 ) then { Nested too deep }
    begin
        Find_File := '' ; { Not found }
        exit ;
    end ;

    { Now try to find the file }

    while length( X ) > 0 do { Check all items in list }
    begin
        { Get next item in list }

        if( Literal ) then
        begin
            A := 0 ;
        end else
        begin
            A := pos( ';', X ) ;
        end ;
        if( A = 0 ) then
        begin
            A := length( X ) + 1 ;
        end ;
        B := copy( X, 1, A - 1 ) ; {Get list item}
        X := copy( X, A + 1, 254 ) ; {What remains}

        { Check for literal device name }

        if( B[ 1 ] = '_' ) then {Use literal device}
        begin
            assign( F, copy( B, 2, 254 ) ) ;
            Saved := FileMode ;
            FileMode := 0 ;
            Clear_IOResult ; { Clear any pending errors }
            {$I-}
            reset( F ) ;
            {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
            if( Dummy = 0 ) then
            begin
                close( F ) ;
                Find_File := copy( B, 2, 254 ) ;
                exit ;
            end ; {if( IOResult = 0 ) then}
        end ; {if( B[ 1 ] = '_' ) then}

        {Handle case where there is no device/logical specified}

        A := pos( ':', B ) ;
        if( A = 0 ) then {No device/logical}
        begin
            Clear_IOResult ; { Clear any pending errors }
            assign( F, B ) ;
            Saved := FileMode ;
            FileMode := 0 ;
            {$I-}
            reset( F ) ;
            {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
            if( Dummy = 0 ) then
            begin
                close( F ) ;
                Find_File := B ;
                exit ;
            end ; {if( IOResult = 0 ) then}
            continue ;
        end ; {if( A = 0 ) then}

        { Try opening with symbol as a logical... }

        Path := '' ;
        Env := OS^.Get_Symbol( copy( B, 1, A - 1 ) ) ; { Get device }
        Work := copy( B, A + 1, 254 ) ; {Get only file name/extension}
        while length( Env ) > 0 Do {Try all specifications}
        begin
            A := Pos( ';', Env ) ;
            if( A = 0 )
                then
                    A := length( Env ) + 1 ;
            Path := Copy( Env, 1, A - 1 ) ; {Get next specification}
            Env := Copy( Env, A + 1, Length( Env ) - 1 ) ;
            if( length( Path ) > 0 ) then
            begin
                if( Path[ Length( Path ) ] = ':' ) then {Nested logical}
                begin
                    Path := Find_File( Combine( Path, Work ), False ) ;
                    if( Path <> '' ) then {Found it}
                    begin
                        Find_File := Path ;
                        exit ;
                    end ;
                    goto Skip ;
                end ;
                if( Path[ Length( Path ) ] <> '\' ) then
                begin
                    Path := Path + '\' ;
                end ;
            end ;
            Clear_IOResult ; { Clear any pending errors }
            assign( F, Combine( Path, Work ) ) ;
            Saved := FileMode ;
            FileMode := 0 ;
            {$I-}
            reset( F ) ;
            {$I+}
            FileMode := Saved ;
            Dummy := IOResult ;
            if( Dummy = 0 ) Then
            begin
                close( F ) ;
                Find_File := Combine( Path, Work ) ;
                exit ;
            end ;
        end ; {while length( Env ) > 0 do}

Skip:
        { Now try device as a real device name... }

        Clear_IOResult ; { Clear any pending errors }
        assign( F , B ) ;
        Saved := FileMode ;
        FileMode := 0 ;
        {$I-}
        reset( F ) ;
        {$I+}
        FileMode := Saved ;
        Dummy := IOResult ;
        if( Dummy = 0 ) then
        begin
            close( F ) ;
            Find_File := B ;
            exit ;
        end ; { if( IOResult = 0 ) }
    end ;

    Find_File := '' ; { File was not found }
end ; { Find_File }


function Normalize_Hierarchy( Name, Separator, Self_Reference, Parent_Reference,
    Prefix : Ansistring ) : Ansistring ; { UNTESTED }

var Dummy, Dummy1 : integer ;
    S : Ansistring ;

begin
    { Handle initial prefix }
    if( Prefix <> '' ) then
    begin
        Dummy1 := pos( Prefix, Name ) ;
        if( ( copy( Name, Dummy1 + 1, 1 ) <> Separator ) and ( Dummy1 > 0 ) ) then
        begin
            Name := copy( Name, 1, Dummy1 ) + Separator +
                copy( Name, Dummy1 + 1, length( Name ) ) ;
        end ;
    end ;

    { Strip out self-references... }
    if( Self_Reference <> '' ) then
    begin
        S := Separator + Self_Reference + Separator ;
        Dummy1 := pos( S, Name + Separator ) ;
        while( Dummy1 <> 0 ) do
        begin
            Name := copy( Name, 1, Dummy1 - 1 ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            Dummy1 := pos( S, Name + Separator ) ;
        end ;
        S := Self_Reference + Separator ;
        while( copy( Name, 1, length( S ) ) = S ) do
        begin
            Name := copy( Name, length( S ) + 1, length( Name ) ) ;
        end ;
    end ;

    { Process parent references... }
    if( Parent_Reference <> '' ) then
    begin
        S := Separator + Parent_Reference + Separator ;
        Dummy1 := pos( S, Name + Separator ) ;
        while( Dummy1 <> 0 ) do
        begin
            Dummy := Dummy1 - 1 ;
            while( ( Dummy > 0 ) and ( copy( Name, Dummy, length( Separator ) ) <> Separator ) ) do
            begin
                dec( Dummy ) ;
            end ;
            if(
                ( Dummy > 0 )
                and
                ( Prefix <> '' )
                and
                ( Dummy < pos( Prefix, Name ) )
              ) then { Indicates a prefix immediately followed by a parent reference }
            begin
                Name := copy( Name, 1, Dummy1 - 1 ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            end else
            begin
                Name := copy( Name, 1, Dummy ) + copy( Name, Dummy1 + length( S ), length( Name ) ) ;
            end ;
            Dummy1 := pos( S, Name + Separator ) ;
        end ;
    end ;

    Normalize_Hierarchy := Name ;
end ; { Normalize_Hierarchy }


function Explicit( X : Ansistring ) : Ansistring ;

var Dummy, Temp : integer ;
    Work : Ansistring ;

begin
    { Handle missing drive and current directory }
    if( copy( X, 1, 2 ) <> '\\' ) then { Don't process network specification }
    begin
        if( copy( X, 2, 1 ) <> ':' ) then { Add drive, if needed }
        begin
            GetDir( 0, Work ) ;
            X := copy( Work, 1, 2 ) + X ;
        end ;
        if( copy( X, 3, 1 ) <> '\' ) then { Add current path, if needed }
        begin
            GetDir( ord( X[ 1 ] ) - 64, Work ) ;
            if( copy( Work, length( Work ), 1 ) <> '\' ) then
            begin
                Work := Work + '\' ;
            end ;
            X := Work + copy( X, 3, length( X ) ) ;
        end ;
    end ;

    { Remove \. }
    Dummy := pos( '\.\', X + '\' ) ;
    while( Dummy > 0 ) do
    begin
        X := copy( X, 1, Dummy - 1 ) + copy( X, Dummy + 2, length( X ) ) ;
        Dummy := pos( '\.\', X + '\' ) ;
    end ;

    { Now process \.. }
    Dummy := pos( '\..\', X + '\' ) ;
    while( Dummy <> 0 ) do
    begin
        X := copy( X, 1, Dummy - 1 ) + copy( X, Dummy + 3, length( X ) ) ;
        Temp := Dummy - 1 ;
        while( ( Temp > 2 ) and ( X[ Temp ] <> '\' ) ) do
        begin
            dec( Temp ) ; { Backup to beginning of previous directory in string }
        end ;
        if( X[ Temp ] = '\' ) then { Found one }
        begin
            X := copy( X, 1, Temp - 1 ) + copy( X, Dummy, length( X ) ) ;
        end ;
        Dummy := pos( '\..\', X + '\' ) ;
    end ;

    { Return Res }
    Explicit := X ;
end ;


{$IFDEF FPC_OR_DELPHI}
{ TWin32_File_System methods... }

{ Constructors and destructors... }

constructor TWin32_File_System.Init ;

begin
    inherited Init ;

    Handles.Init( 0, 4 ) ;
    Names.Init( 0, 4 ) ;
end ;


destructor TWin32_File_System.Done ;

var Dummy : integer ;

begin
    for Dummy := Handles.Count - 1 downto 0 do
    begin
        Close( Handles.At( Dummy ) ) ;
    end ;
    Handles.Done ;
    Names.Done ;

    inherited Done ;
end ;


{ Instance support... }

procedure TWin32_File_System._Set_Last_Error ;

begin
{$IFDEF MSWINDOWS}
    Set_Error( Create_PSWindows_UE( GetLastError, '' ) ) ;
{$ENDIF}
end ;


{ API... }

procedure TWin32_File_System.Close( Handle : integer ) ;

begin
{$IFDEF MSWINDOWS}
    if( Handle <> 0 ) then
    begin
        CloseHandle( Handle ) ;
        Handle := Handles.IndexOf( Handle ) ;
        if( Handle <> -1 ) then
        begin
            Handles.AtDelete( Handle ) ;
            Names.AtDelete( Handle ) ;
        end ;
        _Set_Last_Error ;
    end else
    begin
        inherited Close( Handle ) ;
    end ;
{$ELSE}
    inherited Close( Handle ) ;
{$ENDIF}
end ;


function TWin32_File_System.Create( const Name : Ansistring ;
    Mode : integer ) : integer ;

var Handle : integer ;

begin {TODO: Handle Mode}
    Handle := FileCreate( Name ) ;
    if( Handle < 0 ) then
    begin
        if( GetLastError = 2 ) then
        begin
            Set_Error( Create_File_System_Exception( File_SystemErr_FileNotFound ) ) ;
        end else
        begin
{$IFDEF MSWINDOWS}
            Set_Error( Create_PSWindows_UE( Handle, '' ) ) ;
{$ELSE}
//TODO:Linux
{$ENDIF}
        end ;
        Create := 0 ;
        exit ;
    end ;

    Set_Error( nil ) ;
    Create := Handle ;
    Handles.Insert( Handle ) ;
    Names.Insert( Name ) ;
end ;


{$IFNDEF MOBILE}
function TWin32_File_System.Debugger : PSDebug_Interface ;

begin
    Debugger := nil ; { Not yet supported }
end ;
{$ENDIF}


function TWin32_File_System.File_Name( Handle : integer ) : Ansistring ;

begin
    Handle := Handles.IndexOf( Handle ) ;
    if( Handle <> -1 ) then
    begin
        File_Name := Names.At( Handle ) ;
    end else
    begin
        File_Name := inherited File_Name( Handle ) ;
    end ;
end ;


procedure TWin32_File_System.Flush( Handle : integer ) ;

begin
{$IFDEF MSWINDOWS}
    FlushFileBuffers( Handle ) ;
{$ENDIF}
    _Set_Last_Error ;
end ;


function TWin32_File_System.Get_Size( Handle : integer ) : TStore_Address ;

begin
{$IFDEF MSWINDOWS}
    Get_Size := GetFileSize( Handle, nil ) ;
{$ELSE}
    Get_Size := 0 ; //
{$ENDIF}
    _Set_Last_Error ;
end ;


function TWin32_File_System.Get_Time( Handle, Typ : integer ) : Longint ;

begin
    Get_Time := FileGetDate( Handle ) ;
    _Set_Last_Error ;
end ;


function TWin32_File_System.Open( const Name : Ansistring ;
    Mode : integer ) : integer ;

var Handle : integer ;
{$IFDEF MSWINDOWS}
    OFStruct : TOFStruct ;
{$ENDIF}

begin {TODO: Handle Mode}
{$IFDEF MSWINDOWS}
    fillchar( OFStruct, sizeof( OFStruct ), 0 ) ;
    OFStruct.cBytes := sizeof( OFStruct ) ;
    if( ( Mode and FSOM_Read_Only ) <> 0 ) then
    begin
        Handle := openfile( PAnsiChar( Name ), OFStruct, OF_READ or OF_SHARE_COMPAT ) ;
    end else
    begin
        Handle := openfile( PAnsiChar( Name ), OFStruct, OF_READWRITE or OF_SHARE_COMPAT ) ;
    end ;
    if( Handle < 0 ) then
    begin
        Handle := GetLastError ;
        if( Handle = 2 ) then
        begin
            Set_Error( Create_File_System_Exception( File_SystemErr_FileNotFound ) ) ;
        end else
        begin
            Set_Error( Create_PSWindows_UE( Handle, '' ) ) ;
        end ;
        Handle := 0 ;
    end else
    begin
        Set_Error( nil ) ;
        Handles.Insert( Handle ) ;
        Names.Insert( Name ) ;
    end ;
    Open := Handle ;
{$ELSE}
    Open := 0 ;
{$ENDIF}
end ; { TWin32_File_System.Open }


procedure TWin32_File_System.Read( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

{$IFDEF MSWINDOWS}
var C : DWord ;
{$ENDIF}

begin
    if( FileSeek( Handle, Position, 0 ) < 0 ) then
    begin
        _Set_Last_Error ;
        exit ;
    end ;
{$IFDEF MSWINDOWS}
    if( not ReadFile( Handle, Buffer, Len, C, nil ) ) then
    begin
        Count := 0 ;
    end else
    begin
        Count := C ;
    end ;
{$ELSE}
    Count := 0 ;
{$ENDIF}
    _Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_Time( Handle : integer ; Typ, Time : Longint ) ;

begin
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
{$IFDEF MSWINDOWS}
    FileSetDate( Handle, Time ) ;
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    _Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_Size( Handle : integer ; New_Size : TStore_Address ) ;

begin
    if( New_Size < Get_Size( Handle ) ) then { Truncating }
    begin
        if( FileSeek( Handle, New_Size, 0 ) < 0 ) then
        begin
            _Set_Last_Error ;
            exit ;
        end ;
{$IFDEF MSWINDOWS}
        SetEndOfFile( Handle ) ;
{$ENDIF}
        _Set_Last_Error ;
    end else
    begin
        inherited Set_Size( Handle, New_Size ) ;
    end ;
end ;


procedure TWin32_File_System.Write( Handle : integer ; Position : TStore_Address ;
    Len : TStore_Address ; var Buffer ; var Count : TStore_Address ) ;

begin
    if( FileSeek( Handle, Position, 0 ) < 0 ) then
    begin
        _Set_Last_Error ;
        exit ;
    end ;
    Count := FileWrite( Handle, Buffer, Len ) ;
    if( Count < 0 ) then
    begin
        Set_Error( Create_PSPascal_UE( 101, '' ) ) ;
        Count := 0 ;
    end else
    begin
        Set_Error( nil ) ;
    end ;
end ;


function TWin32_File_System.Get_File_Attributes( const Name : Ansistring ) : longint ;

var Attr : longint ;

begin
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
    Attr := FileGetAttr( Name ) ;
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    Get_File_Attributes := Convert_DOS_Attributes( Attr ) ;
    _Set_Last_Error ;
end ;


procedure TWin32_File_System.Set_File_Attributes( const Name : Ansistring ;
    Value : longint ) ;

var Attr : longint ;

begin
    Attr := Convert_Attributes( Value ) ;
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS OFF}
{$ENDIF}
{$IFDEF MSWINDOWS}
    FileSetAttr( Name, Attr ) ;
{$ENDIF}
{$IFDEF FPC_OR_DELPHI}
  {$WARNINGS ON}
{$ENDIF}
    _Set_Last_Error ;
end ;


function TWin32_File_System.Facility : longint ;

begin
    Facility := 46 ;
end ;


function TWin32_File_System.Version : longint ;

begin
    Version := 10 ; { Version 1.0 }
end ;

{$ENDIF}


{ TSDefault_File_System_Server methods (abstract class)... }

{ API... }

function TSFile_System_Server.Get_File_System_For_Store( Store : PSStore ;
    Own : boolean ) : PSFile_System ;

begin
    Get_File_System_For_Store := nil ;
end ;


function TSFile_System_Server.Get_File_System( const Nam : Ansistring ) : PSFile_System ;

begin
    Get_File_System := nil ;
end ;


procedure TSFile_System_Server.Initialize( const Nam, ID : Ansistring ;
    Flags : integer ; var UEC : PSUnified_Exception ) ;

begin
    UEC := nil {1} ;
end ;



procedure Terminate ;

begin
    DOS_FS.Done ;
end ;


{$IFDEF UNIT_INIT_AND_FINAL}
initialization
{$ELSE}
begin
{$ENDIF}
    { Initialize data... }
    DOS_FS.Init ;
    { Note: We do not destruct DOS_FS on exit since other units may shut down
      AFTER this one and they may access files opened via DOS_FS, which would
      result in the application blowing up.  Since DOS_FS is static, no memory
      leak should result. }

{$IFDEF UNIT_INIT_AND_FINAL}
finalization
  {$IFDEF CC_Auto_Free_Application_Manager}
      Terminate ;
  {$ENDIF}
{$ENDIF}
end.
