unit icsftp ;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses // Borland...
     Classes, Forms,
     lFTP, lNet ;

type TIdStatus = integer ;

type TFTPStatus =
         procedure(ASender: TObject; const AStatus: TIdStatus; const AStatusText: String) of object ;

type TICS_FTP = class
                    private // Instance data...
                        FTP : TLFTPClient ;

                        _Host : string ;
                        _Password : string ;
                        _Username : string ;
                        _OnStatus : TFTPStatus ;

                    private // Internal utility routines...
                        procedure OnConnect(aSocket: TLSocket);
                        procedure OnControl(aSocket: TLSocket);
                        procedure OnError(const msg: string; aSocket: TLSocket);

                    public // Constructors and destructors...
                        constructor Create( Owner : TComponent ) ;
                        destructor Destroy ; override ;

                    public // API...
                        procedure ChangeDirUp ;
                        procedure ChangeDir( S : string ) ;
                        function Connected : boolean ;
                        function Connect : boolean ;
                        procedure Disconnect ;
                        procedure MakeDir( S : string ) ;
                        procedure Put( L, R : string ) ;

                    public // Properties...
                        property Host : string
                            read _Host
                            write _Host ;
                        property OnStatus : TFTPStatus
                            read _OnStatus
                            write _OnStatus ;
                        property Password : string
                            read _Password
                            write _Password ;
                        property Username : string
                            read _Username
                            write _Username ;
                end ;

implementation


// TICS_FTP methods...

// Constructors and destructors...

constructor TICS_FTP.Create( Owner : TComponent ) ;

begin
    inherited Create ;

    FTP := TLFTPClient.Create( Owner ) ;
    FTP.Timeout := 50 ; // 50 milliseconds
    FTP.OnConnect := OnConnect ; // assign all events
    FTP.OnControl := OnControl ;
    FTP.OnError := OnError ;
end ;


destructor TICS_FTP.Destroy ;

begin
    FTP.Free ;

    inherited Destroy ;
end ;



// Internal utility routines...

procedure TICS_FTP.OnConnect( aSocket : TLSocket ) ;

begin
    if( assigned( OnStatus ) ) then
    begin
        OnStatus( self, 0, 'Connected' ) ;
    end ;
end ;


procedure TICS_FTP.OnControl( aSocket : TLSocket ) ;

var s : string ;

begin
    while( FTP.GetMessage( s ) > 0 ) do // if we got some new message about FTP status, write it
    begin
        if( assigned( OnStatus ) ) then
        begin
            OnStatus( self, 0, S ) ;
        end ;
    end ;
end ;


procedure TICS_FTP.OnError( const msg: string; aSocket: TLSocket ) ;

begin
    if( assigned( OnStatus ) ) then
    begin
        OnStatus( self, 0, Msg ) ;
    end ;
end ;



// API...

procedure TICS_FTP.ChangeDirUp ;

begin
    FTP.cdup ;
    FTP.CallAction ;
end ;


procedure TICS_FTP.ChangeDir( S : string ) ;

begin
    FTP.ChangeDirectory( S ) ;
    FTP.CallAction ;
end ;


function TICS_FTP.Connect : boolean ;

var S : string ;

begin
    if FTP.Connect( _Host, 21 ) then
    begin // if initial connect call worked
        repeat
            FTP.CallAction ; // repeat this until we either get connected or fail
            Application.ProcessMessages ;
        until FTP.Connected ;
    end else
    begin
        Result := False ;
        exit ;
    end ;

    if( not FTP.Authenticate( _Username, _Password ) ) then
    begin // if authentication with server failed
        FTP.GetMessage( s ) ; // if the authentication failed, get reason from server
        //ShowMessage( 'Authentication failed: ' + S ) ;
        Result := False ;
    end ;
end ;


function TICS_FTP.Connected : boolean ;

begin
    Result := FTP.Connected ;
end ;


procedure TICS_FTP.Disconnect ;

begin
    FTP.Disconnect ;
end ;


procedure TICS_FTP.MakeDir( S : string ) ;

begin
    FTP.MakeDirectory( S ) ;
    FTP.CallAction ;
end ;


procedure TICS_FTP.Put( L, R : string ) ;

begin
    FTP.Put( L ) ;
    FTP.CallAction ;
end ;



end.
