{
        Program Name : Standard
        Package Name : Subroutine Library - Layer 3
        Purpose      : Standard Pascal routines
        Institution  : Conroy & Conroy Co.
        Date Written : 28-Nov-88
        Written By   : Alan Conroy
        Version      : 2.1I

        Copyright (c) 2016 by Alan Conroy.  Released to the Public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *********************************************************
        *                                                       *
        *        M O D I F I C A T I O N   H I S T O R Y        *
        *                                                       *
        *********************************************************

             DATE      BY             REASON

        *********************************************************
        *                                                       *
        *           P R O G R A M   P U R P O S E               *
        *                                                       *
        *********************************************************

        64-bit integer list class.            
}

unit Intlist64 ;

interface

uses // C&C Subroutine Library...
     Classes ; // TList

{$I EDEFINES.INC}

{$IFDEF DYNAMIC_OBJECTS}
type TInteger64_List = class( TList )
                           protected { Property handlers }
                               function Get_Index( Index : integer ) : int64 ;
                                   virtual ;
                               procedure Set_Index( Index : integer ;
                                   Value : int64 ) ; virtual ;

                           public { API... }
                               Tag : integer ;

                               function Add( Value : int64 ) : integer ;
                                   virtual ;
                               function Count : integer ;
                               procedure Delete( Index : integer ) ;
                               procedure Remove( Value : int64 ) ;
                               procedure Increment( Index : integer ) ;
                               function IndexOf( Value : int64 ) : integer ;
                               procedure Insert( Index : integer ;
                                   Value : int64 ) ;
                               function Index_With_Max : integer ;
                               procedure Sort( Mask : int64 ) ;

                               property Items[ Index : integer ] : int64
                                   read Get_Index
                                   write Set_Index ; default ;
                       end ;
{$ENDIF}

implementation

{$IFDEF DYNAMIC_OBJECTS}
{ TInteger64_List methods... }

{ Property handlers... }

function TInteger64_List.Get_Index( Index : integer ) : int64 ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    Res._Low := integer( inherited Items[ Index * 2 ] ) ;
    Res._High := integer( inherited Items[ Index * 2 + 1 ] ) ;
    Result := Res._Full ;
end ;


procedure TInteger64_List.Set_Index( Index : integer ; Value : int64 ) ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer) ;
          end ;

begin
    Res._Full := Value ;
    inherited Items[ Index * 2 ] := pointer( Res._Low ) ;
    inherited Items[ Index * 2 + 1 ] := pointer( Res._High ) ;
end ;


function TInteger64_List.Add( Value : int64 ) : integer ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    if( Count = Capacity ) then { Need to expand }
    begin
        if( Count < 1024 ) then
        begin
            Capacity := Capacity * 4 ;
        end else
        begin
            Capacity := Capacity + 1024 ;
        end ;
    end ;
    Res._Full := Value ;
    inherited Add( pointer( Res._Low ) ) ;
    inherited Add( pointer( Res._High ) ) ;
    Result := Count - 1 ;
end ;


function TInteger64_List.Count : integer ;

begin
    Result := inherited Count ;
    if( Result > 0 ) then
    begin
        Result := Result div 2 ;
    end ;
end ;


procedure TInteger64_List.Delete( Index : integer ) ;

begin
    inherited Delete( Index * 2 ) ;
    inherited Delete( Index * 2 ) ;
end ;


procedure TInteger64_List.Remove( Value : int64 ) ;

var Dummy : integer ;

begin
    Dummy := IndexOf( Value ) ;
    if( Dummy <> -1 ) then
    begin
        Delete( Dummy ) ;
    end ;
end ;


procedure TInteger64_List.Increment( Index : integer ) ;

begin
    Items[ Index ] := Items[ Index ] + 1 ;
end ;


function TInteger64_List.IndexOf( Value : int64 ) : integer ;

var Loop : integer ;

begin
    IndexOf := -1 ;
    for Loop := 0 to Count - 1 do
    begin
        if( Items[ Loop ] = Value ) then
        begin
            IndexOf := Loop ;
            exit ;
        end ;
    end ;
end ;


procedure TInteger64_List.Insert( Index : integer ; Value : int64 ) ;

var Res : record
              case integer of
                  0 : ( _Full : int64 ) ;
                  1 : ( _Low, _High : integer ) ;
          end ;

begin
    Res._Full := Value ;
    Index := Index * 2 ; // Convert from our 64-bit index to the 32-bit index
    inherited Insert( Index, pointer( Res._High ) ) ; // Insert high
    inherited Insert( Index, pointer( Res._Low ) ) ; // Insert low before High
end ;


function TInteger64_List.Index_With_Max : integer ;

var Loop : integer ;
    Max : int64 ;

begin
    Result := -1 ;
    Max := -( $7FFFFFFFFFFFFFFF ) ;
    for Loop := 0 to Count - 1 do
    begin
        if( Items[ Loop ] > Max ) then
        begin
            Max := Items[ Loop ] ;
            Result := Loop ;
        end ;
    end ;
end ;


procedure TInteger64_List.Sort( Mask : int64 ) ;

var Loop, Loop1 : integer ;
    I, I1, M, M1 : int64 ;

begin
    for Loop := 0 to Count - 1 do
    begin
        I := Items[ Loop ] ;
        M := I and Mask ;
        for Loop1 := Loop to Count - 1 do
        begin
            I1 := Items[ Loop1 ] ;
            M1 := I1 and Mask ;
            if( M > M1 ) then
            begin
                Items[ Loop ] := I1 ;
                Items[ Loop1 ] := I ;
                I := I1 ;
                M := M1 ;
            end ;
        end ;
    end ;
end ;
{$ENDIF}


end.
