{
        Unit Name    : LongArray
        Package Name : Subroutine Library - Layer 1
        Purpose      : Self-managed dynamic longint array
        Institution  : Conroy & Conroy Co.
        Date Written : 29-Sep-2015
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (c) 2015 by Alan Conroy.  Released to the Public Domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *************************************************************

            DATE        BY          REASON

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

          TDynamic_Longint_Array is a self-managed dynamic longint array class.
}

unit LongArray ;

interface

uses // Subroutine Library - Layer 1...
     HeapBuff, // THeap_Buffer
     _Stores ; // TStore_Address

type TDynamic_Longint_Array = class( THeap_Buffer )
                                  protected // Property handlers...
                                      function Get_Item( Index : longint ) : longint ;
                                          virtual ;
                                      procedure Set_Item( Index, Value : longint ) ;
                                          virtual ;

                                  public // API...
                                      // Add an item to the end of the array. Returns index of added item
                                      function Append( Val : longint ) : longint ;
                                          virtual ;

                                      // Returns highest index in array
                                      function High : longint ; virtual ;

                                  published // Properties...
                                      property Items[ Index : longint ]
                                          read Get_Item
                                          write Set_Item
                                          default ;
                              end ; // TDynamic_Longint_Array

implementation

// TDynamic_Longint_Array methods...

// Property handlers...

function TDynamic_Longint_Array.Get_Item( Index : longint ) : longint ;

begin
    Read( Index * sizeof( Result ), Result, sizeof( Result ) ) ;
end ;


procedure TDynamic_Longint_Array.Set_Item( Index, Value : longint ) ;

begin
    Write( Index * sizeof( Value ), Value, sizeof( Value ) ) ;
end ;


// API...

function TDynamic_Longint_Array.Append( Val : longint ) : longint ;

begin
    Add( Val, sizeof( Val ) ) ;
    Result := High ;
end ;


function TDynamic_Longint_Array.High : longint ;

begin
    Result := ( Size div sizeof( Val ) ) - 1 ;
end ;



end.
